#include <stddef.h>
#include <errormanager.h>
#include <fsmanager.h>
#include <intermanager.h>
#include <keyboard.h>
#include <usermodmanager.h>
#include <userspacemanager.h>
#include <textvideo.h>
#include <lib/exception.h>
#include <lib/stringman.h>

void Start(int argc, char *argv[], char *envp[]) __attribute__ ((section (".init")));

unsigned int (*TV_GetCurrentMode)(void);
int (*TV_GetModeDesc)(unsigned int,TextVideo_ModeDesc*);
int (*TV_GetStatus)(TextVideo_Status*);
int (*TV_SetCursorPos)(unsigned int,unsigned int);
int (*TV_Write)(const char*,int);

void ExceptionHandler(User_Exception_Exception *e) {
    char buf[16];

    TV_Write("Shell caught exception:\nType: ", sizeof("Shell caught exception:\nType: ")-1);
    User_StringMan_BufPrintF(buf, "%x ", e->type);
    TV_Write(buf, (int)User_StringMan_Length(buf));
    TV_Write("Address: ", sizeof("Address: ")-1);
    User_StringMan_BufPrintF(buf, "%x ", e->address);
    TV_Write(buf, (int)User_StringMan_Length(buf));
    TV_Write("\nShell has terminated.\n", sizeof("\nShell has terminated.\n")-1);
    UserModManager_UExit(1);
}

void StartLibs(void) {
    unsigned int i;
    int (*start)(void);

    i = 0;
    while ((start = UserModManager_UGetStartFunctionI(i++))) {
        if ((unsigned int)start != 0xffffffff && start()) {
            TV_Write("Error starting lib.\n", sizeof("Error starting lib.\n")-1);
            UserModManager_UExit(2);
        }
    }
}

#define E_INVALID_COMMAND 0xff

void PrintError(int error) {
    char buf[32];

    TV_Write("ss: ", sizeof("ss: ")-1);
    switch(error) {
        case E_INVALID_COMMAND:
            TV_Write("Command or program not found.\n", sizeof("Command or program not found.\n")-1);
            break;
        case E_TOO_BIG:
            TV_Write("Path too big.\n", sizeof("Path too big.\n")-1);
            break;
        case E_NOT_FOUND:
            TV_Write("Path not found.\n", sizeof("Path not found.\n")-1);
            break;
        case E_NOT_A_DIRECTORY:
            TV_Write("Path is not a dir.\n", sizeof("Path is not a dir.\n")-1);
            break;
        case E_IO_ERROR:
            TV_Write("IO error.\n", sizeof("IO error.\n")-1);
            break;
        case E_NO_MEMORY:
            TV_Write("Not enough memory.\n", sizeof("Not enough memory.\n")-1);
            break;
        case E_LOOP:
            TV_Write("Too many symbolic links encountered while traversing path.\n", sizeof("Too many symbolic links encountered while traversing path.\n")-1);
            break;
        default:
            TV_Write("Undefined error: ", sizeof("Undefined error: ")-1);
            User_StringMan_BufPrintF(buf, "0x%x\n", error);
            TV_Write(buf, (int)User_StringMan_Length(buf));
            break;
    }
}

#define LINE_SIZE 1024
char line[LINE_SIZE];
char *pLine = line;

void ProcessCmd(void) {
    int ret;

    ret = 0;
    if (!User_StringMan_NCompare(line, "cd", 2) && (line[2] == ' ' || line[2] == 0)) {
        /* cd command */
        /* I know this could be better, and will be in the future */
        if (line[2] == ' ') {
            ret = FSManager_UChangeDir(line+3);
            if (ret)
                PrintError(ret);
        }
        TV_Write("ss> ", sizeof("ss> ")-1);
    } else if (!User_StringMan_NCompare(line, "exit", 4) && (line[4] == ' ' || line[4] == 0)) {
        /* exit command */
        UserModManager_UExit(0);
    } else {
        /* it's a program, so execute it */
        char *argv[30];
        char *envp[1];
        char *arg;
        int i;

        /* yes, very simple! */
        envp[0] = NULL;
        i = 1;
        arg = line;
        while ((arg = User_StringMan_FindChar(arg, ' '))) {
            *arg = 0;
            arg++;
            argv[i++] = arg;
        }
        argv[0] = line;
        argv[i] = NULL;
        ret = UserSpaceManager_URun(line, argv, envp);
        if (ret > 0) {
            /* flush keyboard events */
            InterManager_UFlushEventQueue(0xfffffff0);
            /* set focus */
            InterManager_USetFocus(ret);
        } else {
            ret = -ret;
            if (ret == E_NOT_FOUND)
                ret = E_INVALID_COMMAND;
            PrintError(ret);
            TV_Write("ss> ", sizeof("ss> ")-1);
        }
    }
}

void UserModManagerEvent(UserModManager_Event *event) {
    if (event->uEvent.type == UserModManager_EVENT_CHILD) {
        if (event->uEvent.child.code == UserModManager_CHILD_DUMPED) {
            TV_Write("Child dump: ", sizeof("Child dump: ")-1);
            switch (event->uEvent.child.exitCode) {
                case UserModManager_CHILD_DUMPED_MEMREF:
                    TV_Write("Invalid memory reference.\n", sizeof("Invalid memory reference.\n")-1);
                    break;
                case UserModManager_CHILD_DUMPED_BADMEM:
                    TV_Write("Bad memory access.\n", sizeof("Bad memory access.\n")-1);
                    break;
                case UserModManager_CHILD_DUMPED_ILINST:
                    TV_Write("Illegal instruction.\n", sizeof("Illegal instruction.\n")-1);
                    break;
                case UserModManager_CHILD_DUMPED_FPE:
                    TV_Write("Floating point exception.\n", sizeof("Floating point exception.\n")-1);
                    break;
                case UserModManager_CHILD_DUMPED_DIV0:
                    TV_Write("Divide by zero.\n", sizeof("Divide by zero.\n")-1);
                    break;
                case UserModManager_CHILD_DUMPED_INTO:
                    TV_Write("INTO instruction.\n", sizeof("INTO instruction.\n")-1);
                    break;
                case UserModManager_CHILD_DUMPED_BOUND:
                    TV_Write("BOUND instruction.\n", sizeof("BOUND instruction.\n")-1);
                    break;
            }
        } else if (event->uEvent.child.code == UserModManager_CHILD_KILLED)
            TV_Write("Child killed.\n", sizeof("Child killed.\n")-1);
        TV_Write("ss> ", sizeof("ss> ")-1);
    }
}

void KeyboardEvent(Keyboard_Event *event) {
    static int curr;
    TextVideo_Status status;

    if (curr < LINE_SIZE)
        if (!event->type)
            switch (event->vKey) {
                case Keyboard_VK_RETURN:
                case Keyboard_VK_KEYBAD_RETURN:
                    *pLine = 0;
                    curr = 0;
                    pLine = line;
                    TV_Write("\n", sizeof("\n")-1);
                    ProcessCmd();
                    break;
                case Keyboard_VK_BACK:
                    if (!curr)
                        break;
                    pLine--;
                    curr--;
                    *pLine = 0;
                    TV_GetStatus(&status);
                    /* check if we're in the start of the line */
                    if (status.col)
                        status.col--;
                    else {
                        TextVideo_ModeDesc mode;
                        TV_GetModeDesc(TV_GetCurrentMode(), &mode);
                        status.row--;
                        status.col = mode.cols-1;
                    }
                    TV_SetCursorPos(status.row, status.col);
                    TV_Write(" ", sizeof(" ")-1);
                    TV_SetCursorPos(status.row, status.col);
                    break;
                default:
                    if (event->charCode) {
                        *pLine = *(char*)&event->charCode;
                        TV_Write((char*)&event->charCode, 1);
                        pLine++;
                        curr++;
                    }
            }
}

void DispatchEvents(void) {
    InterManager_Event event;

    while (1) {
        int ret;
        ret = InterManager_UWaitEvent(&event, 0);
        if (ret) {
            TV_Write("\nShell: Error getting event!\n", sizeof("\nShell: Error getting event!\n")-1);
            continue;
        }
        if (event.type < 0x10)
            KeyboardEvent((Keyboard_Event*)&event);
        else if (event.type < 0x20)
            UserModManagerEvent((UserModManager_Event*)&event);
    }
}

int GetTVFunction(void) {
    int ret;
    char outBuf[256];
    InterManager_OutputInfo *outInfo;

    /* get TextVideo functions */
    outInfo = (InterManager_OutputInfo*)outBuf;
    ret = InterManager_UGetOutputInfo(outInfo, sizeof(outBuf));
    if (ret) {
        return ret;
    }
    TV_GetCurrentMode = UserModManager_UGetFunction(outInfo->names, outInfo->names+outInfo->interSize, TextVideo__UGETCURRENTMODE, UserModManager_SYS_FUNCTION);
    if (IS_PTR_ERROR(TV_GetCurrentMode))
        return PTR_TO_ERROR(TV_GetCurrentMode);
    TV_GetModeDesc = UserModManager_UGetFunction(outInfo->names, outInfo->names+outInfo->interSize, TextVideo__UGETMODEDESC, UserModManager_SYS_FUNCTION);
    if (IS_PTR_ERROR(TV_GetModeDesc))
        return PTR_TO_ERROR(TV_GetModeDesc);
    TV_GetStatus = UserModManager_UGetFunction(outInfo->names, outInfo->names+outInfo->interSize, TextVideo__UGETSTATUS, UserModManager_SYS_FUNCTION);
    if (IS_PTR_ERROR(TV_GetStatus))
        return PTR_TO_ERROR(TV_GetStatus);
    TV_SetCursorPos = UserModManager_UGetFunction(outInfo->names, outInfo->names+outInfo->interSize, TextVideo__USETCURSORPOS, UserModManager_SYS_FUNCTION);
    if (IS_PTR_ERROR(TV_SetCursorPos))
        return PTR_TO_ERROR(TV_SetCursorPos);
    TV_Write = UserModManager_UGetFunction(outInfo->names, outInfo->names+outInfo->interSize, TextVideo__UWRITE, UserModManager_SYS_FUNCTION);
    if (IS_PTR_ERROR(TV_Write))
        return PTR_TO_ERROR(TV_Write);
    return E_OK;
}

void Start(int argc, char *argv[], char *envp[]) {

    if (GetTVFunction()) {
        UserModManager_UExit(1);
    }
    StartLibs();
    /* set exception handler */
    User_Exception_SetCallBack(ExceptionHandler);
    TV_Write("Simple shell version 0.1.6\n\nValid internal commands: cd, exit\n\n", sizeof("Simple shell version 0.1.6\n\nValid internal commands: cd, exit\n\n")-1);
    /* main loop */
    while (1) {
        TV_Write("ss> ", sizeof("ss> ")-1);
        /* process events */
        DispatchEvents();
    }
}
