/*
* Program to generate executable module files for ModulOS.
* Copyright (C) 2000-2005, Luiz Henrique Shigunov (luizshigunov@users.sourceforge.net)
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2, or (at your option)
* any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330,
* Boston, MA 02111-1307, USA.
*/

#include <elf.h>
#include <algorithm>
#include "elf_file.h"
#include "syntaxe.h"
#include "module_file.h"
#include "common.h"
#include "exceptions_file.h"

#define ALIGN_DATA(data, align) ((((unsigned int)(data)) + (align)-1) & ~((align)-1))

#define BASE 0x400000

#define VERSION_MAJOR 0
#define VERSION_MID 2
#define VERSION_MINOR 8

ELFFile elf;
Syntaxe spec;     /* get information from the spec file */
ModuleFile module;

using namespace std;

static bool cmpRelocOffset(ModuleFile::UsedFuncRelocation *i, ModuleFile::UsedFuncRelocation *j) {
    return i->offset < j->offset;
}

int main(int argc, char *argv[]) {
    unsigned int initS, textS, roDataS, dataS, bssS, initSSize, textSSize, codeSize, roDataSSize, dataSSize, bssSSize;
    unsigned int textStart, roDataStart, dataStart, bssStart;
    char *codeBuffer, *roDataBuffer, *dataBuffer;
    vector<Interface*> *usedFunctions;
    vector<ModuleFile::UsedFuncRelocation*> usedFRelocations;
    ModuleFile::Header header;

    if (argc < 4) {
        cout << "elf2emod version " << VERSION_MAJOR << "." << VERSION_MID << "." << VERSION_MINOR << endl;
        cout << "Copyright (C) 2000-2005, Luiz Henrique Shigunov (luizshigunov@users.sourceforge.net)\nThis program may be freely redistributed under the terms of the GNU GPL\n\n";
        cout << "Usage:\nelf2emod spec_file elf_file output_mod_file\n";
        return 1;
    }
    try {
        elf.open(argv[2]);
        if ((elf.getClass() != ELFCLASS32) || (elf.getDataEnc() != ELFDATA2LSB) || (elf.getMachine() != EM_386)) {
            cout << "ELF file isn't for Intel architecture.\n";
            return 1;
        }
        if (elf.getVersion() != 1) {
            cout << "ELF file version unsupported.\n";
            return 1;
        }
        if (elf.getType() != ET_REL) {
            cout << "ELF file type invalid.\n";
            return 1;
        }
        spec.open(argv[1]);
        module.open(argv[3]);

        // get sections number
        initS = elf.getSectionNumber(".init");
        textS = elf.getSectionNumber(".text");
        roDataS = elf.getSectionNumber(".rodata");
        dataS = elf.getSectionNumber(".data");
        bssS = elf.getSectionNumber(".bss");
        if (!initS || !textS) {
            cout << "Error. No .init or .text section in ELF file.\n";
            return 1;
        }

        // define sections start point
        initSSize = elf.getSectionSize(initS);
        textSSize = elf.getSectionSize(textS);
        roDataSSize = elf.getSectionSize(roDataS);
        dataSSize = elf.getSectionSize(dataS);
        bssSSize = elf.getSectionSize(bssS);
        codeSize = initSSize;
        // align in 4 bytes
        codeSize = ALIGN_DATA(codeSize, 4);
        textStart = codeSize;
        codeSize += textSSize;
        codeBuffer = new char[codeSize];
        if (!codeBuffer) {
            cout << "Error. Not enough memory.\n";
            module.closeFD();
            remove(argv[3]);
            exit(1);
        }
        // align ro data in 4 bytes
        roDataStart = ALIGN_DATA(codeSize, 4);
        dataStart = roDataStart + roDataSSize;
        // align data in 0x400000
        dataStart = ALIGN_DATA(dataStart, 0x400000);
        bssStart = dataStart + dataSSize;
        // align data in 4
        bssStart = ALIGN_DATA(bssStart, 4);

        // read .init code
        elf.read(codeBuffer, elf.getSectionStart(initS), initSSize);
        // do .init section relocation
        {
            unsigned int i = 0;
            Elf32_Rel *reloc = elf.getInitReloc(i);
            while (reloc) {
                Elf32_Sym *symbol = elf.getSymbol(reloc->r_info>>8);
                if (!symbol->st_shndx) {
                    // it's a used function relocation
                    ModuleFile::UsedFuncRelocation *newReloc = new ModuleFile::UsedFuncRelocation;
                    newReloc->offset = reloc->r_offset;
                    newReloc->type = (unsigned char)reloc->r_info == 1 ? ModuleFile::Abs: ModuleFile::Relative;
                    newReloc->name = elf.getString(symbol->st_name);
                    usedFRelocations.push_back(newReloc);
                } else if (symbol->st_shndx == initS) {
                    if ((unsigned char)reloc->r_info == 1)
                        *(unsigned int*)(codeBuffer+reloc->r_offset) += symbol->st_value + BASE;
                    else
                        *(unsigned int*)(codeBuffer+reloc->r_offset) += symbol->st_value + BASE - reloc->r_offset - BASE;
                } else if (symbol->st_shndx == textS) {
                    if ((unsigned char)reloc->r_info == 1)
                        *(unsigned int*)(codeBuffer+reloc->r_offset) += symbol->st_value + BASE + textStart;
                    else
                        *(unsigned int*)(codeBuffer+reloc->r_offset) += symbol->st_value + BASE + textStart - reloc->r_offset - BASE;
                } else if (symbol->st_shndx == roDataS) {
                    *(unsigned int*)(codeBuffer+reloc->r_offset) += symbol->st_value + BASE + roDataStart;
                } else if (symbol->st_shndx == dataS) {
                    *(unsigned int*)(codeBuffer+reloc->r_offset) += symbol->st_value + BASE + dataStart;
                } else if (symbol->st_shndx == bssS) {
                    *(unsigned int*)(codeBuffer+reloc->r_offset) += symbol->st_value + BASE + bssStart;
                } else {
                    cout << "Error. Unsupported relocation section at .init section.\n";
                    module.closeFD();
                    remove(argv[3]);
                    exit(1);
                }
                i++;
                reloc = elf.getInitReloc(i);
            }
        }

        // read .text code
        elf.read(codeBuffer+textStart, elf.getSectionStart(textS), textSSize);
        // do .text section relocation
        {
            unsigned int i = 0;
            Elf32_Rel *reloc = elf.getTextReloc(i);
            while (reloc) {
                Elf32_Sym *symbol = elf.getSymbol(reloc->r_info>>8);
                if (!symbol->st_shndx) {
                    // it's a used function relocation
                    ModuleFile::UsedFuncRelocation *newReloc = new ModuleFile::UsedFuncRelocation;
                    newReloc->offset = reloc->r_offset + textStart;
                    newReloc->type = (unsigned char)reloc->r_info == 1 ? ModuleFile::Abs: ModuleFile::Relative;
                    newReloc->name = elf.getString(symbol->st_name);
                    usedFRelocations.push_back(newReloc);
                } else if (symbol->st_shndx == initS) {
                    if ((unsigned char)reloc->r_info == 1)
                        *(unsigned int*)(codeBuffer+textStart+reloc->r_offset) += symbol->st_value + BASE;
                    else
                        *(unsigned int*)(codeBuffer+textStart+reloc->r_offset) += symbol->st_value - reloc->r_offset - textStart;
                } else if (symbol->st_shndx == textS) {
                    if ((unsigned char)reloc->r_info == 1)
                        *(unsigned int*)(codeBuffer+textStart+reloc->r_offset) += symbol->st_value + BASE + textStart;
                    else
                        *(unsigned int*)(codeBuffer+textStart+reloc->r_offset) += symbol->st_value - reloc->r_offset;
                } else if (symbol->st_shndx == roDataS) {
                    *(unsigned int*)(codeBuffer+textStart+reloc->r_offset) += symbol->st_value + BASE + roDataStart;
                } else if (symbol->st_shndx == dataS) {
                    *(unsigned int*)(codeBuffer+textStart+reloc->r_offset) += symbol->st_value + BASE + dataStart;
                } else if (symbol->st_shndx == bssS) {
                    *(unsigned int*)(codeBuffer+textStart+reloc->r_offset) += symbol->st_value + BASE + bssStart;
                } else {
                    cout << "Error. Unsupported relocation section at .text section.\n";
                    module.closeFD();
                    remove(argv[3]);
                    exit(1);
                }
                i++;
                reloc = elf.getTextReloc(i);
            }
        }

        // now write the code
        module.writeSection(codeBuffer, codeSize, ModuleFile::Code);

        // now ro data
        roDataBuffer = new char[roDataSSize];
        if (!roDataBuffer) {
            cout << "Error. Not enough memory.\n";
            module.closeFD();
            remove(argv[3]);
            exit(1);
        }
        // read .rodata section
        elf.read(roDataBuffer, elf.getSectionStart(roDataS), roDataSSize);
        // do .rodata section relocation
        {
            unsigned int i = 0;
            Elf32_Rel *reloc = elf.getRODataReloc(i);
            while (reloc) {
                Elf32_Sym *symbol = elf.getSymbol(reloc->r_info>>8);
                if (symbol->st_shndx == initS) {
                    *(unsigned int*)(roDataBuffer+reloc->r_offset) += symbol->st_value + BASE;
                } else if (symbol->st_shndx == textS) {
                    *(unsigned int*)(roDataBuffer+reloc->r_offset) += symbol->st_value + BASE + textStart;
                } else if (symbol->st_shndx == roDataS) {
                    *(unsigned int*)(roDataBuffer+reloc->r_offset) += symbol->st_value + BASE + roDataStart;
                } else if (symbol->st_shndx == dataS) {
                    *(unsigned int*)(roDataBuffer+reloc->r_offset) += symbol->st_value + BASE + dataStart;
                } else if (symbol->st_shndx == bssS) {
                    *(unsigned int*)(roDataBuffer+reloc->r_offset) += symbol->st_value + BASE + bssStart;
                } else {
                    cout << "Error. Unsupported relocation section at .rodata section.\n";
                    module.closeFD();
                    remove(argv[3]);
                    exit(1);
                }
                i++;
                reloc = elf.getRODataReloc(i);
            }
        }

        // now write the ro data
        module.writeSection(roDataBuffer, roDataSSize, ModuleFile::ROData);

        // now data
        dataBuffer = new char[dataSSize];
        if (!dataBuffer) {
            cout << "Error. Not enough memory.\n";
            module.closeFD();
            remove(argv[3]);
            exit(1);
        }
        // read .data section
        elf.read(dataBuffer, elf.getSectionStart(dataS), dataSSize);
        // do .data section relocation
        {
            unsigned int i = 0;
            Elf32_Rel *reloc = elf.getDataReloc(i);
            while (reloc) {
                Elf32_Sym *symbol = elf.getSymbol(reloc->r_info>>8);
                if (symbol->st_shndx == initS) {
                    *(unsigned int*)(dataBuffer+reloc->r_offset) += symbol->st_value + BASE;
                } else if (symbol->st_shndx == textS) {
                    *(unsigned int*)(dataBuffer+reloc->r_offset) += symbol->st_value + BASE + textStart;
                } else if (symbol->st_shndx == roDataS) {
                    *(unsigned int*)(dataBuffer+reloc->r_offset) += symbol->st_value + BASE + roDataStart;
                } else if (symbol->st_shndx == dataS) {
                    *(unsigned int*)(dataBuffer+reloc->r_offset) += symbol->st_value + BASE + dataStart;
                } else if (symbol->st_shndx == bssS) {
                    *(unsigned int*)(dataBuffer+reloc->r_offset) += symbol->st_value + BASE + bssStart;
                } else {
                    cout << "Error. Unsupported relocation section at .rodata section.\n";
                    module.closeFD();
                    remove(argv[3]);
                    exit(1);
                }
                i++;
                reloc = elf.getDataReloc(i);
            }
        }

        // now write the data
        module.writeSection(dataBuffer, dataSSize, ModuleFile::Data);

        // and write bss data
        module.writeSection(NULL, bssStart-dataStart-dataSSize+bssSSize, ModuleFile::BSS);

        //define header
        header.comment = spec.getComment();
        header.stackSize = spec.getStackSize();
        module.defHeader(header);

        //write relocation information
        if (!usedFRelocations.empty()) {
            sort(usedFRelocations.begin(), usedFRelocations.end(), cmpRelocOffset);
            //get the list of used functions
            usedFunctions = spec.getUsedFunctions();
            module.writeUsedFunctions(usedFunctions);
            module.writeUsedFuncRelocations(&usedFRelocations);
        }

        // finish module file
        module.close();
        return 0;
    } catch (ESyntaxeError e) {
        cout << argv[1] << ":" << spec.getLine() << ":" << e.what() << "\n";
        module.closeFD();
        remove(argv[3]);
        return 1;
    } catch (logic_error e) {
        cout << e.what() << "\n";
        module.closeFD();
        remove(argv[3]);
        return 1;
    }
}
