/*
* Class to read elf files.
* Copyright (C) 2000-2003, Luiz Henrique Shigunov (luizshigunov@users.sourceforge.net)
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2, or (at your option)
* any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330,
* Boston, MA 02111-1307, USA.
*/

#include <iostream>
#include "elf_file.h"

using namespace std;

ELFFile::ELFFile() {
	status = 0;
}

ELFFile::ELFFile(char *fileName) throw (EFileError) {
	fileIn.open(fileName, ios::in | ios::binary);
	if(fileIn.bad())
		throw EFileError(string("Error opening ELF file."));
}

ELFFile::~ELFFile() {
	if (fileIn.is_open())
		fileIn.close();
	delete [] shstrtab;
}

void ELFFile::open(char *fileName) throw (EFileError) {
	int i, finded;
	streampos tmp;

	if (status)
		return;
	fileIn.open(fileName, ios::in | ios::binary);
	if(fileIn.bad())
		throw EFileError(string("Error opening ELF file."));

	// read header
	fileIn.read((char *)&header,sizeof(Elf32_Ehdr));
	if(fileIn.bad())
		throw EFileError(string("Error reading ELF file."));

	// is it an ELF file?
	if ((header.e_ident[EI_MAG0] != ELFMAG0) || (header.e_ident[EI_MAG1] != ELFMAG1) || (header.e_ident[EI_MAG2] != ELFMAG2) || (header.e_ident[EI_MAG3] != ELFMAG3))
		throw EFileError(string("ELF file invalid."));

	// read section string table
	fileIn.seekg(header.e_shoff+(sizeof(Elf32_Shdr)*header.e_shstrndx), ios::beg);
	fileIn.read((char*)&section, sizeof(Elf32_Shdr));
	shstrtab = new char[section.sh_size];
	fileIn.seekg(section.sh_offset, ios::beg);
	fileIn.read((char*)shstrtab, section.sh_size);
	
	// seek to the first valid section header
	fileIn.seekg(header.e_shoff+sizeof(Elf32_Shdr), ios::beg);

	// search for string table and symbol table
	i = header.e_shnum-1;
	finded = 0;
	while (i--) {
		fileIn.read((char*)&section, sizeof(Elf32_Shdr));
		if ((section.sh_type == SHT_SYMTAB) && (strcmp(&shstrtab[section.sh_name], ".symtab")==0)) {
			nSymb = section.sh_size/sizeof(Elf32_Sym);
			symtab = new Elf32_Sym[nSymb];
			// read symbol table
			tmp = fileIn.tellg();
			fileIn.seekg(section.sh_offset, ios::beg);
			fileIn.read((char*)symtab, section.sh_size);
			fileIn.seekg(tmp);
			finded++;
		} else if ((section.sh_type == SHT_STRTAB) && (strcmp(&shstrtab[section.sh_name], ".strtab")==0)) {
			sizeStr = section.sh_size;
			strtab = new char[section.sh_size];
			// read string table
			tmp = fileIn.tellg();
			fileIn.seekg(section.sh_offset, ios::beg);
			fileIn.read((char*)strtab, section.sh_size);
			fileIn.seekg(tmp);
			finded++;
		}
	}

	if (finded != 2)
		throw EFileError(string("Error. Can't find symbol table or string table of ELF file."));
	status = 1;
}

unsigned char ELFFile::getClass() {
	return header.e_ident[EI_CLASS];
}

unsigned char ELFFile::getDataEnc() {
	return header.e_ident[EI_DATA];
}

unsigned char ELFFile::getVersion() {
	return header.e_ident[EI_VERSION];
}

Elf32_Half ELFFile::getType() {
	return header.e_type;
}

Elf32_Half ELFFile::getMachine() {
	return header.e_machine;
}

Elf32_Sym *ELFFile::getSymbol(unsigned int n) {
	if (!status | (n >= nSymb))
		return NULL;
	return &symtab[n];
}

Elf32_Sym *ELFFile::getSymbol(unsigned int *n, int section) {
	if (!status)
		return NULL;
	for(;*n<nSymb;(*n)++)
		if (symtab[*n].st_shndx == section)
			return &symtab[(*n)++];
	return NULL;
}

Elf32_Sym *ELFFile::getSymbol(const char *name) {
	unsigned int i;

	if (!status)
		return NULL;
	i=1;
	while (i<sizeStr) {
		if (!strcmp(name,strtab+i)) {
			for(unsigned int j=0;j<nSymb;j++)
				if (symtab[j].st_name == i)
					return &symtab[j];
			return NULL;
		}
		i += strlen(strtab+i) + 1;
	}
	return NULL;
}

char *ELFFile::getString(unsigned int pos) {
	if (!status | (pos > sizeStr))
		return NULL;
	return &strtab[pos];
}

int ELFFile::getSectionNumber(const char *name) {
    int i;

    // seek to the first valid section header
    fileIn.seekg(header.e_shoff+sizeof(Elf32_Shdr), ios::beg);
    i = header.e_shnum-1;
    while (i--) {
        fileIn.read((char*)&section, sizeof(Elf32_Shdr));
        if (strcmp(&shstrtab[section.sh_name], name)==0)
            return header.e_shnum-i-1;
    }
    return 0;
}

int ELFFile::getSectionStart(int sNumb) {
    if (sNumb >= header.e_shnum)
        return 0;
    // seek to the section header
    fileIn.seekg(header.e_shoff+(sizeof(Elf32_Shdr)*sNumb), ios::beg);
    fileIn.read((char*)&section, sizeof(Elf32_Shdr));
    return section.sh_offset;
}

int ELFFile::getSectionSize(int sNumb) {
    if (sNumb >= header.e_shnum)
        return 0;
    // seek to the section header
    fileIn.seekg(header.e_shoff+(sizeof(Elf32_Shdr)*sNumb), ios::beg);
    fileIn.read((char*)&section, sizeof(Elf32_Shdr));
    return section.sh_size;
}

void ELFFile::read(char *buffer, int start, int size) throw (EFileError) {
    fileIn.seekg(start, ios::beg);
    fileIn.read(buffer, size);
    if (fileIn.bad() || (fileIn.gcount() < size))
        throw EFileError(string("Error reading ELF file."));
}

Elf32_Rel *ELFFile::getDataReloc(unsigned int n) {
	if (!dataReloc) {
		unsigned int relocN, relocSize;
		// has this file data relocations?
		relocN = getSectionNumber(".rel.data");
		relocSize = getSectionSize(relocN);
		nDataReloc = relocSize/sizeof(Elf32_Rel);
		if (relocN) {
			dataReloc = new Elf32_Rel[nDataReloc];
			if (!dataReloc) {
				cout << "Error. Not enough memory.\n";
				exit(1);
			}
			read((char*)dataReloc, getSectionStart(relocN), relocSize);
		} else
			return NULL;
	}
	if (n >= nDataReloc)
		return NULL;
	return &dataReloc[n];
}

Elf32_Rel *ELFFile::getRODataReloc(unsigned int n) {
	if (!roDataReloc) {
		unsigned int relocN, relocSize;
		// has this file ro data relocations?
		relocN = getSectionNumber(".rel.rodata");
		relocSize = getSectionSize(relocN);
		nRODataReloc = relocSize/sizeof(Elf32_Rel);
		if (relocN) {
			roDataReloc = new Elf32_Rel[nRODataReloc];
			if (!roDataReloc) {
				cout << "Error. Not enough memory.\n";
				exit(1);
			}
			read((char*)roDataReloc, getSectionStart(relocN), relocSize);
		} else
			return NULL;
	}
	if (n >= nRODataReloc)
		return NULL;
	return &roDataReloc[n];
}

Elf32_Rel *ELFFile::getTextReloc(unsigned int n) {
	if (!codeReloc) {
		unsigned int relocN, relocSize;
		// has this file code relocations?
		relocN = getSectionNumber(".rel.text");
		relocSize = getSectionSize(relocN);
		nCodeReloc = relocSize/sizeof(Elf32_Rel);
		if (relocN) {
			codeReloc = new Elf32_Rel[nCodeReloc];
			if (!codeReloc) {
				cout << "Error. Not enough memory.\n";
				exit(1);
			}
			read((char*)codeReloc, getSectionStart(relocN), relocSize);
		} else
			return NULL;
	}
	if (n >= nCodeReloc)
		return NULL;
	return &codeReloc[n];
}

Elf32_Rel *ELFFile::getInitReloc(unsigned int n) {
	if (!initReloc) {
		unsigned int relocN, relocSize;
		// has this file code relocations?
		relocN = getSectionNumber(".rel.init");
		relocSize = getSectionSize(relocN);
		nInitReloc = relocSize/sizeof(Elf32_Rel);
		if (relocN) {
			initReloc = new Elf32_Rel[nInitReloc];
			if (!initReloc) {
				cout << "Error. Not enough memory.\n";
				exit(1);
			}
			read((char*)initReloc, getSectionStart(relocN), relocSize);
		} else
			return NULL;
	}
	if (n >= nInitReloc)
		return NULL;
	return &initReloc[n];
}
