/*
* Program to generate library module files for ModulOS.
* Copyright (C) 2000-2005, Luiz Henrique Shigunov (luizshigunov@users.sourceforge.net)
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2, or (at your option)
* any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330,
* Boston, MA 02111-1307, USA.
*/

#include <elf.h>
#include <algorithm>
#include "elf_file.h"
#include "syntaxe.h"
#include "module_file.h"
#include "common.h"
#include "exceptions_file.h"

#define VERSION_MAJOR 0
#define VERSION_MID 2
#define VERSION_MINOR 7

ELFFile elf;
Syntaxe spec;     /* get information from the spec file */
ModuleFile module;

using namespace std;

static bool cmpRelocOffset(ModuleFile::Relocation *i, ModuleFile::Relocation *j) {
    return i->offset < j->offset;
}

static bool cmpUsedFuncRelocOffset(ModuleFile::UsedFuncRelocation *i, ModuleFile::UsedFuncRelocation *j) {
    return i->offset < j->offset;
}

int main(int argc, char *argv[]) {
    unsigned int dataS, roDataS, codeS, bssS, bssStart;
    char *codeBuffer, *roDataBuffer, *dataBuffer;
    vector<Interface*> *interfaces;
    vector<UsedInterface*> *usedFunctions;
    vector<ModuleFile::Relocation*> roDataRelocs, dataRelocs, codeRelocs;
    vector<ModuleFile::UsedFuncRelocation*> usedFRelocations;
    ModuleFile::Header header;

    if (argc < 4) {
        cout << "elf2lmod version " << VERSION_MAJOR << "." << VERSION_MID << "." << VERSION_MINOR << endl;
        cout << "Copyright (C) 2000-2005, Luiz Henrique Shigunov (luizshigunov@users.sourceforge.net)\nThis program may be freely redistributed under the terms of the GNU GPL\n\n";
        cout << "Usage:\nelf2lmod spec_file elf_file output_mod_file\n";
        return 1;
    }
    try {
        elf.open(argv[2]);
        if ((elf.getClass() != ELFCLASS32) || (elf.getDataEnc() != ELFDATA2LSB) || (elf.getMachine() != EM_386)) {
            cout << "ELF file isn't for Intel architecture.\n";
            return 1;
        }
        if (elf.getVersion() != 1) {
            cout << "ELF file version unsupported.\n";
            return 1;
        }
        if (elf.getType() != ET_REL) {
            cout << "ELF file type invalid.\n";
            return 1;
        }
        spec.open(argv[1]);
        module.open(argv[3]);

        dataBuffer = NULL;
        bssStart = 0;
        dataS = elf.getSectionNumber(".data");
        bssS = elf.getSectionNumber(".bss");
        //if there's data or bss put it in the module file
        if (dataS || bssS) {
            unsigned int dataSSize = elf.getSectionSize(dataS);
            unsigned int dataSStart = elf.getSectionStart(dataS);
            unsigned int bssSSize = elf.getSectionSize(bssS);
            unsigned int bssAlign = 0;
            // align data section
            if (bssS) {
                bssAlign = (4-((dataSSize) & 0x3)) & 0x3;
                bssStart = dataSSize+bssAlign;
                module.writeSection(NULL, bssSSize, ModuleFile::BSS);
            }
            dataBuffer = new char[dataSSize+bssAlign];
            if (!dataBuffer) {
                cout << "Error. Not enough memory.\n";
                module.closeFD();
                remove(argv[3]);
                exit(1);
            }
            if (dataS)
                elf.read(dataBuffer, dataSStart, dataSSize);
            dataSSize += bssAlign;
            module.writeSection(dataBuffer, dataSSize, ModuleFile::Data);
        }

        roDataBuffer = NULL;
        roDataS = elf.getSectionNumber(".rodata");
        //if there's ro data put it in the module file
        if (roDataS) {
            unsigned int roDataSSize = elf.getSectionSize(roDataS);
            unsigned int roDataSStart = elf.getSectionStart(roDataS);
            roDataBuffer = new char[roDataSSize];
            if (!roDataBuffer) {
                cout << "Error. Not enough memory.\n";
                module.closeFD();
                remove(argv[3]);
                exit(1);
            }
            elf.read(roDataBuffer, roDataSStart, roDataSSize);
            module.writeSection(roDataBuffer, roDataSSize, ModuleFile::ROData);
        }

        //put the code in the module file
        codeS = elf.getSectionNumber(".text");
        unsigned int codeSSize = elf.getSectionSize(codeS);
        unsigned int codeSStart = elf.getSectionStart(codeS);
        codeBuffer = new char[codeSSize];
        if (!codeBuffer) {
            cout << "Error. Not enough memory.\n";
            module.closeFD();
            remove(argv[3]);
            exit(1);
        }
        elf.read(codeBuffer, codeSStart, codeSSize);
        module.writeSection(codeBuffer, codeSSize, ModuleFile::Code);

        //define header
        header.comment = spec.getComment();
        header.version = spec.getVersion();
        header.prop = spec.getProp();
        header.startF = 0xffffffff;
        header.shutdownF = 0xffffffff;
        string *startF = spec.getStartF();
        //get the start function position in the code section
        if (startF) {
            Elf32_Sym *symbol = elf.getSymbol(startF->c_str());
            if (!symbol) {
                cout << "Module start function '" << startF->c_str() << "' not found.\n";
                return 1;
            }
            //is it a function?
            if ((symbol->st_info & 0xf) != 2 || symbol->st_shndx != codeS) {
                cout << "Module start function '" << startF->c_str() << "' isn't a function.\n";
                return 1;
            }
            header.startF = symbol->st_value;
        }
        string *shutdownF = spec.getShutdownF();
        //get the shutdown function position in the code section
        if (shutdownF) {
            Elf32_Sym *symbol = elf.getSymbol(shutdownF->c_str());
            if (!symbol) {
                cout << "Module shutdown function '" << shutdownF->c_str() << "' not found.\n";
                return 1;
            }
            //is it a function?
            if ((symbol->st_info & 0xf) != 2 || symbol->st_shndx != codeS) {
                cout << "Module shutdown function '" << shutdownF->c_str() << "' isn't a function.\n";
                return 1;
            }
            header.shutdownF = symbol->st_value;
        }
        module.defHeader(header);

        // get the list of interfaces
        interfaces = spec.getInterfaces();
        // now get functions start point
        {
            vector<Interface*>::iterator currInt = interfaces->begin();
            do {
                vector<Implementation*>::iterator currImp = (*currInt)->implementations.begin();
                do {
                    vector<Function*>::iterator currF = (*currImp)->functions.begin();
                    do {
                        // not implemented function?
                        if ((*currF)->name->compare("NotImplemented")) {
                            Elf32_Sym *symbol = elf.getSymbol((*currF)->name->c_str());
                            if (!symbol) {
                                cout << "Function '" << *(*currF)->name << "' from interface '" << *(*currInt)->name << "' from implementation '" << *(*currImp)->name << "' not found.\n";
                                return 1;
                            }
                            //is it a function?
                            if ((symbol->st_info & 0xf) != 2 || symbol->st_shndx != codeS) {
                                cout << "Function '" << *(*currF)->name << "' from interface '" << *(*currInt)->name << "' from implementation '" << *(*currImp)->name << "' isn't a function.\n";
                                return 1;
                            }
                            (*currF)->start = symbol->st_value;
                        } else {
                            (*currF)->prop |= NOT_IMPLEMENTED_FUNC;
                            (*currF)->start = 0;
                        }
                    } while (++currF != (*currImp)->functions.end());
                } while (++currImp != (*currInt)->implementations.end());
            } while (++currInt != interfaces->end());
            module.writeInterfaces(interfaces);
        }

        // get data relocation information
        {
            unsigned int i = 0;
            Elf32_Rel *reloc = elf.getDataReloc(i);
            while (reloc) {
                Elf32_Sym *symbol = elf.getSymbol(reloc->r_info>>8);
                ModuleFile::Relocation *newReloc = new ModuleFile::Relocation;
                newReloc->offset = reloc->r_offset;
                newReloc->add = *(int*)(dataBuffer+reloc->r_offset) + symbol->st_value;
                if (symbol->st_shndx == dataS || symbol->st_shndx == bssS) {
                    newReloc->what = ModuleFile::Data;
                    if (symbol->st_shndx == bssS)
                        newReloc->add += bssStart;
                } else if (symbol->st_shndx == roDataS)
                    newReloc->what = ModuleFile::ROData;
                else if (symbol->st_shndx == codeS)
                    newReloc->what = ModuleFile::Code;
                else {
                    cout << "Error. Unsupported relocation section.\n";
                    module.closeFD();
                    remove(argv[3]);
                    exit(1);
                }
                dataRelocs.push_back(newReloc);
                i++;
                reloc = elf.getDataReloc(i);
            }
        }

        // get ro data relocation information
        {
            unsigned int i = 0;
            Elf32_Rel *reloc = elf.getRODataReloc(i);
            while (reloc) {
                Elf32_Sym *symbol = elf.getSymbol(reloc->r_info>>8);
                ModuleFile::Relocation *newReloc = new ModuleFile::Relocation;
                newReloc->offset = reloc->r_offset;
                newReloc->add = *(int*)(roDataBuffer+newReloc->offset) + symbol->st_value;
                if (symbol->st_shndx == dataS || symbol->st_shndx == bssS) {
                    newReloc->what = ModuleFile::Data;
                    if (symbol->st_shndx == bssS)
                        newReloc->add += bssStart;
                } else if (symbol->st_shndx == roDataS)
                    newReloc->what = ModuleFile::ROData;
                else if (symbol->st_shndx == codeS)
                    newReloc->what = ModuleFile::Code;
                else {
                    cout << "Error. Unsupported relocation section.\n";
                    module.closeFD();
                    remove(argv[3]);
                    exit(1);
                }
                roDataRelocs.push_back(newReloc);
                i++;
                reloc = elf.getRODataReloc(i);
            }
        }

        // get code relocation information
        {
            unsigned int i = 0;
            Elf32_Rel *reloc = elf.getCodeReloc(i);
            while (reloc) {
                Elf32_Sym *symbol = elf.getSymbol(reloc->r_info>>8);
                if (!symbol->st_shndx) {
                    // it's a used function relocation
                    ModuleFile::UsedFuncRelocation *newReloc = new ModuleFile::UsedFuncRelocation;
                    newReloc->offset = reloc->r_offset;
                    newReloc->type = (unsigned char)reloc->r_info == 1 ? ModuleFile::Abs: ModuleFile::Relative;
                    newReloc->name = elf.getString(symbol->st_name);
                    usedFRelocations.push_back(newReloc);
                } else if (symbol->st_shndx == codeS && (unsigned char)reloc->r_info == 2) {
                    // it's a internal function relocation
                    *(int*)(codeBuffer+reloc->r_offset) = symbol->st_value -4 - reloc->r_offset;
                } else {
                    // it's a data or absolute function relocation
                    ModuleFile::Relocation *newReloc = new ModuleFile::Relocation;
                    newReloc->offset = reloc->r_offset;
                    newReloc->add = *(int*)(codeBuffer+reloc->r_offset) + symbol->st_value;
                    if (symbol->st_shndx == dataS || symbol->st_shndx == bssS) {
                        newReloc->what = ModuleFile::Data;
                        if (symbol->st_shndx == bssS)
                            newReloc->add += bssStart;
                    } else if (symbol->st_shndx == roDataS)
                        newReloc->what = ModuleFile::ROData;
                    else if (symbol->st_shndx == codeS)
                        newReloc->what = ModuleFile::Code;
                    else {
                        cout << "Error. Unsupported relocation section.\n";
                        module.closeFD();
                        remove(argv[3]);
                        exit(1);
                    }
                    codeRelocs.push_back(newReloc);
                }
                i++;
                reloc = elf.getCodeReloc(i);
            }
        }

        //write relocation information
        if (!dataRelocs.empty()) {
            sort(dataRelocs.begin(), dataRelocs.end(), cmpRelocOffset);
            module.writeDataRelocations(&dataRelocs);
        }
        if (!roDataRelocs.empty()) {
            sort(roDataRelocs.begin(), roDataRelocs.end(), cmpRelocOffset);
            module.writeRODataRelocations(&roDataRelocs);
        }
        if (!codeRelocs.empty()) {
            sort(codeRelocs.begin(), codeRelocs.end(), cmpRelocOffset);
            module.writeCodeRelocations(&codeRelocs);
        }
        if (!usedFRelocations.empty()) {
            sort(usedFRelocations.begin(), usedFRelocations.end(), cmpUsedFuncRelocOffset);
            usedFunctions = spec.getUsedFunctions();
            module.writeUsedFunctions(usedFunctions);
            module.writeUsedFuncRelocations(&usedFRelocations);
        }

        // finish module file
        module.close();
        return 0;
    } catch (ESyntaxeError e) {
        cout << argv[1] << ":" << spec.getLine() << ":" << e.what() << "\n";
        module.closeFD();
        remove(argv[3]);
        return 1;
    } catch (logic_error e) {
        cout << e.what() << "\n";
        module.closeFD();
        remove(argv[3]);
        return 1;
    }
}
