/*
* Class to generate library module files for ModulOS.
* Copyright (C) 2000-2003, Luiz Henrique Shigunov (luizshigunov@users.sourceforge.net)
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2, or (at your option)
* any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330,
* Boston, MA 02111-1307, USA.
*/

#include <iostream>
#include "module_file.h"
#include "md5.h"

// it's necessary because the C struct can be bigger
#define MOD_HEADER_SIZE 116

/* macro to align data */
#define ALIGN_DATA(data, align) ((((unsigned int)(data)) + (align)-1) & ~((align)-1))

using namespace std;

ModuleFile::ModuleFile() {
    header.sign[0] = 'L';
    header.sign[1] = 'M';
    header.sign[2] = '0';
    header.sign[3] = '4';
    header.uDataSize = 0;
    nextString = 1;
    roDataRelocs = NULL;
    dataRelocs = NULL;
    codeRelocs = NULL;
    usedFuncRelocations = NULL;
    usedFunctions = NULL;
    roData.size = 0;
    data.size = 0;
}

ModuleFile::~ModuleFile() {
	if (fileOut.is_open())
		fileOut.close();
}
static string *str2Find;

bool findStr(ModuleFile::InternalString *i) {
	return i->str == *str2Find;
}

void ModuleFile::open(const char *fileName) throw (EFileError) {
	fileOut.open(fileName, ios::out | ios::in | ios::binary | ios::trunc);
	if(fileOut.bad())
		throw EFileError(string("Error creating output file."));
}

bool ModuleFile::GetUsedFuncNumber(unsigned int *n, string &name) {
    unsigned int numFunc = 0;

    vector<UsedInterface*>::iterator curr = usedFunctions->begin();
    while (curr != usedFunctions->end()) {
        vector<UsedImplementation*>::iterator currImp = (*curr)->imps.begin();
        while (currImp != (*curr)->imps.end()) {
            vector<UsedFunction*>::iterator currFunc = (*currImp)->functions.begin();
            while (currFunc != (*currImp)->functions.end()) {
                if (*((*currFunc)->name) == name) {
                    *n = numFunc;
                    return true;
                }
                currFunc++;
                numFunc++;
            }
            currImp++;
        }
        curr++;
    }
    return false;
}

void ModuleFile::closeFD() throw (EFileError) {
    fileOut.close();
}

void ModuleFile::close() throw (EFileError) {
	// write the header - it'll be fixed later
	fileOut.write((char *)&header, MOD_HEADER_SIZE);

	// write module interfaces
	{
		header.interfaces = fileOut.tellp();
		header.interfacesSize = 0;
		vector<Interface*>::iterator currI = interfaces->begin();
		do {
			vector<ModuleFile::InternalString*>::iterator i;
			// see if we already have this string
			str2Find = (*currI)->name;
			i = find_if(strings.begin(), strings.end(), findStr);
			if (i == strings.end()) {
				ModuleFile::InternalString *str = new ModuleFile::InternalString;
				str->str = *(*currI)->name;
				str->iStr = nextString;
				nextString += (*currI)->name->size()+1;
				strings.push_back(str);
				i = strings.end();
				i--;
			}
			fileOut.write((char*)&(*i)->iStr, 2);
			vector<Implementation*>::iterator currImp = (*currI)->implementations.begin();
			unsigned int numFunctions = (*currImp)->functions.size();
			unsigned int numImp = (*currI)->implementations.size();
			fileOut.write((char*)&numFunctions, 2);
			fileOut.write((char*)&numImp, 2);
			do {
				// it will be fixed later
				fileOut.write((char*)&numImp, 4);
				// see if we already have this string
				str2Find = (*currImp)->name;
				i = find_if(strings.begin(), strings.end(), findStr);
				if (i == strings.end()) {
					ModuleFile::InternalString *str = new ModuleFile::InternalString;
					str->str = *(*currImp)->name;
					str->iStr = nextString;
					nextString += (*currImp)->name->size()+1;
					strings.push_back(str);
					i = strings.end();
					i--;
				}
				fileOut.write((char*)&(*i)->iStr, 2);
				header.interfacesSize += 6;
			} while (++currImp != (*currI)->implementations.end());
			header.interfacesSize += 6;
		} while (++currI != interfaces->end());
	}

	// write module string section
	{
		// string section too big?
		if (nextString > 0xffff)
			throw EFileError(string("Error. String section bigger then 64KB."));
		header.strings = fileOut.tellp();
		header.stringsSize = nextString;
		vector<ModuleFile::InternalString*>::iterator curr = strings.begin();
		// first string is null
		fileOut << (char)0;
		do {
			fileOut << (*curr)->str;
			fileOut << (char)0;
		} while (++curr != strings.end());
	}

	// write module functions information
	{
		vector<unsigned int> positions;
		vector<Interface*>::iterator currI = interfaces->begin();
		do {
			vector<Implementation*>::iterator currImp = (*currI)->implementations.begin();
			unsigned int numFunctions = (*currImp)->functions.size();
			do {
				if ((*currImp)->functions.size() != numFunctions) {
					string str("Error. Interface '");
					str += *(*currI)->name;
					str += "' has implementations with different number of functions.";
					throw EFileError(str);
				}
				// save where these functions are in the file
				positions.push_back(fileOut.tellp());
				vector<Function*>::iterator currF = (*currImp)->functions.begin();
				do {
					fileOut.write((char*)&(*currF)->start, 4);
					fileOut.write((char*)&(*currF)->prop, 2);
				} while (++currF != (*currImp)->functions.end());
			} while (++currImp != (*currI)->implementations.end());
		} while (++currI != interfaces->end());

		// now fix interfaces section
		fileOut.seekp(header.interfaces, ios::beg);
		vector<unsigned int>::iterator curr = positions.begin();
		currI = interfaces->begin();
		do {
			vector<Implementation*>::iterator currImp = (*currI)->implementations.begin();
			fileOut.seekp(6, ios::cur);
			do {
				fileOut.write((char*)&(*curr), 4);
				fileOut.seekp(2, ios::cur);
				curr++;
			} while (++currImp != (*currI)->implementations.end());
		} while (++currI != interfaces->end());
		fileOut.seekp(0, ios::end);
	}

    // write module used functions
    header.usedFuncsSize = 0;
    if (usedFunctions) {
        unsigned int numFunc;
        numFunc = 0;
        header.usedFuncs = fileOut.tellp();
        vector<UsedInterface*>::iterator curr = usedFunctions->begin();
        while (curr != usedFunctions->end()) {
            unsigned int interN, impN;
            vector<ModuleFile::InternalString*>::iterator i;
            str2Find = (*curr)->name;
            i = find_if(strings.begin(), strings.end(), findStr);
            interN = (*i)->iStr;
            vector<UsedImplementation*>::iterator currImp = (*curr)->imps.begin();
            while (currImp != (*curr)->imps.end()) {
                str2Find = (*currImp)->name;
                impN = 0;
                if (str2Find) {
                    i = find_if(strings.begin(), strings.end(), findStr);
                    impN = (*i)->iStr;
                }
                vector<UsedFunction*>::iterator currFunc = (*currImp)->functions.begin();
                while (currFunc != (*currImp)->functions.end()) {
                    fileOut.write((char*)&interN, 2);
                    fileOut.write((char*)&impN, 2);
                    fileOut.write((char*)&(*currFunc)->number, 3);
                    fileOut.write((char*)&(*curr)->prop, 1);
                    currFunc++;
                    numFunc++;
                }
                currImp++;
            }
            curr++;
        }
        header.usedFuncsSize = numFunc * 8;
    }

    // write module used functions relocation
    header.usedFuncRelocSize = 0;
    if (usedFuncRelocations) {
        header.usedFuncReloc = fileOut.tellp();
        header.usedFuncRelocSize = usedFuncRelocations->size() * 8;
        vector<ModuleFile::UsedFuncRelocation*>::iterator curr = usedFuncRelocations->begin();
        if (!usedFunctions) {
            string str("Error. Relocating inexisting used function '");
            str += (*curr)->name;
            str += "'.";
            throw EFileError(str);
        }
        while (curr != usedFuncRelocations->end()) {
            fileOut.write((char*)&(*curr)->offset, 4);
            unsigned char type = 0;
            if ((*curr)->type == ModuleFile::Abs)
                type = 1;
            fileOut << type;
            unsigned int numFunc;
            if (!GetUsedFuncNumber(&numFunc, (*curr)->name)) {
                string str("Error. Relocating inexisting used function '");
                str += (*curr)->name;
                str += "'.";
                throw EFileError(str);
            }
            fileOut.write((char*)&numFunc, 3);
            curr++;
        }
    }

	// write module code relocation
	header.codeRelocSize = 0;
	if (codeRelocs) {
		header.codeReloc = fileOut.tellp();
		// skip size
		fileOut.seekp(12, ios::cur);
		unsigned int roDataR = 0;
		unsigned int dataR = 0;
		unsigned int codeR = 0;
		// write ro data relocation
		vector<ModuleFile::Relocation*>::iterator curr = codeRelocs->begin();
		while (curr != codeRelocs->end()) {
			if ((*curr)->what == ModuleFile::ROData) {
				fileOut.write((char*)&(*curr)->offset, 4);
				*(int*)(code.buffer+(*curr)->offset) = (*curr)->add;
				roDataR++;
			}
			curr++;
		}
		// write data relocation
		curr = codeRelocs->begin();
		while (curr != codeRelocs->end()) {
			if ((*curr)->what == ModuleFile::Data) {
				fileOut.write((char*)&(*curr)->offset, 4);
				*(int*)(code.buffer+(*curr)->offset) = (*curr)->add;
				dataR++;
			}
			curr++;
		}
		// write code relocation
		curr = codeRelocs->begin();
		while (curr != codeRelocs->end()) {
			if ((*curr)->what == ModuleFile::Code) {
				fileOut.write((char*)&(*curr)->offset, 4);
				*(int*)(code.buffer+(*curr)->offset) = (*curr)->add;
				codeR++;
			}
			curr++;
		}
		//fix size
		roDataR *= 4;
		dataR *= 4;
		codeR *= 4;
		header.codeRelocSize = roDataR+dataR+codeR+12;
		fileOut.seekp(header.codeReloc, ios::beg);
		fileOut.write((char*)&roDataR, 4);
		fileOut.write((char*)&dataR, 4);
		fileOut.write((char*)&codeR, 4);
		fileOut.seekp(0, ios::end);
	}

	// write module ro data relocation
	header.dataRelocSize = 0;
	if (dataRelocs) {
		header.dataReloc = fileOut.tellp();
		// skip size
		fileOut.seekp(12, ios::cur);
		unsigned int roDataR = 0;
		unsigned int dataR = 0;
		unsigned int codeR = 0;
		// write ro data relocation
		vector<ModuleFile::Relocation*>::iterator curr = dataRelocs->begin();
		while (curr != dataRelocs->end()) {
			if ((*curr)->what == ModuleFile::ROData) {
				fileOut.write((char*)&(*curr)->offset, 4);
				*(int*)(data.buffer+(*curr)->offset) = (*curr)->add;
				roDataR++;
			}
			curr++;
		}
		// write data relocation
		curr = dataRelocs->begin();
		while (curr != dataRelocs->end()) {
			if ((*curr)->what == ModuleFile::Data) {
				fileOut.write((char*)&(*curr)->offset, 4);
				*(int*)(data.buffer+(*curr)->offset) = (*curr)->add;
				dataR++;
			}
			curr++;
		}
		// write code relocation
		curr = dataRelocs->begin();
		while (curr != dataRelocs->end()) {
			if ((*curr)->what == ModuleFile::Code) {
				fileOut.write((char*)&(*curr)->offset, 4);
				*(int*)(data.buffer+(*curr)->offset) = (*curr)->add;
				codeR++;
			}
			curr++;
		}
		//fix size
		roDataR *= 4;
		dataR *= 4;
		codeR *= 4;
		header.dataRelocSize = roDataR+dataR+codeR+12;
		fileOut.seekp(header.dataReloc, ios::beg);
		fileOut.write((char*)&roDataR, 4);
		fileOut.write((char*)&dataR, 4);
		fileOut.write((char*)&codeR, 4);
		fileOut.seekp(0, ios::end);
	}

	// write module data relocation
	header.dataRelocSize = 0;
	if (dataRelocs) {
		header.dataReloc = fileOut.tellp();
		// skip size
		fileOut.seekp(12, ios::cur);
		unsigned int roDataR = 0;
		unsigned int dataR = 0;
		unsigned int codeR = 0;
		// write ro data relocation
		vector<ModuleFile::Relocation*>::iterator curr = dataRelocs->begin();
		while (curr != dataRelocs->end()) {
			if ((*curr)->what == ModuleFile::ROData) {
				fileOut.write((char*)&(*curr)->offset, 4);
				*(int*)(data.buffer+(*curr)->offset) = (*curr)->add;
				roDataR++;
			}
			curr++;
		}
		// write data relocation
		curr = dataRelocs->begin();
		while (curr != dataRelocs->end()) {
			if ((*curr)->what == ModuleFile::Data) {
				fileOut.write((char*)&(*curr)->offset, 4);
				*(int*)(data.buffer+(*curr)->offset) = (*curr)->add;
				dataR++;
			}
			curr++;
		}
		// write code relocation
		curr = dataRelocs->begin();
		while (curr != dataRelocs->end()) {
			if ((*curr)->what == ModuleFile::Code) {
				fileOut.write((char*)&(*curr)->offset, 4);
				*(int*)(data.buffer+(*curr)->offset) = (*curr)->add;
				codeR++;
			}
			curr++;
		}
		//fix size
		roDataR *= 4;
		dataR *= 4;
		codeR *= 4;
		header.dataRelocSize = roDataR+dataR+codeR+12;
		fileOut.seekp(header.dataReloc, ios::beg);
		fileOut.write((char*)&roDataR, 4);
		fileOut.write((char*)&dataR, 4);
		fileOut.write((char*)&codeR, 4);
		fileOut.seekp(0, ios::end);
	}

	// write module code
	header.code = fileOut.tellp();
	header.codeSize = code.size;
	fileOut.write(code.buffer, code.size);

	// write module ro data
	header.roDataSize = roData.size;
	if (roData.size) {
		header.roData = fileOut.tellp();
		fileOut.write(roData.buffer, roData.size);
	}

	// write module data
	header.dataSize = data.size;
	if (data.size) {
		header.data = fileOut.tellp();
		fileOut.write(data.buffer, data.size);
	}

	// fix header
	fileOut.seekp(0, ios::beg);
	fileOut.write((char *)&header, MOD_HEADER_SIZE);

    // build md5 message digest
    DoMD5();

	fileOut.flush();
	fileOut.close();
}

void ModuleFile::DoMD5() throw (EFileError) {
    unsigned char buf[4096];
    MD5 md5;
    unsigned char digest[16];

    fileOut.seekg(16, ios::beg);
    while (!fileOut.eof()) {
        fileOut.read((char*)buf, 4096);
        if(fileOut.bad())
            throw EFileError(string("Error reading file."));
        md5.Update(buf, fileOut.gcount());
    }
    md5.Final(digest);
    fileOut.clear(ios::goodbit);
    fileOut.seekp(0, ios::beg);
    fileOut.write((char*)digest, 16);
}

void ModuleFile::defHeader(ModuleFile::Header &header) {
	this->header.version = header.version;
	this->header.prop = header.prop;
    this->header.startF = header.startF;
    this->header.shutdownF = header.shutdownF;
	vector<ModuleFile::InternalString*>::iterator i;
	// see if we already have this string
	this->header.comment = 0;
	str2Find = header.comment;
	if (str2Find) {
		i = find_if(strings.begin(), strings.end(), findStr);
		if (i == strings.end()) {
			ModuleFile::InternalString *str = new ModuleFile::InternalString;
			str->str = *header.comment;
			str->iStr = nextString;
			nextString += header.comment->size()+1;
			strings.push_back(str);
			i = strings.end();
			i--;
		}
		this->header.comment = (*i)->iStr;
	}
}

void ModuleFile::writeRODataRelocations(vector<ModuleFile::Relocation*> *relocations) {
	this->roDataRelocs = relocations;
}

void ModuleFile::writeDataRelocations(vector<ModuleFile::Relocation*> *relocations) {
	this->dataRelocs = relocations;
}

void ModuleFile::writeCodeRelocations(vector<ModuleFile::Relocation*> *relocations) {
	this->codeRelocs = relocations;
}

void ModuleFile::writeInterfaces(vector<Interface*> *interfaces) {
	this->interfaces = interfaces;
}

void ModuleFile::writeSection(char *buffer, unsigned int size, ModuleFile::SectionType type) {
	switch (type) {
		case ModuleFile::ROData:
			roData.buffer = buffer;
			roData.size = size;
			break;
		case ModuleFile::Data:
			data.buffer = buffer;
			data.size = size;
			break;
		case ModuleFile::BSS:
			header.uDataSize = size;
			break;
		case ModuleFile::Code:
			code.buffer = buffer;
			code.size = size;
			break;
	}
}

void ModuleFile::writeUsedFuncRelocations(vector<ModuleFile::UsedFuncRelocation*> *relocations) {
	this->usedFuncRelocations = relocations;
}

void ModuleFile::writeUsedFunctions(vector<UsedInterface*> *usedFunctions) {
    this->usedFunctions = usedFunctions;
    // we do this now so we can put the string section before the used function section
    vector<UsedInterface*>::iterator curr = usedFunctions->begin();
    while (curr != usedFunctions->end()) {
        vector<ModuleFile::InternalString*>::iterator i;
        // see if we already have this string
        str2Find = (*curr)->name;
        i = find_if(strings.begin(), strings.end(), findStr);
        if (i == strings.end()) {
            ModuleFile::InternalString *str = new ModuleFile::InternalString;
            str = new ModuleFile::InternalString;
            str->str = *(*curr)->name;
            str->iStr = nextString;
            nextString += (*curr)->name->size()+1;
            strings.push_back(str);
        }
        vector<UsedImplementation*>::iterator currImp = (*curr)->imps.begin();
        while (currImp != (*curr)->imps.end()) {
            // see if we already have this string
            str2Find = (*currImp)->name;
            if (str2Find) {
                i = find_if(strings.begin(), strings.end(), findStr);
                if (i == strings.end()) {
                    ModuleFile::InternalString *str = new ModuleFile::InternalString;
                    str->str = *(*currImp)->name;
                    str->iStr = nextString;
                    nextString += (*currImp)->name->size()+1;
                    strings.push_back(str);
                }
            }
            currImp++;
        }
        curr++;
    }
}
