/*
* Class to generate system modules files for ModulOS.
* Copyright (C) 2000-2003, Luiz Henrique Shigunov (luizshigunov@users.sourceforge.net)
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2, or (at your option)
* any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330,
* Boston, MA 02111-1307, USA.
*/

#include "module_file.h"
#include "md5.h"

// it's necessary because the C struct can be bigger
#define MOD_HEADER_SIZE 104

/* macro to align data */
#define ALIGN_DATA(data, align) ((((unsigned int)(data)) + (align)-1) & ~((align)-1))

ModuleFile::ModuleFile() {
    header.sign[0] = 'S';
    header.sign[1] = 'M';
    header.sign[2] = '0';
    header.sign[3] = '3';
    header.uDataSize = 0;
    nextString = 1;
    dataRelocs = NULL;
    codeRelocs = NULL;
    usedFuncRelocations = NULL;
    usedFunctions = NULL;
    data.size = 0;
}

ModuleFile::~ModuleFile() {
	if (fileOut.is_open())
		fileOut.close();
}
static string *str2Find;

bool findStr(ModuleFile::InternalString *i) {
	return i->str == *str2Find;
}

void ModuleFile::open(const char *fileName) throw (EFileError) {
	fileOut.open(fileName, ios::out | ios::in | ios::binary | ios::trunc);
	if(fileOut.bad())
		throw EFileError(string("Error creating output file."));
}

void ModuleFile::closeFD() throw (EFileError) {
    fileOut.close();
}

void ModuleFile::close() throw (EFileError) {
    // write the header - it'll be fixed later
    fileOut.write((char*)&header, MOD_HEADER_SIZE);

    // write module interfaces
    {
        header.interfaces = fileOut.tellp();
        header.interfacesSize = 0;
        vector<Interface*>::iterator currI = interfaces->begin();
        do {
            vector<ModuleFile::InternalString*>::iterator i;
            // see if we already have this string
            str2Find = (*currI)->name;
            i = find_if(strings.begin(), strings.end(), findStr);
            if (i == strings.end()) {
                ModuleFile::InternalString *str = new ModuleFile::InternalString;
                str->str = *(*currI)->name;
                str->iStr = nextString;
                nextString += (*currI)->name->size()+1;
                strings.push_back(str);
                i = strings.end();
                i--;
            }
            fileOut.write((char*)&(*i)->iStr, 2);
            vector<Implementation*>::iterator currImp = (*currI)->implementations.begin();
            unsigned int numFunctions = (*currImp)->functions.size();
            unsigned int numImp = (*currI)->implementations.size();
            fileOut.write((char*)&numFunctions, 2);
            fileOut.write((char*)&numImp, 2);
            do {
                // it will be fixed later
                fileOut.write((char*)&numImp, 4);
                // see if we already have this string
                str2Find = (*currImp)->name;
                i = find_if(strings.begin(), strings.end(), findStr);
                if (i == strings.end()) {
                    ModuleFile::InternalString *str = new ModuleFile::InternalString;
                    str->str = *(*currImp)->name;
                    str->iStr = nextString;
                    nextString += (*currImp)->name->size()+1;
                    strings.push_back(str);
                    i = strings.end();
                    i--;
                }
                fileOut.write((char*)&(*i)->iStr, 2);
                header.interfacesSize += 6;
            } while (++currImp != (*currI)->implementations.end());
            header.interfacesSize += 6;
        } while (++currI != interfaces->end());
    }

    // write module string section
    {
        // string section too big?
        if (nextString > 0xffff)
            throw EFileError(string("Error. String section bigger then 64KB."));
        header.strings = fileOut.tellp();
        header.stringsSize = nextString;
        vector<ModuleFile::InternalString*>::iterator curr = strings.begin();
        // first string is null
        fileOut << (char)0;
        do {
            fileOut << (*curr)->str;
            fileOut << (char)0;
        } while (++curr != strings.end());
    }

    // write module used functions
    header.usedFuncsSize = 0;
    if (usedFunctions) {
        header.usedFuncs = fileOut.tellp();
        header.usedFuncsSize = usedFunctions->size() * 6;
        vector<UsedFunction*>::iterator curr = usedFunctions->begin();
        while (curr != usedFunctions->end()) {
            vector<ModuleFile::InternalString*>::iterator i;
            str2Find = (*curr)->interface;
            i = find_if(strings.begin(), strings.end(), findStr);
            fileOut.write((char*)&(*i)->iStr, 2);
            str2Find = (*curr)->implementation;
            unsigned int tmp = 0;
            if (str2Find) {
                i = find_if(strings.begin(), strings.end(), findStr);
                tmp = (*i)->iStr;
            }
            fileOut.write((char*)&tmp, 2);
            fileOut.write((char*)&(*curr)->number, 2);
            curr++;
        }
    }

    // write module functions information
    {
        vector<unsigned int> positions;
        vector<Interface*>::iterator currI = interfaces->begin();
        do {
            vector<Implementation*>::iterator currImp = (*currI)->implementations.begin();
            unsigned int numFunctions = (*currImp)->functions.size();
            do {
                if ((*currImp)->functions.size() != numFunctions) {
                    string str("Error. Interface '");
                    str += *(*currI)->name;
                    str += "' has implementations with different number of functions.";
                    throw EFileError(str);
                }
                // save where these functions are in the file
                positions.push_back(fileOut.tellp());
                vector<Function*>::iterator currF = (*currImp)->functions.begin();
                do {
                    fileOut.write((char*)&(*currF)->start, 4);
                    fileOut.write((char*)&(*currF)->prop, 2);
                } while (++currF != (*currImp)->functions.end());
            } while (++currImp != (*currI)->implementations.end());
        } while (++currI != interfaces->end());

        // now fix interfaces section
        fileOut.seekp(header.interfaces, ios::beg);
        vector<unsigned int>::iterator curr = positions.begin();
        currI = interfaces->begin();
        do {
            vector<Implementation*>::iterator currImp = (*currI)->implementations.begin();
            fileOut.seekp(6, ios::cur);
            do {
                fileOut.write((char*)&(*curr), 4);
                fileOut.seekp(2, ios::cur);
                curr++;
            } while (++currImp != (*currI)->implementations.end());
        } while (++currI != interfaces->end());
        fileOut.seekp(0, ios::end);
    }

    // write module used functions relocation
    header.usedFuncRelocSize = 0;
    if (usedFuncRelocations) {
        header.usedFuncReloc = fileOut.tellp();
        header.usedFuncRelocSize = usedFuncRelocations->size() * 8;
        vector<ModuleFile::UsedFuncRelocation*>::iterator curr = usedFuncRelocations->begin();
        if (!usedFunctions) {
            string str("Error. Relocating inexisting used function '");
            str += (*curr)->name;
            str += "'.";
            throw EFileError(str);
        }
        while (curr != usedFuncRelocations->end()) {
            fileOut.write((char*)&(*curr)->offset, 4);
            unsigned char type = 0;
            if ((*curr)->type == ModuleFile::Abs)
                type = 1;
            fileOut << type;
            unsigned int j =  usedFunctions->size();
            while (j--)
                if (*((*usedFunctions)[j]->name) == (*curr)->name)
                    break;
            if (j==0xffffffff) {
                string str("Error. Relocating inexisting used function '");
                str += (*curr)->name;
                str += "'.";
                throw EFileError(str);
            }
            fileOut.write((char*)&j, 3);
            curr++;
        }
    }

    // write module data relocation
    header.dataRelocSize = 0;
    if (dataRelocs) {
        header.dataReloc = fileOut.tellp();
        // skip size
        fileOut.seekp(8, ios::cur);
        unsigned int dataR = 0;
        unsigned int codeR = 0;
        // write data relocation
        vector<ModuleFile::Relocation*>::iterator curr = dataRelocs->begin();
        while (curr != dataRelocs->end()) {
            if ((*curr)->what == ModuleFile::Data) {
                fileOut.write((char*)&(*curr)->offset, 4);
                *(int*)(data.buffer+(*curr)->offset) = (*curr)->add;
                dataR++;
            }
            curr++;
        }
        // write code relocation
        curr = dataRelocs->begin();
        while (curr != dataRelocs->end()) {
            if ((*curr)->what == ModuleFile::Code) {
                fileOut.write((char*)&(*curr)->offset, 4);
                *(int*)(data.buffer+(*curr)->offset) = (*curr)->add;
                codeR++;
            }
            curr++;
        }
        //fix size
        dataR *= 4;
        codeR *= 4;
        header.dataRelocSize = dataR+codeR+8;
        fileOut.seekp(header.dataReloc, ios::beg);
        fileOut.write((char*)&dataR, 4);
        fileOut.write((char*)&codeR, 4);
        fileOut.seekp(0, ios::end);
    }

    // write module code relocation
    header.codeRelocSize = 0;
    if (codeRelocs) {
        header.codeReloc = fileOut.tellp();
        // skip size
        fileOut.seekp(8, ios::cur);
        unsigned int dataR = 0;
        unsigned int codeR = 0;
        // write data relocation
        vector<ModuleFile::Relocation*>::iterator curr = codeRelocs->begin();
        while (curr != codeRelocs->end()) {
            if ((*curr)->what == ModuleFile::Data) {
                fileOut.write((char*)&(*curr)->offset, 4);
                *(int*)(code.buffer+(*curr)->offset) = (*curr)->add;
                dataR++;
            }
            curr++;
        }
        // write code relocation
        curr = codeRelocs->begin();
        while (curr != codeRelocs->end()) {
            if ((*curr)->what == ModuleFile::Code) {
                fileOut.write((char*)&(*curr)->offset, 4);
                *(int*)(code.buffer+(*curr)->offset) = (*curr)->add;
                codeR++;
            }
            curr++;
        }
        //fix size
        dataR *= 4;
        codeR *= 4;
        header.codeRelocSize = dataR+codeR+8;
        fileOut.seekp(header.codeReloc, ios::beg);
        fileOut.write((char*)&dataR, 4);
        fileOut.write((char*)&codeR, 4);
        fileOut.seekp(0, ios::end);
    }

    // write module code
    header.code = fileOut.tellp();
    header.codeSize = code.size;
    fileOut.write(code.buffer, code.size);

    // write module data
    header.dataSize = data.size;
    if (data.size) {
        header.data = fileOut.tellp();
        fileOut.write(data.buffer, data.size);
    }

    // fix header
    fileOut.seekp(0, ios::beg);
    fileOut.write((char *)&header, MOD_HEADER_SIZE);

    // build md5 message digest
    DoMD5();

    fileOut.flush();
    fileOut.close();
}

void ModuleFile::DoMD5() throw (EFileError) {
    unsigned char buf[4096];
    MD5 md5;
    unsigned char digest[16];

    fileOut.seekg(16, ios::beg);
    while (!fileOut.eof()) {
        fileOut.read((char*)buf, 4096);
        if(fileOut.bad())
            throw EFileError(string("Error reading file."));
        md5.Update(buf, fileOut.gcount());
    }
    md5.Final(digest);
    fileOut.clear(ios::goodbit);
    fileOut.seekp(0, ios::beg);
    fileOut.write((char*)digest, 16);
}

void ModuleFile::defHeader(ModuleFile::Header &header) {
    this->header.version = header.version;
    this->header.prop = header.prop;
    this->header.phase0StartF = header.phase0StartF;
    this->header.phase1StartF = header.phase1StartF;
    this->header.shutdownF = header.shutdownF;
    vector<ModuleFile::InternalString*>::iterator i;
    // see if we already have this string
    this->header.comment = 0;
    str2Find = header.comment;
    if (str2Find) {
        i = find_if(strings.begin(), strings.end(), findStr);
        if (i == strings.end()) {
            ModuleFile::InternalString *str = new ModuleFile::InternalString;
            str->str = *header.comment;
            str->iStr = nextString;
            nextString += header.comment->size()+1;
            strings.push_back(str);
            i = strings.end();
            i--;
        }
        this->header.comment = (*i)->iStr;
    }
}

void ModuleFile::writeDataRelocations(vector<ModuleFile::Relocation*> *relocations) {
	this->dataRelocs = relocations;
}

void ModuleFile::writeCodeRelocations(vector<ModuleFile::Relocation*> *relocations) {
	this->codeRelocs = relocations;
}

void ModuleFile::writeInterfaces(vector<Interface*> *interfaces) {
	this->interfaces = interfaces;
}

void ModuleFile::writeSection(char *buffer, unsigned int size, ModuleFile::SectionType type) {
	switch (type) {
		case ModuleFile::Data:
			data.buffer = buffer;
			data.size = size;
			break;
		case ModuleFile::BSS:
			header.uDataSize = size;
			break;
		case ModuleFile::Code:
			code.buffer = buffer;
			code.size = size;
			break;
	}
}

void ModuleFile::writeUsedFuncRelocations(vector<ModuleFile::UsedFuncRelocation*> *relocations) {
	this->usedFuncRelocations = relocations;
}

void ModuleFile::writeUsedFunctions(vector<UsedFunction*> *usedFunctions) {
	this->usedFunctions = usedFunctions;
	// we do this now so we can put the string section before the used function section
	vector<UsedFunction*>::iterator curr = usedFunctions->begin();
	while (curr != usedFunctions->end()) {
		vector<ModuleFile::InternalString*>::iterator i;
		// see if we already have this string
		str2Find = (*curr)->interface;
		i = find_if(strings.begin(), strings.end(), findStr);
		if (i == strings.end()) {
			ModuleFile::InternalString *str = new ModuleFile::InternalString;
			str = new ModuleFile::InternalString;
			str->str = *(*curr)->interface;
			str->iStr = nextString;
			nextString += (*curr)->interface->size()+1;
			strings.push_back(str);
		}
		// see if we already have this string
		str2Find = (*curr)->implementation;
		if (str2Find) {
			i = find_if(strings.begin(), strings.end(), findStr);
			if (i == strings.end()) {
				ModuleFile::InternalString *str = new ModuleFile::InternalString;
				str->str = *(*curr)->implementation;
				str->iStr = nextString;
				nextString += (*curr)->implementation->size()+1;
				strings.push_back(str);
			}
		}
		curr++;
	}
}
