/*
* Program to make ext2 filesystem bootable for ModulOS.
* Copyright (C) 2000-2002, Luiz Henrique Shigunov (luizshigunov@users.sourceforge.net)
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2, or (at your option)
* any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330,
* Boston, MA 02111-1307, USA.
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <sys/stat.h>
#include <linux/ext2_fs.h>
#include <ext2fs/ext2fs.h>

#define VERSION_MAJOR 0
#define VERSION_MID 3
#define VERSION_MINOR 0

char boot_sec[] =
#include "bootsec.h"
;

char stage2[] =
#include "stage2.h"
;

unsigned short *listEntry = (unsigned short*)(boot_sec+510);
blk_t lastBlock = -1;
unsigned int numBlocksBoot;

static int process_blocks(ext2_filsys fs, blk_t *block, int blk_count, void *pri) {
    if (blk_count >= 0) {
        /* exceeded block number support? */
        if (*block > 0xffff) {
            printf("Error. Block number of bootloader too big.\n");
            return BLOCK_ABORT;
        }
        /* new entry? */
        if (lastBlock+1 != *block) {
            listEntry -= 2;
            if (*(listEntry+1)) {
                printf("Error. Too many sectors entries to read.\n");
                return BLOCK_ABORT;
            }
            *listEntry = *block * (fs->blocksize / 512);
            *(listEntry+1) = fs->blocksize / 512;
            lastBlock = *block;
        } else {
            /* just increment sector count */
            *(listEntry+1) += fs->blocksize / 512;
            lastBlock = *block;
        }
        numBlocksBoot++;
    }
    return 0;
}

int main(int argc, char *argv[]) {
    ext2_filsys ext2fs;
    errcode_t error;
    ext2_file_t bootFile;
    FILE *file;
    struct ext2_inode boot_inode;
    unsigned int written;

    if (argc < 2) {
        printf("mkext2boot version %d.%d.%d\n", VERSION_MAJOR, VERSION_MID, VERSION_MINOR);
        printf("Copyright (C) 2000-2002, Luiz Henrique Shigunov (luizshigunov@users.sourceforge.net)\nThis program may be freely redistributed under the terms of the GNU GPL\n\n");
        printf("Usage:\nmkext2boot file\n\nWhere file is a file or device with an ext2 file system.\n");
        return 1;
    }

    /* open ext2 filesystem */
    error = ext2fs_open(argv[1], EXT2_FLAG_RW, 0, 0, unix_io_manager, &ext2fs);
    if (error) {
        com_err(argv[1], error, "while opening ext2fs.");
        return 1;
    }

    /* open bootloader inode */
    error = ext2fs_file_open(ext2fs, EXT2_BOOT_LOADER_INO, EXT2_FILE_WRITE, &bootFile);
    if (error) {
        com_err(argv[1], error, "while opening bootloader.");
        return 1;
    }

    /* if we already have a bootloader, set it to 0 */
    if (ext2fs_file_get_size(bootFile)) {
        error = ext2fs_file_set_size(bootFile, 0);
        if (error) {
            com_err(argv[1], error, "while setting bootloader size to 0.");
            return 1;
        }
    }

    /* write stage2 of bootloader */
    error = ext2fs_file_write(bootFile, stage2, sizeof(stage2), &written);
    if (error) {
        com_err(argv[1], error, "while writing bootloader.");
        return 1;
    }

    ext2fs_file_close(bootFile);

    error = ext2fs_block_iterate(ext2fs, EXT2_BOOT_LOADER_INO, BLOCK_FLAG_DATA_ONLY, NULL, process_blocks, NULL);
    if (error) {
        com_err(argv[1], error, "while iterating on bootloader blocks.");
        return 1;
    }

    /* setup EXT2_BOOT_LOADER_INO inode */
    error = ext2fs_read_inode(ext2fs, EXT2_BOOT_LOADER_INO, &boot_inode);
    if (error) {
        com_err(argv[1], error, "while reading boot loader inode.");
        return 1;
    }

    boot_inode.i_mode = S_IFREG;
    boot_inode.i_size = numBlocksBoot * ext2fs->blocksize;

    error = ext2fs_write_inode(ext2fs, EXT2_BOOT_LOADER_INO, &boot_inode);
    if (error) {
        com_err(argv[1], error, "while writing boot loader inode.");
        return 1;
    }

    ext2fs_close(ext2fs);

    /* let's write boot sector */
    file = fopen(argv[1], "r+");
    if (!file) {
        printf("Error. Cannot open %s to write boot sector.\n", argv[1]);
        return 1;
    }

    if (fwrite(boot_sec, 1, 512, file) != 512) {
        printf("Error. Cannot write boot sector.\n");
        return 1;
    }
    return 0;
}

