/*
*****************************************
* This file implements the BlockDev interface for floppy disk controller
*
* Version: 0.3.0
* Author: Luiz Henrique Shigunov
* E-mail: luizshigunov@users.sourceforge.net
* WEB: http://www.sourceforge.net/projects/modulos
* Last change: 08/06/2005
*
* See the interface specification to know what each function does.
*
* Copyright (C) 2000-2005, Luiz Henrique Shigunov
*
* This module is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2.1 of the License, or (at your option) any later version.
*
* This module is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this module; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*****************************************
*/

#include <stddef.h>
#include <blockdev.h>
#include <debug.h>
#include <dmacontroller_i386.h>
#include <errormanager.h>
#include <intmanager_i386.h>
#include <ioportmanager.h>
#include <memman.h>
#include <sysmodmanager.h>
#include <taskmanager.h>

typedef struct {
    char sectorSize;        /* 0=128;1=256;... */
    char trackSize;         /* sector per track */
    char nCylinders;        /* num of cylinders */
    char lastCyl;           /* current cylinder */
    char transRate;
    char gap3;
} Drive;

SysModManager_Module *ID;
Drive drives = {2, 18, 80, 81, 0, 0x1b};
TaskManager_Semaphore mutex = {1, NULL};
TaskManager_Semaphore intMutex;
char *dmaBuffer;
char DOR = 0x0c;
char result[9];
char commands[10];
TaskManager_Thread *offThread;
TaskManager_Task *offTask;       /* task to switch off drive motor */

void Reset(void);
int Rw(BlockDev_IO*,int);
int Seek(unsigned int,unsigned int);
int SendCommand(void);
int SenseInterrupt(void);

int GetResult(unsigned int n) {
    for (; n>0; n--) {
        unsigned char msr;
        unsigned int try;
        for (try = 0; try < 10000; try++) {
            __asm__ __volatile__ ("inb %%dx, %%al" : "=a" (msr) : "d" (0x3f4));
            if ((msr & 0xc0) == 0xc0) {
                unsigned char resp;
                __asm__ __volatile__ ("inb %%dx, %%al" : "=a" (resp) : "d" (0x3f5));
                result[n-1] = resp;
                goto end;
            }
        }
        DEBUG_MSG_PRE((int)msr);
        /* controller wants bytes */
        if (!(msr & 0x40))
            return E_IO_ERROR;
        if ((msr & 0xc0) != 0xc0)
            return E_TIMED_OUT;
end:
        /* make gcc happy */
        ;
    }
    return E_OK;
}

int GetUnitDesc(unsigned int unit, BlockDev_UnitDesc *unitDesc, unsigned int size, int prop) {
    unsigned int unitSize;

    ASSERT_RET(unitDesc, E_BUG);

    if (unit)
        return E_NO_SUCH_UNIT;

    unitSize = sizeof(BlockDev_UnitDesc);
    if (!(prop & BlockDev_NO_NAME))
        unitSize += sizeof("Floppy 0");
    if (size < unitSize)
        return E_BUFFER_TOO_SMALL;

    unitDesc->unit = 0;
    unitDesc->type = BlockDev_TYPE_FLOPPY;
    unitDesc->flags = BlockDev_F_REMOVABLE;
    unitDesc->blockSize = 128 << drives.sectorSize;
    unitDesc->nBlocks = 0;
    if (!(prop & BlockDev_NO_NAME)) {
        unitDesc->unitName = ((char*)&unitDesc->nBlocks)+sizeof(unitDesc->nBlocks);
        MemMan_Copy(unitDesc->unitName, "Floppy 0", sizeof("Floppy 0"));
    } else
        unitDesc->unitName = NULL;
    return E_OK;
}

void MotorOff(void) {
    while (1) {
        DOR &= 0x0f;
        __asm__ __volatile__ ("outb %%al, %%dx" :: "a" (DOR), "d" (0x3f2));
        TaskManager_Suspend(NULL);
    }
}

void MotorOn(unsigned int unit) {
    unit = 0x10 << unit;
    /* motor already on? */
    if (DOR & unit)
        /* if so, cancel task to switch off motor */
        TaskManager_CancelWait(offThread, TaskManager_WAIT_SUS);
    else {
        DOR |= unit;
        __asm__ __volatile__ ("outb %%al, %%dx" :: "a" (DOR), "d" (0x3f2));
        /* wait a little to stabilize */
        TaskManager_Wait(NULL, 500);
    }
    return;
}

int Read(BlockDev_IO *request) {

    ASSERT_RET(request, E_BUG);
    ASSERT_RET(request->n, E_BUG);
    ASSERT_RET(request->buffer, E_BUG);

    if (request->unit)
        return E_NO_SUCH_UNIT;
    return Rw(request, 0x44);
}

int Recalibrate(unsigned int unit) {
    int try, error;

    try = 0;
tryAgain:
    try += 1;
    commands[0] = 2;
    commands[1] = 0x7;
    commands[2] = unit;
    intMutex.count = 0;
    intMutex.data = NULL;
    error = SendCommand();
    if (error)
        goto try_reset;
    /* wait int */
    error = TaskManager_P(&intMutex, 0, 3000);
    if (error) {
        DEBUG_MSG_PRE(error);
        goto try_reset;
    }
    error = SenseInterrupt();
    if (!error) {
        drives.lastCyl = 0;
        return error;
    }
    DEBUG_MSG_PRE(error);
    if (try < 2)
        goto tryAgain;
    return error;

try_reset:
    if (try < 2) {
        Reset();
        MotorOn(unit);
        goto tryAgain;
    }
    return error;
}

void Reset(void) {
    DOR = 0;
    __asm__ __volatile__ ("outb %%al, %%dx" :: "a" (DOR), "d" (0x3f2));
    DOR = 0x0c;
    __asm__ __volatile__ ("outb %%al, %%dx" :: "a" (DOR), "d" (0x3f2));
    commands[0] = 3;
    commands[1] = 0x3;
    commands[2] = 0xdf;
    commands[3] = 0x2;
    SendCommand();
    return;
}

int Rw(BlockDev_IO *request, int prop) {
    unsigned int offset, toDo, block;
    int error, ret;

    /* this thread can wait for a long time, so set it interruptible */
    ret = TaskManager_P(&mutex, TaskManager_THREAD_INT, 0);
    if (ret)
        return ret;
    MotorOn(request->unit);
    error = 0;
    offset = 0;
    toDo = request->n;
    block = request->block;
    while(toDo > 0) {
        char com, sector, head, cylinder, try;
        int nSectors, maxSectors;

        sector = (block % drives.trackSize) + 1;
        head = (block / drives.trackSize) % 2;
        cylinder = block / (drives.trackSize << 1);
        /* can we process more than one sector? */
        nSectors = toDo;
        maxSectors = drives.trackSize - sector + 1;
        if (nSectors > maxSectors)
            nSectors = maxSectors;
        com = 0xe6;
        /* write? */
        if (prop == 0x48) {
            com = 0xc5;
            MemMan_Copy(dmaBuffer, request->buffer+offset, (nSectors << 7) << drives.sectorSize);
        }
        if (drives.lastCyl != cylinder) {
            error = Seek(request->unit, cylinder);
            if (error)
                goto out;
        }
        /* try to send command 2 times */
        try = 0;
tryAgain:
        try += 1;
        __asm__ __volatile__ ("outb %%al, %%dx" :: "a" (drives.transRate), "d" (0x3f7));
        error = E_IO_ERROR;
        if (DMAController_i386_Setup(ID, 2, 0, (nSectors << 7) << drives.sectorSize, prop))
            goto out;
        /* commands to send */
        commands[0] = 9;
        commands[1] = com;
        commands[2] = head << 2;
        commands[3] = cylinder;
        commands[4] = head;
        commands[5] = sector;
        commands[6] = drives.sectorSize;
        commands[7] = nSectors;
        commands[8] = drives.gap3;
        commands[9] = 0xff;
        intMutex.count = 0;
        intMutex.data = NULL;
        error = SendCommand();
        if (error)
            goto try_reset;
        /* wait 3s for int */
        error = TaskManager_P(&intMutex, 0, 3000);
        /* timed out? */
        if (error) {
            DEBUG_MSG_PRE(error);
            goto try_reset;
        }
        error = GetResult(7);
        if (error)
            goto try_reset;
        /* error? */
        if ((result[0] & 0xc0) == 0)
            goto nextSectors;
        DEBUG_MSG_PRE((int)result[0]);
        /* write protected error? */
        error = E_READ_ONLY_DEVICE;
        if ((prop & 0x48) && (result[1] & 0x02))
            goto out;
        error = E_IO_ERROR;
        /* time out error? */
        if (result[1] & 0x10)
            error = E_TIMED_OUT;

try_reset:
        if (try < 2) {
            Reset();
            MotorOn(request->unit);
            goto tryAgain;
        }
        goto out;

nextSectors:
        /* read command? */
        if (prop == 0x44)
            MemMan_Copy(request->buffer+offset, dmaBuffer, (nSectors << 7) << drives.sectorSize);
        toDo -= nSectors;
        block += nSectors;
        offset += ((nSectors << 7) << drives.sectorSize);
    }
out:
    TaskManager_Wait(offThread, 3000);
    TaskManager_V(&mutex, 0);
    request->done = request->n - toDo;
    return error;
}

int Seek(unsigned int unit, unsigned int cylinder) {
    int try, error;

    try = 0;
tryAgain:
    try += 1;
    commands[0] = 3;
    commands[1] = 0xf;
    commands[2] = unit;
    commands[3] = cylinder;
    intMutex.count = 0;
    intMutex.data = NULL;
    error = SendCommand();
    if (error)
        goto try_reset;
    /* wait int */
    error = TaskManager_P(&intMutex, 0, 3000);
    if (error) {
        DEBUG_MSG_PRE(error);
        goto try_reset;
    }
    error = SenseInterrupt();
    if (error) {
        DEBUG_MSG_PRE(error);
        goto try_recalibrate;
    }
    drives.lastCyl = cylinder;
    return E_OK;

try_recalibrate:
    if (try < 2) {
        error = Recalibrate(unit);
        if (error == 0)
            goto tryAgain;
    }
    return error;

try_reset:
    if (try < 2) {
        Reset();
        MotorOn(unit);
        goto tryAgain;
    }
    return error;
}

int SendCommand(void) {
    unsigned int i;

    for (i=1; i<=commands[0]; i++) {
        unsigned char msr;
        unsigned int try;
        for (try = 0; try < 10000; try++) {
            __asm__ __volatile__ ("inb %%dx, %%al" : "=a" (msr) : "d" (0x3f4));
            if ((msr & 0xc0) == 0x80) {
                __asm__ __volatile__ ("outb %%al, %%dx" :: "a" (commands[i]), "d" (0x3f5));
                goto end;
            }
        }
        DEBUG_MSG_PRE((int)msr);
        /* controller wants to send bytes */
        if (msr & 0x40)
            return E_IO_ERROR;
        if ((msr & 0xc0) != 0x80)
            return E_TIMED_OUT;
end:
        /* make gcc happy */
        ;
    }
    return E_OK;
}

int SenseInterrupt(void) {
    int error;

    commands[0] = 1;
    commands[1] = 0x08;
    error = SendCommand();
    if (error)
        return error;
    error = GetResult(2);
    if (error)
        return error;
    if (result[0] & 0xc0) {
        if (result[0] & 0x08)
            return E_BUSY;
        return E_IO_ERROR;
    }
    return E_OK;
}

int ServeInterrupt(IntManager_i386_Regs *regs) {
    TaskManager_V(&intMutex, TaskManager_RESUME_KEEP_CPU);
    return E_HANDLED_INT;
}

int Shutdown(void) {
    int ret;

    /* switch off drives motor */
    __asm__ __volatile__ ("outb %%al, %%dx" :: "a" (0x0c), "d" (0x3f2));
    ret = IOPortManager_Free(ID, 0x3f1);
    ASSERT_ERROR(ret);
    ret = IOPortManager_Free(ID, 0x3f7);
    ASSERT_ERROR(ret);
    ret = DMAController_i386_Free(ID, 2);
    ASSERT_ERROR(ret);
    ret = IntManager_i386_Free(ID, 6, 0x01);
    ASSERT_ERROR(ret);
    ret = TaskManager_DestroyTask(offTask);
    ASSERT_ERROR(ret);
    return 0;
}

int Phase1Start(SysModManager_Module *modID) {
    int ret;
    void *ptrRet;

    ID = modID;
    /* start all used modules */
    if ((ret = SysModManager_StartModule((unsigned int)DMAController_i386_Alloc)) ||
        (ret = SysModManager_StartModule((unsigned int)IntManager_i386_Catch)) ||
        (ret = SysModManager_StartModule((unsigned int)IOPortManager_Alloc)) ||
        (ret = SysModManager_StartModule((unsigned int)MemMan_Copy)) ||
        (ret = SysModManager_StartModule((unsigned int)TaskManager_P))) {
        DEBUG_MSG_PRE(ret);
        goto errorReturn;
    }
    if ((ret = IOPortManager_Alloc(modID, 0x3f1, 0x3f5, 0))) {
        DEBUG_MSG_PRE(ret);
        goto errorReturn;
    }
    if ((ret = IOPortManager_Alloc(modID, 0x3f7, 0x3f7, 0))) {
        DEBUG_MSG_PRE(ret);
        goto errorFreeIO3f1;
    }
    ptrRet = DMAController_i386_Alloc(modID, 2, 3);
    if (IS_PTR_ERROR(ptrRet)) {
        DEBUG_MSG_PRE(PTR_TO_ERROR(ptrRet));
        goto errorFreeIO3f7;
    }
    dmaBuffer = ptrRet;
    if ((ret = IntManager_i386_Catch(modID, 6, ServeInterrupt, 0x01+0x04))) {
        DEBUG_MSG_PRE(ret);
        goto errorFreeDMA;
    }
    /* create task to switch off motor */
    if ((ret = TaskManager_CreateTask(MotorOff, NULL, 0x3c000003, &offTask, &offThread))) {
        DEBUG_MSG_PRE(ret);
        goto errorFreeEx;
    }
    /* switch off drives motor */
    __asm__ __volatile__ ("outb %%al, %%dx" :: "a" (DOR), "d" (0x3f2));
    return 0;

errorFreeEx:
    ret = IntManager_i386_Free(modID, 6, 0x01);
    ASSERT_ERROR(ret);
errorFreeDMA:
    ret = DMAController_i386_Free(modID, 2);
    ASSERT_ERROR(ret);
errorFreeIO3f7:
    ret = IOPortManager_Free(modID, 0x3f7);
    ASSERT_ERROR(ret);
errorFreeIO3f1:
    ret = IOPortManager_Free(modID, 0x3f1);
    ASSERT_ERROR(ret);
errorReturn:
    return 1;
}

int Write(BlockDev_IO *request) {

    ASSERT_RET(request, E_BUG);
    ASSERT_RET(request->n, E_BUG);
    ASSERT_RET(request->buffer, E_BUG);

    if (request->unit)
        return E_NO_SUCH_UNIT;
    return Rw(request, 0x48);
}
