/*
*****************************************
* This file implements the BlockDev interface for a generic IDE disk
*
* Version: 0.1.0
* Author: Luiz Henrique Shigunov
* E-mail: luizshigunov@users.sourceforge.net
* WEB: http://www.sourceforge.net/projects/modulos
* Last change: 25/01/2004
*
* See the interface specification to know what each function does.
*
* Copyright (C) 2003-2004, Luiz Henrique Shigunov
*
* This module is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2.1 of the License, or (at your option) any later version.
*
* This module is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this module; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*****************************************
*/

/*
 * Some code used from Linux kernel 2.4.21
*/

#include <stddef.h>
#include <blockdev.h>
#include <debug.h>
#include <errormanager.h>
#include <idecontroller.h>
#include <memman.h>
#include <memmanager.h>
#include <sysmodmanager.h>

#ifdef DEBUG
    #define DUMP_ERROR DumpError
#else
    #define DUMP_ERROR(status,error) do {;} while(0)
#endif

/* status reg */
#define STATUS_ERROR     0x01
#define STATUS_INDEX     0x02
#define STATUS_CORR      0x04
#define STATUS_DRQ       0x08
#define STATUS_SEEK      0x10
#define STATUS_WRITE_ERR 0x20
#define STATUS_READY     0x40
#define STATUS_BUSY      0x80

/* error reg */
#define ERROR_MARK      0x01    /* address mark not found */
#define ERROR_TRACK     0x02    /* track 0 not found */
#define ERROR_ABORT     0x04    /* command aborted */
#define ERROR_MCR       0x08    /* media change request */
#define ERROR_ID        0x10    /* ID mark not found */
#define ERROR_MC        0x20    /* media changed */
#define ERROR_ECC       0x40    /* Uncorrectable ECC error */
#define ERROR_BBD       0x80    /* pre-EIDE: block marked bad */
#define ERROR_ICRC      0x80    /* CRC error */

/* ATA commands */
#define CMD_NOP                 0x00
#define CMD_READ                0x20
#define CMD_READ_EXT            0x24
#define CMD_WRITE               0x30
#define CMD_WRITE_EXT           0x34
#define CMD_IDENTIFY            0xec

#define DO_READ     0x01
#define DO_WRITE    0x02

/* io ports */
#define FEATURE_OFF 0
#define SECTOR_C_OFF 1
#define SECTOR_N_OFF 2
#define LOW_CYL_OFF 3
#define HIGH_CYL_OFF 4
#define DRIVE_OFF 5
#define COMMAND_OFF 6

/* ANSI NCITS ATA6 rev.1b spec */
typedef struct {
    unsigned short  config;         /* lots of obsolete bit flags */
    unsigned short  cyls;           /* Obsolete, "physical" cyls */
    unsigned short  reserved2;      /* reserved (word 2) */
    unsigned short  heads;          /* Obsolete, "physical" heads */
    unsigned short  track_bytes;    /* unformatted bytes per track */
    unsigned short  sector_bytes;   /* unformatted bytes per sector */
    unsigned short  sectors;        /* Obsolete, "physical" sectors per track */
    unsigned short  vendor0;        /* vendor unique */
    unsigned short  vendor1;        /* vendor unique */
    unsigned short  vendor2;        /* Retired vendor unique */
    unsigned char   serial_no[20];  /* 0 = not_specified */
    unsigned short  buf_type;       /* Retired */
    unsigned short  buf_size;       /* Retired, 512 byte increments
                                    * 0 = not_specified
                                    */
    unsigned short  ecc_bytes;      /* for r/w long cmds; 0 = not_specified */
    unsigned char   fw_rev[8];      /* 0 = not_specified */
    unsigned char   model[40];      /* 0 = not_specified */
    unsigned char   max_multsect;   /* 0=not_implemented */
    unsigned char   vendor3;        /* vendor unique */
    unsigned short  dword_io;       /* 0=not_implemented; 1=implemented */
    unsigned char   vendor4;        /* vendor unique */
    unsigned char   capability;     /* (upper byte of word 49)
                                    *  3:  IORDYsup
                                    *  2:  IORDYsw
                                    *  1:  LBA
                                    *  0:  DMA
                                    */
    unsigned short  reserved50;     /* reserved (word 50) */
    unsigned char   vendor5;        /* Obsolete, vendor unique */
    unsigned char   tPIO;           /* Obsolete, 0=slow, 1=medium, 2=fast */
    unsigned char   vendor6;        /* Obsolete, vendor unique */
    unsigned char   tDMA;           /* Obsolete, 0=slow, 1=medium, 2=fast */
    unsigned short  field_valid;    /* (word 53)
                                    *  2:  ultra_ok        word  88
                                    *  1:  eide_ok         words 64-70
                                    *  0:  cur_ok          words 54-58
                                    */
    unsigned short  cur_cyls;       /* Obsolete, logical cylinders */
    unsigned short  cur_heads;      /* Obsolete, l heads */
    unsigned short  cur_sectors;    /* Obsolete, l sectors per track */
    unsigned short  cur_capacity0;  /* Obsolete, l total sectors on drive */
    unsigned short  cur_capacity1;  /* Obsolete, (2 words, misaligned int)     */
    unsigned char   multsect;       /* current multiple sector count */
    unsigned char   multsect_valid; /* when (bit0==1) multsect is ok */
    unsigned int    lba_capacity;   /* Obsolete, total number of sectors */
    unsigned short  dma_1word;      /* Obsolete, single-word dma info */
    unsigned short  dma_mword;      /* multiple-word dma info */
    unsigned short  eide_pio_modes; /* bits 0:mode3 1:mode4 */
    unsigned short  eide_dma_min;   /* min mword dma cycle time (ns) */
    unsigned short  eide_dma_time;  /* recommended mword dma cycle time (ns) */
    unsigned short  eide_pio;       /* min cycle time (ns), no IORDY  */
    unsigned short  eide_pio_iordy; /* min cycle time (ns), with IORDY */
    unsigned short  words69_70[2];  /* reserved words 69-70
                                    * future command overlap and queuing
                                    */
    /* HDIO_GET_IDENTITY currently returns only words 0 through 70 */
    unsigned short  words71_74[4];  /* reserved words 71-74
                                    * for IDENTIFY PACKET DEVICE command
                                    */
    unsigned short  queue_depth;    /* (word 75)
                                    * 15:5 reserved
                                    *  4:0 Maximum queue depth -1
                                    */
    unsigned short  words76_79[4];  /* reserved words 76-79 */
    unsigned short  major_rev_num;  /* (word 80) */
    unsigned short  minor_rev_num;  /* (word 81) */
    unsigned short  command_set_1;  /* (word 82) supported
                                    * 15:  Obsolete
                                    * 14:  NOP command
                                    * 13:  READ_BUFFER
                                    * 12:  WRITE_BUFFER
                                    * 11:  Obsolete
                                    * 10:  Host Protected Area
                                    *  9:  DEVICE Reset
                                    *  8:  SERVICE Interrupt
                                    *  7:  Release Interrupt
                                    *  6:  look-ahead
                                    *  5:  write cache
                                    *  4:  PACKET Command
                                    *  3:  Power Management Feature Set
                                    *  2:  Removable Feature Set
                                    *  1:  Security Feature Set
                                    *  0:  SMART Feature Set
                                    */
    unsigned short  command_set_2;  /* (word 83)
                                    * 15:  Shall be ZERO
                                    * 14:  Shall be ONE
                                    * 13:  FLUSH CACHE EXT
                                    * 12:  FLUSH CACHE
                                    * 11:  Device Configuration Overlay
                                    * 10:  48-bit Address Feature Set
                                    *  9:  Automatic Acoustic Management
                                    *  8:  SET MAX security
                                    *  7:  reserved 1407DT PARTIES
                                    *  6:  SetF sub-command Power-Up
                                    *  5:  Power-Up in Standby Feature Set
                                    *  4:  Removable Media Notification
                                    *  3:  APM Feature Set
                                    *  2:  CFA Feature Set
                                    *  1:  READ/WRITE DMA QUEUED
                                    *  0:  Download MicroCode
                                    */
    unsigned short  cfsse;          /* (word 84)
                                    * cmd set-feature supported extensions
                                    * 15:  Shall be ZERO
                                    * 14:  Shall be ONE
                                    * 13:3 reserved
                                    *  2:  Media Serial Number Valid
                                    *  1:  SMART selt-test supported
                                    *  0:  SMART error logging
                                    */
    unsigned short  cfs_enable_1;   /* (word 85)
                                    * command set-feature enabled
                                    * 15:  Obsolete
                                    * 14:  NOP command
                                    * 13:  READ_BUFFER
                                    * 12:  WRITE_BUFFER
                                    * 11:  Obsolete
                                    * 10:  Host Protected Area
                                    *  9:  DEVICE Reset
                                    *  8:  SERVICE Interrupt
                                    *  7:  Release Interrupt
                                    *  6:  look-ahead
                                    *  5:  write cache
                                    *  4:  PACKET Command
                                    *  3:  Power Management Feature Set
                                    *  2:  Removable Feature Set
                                    *  1:  Security Feature Set
                                    *  0:  SMART Feature Set
                                    */
    unsigned short  cfs_enable_2;   /* (word 86)
                                    * command set-feature enabled
                                    * 15:  Shall be ZERO
                                    * 14:  Shall be ONE
                                    * 13:  FLUSH CACHE EXT
                                    * 12:  FLUSH CACHE
                                    * 11:  Device Configuration Overlay
                                    * 10:  48-bit Address Feature Set
                                    *  9:  Automatic Acoustic Management
                                    *  8:  SET MAX security
                                    *  7:  reserved 1407DT PARTIES
                                    *  6:  SetF sub-command Power-Up
                                    *  5:  Power-Up in Standby Feature Set
                                    *  4:  Removable Media Notification
                                    *  3:  APM Feature Set
                                    *  2:  CFA Feature Set
                                    *  1:  READ/WRITE DMA QUEUED
                                    *  0:  Download MicroCode
                                    */
    unsigned short  csf_default;    /* (word 87)
                                    * command set-feature default
                                    * 15:  Shall be ZERO
                                    * 14:  Shall be ONE
                                    * 13:3 reserved
                                    *  2:  Media Serial Number Valid
                                    *  1:  SMART selt-test supported
                                    *  0:  SMART error logging
                                    */
    unsigned short  dma_ultra;      /* (word 88) */
    unsigned short  word89;         /* reserved (word 89) */
    unsigned short  word90;         /* reserved (word 90) */
    unsigned short  CurAPMvalues;   /* current APM values */
    unsigned short  word92;         /* reserved (word 92) */
    unsigned short  hw_config;      /* hardware config (word 93)
                                    * 15:
                                    * 14:
                                    * 13:
                                    * 12:
                                    * 11:
                                    * 10:
                                    *  9:
                                    *  8:
                                    *  7:
                                    *  6:
                                    *  5:
                                    *  4:
                                    *  3:
                                    *  2:
                                    *  1:
                                    *  0:
                                    */
    unsigned short  acoustic;       /* (word 94)
                                    * 15:8 Vendor's recommended value
                                    *  7:0 current value
                                    */
    unsigned short  words95_99[5];  /* reserved words 95-99 */
    unsigned long long lba_capacity_2;/* 48-bit total number of sectors */
    unsigned short  words104_125[22];/* reserved words 104-125 */
    unsigned short  last_lun;       /* (word 126) */
    unsigned short  word127;        /* (word 127) Feature Set
                                    * Removable Media Notification
                                    * 15:2 reserved
                                    *  1:0 00 = not supported
                                    *      01 = supported
                                    *      10 = reserved
                                    *      11 = reserved
                                    */
    unsigned short  dlf;            /* (word 128)
                                    * device lock function
                                    * 15:9 reserved
                                    *  8   security level 1:max 0:high
                                    *  7:6 reserved
                                    *  5   enhanced erase
                                    *  4   expire
                                    *  3   frozen
                                    *  2   locked
                                    *  1   en/disabled
                                    *  0   capability
                                    */
    unsigned short  csfo;           /*  (word 129)
                                    * current set features options
                                    * 15:4 reserved
                                    *  3:  auto reassign
                                    *  2:  reverting
                                    *  1:  read-look-ahead
                                    *  0:  write cache
                                    */
    unsigned short  words130_155[26];/* reserved vendor words 130-155 */
    unsigned short  word156;        /* reserved vendor word 156 */
    unsigned short  words157_159[3];/* reserved vendor words 157-159 */
    unsigned short  cfa_power;      /* (word 160) CFA Power Mode
                                    * 15 word 160 supported
                                    * 14 reserved
                                    * 13
                                    * 12
                                    * 11:0
                                    */
    unsigned short  words161_175[14];/* Reserved for CFA */
    unsigned short  words176_205[31];/* Current Media Serial Number */
    unsigned short  words206_254[48];/* reserved words 206-254 */
    unsigned short  integrity_word; /* (word 255)
                                    * 15:8 Checksum
                                    *  7:0 Signature
                                    */
} IdeID;

typedef struct _Drive {
    struct _Drive *next;
    unsigned int    driveID;        /* 8-15: interface. 0-7: drive */
    unsigned char   res;
    unsigned char   lba;            /* 0: disabled, 1: enabled */
    unsigned char   head;           /* "real" number of heads */
    unsigned char   sect;           /* "real" sectors per track */
    unsigned int    cyl;            /* "real" number of cyls */
    unsigned long   capacity;       /* total number of sectors */
    unsigned long long capacity48;  /* total number of sectors */
    unsigned        addressing;     /* : 2; 0=28-bit, 1=48-bit, 2=64-bit */
    IdeID id;
} Drive;

SysModManager_Module *ID;
Drive *drives;

int Rw(Drive*,BlockDev_IO*,int);
void SetupDriveCapacity(Drive*);
int SetupTaskFile(Drive*,IdeController_Request*,unsigned int,unsigned int,int);
#ifdef DEBUG
void DumpError(unsigned char,unsigned char);
#endif

/*
 * performs a sanity check on the claimed "lba_capacity"
 * value for this drive (from its reported identification information).
 *
 * Returns: 1 if lba_capacity looks sensible
 *          0 otherwise
 */
int CheckLBACapacity(IdeID *id) {
    unsigned long lba_sects, chs_sects, head, tail;

    if ((id->command_set_2 & 0x0400) && (id->cfs_enable_2 & 0x0400)) {
        DEBUG_MSG(("idedisk: 48-bit Drive: %llu \n", id->lba_capacity_2));
        return 1;
    }

    /*
    * The ATA spec tells large drives to return
    * C/H/S = 16383/16/63 independent of their size.
    * Some drives can be jumpered to use 15 heads instead of 16.
    * Some drives can be jumpered to use 4092 cyls instead of 16383.
    */
    if ((id->cyls == 16383
        || (id->cyls == 4092 && id->cur_cyls == 16383)) &&
        id->sectors == 63 &&
        (id->heads == 15 || id->heads == 16) &&
        id->lba_capacity >= 16383*63*id->heads)
        return 1;

    lba_sects   = id->lba_capacity;
    chs_sects   = id->cyls * id->heads * id->sectors;

    /* perform a rough sanity check on lba_sects:  within 10% is OK */
    if ((lba_sects - chs_sects) < chs_sects/10)
        return 1;

    /* some drives have the word order reversed */
    head = ((lba_sects >> 16) & 0xffff);
    tail = (lba_sects & 0xffff);
    lba_sects = (head | (tail << 16));
    if ((lba_sects - chs_sects) < chs_sects/10) {
        id->lba_capacity = lba_sects;
        return 1;   /* lba_capacity is (now) good */
    }

    return 0;   /* lba_capacity value may be bad */
}

#ifdef DEBUG
void DumpError(unsigned char status, unsigned char error) {
    DEBUG_MSG(("IDE error. status=0x%02x { ", status));
    if (status & STATUS_BUSY)
        DEBUG_MSG(("Busy "));
    else {
        if (status & STATUS_READY)
            DEBUG_MSG(("Ready "));
        if (status & STATUS_WRITE_ERR)
            DEBUG_MSG(("Fault "));
        if (status & STATUS_SEEK)
            DEBUG_MSG(("Seek "));
        if (status & STATUS_DRQ)
            DEBUG_MSG(("Data "));
        if (status & STATUS_CORR)
            DEBUG_MSG(("Corrected "));
        if (status & STATUS_INDEX)
            DEBUG_MSG(("Index "));
        if (status & STATUS_ERROR)
            DEBUG_MSG(("Error "));
    }
    DEBUG_MSG(("}\n"));
    DEBUG_MSG(("IDE error. error=0x%02x { ", error));
    if (error & ERROR_MARK)
        DEBUG_MSG(("AddrMarkNotFound "));
    if (error & ERROR_TRACK)
        DEBUG_MSG(("Track0NotFound "));
    if (error & ERROR_ABORT)
        DEBUG_MSG(("Aborted "));
    if (error & ERROR_MCR)
        DEBUG_MSG(("MediaChangeReq "));
    if (error & ERROR_ID)
        DEBUG_MSG(("IDNotFound "));
    if (error & ERROR_MC)
        DEBUG_MSG(("MediaChanged "));
    if (error & ERROR_ECC)
        DEBUG_MSG(("UncorrectableError "));
    if (error & ERROR_ICRC)
        DEBUG_MSG(("CRC "));
    DEBUG_MSG(("}\n"));
}
#endif

void FixString(char *s, int bytecount, int byteswap) {
    char *p = s;
    /* bytecount must be even */
    char *end = &s[bytecount & ~1];

    if (byteswap) {
        /* convert from big-endian to host byte order */
        for (p = end ; p != s;) {
            unsigned short *pp = (unsigned short *) (p -= 2);
            *pp = (*pp << 8) | (*pp >> 8);
        }
    }

    /* strip leading blanks */
    while (s != end && *s == ' ')
        ++s;

    /* compress internal blanks and strip trailing blanks */
    while (s != end && *s) {
        if (*s++ != ' ' || (s != end && *s && *s != ' '))
            *p++ = *(s-1);
    }

    /* wipe out trailing garbage */
    while (p != end)
        *p++ = '\0';
}

static __inline__ int GetCommand(Drive *drive, int cmd) {
    int lba48bit;

    lba48bit = drive->addressing;
    if (cmd == CMD_READ)
        return (lba48bit) ? CMD_READ_EXT : CMD_READ;
    if (cmd == CMD_WRITE)
        return (lba48bit) ? CMD_WRITE_EXT : CMD_WRITE;
    return CMD_NOP;
}

int DoGetUnitDesc(Drive *drive, BlockDev_UnitDesc *unitDesc, unsigned int size, int prop) {
    unsigned int unitSize;

    ASSERT_RET(drive, E_BUG);

    unitSize = sizeof(BlockDev_UnitDesc);
    if (!(prop & BlockDev_NO_NAME)) {
        /* bigest string */
        unitSize += sizeof("IDE 0,0");
    }
    if (size < unitSize)
        return E_BUFFER_TOO_SMALL;

    unitDesc->unit = drive->driveID;
    unitDesc->type = BlockDev_TYPE_DISK;
    unitDesc->flags = 0;
    unitDesc->blockSize = 512;
    unitDesc->nBlocks = drive->capacity;
    if (!(prop & BlockDev_NO_NAME)) {
        unitDesc->unitName = ((char*)&unitDesc->nBlocks)+sizeof(unitDesc->nBlocks);
        MemMan_Copy(unitDesc->unitName, "IDE 0,0", sizeof("IDE 0,0"));
        /* interface number */
        unitDesc->unitName[4] += drive->driveID >> 8;
        /* unit number */
        unitDesc->unitName[6] += drive->driveID & 0xff;
    } else
        unitDesc->unitName = NULL;
    return E_OK;
}

int GetUnitDesc(unsigned int unit, BlockDev_UnitDesc *unitDesc, unsigned int size, int prop) {
    Drive *drive;

    ASSERT_RET(unitDesc, E_BUG);

    drive = drives;
    while (drive) {
        if (drive->driveID == unit)
            return DoGetUnitDesc(drive, unitDesc, size, prop);
        drive = drive->next;
    }
    return E_NO_SUCH_UNIT;
}

int GetUnitDescIndex(unsigned int index, BlockDev_UnitDesc *unitDesc, unsigned int size, int prop) {
    Drive *drive;

    ASSERT_RET(unitDesc, E_BUG);

    drive = drives;
    while (index-- && drive)
        drive = drive->next;
    if (drive)
        return DoGetUnitDesc(drive, unitDesc, size, prop);
    else
        return E_NO_SUCH_UNIT;
}

int Identify(unsigned int interface, unsigned int drive) {
    IdeController_Request rq;
    int error;

    rq.version = 1;
    rq.interface = interface;
    rq.drive = drive;
    rq.timeout = 10;
    rq.flags = IdeController_PROBE_MODE;
    rq.numIOs = 1;
    rq.IOs[0].port = IdeController_COMMAND_PORT;
    rq.IOs[0].data = CMD_IDENTIFY;
    error = IdeController_SendRequest(&rq);
    if (!error) {
        Drive *newDrive;

        /* create new drive */
        newDrive = MemManager_Alloc(sizeof(Drive), 0);
        /* FIXME: we should check whether allocation successed */
        newDrive->driveID = (interface << 8) | drive;
        newDrive->next = drives;
        drives = newDrive;
        /* get data */
        IdeController_ReadData(interface, &newDrive->id, 512);
        FixString(newDrive->id.model,     sizeof(newDrive->id.model),     1);
        FixString(newDrive->id.fw_rev,    sizeof(newDrive->id.fw_rev),    1);
        FixString(newDrive->id.serial_no, sizeof(newDrive->id.serial_no), 1);
    }
    return error;
}

int ProcessCommand(Drive *drive, IdeController_Request *rq, int cmd, char *buffer) {
    int error, ret;

    ASSERT_RET((cmd & DO_READ) || (cmd & DO_WRITE), E_BUG);

    if (cmd & DO_READ) {
        int toRead;

        toRead = rq->IOs[SECTOR_C_OFF].data ? rq->IOs[SECTOR_C_OFF].data : 256;
        /* is multisector? */
        if (toRead > 1)
            rq->flags = IdeController_START_MULTISECTOR;
        else
            rq->flags = IdeController_ONE_SECTOR;
        do {
            error = IdeController_SendRequest(rq);
            if (error) {
                DUMP_ERROR(rq->status, rq->error);
                break;
            }
            /* transfer data */
            ret = IdeController_ReadData(rq->interface, buffer, 512);
            ASSERT_ERROR(ret);
            buffer += 512;
            toRead--;
            /* more sectors to read? */
            if (toRead > 1)
                rq->flags = IdeController_CONT_MULTISECTOR;
            else
                rq->flags = IdeController_END_MULTISECTOR;
        } while (toRead);
   } else {
        DEBUG_MSG(("IDE disk: Write not implemented\n"));
        error = E_UNSUPPORTED;
    }
    return error;
}

int Read(BlockDev_IO *rq) {
    Drive *drive;

    ASSERT_RET(rq, E_BUG);
    ASSERT_RET(rq->n, E_BUG);
    ASSERT_RET(rq->buffer, E_BUG);

    drive = drives;
    while (drive) {
        if (drive->driveID == rq->unit)
            return Rw(drive, rq, DO_READ);
        drive = drive->next;
    }
    return E_NO_SUCH_UNIT;
}

int Rw(Drive *drive, BlockDev_IO *brq, int prop) {
    IdeController_Request *rq;
    char buf[sizeof(IdeController_Request)+(6*sizeof(IdeController_RequestPort))];
    unsigned int offset, toDo, blockN;
    int ret, error;

    ASSERT_RET(drive, E_BUG);

    rq = (IdeController_Request*)buf;
    offset = 0;
    error = E_OK;
    toDo = brq->n;
    blockN = brq->block;
    rq->version = 1;
    rq->interface = drive->driveID >> 8;
    rq->drive = drive->driveID & 0xff;
    rq->timeout = 30;
    rq->numIOs = 7;
    rq->IOs[FEATURE_OFF].port  = IdeController_FEATURE_PORT;
    rq->IOs[FEATURE_OFF].data  = 0;
    rq->IOs[SECTOR_C_OFF].port = IdeController_SECTOR_COUNT_PORT;
    rq->IOs[SECTOR_N_OFF].port = IdeController_SECTOR_NUMBER_PORT;
    rq->IOs[LOW_CYL_OFF].port  = IdeController_LOW_CYL_PORT;
    rq->IOs[HIGH_CYL_OFF].port = IdeController_HIGH_CYL_PORT;
    rq->IOs[DRIVE_OFF].port    = IdeController_DRIVE_PORT;
    rq->IOs[COMMAND_OFF].port  = IdeController_COMMAND_PORT;
    /* get exclusive use of interface */
    ret = IdeController_LockInterface(drive->driveID >> 8);
    ASSERT_ERROR(ret);
    do {
        int nSectors, try;

        /* we only support at most 256 sectors each time */
        nSectors = (toDo > 256) ? 256 : toDo;
        SetupTaskFile(drive, rq, blockN, nSectors, prop);
        /* try to send command 2 times */
        try = 0;
        do {
            error = ProcessCommand(drive, rq, prop, brq->buffer+offset);
            if (!error)
                break;
            /* FIXME: implement reset of IDE interface */
            DEBUG_MSG(("IDE disk: need to reset!\n"));
        } while (++try < 2);
        if (error)
            break;
        toDo -= nSectors;
        blockN += nSectors;
        offset += nSectors << 9;
    } while (toDo > 0);
    ret = IdeController_UnlockInterface(drive->driveID >> 8);
    ASSERT_ERROR(ret);
    brq->done = brq->n - toDo;
    return error;
}

void SetupDrive(unsigned int driveID) {
    Drive *drive;
    IdeID *id;

    drive = drives;
    while (drive) {
        if (drive->driveID == driveID)
            break;
        drive = drive->next;
    }

    ASSERT_NO_RET(drive);

    id = &drive->id;

    /* start with physical values */
    drive->cyl  = id->cyls;
    drive->head = id->heads;
    drive->sect = id->sectors;

    /* Handle logical geometry translation by the drive */
    if ((id->field_valid & 1) && id->cur_cyls && id->cur_heads && (id->cur_heads <= 16) && id->cur_sectors) {
        drive->cyl  = id->cur_cyls;
        drive->head = id->cur_heads;
        drive->sect = id->cur_sectors;
    }

    /* Use physical geometry if what we have still makes no sense */
    if (drive->head > 16 && id->heads && id->heads <= 16) {
        drive->cyl  = id->cyls;
        drive->head = id->heads;
        drive->sect = id->sectors;
    }

    /* calculate drive capacity, and select LBA if possible */
    SetupDriveCapacity(drive);

    if (id->cfs_enable_2 & 0x0400)
        drive->addressing = 1;
    else
        drive->addressing = 0;
}

/* Compute drive->capacity, the full capacity of the drive */
void SetupDriveCapacity (Drive *drive) {
    IdeController_Setup setup;
    IdeID *id;
    unsigned long capacity;
    unsigned long long capacity_2;
    int ret;

    id = &drive->id;
    capacity = drive->cyl * drive->head * drive->sect;
    capacity_2 = capacity;
    drive->capacity48 = 0;

    drive->lba = 0;
    setup.version = 1;
    setup.interface = drive->driveID >> 8;
    setup.drive = drive->driveID & 0xff;
    setup.flags = 0;
    ret = IdeController_SetupDrive(&setup);
    ASSERT_ERROR(ret);

    if (id->cfs_enable_2 & 0x0400) {
        capacity_2          = id->lba_capacity_2;
        drive->head         = 255;
        drive->sect         = 63;
        drive->cyl          = (unsigned int) capacity_2 / (drive->head * drive->sect);
        drive->capacity48   = capacity_2;
        drive->capacity     = (unsigned long) capacity_2;
        drive->lba = 1;
        setup.flags |= IdeController_SETUP_LBA;
        ret = IdeController_SetupDrive(&setup);
        ASSERT_ERROR(ret);
        return;
    /* Determine capacity, and use LBA if the drive properly supports it */
    } else if ((id->capability & 2) && CheckLBACapacity(id)) {
        capacity = id->lba_capacity;
        drive->cyl = capacity / (drive->head * drive->sect);
        drive->lba = 1;
        setup.flags |= IdeController_SETUP_LBA;
        ret = IdeController_SetupDrive(&setup);
        ASSERT_ERROR(ret);
    }

    drive->capacity = capacity;

    if ((id->command_set_2 & 0x0400) && (id->cfs_enable_2 & 0x0400)) {
        drive->capacity48 = id->lba_capacity_2;
        drive->head = 255;
        drive->sect = 63;
        drive->cyl = (unsigned long)(drive->capacity48) / (drive->head * drive->sect);
    }
}

static __inline__ int SetupTaskFileCHS(Drive *drive, IdeController_Request *rq, unsigned int blockN, unsigned int n, int prop) {
    unsigned int command;
    unsigned int track  = (blockN / drive->sect);
    unsigned int sect   = (blockN % drive->sect) + 1;
    unsigned int head   = (track % drive->head);
    unsigned int cyl    = (track / drive->head);

    command = (prop & DO_READ) ? CMD_READ : CMD_WRITE;
    rq->IOs[SECTOR_C_OFF].data = (n==256) ? 0x00 : n;
    rq->IOs[SECTOR_N_OFF].data = sect;
    rq->IOs[LOW_CYL_OFF].data  = cyl;
    rq->IOs[HIGH_CYL_OFF].data = (cyl>>8);
    rq->IOs[DRIVE_OFF].data    = head;
    rq->IOs[COMMAND_OFF].data  = GetCommand(drive, command);
    return E_OK;
}

static __inline__ int SetupTaskFileLBA28(Drive *drive, IdeController_Request *rq, unsigned int blockN, unsigned int n, int prop) {
    unsigned int command;

    command = (prop & DO_READ) ? CMD_READ : CMD_WRITE;
    rq->IOs[SECTOR_C_OFF].data = (n==256) ? 0x00 : n;
    rq->IOs[SECTOR_N_OFF].data = blockN;
    rq->IOs[LOW_CYL_OFF].data  = (blockN>>=8);
    rq->IOs[HIGH_CYL_OFF].data = (blockN>>=8);
    rq->IOs[DRIVE_OFF].data    = ((blockN>>8)&0x0f);
    rq->IOs[COMMAND_OFF].data  = GetCommand(drive, command);
    return E_OK;
}

int SetupTaskFile(Drive *drive, IdeController_Request *rq, unsigned int blockN, unsigned int n, int prop) {
    /* 28-bit LBA */
    if (drive->lba)
        return SetupTaskFileLBA28(drive, rq, blockN, n, prop);

    /* 28-bit CHS : DIE DIE DIE piece of legacy crap!!! */
    return SetupTaskFileCHS(drive, rq, blockN, n, prop);
}

int Shutdown(void) {
    Drive *drive, *tmp;
    int ret;

    drive = drives;
    while (drive) {
        tmp = drive->next;
        ret = IdeController_PutDrive(drive->driveID >> 8, drive->driveID & 0xff);
        ASSERT_ERROR(ret);
        ret = MemManager_Free(drive);
        ASSERT_ERROR(ret);
        drive = tmp;
    }
    return 0;
}

int Phase1Start(SysModManager_Module *modID) {
    unsigned int interface, drive;
    int ret, error;

    ID = modID;
    /* start all used modules */
    if ((ret = SysModManager_StartModule((unsigned int)MemManager_Alloc)) ||
        (ret = SysModManager_StartModule((unsigned int)IdeController_GetDrive)) ||
        (ret = SysModManager_StartModule((unsigned int)MemMan_Copy))) {
        DEBUG_MSG_PRE(ret);
        return 1;
    }
    /* find all ide disks */
    interface = 0;
    while (!IdeController_LockInterface(interface)) {
        drive = 0;
        while ((error = IdeController_GetDrive(interface, drive)) != E_NO_SUCH_DRIVE) {
            if (!error) {
                /* check whether it is a disk */
                if (!Identify(interface, drive))
                    SetupDrive((interface << 8) | drive);
                else {
                    ret = IdeController_PutDrive(interface, drive);
                    ASSERT_ERROR(ret);
                }
            }
            drive++;
        }
        IdeController_UnlockInterface(interface);
        interface++;
    }
    return 0;
}

int Write(BlockDev_IO *rq) {
    Drive *drive;

    ASSERT_RET(rq, E_BUG);
    ASSERT_RET(rq->n, E_BUG);
    ASSERT_RET(rq->buffer, E_BUG);

    drive = drives;
    while (drive) {
        if (drive->driveID == rq->unit)
            return Rw(drive, rq, DO_WRITE);
        drive = drive->next;
    }
    return E_NO_SUCH_UNIT;
}
