/* libavl - manipulates AVL trees.
   Copyright (C) 1998, 1999 Free Software Foundation, Inc.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.

   The author may be contacted at <pfaffben@pilot.msu.edu> on the
   Internet, or as Ben Pfaff, 12167 Airport Rd, DeWitt MI 48820, USA
   through more mundane means. */

/*
*****************************************
* This file implements the AVLTree_In interface
*
* Version: 0.2.2
* Author: Luiz Henrique Shigunov
* E-mail: luizshigunov@users.sourceforge.net
* WEB: http://www.sourceforge.net/projects/modulos
* Last change: 25/06/2005
*
* See the interface specification to know what each function does.
*
* Copyright (C) 2000-2005, Luiz Henrique Shigunov
*****************************************
*/

#include <stddef.h>

#define IS_PTR_ERROR(x) ((unsigned int)(x) > (unsigned int)-4096L)
#define PTR_TO_ERROR(x) (-(int)(x))
#define ERROR_TO_PTR(x) ((void*)-(x))

#define E_BAD_VALUE 1
#define E_IN_USE 1
#define E_NOT_FOUND 1
#define E_OK 0

typedef struct {
    struct _AVLNode_In *root;
    int (*cmpFunc)(const void*,const void*);
    int (*cmpAddFunc)(const void*,const void*);
    unsigned int count;
} AVLTree_In;

typedef struct _AVLNode_In {
    struct _AVLNode_In *link[2];
    signed char bal;
    char cache;
    char pad[2];
} AVLNode_In;

AVLNode_In *AVLTree_In_FindAdd(AVLTree_In*,AVLNode_In*);

int AVLTree_In_Add(AVLTree_In *tree, AVLNode_In *info) {
    AVLNode_In *p;

    if (tree == NULL)
        return E_BAD_VALUE;
    p = AVLTree_In_FindAdd(tree, info);
    if (IS_PTR_ERROR(p))
        return PTR_TO_ERROR(p);
    if (p != info)
        return E_IN_USE;
    return E_OK;
}

AVLNode_In *AVLTree_In_Find(const AVLTree_In *tree, const void *info) {
    AVLNode_In *p;

    if (!tree)
        return ERROR_TO_PTR(E_BAD_VALUE);
    for (p = tree->root; p; ) {
        int diff = tree->cmpFunc(info, p);
        if (diff < 0)
            p = p->link[0];
        else if (diff > 0)
            p = p->link[1];
        else
            return p;
    }
    return ERROR_TO_PTR(E_NOT_FOUND);
}

AVLNode_In *AVLTree_In_FindAdd(AVLTree_In *tree, AVLNode_In *info) {
  /* Uses Knuth's Algorithm 6.2.3A (balanced tree search and
     insertion), but caches results of comparisons.  In empirical
     tests this eliminates about 25% of the comparisons seen under
     random insertions.  */

  /* A1. */
    AVLNode_In *t;
    AVLNode_In *s, *p, *q, *r, *found;

    if (!tree)
        return ERROR_TO_PTR(E_BAD_VALUE);
    t = (AVLNode_In *)&tree->root;
    s = p = tree->root;

    if (!s) {
        tree->count = 1;
        tree->root = info;
        info->link[0] = info->link[1] = NULL;
        info->bal = 0;
        return info;
    }
for (;;) {
        /* A2. */
        int diff = tree->cmpAddFunc(info, p);
        /* A3. */
        if (diff < 0) {
            p->cache = 0;
            q = p->link[0];
            if (!q) {
                p->link[0] = q = info;
                break;
            }
        /* A4. */
        } else if (diff > 0) {
            p->cache = 1;
            q = p->link[1];
            if (!q) {
                p->link[1] = q = info;
                break;
            }
        } else {
            /* A2. */
            return p;
        }
        /* A3, A4. */
        if (q->bal != 0)
            t = p, s = q;
        p = q;
    }
    /* A5. */
    tree->count++;
    q->link[0] = q->link[1] = NULL;
    q->bal = 0;
    found = q;
    /* A6. */
    r = p = s->link[(int) s->cache];
    while (p != q) {
        p->bal = p->cache * 2 - 1;
        p = p->link[(int) p->cache];
    }
    /* A7. */
    if (!s->cache) {
        /* a = -1. */
        if (!s->bal) {
            s->bal = -1;
            return found;
        } else if (s->bal == +1) {
            s->bal = 0;
            return found;
        }
        if (r->bal == -1) {
            /* A8. */
            p = r;
            s->link[0] = r->link[1];
            r->link[1] = s;
            s->bal = r->bal = 0;
        } else {
            /* A9. */
            p = r->link[1];
            r->link[1] = p->link[0];
            p->link[0] = r;
            s->link[0] = p->link[1];
            p->link[1] = s;
            if (p->bal == -1)
                s->bal = 1, r->bal = 0;
            else if (p->bal == 0)
                s->bal = r->bal = 0;
            else
                s->bal = 0, r->bal = -1;
            p->bal = 0;
        }
    } else {
        /* a == +1. */
        if (!s->bal) {
            s->bal = 1;
            return found;
        } else if (s->bal == -1) {
            s->bal = 0;
            return found;
        }
        if (r->bal == +1) {
            /* A8. */
            p = r;
            s->link[1] = r->link[0];
            r->link[0] = s;
            s->bal = r->bal = 0;
        } else {
            /* A9. */
            p = r->link[0];
            r->link[0] = p->link[1];
            p->link[1] = r;
            s->link[1] = p->link[0];
            p->link[0] = s;
            if (p->bal == +1)
                s->bal = -1, r->bal = 0;
            else if (p->bal == 0)
                s->bal = r->bal = 0;
            else
                s->bal = 0, r->bal = 1;
            p->bal = 0;
        }
    }
    /* A10. */
    if (t != (AVLNode_In *)&tree->root && s == t->link[1])
        t->link[1] = p;
    else
        t->link[0] = p;
    return found;
}
