/*
* Read system configuration file
* Copyright (C) 2002, Luiz Henrique Shigunov (luizshigunov@users.sourceforge.net)
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2, or (at your option)
* any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330,
* Boston, MA 02111-1307, USA.
*/

#include <stddef.h>

#include "utils.h"
#include "cfgfile.h"

static char *currMod = NULL;
static char *end = NULL;

/* returns: 0 - has more chars; != 0 - no more chars */
static __inline__ int RemoveBlanks() {
    while (currMod < end && (*currMod == ' ' || *currMod == '\t' || *currMod == '\n')) {
        currMod++;
    }
    return (currMod == end);
}

/* returns: number of chars of the id */
static __inline__ unsigned int ReadID() {
    unsigned int len;

    len = 0;
    while (currMod < end && ((*currMod >= '0' && *currMod <= '9') || ((*currMod & 0xdf) >= 'A' && (*currMod & 0xdf) <= 'Z') || *currMod == '_')) {
        currMod++;
        len++;
    }
    return len;
}

/* returns: number of chars of the data */
static __inline__ unsigned int ReadKeyData() {
    unsigned int len;

    len = 0;
    while (currMod < end && *currMod != '\n') {
        currMod++;
        len++;
    }
    return len;
}

#define CfgManager_MOD_INT 0x01
#define CfgManager_MOD_INT_CLOSE 0x02
#define CfgManager_MAY_END 0x03
#define CfgManager_GROUP_KEY 0x04

#define CFGFILE_HAS_SYSMODMANAGER 0x01
#define CFGFILE_HAS_MODULES 0x02
#define CFGFILE_NO_MORE 0x03

static int state = 0;
static int openGroups = 0;
static int hasModInt = 0;
static int cfgfileState = 0;
static int modulesOpenGroups = 0;
static unsigned int idLen = 0;
static char *id = NULL;
static char *path = NULL;
static unsigned int pathSize = 0;
static int hasProp = 0;
static int prop = 0;

static void ParseCfgFile() {
    if (cfgfileState == CFGFILE_NO_MORE) {
        return;
    }
    while (currMod < end) {
        if (RemoveBlanks()) {
            break;
        }
        switch (state) {
            case CfgManager_MOD_INT:
                id = currMod;
                if ((idLen = ReadID())) {
                    if (!util_strncmp(id, "SysModManager", idLen)) {
                        cfgfileState = CFGFILE_HAS_SYSMODMANAGER;
                    } else {
                        cfgfileState = CFGFILE_NO_MORE;
                    }
                    state = CfgManager_MOD_INT_CLOSE;
                    continue;
                }
                goto endWhile;
            case CfgManager_MOD_INT_CLOSE:
                if (*currMod == ']') {
                    state = CfgManager_MAY_END;
                    currMod++;
                    continue;
                }
                goto endWhile;
            case CfgManager_MAY_END:
                if (*currMod == '}') {
                    if (openGroups) {
                        if (cfgfileState == CFGFILE_HAS_MODULES && openGroups == modulesOpenGroups) {
                            cfgfileState = CFGFILE_NO_MORE;
                            return;
                        }
                        openGroups--;
                        currMod++;
                        continue;
                    }
                } else if (*currMod == '[') {
                    if (!openGroups) {
                        currMod++;
                        state = CfgManager_MOD_INT;
                        hasModInt = 1;
                        continue;
                    }
                } else if (hasModInt) {
                    id = currMod;
                    if ((idLen = ReadID())) {
                        state = CfgManager_GROUP_KEY;
                        continue;
                    }
                }
                goto endWhile;
            case CfgManager_GROUP_KEY:
                if (*currMod == '{') {
                    currMod++;
                    openGroups++;
                    state = CfgManager_MAY_END;
                    if (cfgfileState == CFGFILE_HAS_SYSMODMANAGER &&
                        !util_strncmp(id, "modules", idLen)) {
                        cfgfileState = CFGFILE_HAS_MODULES;
                        modulesOpenGroups = openGroups;
                        return;
                    }
                    continue;
                } else if (*currMod == '=') {
                    currMod++;
                    state = CfgManager_MAY_END;
                    if (cfgfileState == CFGFILE_HAS_MODULES) {
                        if (!util_strncmp(id, "path", idLen)) {
                            path = currMod;
                            pathSize = ReadKeyData();
                            return;
                        } else if (!util_strncmp(id, "prop", idLen)) {
                            hasProp = 1;
                            prop = util_strtol(currMod, NULL, 0);
                            ReadKeyData();
                            return;
                        }
                    }
                    continue;
                }
                goto endWhile;
        }
    }
endWhile:
    cfgfileState = CFGFILE_NO_MORE;
    return;
}

void CfgFile_Start(char *file, unsigned int size) {
    end = file + size;
    /* find SysModManager */
    currMod = file;
    state = CfgManager_MAY_END;
    ParseCfgFile();
}

const char *CfgFile_GetModulePath(unsigned int *size) {
    path = NULL;
    ParseCfgFile();
    *size = pathSize;
    return path;
}

int CfgFile_GetModuleProp() {
    if (!hasProp) {
        prop = 0;
        ParseCfgFile();
    }
    hasProp = 0;
    return prop;
}
