/*
* Read system module file
* Copyright (C) 2000-2002, Luiz Henrique Shigunov (luizshigunov@users.sourceforge.net)
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2, or (at your option)
* any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330,
* Boston, MA 02111-1307, USA.
*/

#include <stddef.h>

#include "utils.h"
#include "module_file.h"

/* SM03 */
#define HEADER_SIGN 0x33304d53

struct InternalHeader {
    char md5[16];
    char sign[4];
    unsigned int code;
    unsigned int codeSize;
    unsigned int data;
    unsigned int dataSize;
    unsigned int uDataSize;
    unsigned int usedFuncs;
    unsigned int usedFuncsSize;
    unsigned int usedFuncReloc;
    unsigned int usedFuncRelocSize;
    unsigned int interfaces;
    unsigned int interfacesSize;
    unsigned int dataReloc;
    unsigned int dataRelocSize;
    unsigned int codeReloc;
    unsigned int codeRelocSize;
    unsigned int strings;
    unsigned short int stringsSize;
    unsigned short int version;
    unsigned short int prop;
    unsigned short int comment;
    unsigned int phase0StartF;
    unsigned int phase1StartF;
    unsigned int shutdownF;
} __attribute__ ((packed));

typedef struct InternalFunction InternalFunction;
struct InternalFunction {
    unsigned int code;
    unsigned short prop;
} __attribute__ ((packed));

typedef struct InternalImplementation InternalImplementation;
struct InternalImplementation {
    struct InternalFunction *functions;
    unsigned short name;
} __attribute__ ((packed));

typedef struct InternalInterface InternalInterface;
struct InternalInterface {
    unsigned short name;
    unsigned short nFunctions;
    unsigned short nImp;
    struct InternalImplementation imps[1];
} __attribute__ ((packed));

typedef struct InternalUsedFunc InternalUsedFunc;
struct InternalUsedFunc {
    unsigned short interface;
    unsigned short imp;
    unsigned short function;
} __attribute__ ((packed));

typedef struct InternalUsedFuncReloc InternalUsedFuncReloc;
struct InternalUsedFuncReloc {
    unsigned int offset;
    unsigned int propInd;
} __attribute__ ((packed));

#define HEADER(module) ((struct InternalHeader*)(module)->fileHeader)
#define INTERFACES(module) ((InternalInterface*)((module)->fileHeader + HEADER(module)->interfaces))
#define STRINGS(module) ((module)->fileHeader + HEADER(module)->strings)
#define USED_FUNC_RELOCS(module) ((InternalUsedFuncReloc*)((module)->fileHeader + HEADER(module)->usedFuncReloc))
#define USED_FUNCS(module) ((InternalUsedFunc*)((module)->fileHeader + HEADER(module)->usedFuncs))
#define CODE_RELOC(module) ((module)->fileHeader + HEADER(module)->codeReloc)
#define DATA_RELOC(module) ((module)->fileHeader + HEADER(module)->dataReloc)

static Module_Implementation imp = {NULL, };
static Module_Interface inter = {NULL, };
static Module_Module mod = {{0, }, };
static Module_UsedFuncReloc usedFReloc = {NULL, };

int Module_CheckFile(Module *module) {
    if (*(unsigned int*)HEADER(module)->sign != HEADER_SIGN) {
        return 1;
    }
    return 0;
}

void Module_GetFunctionsInfo(Module *module, unsigned int interface, unsigned int implementation, Module_Function *functions) {
    InternalInterface *currInt;

    currInt = INTERFACES(module);
    for(; interface; interface--) {
        currInt = (InternalInterface*)((char*)currInt+sizeof(InternalInterface)-sizeof(InternalImplementation)+sizeof(InternalImplementation)*currInt->nImp);
        if ((unsigned int)currInt >= ((unsigned int)INTERFACES(module) + HEADER(module)->interfacesSize)) {
            return;
        }
    }
    if (implementation < currInt->nImp) {
        InternalFunction *funcs;
        unsigned int i;

        funcs = (InternalFunction*)(module->fileHeader + (unsigned int)currInt->imps[implementation].functions);
        for (i=0; i < currInt->nFunctions; i++) {
            functions[i].codeStart = funcs[i].code;
            functions[i].prop = funcs[i].prop;
        }
    }
    return;
}

Module_Implementation *Module_GetImplementationInfo(Module *module, unsigned int interface, unsigned int implementation) {
    InternalInterface *currInt;

    currInt = INTERFACES(module);
    for(; interface; interface--) {
        currInt = (InternalInterface*)((char*)currInt+sizeof(InternalInterface)-sizeof(InternalImplementation)+sizeof(InternalImplementation)*currInt->nImp);
        if ((unsigned int)currInt >= ((unsigned int)INTERFACES(module) + HEADER(module)->interfacesSize)) {
            return NULL;
        }
    }
    if (implementation < currInt->nImp) {
        imp.name = STRINGS(module) + currInt->imps[implementation].name;
        return &imp;
    }
    return NULL;
}

Module_Interface *Module_GetInterfaceInfo(Module *module, unsigned int interface) {
    InternalInterface *currInt;

    currInt = INTERFACES(module);
    for(; interface; interface--) {
        currInt = (InternalInterface*)((char*)currInt+sizeof(InternalInterface)-sizeof(InternalImplementation)+sizeof(InternalImplementation)*currInt->nImp);
        if ((unsigned int)currInt >= ((unsigned int)INTERFACES(module) + HEADER(module)->interfacesSize)) {
            return NULL;
        }
    }
    inter.name = STRINGS(module) + currInt->name;
    inter.nFunctions = currInt->nFunctions;
    return &inter;
}

Module_Module *Module_GetModuleInfo(Module *module) {
    util_memmove(mod.md5, HEADER(module)->md5, 16);
    mod.comment = STRINGS(module) + HEADER(module)->comment;
    mod.phase0Start = (void*)HEADER(module)->phase0StartF;
    mod.phase1Start = (void*)HEADER(module)->phase1StartF;
    mod.shutdown = (void*)HEADER(module)->shutdownF;
    mod.codeSize = HEADER(module)->codeSize;
    mod.dataSize = HEADER(module)->dataSize + HEADER(module)->uDataSize;
    mod.version = HEADER(module)->version;
    mod.prop = HEADER(module)->prop;
    return &mod;
}

Module_UsedFuncReloc *Module_GetUsedFuncReloc(Module *module, unsigned int i) {
    unsigned int ind;

    if (i >= (HEADER(module)->usedFuncRelocSize >> 3)) {
        return NULL;
    }
    usedFReloc.offset = USED_FUNC_RELOCS(module)[i].offset;
    usedFReloc.prop = USED_FUNC_RELOCS(module)[i].propInd & 0xff;
    ind = USED_FUNC_RELOCS(module)[i].propInd >> 8;
    usedFReloc.interface = STRINGS(module) + USED_FUNCS(module)[ind].interface;
    usedFReloc.imp = NULL;
    if (USED_FUNCS(module)[ind].imp) {
        usedFReloc.imp = STRINGS(module) + USED_FUNCS(module)[ind].imp;
    }
    usedFReloc.function = USED_FUNCS(module)[ind].function;
    return &usedFReloc;
}

void Module_ReadCode(Module *module, char *buf, char *codePos, char *dataPos) {
    util_memmove(buf, module->fileHeader + HEADER(module)->code, HEADER(module)->codeSize);
    /* do relocations */
    if (HEADER(module)->codeRelocSize) {
        char *currR;
        unsigned int dataRSize, codeRSize;

        currR = CODE_RELOC(module);
        dataRSize = (*(unsigned int*)currR) >> 2;
        codeRSize = (*(unsigned int*)(currR+4)) >> 2;
        currR += 8;
        for(; dataRSize; dataRSize--, currR += 4) {
            *(int*)(buf+*(unsigned int*)currR) += (int)dataPos;
        }
        for(; codeRSize; codeRSize--, currR += 4) {
            *(int*)(buf+*(unsigned int*)currR) += (int)codePos;
        }
    }
}

void Module_ReadData(Module *module, char *buf, char *codePos, char *dataPos) {
    if (!(HEADER(module)->dataSize + HEADER(module)->uDataSize)) {
        return;
    }
    if (HEADER(module)->dataSize) {
        util_memmove(buf, module->fileHeader + HEADER(module)->data, HEADER(module)->dataSize);
    }
    if (HEADER(module)->uDataSize) {
        util_memset(buf + HEADER(module)->dataSize, 0, HEADER(module)->uDataSize);
    }
    /* do relocations */
    if (HEADER(module)->dataRelocSize) {
        char *currR;
        unsigned int dataRSize, codeRSize;

        currR = DATA_RELOC(module);
        dataRSize = (*(unsigned int*)currR) >> 2;
        codeRSize = (*(unsigned int*)(currR+4)) >> 2;
        currR += 8;
        for(; dataRSize; dataRSize--, currR += 4) {
            *(int*)(buf+*(unsigned int*)currR) += (int)dataPos;
        }
        for(; codeRSize; codeRSize--, currR += 4) {
            *(int*)(buf+*(unsigned int*)currR) += (int)codePos;
        }
    }
}
