/*
* starter - loads config file and all modules marked to be loaded at boot time
* Copyright (C) 2002-2005, Luiz Henrique Shigunov (luizshigunov@users.sourceforge.net)
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2, or (at your option)
* any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330,
* Boston, MA 02111-1307, USA.
*/

#include <stddef.h>

#include "fsys.h"
#include "asm.h"
#include "utils.h"
#include "common.h"
#include "avltree_in.h"
#include "module_file.h"
#include "cfgfile.h"
#include "multiboot.h"

#define IS_PTR_ERROR(x) ((unsigned int)(x) > (unsigned int)-4096L)

#define SYSTEM_MEM 0xe0000000

typedef struct _TaskManager_RWSync TaskManager_RWSync;
struct _TaskManager_RWSync {void *data[4];};

#define SysModManager_LOAD_AT_BOOT 0x000004

typedef struct _SysModManager_Module SysModManager_Module;
struct _SysModManager_Module {
    SysModManager_Module *next;
    int prop;
    /*
    0x000001 - permanent module
    0x000002 - automatic start
    0x000004 - load at boot
    0x010000 - module phase 0 started
    0x020000 - module phase 1 started
    */
    unsigned int memStart;  /* = 0xffffffff module not loaded */
    unsigned int memSize;
    unsigned int refCounter;
    struct _SysModManager_ModData *data;
    TaskManager_RWSync dataSync;
    char *path;
    int (*phase0Start)(SysModManager_Module*);
    int (*phase1Start)(SysModManager_Module*);
    void (*shutdown)(void);
    char md5[16];
};

typedef struct _SysModManager_ModData {
    struct _SysModManager_ModData *next;
    SysModManager_Module *module;
    void *info;
} SysModManager_ModData;

typedef struct _SysModManager_Interface {
    AVLTree_In_Node node;
    char *name;
    unsigned int nFunctions;    /* # functions */
    struct _SysModManager_Implementation *imps; /* implementation list */
} SysModManager_Interface;

typedef struct _SysModManager_Implementation {
    struct _SysModManager_Implementation *next;
    struct _SysModManager_Module *module;
    SysModManager_Interface *interface;
    char *name;
    unsigned short interN;
    unsigned short impN;
    struct _SysModManager_Function *functions;
} SysModManager_Implementation;

typedef struct _SysModManager_Function {
    char *code;
    unsigned char prop;
    /*
    0x01 - system function
    0x02 - not implemented
    */
    unsigned char words;
    short reserved;
} SysModManager_Function;

int CmpInterfaces(const void*,const AVLTree_In_Node*);
int CmpAddInterfaces(const AVLTree_In_Node*,const AVLTree_In_Node*);

char *currAddr = NULL;
char *startAddr = NULL;
char *tmpAddr = (char*)DISK_BUFFERADDR + DISK_BUFFERLEN;
int cfgSize = 0;
char *cfgStart = NULL;
Module *modules = NULL;
Interface *interfaces = NULL;
AVLTree_In_Tree interfacesRoot = {NULL, CmpInterfaces, CmpAddInterfaces};

int CmpInterfaces(const void *info, const AVLTree_In_Node *node) {
    return util_strcmp(info, ((Interface*)node)->name);
}

int CmpAddInterfaces(const AVLTree_In_Node *addNode, const AVLTree_In_Node *node) {
    return util_strcmp(((Interface*)addNode)->name, ((Interface*)node)->name);
}

/* returns: 0 - success; 1 - error */
static __inline__ int StartInterfaces() {
    Module *currMod;
    Module_Interface *interInfo;
    Module_Implementation *impInfo;
    Interface *currInt;
    Implementation *currImp;
    unsigned int i, j;

    currMod = modules;
    do {
        i = 0;
        interInfo = Module_GetInterfaceInfo(currMod, i);
        do {
            currInt = (Interface*)AVLTree_In_Find(&interfacesRoot, interInfo->name);
            if (IS_PTR_ERROR(currInt)) {
                /* register a new interface */
                currInt = (Interface*)ALLOC_TMP_MEM(sizeof(Interface));
                if (!currInt) {
                    util_print("Not enough memory.\n");
                    return 1;
                }
                currInt->name = interInfo->name;
                currInt->nFunctions = interInfo->nFunctions;
                currInt->imps = NULL;
                currInt->next = interfaces;
                interfaces = currInt;
                AVLTree_In_Add(&interfacesRoot, (AVLTree_In_Node*)currInt);
            }
            j = 0;
            impInfo = Module_GetImplementationInfo(currMod, i, j);
            do {
                /* register a new implementation */
                currImp = (Implementation*)ALLOC_TMP_MEM(sizeof(Implementation));
                if (!currImp) {
                    util_print("Not enough memory.\n");
                    return 1;
                }
                currImp->module = currMod;
                currImp->interface = currInt;
                currImp->name = impInfo->name;
                currImp->interN = i;
                currImp->impN = j;
                currImp->functions = (Function*)ALLOC_TMP_MEM(currInt->nFunctions * sizeof(Module_Function));
                if (!currImp->functions) {
                    util_print("Not enough memory.\n");
                    return 1;
                }
                Module_GetFunctionsInfo(currMod, i, j, (Module_Function*)currImp->functions);
                currImp->next = currInt->imps;
                currInt->imps = currImp;
                j++;
                impInfo = Module_GetImplementationInfo(currMod, i, j);
            } while (impInfo);
            i++;
            interInfo = Module_GetInterfaceInfo(currMod, i);
        } while (interInfo);
        currMod = currMod->next;
    } while (currMod);
    return 0;
}

/* returns: 0 - success; 1 - error */
static __inline__ int DoFunctionRelocs(Module *module) {
    Module_UsedFuncReloc *usedFReloc;
    unsigned int i;

    i = 0;
    while ((usedFReloc = Module_GetUsedFuncReloc(module, i++))) {
        Interface *currInt;
        Implementation *currImp;

        /* get the interface */
        currInt = (Interface*)AVLTree_In_Find(&interfacesRoot, usedFReloc->interface);
        if (IS_PTR_ERROR(currInt)) {
            util_print("Cannot find '");
            util_print(usedFReloc->interface);
            util_print("' interface\n");
            return 1;
        }
        /* find the implementation if any */
        currImp = currInt->imps;
        if (usedFReloc->imp) {
            while (currImp && util_strcmp(usedFReloc->imp, currImp->name)) {
                currImp = currImp->next;
            }
            if (!currImp) {
                util_print("Cannot find '");
                util_print(usedFReloc->imp);
                util_print("' implementation from '");
                util_print(usedFReloc->interface);
                util_print("' interface\n");
                return 1;
            }
        }
        if (usedFReloc->function >= currInt->nFunctions) {
            util_print("Invalid function number.");
            return 1;
        }
        currImp->module->refCounter++;
        /* relative or absolute relocation? */
        if (usedFReloc->prop) {
            *(int*)(module->codePos+usedFReloc->offset) = (int)currImp->functions[usedFReloc->function].code;
        } else {
            *(int*)(module->codePos+usedFReloc->offset) = currImp->functions[usedFReloc->function].code - (module->codePos+usedFReloc->offset+4+SYSTEM_MEM);
        }
    }
    return 0;
}

/* returns: 0 - success; 1 - error */
static __inline__ int LoadModule(Module *module) {
    Module_Module *moduleInfo;
    char *start;

    moduleInfo = Module_GetModuleInfo(module);
    start = currAddr;
    module->memStart = currAddr + SYSTEM_MEM;
    module->codePos = currAddr;
    currAddr = (char*)ALIGN_DATA_UP(currAddr + moduleInfo->codeSize, 4);
    module->dataPos = currAddr;
    currAddr += moduleInfo->dataSize;
    /* page align */
    currAddr = (char*)ALIGN_DATA_UP(currAddr, 0x1000);
    module->memSize = currAddr - start;
    Module_ReadCode(module, module->codePos, module->codePos + SYSTEM_MEM, module->dataPos + SYSTEM_MEM);
    Module_ReadData(module, module->dataPos, module->codePos + SYSTEM_MEM, module->dataPos + SYSTEM_MEM);
    return 0;
}

/* returns: 0 - success; 1 - error */
static __inline__ int LoadModules() {
    Interface *currInt;
    Module *currMod;
    const char *path;
    unsigned int size, i;

    /* read all modules that must be loaded at boot */
    CfgFile_Start(cfgStart, cfgSize);
    while ((path = CfgFile_GetModulePath(&size))) {
        int prop;

        prop = CfgFile_GetModuleProp();
        if (!(prop & SysModManager_LOAD_AT_BOOT)) {
            continue;
        }
        currMod = (Module*)ALLOC_TMP_MEM(sizeof(Module));
        if (!currMod) {
            util_print("Not enough memory.\n");
            return 1;
        }
        currMod->path = ALLOC_TMP_MEM(size+1);
        if (!currMod->path) {
            util_print("Not enough memory.\n");
            return 1;
        }
        util_memmove(currMod->path, path, size);
        currMod->path[size] = 0;
        if (fsys_read_file(currMod->path, tmpAddr, &currMod->fileSize)) {
            util_print("Cannot read file '");
            util_print(currMod->path);
            util_print("'\n");
            return 1;
        }
        currMod->prop = prop;
        currMod->refCounter = 0;
        currMod->fileHeader = ALLOC_TMP_MEM(currMod->fileSize);
        if (!currMod->fileHeader) {
            util_print("Not enough memory.\n");
            return 1;
        }
        currMod->next = modules;
        modules = currMod;
    }

    /* no module available? Config error */
    if (!modules) {
        util_print("Configuration error: no modules!\n");
        return 1;
    }

    /* start interfaces register */
    if (StartInterfaces()) {
        return 1;
    }

    /* page align */
    currAddr = (char*)ALIGN_DATA_UP(currAddr, 0x1000);

    /* load modules */
    currMod = modules;
    do {
        if (LoadModule(currMod)) {
            return 1;
        }
        currMod = currMod->next;
    } while (currMod);

    /* set functions position */
    currInt = interfaces;
    do {
        Implementation *currImp;

        currImp = currInt->imps;
        do {
            Function *functions;

            functions = currImp->functions;
            for (i = 0; i < currInt->nFunctions; i++) {
                functions[i].code += (unsigned int)currImp->module->memStart;
            }
            currImp = currImp->next;
        } while (currImp);
        currInt = currInt->next;
    } while (currInt);

    /* do used functions reloc */
    currMod = modules;
    do {
        if (DoFunctionRelocs(currMod)) {
            return 1;
        }
        currMod = currMod->next;
    } while (currMod);
    return 0;
}

/* returns: 0 - success; 1 - error */
static __inline__ int StoreRegs() {
    Interface *currInt;
    Implementation *currImp;
    Module *currMod;
    unsigned int len;

    /* page align */
    currAddr = (char*)ALIGN_DATA_UP(currAddr, 0x1000);
    /* first put all strings */
    currInt = interfaces;
    do {
        currInt->namePos = currAddr;
        len = util_strlen(currInt->name)+1;
        util_memmove(currAddr, currInt->name, len);
        currAddr += len;
        currImp = currInt->imps;
        do {
            currImp->namePos = currAddr;
            len = util_strlen(currImp->name)+1;
            util_memmove(currAddr, currImp->name, len);
            currAddr += len;
            currImp = currImp->next;
        } while (currImp);
        currInt = currInt->next;
    } while (currInt);

    currMod = modules;
    do {
        currMod->pathPos = currAddr;
        len = util_strlen(currMod->path)+1;
        util_memmove(currAddr, currMod->path, len);
        currAddr += len;
        currMod = currMod->next;
    } while (currMod);

    /* align */
    currAddr = (char*)ALIGN_DATA_UP(currAddr, 4);

    /* modules register */
    currMod = modules;
    do {
        Module_Module *moduleInfo;

        currMod->pos = currAddr;
        currAddr += sizeof(SysModManager_Module);
        /* write next */
        if (currMod->next) {
            ((SysModManager_Module*)currMod->pos)->next = (SysModManager_Module*)(currAddr + SYSTEM_MEM);
        } else {
            ((SysModManager_Module*)currMod->pos)->next = NULL;
        }
        moduleInfo = Module_GetModuleInfo(currMod);
        ((SysModManager_Module*)currMod->pos)->phase0Start = moduleInfo->phase0Start;
        if ((unsigned int)moduleInfo->phase0Start != 0xffffffff) {
            ((SysModManager_Module*)currMod->pos)->phase0Start += (unsigned int)currMod->memStart;
        } else {
            /* no start function, set it as started */
            currMod->prop |= 0x010000;
        }
        ((SysModManager_Module*)currMod->pos)->phase1Start = moduleInfo->phase1Start;
        if ((unsigned int)moduleInfo->phase1Start != 0xffffffff) {
            ((SysModManager_Module*)currMod->pos)->phase1Start += (unsigned int)currMod->memStart;
        } else {
            /* no start function, set it as started */
            currMod->prop |= 0x020000;
        }
        ((SysModManager_Module*)currMod->pos)->shutdown = moduleInfo->shutdown;
        if ((unsigned int)moduleInfo->shutdown != 0xffffffff) {
            ((SysModManager_Module*)currMod->pos)->shutdown += (unsigned int)currMod->memStart;
        }
        ((SysModManager_Module*)currMod->pos)->prop = currMod->prop;
        ((SysModManager_Module*)currMod->pos)->memStart = (unsigned int)currMod->memStart;
        ((SysModManager_Module*)currMod->pos)->memSize = currMod->memSize;
        ((SysModManager_Module*)currMod->pos)->refCounter = currMod->refCounter;
        ((SysModManager_Module*)currMod->pos)->data = NULL;
        ((SysModManager_Module*)currMod->pos)->dataSync.data[0] = NULL;
        ((SysModManager_Module*)currMod->pos)->dataSync.data[1] = NULL;
        ((SysModManager_Module*)currMod->pos)->dataSync.data[2] = NULL;
        ((SysModManager_Module*)currMod->pos)->dataSync.data[3] = NULL;
        ((SysModManager_Module*)currMod->pos)->path = currMod->pathPos + SYSTEM_MEM;
        util_memmove(((SysModManager_Module*)currMod->pos)->md5, moduleInfo->md5, 16);
        currMod = currMod->next;
    } while (currMod);

    /* set interface position */
    currInt = interfaces;
    do {
        currInt->pos = currAddr;
        currAddr += sizeof(SysModManager_Interface);
        currImp = currInt->imps;
        do {
            currImp->pos = currAddr;
            currAddr += sizeof(SysModManager_Implementation);
            currImp->functionPos = currAddr;
            currAddr += currInt->nFunctions * sizeof(SysModManager_Function);
            currImp = currImp->next;
        } while (currImp);
        currInt = currInt->next;
    } while (currInt);

    /* now store it */
    currInt = interfaces;
    do {
        /* write avl tree node */
        if (currInt->node.left) {
            ((SysModManager_Interface*)currInt->pos)->node.left = (AVLTree_In_Node*)(((Interface*)currInt->node.left)->pos + SYSTEM_MEM);
        } else {
            ((SysModManager_Interface*)currInt->pos)->node.left = NULL;
        }
        if (currInt->node.right) {
            ((SysModManager_Interface*)currInt->pos)->node.right = (AVLTree_In_Node*)(((Interface*)currInt->node.right)->pos + SYSTEM_MEM);
        } else {
            ((SysModManager_Interface*)currInt->pos)->node.right = NULL;
        }
        ((SysModManager_Interface*)currInt->pos)->node.bal = currInt->node.bal;
        ((SysModManager_Interface*)currInt->pos)->node.cache = currInt->node.cache;
        ((SysModManager_Interface*)currInt->pos)->node.pad[0] = currInt->node.pad[0];
        ((SysModManager_Interface*)currInt->pos)->node.pad[1] = currInt->node.pad[1];
        ((SysModManager_Interface*)currInt->pos)->name = currInt->namePos + SYSTEM_MEM;
        ((SysModManager_Interface*)currInt->pos)->nFunctions = currInt->nFunctions;
        ((SysModManager_Interface*)currInt->pos)->imps = (SysModManager_Implementation*)(currInt->imps->pos + SYSTEM_MEM);
        currImp = currInt->imps;
        do {
            if (currImp->next) {
                ((SysModManager_Implementation*)currImp->pos)->next = (SysModManager_Implementation*)(currImp->next->pos + SYSTEM_MEM);
            } else {
                ((SysModManager_Implementation*)currImp->pos)->next = NULL;
            }
            ((SysModManager_Implementation*)currImp->pos)->module = (SysModManager_Module*)(currImp->module->pos + SYSTEM_MEM);
            ((SysModManager_Implementation*)currImp->pos)->interface = (SysModManager_Interface*)(currImp->interface->pos + SYSTEM_MEM);
            ((SysModManager_Implementation*)currImp->pos)->name = currImp->namePos + SYSTEM_MEM;
            ((SysModManager_Implementation*)currImp->pos)->interN = currImp->interN;
            ((SysModManager_Implementation*)currImp->pos)->impN = currImp->impN;
            ((SysModManager_Implementation*)currImp->pos)->functions = (SysModManager_Function*)(currImp->functionPos + SYSTEM_MEM);
            /* write all functions */
            util_memmove(currImp->functionPos, currImp->functions, currInt->nFunctions * sizeof(SysModManager_Function));
            currImp = currImp->next;
        } while (currImp);
        currInt = currInt->next;
    } while (currInt);
    return 0;
}

static __inline__ void StartPaging(unsigned int *pgDir, unsigned *pgTable) {
    unsigned int currPg, lastPg;

    util_memset(pgDir, 0, 0x2000);
    pgDir[0] = (unsigned int)pgTable | 0x3;
    /* identical mapping at 0xe0000000 */
    pgDir[((unsigned int)SYSTEM_MEM)>>22] = pgDir[0];
    currPg = 0x100000 >> 12;
    lastPg = ((unsigned int)ALIGN_DATA_UP(currAddr, 0x1000)) >> 12;
    for (; currPg <= lastPg; currPg++) {
        pgTable[currPg] = (currPg<<12) | 0x3;
    }
}

static char asm_below_1mb[] =
#include "asm_below_1mb.h"
;

extern multiboot_header_t multiboot_header;
void asm_StartSysModManager(char *stack, char *pgDir, StarterInfo *starterInfo, char *phase0Start);

void cmain(unsigned int magic, multiboot_info_t *mbi) {
    char *stack, *pgTable, *pgDir;
    StarterInfo *starterInfo;
    Interface *sysModInt;
    Module_Module *moduleInfo;

    /* copy code to switch to/from real mode to low mem */
    util_memmove((char*)LOW1MB_ASM, asm_below_1mb, sizeof(asm_below_1mb));

    /* say we're loading ModulOS */
    util_print("Loading ModulOS...\n");

    /* we start storing things from the end of starter and up */
    currAddr = (char*)ALIGN_DATA_UP(multiboot_header.load_end_addr, 0x1000);
    startAddr = currAddr;
    /* stack first */
    currAddr += 0x2000;
    stack = currAddr;

    /* page table and page dir next */
    pgDir = currAddr;
    currAddr += 0x1000;
    pgTable = currAddr;
    currAddr += 0x1000;

    /* mount file system */
    if (fsys_mount()) {
        goto end;
    }
    /* config file next */
    if (fsys_read_file("/system/config/system.cfg", currAddr, &cfgSize)) {
        goto end;
    }

    /* info to pass to SysModManager start function */
    starterInfo = (StarterInfo*)ALIGN_DATA_UP(currAddr + cfgSize, 4);
    starterInfo->cfgStart = currAddr + SYSTEM_MEM;
    starterInfo->cfgSize = cfgSize;
    cfgStart = currAddr;

    currAddr = (char*)starterInfo + sizeof(StarterInfo);

    if (LoadModules()) {
        goto end;
    }

    util_print("store\n");
    /* store modules and interfaces AVL tree */
    if (StoreRegs()) {
        goto end;
    }

    util_print("pag\n");
    /* setup paging */
    StartPaging((unsigned int*)pgDir, (unsigned int*)pgTable);

    sysModInt = (Interface*)AVLTree_In_Find(&interfacesRoot, "SysModManager");
    moduleInfo = Module_GetModuleInfo(sysModInt->imps->module);

    starterInfo->modules = modules->pos + SYSTEM_MEM;
    starterInfo->interfaces = ((Interface*)interfacesRoot.root)->pos + SYSTEM_MEM;
    starterInfo->pgDir = pgDir + SYSTEM_MEM;
    starterInfo->start = startAddr + SYSTEM_MEM;
    /* call Phase0Start from SysModManager */
    util_print("start\n");
    asm_StartSysModManager(stack + SYSTEM_MEM, pgDir, (StarterInfo*)((char*)starterInfo + SYSTEM_MEM), sysModInt->imps->module->memStart + (unsigned int)moduleInfo->phase0Start);

end:
    util_print("end\n");
    /* put it on a known state */
    asm_stop_floppy();
    /* something went wrong, stop! */
    asm_stop();
}
