/*
*****************************************
* This file implements the DMAController_i386 interface
*
* Version: 0.2.4
* Author: Luiz Henrique Shigunov
* E-mail: luizshigunov@users.sourceforge.net
* WEB: http://www.sourceforge.net/projects/modulos
* Last change: 08/06/2005
*
* See the interface specification to know what each function does.
*
* Copyright (C) 2000-2005, Luiz Henrique Shigunov
*
* This module is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2.1 of the License, or (at your option) any later version.
*
* This module is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this module; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*****************************************
*/

#include "common_def.h"

static TaskManager_RWSync regSync;
static int spin;
DMAController_i386_DMA_Channel dmaChannels[8];

void *DMAController_i386_Alloc(SysModManager_Module *module, unsigned int channel, unsigned int bufferSize) {
    int ret;
    void *error;

    error = ERROR_TO_PTR(E_BAD_VALUE);
    if (channel > DMAController_i386_MAX_DMA_CHANNEL)
        goto errorEnd;
    TaskManager_LockWrite(&regSync, 0);
    error = ERROR_TO_PTR(E_IN_USE);
    if (dmaChannels[channel].module)
        goto errorSync;
    error = ERROR_TO_PTR(E_BAD_VALUE);
    if ((channel < 4 && bufferSize > 16) || (channel > 4 && bufferSize > 32))
        goto errorSync;
    if (!dmaChannels[channel].logAddr || dmaChannels[channel].size < bufferSize) {
        if (dmaChannels[channel].logAddr && dmaChannels[channel].size < bufferSize) {
            ret = MemManager_i386_FreePages(dmaChannels[channel].logAddr, dmaChannels[channel].size, 0);
            ASSERT_ERROR(ret);
            dmaChannels[channel].logAddr = NULL;
        }
        dmaChannels[channel].size = bufferSize;
        dmaChannels[channel].physAddr = MemManager_i386_AllocPhysPages(bufferSize, MemManager_i386_BLOCK_PAGE+MemManager_i386_LESS_1MB+MemManager_i386_BOUNDARY_64KB);
        if (IS_PTR_ERROR(dmaChannels[channel].physAddr)) {
            error = (void*)dmaChannels[channel].physAddr;
            goto errorSync;
        }
        dmaChannels[channel].logAddr = MemManager_i386_MapPhysPages(NULL, dmaChannels[channel].physAddr, bufferSize, MemManager_i386_RW_PAGE+MemManager_i386_ALLOC_PAGES);
        if (IS_PTR_ERROR(dmaChannels[channel].logAddr)) {
            ret = MemManager_i386_FreePhysPages(dmaChannels[channel].physAddr, bufferSize);
            ASSERT_ERROR(ret);
            error = dmaChannels[channel].logAddr;
            dmaChannels[channel].logAddr = NULL;
            goto errorSync;
        }
    }
    dmaChannels[channel].module = module;
    error = dmaChannels[channel].logAddr;
errorSync:
    TaskManager_UnlockRW(&regSync, 0);
errorEnd:
    return error;
}

int DMAController_i386_Free(SysModManager_Module *module, unsigned int channel) {
	if (channel > DMAController_i386_MAX_DMA_CHANNEL)
		return E_BAD_VALUE;
	TaskManager_LockWrite(&regSync, 0);
	if (dmaChannels[channel].module != module) {
		TaskManager_UnlockRW(&regSync, 0);
		return E_NOT_IN_USE;
	}
	dmaChannels[channel].module = NULL;
	TaskManager_UnlockRW(&regSync, 0);
	/* disable channel */
	TaskManager_SpinLock(&spin);
	if (channel > 3)
		__asm__ __volatile__ (
			"outb %%al, $0xde\n"
			:: "a" (1 << (channel - 4)));
	else
		__asm__ __volatile__ (
			"outb %%al, $0xf\n"
			:: "a" (1 << channel));
	TaskManager_SpinUnlock(&spin);
	return E_OK;
}

int DMAController_i386_Setup(SysModManager_Module *module, unsigned int channel, unsigned int bufferOffset, unsigned int size, int prop) {
	char addrp, pagep, countp, modep, comp, ffp, maskp, mode, chan;

	if (channel > DMAController_i386_MAX_DMA_CHANNEL)
		return E_BAD_VALUE;
	if (dmaChannels[channel].module != module)
		return E_NOT_IN_USE;
	if ((bufferOffset+size) > (dmaChannels[channel].size << 12))
		return E_BAD_VALUE;
	if (((prop & 0xc0)==0xc0) || ((prop & 0x0c)==0x0c))
		return E_BAD_VALUE;
	mode = (prop & 0xcc);
	switch (channel) {
		case 0:
			addrp = 0;
			countp = 1;
			pagep = 0x87;
			break;
		case 1:
			addrp = 2;
			countp = 3;
			pagep = 0x83;
			break;
		case 2:
			addrp = 4;
			countp = 5;
			pagep = 0x81;
			break;
		case 3:
			addrp = 6;
			countp = 7;
			pagep = 0x82;
			break;
		case 5:
			addrp = 0xc2;
			countp = 0xc3;
			pagep = 0x8b;
			break;
		case 6:
			addrp = 0xc4;
			countp = 0xc5;
			pagep = 0x89;
			break;
		default:
			addrp = 0xc6;
			countp = 0xc7;
			pagep = 0x8a;
	}
	if (channel > 3) {
		comp = 0xd0;
		modep = 0xd6;
		ffp = 0xd8;
		maskp = 0xd4;
		chan = channel - 4;
		bufferOffset = bufferOffset >> 1;
		size = size >> 1;
	} else {
		comp = 8;
		modep = 0xb;
		ffp = 0xc;
		maskp = 0xa;
		chan = channel;
	}
	mode |= chan;
	size -= 1;
	TaskManager_SpinLock(&spin);
	__asm__ __volatile__ (
		"xorl %%edx, %%edx" ::: "edx");
	__asm__ __volatile__ (
		"outb %%al, %%dx\n"     /* send mode and type of transfer */
		::"d" (modep), "a" (mode));
	__asm__ __volatile__ (
		"outb %%al, %%dx\n"     /* reset flip-flop */
		::"d" (ffp));
	__asm__ __volatile__ (
		"outb %%al, %%dx\n"     /* send low part of address */
		"shrl	$8, %%eax\n"
		"outb %%al, %%dx\n"     /* send high part of address */
		"shrl	$8, %%eax\n"
		"movb %%cl, %%dl\n"
		"outb %%al, %%dx\n"     /* send page address */
		::"d" (addrp), "c" (pagep), "a" (dmaChannels[channel].physAddr + bufferOffset));
	__asm__ __volatile__ (
		"outb %%al, %%dx\n"     /* reset flip-flop */
		::"d" (ffp));
	__asm__ __volatile__ (
		"outb %%al, %%dx\n"
		"shrl $8, %%eax\n"      /* send counter */
		"outb %%al, %%dx\n"
		:: "d" (countp), "a" (size));
	__asm__ __volatile__ (
		"outb %%al, %%dx\n"     /* enable channel */
		:: "d" (maskp), "a" (chan));
	TaskManager_SpinUnlock(&spin);
	return E_OK;
}
