/*
*****************************************
* This file implements the IdeController interface for a generic IDE controller
*
* Version: 0.1.0
* Author: Luiz Henrique Shigunov
* E-mail: luizshigunov@users.sourceforge.net
* WEB: http://www.sourceforge.net/projects/modulos
* Last change: 28/12/2003
*
* See the interface specification to know what each function does.
*
* Copyright (C) 2003, Luiz Henrique Shigunov
*
* This module is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2.1 of the License, or (at your option) any later version.
*
* This module is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this module; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*****************************************
*/

/*
 * Some code used from Linux kernel 2.4.21
*/


#include <stddef.h>
#include <debug.h>
#include <errormanager.h>
#include <idecontroller.h>
#include <intmanager_i386.h>
#include <ioportmanager.h>
#include <ioportmanager_i386.h>
#include <sysmodmanager.h>
#include <taskmanager.h>

/* status reg */
#define STATUS_ERROR     0x01
#define STATUS_INDEX     0x02
#define STATUS_CORR      0x04
#define STATUS_DRQ       0x08
#define STATUS_SEEK      0x10
#define STATUS_WRITE_ERR 0x20
#define STATUS_READY     0x40
#define STATUS_BUSY      0x80

/* io ports */
#define NR_IOPORTS 9
#define DATA_OFF        7
#define ERROR_OFF       IdeController_FEATURE_PORT
#define FEATURE_OFF     IdeController_FEATURE_PORT
#define SECTOR_C_OFF    IdeController_SECTOR_COUNT_PORT
#define SECTOR_N_OFF    IdeController_SECTOR_NUMBER_PORT
#define LOW_CYL_OFF     IdeController_LOW_CYL_PORT
#define HIGH_CYL_OFF    IdeController_HIGH_CYL_PORT
#define DRIVE_OFF       IdeController_DRIVE_PORT
#define STATUS_OFF      IdeController_COMMAND_PORT
#define COMMAND_OFF     IdeController_COMMAND_PORT
#define CONTROL_OFF     8
#define ALT_STATUS_OFF  8

typedef union {
    unsigned all                : 8;    /* all of the bits together */
    struct {
        unsigned head           : 4;    /* always zeros here */
        unsigned unit           : 1;    /* drive select number, 0 or 1 */
        unsigned bit5           : 1;    /* always 1 */
        unsigned lba            : 1;    /* using LBA instead of CHS */
        unsigned bit7           : 1;    /* always 1 */
    } b;
} SelectReg;

typedef struct {
    SelectReg select;         /* basic drive/head select reg value */
    unsigned char ctl;        /* "normal" value for IDE_CONTROL_REG */
    unsigned int locked;
} Drive;

typedef struct _Interface Interface;
struct _Interface {
    TaskManager_Semaphore mutex;
    TaskManager_Semaphore intMutex;
    Drive drives[2];
    unsigned short ioports[NR_IOPORTS];
};

SysModManager_Module *ID;
Interface interfaces[2];

int ServeInterrupt(Interface*);

int IdeController_GetDrive(unsigned int interface, unsigned int drive) {
    if (interface > 1)
        return E_NO_SUCH_INTERFACE;
    if (drive > 2)
        return E_NO_SUCH_DRIVE;

    ASSERT_RET(interfaces[interface].mutex.count < 1, E_BUG);

    if (interfaces[interface].drives[drive].locked)
        return E_BUSY;
    interfaces[interface].drives[drive].locked = 1;
    return E_OK;
}

int IdeController_LockInterface(unsigned int interface) {
    if (interface > 1)
        return E_NO_SUCH_INTERFACE;

    ASSERT_RET(interfaces[interface].mutex.count == 1, E_BUG);

    /* this thread can wait for a long time, so set it interruptible */
    return TaskManager_P(&interfaces[interface].mutex, TaskManager_THREAD_INT, 0);
}

int IdeController_PutDrive(unsigned int interface, unsigned int drive) {
    if (interface > 1)
        return E_NO_SUCH_INTERFACE;
    if (drive > 2)
        return E_NO_SUCH_DRIVE;

    ASSERT_RET(interfaces[interface].mutex.count < 1, E_BUG);

    interfaces[interface].drives[drive].locked = 0;
    return E_OK;
}

int IdeController_ReadData(unsigned int interface, void *buf, unsigned int size) {
    if (interface > 1)
        return E_NO_SUCH_INTERFACE;

    ASSERT_RET(buf, E_BUG);
    ASSERT_RET(interfaces[interface].mutex.count < 1, E_BUG);

    IOPortManager_i386_InShorts(interfaces[interface].ioports[DATA_OFF], buf, size >> 1);
    return E_OK;
}

int IdeController_SendRequest(IdeController_Request *rq) {
    int error, i, mode;

    ASSERT_RET(rq, E_BUG);

    if (rq->version != 1)
        return E_INVALID_VERSION;
    if (rq->interface > 1)
        return E_NO_SUCH_INTERFACE;
    if (rq->drive > 2)
        return E_NO_SUCH_DRIVE;

    ASSERT_RET(interfaces[rq->interface].mutex.count < 1, E_BUG);

    /* select drive */
    IOPortManager_i386_OutChar(interfaces[rq->interface].ioports[DRIVE_OFF],
                               interfaces[rq->interface].drives[rq->drive].select.all);
    if ((rq->flags & IdeController_MODE_MASK) == IdeController_PROBE_MODE) {
        /* check whether select worked */
        TaskManager_Wait(NULL, 50);
        if (IOPortManager_i386_InChar(interfaces[rq->interface].ioports[DRIVE_OFF]) !=
            interfaces[rq->interface].drives[rq->drive].select.all)
            return E_PROBE_FAILED;
    }
    interfaces[rq->interface].intMutex.count = 0;
    interfaces[rq->interface].intMutex.data = NULL;

    mode = rq->flags & IdeController_MODE_MASK;
    if (mode == IdeController_PROBE_MODE || mode == IdeController_START_MULTISECTOR || mode == IdeController_ONE_SECTOR) {
        /* write io ports */
        for (i = 0; i < rq->numIOs; i++) {
            unsigned char data;
            if (rq->IOs[i].port != IdeController_DRIVE_PORT)
                data = rq->IOs[i].data;
            else
                data = (rq->IOs[i].data & 0x0f) | interfaces[rq->interface].drives[rq->drive].select.all;
            IOPortManager_i386_OutChar(interfaces[rq->interface].ioports[rq->IOs[i].port], data);
        }
    }
    /* wait for int */
    error = TaskManager_P(&interfaces[rq->interface].intMutex, TaskManager_THREAD_INT, rq->timeout*1000);
    rq->status = IOPortManager_i386_InChar(interfaces[rq->interface].ioports[STATUS_OFF]);
    if (rq->status & STATUS_ERROR)
        rq->error = IOPortManager_i386_InChar(interfaces[rq->interface].ioports[ERROR_OFF]);
    else
        rq->error = 0;
    if (!error) {
        if ((rq->status & (STATUS_DRQ|STATUS_BUSY|STATUS_ERROR)) == STATUS_DRQ)
            error = E_OK;
        else
            error = E_IO_ERROR;
    }
    return error;
}

int IdeController_SetupDrive(const IdeController_Setup *set) {

    ASSERT_RET(set, E_BUG);

    if (set->version != 1)
        return E_INVALID_VERSION;
    if (set->interface > 1)
        return E_NO_SUCH_INTERFACE;
    if (set->drive > 2)
        return E_NO_SUCH_DRIVE;

    ASSERT_RET(interfaces[set->interface].mutex.count < 1, E_BUG);

    interfaces[set->interface].drives[set->drive].select.b.lba = set->flags & IdeController_SETUP_LBA;
    return E_OK;
}

int IdeController_UnlockInterface(unsigned int interface) {
    if (interface > 1)
        return E_NO_SUCH_INTERFACE;

    ASSERT_RET(interfaces[interface].mutex.count < 1, E_BUG);

    TaskManager_V(&interfaces[interface].mutex, 0);
    return E_OK;
}

int IdeController_WriteData(unsigned int interface, const void *buf, unsigned int size) {
    if (interface > 1)
        return E_NO_SUCH_INTERFACE;

    ASSERT_RET(buf, E_BUG);
    ASSERT_RET(interfaces[interface].mutex.count < 1, E_BUG);

    IOPortManager_i386_OutShorts(interfaces[interface].ioports[DATA_OFF], buf, size >> 1);
    return E_OK;
}

int Serve14Interrupt(IntManager_i386_Regs *regs) {
    return ServeInterrupt(&interfaces[0]);
}

int Serve15Interrupt(IntManager_i386_Regs *regs) {
    return ServeInterrupt(&interfaces[1]);
}

int ServeInterrupt(Interface *hwif) {
    unsigned char statusReg;

    /* our int? */
    statusReg = IOPortManager_i386_InChar(hwif->ioports[STATUS_OFF]);
    if ((statusReg & (STATUS_DRQ|STATUS_BUSY|STATUS_ERROR)) != STATUS_DRQ) {
        /* not our int ? */
        if (!(statusReg & (STATUS_DRQ|STATUS_ERROR)))
            return E_UNHANDLED_INT;
    }
    TaskManager_V(&hwif->intMutex, TaskManager_RESUME_KEEP_CPU);
    return E_HANDLED_INT;
}

void SetupInterface(unsigned int interface, unsigned short basePort) {
    interfaces[interface].mutex.count = 1;
    interfaces[interface].mutex.data  = NULL;
    interfaces[interface].drives[0].locked        = 0;
    interfaces[interface].drives[0].ctl           = 0x08;
    interfaces[interface].drives[0].select.all    = 0;
    interfaces[interface].drives[0].select.b.bit5 = 1;
    interfaces[interface].drives[0].select.b.bit7 = 1;
    interfaces[interface].drives[1].locked        = 0;
    interfaces[interface].drives[1].ctl           = 0x08;
    interfaces[interface].drives[1].select.all    = 0;
    interfaces[interface].drives[1].select.b.unit = 1;
    interfaces[interface].drives[1].select.b.bit5 = 1;
    interfaces[interface].drives[1].select.b.bit7 = 1;
    interfaces[interface].ioports[DATA_OFF]       = basePort;
    interfaces[interface].ioports[ERROR_OFF]      = basePort + 1;
    interfaces[interface].ioports[SECTOR_C_OFF]   = basePort + 2;
    interfaces[interface].ioports[SECTOR_N_OFF]   = basePort + 3;
    interfaces[interface].ioports[LOW_CYL_OFF]    = basePort + 4;
    interfaces[interface].ioports[HIGH_CYL_OFF]   = basePort + 5;
    interfaces[interface].ioports[DRIVE_OFF]      = basePort + 6;
    interfaces[interface].ioports[COMMAND_OFF]    = basePort + 7;
    interfaces[interface].ioports[CONTROL_OFF]    = basePort + 0x206;
    /* enable interrupts */
    IOPortManager_i386_OutChar(interfaces[interface].ioports[CONTROL_OFF], interfaces[interface].drives[0].ctl);
}

int Shutdown(void) {
    int ret;

    ret = IOPortManager_Free(ID, 0x1f0);
    ASSERT_ERROR(ret);
    ret = IOPortManager_Free(ID, 0x3f6);
    ASSERT_ERROR(ret);
    ret = IOPortManager_Free(ID, 0x170);
    ASSERT_ERROR(ret);
    ret = IOPortManager_Free(ID, 0x376);
    ASSERT_ERROR(ret);
    ret = IntManager_i386_Free(ID, 14, IntManager_i386_HARD_INT);
    ASSERT_ERROR(ret);
    ret = IntManager_i386_Free(ID, 15, IntManager_i386_HARD_INT);
    ASSERT_ERROR(ret);
    return 0;
}

int Phase1Start(SysModManager_Module *modID) {
    int ret;

    ID = modID;
    /* start all used modules */
    if ((ret = SysModManager_StartModule((unsigned int)IntManager_i386_Catch)) ||
        (ret = SysModManager_StartModule((unsigned int)IOPortManager_Alloc)) ||
        (ret = SysModManager_StartModule((unsigned int)TaskManager_P))) {
        DEBUG_MSG_PRE(ret);
        goto errorReturn;
    }
    if ((ret = IOPortManager_Alloc(modID, 0x1f0, 0x1f7, 0))) {
        DEBUG_MSG_PRE(ret);
        goto errorReturn;
    }
    if ((ret = IOPortManager_Alloc(modID, 0x3f6, 0x3f6, 0))) {
        DEBUG_MSG_PRE(ret);
        goto errorFreeIO1f0;
    }
    if ((ret = IOPortManager_Alloc(modID, 0x170, 0x177, 0))) {
        DEBUG_MSG_PRE(ret);
        goto errorFreeIO3f6;
    }
    if ((ret = IOPortManager_Alloc(modID, 0x376, 0x376, 0))) {
        DEBUG_MSG_PRE(ret);
        goto errorFreeIO170;
    }
    if ((ret = IntManager_i386_Catch(modID, 14, Serve14Interrupt, IntManager_i386_HARD_INT+IntManager_i386_LAST_INT))) {
        DEBUG_MSG_PRE(ret);
        goto errorFreeIO376;
    }
    if ((ret = IntManager_i386_Catch(modID, 15, Serve15Interrupt, IntManager_i386_HARD_INT+IntManager_i386_LAST_INT))) {
        DEBUG_MSG_PRE(ret);
        goto errorFreeInt14;
    }
    /* setup interfaces */
    SetupInterface(0, 0x1f0);
    SetupInterface(1, 0x170);
    return 0;

errorFreeInt14:
    ret = IntManager_i386_Free(ID, 14, IntManager_i386_HARD_INT);
    ASSERT_ERROR(ret);
errorFreeIO376:
    ret = IOPortManager_Free(modID, 0x376);
    ASSERT_ERROR(ret);
errorFreeIO170:
    ret = IOPortManager_Free(modID, 0x170);
    ASSERT_ERROR(ret);
errorFreeIO3f6:
    ret = IOPortManager_Free(modID, 0x3f6);
    ASSERT_ERROR(ret);
errorFreeIO1f0:
    ret = IOPortManager_Free(modID, 0x1f0);
    ASSERT_ERROR(ret);
errorReturn:
    return 1;
}
