/* libavl - manipulates AVL trees.
   Copyright (C) 1998, 1999 Free Software Foundation, Inc.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.

   The author may be contacted at <pfaffben@pilot.msu.edu> on the
   Internet, or as Ben Pfaff, 12167 Airport Rd, DeWitt MI 48820, USA
   through more mundane means. */

/*
*****************************************
* This file implements the AVLTree_In interface
*
* Version: 0.2.2
* Author: Luiz Henrique Shigunov
* E-mail: luizshigunov@users.sourceforge.net
* WEB: http://www.sourceforge.net/projects/modulos
* Last change: 24/06/2005
*
* See the interface specification to know what each function does.
*
* Copyright (C) 2000-2005, Luiz Henrique Shigunov
*****************************************
*/

#include <stddef.h>
#include "errormanager.h"

typedef struct {
    struct _AVLNode_In *root;
    int (*cmpFunc)(const void*,const void*);
    int (*cmpAddFunc)(const void*,const void*);
    unsigned int count;
} AVLTree_In;

typedef struct _AVLNode_In {
    struct _AVLNode_In *link[2];
    signed char bal;
    char cache;
    char pad[2];
} AVLNode_In;

AVLNode_In *FindAdd(AVLTree_In*,AVLNode_In*);

int Add(AVLTree_In *tree, AVLNode_In *info) {
    AVLNode_In *p;

    if (tree == NULL)
        return E_BAD_VALUE;
    p = FindAdd(tree, info);
    if (IS_PTR_ERROR(p))
        return PTR_TO_ERROR(p);
    if (p != info)
        return E_IN_USE;
    return E_OK;
}

unsigned int Count(const AVLTree_In *tree) {
    if (!tree)
        return 0;
    return tree->count;
}

AVLNode_In *Find(const AVLTree_In *tree, const void *info) {
    AVLNode_In *p;

    if (!tree)
        return ERROR_TO_PTR(E_BAD_VALUE);
    for (p = tree->root; p; ) {
        int diff = tree->cmpFunc(info, p);
        if (diff < 0)
            p = p->link[0];
        else if (diff > 0)
            p = p->link[1];
        else
            return p;
    }
    return ERROR_TO_PTR(E_NOT_FOUND);
}

AVLNode_In *FindAdd(AVLTree_In *tree, AVLNode_In *info) {
  /* Uses Knuth's Algorithm 6.2.3A (balanced tree search and
     insertion), but caches results of comparisons.  In empirical
     tests this eliminates about 25% of the comparisons seen under
     random insertions.  */

  /* A1. */
    AVLNode_In *t;
    AVLNode_In *s, *p, *q, *r, *found;

    if (!tree)
        return ERROR_TO_PTR(E_BAD_VALUE);
    t = (AVLNode_In *)&tree->root;
    s = p = tree->root;

    if (!s) {
        tree->count = 1;
        tree->root = info;
        info->link[0] = info->link[1] = NULL;
        info->bal = 0;
        return info;
    }
for (;;) {
        /* A2. */
        int diff = tree->cmpAddFunc(info, p);
        /* A3. */
        if (diff < 0) {
            p->cache = 0;
            q = p->link[0];
            if (!q) {
                p->link[0] = q = info;
                break;
            }
        /* A4. */
        } else if (diff > 0) {
            p->cache = 1;
            q = p->link[1];
            if (!q) {
                p->link[1] = q = info;
                break;
            }
        } else {
            /* A2. */
            return p;
        }
        /* A3, A4. */
        if (q->bal != 0)
            t = p, s = q;
        p = q;
    }
    /* A5. */
    tree->count++;
    q->link[0] = q->link[1] = NULL;
    q->bal = 0;
    found = q;
    /* A6. */
    r = p = s->link[(int) s->cache];
    while (p != q) {
        p->bal = p->cache * 2 - 1;
        p = p->link[(int) p->cache];
    }
    /* A7. */
    if (!s->cache) {
        /* a = -1. */
        if (!s->bal) {
            s->bal = -1;
            return found;
        } else if (s->bal == +1) {
            s->bal = 0;
            return found;
        }
        if (r->bal == -1) {
            /* A8. */
            p = r;
            s->link[0] = r->link[1];
            r->link[1] = s;
            s->bal = r->bal = 0;
        } else {
            /* A9. */
            p = r->link[1];
            r->link[1] = p->link[0];
            p->link[0] = r;
            s->link[0] = p->link[1];
            p->link[1] = s;
            if (p->bal == -1)
                s->bal = 1, r->bal = 0;
            else if (p->bal == 0)
                s->bal = r->bal = 0;
            else
                s->bal = 0, r->bal = -1;
            p->bal = 0;
        }
    } else {
        /* a == +1. */
        if (!s->bal) {
            s->bal = 1;
            return found;
        } else if (s->bal == -1) {
            s->bal = 0;
            return found;
        }
        if (r->bal == +1) {
            /* A8. */
            p = r;
            s->link[1] = r->link[0];
            r->link[0] = s;
            s->bal = r->bal = 0;
        } else {
            /* A9. */
            p = r->link[0];
            r->link[0] = p->link[1];
            p->link[1] = r;
            s->link[1] = p->link[0];
            p->link[0] = s;
            if (p->bal == +1)
                s->bal = -1, r->bal = 0;
            else if (p->bal == 0)
                s->bal = r->bal = 0;
            else
                s->bal = 0, r->bal = 1;
            p->bal = 0;
        }
    }
    /* A10. */
    if (t != (AVLNode_In *)&tree->root && s == t->link[1])
        t->link[1] = p;
    else
        t->link[0] = p;
    return found;
}

AVLNode_In *FindClose(const AVLTree_In *tree, const void *info) {
    AVLNode_In *p;

    if (!tree)
        return ERROR_TO_PTR(E_BAD_VALUE);
    p = tree->root;
    if (!p)
        return ERROR_TO_PTR(E_EMPTY);
    for (;;) {
        int diff = tree->cmpFunc(info, p);
        int t;

        if (diff < 0)
            t = 0;
        else if (diff > 0)
            t = 1;
        else
            return p;
        if (p->link[t])
            p = p->link[t];
        else
            return p;
    }
}

/* Uses Knuth's Algorithm 2.3.1T as modified in exercise 13
(postorder traversal). */
int Free(AVLTree_In *tree, void (*freeFunc)(AVLNode_In*)) {
    /* T1. */
    AVLNode_In *an[32]; /* Stack A: nodes. */
    char ab[32];		 /* Stack A: bits. */
    int ap = 0;			 /* Stack A: height. */
    AVLNode_In *p;

    if (!tree)
        return E_BAD_VALUE;
    p = tree->root;
    for (;;) {
        /* T2. */
        while (p) {
            /* T3. */
            ab[ap] = 0;
            an[ap++] = p;
            p = p->link[0];
        }
        /* T4. */
        for (;;) {
            if (!ap)
                return E_OK;
            p = an[--ap];
            if (!ab[ap]) {
                ab[ap++] = 1;
                p = p->link[1];
                break;
            }
            freeFunc(p);
        }
    }
    return E_OK;
}

int IsEmpty(const AVLTree_In *tree) {
    if (!tree)
        return E_BAD_VALUE;
    if (!tree->root)
        return E_EMPTY;
    return E_NOT_EMPTY;
}

AVLNode_In *Remove(AVLTree_In *tree, const void *info) {
  /* Uses my Algorithm D, which can be found at
     http://www.msu.edu/user/pfaffben/avl.  Algorithm D is based on
     Knuth's Algorithm 6.2.2D (Tree deletion) and 6.2.3A (Balanced
     tree search and insertion), as well as the notes on pages 465-466
     of Vol. 3. */

    /* D1. */
    AVLNode_In *pa[32];	/* Stack P: Nodes. */
    char a[32];				/* Stack P: Bits. */
    int k = 1;				/* Stack P: Pointer. */

    AVLNode_In **q;
    AVLNode_In *p, *removed;

    if (!tree)
        return ERROR_TO_PTR(E_BAD_VALUE);
    a[0] = 0;
    pa[0] = (AVLNode_In *)&tree->root;
    p = tree->root;
    for (;;) {
        /* D2. */
        int diff;
        if (!p)
            return ERROR_TO_PTR(E_NOT_FOUND);
        diff = tree->cmpFunc(info, p);
        if (!diff)
            break;
        /* D3, D4. */
        pa[k] = p;
        if (diff < 0) {
            p = p->link[0];
            a[k] = 0;
        } else {
            p = p->link[1];
            a[k] = 1;
        }
        k++;
    }
    removed = p;
    tree->count--;
    /* D5. */
    q = &pa[k - 1]->link[(int) a[k - 1]];
    if (!p->link[1]) {
        *q = p->link[0];
        if (*q)
            (*q)->bal = 0;
    } else {
        /* D6. */
        AVLNode_In *r = p->link[1];
        if (!r->link[0]) {
            r->link[0] = p->link[0];
            *q = r;
            r->bal = p->bal;
            a[k] = 1;
            pa[k++] = r;
        } else {
            /* D7. */
            AVLNode_In *s = r->link[0];
            int l = k++;
            a[k] = 0;
            pa[k++] = r;
            /* D8. */
            while (s->link[0]) {
                r = s;
                s = r->link[0];
                a[k] = 0;
                pa[k++] = r;
            }
            /* D9. */
            a[l] = 1;
            pa[l] = s;
            s->link[0] = p->link[0];
            r->link[0] = s->link[1];
            s->link[1] = p->link[1];
            s->bal = p->bal;
            *q = s;
        }
    }
    /* D10. */
    while (--k) {
        AVLNode_In *s = pa[k], *r;
        if (!a[k]) {
            /* D10. */
            if (s->bal == -1) {
                s->bal = 0;
                continue;
            } else if (!s->bal) {
                s->bal = 1;
                break;
            }
            r = s->link[1];
            if (!r->bal) {
                /* D11. */
                s->link[1] = r->link[0];
                r->link[0] = s;
                r->bal = -1;
                pa[k - 1]->link[(int) a[k - 1]] = r;
                break;
            } else if (r->bal == +1) {
                /* D12. */
                s->link[1] = r->link[0];
                r->link[0] = s;
                s->bal = r->bal = 0;
                pa[k - 1]->link[(int) a[k - 1]] = r;
            } else {
                /* D13. */
                p = r->link[0];
                r->link[0] = p->link[1];
                p->link[1] = r;
                s->link[1] = p->link[0];
                p->link[0] = s;
                if (p->bal == +1)
                    s->bal = -1, r->bal = 0;
                else if (!p->bal)
                    s->bal = r->bal = 0;
                else
                    s->bal = 0, r->bal = +1;
                p->bal = 0;
                pa[k - 1]->link[(int) a[k - 1]] = p;
            }
        } else {
            /* D10. */
            if (s->bal == +1) {
                s->bal = 0;
                continue;
            } else if (!s->bal) {
                s->bal = -1;
                break;
            }
            r = s->link[0];
            if (!r || !r->bal) {
                /* D11. */
                s->link[0] = r->link[1];
                r->link[1] = s;
                r->bal = 1;
                pa[k - 1]->link[(int) a[k - 1]] = r;
                break;
            } else if (r->bal == -1) {
                /* D12. */
                s->link[0] = r->link[1];
                r->link[1] = s;
                s->bal = r->bal = 0;
                pa[k - 1]->link[(int) a[k - 1]] = r;
            } else if (r->bal == +1) {
                /* D13. */
                p = r->link[1];
                r->link[1] = p->link[0];
                p->link[0] = r;
                s->link[0] = p->link[1];
                p->link[1] = s;
                if (p->bal == -1)
                    s->bal = 1, r->bal = 0;
                else if (!p->bal)
                    s->bal = r->bal = 0;
                else
                    s->bal = 0, r->bal = -1;
                p->bal = 0;
                pa[k - 1]->link[(int) a[k - 1]] = p;
            }
        }
    }
    return removed;
}

int Walk(const AVLTree_In *tree,void (*walkFunc)(void *info, void *param1, void *param2), void *param1, void *param2) {
    /* Uses Knuth's algorithm 2.3.1T (inorder traversal).*/
    /* T1. */
    const AVLNode_In *an[32];		/* Stack A: nodes. */
    const AVLNode_In **ap = an;	/* Stack A: stack pointer. */
    const AVLNode_In *p = tree->root;
    if (!tree || !walkFunc)
        return E_BAD_VALUE;
    for (;;) {
        /* T2. */
        while (p) {
            /* T3. */
            *ap++ = p;
            p = p->link[0];
        }
        /* T4. */
        if (ap == an)
            return E_OK;
        p = *--ap;
        /* T5. */
        walkFunc((AVLNode_In*)p, param1, param2);
        p = p->link[1];
    }
}
