/*
*****************************************
* This file implements the FileSystem interface for ext2fs
*
* Version: 0.3.2
* Author: Luiz Henrique Shigunov
* E-mail: luizshigunov@users.sourceforge.net
* WEB: http://www.sourceforge.net/projects/modulos
* Last change: 24/06/2005
*
* See the interface specification to know what each function does.
*
* Copyright (C) 2000-2005, Luiz Henrique Shigunov
*
* This module is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2.1 of the License, or (at your option) any later version.
*
* This module is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this module; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*****************************************
*/

#include <stddef.h>
#include <blockbuffer.h>
#include <debug.h>
#include <errormanager.h>
#include <filesystem.h>
#include <memman.h>
#include <memmanager.h>
#include <stringman.h>
#include <sysmodmanager.h>
#include <taskmanager.h>
#include <userspacemanager.h>

#define HASHBITS 10
#define HASHMASK 0x3ff

#define IMCOMPAT_FEATURES_MASK ~0x0002

#define IS_LINK(mode) ((mode & 0170000)==0120000)
#define IS_DIR(mode)  ((mode & 0170000)==0040000)
#define IS_FILE(mode) ((mode & 0170000)==0100000)

/*
 * Copyright (C) 1992, 1993, 1994, 1995
 * Remy Card (card@masi.ibp.fr)
 * Laboratoire MASI - Institut Blaise Pascal
 * Universite Pierre et Marie Curie (Paris VI)
 *
 *  from
 *
 *  linux/include/linux/minix_fs.h
 *
 *  Copyright (C) 1991, 1992  Linus Torvalds
*/
typedef struct {
    unsigned int    inodes_count;       /* Inodes count */
    unsigned int    blocks_count;       /* Blocks count */
    unsigned int    r_blocks_count;     /* Reserved blocks count */
    unsigned int    free_blocks_count;  /* Free blocks count */
    unsigned int    free_inodes_count;  /* Free inodes count */
    unsigned int    first_data_block;	/* First Data Block */
    unsigned int    log_block_size;     /* Block size */
    unsigned int    log_frag_size;      /* Fragment size */
    unsigned int    blocks_per_group;   /* # Blocks per group */
    unsigned int    frags_per_group;    /* # Fragments per group */
    unsigned int    inodes_per_group;   /* # Inodes per group */
    unsigned int    mtime;              /* Mount time */
    unsigned int    wtime;              /* Write time */
    unsigned short  mnt_count;          /* Mount count */
    short           max_mnt_count;      /* Maximal mount count */
    unsigned short  magic;              /* Magic signature */
    unsigned short  state;              /* File system state */
    unsigned short  errors;             /* Behaviour when detecting errors */
    unsigned short  minor_rev_level;    /* minor revision level */
    unsigned int    lastcheck;          /* time of last check */
    unsigned int    checkinterval;      /* max. time between checks */
    unsigned int    creator_os;         /* OS */
    unsigned int    rev_level;          /* Revision level */
    unsigned short  def_resuid;         /* Default uid for reserved blocks */
    unsigned short  def_resgid;         /* Default gid for reserved blocks */
    /*
     * These fields are for EXT2_DYNAMIC_REV superblocks only.
     *
     * Note: the difference between the compatible feature set and
     * the incompatible feature set is that if there is a bit set
     * in the incompatible feature set that the kernel doesn't
     * know about, it should refuse to mount the filesystem.
     *
     * e2fsck's requirements are more strict; if it doesn't know
     * about a feature in either the compatible or incompatible
     * feature set, it must abort and not try to meddle with
     * things it doesn't understand...
     */
    unsigned int    first_ino;          /* First non-reserved inode */
    unsigned short  inode_size;         /* size of inode structure */
    unsigned short  block_group_nr;     /* block group # of this superblock */
    unsigned int    feature_compat;     /* compatible feature set */
    unsigned int    feature_incompat;   /* incompatible feature set */
    unsigned int    feature_ro_compat;  /* readonly-compatible feature set */
    unsigned char   uuid[16];           /* 128-bit uuid for volume */
    char            volume_name[16];    /* volume name */
    char            last_mounted[64];   /* directory where last mounted */
    unsigned int    algorithm_usage_bitmap; /* For compression */
    /*
     * Performance hints.  Directory preallocation should only
     * happen if the EXT2_COMPAT_PREALLOC flag is on.
     */
    unsigned char   prealloc_blocks;    /* Nr of blocks to try to preallocate*/
    unsigned char   prealloc_dir_blocks;/* Nr to preallocate for dirs */
    unsigned short  padding1;
    unsigned int    reserved[204];      /* Padding to the end of the block */
} SB;

typedef struct {
    unsigned int    bg_block_bitmap;        /* Blocks bitmap block */
    unsigned int    bg_inode_bitmap;        /* Inodes bitmap block */
    unsigned int    bg_inode_table;         /* Inodes table block */
    unsigned short  bg_free_blocks_count;   /* Free blocks count */
    unsigned short  bg_free_inodes_count;   /* Free inodes count */
    unsigned short  bg_used_dirs_count;     /* Directories count */
    unsigned short  bg_pad;
    unsigned int    bg_reserved[3];
} GroupDesc;

typedef struct {
    unsigned short  i_mode;     /* File mode */
    unsigned short  i_uid;      /* Owner Uid */
    unsigned int    i_size;     /* Size in bytes */
    unsigned int    i_atime;    /* Access time */
    unsigned int    i_ctime;    /* Creation time */
    unsigned int    i_mtime;    /* Modification time */
    unsigned int    i_dtime;    /* Deletion Time */
    unsigned short  i_gid;      /* Group Id */
    unsigned short  i_links_count;  /* Links count */
    unsigned int    i_blocks;   /* Blocks count */
    unsigned int    i_flags;    /* File flags */
    union {
        struct {
            unsigned int  l_i_reserved1;
        } linux1;
        struct {
            unsigned int  h_i_translator;
        } hurd1;
        struct {
            unsigned int  m_i_reserved1;
        } masix1;
    } osd1;                     /* OS dependent 1 */
    unsigned int    i_block[15];/* Pointers to blocks */
    unsigned int    i_version;  /* File version (for NFS) */
    unsigned int    i_file_acl; /* File ACL */
    unsigned int    i_dir_acl;  /* Directory ACL */
    unsigned int    i_faddr;    /* Fragment address */
    union {
        struct {
            unsigned char   l_i_frag;   /* Fragment number */
            unsigned char   l_i_fsize;  /* Fragment size */
            unsigned short  i_pad1;
            unsigned int    l_i_reserved2[2];
        } linux2;
        struct {
            unsigned char   h_i_frag;   /* Fragment number */
            unsigned char   h_i_fsize;  /* Fragment size */
            unsigned short  h_i_mode_high;
            unsigned short  h_i_uid_high;
            unsigned short  h_i_gid_high;
            unsigned int    h_i_author;
        } hurd2;
        struct {
            unsigned char   m_i_frag;   /* Fragment number */
            unsigned char   m_i_fsize;  /* Fragment size */
            unsigned short  m_pad1;
            unsigned int    m_i_reserved2[2];
        } masix2;
    } osd2;                 /* OS dependent 2 */
} Inode;

typedef struct {
    unsigned int inode;         /* Inode number */
    unsigned short rec_len;     /* Directory entry length */
    unsigned short name_len;    /* Name length */
    char name[255];             /* File name */
} DirE;

typedef struct {
    unsigned int blockSize;      /* block size in bytes */
    BlockBuffer_Unit *unit;      /* device mounted */
    BlockBuffer_Block **groupBD; /* group descriptors */
    BlockBuffer_Block *sbb;      /* FS superblock */
    unsigned int refCount;
} FS;

typedef struct _DentryData {
    unsigned int inodeN;
    Inode *inode;
    FS *fs;
    struct _DentryData *next;
    struct _DentryData *prior;
    BlockBuffer_Block *iBlock;
    unsigned int refCount;
} DentryData;

struct _FileSystem_Handle {
    DentryData *dd;
    unsigned int pos;
    int mode;
};

SysModManager_Module *ID;
DentryData **dentryDataHT;
TaskManager_Semaphore dentryDataSem = {1, NULL};

MemManager_Cache *dentryDataObjs, *fdObjs;

int GetDentryData(unsigned int,DentryData**,FS*);
int GetPhysBlock(DentryData*,unsigned int,unsigned int*);

static __inline__ int CheckGroupDescriptors(unsigned int groupC, BlockBuffer_Block **groupBD, unsigned int blockSize, SB *sb) {
    unsigned int i, block, itb_per_group;
    GroupDesc *gd;

    DEBUG_MSG(("Ext2fs: Checking descriptors.\n"));
    gd = NULL;
    block = sb->first_data_block;
    itb_per_group = sb->inodes_per_group / (blockSize/128);
    for (i=0; i < groupC; i++) {
        if (!(i % (blockSize / sizeof(GroupDesc))))
            gd = (GroupDesc*)(groupBD[i / (blockSize / sizeof(GroupDesc))]->data);
        if (gd->bg_block_bitmap < block || gd->bg_block_bitmap >= block + sb->blocks_per_group) {
            DEBUG_MSG(("Ext2fs: Block bitmap for group %x not in group (block %x)!\n", i, gd->bg_block_bitmap));
            return E_CORRUPTED_FS;
        }
        if (gd->bg_inode_bitmap < block || gd->bg_inode_bitmap >= block + sb->blocks_per_group) {
            DEBUG_MSG(("Ext2fs: Inode bitmap for group %x not in group (block %x)!\n", i, gd->bg_inode_bitmap));
            return E_CORRUPTED_FS;
        }
        if (gd->bg_inode_table < block || gd->bg_inode_table + itb_per_group >= block + sb->blocks_per_group) {
            DEBUG_MSG(("Ext2fs: Inode table for group %x not in group (block %x)!\n", i, gd->bg_inode_table));
            return E_CORRUPTED_FS;
        }
        block += sb->blocks_per_group;
        gd++;
    }
    return E_OK;
}

void Close(FileSystem_Handle *fd) {
    int ret;

#ifdef DEBUG_ASSERT
    if (!fd) {
        ASSERT_MSG_PRE(0);
        return;
    }
#endif
    TaskManager_AtomicDec(&fd->dd->refCount);
    ret = MemManager_FreeObj(fdObjs, fd);
    ASSERT_ERROR(ret);
    return;
}

static __inline__ int DoMount(BlockBuffer_Unit *unit, DentryData **root) {
    int error, ret;
    unsigned int blockSize, blockN, blockOff, groupC, groupBC, groupB, i;
    SB *sb;
    BlockBuffer_Block *block, **groupBD;
    FS *mountFS;

    error = E_BAD_VALUE;
    if (BlockBuffer_GetHardBlockSize(unit, &blockSize))
        goto returnEnd;
    if (blockSize < 1024) {
        blockSize = 1024;
        BlockBuffer_SetBlockSize(unit, 1024);
    }
    /* superblock always start at byte 1024. So we need to figure out the block */
    blockN = 1024 / blockSize;
    blockOff = 1024 % blockSize;
    error = E_IO_ERROR;
    if (BlockBuffer_GetBlock(unit, blockN, 0x01, &block))
        goto returnEnd;
    sb = (SB*)(block->data+blockOff);
    /* is an ext2 fs? */
    error = E_BAD_TYPE;
    if (sb->magic != 0xEF53)
        goto returnFreeBlock;
    /* check feature flags */
    if (sb->feature_incompat & IMCOMPAT_FEATURES_MASK) {
        DEBUG_MSG(("Ext2fs: Incompatible features: %x\n", sb->feature_incompat));
        goto returnFreeBlock;
    }
    /* bad block size */
    if ((1024 << sb->log_block_size) != 1024 && (1024 << sb->log_block_size) != 2048 &&
       (1024 << sb->log_block_size) != 4096) {
        DEBUG_MSG(("Ext2fs: Unsupported block size: %x\n", 1024 << sb->log_block_size));
        goto returnFreeBlock;
    }
    /* read superblock again? */
    if ((1024 << sb->log_block_size) != blockSize) {
        /* block size small than device block? */
        error = E_CORRUPTED_FS;
        if ((1024 << sb->log_block_size) < blockSize)
            goto returnFreeBlock;
        blockSize = 1024 << sb->log_block_size;
        BlockBuffer_Release(unit, block);
        BlockBuffer_SetBlockSize(unit, blockSize);
        blockN = 1024 / blockSize;
        blockOff = 1024 % blockSize;
        error = E_IO_ERROR;
        if (BlockBuffer_GetBlock(unit, blockN, 0x01, &block))
            goto returnEnd;
        sb = (SB*)(block->data+blockOff);
        /* is an ext2 fs? */
        error = E_BAD_TYPE;
        if (sb->magic != 0xEF53)
            goto returnFreeBlock;
    }
    /* check revision */
    if (sb->rev_level != 0x00) {
        /* we only support 128 bytes inode size */
        if (sb->inode_size != 128) {
            DEBUG_MSG(("Ext2fs: Unsupported inode size: %x\n", sb->inode_size));
            goto returnFreeBlock;
        }
    }
    /* we don't support fragments */
    if (sb->log_frag_size) {
        DEBUG_MSG(("Ext2fs: Unsupported fragmente size: %x\n", 1024 << sb->log_frag_size));
        goto returnFreeBlock;
    }
    error = E_CORRUPTED_FS;
    if (sb->blocks_per_group > blockSize * 8) {
        DEBUG_MSG(("Ext2fs: # blocks per group too big: %x\n", sb->blocks_per_group));
        goto returnFreeBlock;
    }
    if (sb->frags_per_group > blockSize * 8) {
        DEBUG_MSG(("Ext2fs: # fragments per group too big: %x\n", sb->frags_per_group));
        goto returnFreeBlock;
    }
    if (sb->inodes_per_group > blockSize * 8) {
        DEBUG_MSG(("Ext2fs: # inodes per group too big: %x\n", sb->inodes_per_group));
        goto returnFreeBlock;
    }
    /* get group descriptors */
    /* calculate how many groups */
    groupC = (sb->blocks_count - sb->first_data_block + sb->blocks_per_group - 1) / sb->blocks_per_group;
    /* calculate how many group blocks */
    groupBC = (groupC + (blockSize / sizeof(GroupDesc)) - 1) / (blockSize / sizeof(GroupDesc));
    groupBD = MemManager_Alloc(groupBC*sizeof(BlockBuffer_Block*), 0);
    if (IS_PTR_ERROR(groupBD)) {
        error = PTR_TO_ERROR(groupBD);
        goto returnFreeBlock;
    }
    MemMan_Set((char*)groupBD, 0, groupBC*sizeof(BlockBuffer_Block*));
    groupB = sb->first_data_block + 1;
    error = E_IO_ERROR;
    for (i=0; i<groupBC; i++, groupB++)
        if (BlockBuffer_GetBlock(unit, groupB, 0x01, &groupBD[i]))
            goto returnFreeGroupD;
    /* now check group descriptors */
    error = CheckGroupDescriptors(groupC, groupBD, blockSize, sb);
    if (error)
        goto returnFreeGroupD;
    mountFS = MemManager_Alloc(sizeof(FS), MemManager_BLOCK_MEM);
    if (IS_PTR_ERROR(mountFS)) {
        error = PTR_TO_ERROR(mountFS);
        goto returnFreeGroupD;
    }
    mountFS->blockSize = blockSize;
    mountFS->unit = unit;
    mountFS->groupBD = groupBD;
    mountFS->sbb = block;
    mountFS->refCount = 1;
    /* read root inode */
    error = GetDentryData(2, root, mountFS);
    if (error)
        goto returnFreeFS;
    goto returnEnd;

returnFreeFS:
    ret = MemManager_Free(mountFS);
    ASSERT_ERROR(ret);
returnFreeGroupD:
    {
        unsigned int j;
        for (j=0; j<groupBC; j++) {
            if (!groupBD[j])
                break;
            ret = BlockBuffer_Release(unit, groupBD[j]);
            ASSERT_ERROR(ret);
        }
        ret = MemManager_Free(groupBD);
        ASSERT_ERROR(ret);
    }
returnFreeBlock:
    ret = BlockBuffer_Release(unit, block);
    ASSERT_ERROR(ret);
returnEnd:
    return error;
}

static __inline__ int FindDirEntry(DentryData *dd, FSManager_ExStr *name, unsigned int *inodeN) {
    unsigned int blockN, pb;
    BlockBuffer_Block *block;
    int error, ret;

    blockN = 0;
    while ((error = GetPhysBlock(dd, blockN, &pb)) == E_OK) {
        unsigned int offset;
        DirE *dir;

        if (BlockBuffer_GetBlock(dd->fs->unit, pb, 0, &block))
            return E_IO_ERROR;
        offset = 0;
        while (offset < dd->fs->blockSize) {
            dir = (DirE*)(block->data+offset);
            /* some check */
            if (((offset + dir->rec_len) > dd->fs->blockSize) || (dir->rec_len < 8) ||
               (((dir->name_len & 0xFF)+8) > dir->rec_len)) {
                DEBUG_MSG(("Ext2fs: Invalid dir entry in block %x offset %x\n", block->blockN, offset));
                ret = BlockBuffer_Release(dd->fs->unit, block);
                ASSERT_ERROR(ret);
                return E_CORRUPTED_FS;
            }
            if (dir->inode && (dir->name_len & 0xFF) == name->len && !StringMan_NCompare(name->str, dir->name, name->len)) {
                *inodeN = dir->inode;
                ret = BlockBuffer_Release(dd->fs->unit, block);
                ASSERT_ERROR(ret);
                return E_OK;
            }
            offset += dir->rec_len;
        }
        blockN++;
        ret = BlockBuffer_Release(dd->fs->unit, block);
        ASSERT_ERROR(ret);
    }
    return error;
}

static __inline__ DentryData **GetHashEntry(unsigned int inode, FS *fs) {
    unsigned int hash;

    hash = (unsigned int)fs >> 5;
    hash = hash ^ (inode >> HASHBITS) ^ (inode << HASHBITS);
    return dentryDataHT + (hash & HASHMASK);
}

static __inline__ DentryData *GetCacheDentryData(unsigned int inodeN, FS *fs) {
    DentryData **head, *tmp, *ret;

    head = GetHashEntry(inodeN, fs);
    TaskManager_P(&dentryDataSem, 0, 0);
    tmp = *head;
    ret = NULL;
    while (tmp) {
        DentryData *dentry;
        dentry = tmp;
        tmp = tmp->next;
        if (dentry->inodeN != inodeN || dentry->fs != fs || dentry->fs->unit != fs->unit)
            continue;
        dentry->refCount++;
        ret = dentry;
        break;
    }
    TaskManager_V(&dentryDataSem, 0);
    return ret;
}

static __inline__ int ReadDentryData(unsigned int inodeN, DentryData **dd, FS *fs) {
    unsigned int group, groupB, offset, blockN, firstInodeN;
    BlockBuffer_Block *block;
    unsigned int i;
    DentryData *newDD;
    int error, ret;

    /* first read inode block */
    group = (inodeN-1) / ((SB*)fs->sbb->data)->inodes_per_group;
    groupB = group / (fs->blockSize / sizeof(GroupDesc));
    group = group % (fs->blockSize / sizeof(GroupDesc));
    offset = ((inodeN-1) % ((SB*)fs->sbb->data)->inodes_per_group) * 128;
    blockN = offset >> (10+((SB*)fs->sbb->data)->log_block_size);
    firstInodeN = offset & (fs->blockSize-1);
    firstInodeN >>= 7;
    firstInodeN = inodeN - firstInodeN;
    blockN += ((GroupDesc*)fs->groupBD[groupB]->data)[group].bg_inode_table;
    if (BlockBuffer_GetBlock(fs->unit, blockN, 0, &block))
        return E_IO_ERROR;
    TaskManager_P(&dentryDataSem, 0, 0);
    if (block->avail2 != (void*)0xdf657890) {
        block->avail2 = (void*)0xdf657890;
        error = E_OK;
    } else
        error = E_IN_USE;
    TaskManager_V(&dentryDataSem, 0);
    if (error) {
        /* someone has already get the block.
           give some time so it can add all DentryData in the hash.
        */
        ret = BlockBuffer_Release(fs->unit, block);
        ASSERT_ERROR(ret);
        TaskManager_Yield();
        return E_OK;
    }
    /* add all DentryData structures in the hash table */
    i = fs->blockSize / 128;
    /* first allocate all memory.
       newDD will hold a linked list of DentryData.
    */
    newDD = NULL;
    do {
        DentryData *tmp;
        tmp = MemManager_AllocObj(dentryDataObjs);
        if (IS_PTR_ERROR(tmp)) {
            /* no memory, free all obj allocated */
            while (newDD) {
                tmp = newDD->next;
                ret = MemManager_FreeObj(dentryDataObjs, newDD);
                ASSERT_ERROR(ret);
                newDD = tmp;
            }
            ret = BlockBuffer_Release(fs->unit, block);
            ASSERT_ERROR(ret);
            return E_NO_MEMORY;
        }
        tmp->next = newDD;
        newDD = tmp;
    } while (--i);
    /* now add them */
    TaskManager_P(&dentryDataSem, 0, 0);
    i = 0;
    do {
        DentryData *tmp, **entry;
        if (firstInodeN == inodeN) {
            *dd = newDD;
            newDD->refCount = 1;
        } else
            newDD->refCount = 0;
        newDD->inodeN = firstInodeN;
        newDD->fs = fs;
        newDD->iBlock = block;
        newDD->inode = ((Inode*)block->data)+i;
        /* add to hash table */
        entry = GetHashEntry(inodeN, fs);
        tmp = newDD->next;
        newDD->next = *entry;
        newDD->prior = NULL;
        if (*entry)
            (*entry)->prior = newDD;
        *entry = newDD;
        newDD = tmp;
        firstInodeN++;
    } while (++i < (fs->blockSize / 128));
    block->avail1 = (void*)i;
    TaskManager_V(&dentryDataSem, 0);
    return E_OK;
}

int GetDentryData(unsigned int inodeN, DentryData **dd, FS *fs) {
    DentryData *ret;

    do {
        ret = GetCacheDentryData(inodeN, fs);
        if (!ret) {
            int error;
            /* we have to read from disk */
            error = ReadDentryData(inodeN, &ret, fs);
            if (error)
                return error;
        }
    } while (!ret);
    *dd = ret;
    return E_OK;
}

static __inline__ int IndBlockMap(DentryData *dd, unsigned int ind, unsigned int blockN, unsigned int *pb) {
    BlockBuffer_Block *block;
    unsigned int b;
    int ret;

    if (!ind)
        return E_NOT_FOUND;
    if (BlockBuffer_GetBlock(dd->fs->unit, ind, 0, &block))
        return E_IO_ERROR;
    b = ((unsigned int *)block->data)[blockN];
    ret = BlockBuffer_Release(dd->fs->unit, block);
    ASSERT_ERROR(ret);
    if (!b)
        return E_NOT_FOUND;
    *pb = b;
    return E_OK;
}

int GetPhysBlock(DentryData *dd, unsigned int blockN, unsigned int *pb) {
    unsigned int nMap, dnMap, indBlock;

    /* is a direct mapped block? */
    if (blockN < 12) {
        if (dd->inode->i_block[blockN]) {
            *pb = dd->inode->i_block[blockN];
            return E_OK;
        }
        return E_NOT_FOUND;
    }
    nMap = dd->fs->blockSize >> 2;
    dnMap = nMap*nMap;
    blockN -= 12;
    /* inexistent block? */
    if (blockN >= ((dnMap*nMap)+dnMap+nMap))
        return E_NOT_FOUND;
    indBlock = 0;
    /* is it a three indirect block? */
    if (blockN >= (nMap + dnMap)) {
        int error;
        /* remove indirect and double-indirect blocks */
        blockN -= dnMap;
        blockN -= nMap;
        error = IndBlockMap(dd, dd->inode->i_block[14], blockN/dnMap, &indBlock);
        if (error)
            return error;
        /* remove blocks to turn into a double-indirect block */
        blockN = blockN % dnMap;
    }
    /* is it a two indirect block? */
    if (blockN >= nMap || indBlock) {
        int error;
        if (!indBlock) {
            /* remove indirect blocks */
            blockN -= nMap;
            indBlock = dd->inode->i_block[13];
        }
        error = IndBlockMap(dd, indBlock, blockN/nMap, &indBlock);
        if (error)
            return error;
        /* remove blocks to turn into a indirect block */
        blockN = blockN % nMap;
    }
    /* it is a indirect block */
    if (!indBlock)
        indBlock = dd->inode->i_block[12];
    return IndBlockMap(dd, indBlock, blockN, pb);
}

int Lookup(FSManager_Dentry *parent, FSManager_Dentry *newDentry) {
    int error;
    unsigned int inodeN;

#ifdef DEBUG_ASSERT
    if (!parent || !newDentry) {
        ASSERT_MSG_PRE(0);
        return E_BAD_VALUE;
    }
#endif
    /* get inode number */
    error = FindDirEntry(parent->fsData, &newDentry->name, &inodeN);
    if (error)
        return error;
    /* ok, now get the inode */
    error = GetDentryData(inodeN, (DentryData**)&newDentry->fsData, ((DentryData*)parent->fsData)->fs);
    if (!error)
        newDentry->mode = ((DentryData*)newDentry->fsData)->inode->i_mode;
    return error;
}

int Mount(const char *devImp, unsigned int unit, int prop, const char *data, FSManager_Dentry *dentry) {
    int error;
    BlockBuffer_Unit *devUnit;

#ifdef DEBUG_ASSERT
    if (!devImp || !dentry) {
        ASSERT_MSG_PRE(0);
        return E_BAD_VALUE;
    }
#endif
    /* only read only now */
    error = E_READ_ONLY_FS;
    if (!(prop & 0x01))
        return error;
    /* get unit ID */
    error = BlockBuffer_RegisterUnit(devImp, unit, &devUnit);
    if (error)
        return error;
    error = DoMount(devUnit, (DentryData**)&dentry->fsData);
    if (error) {
        int ret;
        ret = BlockBuffer_FreeUnit(devUnit);
        ASSERT_ERROR(ret);
    }
    return error;
}

int Open(FSManager_Dentry *dentry, int prop, FileSystem_Handle **fd) {
#ifdef DEBUG_ASSERT
    if (!dentry || !fd) {
        ASSERT_MSG_PRE(0);
        return E_BAD_VALUE;
    }
#endif
    /* we only support read now */
    if (prop & 0x3e)
        return E_READ_ONLY_FS;
    *fd = MemManager_AllocObj(fdObjs);
    if (IS_PTR_ERROR(*fd))
        return PTR_TO_ERROR(*fd);
    (*fd)->dd = (DentryData*)dentry->fsData;
    (*fd)->pos = 0;
    TaskManager_AtomicInc(&(*fd)->dd->refCount);
    return E_OK;
}

int ReadFile(FSManager_FDIO *request) {
    int read, error, ret;
    DentryData *dd;
    FileSystem_Handle *fd;

#ifdef DEBUG_ASSERT
    if (!request) {
        ASSERT_MSG_PRE(0);
        return E_BAD_VALUE;
    }
#endif
    read = 0;
    fd = request->fd->fd;
    dd = fd->dd;
    while ((fd->pos < dd->inode->i_size) && (read != request->size)) {
        unsigned int blockN, copy, start, left, pb;
        BlockBuffer_Block *block;

        blockN = fd->pos / dd->fs->blockSize;
        /* get physical block number */
        error = E_IO_ERROR;
        if (GetPhysBlock(dd, blockN, &pb))
            goto returnRead;
        /* get block */
        if (BlockBuffer_GetBlock(dd->fs->unit, pb, 0, &block))
            goto returnRead;
        start = fd->pos % dd->fs->blockSize;
        /* bytes to copy */
        copy = dd->fs->blockSize - start;
        if (copy > (request->size - read))
            copy = request->size - read;
        /* will pass end of file? */
        left = dd->inode->i_size - fd->pos;
        if (copy > left)
            copy = left;
        /* system buffer? */
        error = E_OK;
        if (!request->flags)
            MemMan_Copy(request->buf+read, block->data+start, copy);
        else
            error = UserSpaceManager_CopyToUser(request->buf+read, block->data+start, copy);
        ret = BlockBuffer_Release(dd->fs->unit, block);
        ASSERT_ERROR(ret);
        if (error)
            goto returnRead;
        fd->pos += copy;
        read += copy;
    }
    if (read)
        error = E_OK;
    else
        error = E_END_REACHED;
returnRead:
    request->size = read;
    return error;
}

int ReadDir(FSManager_FDIO *request) {
    int read, error, ret;
    DentryData *dd;
    FileSystem_Handle *fd;

#ifdef DEBUG_ASSERT
    if (!request) {
        ASSERT_MSG_PRE(0);
        return E_BAD_VALUE;
    }
#endif
    read = 0;
    fd = request->fd->fd;
    dd = fd->dd;
    while ((fd->pos < dd->inode->i_size) && (read != request->size)) {
        unsigned int blockN, offset, pb;
        BlockBuffer_Block *block;

        blockN = fd->pos / dd->fs->blockSize;
        offset = fd->pos % dd->fs->blockSize;
        /* get physical block number */
        error = E_IO_ERROR;
        if (GetPhysBlock(dd, blockN, &pb))
            goto returnRead;
        /* get block */
        if (BlockBuffer_GetBlock(dd->fs->unit, pb, 0, &block))
            goto returnRead;
        while (offset < dd->fs->blockSize) {
            DirE *rawDir;
            FSManager_DirEntry *dire;

            rawDir = (DirE*)(block->data+offset);
            /* some check */
            error = E_CORRUPTED_FS;
            if (((offset + rawDir->rec_len) > dd->fs->blockSize) || (rawDir->rec_len < 8) ||
               (((rawDir->name_len & 0xFF)+8) > rawDir->rec_len)) {
                DEBUG_MSG(("Ext2fs: Invalid dir entry in block %x offset %x\n", block->blockN, offset));
                ret = BlockBuffer_Release(dd->fs->unit, block);
                ASSERT_ERROR(ret);
                goto returnRead;
            }
            if (rawDir->inode) {
                /* dir entry too big for buffer? */
                if (((rawDir->name_len & 0xFF)+3) > (request->size-read)) {
                    ret = BlockBuffer_Release(dd->fs->unit, block);
                    ASSERT_ERROR(ret);
                    goto returnOK;
                }
                /* copy dir entry */
                dire = (FSManager_DirEntry*)((char*)request->buf+read);
                /* system buffer? */
                if (!request->flags) {
                    dire->size = (rawDir->name_len & 0xFF)+3;
                    MemMan_Copy(dire->name, rawDir->name, rawDir->name_len & 0xFF);
                    dire->name[rawDir->name_len & 0xFF] = 0;
                } else {
                    int tmp;
                    tmp = (rawDir->name_len & 0xFF)+3;
                    error = UserSpaceManager_CopyToUser((char*)&dire->size, (char*)&tmp, 2);
                    error |= UserSpaceManager_CopyToUser(dire->name, rawDir->name, rawDir->name_len & 0xFF);
                    tmp = 0;
                    error |= UserSpaceManager_CopyToUser(&dire->name[rawDir->name_len & 0xFF], (char*)&tmp, 1);
                    if (error) {
                        ret = BlockBuffer_Release(dd->fs->unit, block);
                        ASSERT_ERROR(ret);
                        goto returnRead;
                    }
                }
                read += (rawDir->name_len & 0xFF)+3;
            }
            offset += rawDir->rec_len;
        }
        ret = BlockBuffer_Release(dd->fs->unit, block);
        ASSERT_ERROR(ret);
        fd->pos += offset;
    }
returnOK:
    if (read)
        error = E_OK;
    else
        error = E_END_REACHED;
returnRead:
    request->size = read;
    return error;
}

int Read(FSManager_FDIO *request) {
    if (IS_FILE(request->fd->fd->dd->inode->i_mode))
        return ReadFile(request);
    else
        return ReadDir(request);
}

int ReadLink(FSManager_Dentry *dentry, char *buf, unsigned int size, int prop) {
    DentryData *dd;
    unsigned int pos;
    int ret;

#ifdef DEBUG_ASSERT
    if (!dentry || !buf) {
        ASSERT_MSG_PRE(0);
        return E_BAD_VALUE;
    }
#endif
    dd = (DentryData*)dentry->fsData;
    if (dd->inode->i_size > size)
        return E_TOO_BIG;
    if (!dd->inode->i_blocks) {
        /* symbolic link data in the inode */
        /* system buffer? */
        if (!prop) {
            MemMan_Copy(buf, (char*)dd->inode->i_block, dd->inode->i_size);
            return E_OK;
        }
        return UserSpaceManager_CopyToUser(buf, (char*)dd->inode->i_block, dd->inode->i_size);
    }
    pos = 0;
    while (pos < dd->inode->i_size) {
        unsigned int blockN, copy, left, pb;
        int error;
        BlockBuffer_Block *block;

        blockN = pos / dd->fs->blockSize;
        /* get physical block number */
        error = E_IO_ERROR;
        if (GetPhysBlock(dd, blockN, &pb))
            return error;
        /* get block */
        if (BlockBuffer_GetBlock(dd->fs->unit, pb, 0, &block))
            return error;
        /* bytes to copy */
        copy = dd->fs->blockSize;
        /* will pass end of file? */
        left = dd->inode->i_size - pos;
        if (copy > left)
            copy = left;
        /* system buffer? */
        if (!prop)
            MemMan_Copy(buf+pos, block->data, copy);
        else {
            error = UserSpaceManager_CopyToUser(buf+pos, block->data, copy);
            if (error) {
                ret = BlockBuffer_Release(dd->fs->unit, block);
                ASSERT_ERROR(ret);
                return error;
            }
        }
        ret = BlockBuffer_Release(dd->fs->unit, block);
        ASSERT_ERROR(ret);
        pos += copy;
    }
    return E_OK;
}

int Remove(FSManager_Dentry *dentry) {
    return E_READ_ONLY_FS;
}

int RewindDir(FileSystem_Handle *fd) {
#ifdef DEBUG_ASSERT
    if (!fd) {
        ASSERT_MSG_PRE(0);
        return E_BAD_VALUE;
    }
#endif
    if (IS_FILE(fd->dd->inode->i_mode))
        return E_NOT_A_DIRECTORY;
    fd->pos = 0;
    return E_OK;
}

int Seek(FileSystem_Handle *fd, int count, int prop, int *newPos) {
#ifdef DEBUG_ASSERT
    if (!fd) {
        ASSERT_MSG_PRE(0);
        return E_BAD_VALUE;
    }
#endif
    if (IS_DIR(fd->dd->inode->i_mode))
        return E_IS_A_DIRECTORY;
    if (prop & FSManager_SEEK_CURRENT)
        count += (int)fd->pos;
    else if (prop & FSManager_SEEK_END)
        count = count + fd->dd->inode->i_size;
    if (count < 0)
        return E_TOO_BIG;
    fd->pos = (unsigned int)count;
    if (newPos)
        *newPos = count;
    return E_OK;
}

int Shutdown(void) {
    int ret;

    ret = MemManager_DestroyObjCache(dentryDataObjs);
    ASSERT_ERROR(ret);
    ret = MemManager_DestroyObjCache(fdObjs);
    ASSERT_ERROR(ret);
    ret = MemManager_Free(dentryDataHT);
    ASSERT_ERROR(ret);
    return E_OK;
}

int Phase1Start(SysModManager_Module *modID) {
    int ret;

    ID = modID;
    /* start all used modules. Don't need to start UserSpaceManager because
       we only use it if it's already started */
    if ((ret = SysModManager_StartModule((unsigned int)BlockBuffer_FreeUnit)) ||
        (ret = SysModManager_StartModule((unsigned int)MemMan_Copy)) ||
        (ret = SysModManager_StartModule((unsigned int)MemManager_Alloc)) ||
        (ret = SysModManager_StartModule((unsigned int)StringMan_NCompare)) ||
        (ret = SysModManager_StartModule((unsigned int)TaskManager_P))) {
        DEBUG_MSG_PRE(ret);
        goto returnEnd;
    }
    dentryDataObjs = MemManager_CreateObjCache(sizeof(DentryData), 0, 0, NULL, NULL);
    if (IS_PTR_ERROR(dentryDataObjs)) {
        DEBUG_MSG_PRE(PTR_TO_ERROR(dentryDataObjs));
        goto returnEnd;
    }
    fdObjs = MemManager_CreateObjCache(sizeof(FileSystem_Handle), 0, 0, NULL, NULL);
    if (IS_PTR_ERROR(fdObjs)) {
        DEBUG_MSG_PRE(PTR_TO_ERROR(fdObjs));
        goto returnFreeDentryObjs;
    }
    /* for now 4096 bytes of hash table */
    dentryDataHT = MemManager_Alloc(4096, 0);
    if (IS_PTR_ERROR(dentryDataHT)) {
        DEBUG_MSG_PRE(PTR_TO_ERROR(dentryDataHT));
        goto returnFreeFdObjs;
    }
    MemMan_Set((char*)dentryDataHT, 0, 4096);
    return 0;

returnFreeFdObjs:
    ret = MemManager_DestroyObjCache(fdObjs);
    ASSERT_ERROR(ret);
returnFreeDentryObjs:
    ret = MemManager_DestroyObjCache(dentryDataObjs);
    ASSERT_ERROR(ret);
returnEnd:
    return 1;
}

int StatusDD(DentryData *dd, FSManager_EntryStatus *buf, int prop) {
    FSManager_EntryStatus tmp, *pBuf;

    if (prop & FSManager_USER_BUF)
        pBuf = &tmp;
    else
        pBuf = buf;
    pBuf->mode = dd->inode->i_mode;
    pBuf->nlink = dd->inode->i_links_count;
    pBuf->uid = dd->inode->i_uid;
    pBuf->gid = dd->inode->i_gid;
    pBuf->size = (int)dd->inode->i_size;
    pBuf->blksize = dd->fs->blockSize;
    pBuf->blocks = dd->inode->i_blocks << (dd->fs->blockSize >> 9);
    pBuf->atime = dd->inode->i_atime;
    pBuf->ctime = dd->inode->i_ctime;
    pBuf->mtime = dd->inode->i_mtime;
    if (prop & FSManager_USER_BUF)
        return UserSpaceManager_CopyToUser((char*)buf, (char*)pBuf, sizeof(FSManager_EntryStatus));
    return E_OK;
}

int Status(FSManager_Dentry *dentry, FSManager_EntryStatus *buf, int prop) {
#ifdef DEBUG_ASSERT
    if (!dentry || !buf) {
        ASSERT_MSG_PRE(0);
        return E_BAD_VALUE;
    }
#endif
    return StatusDD((DentryData*)dentry->fsData, buf, prop);
}

int StatusFD(FileSystem_Handle *fd, FSManager_EntryStatus *buf, int prop) {
#ifdef DEBUG_ASSERT
    if (!fd || !buf) {
        ASSERT_MSG_PRE(0);
        return E_BAD_VALUE;
    }
#endif
    return StatusDD(fd->dd, buf, prop);
}

int Unmount(FSManager_Dentry *dentry, int prop) {
    return E_OK;
}

int Write(FSManager_FDIO *request) {
    return E_READ_ONLY_FS;
}
