/*
*****************************************
* This file implements the Keyboard and Input interfaces
*
* Version: 0.3.3
* Author: Luiz Henrique Shigunov
* E-mail: luizshigunov@users.sourceforge.net
* WEB: http://www.sourceforge.net/projects/modulos
* Last change: 17/06/2005
*
* See the interface specification to know what each function does.
*
* Copyright (C) 2000-2005, Luiz Henrique Shigunov
*
* This module is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2.1 of the License, or (at your option) any later version.
*
* This module is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this module; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*****************************************
*/

#include <stddef.h>
#include <cfgmanager.h>
#include <debug.h>
#include <errormanager.h>
#include <kbdlayout.h>
#include <keyboard.h>
#include <inputdev.h>
#include <intmanager_i386.h>
#include <ioportmanager.h>
#include <memman.h>
#include <memmanager.h>
#include <stringman.h>
#include <sysmodmanager.h>
#include <taskmanager.h>
#include <userspacemanager.h>

#define SCROLLLOCK_FLAG 0x001
#define NUMLOCK_FLAG 0x002
#define CAPSLOCK_FLAG 0x004
#define LSHIFT_FLAG 0x008
#define RSHIFT_FLAG 0x010
#define LCTRL_FLAG 0x020
#define RCTRL_FLAG 0x040
#define LALT_FLAG 0x080
#define RALT_FLAG 0x100
#define MACRO_FLAG 0x200

#define SHIFT_CTRL_ALT_MASK (LSHIFT_FLAG+RSHIFT_FLAG+LCTRL_FLAG+RCTRL_FLAG+LALT_FLAG+RALT_FLAG)

#define SHIFT_MASK (LSHIFT_FLAG+RSHIFT_FLAG)
#define CTRL_MASK (LCTRL_FLAG+RCTRL_FLAG)
#define ALT_MASK (LALT_FLAG+RALT_FLAG)

typedef int KbdData;

SysModManager_Module *ID;
TaskManager_Semaphore hasData;
unsigned char VKBuffer[256]; /* virtual key followed by scancode (to get up/down key) */
unsigned char inPos, outPos;
KbdLayout_MapTables mapTables;
unsigned short vk2CharS0[256];
unsigned short vk2CharS1[256];
unsigned short vk2CharS2[256];
unsigned short vk2CharS3[256];
unsigned short vk2CharS4[256];
unsigned short vk2CharS5[256];
unsigned short vk2CharS6[256];
unsigned short vk2CharS7[256];
char layoutImp[256];
TaskManager_Semaphore layoutMutex = {1, NULL};
int shiftState;
int ack, ackExpected;
TaskManager_Task *focus;
TaskManager_Task *sendTask;
TaskManager_Thread *sendThread;

void SendEventsThread(void);
void Wait(void);

const char *GetLayout(void) {
    if (layoutImp[0])
        return layoutImp;
    return NULL;
}

int SetLayout(const char *imp) {
    KbdLayout_MapTables *(*getMap)(void);
    SysModManager_Imp *tmp;
    KbdLayout_MapTables *mTables;
    int error, ret;
    unsigned int i;

    tmp = SysModManager_GetImpID("KbdLayout", imp);
    error = E_MODULE;
    if (IS_PTR_ERROR(tmp))
        goto errorEnd;
    getMap = SysModManager_GetFunction(tmp, 0, 0, NULL);
    if (IS_PTR_ERROR(getMap))
        goto errorPutImpID;
    mTables = getMap();
    TaskManager_P(&layoutMutex, 0, 0);
    /* copy tables */
    MemMan_Copy(&mapTables, mTables, sizeof(mapTables.sc2VK)+sizeof(mapTables.e0sc2VK)+sizeof(mapTables.vk2CharN));
    for (i = 0; i < 8; i++) {
        if (mTables->vk2CharS[i])
            MemMan_Copy(mapTables.vk2CharS[i], mTables->vk2CharS[i], sizeof(unsigned short)*256);
        else
            MemMan_Set(mapTables.vk2CharS[i], 0, sizeof(unsigned short)*256);
    }
    /* free old dead keys */
    if (mapTables.deadKeys) {
        ret = MemManager_Free(mapTables.deadKeys);
        ASSERT_ERROR(ret);
        mapTables.deadKeys = NULL;
    }
    if (mTables->deadKeys) {
        unsigned int dk;

        /* get number of dead keys */
        dk = 0;
        while (mTables->deadKeys[dk].key.all)
            dk++;
        dk++;
        mapTables.deadKeys = MemManager_Alloc(sizeof(KbdLayout_VKDeadKey)*dk, 0);
        if (IS_PTR_ERROR(mapTables.deadKeys)) {
            error = PTR_TO_ERROR(mapTables.deadKeys);
            /* clean up */
            MemMan_Set(&mapTables, 0, sizeof(mapTables.sc2VK)+sizeof(mapTables.e0sc2VK)+sizeof(mapTables.vk2CharN));
            for (i = 0; i < 8; i++) {
                MemMan_Set(mapTables.vk2CharS[i], 0, sizeof(unsigned short)*256);
            }
            goto errorV;
        }
        MemMan_Copy(mapTables.deadKeys, mTables->deadKeys, sizeof(KbdLayout_VKDeadKey)*dk);
    }
    /* copy imp name */
    StringMan_NCopy(layoutImp, imp, sizeof(layoutImp)-1);
    error = E_OK;
errorV:
    TaskManager_V(&layoutMutex, 0);
errorPutImpID:
    ret = SysModManager_PutImpID(tmp);
    ASSERT_ERROR(ret);
errorEnd:
    return error;
}

int SendByte(char byte) {
    int retries;

    retries = 3;
    do {
        int timeout;
        timeout = 3;
        ack = 0;
        ackExpected = 1;
        __asm__ __volatile__ ("outb %%al, %%dx" :: "a" (byte), "d" (0x60));
        do {
            if (ack)
                return E_OK;
            TaskManager_Wait(NULL, 1);
        } while (--timeout);
    } while (--retries);
    ackExpected = 0;
    DEBUG_MSG(("Kbd: Timeout sending command to keyboard.\n"));
    return E_TIMED_OUT;
}

int SetLeds(int status) {
    /* send led command and status */
    status &= 0x07;
    if (SendByte(0xed) || SendByte(status))
        return E_UNDEFINED;
    shiftState &= 0xfffffff8;
    shiftState |= status;
    return E_OK;
}

static __inline__ unsigned char Translate(unsigned char scancode) {
    static int prevScanCode;

    if (ackExpected) {
        /* we're waiting for an ACK */
        if (scancode == 0xfa) {
            ackExpected = 0;
            ack = 1;
            return 0;
        }
        DEBUG_MSG(("Kbd: ack expected, but received: %x\n", (int)scancode));
        return 0;
    }

    /* special prefix */
    if (scancode == 0xe0 || scancode == 0xe1) {
        prevScanCode = scancode;
        return 0;
    }

    /* 0xff or 0x00 - overflow or key error */
    if (scancode == 0x00 || scancode == 0xff) {
        DEBUG_MSG(("Kbd: Error scancode= %x\n", (int)scancode));
        prevScanCode = 0;
        return 0;
    }

    scancode &= 0x7f;

    if (!prevScanCode)
        return mapTables.sc2VK[scancode];

    /* special cases */
    if (prevScanCode == 0xe0) {
        prevScanCode = 0;
        return mapTables.e0sc2VK[scancode];
    } else if (prevScanCode == 0xe1) {
        /* first scancode of PAUSE key */
        if (scancode == 0x1d) {
            prevScanCode = 0x100;
            return 0;
        }
        DEBUG_MSG(("Kbd: Unknown 0xe1 escape sequence: 0xe1 0x%x\n", (int)scancode));
        prevScanCode = 0;
        return 0;
    } else {
        prevScanCode = 0;
        /* second scancode of PAUSE key */
        if (scancode == 0x45)
            return Keyboard_VK_PAUSE;
        DEBUG_MSG(("Kbd: Unknown 0xe1 escape sequence: 0xe1 0x1d 0x%x\n", (int)scancode));
        return 0;
    }
}

int ServeInterrupt(IntManager_i386_Regs *regs) {
    unsigned char status, scancode, vk;

    /* read scancode */
    __asm__ __volatile__ ("inb $0x60, %%al" : "=a" (scancode));
    /* read keyboard status */
    __asm__ __volatile__ ("inb $0x64, %%al" : "=a" (status));

    DEBUG_MSG(("Kbd: status= %x scancode= %x\n", (int)status, (int)scancode));
    vk = Translate(scancode);
    if (vk) {
        VKBuffer[inPos++] = vk;
        VKBuffer[inPos++] = scancode;
        TaskManager_V(&hasData, 0x01);
    }
    return E_HANDLED_INT;
}

int Shutdown(void) {
    IOPortManager_Free(ID, 0x60);
    IntManager_i386_Free(ID, 1, 0x01);
    return E_OK;
}

#define Keyboard_NO_DATA (-1)

__inline__ static int ReadByte(void) {
    char status, scancode;
    int ret;

    ret = Keyboard_NO_DATA;
    /* read keyboard status */
    __asm__ __volatile__ ("inb $0x64, %%al" : "=a" (status));
    if (status & 0x01) {
        /* read scancode */
        __asm__ __volatile__ ("inb $0x60, %%al" : "=a" (scancode));
        ret = scancode;
    }

    return ret;
}

__inline__ static void ClearInput(void) {
    int maxRead;

    maxRead = 100;
    do {
        if (ReadByte() == Keyboard_NO_DATA)
            break;
    } while (--maxRead);
}

int Phase1Start(SysModManager_Module *modID) {
    CfgManager_Group *baseGroup;
    char buf[128];
    unsigned int bufSize;
    int ret, ret2;

    ID = modID;
    /* start all used modules, except UserSpaceManager cause we only use it if it's already started */
    if ((ret = SysModManager_StartModule((unsigned int)IntManager_i386_Catch)) ||
        (ret = SysModManager_StartModule((unsigned int)IOPortManager_Alloc)) ||
        (ret = SysModManager_StartModule((unsigned int)TaskManager_P)) ||
        (ret = SysModManager_StartModule((unsigned int)StringMan_Length)) ||
        (ret = SysModManager_StartModule((unsigned int)CfgManager_OpenGroup)) ||
        (ret = SysModManager_StartModule((unsigned int)MemManager_Alloc)) ||
        (ret = SysModManager_StartModule((unsigned int)InterManager_SendEvent)) ||
        (ret = SysModManager_StartModule((unsigned int)MemMan_Copy))) {
        DEBUG_MSG_PRE(ret);
        goto returnEnd;
    }
    if ((ret = IOPortManager_Alloc(modID, 0x60, 0x64, 0))) {
        DEBUG_MSG_PRE(ret);
        goto returnEnd;
    }
    /* remove pending input */
    ClearInput();
    /* disable aux device so we can get data from ps/2 kbds */
    __asm__ __volatile__ ("outb %%al, %%dx" :: "a" (0xa7), "d" (0x64));
    Wait();
    __asm__ __volatile__ ("outb %%al, %%dx" :: "a" (0x60), "d" (0x64));
    Wait();
    __asm__ __volatile__ ("outb %%al, %%dx" :: "a" (0x65), "d" (0x60));
    if ((ret = IntManager_i386_Catch(modID, 1, ServeInterrupt, 0x01+0x04))) {
        DEBUG_MSG_PRE(ret);
        goto returnFreeIO;
    }
    /* disable leds */
    ret = SetLeds(0);
    ASSERT_ERROR(ret);
    /* start mapTables */
    mapTables.vk2CharS[0] = vk2CharS0;
    mapTables.vk2CharS[1] = vk2CharS1;
    mapTables.vk2CharS[2] = vk2CharS2;
    mapTables.vk2CharS[3] = vk2CharS3;
    mapTables.vk2CharS[4] = vk2CharS4;
    mapTables.vk2CharS[5] = vk2CharS5;
    mapTables.vk2CharS[6] = vk2CharS6;
    mapTables.vk2CharS[7] = vk2CharS7;
    /* set mapping */
    baseGroup = CfgManager_OpenGroup(NULL, "/system/Keyboard");
    if (!baseGroup) {
        DEBUG_MSG_PRE(0);
        goto returnFreeIRQ;
    }
    bufSize = 128;
    ret2 = CfgManager_GetValue(baseGroup, "layout", CfgManager_STRING, buf, &bufSize);
    ret = CfgManager_CloseGroup(baseGroup);
    ASSERT_ERROR(ret);
    if (ret2) {
        DEBUG_MSG_PRE(ret2);
        goto returnFreeIRQ;
    }
    if ((ret = SetLayout(buf))) {
        DEBUG_MSG_PRE(ret);
        goto returnFreeIRQ;
    }
    /* thread to send events */
    if ((ret = TaskManager_CreateTask(SendEventsThread, NULL, 0x40000000 | TaskManager_CREATE_SYSTEM, &sendTask, &sendThread))) {
        DEBUG_MSG_PRE(ret);
        goto returnFreeIRQ;
    }
    return E_OK;

returnFreeIRQ:
    ret = IntManager_i386_Free(modID, 1, 0x01);
    ASSERT_ERROR(ret);
returnFreeIO:
    ret = IOPortManager_Free(modID, 0x60);
    ASSERT_ERROR(ret);
returnEnd:
    return 1;
}

/* wait for clear input buffer */
void Wait(void) {
    int timeout;

    timeout = 250;
    do {
        unsigned char status;
        /* read keyboard status */
        __asm__ __volatile__ ("inb $0x64, %%al" : "=a" (status));
        if (!(status & 0x02))
            return;
        TaskManager_Wait(NULL, 1);
    } while (--timeout);
    DEBUG_MSG(("Kbd: timed out waiting for clear input buffer\n"));
}

TaskManager_Task *InputDev_GetFocus(void) {
    return focus;
}

int InputDev_GetImpInfo(InputDev_ImpInfo *info) {
#ifdef DEBUG_ASSERT
    if (!info) {
        ASSERT_MSG_PRE(0);
        return E_BAD_VALUE;
    }
#endif
    info->interface = "Keyboard";
    info->implementation = "MF2KBD";
    return E_OK;
}

int InputDev_Register(TaskManager_Task *task) {
    void *voidRet;

    voidRet = TaskManager_SetTaskInfo(task, ID, (void*)0x1234, 0);
    if (IS_PTR_ERROR(voidRet))
        return PTR_TO_ERROR(voidRet);
    return E_OK;
}

void InputDev_SetFocus(TaskManager_Task *task) {
    focus = task;
}

static __inline__ int WaitEvent(Keyboard_Event *event) {
    static KbdLayout_VKDeadKey deadKey;
    unsigned char vk;
    int shift, charCode, flag;

    TaskManager_P(&hasData, 0, 0);
    /* get virtual key and flag */
    vk = VKBuffer[outPos++];
    flag = VKBuffer[outPos++] >> 7;

    event->type = flag;
    event->vKey = vk;
    charCode = 0;

    /* translate to UNICODE char and update shift state */
    switch (vk) {
        case Keyboard_VK_NUMLOCK:
            shiftState ^= (flag-1) & NUMLOCK_FLAG;
            SetLeds(shiftState);
            break;
        case Keyboard_VK_CAPSLOCK:
            shiftState ^= (flag-1) & CAPSLOCK_FLAG;
            SetLeds(shiftState);
            break;
        case Keyboard_VK_SCROLLLOCK:
            shiftState ^= (flag-1) & SCROLLLOCK_FLAG;
            SetLeds(shiftState);
            break;
        case Keyboard_VK_LSHIFT:
            shiftState &= ~LSHIFT_FLAG;
            shiftState |= (flag-1) & LSHIFT_FLAG;
            break;
        case Keyboard_VK_RSHIFT:
            shiftState &= ~RSHIFT_FLAG;
            shiftState |= (flag-1) & RSHIFT_FLAG;
            break;
        case Keyboard_VK_LCTRL:
            shiftState &= ~LCTRL_FLAG;
            shiftState |= (flag-1) & LCTRL_FLAG;
            break;
        case Keyboard_VK_RCTRL:
            shiftState &= ~RCTRL_FLAG;
            shiftState |= (flag-1) & RCTRL_FLAG;
            break;
        case Keyboard_VK_LALT:
            shiftState &= ~LALT_FLAG;
            shiftState |= (flag-1) & LALT_FLAG;
            break;
        case Keyboard_VK_RALT:
            shiftState &= ~RALT_FLAG;
            shiftState |= (flag-1) & RALT_FLAG;
            break;
        case Keyboard_VK_MACRO:
            shiftState &= ~MACRO_FLAG;
            shiftState |= (flag-1) & MACRO_FLAG;
            break;
        default:
            shift = 0;
            /* get shift table */
            /* capslock is like shift if key uses it */
            if (shiftState & CAPSLOCK_FLAG && mapTables.vk2CharN[vk].flags)
                shift = 0x01;
            if (shiftState & SHIFT_MASK)
                /* shift will enable or disable shift */
                shift ^= 0x01;
            if (shiftState & CTRL_MASK)
                shift |= 0x02;
            if (shiftState & LALT_FLAG)
                shift |= 0x04;
            /* alt gr only used if alone */
            if (!shift && shiftState & RALT_FLAG)
                shift = 0x08;
            if (!shift)
                charCode = mapTables.vk2CharN[vk].charCode;
            else {
                unsigned short *charTable;
                charTable = mapTables.vk2CharS[shift-1];
                charCode = charTable[vk];
            }

            if (!flag) {
                if (deadKey.key.all) {
                    int dk;
                    /* previous key was a deadkey - search for charCode */
                    deadKey.key.vks.nextVK = vk;
                    deadKey.key.vks.nextShift = shift;
                    charCode = 0;
                    dk = 0;
                    do {
                        if (mapTables.deadKeys[dk].key.all == deadKey.key.all) {
                            charCode = mapTables.deadKeys[dk].charCode;
                            break;
                        }
                        dk++;
                    } while (mapTables.deadKeys[dk].key.all);
                    deadKey.key.all = 0;
                }

                /* is a deadkey? */
                if (charCode == 0xf000) {
                    deadKey.key.vks.firstVK = vk;
                    deadKey.key.vks.firstShift = shift;
                    charCode = 0;
                }
            }
    }

    event->flags = shiftState;
    event->charCode = charCode;
    return E_OK;
}

void SendEventsThread(void) {
    Keyboard_Event event;
    int ret;

    while (1) {
        ret = WaitEvent(&event);
        ASSERT_ERROR(ret);
        if (focus) {
            InterManager_SendEvent(focus, (InterManager_Event*)&event);
        }
    }
}

int UGetLayout(Keyboard_LayoutInfo *info, unsigned int size) {
    KbdData *data;
    unsigned short interSize, impSize;
    int error;

    data = TaskManager_GetTaskInfo(NULL, ID);
    if (IS_PTR_ERROR(data))
        return E_NOT_REGISTERED;
    interSize = sizeof("KbdLayout");
    impSize = StringMan_Length(layoutImp) + 1;
    if (size < (interSize + impSize + sizeof(info->interSize) + sizeof(info->impSize)))
        return E_TOO_BIG;
    error = UserSpaceManager_CopyToUser(&info->interSize, &interSize, sizeof(interSize));
    error |= UserSpaceManager_CopyToUser(&info->impSize, &impSize, sizeof(impSize));
    error |= UserSpaceManager_CopyToUser(info->names, "KbdLayout", interSize);
    error |= UserSpaceManager_CopyToUser(info->names+interSize, layoutImp, impSize);
    if (error)
        error = E_FAULT;
    return error;
}

int USetLayout(const char *imp) {
    KbdData *info;
    char tmpImp[256];
    unsigned int size;
    int error;

    info = TaskManager_GetTaskInfo(NULL, ID);
    if (IS_PTR_ERROR(info))
        return E_NOT_REGISTERED;
    error = UserSpaceManager_StrNCopy(tmpImp, imp, sizeof(tmpImp), &size);
    if (error)
        return error;
    return SetLayout(tmpImp);
}
