/*
*****************************************
* This file implements the CfgManager interface
*
* Version: 0.1.2
* Author: Luiz Henrique Shigunov
* E-mail: luizshigunov@users.sourceforge.net
* WEB: http://www.sourceforge.net/projects/modulos
* Last change: 01/07/2005
*
* See the interface specification to know what each function does.
*
* Copyright (C) 2000-2005, Luiz Henrique Shigunov
*
* This module is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2.1 of the License, or (at your option) any later version.
*
* This module is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this module; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*****************************************
*/

#include <stddef.h>
#include <cfgmanager.h>
#include <debug.h>
#include <errormanager.h>
#include <memman.h>
#include <memmanager.h>
#include <stringman.h>
#include <sysdef.h>
#include <sysmodmanager.h>

typedef struct _CfgManager_Key {
    struct _CfgManager_Key *next;
    const char *name;
    unsigned int nameLen;
    const char *data;
    unsigned int dataLen;
    /* so key and group can use the same obj cache */
    int reserved;
} CfgManager_Key;

struct _CfgManager_Group {
    struct _CfgManager_Group *next;
    struct _CfgManager_Group *parent;
    const char *name;
    unsigned int nameLen;
    struct _CfgManager_Group *groups;
    CfgManager_Key *keys;
};

MemManager_Cache *dataCache;
CfgManager_Group root;

/* in the future we need something better */
char *systemFile;

int CloseGroup(CfgManager_Group *group) {
    return E_OK;
}

int EnumGroup(CfgManager_Group *group, unsigned int index, char *name, unsigned int *nameSize) {
    CfgManager_Group *currGroup;

    ASSERT_RET(group, E_BUG);
    ASSERT_RET(name, E_BUG);
    ASSERT_RET(nameSize, E_BUG);

    currGroup = group->groups;
    while (index && currGroup) {
        index--;
        currGroup = currGroup->next;
    }
    if (!currGroup) {
        return E_BAD_INDEX;
    }
    if ((currGroup->nameLen+1) > *nameSize) {
        *nameSize = currGroup->nameLen+1;
        return E_TOO_BIG;
    }
    *nameSize = currGroup->nameLen;
    MemMan_Copy(name, currGroup->name, currGroup->nameLen);
    name[currGroup->nameLen] = 0;
    return E_OK;
}

int EnumKey(CfgManager_Group *group, unsigned int index, char *name, unsigned int *nameSize) {
    CfgManager_Key *currKey;

    ASSERT_RET(group, E_BUG);
    ASSERT_RET(name, E_BUG);
    ASSERT_RET(nameSize, E_BUG);

    currKey = group->keys;
    while (index && currKey) {
        index--;
        currKey = currKey->next;
    }
    if (!currKey) {
        return E_BAD_INDEX;
    }
    if ((currKey->nameLen+1) > *nameSize) {
        *nameSize = currKey->nameLen+1;
        return E_TOO_BIG;
    }
    *nameSize = currKey->nameLen;
    MemMan_Copy(name, currKey->name, currKey->nameLen);
    name[currKey->nameLen] = 0;
    return E_OK;
}

int GetValue(CfgManager_Group *group, const char *key, int type, void *buf, unsigned int *bufSize) {
    CfgManager_Key *currKey;
    unsigned int keyLen, tmpBufSize;

    ASSERT_RET(group, E_BUG);
    ASSERT_RET(key, E_BUG);
    ASSERT_RET(*key, E_BUG);
    ASSERT_RET(buf, E_BUG);
    ASSERT_RET(bufSize, E_BUG);

    currKey = group->keys;
    keyLen = StringMan_Length(key);
    while (currKey) {
        if (currKey->nameLen == keyLen && !MemMan_Compare(currKey->name, key, keyLen)) {
            break;
        }
        currKey = currKey->next;
    }
    if (!currKey) {
        return E_NOT_FOUND;
    }
    switch (type) {
        case CfgManager_HEXNUMBER:
            tmpBufSize = *bufSize;
            *bufSize = 4;
            if (tmpBufSize < 4) {
                return E_TOO_BIG;
            }
            *(long int*)buf = StringMan_StrToInt(currKey->data, NULL, 16, NULL);
            break;
        case CfgManager_STRING:
            tmpBufSize = *bufSize;
            *bufSize = currKey->dataLen + 1;
            if (tmpBufSize < (currKey->dataLen + 1)) {
                return E_TOO_BIG;
            }
            MemMan_Copy(buf, currKey->data, currKey->dataLen);
            ((char*)buf)[currKey->dataLen] = 0;
            break;
        default:
            return E_BAD_VALUE;
    }
    return E_OK;
}

CfgManager_Group *OpenGroup(CfgManager_Group *base, const char *path) {
    CfgManager_Group *group;

    ASSERT_RET(path, NULL);
    ASSERT_RET(*path, NULL);

    if (!base) {
        base = &root;
    }
    /* parse path */
    while (*path == '/') {
        path++;
    }
    group = base;
    while (*path) {
        const char *nameStart;
        unsigned int nameLen;
        CfgManager_Group *currGroup;
        char c;

        nameStart = path;
        c = *path;
        do {
            path++;
            /* valid char? */
            if (!(c >= '0' && c <= '9') && !((c & 0xdf) >= 'A' && (c & 0xdf) <= 'Z') && c != '_') {
                return NULL;
            }
            c = *path;
        } while (c && c != '/');
        nameLen = path - nameStart;
        /* remove trailing slashes */
        while (*path == '/') {
            path++;
        }
        /* lookup name */
        currGroup = group->groups;
        while (currGroup) {
            if (currGroup->nameLen == nameLen && !StringMan_NCompare(currGroup->name, nameStart, nameLen)) {
                group = currGroup;
                break;
            }
            currGroup = currGroup->next;
        }
        if (!currGroup) {
            return NULL;
        }
    }
    return group;
}

CfgManager_Group *OpenGroupIndex(CfgManager_Group *base, unsigned int index) {
    CfgManager_Group *group;

    if (!base) {
        base = &root;
    }
    group = base->groups;
    while (index && group) {
        index--;
        group = group->next;
    }
    return group;
}

/* returns: 0 - has more chars; != 0 - no more chars */
static __inline__ int RemoveBlanks(char **c, char *end) {
    while (*c < end && (**c == ' ' || **c == '\t' || **c == '\n')) {
        (*c)++;
    }
    return (*c >= end);
}

/* returns: number of chars of the id */
static __inline__ unsigned int ReadID(char **c, char *end) {
    unsigned int len;

    len = 0;
    while (*c < end && ((**c >= '0' && **c <= '9') || ((**c & 0xdf) >= 'A' && (**c & 0xdf) <= 'Z') || **c == '_')) {
        (*c)++;
        len++;
    }
    return len;
}

/* returns: number of chars of the data */
static __inline__ unsigned int ReadKeyData(char **c, char *end) {
    unsigned int len;

    len = 0;
    while (*c < end && **c != '\n') {
        (*c)++;
        len++;
    }
    return len;
}

#define CfgManager_MOD_INT 0x01
#define CfgManager_MOD_INT_CLOSE 0x02
#define CfgManager_MAY_END 0x03
#define CfgManager_GROUP_KEY 0x04

int ParseCfgFile(CfgManager_Group *group, char *file, unsigned int size) {
    CfgManager_Group *currGroup, *tmpGroup;
    CfgManager_Key *key;
    int error, ret, state, openGroups, hasModInt;
    unsigned int idLen;
    char *c, *id, *end;

    /* for parser error */
    error = E_UNDEFINED;
    openGroups = 0;
    hasModInt = 0;
    c = file;
    end = file + size;
    state = CfgManager_MAY_END;
    id = NULL;
    idLen = 0;
    currGroup = group;
    while (c < end) {
        if (RemoveBlanks(&c, end)) {
            break;
        }
        switch (state) {
            case CfgManager_MOD_INT:
                id = c;
                if ((idLen = ReadID(&c, end))) {
                    CfgManager_Group *newGroup;

                    newGroup = MemManager_AllocObj(dataCache);
                    if (!IS_PTR_ERROR(newGroup)) {
                        newGroup->name = id;
                        newGroup->nameLen = idLen;
                        newGroup->parent = currGroup;
                        newGroup->next = currGroup->groups;
                        currGroup->groups = newGroup;
                        currGroup = newGroup;
                        state = CfgManager_MOD_INT_CLOSE;
                        continue;
                    }
                    error = PTR_TO_ERROR(newGroup);
                }
                goto endWhile;
            case CfgManager_MOD_INT_CLOSE:
                if (*c == ']') {
                    state = CfgManager_MAY_END;
                    c++;
                    continue;
                }
                goto endWhile;
            case CfgManager_MAY_END:
                if (*c == '}') {
                    if (openGroups) {
                        openGroups--;
                        currGroup = currGroup->parent;
                        c++;
                        continue;
                    }
                } else if (*c == '[') {
                    if (!openGroups) {
                        c++;
                        if (currGroup != group) {
                            currGroup = currGroup->parent;
                        }
                        state = CfgManager_MOD_INT;
                        /* now we can get group/keys */
                        hasModInt = 1;
                        continue;
                    }
                } else if (hasModInt) {
                    id = c;
                    if ((idLen = ReadID(&c, end))) {
                        state = CfgManager_GROUP_KEY;
                        continue;
                    }
                }
                goto endWhile;
            case CfgManager_GROUP_KEY:
                if (*c == '{') {
                    CfgManager_Group *newGroup;

                    c++;
                    newGroup = MemManager_AllocObj(dataCache);
                    if (!IS_PTR_ERROR(newGroup)) {
                        newGroup->name = id;
                        newGroup->nameLen = idLen;
                        newGroup->parent = currGroup;
                        newGroup->next = currGroup->groups;
                        currGroup->groups = newGroup;
                        currGroup = newGroup;
                        openGroups++;
                        state = CfgManager_MAY_END;
                        continue;
                    }
                    error = PTR_TO_ERROR(newGroup);
                } else if (*c == '=') {
                    c++;
                    key = MemManager_AllocObj(dataCache);
                    if (!IS_PTR_ERROR(key)) {
                        key->name = id;
                        key->nameLen = idLen;
                        key->data = c;
                        key->dataLen = ReadKeyData(&c, end);
                        key->next = currGroup->keys;
                        currGroup->keys = key;
                        state = CfgManager_MAY_END;
                        continue;
                    }
                    error = PTR_TO_ERROR(key);
                }
                goto endWhile;
        }
    }
endWhile:
    if (!openGroups && c >= end && state == CfgManager_MAY_END) {
        return E_OK;
    }
    /* free all groups and keys */
    currGroup = group;
    do {
        key = currGroup->keys;
        currGroup->keys = NULL;
        while (key) {
            CfgManager_Key *tmp;

            tmp = key->next;
            ret = MemManager_FreeObj(dataCache, key);
            ASSERT_ERROR(ret);
            key = tmp;
        }
        tmpGroup = currGroup->groups;
        if (tmpGroup) {
            currGroup->groups = tmpGroup->next;
            currGroup = tmpGroup;
        } else if (currGroup != group) {
            tmpGroup = currGroup->parent;
            ret = MemManager_FreeObj(dataCache, currGroup);
            ASSERT_ERROR(ret);
            currGroup = tmpGroup;
        } else {
            break;
        }
    } while (1);
    return error;
}

int SetSystemFile(void *file, unsigned int size) {
    CfgManager_Group *newGroup;
    int ret, error;

    ASSERT_RET(file, E_BUG);
    ASSERT_RET(size, E_BUG);

    /* copy data */
    systemFile = MemManager_Alloc(size, MemManager_BLOCK_MEM);
    if (IS_PTR_ERROR(systemFile)) {
        error = PTR_TO_ERROR(systemFile);
        goto errorEnd;
    }
    MemMan_Copy(systemFile, file, size);
    /* create a group for this file */
    newGroup = MemManager_AllocObj(dataCache);
    if (IS_PTR_ERROR(newGroup)) {
        error = PTR_TO_ERROR(newGroup);
        goto errorFreeFile;
    }
    error = ParseCfgFile(newGroup, systemFile, size);
    if (error)
        goto errorFreeGroup;
    newGroup->name = "system";
    newGroup->nameLen = sizeof("system")-1;
    newGroup->parent = &root;
    newGroup->next = root.groups;
    root.groups = newGroup;
    return E_OK;

errorFreeGroup:
    ret = MemManager_FreeObj(dataCache, newGroup);
    ASSERT_MSG_PRE(ret);
errorFreeFile:
    ret = MemManager_Free(systemFile);
    ASSERT_MSG_PRE(ret);
errorEnd:
    return error;
}

int Phase0Start(SysModManager_Module *modID) {
    int ret;

    if ((ret = SysModManager_StartModule((unsigned int)MemManager_AllocObj)) ||
        (ret = SysModManager_StartModule((unsigned int)MemMan_Copy))) {
        DEBUG_MSG_PRE(ret);
        return 1;
    }
    dataCache = MemManager_CreateObjCache(sizeof(CfgManager_Group), 0, MemManager_CLEAR_OBJS, NULL, NULL);
    if (IS_PTR_ERROR(dataCache)) {
        DEBUG_MSG_PRE(PTR_TO_ERROR(dataCache));
        return 1;
    }
    return 0;
}

int Phase1Start(SysModManager_Module *modID) {
    int ret;

    if ((ret = SysModManager_StartModule((unsigned int)MemManager_AllocObj)) ||
        (ret = SysModManager_StartModule((unsigned int)MemMan_Copy)) ||
        (ret = SysModManager_StartModule((unsigned int)StringMan_StrToInt))) {
        DEBUG_MSG_PRE(ret);
        return 1;
    }
    return 0;
}
