/*
*****************************************
* This file implements the FSManager interface
*
* Version: 0.3.2
* Author: Luiz Henrique Shigunov
* E-mail: luizshigunov@users.sourceforge.net
* WEB: http://www.sourceforge.net/projects/modulos
* Last change: 17/06/2005
*
* See the interface specification to know what each function does.
*
* Copyright (C) 2000-2005, Luiz Henrique Shigunov
*
* This module is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2.1 of the License, or (at your option) any later version.
*
* This module is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this module; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*****************************************
*/

/* no dummy types */
#define __NO_FSMANAGER_TYPES 1

#include <stddef.h>
#include <cfgmanager.h>
#include <debug.h>
#include <errormanager.h>
#include <filesystem.h>
#include <fsmanager.h>
#include <memman.h>
#include <memmanager.h>
#include <memmanager_i386.h>
#include <sysmodmanager.h>
#include <taskmanager.h>
#include <userspacemanager.h>

/* most used first */
#define FSManager_READ_FUNC      0x00
#define FSManager_WRITE_FUNC     0x01
#define FSManager_SEEK_FUNC      0x02
#define FSManager_OPEN_FUNC      0x03
#define FSManager_CLOSE_FUNC     0x04
#define FSManager_STATUS_FUNC    0x05
#define FSManager_STATUSFD_FUNC  0x06
#define FSManager_REWINDDIR_FUNC 0x07
#define FSManager_REMOVE_FUNC    0x08
#define FSManager_READLINK_FUNC  0x09
#define FSManager_LOOKUP_FUNC    0x0a

#define FSManager_HASHED 0x01

#define FSManager_HASHBITS 12
#define FSManager_HASHMASK (0xfff>>2)

#define FSManager_LOOKUP_LAST   0x01
#define FSManager_LOOKUP_FOLLOW 0x02
#define FSManager_LOOKUP_DIR    0x04
#define FSManager_LOOKUP_PARENT 0x08

#define FSManager_LAST_ROOT   0x01
#define FSManager_LAST_NORM   0x02
#define FSManager_LAST_DOT    0x03
#define FSManager_LAST_DOTDOT 0x04

#define FSManager_IS_LINK(mode) ((mode & 0170000)==0120000)
#define FSManager_IS_DIR(mode)  ((mode & 0170000)==0040000)
#define FSManager_IS_FILE(mode) ((mode & 0170000)==0100000)

typedef struct {
    FSManager_Dentry *dentry;
    FSManager_Dentry *root;
    unsigned int flags;
    unsigned int linkCount;
    FSManager_ExStr last;
    int lastType;
} FSManager_LookupData;

#define FSManager_NUM_FUNCTIONS  0x0b

typedef struct FSManager__MountPoint {
    struct FSManager__MountPoint *next;
    FSManager_FS *fs;
    FSManager_Dentry *dentry;   /* dentry where fs was mounted */
    unsigned int unit;
    SysModManager_Imp *devImp;
} FSManager_MountPoint;

struct _FSManager_FS {
    struct _FSManager_FS *next;
    void *functions[FSManager_NUM_FUNCTIONS];
    unsigned int count;
    SysModManager_Imp *fsImp;
};

struct _FSManager_Handle {
    FileSystem_Handle *fsData;
    FSManager_FS *fs;
};

#define FSManager_NUM_FDS 10

typedef struct {
    FSManager_Dentry *cwd;
    FSManager_Dentry *root;
    TaskManager_Semaphore lock;
    unsigned int numFDs;
    struct _FSManager_Handle *fds[FSManager_NUM_FDS];
    struct _FSManager_Handle **fds1;
} FSManager_UserData;

SysModManager_Module *ID;

MemManager_Cache *handlerCache;
MemManager_Cache *dentryCache;

TaskManager_Semaphore dentrySem = {1, NULL};
FSManager_Dentry **dentryHashTable;
FSManager_Dentry *firstUnusedDentry;
FSManager_Dentry *lastUnusedDentry;

TaskManager_Semaphore mountSem = {1, NULL};
FSManager_Dentry systemRoot;

FSManager_MountPoint *mps;
FSManager_FS *fss;

unsigned int funcNum[FSManager_NUM_FUNCTIONS] = {
FileSystem__READ,
FileSystem__WRITE,
FileSystem__SEEK,
FileSystem__OPEN,
FileSystem__CLOSE,
FileSystem__STATUS,
FileSystem__STATUSFD,
FileSystem__REWINDDIR,
FileSystem__REMOVE,
FileSystem__READLINK,
FileSystem__LOOKUP
};

static FSManager_Dentry *FSManager_CachedLookup(FSManager_Dentry*,FSManager_ExStr*);
static void FSManager_DPut(FSManager_Dentry*);
static int FSManager_Lookup(const char*,FSManager_LookupData*);

static __inline__ FSManager_Dentry *FSManager_DGet(FSManager_Dentry *dentry) {
#ifdef DEBUG_ASSERT
    if (!dentry->count)
        ASSERT_MSG(("FSManager: DGet on an unused dentry! Address of caller %x\n", __builtin_return_address(0)));
#endif
    TaskManager_AtomicInc(&dentry->count);
    return dentry;
}

static __inline__ FSManager_Dentry *FSManager_DGetUnused(FSManager_Dentry *dentry) {
    TaskManager_AtomicInc(&dentry->count);
    if (dentry->count == 1) {
        /* remove from lru list */
        if (dentry->lruPrior && dentry->lruNext) {
            /* we are in the middle of the list */
            dentry->lruPrior->lruNext = dentry->lruNext;
            dentry->lruNext->lruPrior = dentry->lruPrior;
        } else if (dentry->lruPrior) {
            /* we are the last one */
            lastUnusedDentry = dentry->lruPrior;
            dentry->lruPrior->lruNext = NULL;
        } else {
            /* we are the first one */
            if (dentry->lruNext)
                dentry->lruNext->lruPrior = NULL;
            firstUnusedDentry = dentry->lruNext;
        }
    }
    return dentry;
}

static __inline__ int FSManager_AllocDentry(FSManager_Dentry *parent, FSManager_ExStr *name, FSManager_Dentry **newDentry) {
    char *str;
    FSManager_Dentry *dentry;
    int ret;

    dentry = MemManager_AllocObj(dentryCache);
    if (IS_PTR_ERROR(dentry))
        return PTR_TO_ERROR(dentry);
    if (name->len > FSMANAGER_INAME_LEN - 1) {
        str = MemManager_Alloc(name->len+1, MemManager_BLOCK_MEM);
        if (IS_PTR_ERROR(str)) {
            ret = MemManager_FreeObj(dentryCache, dentry);
            ASSERT_ERROR(ret);
            return PTR_TO_ERROR(str);
        }
    } else
        str = dentry->iname;
    MemMan_Copy(str, name->str, name->len);
    str[name->len] = 0;
    dentry->count = 1;
    dentry->flags = 0;
    dentry->parent = FSManager_DGet(parent);
    /* change parent version */
    dentry->parent->version++;
    dentry->fs = parent->fs;
    dentry->mounted = NULL;
    dentry->version = 0;
    dentry->lock.count = 1;
    dentry->lock.data = NULL;
    dentry->name.str = str;
    dentry->name.len = name->len;
    dentry->name.hash = name->hash;
    dentry->fsData = NULL;
    *newDentry = dentry;
    return E_OK;
}

static __inline__ void FSManager_FreeDentry(FSManager_Dentry *dentry) {
    int ret;

    FSManager_DPut(dentry->parent);
    if (dentry->name.str != dentry->iname)
        MemManager_Free((void*)dentry->name.str);
    ret = MemManager_FreeObj(dentryCache, dentry);
    ASSERT_ERROR(ret);
}

static __inline__ FSManager_Dentry **FSManager_GetHashEntry(FSManager_Dentry *parent, unsigned int hash) {
    hash += (unsigned int)parent >> 5;
    hash = hash ^ (hash >> FSManager_HASHBITS);
    return dentryHashTable + (hash & FSManager_HASHMASK);
}

static __inline__ void FSManager_LookupEnd(FSManager_LookupData *ld) {
    FSManager_DPut(ld->dentry);
    FSManager_DPut(ld->root);
}

static __inline__ void FSManager_LookupInit(const char *path, unsigned int flags, FSManager_LookupData *ld) {
    FSManager_UserData *ud;

    ud = TaskManager_GetTaskInfo(NULL, ID);
    if (IS_PTR_ERROR(ud))
        ud = NULL;
    ld->lastType = FSManager_LAST_ROOT;
    ld->flags = flags;
    ld->linkCount = 0;
    if (ud)
        ld->root = FSManager_DGet(ud->root);
    else
        ld->root = FSManager_DGet(&systemRoot);
    if (*path == '/' || !ud)
        ld->dentry = FSManager_DGet(ld->root);
    else
        ld->dentry = FSManager_DGet(ud->cwd);
}

static __inline__ void FSManager_LookupReInit(FSManager_LookupData *ld) {
    FSManager_UserData *ud;

    ud = TaskManager_GetTaskInfo(NULL, ID);
    if (IS_PTR_ERROR(ud))
        ud = NULL;
    ld->lastType = FSManager_LAST_ROOT;
    if (ud) {
        ld->root = FSManager_DGet(ud->root);
        ld->dentry = FSManager_DGet(ud->root);
    } else {
        ld->root = FSManager_DGet(&systemRoot);
        ld->dentry = FSManager_DGet(&systemRoot);
    }
}

static FSManager_Dentry *FSManager_CachedLookup(FSManager_Dentry *parent, FSManager_ExStr *name) {
    FSManager_Dentry **head, *tmp, *ret;

    head = FSManager_GetHashEntry(parent, name->hash);
    TaskManager_P(&dentrySem, 0, 0);
    tmp = *head;
    ret = NULL;
    while (tmp) {
        FSManager_Dentry *dentry;
        dentry = tmp;
        tmp = tmp->hashNext;
        if (dentry->name.hash != name->hash)
            continue;
        if (dentry->parent != parent)
            continue;
        if (dentry->name.len != name->len)
            continue;
        if (MemMan_Compare(dentry->name.str, name->str, name->len))
            continue;
        FSManager_DGetUnused(dentry);
        ret = dentry;
        break;
    }
    TaskManager_V(&dentrySem, 0);
    return ret;
}

int FSManager_Close(FSManager_Handle *fd) {
    void (*func)(FileSystem_Handle*);
    int ret;

#ifdef DEBUG_ASSERT
    /* it should not happen with system modules! */
    if (!fd) {
        ASSERT_MSG_PRE(0);
        return E_BAD_VALUE;
    }
#endif
    func = fd->fs->functions[FSManager_CLOSE_FUNC];
    func(fd->fsData);
    ret = MemManager_FreeObj(handlerCache, fd);
    ASSERT_ERROR(ret);
    return E_OK;
}

static void FSManager_DPut(FSManager_Dentry *dentry) {
#ifdef DEBUG_ASSERT
    if (!dentry->count) {
        ASSERT_MSG(("FSManager: DPut on an unused dentry! Address of caller %x\n", __builtin_return_address(0)));
        return;
    }
#endif
repeat:
    TaskManager_P(&dentrySem, 0, 0);
    TaskManager_AtomicDec(&dentry->count);
    if (dentry->count)
        goto returnV;
    /* if not in hash table, free dentry */
    if (!(dentry->flags & FSManager_HASHED)) {
       FSManager_Dentry *parent;
       TaskManager_V(&dentrySem, 0);
       parent = dentry->parent;
       FSManager_FreeDentry(dentry);
       dentry = parent;
       goto repeat;
    }
    /* if in hash table, put it in lru list */
    if (!firstUnusedDentry) {
        firstUnusedDentry = dentry;
        dentry->lruPrior = NULL;
    } else {
        lastUnusedDentry->lruNext = dentry;
        dentry->lruPrior = lastUnusedDentry;
    }
    lastUnusedDentry = dentry;
    dentry->lruNext = NULL;
returnV:
    TaskManager_V(&dentrySem, 0);
}

static __inline__ int FSManager_FollowLink(FSManager_LookupData *ld) {
    char *link;
    int (*readlink)(FSManager_Dentry*,char*,unsigned int,int);
    int error, ret;

    error = E_LOOP;
    if (++ld->linkCount >= 8)
        goto returnDecLink;
    link = MemManager_i386_AllocPages(1, MemManager_i386_RW_PAGE+MemManager_i386_ALLOC_PHYS);
    if (IS_PTR_ERROR(link)) {
        error = PTR_TO_ERROR(link);
        goto returnDecLink;
    }
    readlink = ld->dentry->fs->functions[FSManager_READLINK_FUNC];
    error = readlink(ld->dentry, link, 4096, 0);
    if (error)
        goto returnFreeLink;
    if (*link == '/') {
        FSManager_LookupEnd(ld);
        FSManager_LookupReInit(ld);
    }
    error = FSManager_Lookup(link, ld);
    ret = MemManager_i386_FreePages(link, 1, 0);
    ASSERT_ERROR(ret);
    ld->linkCount--;
    return error;

returnFreeLink:
    ret = MemManager_i386_FreePages(link, 1, 0);
    ASSERT_ERROR(ret);
returnDecLink:
    FSManager_LookupEnd(ld);
    ld->linkCount--;
    return error;
}

static __inline__ void FSManager_InsertDentry(FSManager_Dentry *dentry) {
    FSManager_Dentry **entry;

    entry = FSManager_GetHashEntry(dentry->parent, dentry->name.hash);
    TaskManager_P(&dentrySem, 0, 0);
    dentry->flags |= FSManager_HASHED;
    dentry->hashNext = *entry;
    dentry->hashPrior = NULL;
    if (*entry)
        (*entry)->hashPrior = dentry;
    *entry = dentry;
    TaskManager_V(&dentrySem, 0);
}

static int FSManager_Lookup(const char *path, FSManager_LookupData *ld) {
    int error;
    unsigned int lookupFlags = ld->flags;

    while (*path == '/')
        path++;
    if (!*path)
        goto returnBase;
    do {
        FSManager_ExStr comp;
        unsigned int c, parentVersion;
        FSManager_Dentry *dentry;

        comp.str = path;
        c = *(unsigned char*)path;
        comp.hash = 0;
        do {
            path++;
            comp.hash = (comp.hash + (c << 4) + (c >> 4)) * 11;
            c = *(unsigned char *)path;
        } while (c && c != '/');
        comp.len = path - comp.str;
        /* remove trailing slashes */
        while (*path == '/')
            path++;
        /* last component? */
        if (!*path) {
            lookupFlags |= FSManager_LOOKUP_LAST;
            if (*(path-1) == '/')
                lookupFlags |= FSManager_LOOKUP_FOLLOW | FSManager_LOOKUP_DIR;
        }
        /* wants parent */
        if ((lookupFlags & FSManager_LOOKUP_PARENT) && (lookupFlags & FSManager_LOOKUP_LAST)) {
            ld->last = comp;
            ld->lastType = FSManager_LAST_NORM;
            if (comp.str[0] != '.')
                goto returnBase;
            if (comp.len == 1)
                ld->lastType = FSManager_LAST_DOT;
            else if (comp.len == 2 && comp.str[1] == '.')
                ld->lastType = FSManager_LAST_DOTDOT;
            goto returnBase;
        }
        /* handle special cases "." and ".." */
        if (comp.str[0] == '.')
            switch (comp.len) {
                default:
                    break;
                case 2:	
                    if (comp.str[1] != '.')
                        break;
                    if (ld->dentry != ld->root) {
                        FSManager_Dentry *tmp;
                        tmp = FSManager_DGet(ld->dentry->parent);
                        FSManager_DPut(ld->dentry);
                        ld->dentry = tmp;
                    }
                case 1:
                    continue;
            }
        parentVersion = ld->dentry->version;
        dentry = FSManager_CachedLookup(ld->dentry, &comp);
        if (!dentry) {
            /* alloc a dentry and call fs lookup function */
            int (*fsLookup)(FSManager_Dentry*,FSManager_Dentry*);
            /* lock parent dir */
            TaskManager_P(&ld->dentry->lock, 0, 0);
            /* parent changed while we waited for the lock? */
            if (ld->dentry->version != parentVersion) {
                /* if so, redo cache lookup */
                dentry = FSManager_CachedLookup(ld->dentry, &comp);
                if (dentry)
                    goto freeDirLock;
            }
            error = FSManager_AllocDentry(ld->dentry, &comp, &dentry);
            if (error)
                goto freeDirLockError;
            fsLookup = ld->dentry->fs->functions[FSManager_LOOKUP_FUNC];
            error = fsLookup(ld->dentry, dentry);
            if (error) {
                FSManager_FreeDentry(dentry);
freeDirLockError:
                TaskManager_V(&ld->dentry->lock, 0);
                goto returnLookupEnd;
            }
            /* insert dentry in the cache */
            FSManager_InsertDentry(dentry);
freeDirLock:
            TaskManager_V(&ld->dentry->lock, 0);
        }
        FSManager_DPut(ld->dentry);
        /* follow mount point */
        if (dentry->mounted) {
            FSManager_Dentry *tmp;
            tmp = FSManager_DGet(dentry->mounted);
            FSManager_DPut(dentry);
            dentry = tmp;
        }
        ld->dentry = dentry;
        if (FSManager_IS_LINK(ld->dentry->mode) && (!(lookupFlags & FSManager_LOOKUP_LAST) || (lookupFlags & FSManager_LOOKUP_FOLLOW))) {
            error = FSManager_FollowLink(ld);
            if (error)
                goto returnError;
        }
        error = E_NOT_A_DIRECTORY;
        if (!FSManager_IS_DIR(ld->dentry->mode) && (!(lookupFlags & FSManager_LOOKUP_LAST) || (lookupFlags & FSManager_LOOKUP_DIR)))
            goto returnLookupEnd;
    } while (!(lookupFlags & FSManager_LOOKUP_LAST));

returnBase:
    return E_OK;

returnLookupEnd:
    FSManager_LookupEnd(ld);
returnError:
    return error;
}

int FSManager_Mount(const char *fsImpName, const char *devImpName, unsigned int unit, const char *path, int prop, const char *data) {
    int error, ret;
    unsigned int i;
    int (*mountF)(const char*,unsigned int,int,const char*,FSManager_Dentry*);
    int (*unmountF)(FSManager_Dentry*,int);
    FSManager_MountPoint *mp;
    FSManager_FS *fs;
    FSManager_LookupData ld;
    FSManager_Dentry *dentry;
    SysModManager_Imp *fsImp, *devImp;

#ifdef DEBUG_ASSERT
    /* it should not happen with system modules! */
    if (!fsImpName || !devImpName || !path || !*path) {
        ASSERT_MSG_PRE(0);
        return E_BAD_VALUE;
    }
#endif

    FSManager_LookupInit(path, FSManager_LOOKUP_FOLLOW | FSManager_LOOKUP_DIR, &ld);
    error = FSManager_Lookup(path, &ld);
    if (error)
        goto errorEnd;
    error = E_MODULE;
    fsImp = SysModManager_GetImpID("FileSystem", fsImpName);
    if (IS_PTR_ERROR(fsImp)) {
        goto errorEnd;
    }
    devImp = SysModManager_GetImpID("BlockDev", devImpName);
    if (IS_PTR_ERROR(devImp)) {
        goto errorPutFsImp;
    }
    /* only one thread can mount a fs each time */
    TaskManager_P(&mountSem, 0, 0);
    /* dir already mounted? */
    error = E_IN_USE;
    if (!ld.dentry->name.str)
        goto errorUnlock;
    /* device/unit already mounted? */
    mp = mps;
    while (mp && (mp->unit != unit || mp->devImp != devImp))
        mp = mp->next;
    if (mp) {
        /* different fs type? */
        error = E_BAD_TYPE;
        if (mp->fs->fsImp != fsImp)
            goto errorUnlock;
        /* just add dentry */
        dentry = FSManager_DGet(mp->dentry->mounted);
        goto alreadyMounted;
    }
    /* alloc root dentry */
    dentry = MemManager_AllocObj(dentryCache);
    if (IS_PTR_ERROR(dentry)) {
        error = PTR_TO_ERROR(dentry);
        goto errorUnlock;
    }
    error = E_MODULE;
    mountF = SysModManager_GetFunction(fsImp, FileSystem__MOUNT, 0, NULL);
    if (IS_PTR_ERROR(mountF))
        goto errorFreeDentry;
    error = mountF(devImpName, unit, prop & 0xff, data, dentry);
    if (error)
        goto errorFreeDentry;
    mp = MemManager_Alloc(sizeof(FSManager_MountPoint), 0);
    if (IS_PTR_ERROR(mp)) {
        error = PTR_TO_ERROR(mp);
        goto errorUnmount;
    }
    /* FS already in register? */
    fs = fss;
    while (fs && fs->fsImp != fsImp)
        fs = fs->next;
    if (!fs) {
        fs = MemManager_Alloc(sizeof(FSManager_FS), 0);
        if (IS_PTR_ERROR(fs)) {
            error = PTR_TO_ERROR(fs);
            goto errorAllocFS;
        }
        MemMan_Set(&fs->functions, 0, sizeof(void*)*FSManager_NUM_FUNCTIONS);
        i = 0;
        error = E_MODULE;
        do {
            fs->functions[i] = SysModManager_GetFunction(fsImp, funcNum[i], 0, NULL);
            if (IS_PTR_ERROR(fs->functions[i]))
                goto errorGetFunctions;
        } while (++i < FSManager_NUM_FUNCTIONS);
        fs->count = 0;
        fs->fsImp = fsImp;
        fs->next = fss;
        fss = fs;
    }
    fs->count++;
    mp->fs = fs;
    mp->dentry = ld.dentry;
    mp->unit = unit;
    mp->devImp = devImp;
    mp->next = mps;
    mps = mp;
    /* start root dentry */
    dentry->mode = 0040000;
    dentry->count = 1;
    dentry->parent = FSManager_DGet(ld.dentry->parent);
    dentry->fs = fs;
    dentry->mounted = NULL;
    dentry->version = 0;
    dentry->lock.count = 1;
    dentry->lock.data = NULL;
    /* root entry does not have name */
    dentry->name.str = NULL;
    dentry->name.len = 0;
    dentry->name.hash = 0;
alreadyMounted:
    ld.dentry->mounted = dentry;
    error = E_OK;
    goto errorUnlock;

errorGetFunctions:
    ret = MemManager_Free(fs);
    ASSERT_ERROR(ret);
errorAllocFS:
    ret = MemManager_Free(mp);
    ASSERT_ERROR(ret);
errorUnmount:
    unmountF = SysModManager_GetFunction(fsImp, FileSystem__UNMOUNT, 0, NULL);
    if (!IS_PTR_ERROR(unmountF)) {
        unmountF(dentry, 0x01);
    }
errorFreeDentry:
    ret = MemManager_FreeObj(dentryCache, dentry);
    ASSERT_ERROR(ret);
errorUnlock:
    TaskManager_V(&mountSem, 0);
    FSManager_LookupEnd(&ld);
    ret = SysModManager_PutImpID(devImp);
    ASSERT_ERROR(ret);
errorPutFsImp:
    ret = SysModManager_PutImpID(fsImp);
    ASSERT_ERROR(ret);
errorEnd:
    return error;
}

int FSManager_Open(const char *pathname, int prop, FSManager_Handle **fd) {
    FSManager_Handle *newFD;
    FileSystem_Handle *newHandle;
    int error;
    unsigned int flags;
    int (*func)(FSManager_Dentry*,int,FileSystem_Handle**);
    FSManager_LookupData ld;

#ifdef DEBUG_ASSERT
    /* it should not happen with system modules! */
    if (!pathname || !*pathname || !(prop & (FSManager_READ + FSManager_WRITE))) {
        ASSERT_MSG_PRE(0);
        return E_BAD_VALUE;
    }
#endif

    /* we only support read now */
    error = E_READ_ONLY_FS;
    if (prop & 0x3e)
        goto errorEnd;
    /* follow links? */
    flags = FSManager_LOOKUP_FOLLOW;
    if (prop & FSManager_NOFOLLOW)
        flags = 0;
    /* want's dir? */
    if (prop & FSManager_DIRECTORY)
        flags |= FSManager_LOOKUP_DIR;
    FSManager_LookupInit(pathname, flags, &ld);
    error = FSManager_Lookup(pathname, &ld);
    if (error)
        goto errorEnd;
    /* is it a file or dir? */
    error = E_NOT_FILE_DIR;
    if (!FSManager_IS_FILE(ld.dentry->mode) && !FSManager_IS_DIR(ld.dentry->mode))
        goto errorLookupEnd;
    func = ld.dentry->fs->functions[FSManager_OPEN_FUNC];
    error = func(ld.dentry, prop, &newHandle);
    if (error)
        goto errorLookupEnd;
    newFD = MemManager_AllocObj(handlerCache);
    if (IS_PTR_ERROR(newFD)) {
        ((void(*)(FileSystem_Handle*))(ld.dentry->fs->functions[FSManager_CLOSE_FUNC]))(newHandle);
        error = PTR_TO_ERROR(newFD);
        goto errorLookupEnd;
    }
    newFD->fs = ld.dentry->fs;
    newFD->fsData = newHandle;
    *fd = newFD;
errorLookupEnd:
    FSManager_LookupEnd(&ld);
errorEnd:
    return error;
}

int FSManager_Read(FSManager_FDIO *request) {
    int (*func)(FSManager_FDIO*);

#ifdef DEBUG_ASSERT
    /* it should not happen with system modules! */
    if (!request || !request->fd) {
        ASSERT_MSG_PRE(0);
        return E_BAD_VALUE;
    }
#endif
    func = request->fd->fs->functions[FSManager_READ_FUNC];
    return func(request);
}

int FSManager_Remove(const char *pathname) {
    /* not supported yet */
    return E_READ_ONLY_FS;
}

int FSManager_RewindDir(FSManager_Handle *fd) {
    int (*func)(FileSystem_Handle*);

#ifdef DEBUG_ASSERT
    /* it should not happen with system modules! */
    if (!fd) {
        ASSERT_MSG_PRE(0);
        return E_BAD_VALUE;
    }
#endif
    func = fd->fs->functions[FSManager_REWINDDIR_FUNC];
    return func(fd->fsData);
}

int FSManager_Seek(FSManager_Handle *fd, int count, int prop, int *newPos) {
    int (*func)(FileSystem_Handle*,int,int,int*);

#ifdef DEBUG_ASSERT
    /* it should not happen with system modules! */
    if (!fd) {
        ASSERT_MSG_PRE(0);
        return E_BAD_VALUE;
    }
#endif
    func = fd->fs->functions[FSManager_SEEK_FUNC];
    return func(fd->fsData, count, prop, newPos);
}

int FSManager_Status(const char *pathname, FSManager_EntryStatus *buf, int prop) {
    int error;
    int (*func)(FSManager_Dentry*,FSManager_EntryStatus*,int);
    unsigned int flags;
    FSManager_LookupData ld;

#ifdef DEBUG_ASSERT
    /* it should not happen with system modules! */
    if (!pathname || !*pathname) {
        ASSERT_MSG_PRE(0);
        return E_BAD_VALUE;
    }
#endif

    /* follow links? */
    flags = FSManager_LOOKUP_FOLLOW;
    if (prop & FSManager_STATUS_LINK)
        flags = 0;
    FSManager_LookupInit(pathname, flags, &ld);
    error = FSManager_Lookup(pathname, &ld);
    if (error)
        goto errorEnd;
    func = ld.dentry->fs->functions[FSManager_STATUS_FUNC];
    error = func(ld.dentry, buf, prop);
    FSManager_LookupEnd(&ld);
errorEnd:
    return error;
}

int FSManager_StatusFD(FSManager_Handle *fd, FSManager_EntryStatus *buf, int prop) {
    int (*func)(FileSystem_Handle*,FSManager_EntryStatus*,int);

#ifdef DEBUG_ASSERT
    /* it should not happen with system modules! */
    if (!fd) {
        ASSERT_MSG_PRE(0);
        return E_BAD_VALUE;
    }
#endif

    func = fd->fs->functions[FSManager_STATUSFD_FUNC];
    return func(fd->fsData, buf, prop);
}

int FSManager_Unmount(const char *path, int prop) {
    return E_OK;
}

int FSManager_Write(FSManager_FDIO *request) {
    int (*func)(FSManager_FDIO*);

#ifdef DEBUG_ASSERT
    /* it should not happen with system modules! */
    if (!request || !request->fd) {
        ASSERT_MSG_PRE(0);
        return E_BAD_VALUE;
    }
#endif
    func = request->fd->fs->functions[FSManager_WRITE_FUNC];
    return func(request);
}

void FSManager_TaskDestroyed(TaskManager_Task *task) {
    FSManager_UserData *ud;
    unsigned int i;
    int ret;

    ud = TaskManager_DelTaskInfo(task, ID);
    if (IS_PTR_ERROR(ud)) {
        DEBUG_MSG_PRE(PTR_TO_ERROR(ud));
        return;
    }
    FSManager_DPut(ud->cwd);
    FSManager_DPut(ud->root);
    /* close all open files/dirs */
    for (i = 0; ud->numFDs && i < FSManager_NUM_FDS; i++)
        if (ud->fds[i]) {
            ret = FSManager_Close(ud->fds[i]);
            ASSERT_ERROR(ret);
            ud->numFDs--;
        }
    if (ud->fds1) {
        for (i = 0; ud->numFDs; i++)
            if (ud->fds1[i]) {
                ret = FSManager_Close(ud->fds1[i]);
                ASSERT_ERROR(ret);
                ud->numFDs--;
            }
        ret = MemManager_i386_FreePages(ud->fds1, 1, 0);
        ASSERT_ERROR(ret);
    }
    ret = MemManager_Free(ud);
    ASSERT_ERROR(ret);
}

void FSManager_TaskCreated(TaskManager_Task *creator, TaskManager_Task *created) {
    FSManager_UserData *ud, *cud;
    void *voidRet;

    ud = MemManager_Alloc(sizeof(FSManager_UserData), 0);
    if (IS_PTR_ERROR(ud)) {
        DEBUG_MSG_PRE(PTR_TO_ERROR(ud));
        return;
    }
    MemMan_Set(ud, 0, sizeof(FSManager_UserData));
    cud = TaskManager_GetTaskInfo(creator, ID);
    /* use creator cwd and root, if it exists */
    if (!IS_PTR_ERROR(cud)) {
        ud->cwd = FSManager_DGet(cud->cwd);
        ud->root = FSManager_DGet(cud->root);
    } else {
        ud->cwd = FSManager_DGet(&systemRoot);
        ud->root = FSManager_DGet(&systemRoot);
    }
    ud->lock.count = 1;
    /* store it */
    voidRet = TaskManager_SetTaskInfo(created, ID, ud, 0);
    ASSERT_ERROR_PTR(voidRet);
}

/* user module functions */

static int FSManager_UAllocFD(FSManager_Handle*);

static __inline__ FSManager_Handle *FSManager_UGetFileHandler(int fd) {
    FSManager_UserData *ud;
    FSManager_Handle *handler;

    handler = NULL;
    ud = TaskManager_GetTaskInfo(NULL, ID);
    if (!IS_PTR_ERROR(ud)) {
        TaskManager_P(&ud->lock, 0, 0);
        if (fd < FSManager_NUM_FDS)
            handler = ud->fds[fd];
        else {
            fd -= FSManager_NUM_FDS;
            if (fd < 1024 && ud->fds1)
                handler = ud->fds1[fd];
        }
        TaskManager_V(&ud->lock, 0);
    }
    return handler;
}

static __inline__ int FSManager_UGetUserStr(const char *uStr, const char **sysStr) {
    char *page;
    int error, ret;
    unsigned int len;

    error = E_BAD_VALUE;
    if (!uStr)
        goto returnError;
    page = MemManager_i386_AllocPages(1, MemManager_i386_RW_PAGE+MemManager_i386_ALLOC_PHYS);
    if (IS_PTR_ERROR(page)) {
        error = PTR_TO_ERROR(page);
        goto returnError;
    }
    error = UserSpaceManager_StrNCopy(page, uStr, 4096, &len);
    if (!error) {
        if (len != 4096) {
            *sysStr = page;
            goto returnError;
        }
        error = E_TOO_BIG;
    }
    ret = MemManager_i386_FreePages(page, 1, 0);
    ASSERT_ERROR(ret);
returnError:
    return error;
}

static __inline__ void FSManager_UPutUserStr(const char *str) {
    int ret;

    ret = MemManager_i386_FreePages((void*)str, 1, 0);
    ASSERT_ERROR(ret);
}

static int FSManager_UAllocFD(FSManager_Handle *fd) {
    FSManager_UserData *ud;
    int error, i;

    ud = TaskManager_GetTaskInfo(NULL, ID);
    if (IS_PTR_ERROR(ud)) {
        /* too bad, task was created but no UserData */
        error = PTR_TO_ERROR(ud);
        goto returnError;
    }
    error = E_NO_MEMORY;
    /* find a free fd */
    TaskManager_P(&ud->lock, 0, 0);
    for (i = 0; i < FSManager_NUM_FDS; i++)
        if (!ud->fds[i]) {
            ud->fds[i] = fd;
            ud->numFDs++;
            error = -i;
            goto returnV;
        }
    if (!ud->fds1) {
        ud->fds1 = MemManager_i386_AllocPages(1, MemManager_i386_RW_PAGE+MemManager_i386_ALLOC_PHYS+MemManager_i386_CLEAR_PAGES);
        if (IS_PTR_ERROR(ud->fds1)) {
            error = PTR_TO_ERROR(ud->fds1);
            goto returnV;
        }
    }
    error = E_TOO_MANY_OPEN_FILES;
    for (i = 0; i < 1024; i++)
        if (!ud->fds1[i]) {
            ud->fds1[i] = fd;
            ud->numFDs++;
            error = -i;
            break;
        }
returnV:
    TaskManager_V(&ud->lock, 0);
returnError:
    return error;
}

int FSManager_UClose(int fd) {
    FSManager_UserData *ud;
    FSManager_Handle *handler;

    handler = NULL;
    ud = TaskManager_GetTaskInfo(NULL, ID);
    if (!IS_PTR_ERROR(ud)) {
        TaskManager_P(&ud->lock, 0, 0);
        if (fd < FSManager_NUM_FDS) {
            handler = ud->fds[fd];
            ud->numFDs--;
            ud->fds[fd] = NULL;
        } else {
            fd -= FSManager_NUM_FDS;
            if (fd < 1024 && ud->fds1) {
                handler = ud->fds1[fd];
                ud->numFDs--;
                ud->fds1[fd] = NULL;
            }
        }
        TaskManager_V(&ud->lock, 0);
    }
    if (!handler)
        return E_BAD_VALUE;
    return FSManager_Close(handler);
}

int FSManager_UChangeDir(const char *path) {
    const char *sysPath;
    int error;
    FSManager_LookupData ld;
    FSManager_UserData *ud;

    error = FSManager_UGetUserStr(path, &sysPath);
    if (error)
        goto returnError;
    FSManager_LookupInit(sysPath, FSManager_LOOKUP_FOLLOW | FSManager_LOOKUP_DIR, &ld);
    error = FSManager_Lookup(sysPath, &ld);
    if (error)
        goto returnPutStr;
    ud = TaskManager_GetTaskInfo(NULL, ID);
    if (IS_PTR_ERROR(ud)) {
        /* too bad, task was created but no UserData */
        error = PTR_TO_ERROR(ud);
        goto returnLookupEnd;
    }
    /* we already have a cwd */
    TaskManager_P(&ud->lock, 0, 0);
    FSManager_DPut(ud->cwd);
    ud->cwd = ld.dentry;
    TaskManager_V(&ud->lock, 0);
    /* we do not need to do LookupEnd! */
    FSManager_DPut(ld.root);
    error = E_OK;
    goto returnPutStr;

returnLookupEnd:
    FSManager_LookupEnd(&ld);
returnPutStr:
    FSManager_UPutUserStr(sysPath);
returnError:
    return error;
}

int FSManager_UChangeRoot(const char *path) {
    const char *sysPath;
    int error;
    FSManager_LookupData ld;
    FSManager_UserData *ud;

    error = FSManager_UGetUserStr(path, &sysPath);
    if (error)
        goto returnError;
    FSManager_LookupInit(sysPath, FSManager_LOOKUP_FOLLOW | FSManager_LOOKUP_DIR, &ld);
    error = FSManager_Lookup(sysPath, &ld);
    if (error)
        goto returnPutStr;
    ud = TaskManager_GetTaskInfo(NULL, ID);
    if (IS_PTR_ERROR(ud)) {
        /* too bad, task was created but no UserData */
        error = PTR_TO_ERROR(ud);
        goto returnLookupEnd;
    }
    /* we already have a root */
    TaskManager_P(&ud->lock, 0, 0);
    FSManager_DPut(ud->root);
    ud->root = ld.dentry;
    TaskManager_V(&ud->lock, 0);
    /* we do not need to do LookupEnd! */
    FSManager_DPut(ld.root);
    error = E_OK;
    goto returnPutStr;

returnLookupEnd:
    FSManager_LookupEnd(&ld);
returnPutStr:
    FSManager_UPutUserStr(sysPath);
returnError:
    return error;
}

int FSManager_UMount(const char *fsImp, const char *devImp, unsigned int unit, const char *path, int prop, const char *data) {
    const char *uFsImp, *uDevImp, *uPath, *uData;
    int error;

    error = FSManager_UGetUserStr(fsImp, &uFsImp);
    if (error)
        goto returnError;
    error = FSManager_UGetUserStr(devImp, &uDevImp);
    if (error)
        goto returnPutFsImp;
    error = FSManager_UGetUserStr(path, &uPath);
    if (error)
        goto returnPutDevImp;
    error = E_BAD_VALUE;
    if (!*uPath)
        goto returnPutPath;
    error = FSManager_UGetUserStr(data, &uData);
    if (error)
        goto returnPutPath;
    error = FSManager_Mount(uFsImp, uDevImp, unit, uPath, prop, uData);
    FSManager_UPutUserStr(uData);
returnPutPath:
    FSManager_UPutUserStr(uPath);
returnPutDevImp:
    FSManager_UPutUserStr(uDevImp);
returnPutFsImp:
    FSManager_UPutUserStr(uFsImp);
returnError:
    return error;
}

int FSManager_UOpen(const char *pathname, int prop) {
    const char *uPathname;
    int error, ret;
    FSManager_Handle *fd;

    error = FSManager_UGetUserStr(pathname, &uPathname);
    if (error)
        goto returnError;
    error = E_BAD_VALUE;
    if (!*uPathname || !(prop & (FSManager_READ + FSManager_WRITE)))
        goto returnPutStr;
    error = FSManager_Open(uPathname, prop, &fd);
    if (!error) {
        /* ok, now alloc a fd slot */
        error = FSManager_UAllocFD(fd);
        if (error > 0) {
            /* something wrong. close it */
            ret = FSManager_Close(fd);
            ASSERT_ERROR(ret);
        }
    }
returnPutStr:
    FSManager_UPutUserStr(uPathname);
returnError:
    return -error;
}

int FSManager_URead(int fd, void *buf, int size) {
    FSManager_Handle *handler;
    FSManager_FDIO request;
    int error;

    handler = FSManager_UGetFileHandler(fd);
    if (!handler || size < 0)
        return E_BAD_VALUE;
    request.fd = handler;
    request.flags = FSManager_USER_BUF_IO;
    request.buf = buf;
    request.size = size;
    error = FSManager_Read(&request);
    if (error)
        return -error;
    return request.size;
}

int FSManager_URemove(const char *pathname) {
    const char *uPathname;
    int error;

    error = FSManager_UGetUserStr(pathname, &uPathname);
    if (error)
        goto returnError;
    error = E_BAD_VALUE;
    if (!*uPathname)
        goto returnPutStr;
    error = FSManager_Remove(uPathname);
returnPutStr:
    FSManager_UPutUserStr(uPathname);
returnError:
    return error;
}

int FSManager_URewindDir(int fd) {
    FSManager_Handle *handler;

    handler = FSManager_UGetFileHandler(fd);
    if (!handler)
        return E_BAD_VALUE;
    return FSManager_RewindDir(handler);
}

int FSManager_USeek(int fd, int count, int prop) {
    FSManager_Handle *handler;
    int error;
    int newPos;

    handler = FSManager_UGetFileHandler(fd);
    if (!handler)
        return E_BAD_VALUE;
    error = FSManager_Seek(handler, count, prop, &newPos);
    if (error)
        return -error;
    return newPos;
}

int FSManager_UStatus(const char *pathname, FSManager_EntryStatus *buf, int prop) {
    const char *uPathname;
    int error;

    error = FSManager_UGetUserStr(pathname, &uPathname);
    if (error)
        goto returnError;
    error = E_BAD_VALUE;
    if (!*uPathname)
        goto returnPutStr;
    error = FSManager_Status(uPathname, buf, prop|FSManager_USER_BUF);
returnPutStr:
    FSManager_UPutUserStr(uPathname);
returnError:
    return error;
}

int FSManager_UStatusFD(int fd, FSManager_EntryStatus *buf, int prop) {
    FSManager_Handle *handler;

    handler = FSManager_UGetFileHandler(fd);
    if (!handler)
        return E_BAD_VALUE;
    return FSManager_StatusFD(handler, buf, prop|FSManager_USER_BUF);
}

int FSManager_UUnmount(const char *path, int prop) {
    const char *uPath;
    int error;

    error = FSManager_UGetUserStr(path, &uPath);
    if (error)
        goto returnError;
    error = E_BAD_VALUE;
    if (!*uPath)
        goto returnPutStr;
    error = FSManager_Unmount(uPath, prop);
returnPutStr:
    FSManager_UPutUserStr(uPath);
returnError:
    return error;
}

int FSManager_UWrite(int fd, void *buf, int count) {
    FSManager_Handle *handler;
    FSManager_FDIO request;
    int error;

    handler = FSManager_UGetFileHandler(fd);
    if (!handler || count < 0)
        return E_BAD_VALUE;
    request.fd = handler;
    request.flags = FSManager_USER_BUF_IO;
    request.buf = buf;
    request.size = count;
    error = FSManager_Write(&request);
    if (error)
        return -error;
    return request.size;
}

static __inline__ int MountRoot(const char *fsImpName, const char *devImpName, unsigned int unit, const char *data) {
    int error, ret;
    unsigned int i;
    int (*mountF)(const char*,unsigned int,int,const char*,FSManager_Dentry*);
    int (*unmountF)(FSManager_Dentry*,int);
    FSManager_MountPoint *mountP;
    FSManager_FS *fs;
    SysModManager_Imp *fsImp, *devImp;

    error = E_MODULE;
    fsImp = SysModManager_GetImpID("FileSystem", fsImpName);
    if (IS_PTR_ERROR(fsImp)) {
        goto errorEnd;
    }
    devImp = SysModManager_GetImpID("BlockDev", devImpName);
    if (IS_PTR_ERROR(devImp)) {
        goto errorPutFsImp;
    }
    mountF = SysModManager_GetFunction(fsImp, FileSystem__MOUNT, 0, NULL);
    if (IS_PTR_ERROR(mountF)) {
        goto errorPutDevImp;
    }
    error = mountF(devImpName, unit, 0x01, data, &systemRoot);
    if (error)
        goto errorPutDevImp;
    mountP = MemManager_Alloc(sizeof(FSManager_MountPoint), 0);
    if (IS_PTR_ERROR(mountP)) {
        error = PTR_TO_ERROR(mountP);
        goto errorUnmount;
    }
    fs = MemManager_Alloc(sizeof(FSManager_FS), 0);
    if (IS_PTR_ERROR(fs)) {
        error = PTR_TO_ERROR(fs);
        goto errorFreeMountP;
    }
    MemMan_Set(&fs->functions, 0, sizeof(void*)*FSManager_NUM_FUNCTIONS);
    i = 0;
    error = E_MODULE;
    do {
        fs->functions[i] = SysModManager_GetFunction(fsImp, funcNum[i], 0, NULL);
        if (IS_PTR_ERROR(fs->functions[i]))
            goto errorGetFunctions;
    } while (++i < FSManager_NUM_FUNCTIONS);
    fs->count = 1;
    fs->fsImp = fsImp;
    fs->next = fss;
    fss = fs;
    mountP->fs = fs;
    mountP->dentry = &systemRoot;
    mountP->unit = unit;
    mountP->devImp = devImp;
    mountP->next = mps;
    mps = mountP;
    /* start root dentry */
    systemRoot.mode = 0040000;
    systemRoot.count = 1;
    systemRoot.parent = &systemRoot;
    systemRoot.fs = fs;
    systemRoot.mounted = NULL;
    systemRoot.version = 0;
    systemRoot.lock.count = 1;
    systemRoot.lock.data = NULL;
    /* root entry does not have name */
    systemRoot.name.str = NULL;
    systemRoot.name.len = 0;
    systemRoot.name.hash = 0;
    ret = SysModManager_PutImpID(devImp);
    ASSERT_ERROR(ret);
    ret = SysModManager_PutImpID(fsImp);
    ASSERT_ERROR(ret);
    return E_OK;

errorGetFunctions:
    ret = MemManager_Free(fs);
    ASSERT_ERROR(ret);
errorFreeMountP:
    ret = MemManager_Free(mountP);
    ASSERT_ERROR(ret);
errorUnmount:
    unmountF = SysModManager_GetFunction(fsImp, FileSystem__UNMOUNT, 0, NULL);
    if (!IS_PTR_ERROR(unmountF)) {
        unmountF(&systemRoot, 0x01);
    }
errorPutDevImp:
    ret = SysModManager_PutImpID(devImp);
    ASSERT_ERROR(ret);
errorPutFsImp:
    ret = SysModManager_PutImpID(fsImp);
    ASSERT_ERROR(ret);
errorEnd:
    return error;
}

int Shutdown(void) {
    int ret;

    ret = MemManager_i386_FreePages(dentryHashTable, 1, 0);
    ASSERT_ERROR(ret);
    ret = MemManager_DestroyObjCache(dentryCache);
    ASSERT_ERROR(ret);
    ret = MemManager_DestroyObjCache(handlerCache);
    ASSERT_ERROR(ret);
    return 0;
}

int Phase1Start(SysModManager_Module *modID) {
    char fsImp[64], devImp[64], data[64];
    CfgManager_Group *rootGroup;
    unsigned int unit, bufSize;
    int ret;

    ID = modID;
    /* we don't need to start UserSpaceManager because we'll use it if it is started. */
    if ((ret = SysModManager_StartModule((unsigned int)MemMan_Copy)) ||
        (ret = SysModManager_StartModule((unsigned int)MemManager_Alloc)) ||
        (ret = SysModManager_StartModule((unsigned int)CfgManager_OpenGroup)) ||
        (ret = SysModManager_StartModule((unsigned int)TaskManager_P))) {
        DEBUG_MSG_PRE(ret);
        goto returnEnd;
    }
    handlerCache = MemManager_CreateObjCache(sizeof(FSManager_Handle), 0, 0, NULL, NULL);
    if (IS_PTR_ERROR(handlerCache)) {
        DEBUG_MSG_PRE(PTR_TO_ERROR(handlerCache));
        goto returnEnd;
    }
    dentryCache = MemManager_CreateObjCache(sizeof(FSManager_Dentry), 0, 0, NULL, NULL);
    if (IS_PTR_ERROR(dentryCache)) {
        DEBUG_MSG_PRE(PTR_TO_ERROR(dentryCache));
        goto returnFreeFileCache;
    }
    dentryHashTable = MemManager_i386_AllocPages(1, MemManager_i386_RW_PAGE+MemManager_i386_ALLOC_PHYS+MemManager_i386_CLEAR_PAGES);
    if (IS_PTR_ERROR(dentryHashTable)) {
        DEBUG_MSG_PRE(PTR_TO_ERROR(dentryHashTable));
        goto returnFreeDentryCache;
    }
    /* mount root file system */
    rootGroup = CfgManager_OpenGroup(NULL, "/system/FSManager/root");
    if (!rootGroup) {
        DEBUG_MSG_PRE(0);
        goto returnFreeHashTable;
    }
    bufSize = sizeof(fsImp);
    if ((ret = CfgManager_GetValue(rootGroup, "fsImp", CfgManager_STRING, fsImp, &bufSize))) {
        DEBUG_MSG_PRE(ret);
        goto returnCloseGroup;
    }
    bufSize = sizeof(devImp);
    if ((ret = CfgManager_GetValue(rootGroup, "devImp", CfgManager_STRING, devImp, &bufSize))) {
        DEBUG_MSG_PRE(ret);
        goto returnCloseGroup;
    }
    bufSize = sizeof(data);
    if ((ret = CfgManager_GetValue(rootGroup, "data", CfgManager_STRING, data, &bufSize))) {
        DEBUG_MSG_PRE(ret);
        goto returnCloseGroup;
    }
    bufSize = sizeof(unit);
    if ((ret = CfgManager_GetValue(rootGroup, "unit", CfgManager_HEXNUMBER, &unit, &bufSize))) {
        DEBUG_MSG_PRE(ret);
        goto returnCloseGroup;
    }
    ret = CfgManager_CloseGroup(rootGroup);
    ASSERT_ERROR(ret);
    if ((ret = MountRoot(fsImp, devImp, unit, data))) {
        DEBUG_MSG_PRE(ret);
        goto returnFreeHashTable;
    }
    return 0;

returnCloseGroup:
    ret = CfgManager_CloseGroup(rootGroup);
    ASSERT_ERROR(ret);
returnFreeHashTable:
    ret = MemManager_i386_FreePages(dentryHashTable, 1, 0);
    ASSERT_ERROR(ret);
returnFreeDentryCache:
    ret = MemManager_DestroyObjCache(dentryCache);
    ASSERT_ERROR(ret);
returnFreeFileCache:
    ret = MemManager_DestroyObjCache(handlerCache);
    ASSERT_ERROR(ret);
returnEnd:
    return 1;
}
