/*
*****************************************
* This file implements the InterManager interface
*
* Version: 0.3.6
* Author: Luiz Henrique Shigunov
* E-mail: luizshigunov@users.sourceforge.net
* WEB: http://www.sourceforge.net/projects/modulos
* Last change: 17/01/2004
*
* See the interface specification to know what each function does.
*
* Copyright (C) 2000-2004, Luiz Henrique Shigunov
*
* This module is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2.1 of the License, or (at your option) any later version.
*
* This module is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this module; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*****************************************
*/

#include <stddef.h>
#include <debug.h>
#include <errormanager.h>
#include <inputdev.h>
#include <intermanager.h>
#include <memman.h>
#include <memmanager.h>
#include <outputdev.h>
#include <sysmodmanager.h>
#include <stringman.h>
#include <taskmanager.h>
#include <userspacemanager.h>

#define INDEV_FUNCS 4
#define INDEV_GETFOCUS 0
#define INDEV_GETIMPINFO 1
#define INDEV_REGISTER 2
#define INDEV_SETFOCUS 3

#define OUTDEV_FUNCS 2
#define OUTDEV_GETIMPINFO 0
#define OUTDEV_REGISTER 1

typedef struct _QueueEvent {
    struct _QueueEvent *next;
    InterManager_Event event;
} QueueEvent;

typedef struct _InDev {
    struct _InDev *next;
    struct _InDev **pprevNext;
    void *functions[INDEV_FUNCS];
    unsigned int count;
    SysModManager_Imp *imp;
} InDev;

typedef struct _OutDev {
    struct _OutDev *next;
    struct _OutDev **pprevNext;
    void *functions[OUTDEV_FUNCS];
    unsigned int count;
    SysModManager_Imp *imp;
} OutDev;

#define MAX_EVENTS 32

typedef struct _InterData {
    TaskManager_Semaphore sem;
    TaskManager_Semaphore lock;
    QueueEvent *first;
    QueueEvent *last;
    InDev *in;
    OutDev *out;
} InterData;

SysModManager_Module *ID;
InDev *inDevs;
OutDev *outDevs;
TaskManager_Semaphore inSync = {1, NULL};
TaskManager_Semaphore outSync = {1, NULL};
MemManager_Cache *eventCache;

int InsertEvent(InterData*,const InterManager_Event*);
void UserSpaceManagerE_TaskDestroyed(TaskManager_Task*);

void DoFreeInput(InDev *in, TaskManager_Task *task) {
    TaskManager_Task *(*getFocusF)(void);
    void (*setFocusF)(TaskManager_Task*);
    int ret;

    /* first change focus because after dec count we can't use it anymore */
    getFocusF = in->functions[INDEV_GETFOCUS];
    setFocusF = in->functions[INDEV_SETFOCUS];
    if (getFocusF() == task) {
        TaskManager_Task *parent;
        
        parent = TaskManager_GetParentTaskHandle(task);
        setFocusF(parent);
        TaskManager_Destroyable(parent);
    }
    TaskManager_P(&inSync, 0, 0);
    in->count--;
    if (!in->count) {
        *in->pprevNext = in->next;
        ret = SysModManager_PutImpID(in->imp);
        ASSERT_ERROR(ret);
        ret = MemManager_Free(in);
        ASSERT_ERROR(ret);
    }
    TaskManager_V(&inSync, 0);
}

void DoFreeOutput(OutDev *out) {
    int ret;

    TaskManager_P(&outSync, 0, 0);
    out->count--;
    if (!out->count) {
        *out->pprevNext = out->next;
        ret = SysModManager_PutImpID(out->imp);
        ASSERT_ERROR(ret);
        ret = MemManager_Free(out);
        ASSERT_ERROR(ret);
    }
    TaskManager_V(&outSync, 0);
}

int DoInputReg(TaskManager_Task *task, InDev *in) {
    TaskManager_Task *(*getFocusF)(void);
    void (*setFocusF)(TaskManager_Task*);
    InterData *info;
    int (*regFunc)(TaskManager_Task*);

    info = TaskManager_GetTaskInfo(task, ID);
    if (IS_PTR_ERROR(info)) {
        void *voidRet;
        info = MemManager_Alloc(sizeof(InterData), 0);
        if (IS_PTR_ERROR(info))
            return PTR_TO_ERROR(info);
        /* we only support one input device now */
        info->in = in;
        info->out = NULL;
        info->sem.count = 0;
        info->sem.data = NULL;
        info->lock.count = 1;
        info->lock.data = NULL;
        info->first = NULL;
        info->last = NULL;
        voidRet = TaskManager_SetTaskInfo(task, ID, info, 0);
        ASSERT_ERROR_PTR(voidRet);
    } else {
        /* we only support one input device now */
        if (info->in)
            return E_REGISTERED;
        info->in = in;
    }
    regFunc = (int(*)(TaskManager_Task*))in->functions[INDEV_REGISTER];
    regFunc(task);
    /* set focus */
    getFocusF = in->functions[INDEV_GETFOCUS];
    setFocusF = in->functions[INDEV_SETFOCUS];
    if (!getFocusF())
        setFocusF(task);
    return E_OK;
}

int DoOutputReg(TaskManager_Task *task, OutDev *out) {
    InterData *info;
    int (*regFunc)(TaskManager_Task*);

    info = TaskManager_GetTaskInfo(task, ID);
    if (IS_PTR_ERROR(info)) {
        void *voidRet;
        info = MemManager_Alloc(sizeof(InterData), 0);
        if (IS_PTR_ERROR(info))
            return PTR_TO_ERROR(info);
        info->out = out;
        info->in = NULL;
        info->sem.count = 0;
        info->sem.data = NULL;
        info->lock.count = 1;
        info->lock.data = NULL;
        info->first = NULL;
        info->last = NULL;
        voidRet = TaskManager_SetTaskInfo(task, ID, info, 0);
        ASSERT_ERROR_PTR(voidRet);
    } else {
        if (info->out)
            return E_REGISTERED;
        info->out = out;
    }
    regFunc = (int(*)(TaskManager_Task*))out->functions[OUTDEV_REGISTER];
    regFunc(task);
    return E_OK;
}

int GetInputInfo(unsigned int index, InterManager_InputInfo *inInfo, unsigned int size) {
    InterData *info;
    InputDev_ImpInfo impInfo;
    unsigned short interSize, impSize;
    int (*getImpInfo)(InputDev_ImpInfo*);
    int error;

    info = TaskManager_GetTaskInfo(NULL, ID);
    if (index || IS_PTR_ERROR(info) || !info->in)
        return E_NOT_REGISTERED;
    getImpInfo = (int(*)(InputDev_ImpInfo*))info->in->functions[INDEV_GETIMPINFO];
    if (getImpInfo(&impInfo))
        return E_MODULE;
    interSize = StringMan_Length(impInfo.interface) + 1;
    impSize = StringMan_Length(impInfo.implementation) + 1;
    if (size < (interSize + impSize + sizeof(inInfo->interSize) + sizeof(inInfo->impSize)))
        return E_TOO_BIG;
    /* user or system area? */
    if (inInfo < (InterManager_InputInfo*)0xe0000000) {
        error = UserSpaceManager_CopyToUser(&inInfo->interSize, &interSize, sizeof(interSize));
        error |= UserSpaceManager_CopyToUser(&inInfo->impSize, &impSize, sizeof(impSize));
        error |= UserSpaceManager_CopyToUser(inInfo->names, impInfo.interface, interSize);
        error |= UserSpaceManager_CopyToUser(inInfo->names+interSize, impInfo.implementation, impSize);
        if (error)
            error = E_FAULT;
    } else {
        MemMan_Copy(&inInfo->interSize, &interSize, sizeof(interSize));
        MemMan_Copy(&inInfo->impSize, &impSize, sizeof(impSize));
        MemMan_Copy(inInfo->names, impInfo.interface, interSize);
        MemMan_Copy(inInfo->names+interSize, impInfo.implementation, impSize);
        error = E_OK;
    }
    return error;
}

int GetOutputInfo(InterManager_OutputInfo *outInfo, unsigned int size) {
    InterData *info;
    OutputDev_ImpInfo impInfo;
    unsigned short interSize, impSize;
    int (*getImpInfo)(OutputDev_ImpInfo*);
    int error;

    info = TaskManager_GetTaskInfo(NULL, ID);
    if (IS_PTR_ERROR(info) || !info->out)
        return E_NOT_REGISTERED;
    getImpInfo = (int(*)(OutputDev_ImpInfo*))info->out->functions[OUTDEV_GETIMPINFO];
    if (getImpInfo(&impInfo))
        return E_MODULE;
    interSize = StringMan_Length(impInfo.interface) + 1;
    impSize = StringMan_Length(impInfo.implementation) + 1;
    if (size < (interSize + impSize + sizeof(outInfo->interSize) + sizeof(outInfo->impSize)))
        return E_TOO_BIG;
    /* user or system area? */
    if (outInfo < (InterManager_OutputInfo*)0xe0000000) {
        error = UserSpaceManager_CopyToUser(&outInfo->interSize, &interSize, sizeof(interSize));
        error |= UserSpaceManager_CopyToUser(&outInfo->impSize, &impSize, sizeof(impSize));
        error |= UserSpaceManager_CopyToUser(outInfo->names, impInfo.interface, interSize);
        error |= UserSpaceManager_CopyToUser(outInfo->names+interSize, impInfo.implementation, impSize);
        if (error)
            error = E_FAULT;
    } else {
        MemMan_Copy(&outInfo->interSize, &interSize, sizeof(interSize));
        MemMan_Copy(&outInfo->impSize, &impSize, sizeof(impSize));
        MemMan_Copy(outInfo->names, impInfo.interface, interSize);
        MemMan_Copy(outInfo->names+interSize, impInfo.implementation, impSize);
        error = E_OK;
    }
    return error;
}

int InsertEvent(InterData *info, const InterManager_Event *event) {
    int error;
    QueueEvent *qEvent;

    TaskManager_P(&info->lock, 0, 0);
    /* too many events? */
    error = E_TOO_BIG;
    if (info->sem.count == MAX_EVENTS)
        goto returnV;
    qEvent = MemManager_AllocObj(eventCache);
    if (IS_PTR_ERROR(qEvent)) {
        error = PTR_TO_ERROR(qEvent);
        goto returnV;
    }
    MemMan_Copy(&qEvent->event, event, sizeof(InterManager_Event));
    qEvent->next = NULL;
    if (info->last)
        info->last->next = qEvent;
    else
        info->first = qEvent;
    info->last = qEvent;
    TaskManager_V(&info->sem, TaskManager_RESUME_KEEP_CPU);
    error = E_OK;
returnV:
    TaskManager_V(&info->lock, 0);
    return error;
}

int RegisterInput(TaskManager_Task *task, const char *inputImp) {
    int error, ret;
    InDev *curr;
    void *getFocusF, *setFocusF, *regFunc, *getImpInfo;
    SysModManager_Imp *imp;

#ifdef DEBUG_ASSERT
    if (!inputImp) {
        ASSERT_MSG_PRE(0);
        return E_BAD_VALUE;
    }
#endif
    /* get imp ID */
    imp = SysModManager_GetImpID("InputDev", inputImp);
    if (IS_PTR_ERROR(imp))
        return E_MODULE;
    TaskManager_P(&inSync, 0, 0);
    curr = inDevs;
    while (curr && curr->imp != imp)
        curr = curr->next;
    if (!curr) {
        error = E_MODULE;
        getFocusF = SysModManager_GetFunction(imp, InputDev__GETFOCUS, 0, NULL);
        setFocusF = SysModManager_GetFunction(imp, InputDev__SETFOCUS, 0, NULL);
        getImpInfo = SysModManager_GetFunction(imp, InputDev__GETIMPINFO, 0, NULL);
        regFunc = SysModManager_GetFunction(imp, InputDev__REGISTER, 0, NULL);
        if (IS_PTR_ERROR(getFocusF) | IS_PTR_ERROR(setFocusF) | IS_PTR_ERROR(getImpInfo) | IS_PTR_ERROR(regFunc))
            goto errorReg;
        curr = MemManager_Alloc(sizeof(InDev), 0);
        if (IS_PTR_ERROR(curr)) {
            error = PTR_TO_ERROR(curr);
errorReg:
            TaskManager_V(&inSync, 0);
            ret = SysModManager_PutImpID(imp);
            ASSERT_ERROR(ret);
            return error;
        }
        curr->imp = imp;
        curr->count = 1;
        curr->functions[INDEV_GETFOCUS] = getFocusF;
        curr->functions[INDEV_GETIMPINFO] = getImpInfo;
        curr->functions[INDEV_REGISTER] = regFunc;
        curr->functions[INDEV_SETFOCUS] = setFocusF;
        curr->next = inDevs;
        curr->pprevNext = &inDevs;
        if (inDevs)
            inDevs->pprevNext = &curr->next;
        inDevs = curr;
    } else {
        curr->count++;
        ret = SysModManager_PutImpID(imp);
        ASSERT_ERROR(ret);
    }
    TaskManager_V(&inSync, 0);
    error = DoInputReg(task, curr);
    if (error)
        DoFreeInput(curr, task);
    return error;
}

int RegisterOutput(TaskManager_Task *task, const char *outputImp) {
    int error, ret;
    OutDev *curr;
    SysModManager_Imp *imp;

#ifdef DEBUG_ASSERT
    if (!outputImp) {
        ASSERT_MSG_PRE(0);
        return E_BAD_VALUE;
    }
#endif
    /* get imp ID */
    imp = SysModManager_GetImpID("OutputDev", outputImp);
    if (IS_PTR_ERROR(imp))
        return E_MODULE;
    TaskManager_P(&outSync, 0, 0);
    curr = outDevs;
    while (curr && curr->imp != imp)
        curr = curr->next;
    if (!curr) {
        void *regFunc, *getImpInfo;

        error = E_MODULE;
        regFunc = SysModManager_GetFunction(imp, OutputDev__REGISTER, 0, NULL);
        getImpInfo = SysModManager_GetFunction(imp, OutputDev__GETIMPINFO, 0, NULL);
        if (IS_PTR_ERROR(regFunc) | IS_PTR_ERROR(getImpInfo))
            goto errorReg;
        curr = MemManager_Alloc(sizeof(OutDev), 0);
        if (IS_PTR_ERROR(curr)) {
            error = PTR_TO_ERROR(curr);
errorReg:
            TaskManager_V(&outSync, 0);
            ret = SysModManager_PutImpID(imp);
            ASSERT_ERROR(ret);
            return error;
        }
        curr->imp = imp;
        curr->count = 1;
        curr->next = outDevs;
        curr->pprevNext = &outDevs;
        curr->functions[OUTDEV_GETIMPINFO] = getImpInfo;
        curr->functions[OUTDEV_REGISTER] = regFunc;
        if (outDevs)
            outDevs->pprevNext = &curr->next;
        outDevs = curr;
    } else {
        curr->count++;
        ret = SysModManager_PutImpID(imp);
        ASSERT_ERROR(ret);
    }
    TaskManager_V(&outSync, 0);
    error = DoOutputReg(task, curr);
    if (error)
        DoFreeOutput(curr);
    return error;
}

int SendEvent(TaskManager_Task *task, const InterManager_Event *event) {
    InterData *info;
    int error;

#ifdef DEBUG_ASSERT
    if (!task || !event) {
        ASSERT_MSG_PRE(0);
        return E_BAD_VALUE;
    }
#endif
    error = E_BAD_VALUE;
    info = TaskManager_GetTaskInfo(task, ID);
    if (IS_PTR_ERROR(info))
        goto errorEnd;
    error = InsertEvent(info, event);

errorEnd:
    return error;
}

int Shutdown(void) {
    return 0;
}

int Phase1Start(SysModManager_Module *modID) {
    int ret;

    ID = modID;
    /* don't need to start UserSpaceManager cause we only use it if it's already started */
    if ((ret = SysModManager_StartModule((unsigned int)MemManager_Alloc)) ||
        (ret = SysModManager_StartModule((unsigned int)TaskManager_P)) ||
        (ret = SysModManager_StartModule((unsigned int)MemMan_Copy))) {
        DEBUG_MSG_PRE(ret);
        return 1;
    }
    eventCache = MemManager_CreateObjCache(sizeof(QueueEvent), 0, 0, NULL, NULL);
    if (IS_PTR_ERROR(eventCache)) {
        DEBUG_MSG_PRE(PTR_TO_ERROR(eventCache));
        return 1;
    }
    return E_OK;
}

void UserSpaceManagerE_TaskCreated(TaskManager_Task *creator, TaskManager_Task *created) {
    InterData *info;

    info = TaskManager_GetTaskInfo(creator, ID);
    if (IS_PTR_ERROR(info))
        return;
    TaskManager_AtomicInc(&info->in->count);
    if (DoInputReg(created, info->in))
        goto returnFree;
    TaskManager_AtomicInc(&info->out->count);
    if (DoOutputReg(created, info->out)) {
returnFree:
        /* something went wrong, free all resources */
        UserSpaceManagerE_TaskDestroyed(created);
    }
}

void UserSpaceManagerE_TaskDestroyed(TaskManager_Task *task) {
    InterData *info;
    QueueEvent *curr;
    int ret;

    info = TaskManager_DelTaskInfo(task, ID);
    if (IS_PTR_ERROR(info))
        return;
    if (info->in)
        DoFreeInput(info->in, task);
    if (info->out)
        DoFreeOutput(info->out);
    curr = info->first;
    while (curr) {
        QueueEvent *tmp;
        tmp = curr->next;
        ret = MemManager_FreeObj(eventCache, curr);
        ASSERT_ERROR(ret);
        curr = tmp;
    }
    ret = MemManager_Free(info);
    ASSERT_ERROR(ret);
}

void UFlushEventQueue(int mask) {
    InterData *info;

    info = TaskManager_GetTaskInfo(NULL, ID);
    if (IS_PTR_ERROR(info))
        return;
    TaskManager_P(&info->lock, 0, 0);
    if (info->first) {
        QueueEvent *curr, **pprev;
        curr = info->first;
        pprev = &info->first;
        while (curr) {
            if (!(curr->event.type & mask)) {
                int ret;
                *pprev = curr->next;
                ret = MemManager_FreeObj(eventCache, curr);
                ASSERT_ERROR(ret);
                curr = *pprev;
                /* info->sem is only an hint because of this */
            } else {
                pprev = &curr->next;
                curr = curr->next;
            }
        }
        if (!info->first)
            info->last = NULL;
    }
    TaskManager_V(&info->lock, 0);
}

int USetFocus(int task) {
    TaskManager_Task *(*getFocusF)(void);
    void (*setFocusF)(TaskManager_Task*);
    TaskManager_Task *taskH, *currTask;
    InterData *info, *callerInfo;
    int error;

    taskH = TaskManager_GetTaskHandleID(task);
    error = E_BAD_VALUE;
    if (!taskH)
        goto errorEnd;
    info = TaskManager_GetTaskInfo(taskH, ID);
    callerInfo = TaskManager_GetTaskInfo(NULL, ID);
    if (IS_PTR_ERROR(info) || IS_PTR_ERROR(callerInfo) || !callerInfo->in)
        goto errorDestroyable;
    error = E_NOT_ALLOWED;
    if (callerInfo->in != info->in)
        goto errorDestroyable;
    getFocusF = info->in->functions[INDEV_GETFOCUS];
    setFocusF = info->in->functions[INDEV_SETFOCUS];
    currTask = TaskManager_GetTaskHandle(NULL);
    if (getFocusF() != currTask)
        goto errorDestroyableCurr;
    setFocusF(taskH);
    error = E_OK;
    
errorDestroyableCurr:
    TaskManager_Destroyable(currTask);
errorDestroyable:
    TaskManager_Destroyable(taskH);
errorEnd:    
    return error;
}

int UGetInputInfo(unsigned int index, InterManager_InputInfo *inInfo, unsigned int size) {
    if (inInfo >= (InterManager_InputInfo*)0xe0000000)
        return E_FAULT;
    return GetInputInfo(index, inInfo, size);
}

int UGetOutputInfo(InterManager_OutputInfo *outInfo, unsigned int size) {
    if (outInfo >= (InterManager_OutputInfo*)0xe0000000)
        return E_FAULT;
    return GetOutputInfo(outInfo, size);
}

int UWaitEvent(InterManager_Event *event, int prop) {
    InterData *info;
    QueueEvent *qEvent;
    int error, ret;

    info = TaskManager_GetTaskInfo(NULL, ID);
    if (IS_PTR_ERROR(info) || !info->in)
        return E_NOT_REGISTERED;
    do {
        /* this thread can wait for ever, so set it interruptible */
        ret = TaskManager_P(&info->sem, TaskManager_THREAD_INT|(prop & TaskManager_LOCK_DONT_BLOCK), 0);
        if (ret)
            return ret;
        /* we may have an event. It's not guaranteed because we can remove events from queue */
        TaskManager_P(&info->lock, 0, 0);
        qEvent = info->first;
        if (qEvent) {
            info->first = info->first->next;
            if (!info->first)
                info->last = NULL;
        }
        TaskManager_V(&info->lock, 0);
    } while (!qEvent);
    error = UserSpaceManager_CopyToUser((char*)event, (char*)&qEvent->event, sizeof(InterManager_Event));
    ret = MemManager_FreeObj(eventCache, qEvent);
    ASSERT_ERROR(ret);
    return error;
}
