/*
*****************************************
* This file implements the IntManager_i386 interface
*
* Version: 0.3.3
* Author: Luiz Henrique Shigunov
* E-mail: luizshigunov@users.sourceforge.net
* WEB: http://www.sourceforge.net/projects/modulos
* Last change: 28/12/2003
*
* See the interface specification to know what each function does.
*
* Copyright (C) 2000-2003, Luiz Henrique Shigunov
*
* This module is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2.1 of the License, or (at your option) any later version.
*
* This module is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this module; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*****************************************
*/

#include <stddef.h>
#include <debug.h>
#include <errormanager.h>
#include <intcontroller_i386.h>
#include <intmanager_i386.h>
#include <memmanager.h>
#include <sysdef.h>
#include <sysmodmanager.h>
#include <taskmanager.h>

#define IntManager_i386_START_IRQ 240   /* where to remap irqs */

typedef struct _IntManager_i386_IRQHandler {
    struct _IntManager_i386_IRQHandler *next;
    int (*function)(IntManager_i386_Regs*);
    struct _SysModManager_Module *module;
} IntManager_i386_IRQHandler;

typedef struct _IntManager_i386_ExceptionHandler {
    struct _IntManager_i386_ExceptionHandler *next;
    int (*function)(IntManager_i386_ExRegs*);
    unsigned int start, end;
} IntManager_i386_ExceptionHandler;

typedef struct {
    unsigned short lowOffset;
    unsigned short segment;
    unsigned short prop;
    unsigned short highOffset;
} IntManager_i386_IDTDesc;

SysModManager_Module *ID;
IntManager_i386_IDTDesc IDT[256];      /* interrupt array */
IntManager_i386_IDTDesc defDescriptor; /* default software int descriptor */
TaskManager_Semaphore regSync = {1, NULL};
SysModManager_Module *ints[256-32-16]; /* modules using an int */
IntManager_i386_IRQHandler *irqs[16];  /* array of hardware ints handlers */
TaskManager_Semaphore exMutex = {1, NULL};
IntManager_i386_ExceptionHandler *exceptions[17];   /* we only support exception till 16 */

void IntManager_i386_DefaultSoftHandler(void);
void IntManager_i386_Ex00(void);
void IntManager_i386_Ex01(void);
void IntManager_i386_Ex02(void);
void IntManager_i386_Ex03(void);
void IntManager_i386_Ex04(void);
void IntManager_i386_Ex05(void);
void IntManager_i386_Ex06(void);
void IntManager_i386_Ex07(void);
void IntManager_i386_Ex08(void);
void IntManager_i386_Ex09(void);
void IntManager_i386_Ex10(void);
void IntManager_i386_Ex11(void);
void IntManager_i386_Ex12(void);
void IntManager_i386_Ex13(void);
void IntManager_i386_Ex14(void);
void IntManager_i386_Ex15(void);
void IntManager_i386_Ex16(void);
void IntManager_i386_IRQ0(void);
void IntManager_i386_IRQ1(void);
void IntManager_i386_IRQ2(void);
void IntManager_i386_IRQ3(void);
void IntManager_i386_IRQ4(void);
void IntManager_i386_IRQ5(void);
void IntManager_i386_IRQ6(void);
void IntManager_i386_IRQ7(void);
void IntManager_i386_IRQ8(void);
void IntManager_i386_IRQ9(void);
void IntManager_i386_IRQ10(void);
void IntManager_i386_IRQ11(void);
void IntManager_i386_IRQ12(void);
void IntManager_i386_IRQ13(void);
void IntManager_i386_IRQ14(void);
void IntManager_i386_IRQ15(void);

void IntManager_i386_CallExceptionHandler(unsigned int ex, IntManager_i386_ExRegs regs) {
    IntManager_i386_ExceptionHandler *curr;
    SysModManager_ModuleInfo moduleInfo;
    unsigned int index;
    char modulePath[256];
    int ret;

    curr = exceptions[ex];
    while (curr && (regs.EIP < curr->start || regs.EIP > curr->end))
        curr = curr->next;
    while (curr && (regs.EIP >= curr->start && regs.EIP <= curr->end)) {
        ret = ((int(*)(IntManager_i386_ExRegs*))curr->function)(&regs);
        if (ret == E_HANDLED_INT)
            return;
        curr = curr->next;
    }
    SysModManager_DebugMsg("IntManager: Unhandled exception 0x%x at address 0x%x\n", ex, regs.EIP);
    /* if user task, destroy it */
    if (regs.EIP < 0xe0000000) {
        TaskManager_UndestroyableThread();
        TaskManager_DestroyTask(NULL);
    }
    /* find the module */
    index = 0;
    while (!(ret = SysModManager_GetModuleInfoIndex(index++, &moduleInfo, modulePath, sizeof(modulePath)))) {
        if (moduleInfo.memStart && regs.EIP >= moduleInfo.memStart && regs.EIP < (moduleInfo.memStart+moduleInfo.memSize))
            break;
    }
    if (!ret) {
        SysModManager_DebugMsg("Module: %s - Start: 0x%x\n", modulePath, moduleInfo.memStart);
    } else {
        SysModManager_DebugMsg("Error: EIP address has no module: 0x%x\n", ret);
    }
    SysModManager_DebugMsg("Current thread will hang.\n");
    /* FIXME: what about an arch specific function? */
    while(1)
        __asm__ __volatile__ ("hlt");
}

void IntManager_i386_CallHandler(unsigned int irq, IntManager_i386_Regs regs) {
    IntManager_i386_IRQHandler *curr;
    int ret;

    ret = E_HANDLED_INT;
    curr = irqs[irq];
    while (curr) {
        ret = ((int(*)(IntManager_i386_Regs*))curr->function)(&regs);
        /* should we continue? */
        if (ret != E_UNHANDLED_INT)
            break;
        curr = curr->next;
    }
    IntController_i386_EndIRQ(irq);
    if (ret == E_INVOKE_SCHEDULER)
        TaskManager_Schedule();
}

int IntManager_i386_Catch(SysModManager_Module *module, unsigned int interrupt, int (*function)(IntManager_i386_Regs*), int prop) {
    int error, ret;

    error = E_BAD_VALUE;
    TaskManager_P(&regSync, 0, 0);
    if (prop & 0x01) {
        IntManager_i386_IRQHandler *newHandler, **curr;
        if (interrupt > 15)
            goto errorSync;
        newHandler = MemManager_Alloc(sizeof(IntManager_i386_IRQHandler), MemManager_BLOCK_MEM);
        if (IS_PTR_ERROR(newHandler)) {
            error = PTR_TO_ERROR(newHandler);
            goto errorSync;
        }
        curr = &irqs[interrupt];
        /* insert in the end of the list? */
        if (prop & 0x04) {
            while (*curr)
                curr = &(*curr)->next;
        }
        newHandler->function = function;
        newHandler->module = module;
        newHandler->next = *curr;
        *curr = newHandler;
        ret = IntController_i386_UnmaskIRQ(interrupt);
        ASSERT_ERROR(ret);
        goto errorOK;
    }
    interrupt -= 32;
    /* IRQs are above 255-16 */
    if (interrupt > 255-16)
        goto errorSync;
    error = E_IN_USE;
    if (ints[interrupt])
        goto errorSync;
    ints[interrupt] = module;
    IDT[interrupt+32].lowOffset = (unsigned int)function;
    IDT[interrupt+32].highOffset = (unsigned int)function >> 16;
    IDT[interrupt+32].segment = DPL0CodeSegment;
    IDT[interrupt+32].prop = 0xef00;
    if (prop & 0x02)
        IDT[interrupt+32].prop = 0x8F00;
errorOK:
    error = E_OK;
errorSync:
    TaskManager_V(&regSync, 0);
    return error;
}

int IntManager_i386_CatchException(unsigned int exception, int (*function)(IntManager_i386_ExRegs*), void *start, void *end) {
    int error;
    IntManager_i386_ExceptionHandler *newHandler, *curr, **prev;

    error = E_BAD_VALUE;
    /* we only support exception till 16 */
    if (exception > 16)
        goto errorEnd;
    TaskManager_P(&exMutex, 0, 0);
    newHandler = MemManager_Alloc(sizeof(IntManager_i386_ExceptionHandler), MemManager_BLOCK_MEM);
    if (IS_PTR_ERROR(newHandler)) {
        error = PTR_TO_ERROR(newHandler);
        goto errorSync;
    }
    curr = exceptions[exception];
    prev = &exceptions[exception];
    while (curr && curr->start <= (unsigned int)start) {
        prev = &curr->next;
        curr = curr->next;
    }
    newHandler->function = function;
    newHandler->start = (unsigned int)start;
    newHandler->end = (unsigned int)end;
    newHandler->next = curr;
    *prev = newHandler;
    error = E_OK;
errorSync:
    TaskManager_V(&exMutex, 0);
errorEnd:
    return error;
}

#ifdef _INCLUDE_ASM
/* must be modified in assembly */
void IntManager_i386_DefaultSoftHandler(void) {
    __asm__ __volatile__ ("iret");
}
#endif

#ifdef _INCLUDE_ASM
/* must be modified in assembly */
void IntManager_i386_Ex00(void) {
    /* jump to IntManager_i386_SetupExStack */
    /* must push dummy error code */
    __asm__ __volatile__ (
        "pushl $0\n"
        "pushl $0\n"
        "jmp IntManager_i386_SetupExStack");
}
#endif

#ifdef _INCLUDE_ASM
/* must be modified in assembly */
void IntManager_i386_Ex01(void) {
    /* jump to IntManager_i386_SetupExStack */
    /* must push dummy error code */
    __asm__ __volatile__ (
        "pushl $0\n"
        "pushl $1\n"
        "jmp IntManager_i386_SetupExStack");
}
#endif

#ifdef _INCLUDE_ASM
/* must be modified in assembly */
void IntManager_i386_Ex02(void) {
    /* jump to IntManager_i386_SetupExStack */
    /* must push dummy error code */
    __asm__ __volatile__ (
        "pushl $0\n"
        "pushl $2\n"
        "jmp IntManager_i386_SetupExStack");
}
#endif

#ifdef _INCLUDE_ASM
/* must be modified in assembly */
void IntManager_i386_Ex03(void) {
    /* jump to IntManager_i386_SetupExStack */
    /* must push dummy error code */
    __asm__ __volatile__ (
        "pushl $0\n"
        "pushl $3\n"
        "jmp IntManager_i386_SetupExStack");
}
#endif

#ifdef _INCLUDE_ASM
/* must be modified in assembly */
void IntManager_i386_Ex04(void) {
    /* jump to IntManager_i386_SetupExStack */
    /* must push dummy error code */
    __asm__ __volatile__ (
        "pushl $0\n"
        "pushl $4\n"
        "jmp IntManager_i386_SetupExStack");
}
#endif

#ifdef _INCLUDE_ASM
/* must be modified in assembly */
void IntManager_i386_Ex05(void) {
    /* jump to IntManager_i386_SetupExStack */
    /* must push dummy error code */
    __asm__ __volatile__ (
        "pushl $0\n"
        "pushl $5\n"
        "jmp IntManager_i386_SetupExStack");
}
#endif

#ifdef _INCLUDE_ASM
/* must be modified in assembly */
void IntManager_i386_Ex06(void) {
    /* jump to IntManager_i386_SetupExStack */
    /* must push dummy error code */
    __asm__ __volatile__ (
        "pushl $0\n"
        "pushl $6\n"
        "jmp IntManager_i386_SetupExStack");
}
#endif

#ifdef _INCLUDE_ASM
/* must be modified in assembly */
void IntManager_i386_Ex07(void) {
    /* jump to IntManager_i386_SetupExStack */
    /* must push dummy error code */
    __asm__ __volatile__ (
        "pushl $0\n"
        "pushl $7\n"
        "jmp IntManager_i386_SetupExStack");
}
#endif

#ifdef _INCLUDE_ASM
/* must be modified in assembly */
void IntManager_i386_Ex08(void) {
    /* jump to IntManager_i386_SetupExStack */
    __asm__ __volatile__ (
        "pushl $8\n"
        "jmp IntManager_i386_SetupExStack");
}
#endif

#ifdef _INCLUDE_ASM
/* must be modified in assembly */
void IntManager_i386_Ex09(void) {
    /* jump to IntManager_i386_SetupExStack */
    /* must push dummy error code */
    __asm__ __volatile__ (
        "pushl $0\n"
        "pushl $9\n"
        "jmp IntManager_i386_SetupExStack");
}
#endif

#ifdef _INCLUDE_ASM
/* must be modified in assembly */
void IntManager_i386_Ex10(void) {
    /* jump to IntManager_i386_SetupExStack */
    __asm__ __volatile__ (
        "pushl $10\n"
        "jmp IntManager_i386_SetupExStack");
}
#endif

#ifdef _INCLUDE_ASM
/* must be modified in assembly */
void IntManager_i386_Ex11(void) {
    /* jump to IntManager_i386_SetupExStack */
    __asm__ __volatile__ (
        "pushl $11\n"
        "jmp IntManager_i386_SetupExStack");
}
#endif

#ifdef _INCLUDE_ASM
/* must be modified in assembly */
void IntManager_i386_Ex12(void) {
    /* jump to IntManager_i386_SetupExStack */
    __asm__ __volatile__ (
        "pushl $12\n"
        "jmp IntManager_i386_SetupExStack");
}
#endif

#ifdef _INCLUDE_ASM
/* must be modified in assembly */
void IntManager_i386_Ex13(void) {
    /* jump to IntManager_i386_SetupExStack */
    __asm__ __volatile__ (
        "pushl $13\n"
        "jmp IntManager_i386_SetupExStack");
}
#endif

#ifdef _INCLUDE_ASM
/* must be modified in assembly */
void IntManager_i386_Ex14(void) {
    /* jump to IntManager_i386_SetupExStack */
    __asm__ __volatile__ (
        "pushl $14\n"
        "jmp IntManager_i386_SetupExStack");
}
#endif

#ifdef _INCLUDE_ASM
/* must be modified in assembly */
void IntManager_i386_Ex15(void) {
    /* jump to IntManager_i386_SetupExStack */
    /* must push dummy error code */
    __asm__ __volatile__ (
        "pushl $0\n"
        "pushl $15\n"
        "jmp IntManager_i386_SetupExStack");
}
#endif

#ifdef _INCLUDE_ASM
/* must be modified in assembly */
void IntManager_i386_Ex16(void) {
    /* jump to IntManager_i386_SetupExStack */
    /* must push dummy error code */
    __asm__ __volatile__ (
        "pushl $0\n"
        "pushl $16\n"
        "jmp IntManager_i386_SetupExStack");
}
#endif

int IntManager_i386_Free(SysModManager_Module *module, unsigned int interrupt, int prop) {
    int error;

    error = E_BAD_VALUE;
    TaskManager_P(&regSync, 0, 0);
    if (prop & 0x01) {
        IntManager_i386_IRQHandler *curr, **prev;
        if (interrupt > 15)
            goto errorSync;
        curr = irqs[interrupt];
        prev = &irqs[interrupt];
        error = E_NOT_IN_USE;
        while (curr) {
            if (curr->module == module) {
                int ret;
                *prev = curr->next;
                ret = MemManager_Free(curr);
                ASSERT_ERROR(ret);
                if (!irqs[interrupt]) {
                    ret = IntController_i386_MaskIRQ(interrupt);
                    ASSERT_ERROR(ret);
                }
                goto errorOK;
            }
            prev = &curr->next;
            curr = curr->next;
        }
        goto errorSync;
    }
    interrupt -= 32;
    /* IRQs are above 255-16 */
    if (interrupt > 255-16)
        goto errorSync;
    error = E_NOT_IN_USE;
    if (ints[interrupt] != module)
        goto errorSync;
    ints[interrupt] = NULL;
    IDT[interrupt+32] = defDescriptor;
errorOK:
    error = E_OK;
errorSync:
    TaskManager_V(&regSync, 0);
    return error;
}

int IntManager_i386_FreeException(unsigned int exception, int (*function)(IntManager_i386_ExRegs*), void *start) {
    int error;
    IntManager_i386_ExceptionHandler *curr, **prev;

    error = E_BAD_VALUE;
    /* we only support exception till 16 */
    if (exception > 16)
        goto errorEnd;
    TaskManager_P(&exMutex, 0, 0);
    curr = exceptions[exception];
    prev = &exceptions[exception];
    error = E_NOT_FOUND;
    while (curr) {
        if (curr->function == function && curr->start == (unsigned int)start) {
            int ret;
            *prev = curr->next;
            ret = MemManager_Free(curr);
            ASSERT_ERROR(ret);
            error = E_OK;
            break;
        }
        prev = &curr->next;
        curr = curr->next;
    }
    TaskManager_V(&exMutex, 0);
errorEnd:
    return error;
}

#ifdef _INCLUDE_ASM
/* must be modified in assembly */
void IntManager_i386_IRQ0(void) {
    /* jump to IntManager_i386_SetupIntStack */
    __asm__ __volatile__ (
        "pushl $0\n"
        "jmp IntManager_i386_SetupIntStack");
}
#endif

#ifdef _INCLUDE_ASM
/* must be modified in assembly */
void IntManager_i386_IRQ1(void) {
    /* jump to IntManager_i386_SetupIntStack */
    __asm__ __volatile__ (
        "pushl $1\n"
        "jmp IntManager_i386_SetupIntStack");
}
#endif

#ifdef _INCLUDE_ASM
/* must be modified in assembly */
void IntManager_i386_IRQ2(void) {
    /* jump to IntManager_i386_SetupIntStack */
    __asm__ __volatile__ (
        "pushl $2\n"
        "jmp IntManager_i386_SetupIntStack");
}
#endif

#ifdef _INCLUDE_ASM
/* must be modified in assembly */
void IntManager_i386_IRQ3(void) {
    /* jump to IntManager_i386_SetupIntStack */
    __asm__ __volatile__ (
        "pushl $3\n"
        "jmp IntManager_i386_SetupIntStack");
}
#endif

#ifdef _INCLUDE_ASM
/* must be modified in assembly */
void IntManager_i386_IRQ4(void) {
    /* jump to IntManager_i386_SetupIntStack */
    __asm__ __volatile__ (
        "pushl $4\n"
        "jmp IntManager_i386_SetupIntStack");
}
#endif

#ifdef _INCLUDE_ASM
/* must be modified in assembly */
void IntManager_i386_IRQ5(void) {
    /* jump to IntManager_i386_SetupIntStack */
    __asm__ __volatile__ (
        "pushl $5\n"
        "jmp IntManager_i386_SetupIntStack");
}
#endif

#ifdef _INCLUDE_ASM
/* must be modified in assembly */
void IntManager_i386_IRQ6(void) {
    /* jump to IntManager_i386_SetupIntStack */
    __asm__ __volatile__ (
        "pushl $6\n"
        "jmp IntManager_i386_SetupIntStack");
}
#endif

#ifdef _INCLUDE_ASM
/* must be modified in assembly */
void IntManager_i386_IRQ7(void) {
    /* jump to IntManager_i386_SetupIntStack */
    __asm__ __volatile__ (
        "pushl $7\n"
        "jmp IntManager_i386_SetupIntStack");
}
#endif

#ifdef _INCLUDE_ASM
/* must be modified in assembly */
void IntManager_i386_IRQ8(void) {
    /* jump to IntManager_i386_SetupIntStack */
    __asm__ __volatile__ (
        "pushl $8\n"
        "jmp IntManager_i386_SetupIntStack");
}
#endif

#ifdef _INCLUDE_ASM
/* must be modified in assembly */
void IntManager_i386_IRQ9(void) {
    /* jump to IntManager_i386_SetupIntStack */
    __asm__ __volatile__ (
        "pushl $9\n"
        "jmp IntManager_i386_SetupIntStack");
}
#endif

#ifdef _INCLUDE_ASM
/* must be modified in assembly */
void IntManager_i386_IRQ10(void) {
    /* jump to IntManager_i386_SetupIntStack */
    __asm__ __volatile__ (
        "pushl $10\n"
        "jmp IntManager_i386_SetupIntStack");
}
#endif

#ifdef _INCLUDE_ASM
/* must be modified in assembly */
void IntManager_i386_IRQ11(void) {
    /* jump to IntManager_i386_SetupIntStack */
    __asm__ __volatile__ (
        "pushl $11\n"
        "jmp IntManager_i386_SetupIntStack");
}
#endif

#ifdef _INCLUDE_ASM
/* must be modified in assembly */
void IntManager_i386_IRQ12(void) {
    /* jump to IntManager_i386_SetupIntStack */
    __asm__ __volatile__ (
        "pushl $12\n"
        "jmp IntManager_i386_SetupIntStack");
}
#endif

#ifdef _INCLUDE_ASM
/* must be modified in assembly */
void IntManager_i386_IRQ13(void) {
    /* jump to IntManager_i386_SetupIntStack */
    __asm__ __volatile__ (
        "pushl $13\n"
        "jmp IntManager_i386_SetupIntStack");
}
#endif

#ifdef _INCLUDE_ASM
/* must be modified in assembly */
void IntManager_i386_IRQ14(void) {
    /* jump to IntManager_i386_SetupIntStack */
    __asm__ __volatile__ (
        "pushl $14\n"
        "jmp IntManager_i386_SetupIntStack");
}
#endif

#ifdef _INCLUDE_ASM
/* must be modified in assembly */
void IntManager_i386_IRQ15(void) {
    /* jump to IntManager_i386_SetupIntStack */
    __asm__ __volatile__ (
        "pushl $15\n"
        "jmp IntManager_i386_SetupIntStack");
}
#endif

#ifdef _INCLUDE_ASM
/* must be modified in assembly */
int IntManager_i386_NoInt(void) {
    int ret;

    __asm__ __volatile__ (
        "pushfl\n"
        "popl %0\n"
        "cli"
        :"=q" (ret):);
    return ret;
}
#endif

int Phase0Start(SysModManager_Module *modID) {
    unsigned int i;
    unsigned char idtReg[6];
    IntManager_i386_IDTDesc tmp;
    int ret;

    ID = modID;
    /* start with default handlers */
    i = (unsigned int)IntManager_i386_DefaultSoftHandler;
    defDescriptor.lowOffset = i;
    defDescriptor.segment = DPL0CodeSegment;
    defDescriptor.prop = 0x8F00;
    defDescriptor.highOffset = i >> 16;
    for (i=32; i < IntManager_i386_START_IRQ; i++)
        IDT[i] = defDescriptor;
    /* start IDT with exception handlers */
    tmp.segment = DPL0CodeSegment;
    tmp.prop = 0x8F00;
    tmp.lowOffset = (unsigned int)IntManager_i386_Ex00;
    tmp.highOffset = (unsigned int)IntManager_i386_Ex00 >> 16;
    IDT[0] = tmp;
    tmp.lowOffset = (unsigned int)IntManager_i386_Ex01;
    tmp.highOffset = (unsigned int)IntManager_i386_Ex01 >> 16;
    IDT[1] = tmp;
    tmp.lowOffset = (unsigned int)IntManager_i386_Ex02;
    tmp.highOffset = (unsigned int)IntManager_i386_Ex02 >> 16;
    IDT[2] = tmp;
    tmp.lowOffset = (unsigned int)IntManager_i386_Ex03;
    tmp.highOffset = (unsigned int)IntManager_i386_Ex03 >> 16;
    IDT[3] = tmp;
    tmp.lowOffset = (unsigned int)IntManager_i386_Ex04;
    tmp.highOffset = (unsigned int)IntManager_i386_Ex04 >> 16;
    IDT[4] = tmp;
    tmp.lowOffset = (unsigned int)IntManager_i386_Ex05;
    tmp.highOffset = (unsigned int)IntManager_i386_Ex05 >> 16;
    IDT[5] = tmp;
    tmp.lowOffset = (unsigned int)IntManager_i386_Ex06;
    tmp.highOffset = (unsigned int)IntManager_i386_Ex06 >> 16;
    IDT[6] = tmp;
    tmp.lowOffset = (unsigned int)IntManager_i386_Ex07;
    tmp.highOffset = (unsigned int)IntManager_i386_Ex07 >> 16;
    IDT[7] = tmp;
    tmp.lowOffset = (unsigned int)IntManager_i386_Ex08;
    tmp.highOffset = (unsigned int)IntManager_i386_Ex08 >> 16;
    IDT[8] = tmp;
    tmp.lowOffset = (unsigned int)IntManager_i386_Ex09;
    tmp.highOffset = (unsigned int)IntManager_i386_Ex09 >> 16;
    IDT[9] = tmp;
    tmp.lowOffset = (unsigned int)IntManager_i386_Ex10;
    tmp.highOffset = (unsigned int)IntManager_i386_Ex10 >> 16;
    IDT[10] = tmp;
    tmp.lowOffset = (unsigned int)IntManager_i386_Ex11;
    tmp.highOffset = (unsigned int)IntManager_i386_Ex11 >> 16;
    IDT[11] = tmp;
    tmp.lowOffset = (unsigned int)IntManager_i386_Ex12;
    tmp.highOffset = (unsigned int)IntManager_i386_Ex12 >> 16;
    IDT[12] = tmp;
    tmp.lowOffset = (unsigned int)IntManager_i386_Ex13;
    tmp.highOffset = (unsigned int)IntManager_i386_Ex13 >> 16;
    IDT[13] = tmp;
    tmp.lowOffset = (unsigned int)IntManager_i386_Ex15;
    tmp.highOffset = (unsigned int)IntManager_i386_Ex15 >> 16;
    IDT[15] = tmp;
    tmp.lowOffset = (unsigned int)IntManager_i386_Ex16;
    tmp.highOffset = (unsigned int)IntManager_i386_Ex16 >> 16;
    IDT[16] = tmp;
    /* page fault handler is called with interrupts disabled.
       this way the handler can get cr2 value */
    tmp.prop = 0x8E00;
    tmp.lowOffset = (unsigned int)IntManager_i386_Ex14;
    tmp.highOffset = (unsigned int)IntManager_i386_Ex14 >> 16;
    IDT[14] = tmp;
    /* start IDT with IRQ handlers */
    tmp.segment = DPL0CodeSegment;
    /* IRQ0 */
    tmp.lowOffset = (unsigned int)IntManager_i386_IRQ0;
    tmp.highOffset = (unsigned int)IntManager_i386_IRQ0 >> 16;
    IDT[IntManager_i386_START_IRQ] = tmp;
    /* IRQ1 */
    tmp.lowOffset = (unsigned int)IntManager_i386_IRQ1;
    tmp.highOffset = (unsigned int)IntManager_i386_IRQ1 >> 16;
    IDT[IntManager_i386_START_IRQ+1] = tmp;
    /* IRQ2 */
    tmp.lowOffset = (unsigned int)IntManager_i386_IRQ2;
    tmp.highOffset = (unsigned int)IntManager_i386_IRQ2 >> 16;
    IDT[IntManager_i386_START_IRQ+2] = tmp;
    /* IRQ3 */
    tmp.lowOffset = (unsigned int)IntManager_i386_IRQ3;
    tmp.highOffset = (unsigned int)IntManager_i386_IRQ3 >> 16;
    IDT[IntManager_i386_START_IRQ+3] = tmp;
    /* IRQ4 */
    tmp.lowOffset = (unsigned int)IntManager_i386_IRQ4;
    tmp.highOffset = (unsigned int)IntManager_i386_IRQ4 >> 16;
    IDT[IntManager_i386_START_IRQ+4] = tmp;
    /* IRQ5 */
    tmp.lowOffset = (unsigned int)IntManager_i386_IRQ5;
    tmp.highOffset = (unsigned int)IntManager_i386_IRQ5 >> 16;
    IDT[IntManager_i386_START_IRQ+5] = tmp;
    /* IRQ6 */
    tmp.lowOffset = (unsigned int)IntManager_i386_IRQ6;
    tmp.highOffset = (unsigned int)IntManager_i386_IRQ6 >> 16;
    IDT[IntManager_i386_START_IRQ+6] = tmp;
    /* IRQ7 */
    tmp.lowOffset = (unsigned int)IntManager_i386_IRQ7;
    tmp.highOffset = (unsigned int)IntManager_i386_IRQ7 >> 16;
    IDT[IntManager_i386_START_IRQ+7] = tmp;
    /* IRQ8 */
    tmp.lowOffset = (unsigned int)IntManager_i386_IRQ8;
    tmp.highOffset = (unsigned int)IntManager_i386_IRQ8 >> 16;
    IDT[IntManager_i386_START_IRQ+8] = tmp;
    /* IRQ9 */
    tmp.lowOffset = (unsigned int)IntManager_i386_IRQ9;
    tmp.highOffset = (unsigned int)IntManager_i386_IRQ9 >> 16;
    IDT[IntManager_i386_START_IRQ+9] = tmp;
    /* IRQ10 */
    tmp.lowOffset = (unsigned int)IntManager_i386_IRQ10;
    tmp.highOffset = (unsigned int)IntManager_i386_IRQ10 >> 16;
    IDT[IntManager_i386_START_IRQ+10] = tmp;
    /* IRQ11 */
    tmp.lowOffset = (unsigned int)IntManager_i386_IRQ11;
    tmp.highOffset = (unsigned int)IntManager_i386_IRQ11 >> 16;
    IDT[IntManager_i386_START_IRQ+11] = tmp;
    /* IRQ12 */
    tmp.lowOffset = (unsigned int)IntManager_i386_IRQ12;
    tmp.highOffset = (unsigned int)IntManager_i386_IRQ12 >> 16;
    IDT[IntManager_i386_START_IRQ+12] = tmp;
    /* IRQ13 */
    tmp.lowOffset = (unsigned int)IntManager_i386_IRQ13;
    tmp.highOffset = (unsigned int)IntManager_i386_IRQ13 >> 16;
    IDT[IntManager_i386_START_IRQ+13] = tmp;
    /* IRQ14 */
    tmp.lowOffset = (unsigned int)IntManager_i386_IRQ14;
    tmp.highOffset = (unsigned int)IntManager_i386_IRQ14 >> 16;
    IDT[IntManager_i386_START_IRQ+14] = tmp;
    /* IRQ15 */
    tmp.lowOffset = (unsigned int)IntManager_i386_IRQ15;
    tmp.highOffset = (unsigned int)IntManager_i386_IRQ15 >> 16;
    IDT[IntManager_i386_START_IRQ+15] = tmp;
    /* load idt */
    *((unsigned short *)idtReg) = 0x7FF;
    *((unsigned int *)(idtReg+2)) = (int)IDT;
    __asm__ __volatile__ ("lidt (%0)" : : "q" (idtReg));
    if ((ret = SysModManager_StartModule((unsigned int)MemManager_Alloc)) ||
        (ret = SysModManager_StartModule((unsigned int)TaskManager_P)) ||
        (ret = SysModManager_StartModule((unsigned int)IntController_i386_UnmaskIRQ))) {
        DEBUG_MSG_PRE(ret);
        return 1;
    }
    /* must be done here, because Phase1Start must enable interrupts */
    IntController_i386_RedirectIRQ(IntManager_i386_START_IRQ);
    return 0;
}

int Phase1Start(SysModManager_Module *modID) {
    int ret;

    /* enable interrupts */
    __asm__ __volatile__ ("sti");
    if ((ret = SysModManager_StartModule((unsigned int)MemManager_Alloc)) ||
        (ret = SysModManager_StartModule((unsigned int)TaskManager_P)) ||
        (ret = SysModManager_StartModule((unsigned int)IntController_i386_UnmaskIRQ))) {
        DEBUG_MSG_PRE(ret);
        return 1;
    }
    return 0;
}

#ifdef _INCLUDE_ASM
/* must be modified in assembly */
void IntManager_i386_RestoreInt(int flags) {
    __asm__ __volatile__ (
        "pushl	%0\n"
        "popfl"
        :: "q" (flags));
    return;
}
#endif
