/*
*****************************************
* This file implements the IOPortManager interface
*
* Version: 0.3.2
* Author: Luiz Henrique Shigunov
* E-mail: luizshigunov@users.sourceforge.net
* WEB: http://www.sourceforge.net/projects/modulos
* Last change: 01/07/2005
*
* See the interface specification to know what each function does.
*
* Copyright (C) 2000-2005, Luiz Henrique Shigunov
*
* This module is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2.1 of the License, or (at your option) any later version.
*
* This module is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this module; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*****************************************
*/

#include <stddef.h>
#include <debug.h>
#include <errormanager.h>
#include <memmanager.h>
#include <sysmodmanager.h>
#include <taskmanager.h>

typedef struct _IOPortManager_Port {
    struct _IOPortManager_Port *next;
    unsigned int start, end;
    struct _SysModManager_Module *module;
} IOPortManager_Port;

IOPortManager_Port *ports;
TaskManager_RWSync portSync;

int IOPortManager_Alloc(SysModManager_Module *module, unsigned int start, unsigned int end, int prop) {
    IOPortManager_Port *port;
    int error;

    error = E_IN_USE;
    TaskManager_LockWrite(&portSync, 0);
    if (!prop) {
        IOPortManager_Port *curr;
        curr = ports;
        while (curr) {
            if ((start >= curr->start && start <= curr->end) || (end >= curr->start && end <= curr->end))
                goto errorEnd;
            curr = curr->next;
        }
    }
    port = MemManager_Alloc(sizeof(IOPortManager_Port), 0);
    if (IS_PTR_ERROR(port)) {
        error = PTR_TO_ERROR(port);
        goto errorEnd;
    }
    port->next = ports;
    ports = port;
    port->start = start;
    port->end = end;
    port->module = module;
    error = E_OK;
errorEnd:
    TaskManager_UnlockRW(&portSync, 0);
    return error;
}

int IOPortManager_Free(SysModManager_Module *module, unsigned int start) {
    IOPortManager_Port *curr, **prev;
    int error;

    error = E_BAD_VALUE;
    TaskManager_LockWrite(&portSync, 0);
    curr = ports;
    prev = &ports;
    while (curr) {
        if (start == curr->start && module == curr->module) {
            int ret;
            *prev = curr->next;
            ret = MemManager_Free(curr);
            ASSERT_ERROR(ret);
            error = E_OK;
            break;
        }
        prev = &curr->next;
        curr = curr->next;
    }
    TaskManager_UnlockRW(&portSync, 0);
    return error;
}

int IOPortManager_IsInUse(unsigned int portN) {
    IOPortManager_Port *curr;
    int error;

    error = E_NOT_IN_USE;
    TaskManager_LockRead(&portSync, 0);
    curr = ports;
    while (curr) {
        if (portN >= curr->start && portN <= curr->end) {
            error = E_IN_USE;
            break;
        }
        curr = curr->next;
    }
    TaskManager_UnlockRW(&portSync, 0);
    return error;
}

int Phase0Start(SysModManager_Module *modID) {
    int ret;

    if ((ret = SysModManager_StartModule((unsigned int)TaskManager_LockWrite)) ||
        (ret = SysModManager_StartModule((unsigned int)MemManager_Alloc))) {
        DEBUG_MSG_PRE(ret);
        return 1;
    }
    return 0;
}

int Phase1Start(SysModManager_Module *modID) {
    int ret;

    if ((ret = SysModManager_StartModule((unsigned int)TaskManager_LockWrite)) ||
        (ret = SysModManager_StartModule((unsigned int)MemManager_Alloc))) {
        DEBUG_MSG_PRE(ret);
        return 1;
    }
    return 0;
}
