/*
*****************************************
* This file implements the MemManager and MemManager_i386 interfaces
*
* Version: 0.3.5
* Author: Luiz Henrique Shigunov
* E-mail: luizshigunov@users.sourceforge.net
* WEB: http://www.sourceforge.net/projects/modulos
* Last change: 24/06/2005
*
* See the interface specification to know what each function does.
*
* Copyright (C) 2000-2005, Luiz Henrique Shigunov
*
* This module is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2.1 of the License, or (at your option) any later version.
*
* This module is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this module; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*****************************************
*/

#include <stddef.h>
#include <debug.h>
#include <errormanager.h>
#include <intmanager_i386.h>
#include <memcontroller.h>
#include <memman.h>
#include <memmanager.h>
#include <memmanager_i386.h>
#include <sysdef.h>
#include <sysmodmanager.h>
#include <taskmanager.h>

#define ALIGN_DATA_UP(data, align) ((((unsigned int)(data)) + (align)-1) & ~((align)-1))
#define ALIGN_DATA_DOWN(data, align) (((unsigned int)(data)) & ~((align)-1))

/* max # of descriptor in GDT */
#define MemManager_SIZE_GDT 10
/* first free descriptor in GDT */
#define MemManager_FIRST_FREE_GDTE 4

/* get the address of the first PDE */
#define MemManager_PG_DIR ((unsigned int*)0xfffff000)
/* get the address of the first PTE */
#define MemManager_PG_TABS ((unsigned int*)0xffc00000)

/* biggest available slab cache for Alloc */
#define MemManager_LAST_CACHE_SIZE 6816
/* last logical page available for system (zero is the first page) */
#define MemManager_LAST_LG_PG 0xffbff
/* biggest logical page chunck allocated - 4 MB */
#define MemManager_BIG_LG_PG_CHUNCK 1024

#define MemManager_START_SYS_MEM 0xe0000000     /* start of logical memory for system */
#define MemManager_N_SYS_PDE 128                /* # of PDEs used by the system - 512MB */
#define MemManager_FIRST_SYS_PDE	(1024-MemManager_N_SYS_PDE)  /* first system entry in PGD 4GB - 512MB */
#define MemManager_PTES_ADDRESS 0xffc00000      /* address of first PTE */
#define MemManager_FIRST_SYS_PTE 0xe0000        /* # of first system PTE */

/* properties of physical pages */
#define MemManager_PHYS_PG_FREE      0 /* free page */
#define MemManager_PHYS_PG_USED      1 /* used page */
#define MemManager_PHYS_PG_BLOCKED   2 /* blocked page */
#define MemManager_PHYS_PG_RESERVED  4 /* reserved page, can't be used for memory */

#define MemManager_TMP_PGS 400
#define MemManager_TMP_PGS_LOW1MB_RESERVED 0x96
#define MemManager_TMP_PGS_LOW1MB_RESERVED_SIZE (0x100-MemManager_TMP_PGS_LOW1MB_RESERVED)

/* # of pre-defined caches */
#define MemManager_NR_CACHES 37

typedef struct _MemManager_Slab {
    char *free;            /* pointer to the first free obj in slab */
    unsigned int inUse;    /* # of used objs */
    struct _MemManager_Cache *cache;  /* pointer to the cache */
    struct _MemManager_Slab *next;    /* linked list of slabs in cache */
    struct _MemManager_Slab *prev;
} MemManager_Slab;

struct _MemManager_Cache {
    struct _MemManager_Slab *free;         /* first slab with free objs */
    unsigned int size;  /* size of obj in this cache */
    TaskManager_Semaphore sem;  /* semaphore to sync the use of cache */
    unsigned int num;   /* # of obj in each slab */
    struct _MemManager_Slab *last;         /* last slab of the circular linked list of slabs */
    void (*ctor)(void*,unsigned int); /* constructor */
    void (*dtor)(void*,unsigned int); /* destructor */
    unsigned int nPages;    /* # of logical pages per slab */
    unsigned int prop;      /* cache properties:
                            bits 0-4 - align in bytes;
                            bit 5 - slab info inside the slab (=1)
                            bit 6 - objs initialized with zero (=1) */
    unsigned int lastColor; /* last color - when nextColor == lastColor, then nextColor = 0 */
    unsigned int nextColor;
};

typedef struct _MemManager_Exception {
    struct _MemManager_Exception *next;
    unsigned int start;
    unsigned int end;
    void *function;
    int prop;
} MemManager_Exception;

typedef struct {
    unsigned short size;
    struct _MemManager_Cache cache;
} MemManager_Size;

typedef struct _MemManager_MemArea {
    struct _MemManager_MemArea *next;
    unsigned int start;
    unsigned int end;
    int prop;
    struct _MemManager_ShMem *shmem;
    unsigned int offset;
} MemManager_MemArea;

typedef struct _MemManager_ShPage {
    struct _MemManager_ShPage *next;
    unsigned int physPg;
    unsigned int offset;
} MemManager_ShPage;

struct _MemManager_ShMem {
    unsigned int count;
    unsigned int size;
    int prop;
    MemManager_ShPage *pages;
    const struct _MemManager_ShMemFunctions *functions;
    void *arg;
};

typedef struct {
    /* sorted by address in decreasing order */
    MemManager_MemArea *areas;
    TaskManager_Semaphore areasSem;
} MemManager_TaskInfo;

/* PTE structure
    unsigned int
        p: 1,           page present
        rw: 1,          read or write(1)
        us: 1,          system or user(1)
        pwt: 1,         write-through
        pcd: 1,         cache disable
        a: 1,           accessed
        d: 1,           dirty
        reserved: 1,
        global: 1,      global pge
        avail: 3,       available
        base: 20;       base address
*/

/* PDE structure
    unsigned int
        p: 1,           * page present *
        rw: 1,          * read or write *
        us: 1,          * system or user *
        pwt: 1,         * write-through *
        pcd: 1,         * cache disable *
        a: 1,           * accessed *
        reserved: 1,
        psize: 1,       * page size *
        global: 1,      * global page *
        avail: 3,       * available *
        base: 20;       * base address *
*/

/* pre-defined caches for the Alloc function. They were chosen because they have the small memory waste. */
MemManager_Size sizes[MemManager_NR_CACHES] = {
{8, {NULL, 12, {1, NULL}, 341, NULL, NULL, NULL, 1, 0x04, 8, 0}},
{12, {NULL, 16, {1, NULL}, 256, NULL, NULL, NULL, 1, 0x04, 4, 0}},
{16, {NULL, 20, {1, NULL}, 204, NULL, NULL, NULL, 1, 0x04, 20, 0}},
{20, {NULL, 24, {1, NULL}, 169, NULL, NULL, NULL, 1, 0x24, 20, 0}},
{24, {NULL, 28, {1, NULL}, 146, NULL, NULL, NULL, 1, 0x04, 12, 0}},
{28, {NULL, 32, {1, NULL}, 128, NULL, NULL, NULL, 1, 0x04, 4, 0}},
{32, {NULL, 36, {1, NULL}, 113, NULL, NULL, NULL, 1, 0x24, 12, 0}},
{48, {NULL, 52, {1, NULL}, 78, NULL, NULL, NULL, 1, 0x24, 24, 0}},
{64, {NULL, 68, {1, NULL}, 60, NULL, NULL, NULL, 1, 0x04, 20, 0}},
{80, {NULL, 84, {1, NULL}, 48, NULL, NULL, NULL, 1, 0x24, 48, 0}},
{100, {NULL, 104, {1, NULL}, 39, NULL, NULL, NULL, 1, 0x24, 24, 0}},
{136, {NULL, 140, {1, NULL}, 29, NULL, NULL, NULL, 1, 0x24, 20, 0}},
{164, {NULL, 168, {1, NULL}, 24, NULL, NULL, NULL, 1, 0x24, 48, 0}},
{188, {NULL, 192, {1, NULL}, 21, NULL, NULL, NULL, 1, 0x24, 48, 0}},
{220, {NULL, 224, {1, NULL}, 18, NULL, NULL, NULL, 1, 0x24, 48, 0}},
{248, {NULL, 252, {1, NULL}, 16, NULL, NULL, NULL, 1, 0x24, 48, 0}},
{284, {NULL, 288, {1, NULL}, 14, NULL, NULL, NULL, 1, 0x24, 48, 0}},
{332, {NULL, 336, {1, NULL}, 12, NULL, NULL, NULL, 1, 0x24, 48, 0}},
{364, {NULL, 368, {1, NULL}, 11, NULL, NULL, NULL, 1, 0x24, 32, 0}},
{400, {NULL, 404, {1, NULL}, 10, NULL, NULL, NULL, 1, 0x24, 40, 0}},
{444, {NULL, 448, {1, NULL}, 9, NULL, NULL, NULL, 1, 0x24, 48, 0}},
{500, {NULL, 504, {1, NULL}, 8, NULL, NULL, NULL, 1, 0x24, 48, 0}},
{572, {NULL, 576, {1, NULL}, 7, NULL, NULL, NULL, 1, 0x24, 48, 0}},
{672, {NULL, 676, {1, NULL}, 6, NULL, NULL, NULL, 1, 0x24, 24, 0}},
{804, {NULL, 808, {1, NULL}, 5, NULL, NULL, NULL, 1, 0x24, 40, 0}},
{1008, {NULL, 1012, {1, NULL}, 4, NULL, NULL, NULL, 1, 0x24, 32, 0}},
{1352, {NULL, 1356, {1, NULL}, 3, NULL, NULL, NULL, 1, 0x24, 12, 0}},
{1628, {NULL, 1632, {1, NULL}, 5, NULL, NULL, NULL, 2, 0x24, 16, 0}},
{2040, {NULL, 2044, {1, NULL}, 2, NULL, NULL, NULL, 1, 0x04, 12, 0}},
{2448, {NULL, 2452, {1, NULL}, 5, NULL, NULL, NULL, 3, 0x24, 12, 0}},
{2716, {NULL, 2720, {1, NULL}, 3, NULL, NULL, NULL, 2, 0x24, 16, 0}},
{3060, {NULL, 3064, {1, NULL}, 4, NULL, NULL, NULL, 3, 0x24, 16, 0}},
{3404, {NULL, 3408, {1, NULL}, 6, NULL, NULL, NULL, 5, 0x24, 16, 0}},
{4072, {NULL, 4076, {1, NULL}, 1, NULL, NULL, NULL, 1, 0x24, 4, 0}},
{5108, {NULL, 5112, {1, NULL}, 4, NULL, NULL, NULL, 5, 0x24, 16, 0}},
{5448, {NULL, 5452, {1, NULL}, 3, NULL, NULL, NULL, 4, 0x24, 12, 0}},
{6816, {NULL, 6820, {1, NULL}, 3, NULL, NULL, NULL, 5, 0x24, 4, 0}}
};

/* physical pages array used on start. Contains the first 640 KB of memory */
char tmpPhysPgs[MemManager_TMP_PGS];

SysModManager_Module *ID;
TaskManager_Semaphore GDTSem = {1, NULL};
TaskManager_Semaphore lgPgSem = {1, NULL};    /* semaphore to sync the use of logical pages */
TaskManager_Semaphore physPgSem = {1, NULL};  /* semaphore to sync the use of physical pages */
unsigned int firstFreePhysPg; /* first free physical page */
unsigned int freePhysPgs=MemManager_TMP_PGS;	/* amount of free physical pages */
unsigned int physPgVEnd=MemManager_TMP_PGS-1;  /* last position in the physical pages array - the first physical page */
char *physPages = tmpPhysPgs; /* pointer to the physical pages array. physPages[0] is the last physical page.
                                 and physPgVEnd is the last array position, which is the first physical page */
unsigned int freeLgPgs=(MemManager_N_SYS_PDE-1)<<10; /* amount of free logical pages for system */
unsigned int firstFreeLgPg=0xe0000;       /* first free logical system page */
unsigned int GDT[MemManager_SIZE_GDT*2] = {(MemManager_SIZE_GDT*8)-1, 0, 0x0000ffff, 0x00cf9900, 0x0000ffff, 0x00cf9300, 0x0000ffff, 0x00cff900, 0x0000ffff, 0x00cff300};

TaskManager_Semaphore exMutex = {1, NULL};
MemManager_Exception *pageFaultWho;
MemManager_Exception *pageFaultUserMode;

MemManager_Cache *areaCache;
MemManager_Cache *pageCache;

int FreeUserPagesNoLock(void *start, unsigned int size, int prop, TaskManager_Task *task, MemManager_TaskInfo *taskInfo);

void *MemManager_Alloc(unsigned int size, int prop) {
    unsigned int i;

    ASSERT_RET(size, ERROR_TO_PTR(E_BUG));

    /* request too big? */
    if (size > (MemManager_BIG_LG_PG_CHUNCK<<12))
        return ERROR_TO_PTR(E_TOO_BIG);
    /* should we alloc whole pages instead? */
    if (size > MemManager_LAST_CACHE_SIZE) {
        unsigned int nPgs, *area;
        /* 2 words, one for the slab info and other for the # of pages */
        size += 8;
        nPgs = ALIGN_DATA_UP(size, 0x1000) >> 12;
        area = MemManager_i386_AllocPages(nPgs, MemManager_i386_RW_PAGE+MemManager_i386_ALLOC_PHYS);
        if (IS_PTR_ERROR(area))
            return area;
        /* first word - size */
        *area = nPgs;
        /* second - zero to indicate no slab cache */
        *(area+1) = 0;
        return area+2;
    }
    /* find best slab */
    i = 0;
    while (1)
        if (size <= sizes[i++].size)
            return MemManager_AllocObj(&sizes[i-1].cache);
}

void *MemManager_AllocObj(MemManager_Cache *cache) {
    MemManager_Slab *slabMgt;
    char **freePtr;
    void *ret;

    ASSERT_RET(cache, ERROR_TO_PTR(E_BUG));

    TaskManager_P(&cache->sem, 0, 0);
    slabMgt = cache->free;
    /* no slab available */
    if (!slabMgt) {
        char *newSlab, *nextObj, *tmp;
        unsigned int n;
        newSlab = MemManager_i386_AllocPages(cache->nPages, MemManager_i386_RW_PAGE+MemManager_i386_ALLOC_PHYS);
        if (IS_PTR_ERROR(newSlab)) {
            ret = newSlab;
            goto errorV;
        }
        /* slab info inside the slab? */
        if (cache->prop & 0x20)
            slabMgt = (MemManager_Slab*)(newSlab+(cache->nPages<<12)-sizeof(MemManager_Slab));
        else {
            slabMgt = MemManager_Alloc(sizeof(MemManager_Slab), 0);
            if (IS_PTR_ERROR(slabMgt)) {
                int iRet;
                iRet = MemManager_i386_FreePages(newSlab, cache->nPages, 0);
                ASSERT_ERROR(iRet);
                ret = slabMgt;
                goto errorV;
            }
        }
        slabMgt->inUse = 0;
        slabMgt->cache = cache;
        /* insert in the list (totally free slabs must be in the end of the list) */
        if (!cache->last) {
            cache->last = slabMgt;
            slabMgt->next = slabMgt;
        }
        slabMgt->prev = cache->last;
        slabMgt->next = cache->last->next;
        cache->last->next->prev = slabMgt;
        cache->last->next = slabMgt;
        cache->free = slabMgt;
        cache->last = slabMgt;
        /* adjust the address of the first obj */
        slabMgt->free = newSlab+cache->nextColor;
        nextObj = slabMgt->free;
        cache->nextColor += (cache->prop & 0x1f);
        cache->nextColor %= cache->lastColor;
        n = cache->num;
        /* should we start the objs with zero? */
        if (cache->prop & 0x40)
            MemMan_Set(nextObj, 0, n*cache->size);
        do {
            tmp = nextObj;
            nextObj += cache->size;
            /* link in the free list */
            *(char**)tmp = nextObj;
            if (cache->ctor)
                cache->ctor(tmp+4, cache->size);
        } while (--n);
        /* end of free list */
        *(char**)tmp = NULL;
    }
    slabMgt->inUse++;
    ret = slabMgt->free+4;
    /* remove from free list - remenber that the pointer to the next free obj is in the start */
    freePtr = (char**)slabMgt->free;
    slabMgt->free = *freePtr;
    /* now the first word points to the slab */
    *freePtr = (char*)slabMgt;
    if (!slabMgt->free) {
        /* set next free slab */
        cache->free = NULL;
        if (slabMgt->next->free)
            cache->free = slabMgt->next;
    }
errorV:
    TaskManager_V(&cache->sem, 0);
    return ret;
}

void *MemManager_i386_AllocPages(unsigned int n, int prop) {
    unsigned int currPg, remain;
    void *ret;

    ASSERT_RET(n, ERROR_TO_PTR(E_BUG));

    /* request too big */
    if (n > MemManager_BIG_LG_PG_CHUNCK)
        return ERROR_TO_PTR(E_TOO_BIG);
    remain = n;
    TaskManager_P(&lgPgSem, 0, 0);
    /* not enough memory? */
    if (freeLgPgs < n) {
        ret = ERROR_TO_PTR(E_NO_MEMORY);
        goto errorV;
    }
    /* go through all pages */
    ret = ERROR_TO_PTR(E_NO_MEMORY);
    currPg = firstFreeLgPg;
    do {
        /* PTE is free? */
        if (!MemManager_PG_TABS[currPg++]) {
            if (--remain == 0) {
                int pageProp;
                unsigned int *ptes, tmpN;
                currPg -= n;
                tmpN = n;
                ptes = &MemManager_PG_TABS[currPg];
                /* set the pages prop */
                pageProp = ((prop << 1) & 0x1f);
                if (prop & MemManager_i386_ALLOC_PHYS) {
                    pageProp |= 1;
                    do {
                        unsigned int physPage;
                        physPage = MemManager_i386_AllocPhysPages(1, 0);
                        if (IS_PTR_ERROR(physPage)) {
                            /* if we get an error, undo everything */
                            ptes--;
                            for (; tmpN != n; tmpN++) {
                                int error;
                                error = MemManager_i386_FreePhysPages(*ptes, 1);
                                ASSERT_ERROR(error);
                                *ptes = 0;
                                ptes--;
                            }
                            ret = (void*)physPage;
                            goto errorV;
                        }
                        *ptes = physPage | pageProp;
                        ptes++;
                    } while (--tmpN);
                    if (prop & MemManager_i386_CLEAR_PAGES)
                        MemMan_Set((void*)(currPg<<12), 0, n<<12);
                } else
                    do {
                        *ptes = pageProp;
                        ptes++;
                    } while (--tmpN);
                /* we don't need to invalided the TLB, because a page fault would have occurred */
                freeLgPgs -= n;
                if (n==1)
                    firstFreeLgPg = currPg+n;
                ret = (void *)(currPg << 12);
                goto errorV;
            }
        } else
            /* restart */
            remain = n;
    } while (currPg <= MemManager_LAST_LG_PG);
errorV:
    TaskManager_V(&lgPgSem, 0);
    return ret;
}

unsigned int MemManager_i386_AllocPhysPages(unsigned int n, int prop) {
    unsigned int currPg;
    unsigned int ret;

    ASSERT_RET(n, (unsigned int)ERROR_TO_PTR(E_BUG));

    TaskManager_P(&physPgSem, 0, 0);
    ret = (unsigned int)ERROR_TO_PTR(E_NO_MEMORY);
    /* not enough memory? */
    if (freePhysPgs < n)
        goto errorV;
    /* if there's no limit start at the end of memory - start of the array */
    currPg = firstFreePhysPg;
    /* 1MB limit? */
    if ((prop & MemManager_i386_LESS_1MB) && physPgVEnd >= 256 && firstFreePhysPg < (physPgVEnd-256))
        currPg = physPgVEnd-255;
    /* 16MB limit? */
    else if ((prop & MemManager_i386_LESS_16MB) && physPgVEnd >= 4096 && firstFreePhysPg < (physPgVEnd-4096))
        currPg = physPgVEnd-4095;
    while (1) {
        char *finded;
        unsigned int remain;
        remain = n;
        /* end reached? */
        if ((currPg+remain) > (physPgVEnd+1))
            goto errorV;
        while (physPages[currPg++] == MemManager_PHYS_PG_FREE)
            if (--remain==0) {
                /* set pages as used */
                ret = (physPgVEnd-currPg+1) << 12;
                freePhysPgs -= n;
                currPg -= n;
                if (n==1) {
                    if (!(prop & (MemManager_i386_LESS_1MB+MemManager_i386_LESS_16MB))) {
                        firstFreePhysPg = currPg+n;
                    }
                    physPages[currPg] = MemManager_PHYS_PG_USED;
                } else
                    MemMan_Set(physPages+currPg, MemManager_PHYS_PG_USED, n);
errorV:
                TaskManager_V(&physPgSem, 0);
                return ret;
            }
        /* find a free page */
        finded = MemMan_FindByte(physPages+currPg, MemManager_PHYS_PG_FREE, physPgVEnd-currPg);
        if (!finded)
            goto errorV;
        currPg = (unsigned int)finded-(unsigned int)physPages;
        if ((prop & MemManager_i386_BOUNDARY_64KB) && (((currPg & 0xf)+n) > 16))
            currPg = (currPg & 0xfffffff0) + 16;
    }
}

/* Look up the last MemArea which satisfies address < area->end starting which area, NULL if none */
static __inline__ MemManager_MemArea *FindLastMemArea(MemManager_MemArea *area, unsigned int address) {
    MemManager_MemArea *found;

    found = NULL;
    while (area && address < area->end) {
        found = area;
        area = area->next;
    }
    return found;
}

/* Look up the first MemArea which satisfies area->start < address starting which area, NULL if none */
static __inline__ MemManager_MemArea *FindFirstMemArea(MemManager_MemArea *area, unsigned int address) {
    while (area && area->start >= address)
        area = area->next;
    return area;
}

/* Look up the first MemArea which satisfies area->start < address starting which area, NULL if none */
static __inline__ MemManager_MemArea *FindFirstMemAreaPrev(MemManager_MemArea *area, unsigned int address, MemManager_MemArea **prev) {
    *prev = NULL;
    while (area && area->start >= address) {
        *prev = area;
        area = area->next;
    }
    return area;
}

static __inline__ void *GetFreeMemArea(MemManager_TaskInfo *taskInfo, unsigned int start, unsigned int len, unsigned int align) {
    MemManager_MemArea *area;

    if (!align)
        align = 4096;
    if (!start)
        start = MemManager_START_SYS_MEM - len;
    start = ALIGN_DATA_DOWN(start, align);
    area = taskInfo->areas;
    do {
        area = FindLastMemArea(area, start);
        if (!area || (start + len) <= area->start)
            return (void*)start;
        /* will wrap? */
        if (area->start < len)
            break;
        start = ALIGN_DATA_DOWN(area->start - len, align);
    } while (start >= 0x1000);
    return ERROR_TO_PTR(E_NO_MEMORY);
}

/* try to merge areas */
void MergeMemArea(MemManager_TaskInfo *taskInfo, unsigned int start, unsigned int end) {
    MemManager_MemArea *area, *next, *prev;
    int ret;

    area = FindFirstMemAreaPrev(taskInfo->areas, end, &prev);
    if (!area)
        return;

    if (!prev) {
        prev = area;
        area = area->next;
    }

    for (; area && prev->start >= start; prev = area, area = next) {
        next = area->next;
        if (prev->prop != area->prop || prev->shmem != area->shmem || area->end != prev->start ||
            (area->shmem && prev->offset != (area->offset + (area->end-area->start))))
            continue;
        /* merge prev with area */
        prev->start = area->start;
        prev->next = area->next;
        prev->offset = area->offset;
        if (area->shmem)
            area->shmem->count--;
        ret = MemManager_FreeObj(areaCache, area);
        ASSERT_ERROR(ret);
        area = prev;
    }
    return;
}

/* insert mem area into task areas list sorted by address */
static __inline__ void InsertMemArea(MemManager_TaskInfo *taskInfo, MemManager_MemArea *area) {
    MemManager_MemArea **pnext;

    pnext = &taskInfo->areas;
    while (*pnext && (*pnext)->start >= area->start)
        pnext = &(*pnext)->next;
    area->next = *pnext;
    *pnext = area;
}

void *MemManager_AllocUserPages(void* start, unsigned int size, unsigned int align, int prop, TaskManager_Task *task) {
    MemManager_MemArea *area;
    MemManager_TaskInfo *taskInfo;
    int ret;

    ASSERT_RET(size, ERROR_TO_PTR(E_BUG));
    ASSERT_RET(!(size & 0xfff), ERROR_TO_PTR(E_BUG));
    ASSERT_RET(!(align & 0xfff), ERROR_TO_PTR(E_BUG));
    ASSERT_RET(!((unsigned int)start & 0xfff), ERROR_TO_PTR(E_BUG));

    if (size > MemManager_START_SYS_MEM || (unsigned int)start > (MemManager_START_SYS_MEM-size))
        return ERROR_TO_PTR(E_FAULT);
    taskInfo = TaskManager_GetTaskInfo(task, ID);
    if (IS_PTR_ERROR(taskInfo)) {
        DEBUG_MSG_PRE(PTR_TO_ERROR(taskInfo));
        return ERROR_TO_PTR(E_NOT_USER_TASK);
    }
    TaskManager_P(&taskInfo->areasSem, 0, 0);
    /* get address if not FIXED */
    if (!(prop & MemManager_FIXED_ADDR)) {
        start = GetFreeMemArea(taskInfo, (unsigned int)start, size, align);
        if (IS_PTR_ERROR(start))
            goto errorV;
    } else {
        /* check if range is already in use */
        area = FindLastMemArea(taskInfo->areas, (unsigned int)start);
        if (area && ((unsigned int)start + size) > area->start) {
            /* return error or unmap range? */
            if (prop & MemManager_FAIL_MAPPED) {
                start = ERROR_TO_PTR(E_IN_USE);
                goto errorV;
            }
            ret = FreeUserPagesNoLock(start, size, MemManager_CHANGE_SYS, task, taskInfo);
            ASSERT_ERROR(ret);
        }
    }
    /* alloc MemArea struct */
    area = MemManager_AllocObj(areaCache);
    if (IS_PTR_ERROR(area)) {
        start = area;
        goto errorV;
    }
    /* i386 doesn't has an exec flag, but to exec we need to read */
    /* on a i386 if one can write it can read too */
    if (prop & (MemManager_PROT_EXEC + MemManager_PROT_WRITE))
        prop |= MemManager_PROT_READ;
    area->start = (unsigned int)start;
    area->end = (unsigned int)start + size;
    area->prop = prop & MemManager_MAP_FLAGS_MASK;
    area->shmem = NULL;
    /* insert into areas list */
    InsertMemArea(taskInfo, area);
    MergeMemArea(taskInfo, area->start, area->end);
errorV:
    TaskManager_V(&taskInfo->areasSem, 0);
    return start;
}

void *MemManager_AttachShMemory(MemManager_ShMem *shmem, void *start, unsigned int offset, unsigned int size, int prop, TaskManager_Task *task) {
    MemManager_TaskInfo *taskInfo;
    MemManager_MemArea *area;
    int ret;

    ASSERT_RET(shmem, ERROR_TO_PTR(E_BUG));
    ASSERT_RET(size, ERROR_TO_PTR(E_BUG));
    ASSERT_RET(!(size & 0xfff), ERROR_TO_PTR(E_BUG));
    ASSERT_RET(!(offset & 0xfff), ERROR_TO_PTR(E_BUG));
    ASSERT_RET(!((unsigned int)start & 0xfff), ERROR_TO_PTR(E_BUG));

    if ((offset+size) > shmem->size || (unsigned int)start > (MemManager_START_SYS_MEM-size))
        return ERROR_TO_PTR(E_FAULT);
    /* check special SH_PDE conditions */
    if (shmem->prop & MemManager_SH_PDE && ((unsigned int)start & 0x3fffff || offset & 0x3ffff || size & 0x3ffff))
        return ERROR_TO_PTR(E_MISMATCHED_VALUES);
    taskInfo = TaskManager_GetTaskInfo(task, ID);
    if (IS_PTR_ERROR(taskInfo)) {
        DEBUG_MSG_PRE(PTR_TO_ERROR(taskInfo));
        return ERROR_TO_PTR(E_NOT_USER_TASK);
    }
    TaskManager_P(&taskInfo->areasSem, 0, 0);
    /* get address if not FIXED */
    if (!(prop & MemManager_FIXED_ADDR)) {
        start = GetFreeMemArea(taskInfo, (unsigned int)start, size, 0);
        if (IS_PTR_ERROR(start))
            goto errorV;
    } else {
        /* check if range is already in use */
        area = FindLastMemArea(taskInfo->areas, (unsigned int)start);
        if (area && ((unsigned int)start + size) > area->start) {
            /* return error or unmap range? */
            if (prop & MemManager_FAIL_MAPPED) {
                start = ERROR_TO_PTR(E_IN_USE);
                goto errorV;
            }
            ret = FreeUserPagesNoLock(start, size, MemManager_CHANGE_SYS, task, taskInfo);
            ASSERT_ERROR(ret);
        }
    }
    /* alloc MemArea struct */
    area = MemManager_AllocObj(areaCache);
    if (IS_PTR_ERROR(area)) {
        start = area;
        goto errorV;
    }
    /* i386 doesn't has an exec flag, but to exec we need to read */
    /* on a i386 if one can write it can read too */
    if (prop & (MemManager_PROT_EXEC + MemManager_PROT_WRITE))
        prop |= MemManager_PROT_READ;
    area->start = (unsigned int)start;
    area->end = (unsigned int)start + size;
    area->prop = prop & MemManager_MAP_FLAGS_MASK;
    area->shmem = shmem;
    area->offset = offset;
    shmem->count++;
    /* insert into areas list */
    InsertMemArea(taskInfo, area);
    MergeMemArea(taskInfo, area->start, area->end);
errorV:
    TaskManager_V(&taskInfo->areasSem, 0);
    return start;
}

int MemManager_i386_Block(void *ptr, unsigned int size) {
    return E_OK;
}

int MemManager_i386_Catch(void *start, void *end, void *function, int prop) {
    MemManager_Exception *ex;

    ASSERT_RET(function, E_BUG);

    ex = MemManager_Alloc(sizeof(MemManager_Exception), MemManager_BLOCK_MEM);
    if (IS_PTR_ERROR(ex))
        return PTR_TO_ERROR(ex);
    ex->start = (unsigned int)start;
    ex->end = (unsigned int)end;
    ex->function = function;
    ex->prop = prop & MemManager_i386_EX_RETURN;
    TaskManager_P(&exMutex, 0, 0);
    if (prop & MemManager_i386_EX_USER_MODE) {
        ex->next = pageFaultUserMode;
        pageFaultUserMode = ex;
    } else {
        ex->next = pageFaultWho;
        pageFaultWho = ex;
    }
    TaskManager_V(&exMutex, 0);
    return E_OK;
}

int MemManager_i386_CreateDescriptor(MemManager_i386_SegDescriptor descriptor, unsigned short *selector) {
    int i, error;

    ASSERT_RET(selector, E_BUG);

    i = MemManager_FIRST_FREE_GDTE;
    error = E_NO_MEMORY;
    TaskManager_P(&GDTSem, 0, 0);
    do {
        if (!GDT[i<<1]) {
            *(MemManager_i386_SegDescriptor*)(&GDT[i<<1]) = descriptor;
            *selector = i<<3;
            error = E_OK;
            break;
        }
    } while (++i<MemManager_SIZE_GDT);
    TaskManager_V(&GDTSem, 0);
    return error;
}

MemManager_Cache *MemManager_CreateObjCache(unsigned int size, unsigned int align, int prop, void (*ctor)(void *obj, unsigned int size), void (*dtor)(void *obj, unsigned int size)) {
    unsigned int slabSize, small;
    MemManager_Cache *newCache;

    ASSERT_RET(size, ERROR_TO_PTR(E_BUG));

    /* request too big */
    if (size > (MemManager_BIG_LG_PG_CHUNCK<<12))
        return ERROR_TO_PTR(E_TOO_BIG);
    /* max alignment - 32 bytes */
    if (align > 5)
        align = 5;
    else if (align == 0)
        align = 2;
    /* the alignment is in bytes */
    align = 1 << align;
    /* wrong align */
    if (align > size)
        return ERROR_TO_PTR(E_MISMATCHED_VALUES);
    newCache = MemManager_Alloc(sizeof(MemManager_Cache), 0);
    if (IS_PTR_ERROR(newCache))
        return newCache;
    /* each obj has a pointer to the slab or free list */
    size += 4;
    if (size < 12)
        size = 12;
    size = ALIGN_DATA_UP(size, align);
    slabSize = 1;
    small = (1<<12) % size;
    /* find best # of pages? */
    if (size > 120) {
        /* if bigger than 8 KB, allocate directly from logical pages */
        if (size > 8192) {
            slabSize = size >> 12;
            small = size & 0xfff;
            if (small)
                slabSize++;
        } else {
            unsigned int limit;
            /* limit the # of pages we search */
            if (size > 2048)
                limit = 6;
            else
                limit = 4;
            /* figure out the best # of pages */
            if (small > 64) {
                unsigned int i;
                for (i=2;i < limit;i++)
                    if (((i<<12) % size) < small) {
                        small = (i<<12) % size;
                        slabSize = i;
                    }
            }
        }
    }
    newCache->size = size;
    newCache->num = (slabSize<<12) / size;
    newCache->nPages = slabSize;
    newCache->prop = (prop&0x01)<<6;
    /* slab info inside the slab? */
    if (small >= sizeof(MemManager_Slab)) {
        small -= sizeof(MemManager_Slab);
        newCache->prop |= 0x20;
    }
    newCache->prop |= align;
    newCache->free = newCache->last = NULL;
    /* set the # of colors */
    newCache->nextColor = 0;
    newCache->lastColor = (small & ~(align-1))+align;
    newCache->ctor = ctor;
    newCache->dtor = dtor;
    newCache->sem.count = 1;
    newCache->sem.data = NULL;
    return newCache;
}

unsigned int MemManager_i386_CreatePGD(void) {
    unsigned int *newPGD, ret;
    int error;

    newPGD = MemManager_i386_AllocPages(1, MemManager_i386_RW_PAGE+MemManager_i386_ALLOC_PHYS);
    if (IS_PTR_ERROR(newPGD))
        return (unsigned int)newPGD;
    /* copy system map */
    MemMan_Copy(newPGD+MemManager_FIRST_SYS_PDE, &MemManager_PG_DIR[MemManager_FIRST_SYS_PDE], MemManager_N_SYS_PDE<<2);
    /* make the last PDE point to the PGD */
    newPGD[1023] = MemManager_PG_TABS[(unsigned int)newPGD>>12];
    ret = MemManager_PG_TABS[(unsigned int)newPGD>>12] & 0xfffff000;
    /* zero user area */
    MemMan_Set(newPGD, 0, 4*MemManager_FIRST_SYS_PDE);
    error = MemManager_i386_FreePages(newPGD, 1, MemManager_i386_DONT_FREE_PHYS);
    ASSERT_ERROR(error);
    return ret;
}

MemManager_ShMem *MemManager_CreateShMemory(unsigned int size, int prop, const MemManager_ShMemFunctions *functions, void *arg) {
    MemManager_ShMem *shmem;

    ASSERT_RET(size, ERROR_TO_PTR(E_BUG));
    ASSERT_RET(functions, ERROR_TO_PTR(E_BUG));
    ASSERT_RET(!(size & 0xfff), ERROR_TO_PTR(E_BUG));

    if (size > MemManager_START_SYS_MEM)
        return ERROR_TO_PTR(E_TOO_BIG);
    shmem = MemManager_Alloc(sizeof(MemManager_ShMem), 0);
    if (IS_PTR_ERROR(shmem))
        return shmem;
    shmem->count = 0;
    shmem->size = size;
    shmem->prop = prop;
    shmem->pages = NULL;
    shmem->functions = functions;
    shmem->arg = arg;
    return shmem;
}

int MemManager_i386_DestroyDescriptor(unsigned int selector) {
    if (selector < MemManager_FIRST_FREE_GDTE)
        return E_BAD_VALUE;
    TaskManager_P(&GDTSem, 0, 0);
    GDT[selector<<1] = 0;
    GDT[(selector<<1)+1] = 0;
    TaskManager_V(&GDTSem, 0);
    return E_OK;
}

int MemManager_DestroyObjCache(MemManager_Cache *cache) {
    MemManager_Slab *slabMgt;
    int ret;

    ASSERT_RET(cache, E_BUG);

    slabMgt = cache->free;
    if (slabMgt) {
        do {
            void *pgPtr;
            MemManager_Slab *tmp;
            if (slabMgt->inUse)
                return E_IN_USE;
            /* get slab start */
            if (cache->prop & 0x20)
                pgPtr = (void*)((unsigned int)slabMgt & 0xfffff000)-cache->nPages+1;
            else {
                pgPtr = (void*)((unsigned int)slabMgt->free & 0xfffff000);
                if (cache->num != 1 && cache->nPages != 1) {
                    /* we must figure out the start of the slab */
                    unsigned char i = 1;
                    char *nextF = *(char**)slabMgt->free;
                    /* go through all free objs to find the start of the slab */
                    do {
                        if (((unsigned int)nextF & 0xfffff000) != (unsigned int)pgPtr) {
                            if (((unsigned int)nextF & 0xfffff000) < (unsigned int)pgPtr)
                                pgPtr = (void*)((unsigned int)nextF & 0xfffff000);
                            i++;
                            /* if we've went through all objs break */
                            if (i == cache->nPages)
                                break;
                        }
                        nextF = *(char**)nextF;
                    } while (nextF);
                }
            }
            /* call the destructor */
            if (cache->dtor) {
                char *obj = slabMgt->free;
                do {
                    cache->dtor(obj+4, cache->size);
                    obj = *(char**)obj;
                } while (obj);
            }
            /* free slab pages */
            ret = MemManager_i386_FreePages(pgPtr, cache->nPages, 0);
            ASSERT_ERROR(ret);
            /* free slab info */
            tmp = slabMgt->next;
            if (!(cache->prop & 0x20)) {
                ret = MemManager_Free(slabMgt);
                ASSERT_ERROR(ret);
            }
            /* get the next slab */
            slabMgt = tmp;
        } while (slabMgt != cache->free);
    }
    ret = MemManager_Free(cache);
    ASSERT_ERROR(ret);
    return E_OK;
}

int MemManager_DestroyShMemory(MemManager_ShMem *shmem) {
    MemManager_ShPage *currPg;
    unsigned int *ptes;
    int ret;

    ASSERT_RET(shmem, E_BUG);

    if (shmem->count)
        return E_IN_USE;
    ptes = NULL;
    if (shmem->prop & MemManager_SH_PDE) {
        /* alloc logical page to map PDEs */
        ptes = MemManager_i386_AllocPages(1, 0);
        if (IS_PTR_ERROR(ptes))
            return PTR_TO_ERROR(ptes);
    }
    currPg = shmem->pages;
    while (currPg) {
        MemManager_ShPage *tmp;

        if (shmem->prop & MemManager_SH_PDE) {
            unsigned int i;
            void *voidRet;

            /* we must free PDE's pages */
            /* map PDE */
            voidRet = MemManager_i386_MapPhysPages(ptes, currPg->physPg, 1, MemManager_i386_RW_PAGE);
            ASSERT_ERROR_PTR(voidRet);
            for (i=0; i < 1024; i++)
                if (ptes[i]) {
                    ret = MemManager_i386_FreePhysPages(ptes[i], 1);
                    ASSERT_ERROR(ret);
                }
        }
        ret = MemManager_i386_FreePhysPages(currPg->physPg, 1);
        ASSERT_ERROR(ret);
        tmp = currPg->next;
        ret = MemManager_FreeObj(pageCache, currPg);
        ASSERT_ERROR(ret);
        currPg = tmp;
    }
    if (shmem->prop & MemManager_SH_PDE) {
        /* free logical page to map PDEs */
        ret = MemManager_i386_FreePages(ptes, 1, MemManager_i386_DONT_FREE_PHYS);
        ASSERT_ERROR(ret);
    }
    ret = MemManager_Free(shmem);
    ASSERT_ERROR(ret);
    return E_OK;
}

/* free page frames of range start-end, NOT including end */
void FreePTES(unsigned int start, unsigned int end, int freePages, TaskManager_Task *task) {
    int ret;

    if (!task) {
        unsigned int *pdes, *ptes;

        /* current task, use current PGD */
        pdes = &MemManager_PG_DIR[start>>22];
        ptes = &MemManager_PG_TABS[start>>12];
        while (*pdes) {
            do {
                if (*ptes) {
                    if (freePages) {
                        ret = MemManager_i386_FreePhysPages(*ptes, 1);
                        ASSERT_ERROR(ret);
                    }
                    *ptes = 0;
                }
                start += 4096;
                ptes++;
            } while (((unsigned int)ptes & 0xfff) && start < end);
            if (start >= end)
                break;
            pdes++;
        }
    } else {
        int pgd;
        unsigned int *pdes, *ptes, *savedPDES, *savedPTES;
        void *voidRet;

        pgd = TaskManager_GetTaskMMap(task);
        /* map PGD */
        pdes = MemManager_i386_MapPhysPages(NULL, (unsigned int)pgd, 1, MemManager_i386_RW_PAGE+MemManager_i386_ALLOC_PAGES);
        if (IS_PTR_ERROR(pdes)) {
            DEBUG_MSG_PRE(PTR_TO_ERROR(pdes));
            return;
        }
        savedPDES = pdes;
        /* alloc logical page to map PDEs */
        savedPTES = MemManager_i386_AllocPages(1, 0);
        if (IS_PTR_ERROR(savedPTES)) {
            DEBUG_MSG_PRE(PTR_TO_ERROR(savedPTES));
            return;
        }
        /* goto right PDE and PTE */
        pdes += start >> 22;
        ptes = savedPTES + ((start & 0x3fffff) >> 12);
        while (*pdes) {
            /* map PDE */
            voidRet = MemManager_i386_MapPhysPages(savedPTES, *pdes, 1, MemManager_i386_RW_PAGE);
            ASSERT_ERROR_PTR(voidRet);
            do {
                if (*ptes) {
                    if (freePages) {
                        ret = MemManager_i386_FreePhysPages(*ptes, 1);
                        ASSERT_ERROR(ret);
                    }
                    *ptes = 0;
                }
                start += 4096;
                ptes++;
            } while (((unsigned int)ptes & 0xfff) && start < end);
            if (start >= end)
                break;
            ptes = savedPTES;
            pdes++;
        }
        /* free logical page to map PDEs */
        ret = MemManager_i386_FreePages(savedPTES, 1, MemManager_i386_DONT_FREE_PHYS);
        ASSERT_ERROR(ret);
        /* unmap PGD */
        ret = MemManager_i386_FreePages(savedPDES, 1, MemManager_i386_DONT_FREE_PHYS);
        ASSERT_ERROR(ret);
    }
}

/* free pdes of range start-(start+len-1) */
void FreePDEs(unsigned int start, unsigned int end, int freePDE, TaskManager_Task *task) {
    unsigned int len, n;
    int ret;


    len = ALIGN_DATA_UP(end, 0x400000) - ALIGN_DATA_DOWN(start, 0x400000);
    n = len >> 22;
    if (!task) {
        unsigned int *pdes;

        /* start with end because we alloc areas starting from end */
        pdes = &MemManager_PG_DIR[(end-1)>>22];
        do {
            if (*pdes) {
                unsigned int *ptes, i;
                /* see if all ptes are free */
                ptes = (unsigned int*)((unsigned int)pdes << 10);
                for (i=0; i < 1024; i++)
                    if (ptes[i])
                        break;
                if (i == 1024) {
                    /* if so, free pde */
                    if (freePDE) {
                        ret = MemManager_i386_FreePhysPages(*pdes, 1);
                        ASSERT_ERROR(ret);
                    }
                    *pdes = 0;
                }
            }
            pdes--;
        } while (--n);
    } else {
        int pgd;
        unsigned int *pdes, *savedPDES, *ptes;

        pgd = TaskManager_GetTaskMMap(task);
        /* map PGD */
        pdes = MemManager_i386_MapPhysPages(NULL, (unsigned int)pgd, 1, MemManager_i386_RW_PAGE+MemManager_i386_ALLOC_PAGES);
        if (IS_PTR_ERROR(pdes)) {
            DEBUG_MSG_PRE(PTR_TO_ERROR(pdes));
            return;
        }
        savedPDES = pdes;
        /* alloc logical page to map PDEs */
        ptes = MemManager_i386_AllocPages(1, 0);
        if (IS_PTR_ERROR(ptes)) {
            DEBUG_MSG_PRE(PTR_TO_ERROR(ptes));
            return;
        }
        pdes += (end-1) >> 22;
        do {
            if (*pdes) {
                unsigned int i;
                void *voidRet;

                /* map PDE */
                voidRet = MemManager_i386_MapPhysPages(ptes, *pdes, 1, MemManager_i386_RW_PAGE);
                ASSERT_ERROR_PTR(voidRet);
                /* see if all ptes are free */
                for (i=0; i < 1024; i++)
                    if (ptes[i])
                        break;
                if (i == 1024) {
                    /* if so, free pde */
                    if (freePDE) {
                        ret = MemManager_i386_FreePhysPages(*pdes, 1);
                        ASSERT_ERROR(ret);
                    }
                    *pdes = 0;
                }
            }
            pdes--;
        } while (--n);
        /* free logical page to map PDEs */
        ret = MemManager_i386_FreePages(ptes, 1, MemManager_i386_DONT_FREE_PHYS);
        ASSERT_ERROR(ret);
        /* unmap PGD */
        ret = MemManager_i386_FreePages(savedPDES, 1, MemManager_i386_DONT_FREE_PHYS);
        ASSERT_ERROR(ret);
    }
}

int MemManager_DetachShMemory(MemManager_ShMem *shmem, void *start, unsigned int size, int prop, TaskManager_Task *task) {
    MemManager_MemArea *area, *prev, **pp, *extra;
    MemManager_TaskInfo *taskInfo;
    int error, ret;

    ASSERT_RET(shmem, E_BUG);
    ASSERT_RET(size, E_BUG);
    ASSERT_RET(!(size & 0xfff), E_BUG);
    ASSERT_RET(!((unsigned int)start & 0xfff), E_BUG);

    if (size > shmem->size || (unsigned int)start > (MemManager_START_SYS_MEM-size))
        return E_FAULT;
    /* check special SH_PDE conditions */
    if (shmem->prop & MemManager_SH_PDE && ((unsigned int)start & 0x3fffff || size & 0x3ffff))
        return E_MISMATCHED_VALUES;
    taskInfo = TaskManager_GetTaskInfo(task, ID);
    if (IS_PTR_ERROR(taskInfo)) {
        DEBUG_MSG_PRE(PTR_TO_ERROR(taskInfo));
        return E_NOT_USER_TASK;
    }
    TaskManager_P(&taskInfo->areasSem, 0, 0);
    /* find first area that may be in the range */
    area = FindFirstMemAreaPrev(taskInfo->areas, (unsigned int)start + size, &prev);
    /* check if it's in the range */
    if (!area || area->end <= (unsigned int)start)
        goto returnOK;

    /* we may need one additional area if we make a hole */
    extra = MemManager_AllocObj(areaCache);
    if (IS_PTR_ERROR(extra)) {
        error = PTR_TO_ERROR(extra);
        goto returnV;
    }

    /* process each area */
    pp = prev ? &prev->next : &taskInfo->areas;
    do {
        /* can we free this area? */
        if (!area->shmem) {
            error = E_FAULT;
            goto returnV;
        }
        if (area->prop & MemManager_ONLY_SYS && !(prop & MemManager_CHANGE_SYS)) {
            error = E_PERMISSION_DENIED;
            goto returnV;
        }
        if (area->start >= (unsigned int)start && area->end <= ((unsigned int)start+size)) {
            MemManager_MemArea *tmp;
            /* free the whole area */
            if (!(area->shmem->prop & MemManager_SH_PDE)) {
                /* free memory pages */
                FreePTES(area->start, area->end, 0, task);
            }
            /* free pdes */
            FreePDEs(area->start, area->end, !(area->shmem->prop & MemManager_SH_PDE), task);
            *pp = area->next;
            tmp = area->next;
            area->shmem->count--;
            ret = MemManager_FreeObj(areaCache, area);
            ASSERT_ERROR(ret);
            area = tmp;
            continue;
        }

        if (area->end <= ((unsigned int)start+size)) {
            /* free up part */
            if (!(area->shmem->prop & MemManager_SH_PDE)) {
                /* free memory pages */
                FreePTES((unsigned int)start, area->end, 0, task);
            }
            /* free pdes */
            FreePDEs((unsigned int)start, area->end, !(area->shmem->prop & MemManager_SH_PDE), task);
            area->end = (unsigned int)start;
            pp = &area->next;
        } else if (area->start >= (unsigned int)start) {
            /* free down part */
            if (!(area->shmem->prop & MemManager_SH_PDE)) {
                /* free memory pages */
                FreePTES(area->start, (unsigned int)start + size, 0, task);
            }
            /* free pdes */
            FreePDEs(area->start, (unsigned int)start + size, !(area->shmem->prop & MemManager_SH_PDE), task);
            area->start = (unsigned int)start + size;
            area->offset += ((unsigned int)start + size) - area->start;
        } else {
            /* we have a hole */
            if (!(area->shmem->prop & MemManager_SH_PDE)) {
                /* free memory pages */
                FreePTES((unsigned int)start, (unsigned int)start + size, 0, task);
            }
            /* free pdes */
            FreePDEs((unsigned int)start, (unsigned int)start + size, !(area->shmem->prop & MemManager_SH_PDE), task);
            extra->start = (unsigned int)start + size;
            extra->end = area->end;
            extra->prop = area->prop;
            extra->next = area;
            extra->shmem = area->shmem;
            /* one more area using the same shmem */
            area->shmem->count++;
            extra->offset += extra->start - area->start;
            *pp = extra;
            area->end = (unsigned int)start;
            extra = NULL;
            break;
        }
        area = area->next;
    } while (area && area->end > (unsigned int)start);

    if (extra) {
        ret = MemManager_FreeObj(areaCache, extra);
        ASSERT_ERROR(ret);
    }

    if (!task) {
        /* we need to invalidate TLB */
        __asm__ __volatile__ (
            "movl %%cr3, %%ecx\n"
            "movl %%ecx, %%cr3"
            : : : "%ecx" );
    }

returnOK:
    error = E_OK;
returnV:
    TaskManager_V(&taskInfo->areasSem, 0);
    return error;
}

int MemManager_Free(void *ptr) {
    MemManager_Slab *slabMgt;

    ASSERT_RET(ptr, E_BUG);

    /* get the slab */
    slabMgt = (MemManager_Slab *)*((void **)((char*)ptr-4));
    /* no slab, free logical pages */
    if (slabMgt == NULL)
        return MemManager_i386_FreePages(ptr-8, *(unsigned int*)((char*)ptr-8), 0);
    return MemManager_FreeObj(slabMgt->cache, ptr);
}

int MemManager_i386_FreeException(void *start, void *function, int prop) {
    MemManager_Exception *curr, **prev;
    int error;

    TaskManager_P(&exMutex, 0, 0);
    if (prop & MemManager_i386_EX_USER_MODE) {
        prev = &pageFaultUserMode;
        curr = pageFaultUserMode;
    } else {
        prev = &pageFaultWho;
        curr = pageFaultWho;
    }
    error = E_NOT_REGISTERED;
    while (curr) {
        if (curr->start == (unsigned int)start && curr->function == function) {
            int ret;
            *prev = curr->next;
            ret = MemManager_Free(curr);
            ASSERT_ERROR(ret);
            error = E_OK;
            break;
        }
        prev = &curr->next;
        curr = curr->next;
    }
    TaskManager_V(&exMutex, 0);
    return error;
}

int MemManager_FreeObj(MemManager_Cache *cache, void *obj) {
    MemManager_Slab *slabMgt;
    void **freePtr;

    ASSERT_RET(cache, E_BUG);
    ASSERT_RET(obj, E_BUG);

    /* get slab info */
    freePtr = (void **)((unsigned int)obj-4);
    slabMgt = (MemManager_Slab *)*freePtr;
    TaskManager_P(&cache->sem, 0, 0);
    /* insert obj in free list */
    *freePtr = slabMgt->free;
    slabMgt->inUse--;
    slabMgt->free = (char*)((unsigned int)obj-4);
    if (*freePtr) {
        if (slabMgt->inUse) {
func_end:
            TaskManager_V(&cache->sem, 0);
            return E_OK;
        }
        /* totally free slabs must be in the end of the list */
        if (cache->last != slabMgt && slabMgt->next->inUse) {
            /* insert at the end of the list */
            if (cache->free == slabMgt)
                cache->free = slabMgt->next;
            /* remove from list */
            slabMgt->prev->next = slabMgt->next;
            slabMgt->next->prev = slabMgt->prev;
            /* insert in the end of list */
            slabMgt->prev = cache->last;
            slabMgt->next = cache->last->next;
            cache->last->next->prev = slabMgt;
            cache->last->next = slabMgt;
            cache->last = slabMgt;
        }
        goto func_end;
    }
    /* if next slab is full we're at the wrong place */
    if (!slabMgt->next->free) {
        /* remove from list */
        slabMgt->prev->next = slabMgt->next;
        slabMgt->next->prev = slabMgt->prev;
        /* put in the free */
        slabMgt->next = cache->free;
        slabMgt->prev = cache->free->prev;
        cache->free->prev->next = slabMgt;
        cache->free->prev = slabMgt;
    }
    cache->free = slabMgt;
    goto func_end;
}

int MemManager_i386_FreePages(void *pages, unsigned int n, int prop) {
    unsigned int *ptes, currPg;

    ASSERT_RET((unsigned int)pages >= (unsigned int)MemManager_START_SYS_MEM, E_BUG);

    /* request too big? */
    if (n > MemManager_BIG_LG_PG_CHUNCK)
        return E_TOO_BIG;
    currPg = (unsigned int)pages>>12;
    ptes = &MemManager_PG_TABS[currPg];
    TaskManager_P(&lgPgSem, 0, 0);
    if (firstFreeLgPg > currPg)
        firstFreeLgPg = currPg;
    freeLgPgs += n;
    do {
        if (!(prop & MemManager_i386_DONT_FREE_PHYS) && (*ptes & 0x1)) {
            int ret;
            ret = MemManager_i386_FreePhysPages(*ptes, 1);
            ASSERT_ERROR(ret);
        }
        *ptes = 0;
        ptes++;
    } while (--n);
    TaskManager_V(&lgPgSem, 0);
    /* invalidate TLB */
    __asm__ __volatile__ (
        "movl %%cr3, %%ecx\n"
        "movl %%ecx, %%cr3"
        : : : "%ecx" );
    return E_OK;
}

int MemManager_i386_FreePhysPages(unsigned int pages, unsigned int n) {
    pages >>= 12;

    ASSERT_RET(pages <= physPgVEnd, E_BUG);
    ASSERT_RET(n, E_BUG);
    ASSERT_RET(n <= (pages+1), E_BUG);

    TaskManager_P(&physPgSem, 0, 0);
    freePhysPgs += n;
    /* Remenber: the end of the array is the start of mem! */
    pages = physPgVEnd - pages;
    do {
        physPages[pages--] = MemManager_PHYS_PG_FREE;
    } while (--n);
    if (pages < firstFreePhysPg)
        firstFreePhysPg = pages;
    TaskManager_V(&physPgSem, 0);
    return E_OK;
}

/* task areasSem must be locked before calling this function */
int FreeUserPagesNoLock(void *start, unsigned int size, int prop, TaskManager_Task *task, MemManager_TaskInfo *taskInfo) {
    MemManager_MemArea *area, *prev, **pp, *extra;
    int ret;

    /* find first area that may be in the range */
    area = FindFirstMemAreaPrev(taskInfo->areas, (unsigned int)start + size, &prev);
    /* check if it's in the range */
    if (!area || area->end <= (unsigned int)start)
        return E_OK;

    /* we may need one additional area if we make a hole */
    extra = MemManager_AllocObj(areaCache);
    if (IS_PTR_ERROR(extra))
        return PTR_TO_ERROR(extra);

    /* process each area */
    pp = prev ? &prev->next : &taskInfo->areas;
    do {
        /* can we free this area? */
        if (area->shmem)
            return E_FAULT;
        if (area->prop & MemManager_ONLY_SYS && !(prop & MemManager_CHANGE_SYS))
            return E_PERMISSION_DENIED;
        if (area->start >= (unsigned int)start && area->end <= ((unsigned int)start+size)) {
            MemManager_MemArea *tmp;
            /* free the whole area */
            /* free memory pages */
            FreePTES(area->start, area->end, 1, task);
            *pp = area->next;
            tmp = area->next;
            ret = MemManager_FreeObj(areaCache, area);
            ASSERT_ERROR(ret);
            area = tmp;
            continue;
        }

        if (area->end <= ((unsigned int)start+size)) {
            /* free up part */
            /* free memory pages */
            FreePTES((unsigned int)start, area->end, 1, task);
            area->end = (unsigned int)start;
            pp = &area->next;
        } else if (area->start >= (unsigned int)start) {
            /* free down part */
            /* free memory pages */
            FreePTES(area->start, (unsigned int)start + size, 1, task);
            area->start = (unsigned int)start + size;
        } else {
            /* we have a hole */
            /* free memory pages */
            FreePTES((unsigned int)start, (unsigned int)start + size, 1, task);
            extra->start = (unsigned int)start + size;
            extra->end = area->end;
            extra->prop = area->prop;
            extra->next = area;
            extra->shmem = NULL;
            *pp = extra;
            area->end = (unsigned int)start;
            extra = NULL;
            break;
        }
        area = area->next;
    } while (area && area->end > (unsigned int)start);

    if (extra) {
        ret = MemManager_FreeObj(areaCache, extra);
        ASSERT_ERROR(ret);
    }

    /* free pdes */
    FreePDEs((unsigned int)start, (unsigned int)start + size, 1, task);
    if (!task) {
        /* we need to invalidate TLB */
        __asm__ __volatile__ (
            "movl %%cr3, %%ecx\n"
            "movl %%ecx, %%cr3"
            : : : "%ecx" );
    }

    return E_OK;
}

int MemManager_FreeUserPages(void *start, unsigned int size, int prop, TaskManager_Task *task) {
    MemManager_TaskInfo *taskInfo;
    int error;

    ASSERT_RET(size, E_BUG);
    ASSERT_RET(!(size & 0xfff), E_BUG);
    ASSERT_RET(!((unsigned int)start & 0xfff), E_BUG);

    if (size > MemManager_START_SYS_MEM || (unsigned int)start > (MemManager_START_SYS_MEM-size))
        return E_FAULT;
    taskInfo = TaskManager_GetTaskInfo(task, ID);
    if (IS_PTR_ERROR(taskInfo)) {
        DEBUG_MSG_PRE(PTR_TO_ERROR(taskInfo));
        return E_NOT_USER_TASK;
    }
    TaskManager_P(&taskInfo->areasSem, 0, 0);
    error = FreeUserPagesNoLock(start, size, prop, task, taskInfo);
    TaskManager_V(&taskInfo->areasSem, 0);
    return error;
}

int MemManager_MapPageAtUserPage(void *logPage, unsigned int physPage, TaskManager_Task *task) {
    MemManager_MemArea *area;
    MemManager_TaskInfo *taskInfo;
    int ret, error;

    taskInfo = TaskManager_GetTaskInfo(task, ID);
    if (IS_PTR_ERROR(taskInfo)) {
        DEBUG_MSG_PRE(PTR_TO_ERROR(taskInfo));
        return E_NOT_USER_TASK;
    }
    logPage = (void*)((unsigned int)logPage & 0xfffff000);
    TaskManager_P(&taskInfo->areasSem, 0, 0);
    /* find first area that may be in the range */
    area = FindFirstMemArea(taskInfo->areas, (unsigned int)logPage + 4096);
    /* check if it's in the range */
    error = E_FAULT;
    if (!area || area->end <= (unsigned int)logPage)
        goto returnV;
    if (!task) {
        unsigned int *pdes, *ptes;

        /* current task, use current PGD */
        /* do we need to allocate PGT? */
        pdes = &MemManager_PG_DIR[(unsigned int)logPage>>22];
        if (!*pdes) {
            unsigned int physPg;

            physPg = MemManager_i386_AllocPhysPages(1, 0);
            if (IS_PTR_ERROR(physPg)) {
                error = PTR_TO_ERROR(physPg);
                goto returnV;
            }
            *pdes = physPg | 0x7;
            /* we need to invalidate TLB */
            __asm__ __volatile__ (
                "movl %%cr3, %%ecx\n"
                "movl %%ecx, %%cr3"
                : : : "%ecx" );
            MemMan_Set((char*)((unsigned int)pdes<<10), 0, 4096);
        }
        /* page already alloc? */
        ptes = &MemManager_PG_TABS[(unsigned int)logPage>>12];
        error = E_IN_USE;
        if (*ptes)
            goto returnV;

        *ptes = physPage | 0x5;
        if (area->prop & MemManager_PROT_WRITE)
            *ptes |= 0x2;
        /* we need to invalidate TLB */
        __asm__ __volatile__ (
            "movl %%cr3, %%ecx\n"
            "movl %%ecx, %%cr3"
            : : : "%ecx" );
        error = E_OK;
    } else {
        int pgd;
        unsigned int *pdes, *ptes, *savedPDES, *savedPTES;
        void *voidRet;

        pgd = TaskManager_GetTaskMMap(task);
        /* map PGD */
        pdes = MemManager_i386_MapPhysPages(NULL, (unsigned int)pgd, 1, MemManager_i386_RW_PAGE+MemManager_i386_ALLOC_PAGES);
        if (IS_PTR_ERROR(pdes)) {
            error = PTR_TO_ERROR(pdes);
            goto returnV;
        }
        savedPDES = pdes;
        /* goto right PDE */
        pdes += (unsigned int)logPage >> 22;
        if (!*pdes) {
            unsigned int physPg, *tmpPDE;

            physPg = MemManager_i386_AllocPhysPages(1, 0);
            if (IS_PTR_ERROR(physPg)) {
                error = PTR_TO_ERROR(physPg);
                goto errorFreePGD;
            }
            *pdes = physPg | 0x7;
            /* clear PDE */
            tmpPDE = MemManager_i386_MapPhysPages(NULL, physPg, 1, MemManager_i386_RW_PAGE+MemManager_i386_ALLOC_PAGES+MemManager_i386_CLEAR_PAGES);
            if (IS_PTR_ERROR(tmpPDE)) {
                error = PTR_TO_ERROR(tmpPDE);
                goto errorFreePGD;
            }
            /* free logical page */
            ret = MemManager_i386_FreePages(tmpPDE, 1, MemManager_i386_DONT_FREE_PHYS);
            ASSERT_ERROR(ret);
        }
        /* alloc logical page to map PDEs */
        savedPTES = MemManager_i386_AllocPages(1, 0);
        if (IS_PTR_ERROR(savedPTES)) {
            error = PTR_TO_ERROR(savedPTES);
            goto errorFreePGD;
        }
        /* map PDE */
        voidRet = MemManager_i386_MapPhysPages(savedPTES, *pdes, 1, MemManager_i386_RW_PAGE);
        ASSERT_ERROR_PTR(voidRet);
        /* goto right PTE */
        ptes = savedPTES + (((unsigned int)logPage & 0x3fffff) >> 12);
        /* page already alloc? */
        error = E_IN_USE;
        if (*ptes)
            goto errorFreePDEMap;

        *ptes = physPage | 0x5;
        if (area->prop & MemManager_PROT_WRITE)
            *ptes |= 0x2;
        error = E_OK;
errorFreePDEMap:
        /* free logical page to map PDEs */
        ret = MemManager_i386_FreePages(savedPTES, 1, MemManager_i386_DONT_FREE_PHYS);
        ASSERT_ERROR(ret);
errorFreePGD:
        /* unmap PGD */
        ret = MemManager_i386_FreePages(savedPDES, 1, MemManager_i386_DONT_FREE_PHYS);
        ASSERT_ERROR(ret);
    }
returnV:
    TaskManager_V(&taskInfo->areasSem, 0);
    return error;
}

void *MemManager_i386_MapPhysPages(void *to, unsigned int start, unsigned int n, int prop) {
    unsigned int *ptes;
    int pageProp;

    ASSERT_RET(n, ERROR_TO_PTR(E_BUG));

    if (prop & MemManager_i386_ALLOC_PAGES) {
        to = MemManager_i386_AllocPages(n, 0);
        if (IS_PTR_ERROR(to))
            return to;
    } else
        ASSERT_RET((unsigned int)to >= MemManager_START_SYS_MEM, ERROR_TO_PTR(E_BUG));

    pageProp = 1;
    pageProp |= ((prop << 1) & 0x1f);
    ptes = &MemManager_PG_TABS[(unsigned int)to>>12];
    TaskManager_P(&lgPgSem, 0, 0);
    do {
        *ptes = start | pageProp;
        ptes++;
        start += 4096;
    } while (--n);
    TaskManager_V(&lgPgSem, 0);
    return to;
}

static __inline__ int HandleShMemFault(unsigned int where, MemManager_MemArea *area) {
    unsigned int *pdes, *ptes;
    int ret;

    /* do we need to allocate PGT? */
    pdes = &MemManager_PG_DIR[where>>22];
    if (!*pdes) {
        unsigned int physPg;

        /* shared PDE? */
        if (!(area->shmem->prop & MemManager_SH_PDE)) {
            physPg = MemManager_i386_AllocPhysPages(1, 0);
            if (IS_PTR_ERROR(physPg))
                return 1;
            *pdes = physPg | 0x7;
            /* we need to invalidate TLB */
            __asm__ __volatile__ (
                "movl %%cr3, %%ecx\n"
                "movl %%ecx, %%cr3"
                : : : "%ecx" );
            MemMan_Set((char*)((unsigned int)pdes<<10), 0, 4096);
        } else {
            MemManager_ShPage *currPg, *newPg, **pnext;
            unsigned int offset;

            /* try to find PDE */
            offset = area->offset + (ALIGN_DATA_DOWN(where, 0x400000) - area->start);
            pnext = &area->shmem->pages;
            currPg = area->shmem->pages;
            while (currPg && offset <= currPg->offset) {
                if (offset == currPg->offset) {
                    *pdes = currPg->physPg | 0x7;
                    /* we need to invalidate TLB */
                    __asm__ __volatile__ (
                        "movl %%cr3, %%ecx\n"
                        "movl %%ecx, %%cr3"
                        : : : "%ecx" );
                    goto tryPage;
                }
                pnext = &currPg->next;
                currPg = currPg->next;
            }
            /* alloc new PDE */
            physPg = MemManager_i386_AllocPhysPages(1, 0);
            if (IS_PTR_ERROR(physPg))
                return 1;
            newPg = MemManager_AllocObj(pageCache);
            if (IS_PTR_ERROR(newPg)) {
                ret = MemManager_i386_FreePhysPages(physPg, 1);
                ASSERT_ERROR(ret);
                return 1;
            }
            newPg->physPg = physPg;
            newPg->offset = offset;
            /* insert into list */
            *pnext = newPg;
            newPg->next = currPg;
            *pdes = physPg | 0x7;
            /* we need to invalidate TLB */
            __asm__ __volatile__ (
                "movl %%cr3, %%ecx\n"
                "movl %%ecx, %%cr3"
                : : : "%ecx" );
            MemMan_Set((char*)((unsigned int)pdes<<10), 0, 4096);
        }
    }
tryPage:
    /* page already alloc? */
    ptes = &MemManager_PG_TABS[where>>12];
    if (!*ptes) {
        unsigned int physPg, offset;

        offset = area->offset + (ALIGN_DATA_DOWN(where, 0x1000) - area->start);
        if (!(area->shmem->prop & MemManager_SH_PDE)) {
            MemManager_ShPage *currPg, *newPg, **pnext;

            /* try to find page */
            pnext = &area->shmem->pages;
            currPg = area->shmem->pages;
            while (currPg && offset <= currPg->offset) {
                if (offset == currPg->offset) {
                    physPg = currPg->physPg;
                    goto setupPTE;
                }
                pnext = &currPg->next;
                currPg = currPg->next;
            }
            /* alloc new page */
            physPg = MemManager_i386_AllocPhysPages(1, 0);
            if (IS_PTR_ERROR(physPg))
                return 1;
            newPg = MemManager_AllocObj(pageCache);
            if (IS_PTR_ERROR(newPg)) {
                ret = MemManager_i386_FreePhysPages(physPg, 1);
                ASSERT_ERROR(ret);
                return 1;
            }
            newPg->physPg = physPg;
            newPg->offset = offset;
            /* insert into list */
            *pnext = newPg;
            newPg->next = currPg;
        } else {
            physPg = MemManager_i386_AllocPhysPages(1, 0);
            if (IS_PTR_ERROR(physPg))
                return 1;
        }
setupPTE:
        /* don't let user code access this page now */
        *ptes = physPg | 0x3;
        if (area->shmem->prop & MemManager_SH_CLEAR || area->shmem->functions->readpage) {
            /* we need to invalidate TLB */
            __asm__ __volatile__ (
                "movl %%cr3, %%ecx\n"
                "movl %%ecx, %%cr3"
                : : : "%ecx" );
            if (area->shmem->prop & MemManager_SH_CLEAR)
                /* clear page */
                MemMan_Set((char*)((unsigned int)ptes<<10), 0, 4096);
            /* call readpage function */
            if (area->shmem->functions->readpage && area->shmem->functions->readpage(area->shmem->arg, (char*)((unsigned int)ptes<<10), offset))
                return 1;
        }
        /* remove write flag */
        *ptes ^= 0x2;
        if (area->prop & MemManager_PROT_WRITE)
            *ptes |= 0x2;
        /* enable user code access */
        *ptes |= 0x4;
        /* we need to invalidate TLB */
        __asm__ __volatile__ (
            "movl %%cr3, %%ecx\n"
            "movl %%ecx, %%cr3"
            : : : "%ecx" );
    }
    return 0;
}

static __inline__ int HandleUserMemAreaFault(unsigned int where, MemManager_MemArea *area) {
    unsigned int *pdes, *ptes;

    /* do we need to allocate PGT? */
    pdes = &MemManager_PG_DIR[where>>22];
    if (!*pdes) {
        unsigned int physPg;

        physPg = MemManager_i386_AllocPhysPages(1, 0);
        if (IS_PTR_ERROR(physPg))
            return 1;
        *pdes = physPg | 0x7;
        /* we need to invalidate TLB */
        __asm__ __volatile__ (
            "movl %%cr3, %%ecx\n"
            "movl %%ecx, %%cr3"
            : : : "%ecx" );
        MemMan_Set((char*)((unsigned int)pdes<<10), 0, 4096);
    }
    /* page already alloc? */
    ptes = &MemManager_PG_TABS[where>>12];
    if (!*ptes) {
        unsigned int physPg;

        physPg = MemManager_i386_AllocPhysPages(1, 0);
        if (IS_PTR_ERROR(physPg))
            return 1;
        /* don't let user code access this page now */
        *ptes = physPg | 0x3;
        if (area->prop & MemManager_CLEAR_PAGE) {
            /* we need to invalidate TLB */
            __asm__ __volatile__ (
                "movl %%cr3, %%ecx\n"
                "movl %%ecx, %%cr3"
                : : : "%ecx" );
            /* clear page */
            MemMan_Set((char*)((unsigned int)ptes<<10), 0, 4096);
        }
        /* remove write flag */
        *ptes ^= 0x2;
        if (area->prop & MemManager_PROT_WRITE)
            *ptes |= 0x2;
        /* enable user code access */
        *ptes |= 0x4;
        /* we need to invalidate TLB */
        __asm__ __volatile__ (
            "movl %%cr3, %%ecx\n"
            "movl %%ecx, %%cr3"
            : : : "%ecx" );
    }
    return 0;
}

static __inline__ int HandleUserPageFault(unsigned int where, IntManager_i386_ExRegs *regs) {
    MemManager_TaskInfo *taskInfo;
    MemManager_MemArea *area;
    int error;

    taskInfo = TaskManager_GetTaskInfo(NULL, ID);
    if (IS_PTR_ERROR(taskInfo))
        return 1;
    TaskManager_P(&taskInfo->areasSem, 0, 0);
    area = FindLastMemArea(taskInfo->areas, where);
    error = 1;
    if (area && where >= area->start) {
        /* first check protections */
        if (area->prop & MemManager_PROT_READ && (!(regs->ErrorCode & 0x02) || area->prop & MemManager_PROT_WRITE)) {
            if (area->shmem) {
                if (!HandleShMemFault(where, area))
                    error = 0;
            } else {
                if (!HandleUserMemAreaFault(where, area))
                    error = 0;
            }
        }
    }
    TaskManager_V(&taskInfo->areasSem, 0);
    return error;
}

int MemManager_i386_PageFault(IntManager_i386_ExRegs *regs) {
    MemManager_Exception *curr;
    SysModManager_ModuleInfo moduleInfo;
    unsigned int index;
    char modulePath[256];
    unsigned int address;
    unsigned int *esp;
    int j, ret;

    /* get cr2 value and enable interrupts */
    __asm__ __volatile__ ("movl %%cr2, %0; sti;" : "=r" (address));

    /* if in user space, try areas */
    if (address < MemManager_START_SYS_MEM) {
        int error;

        /* thread can't be destroyed while exec this function */
        TaskManager_UndestroyableThread();
        error = HandleUserPageFault(address, regs);
        TaskManager_DestroyableThread(NULL);
        if (!error)
            return E_HANDLED_INT;
    }

    /* FIXME: we need to lock exMutex, but the handler may kill the thread/task.
       What about an atexit function? */

    /* next try user mode handle */
    if (regs->ErrorCode & 0x04) {
        curr = pageFaultUserMode;
        while (curr) {
            int error;

            if (curr->prop) {
                regs->EIP = (unsigned int)curr->function;
                return E_HANDLED_INT;
            }
            error = ((int (*)(unsigned int,IntManager_i386_ExRegs*))curr->function)(address, regs);
            if (error == E_HANDLED_INT)
                return E_HANDLED_INT;
            curr = curr->next;
        }
    }
    /* now who handlers */
    curr = pageFaultWho;
    while (curr) {
        if (regs->EIP >= curr->start && regs->EIP <= curr->end) {
            int error;

            if (curr->prop) {
                regs->EIP = (unsigned int)curr->function;
                return E_HANDLED_INT;
            }
            error = ((int (*)(unsigned int,IntManager_i386_ExRegs*))curr->function)(address, regs);
            if (error == E_HANDLED_INT)
                return E_HANDLED_INT;
        }
        curr = curr->next;
    }
    /* unhandled page fault! */
    SysModManager_DebugMsg("MemManager: Unhandled page fault! Current thread will hang.\n\n");
    SysModManager_DebugMsg("EAX: 0x%08x  EBX: 0x%08x  ECX: 0x%08x  EDX: 0x%08x\n", regs->EAX, regs->EBX, regs->ECX, regs->EDX);
    SysModManager_DebugMsg("EDI: 0x%08x  ESI: 0x%08x  EBP: 0x%08x  ESP: 0x%08x\n\n", regs->EDI, regs->ESI, regs->EBP, regs->ESP);
    SysModManager_DebugMsg("Faulting address: 0x%08x  ErrorCode: 0x%x\n", address, regs->ErrorCode);
    SysModManager_DebugMsg("CS: 0x%x  EIP: 0x%08x  EFLAGS: 0x%x\n\n", regs->CS, regs->EIP, regs->EFLAGS);
    /* find the module */
    index = 0;
    while (!(ret = SysModManager_GetModuleInfoIndex(index++, &moduleInfo, modulePath, sizeof(modulePath)))) {
        if (moduleInfo.memStart && regs->EIP >= moduleInfo.memStart && regs->EIP < (moduleInfo.memStart+moduleInfo.memSize))
            break;
    }
    if (!ret) {
        SysModManager_DebugMsg("Module: %s - Start: 0x%x\n\n", modulePath, moduleInfo.memStart);
    } else {
        SysModManager_DebugMsg("Error: EIP address has no module: 0x%x\n\n", ret);
    }
    SysModManager_DebugMsg("At most 32 words from stack:\n");
    j = 0;
    esp = &regs->ESP;
    do {
        int i;
        i = 0;
        do {
            SysModManager_DebugMsg("0x%08x  ", *esp);
            esp++;
        } while (++i < 4 && ((unsigned int)esp & 0xfff));
        SysModManager_DebugMsg("\n");
        j += 4;
    } while (j < 32 && ((unsigned int)esp & 0xfff));
    /* maybe a panic function? */
    while (1)
        __asm__ __volatile__ ("hlt");
}

int Phase0Start(SysModManager_Module *modID) {
    unsigned int n, *ptes, physPg;
    int ret;

    ID = modID;
    /* load GDT */
    __asm__ __volatile__ (
        "movl %0, 2(%0)\n"
        "lgdtl (%0)"
        : : "r" (GDT));
    physPg = (*(unsigned int*)0xfffffffc) >> 12;
    /* set PGD and first page table PHYSICAL pages blocked */
    physPages[MemManager_TMP_PGS - 1 - physPg] = MemManager_PHYS_PG_BLOCKED;
    physPages[MemManager_TMP_PGS - 1 - (physPg+1)] = MemManager_PHYS_PG_BLOCKED;
    freePhysPgs -= 2;
    /* setup low memory */
    freePhysPgs -= MemManager_TMP_PGS_LOW1MB_RESERVED_SIZE;
    n = MemManager_TMP_PGS_LOW1MB_RESERVED_SIZE;
    do {
        physPages[MemManager_TMP_PGS - 1 - 0xff + n - 1] = MemManager_PHYS_PG_RESERVED;
    } while(--n);
    /* set system PHYSICAL pages blocked */
    ptes = &MemManager_PG_TABS[MemManager_FIRST_SYS_PTE];
    n = MemManager_TMP_PGS;
    do {
        if (*ptes) {
            physPages[MemManager_TMP_PGS - 1 - (*ptes >> 12)] = MemManager_PHYS_PG_BLOCKED;
            freePhysPgs--;
            freeLgPgs--;
        }
        ptes++;
    } while (--n);
    if ((ret = SysModManager_StartModule((unsigned int)IntManager_i386_CatchException)) ||
        (ret = SysModManager_StartModule((unsigned int)MemMan_FindByte)) ||
        (ret = SysModManager_StartModule((unsigned int)TaskManager_P))) {
        DEBUG_MSG_PRE(ret);
        return 1;
    }
    if ((ret = IntManager_i386_CatchException(0x0e, MemManager_i386_PageFault, (void*)0, (void*)0xffffffff))) {
        DEBUG_MSG_PRE(ret);
        return 1;
    }
    return 0;
}

int Phase1Start(SysModManager_Module *modID) {
    unsigned int i, j, nPhysPgFree, lastPg, nPages, expected, startArea, endArea;
    int ret;

    if ((ret = SysModManager_StartModule((unsigned int)MemController_GetMemArea)) ||
        (ret = SysModManager_StartModule((unsigned int)MemMan_FindByte)) ||
        (ret = SysModManager_StartModule((unsigned int)TaskManager_P))) {
        DEBUG_MSG_PRE(ret);
        return 1;
    }
    /* number of physical pages in system */
    i = nPhysPgFree = lastPg = 0;
    while (MemController_GetMemArea(i++, &startArea, &endArea) == E_OK) {
        nPhysPgFree += (endArea-startArea+1);
        lastPg = endArea;
    }
    /* figure out the number of pages to allocate */
    nPages = ALIGN_DATA_UP(lastPg, 0x1000) >> 12;
    physPages = MemManager_i386_AllocPages(nPages, MemManager_i386_RW_PAGE+MemManager_i386_ALLOC_PHYS);
    if (IS_PTR_ERROR(physPages)) {
        DEBUG_MSG_PRE(PTR_TO_ERROR(physPages));
        return 1;
    }
    j = expected = 0;
    while (MemController_GetMemArea(j++, &startArea, &endArea) == 0) {
        /* is a reserved area? */
        if (expected != startArea)
            MemMan_Set(physPages+(lastPg-(startArea-1)), MemManager_PHYS_PG_RESERVED, startArea-expected);
        expected = endArea+1;
        /* set the area as free */
        MemMan_Set(physPages+(lastPg-endArea), MemManager_PHYS_PG_FREE, endArea-startArea+1);
    }
    /* copy 640KB pages */
    MemMan_Copy(physPages+(lastPg-0x100+MemManager_TMP_PGS_LOW1MB_RESERVED_SIZE), tmpPhysPgs+MemManager_TMP_PGS
                -0x100+MemManager_TMP_PGS_LOW1MB_RESERVED_SIZE, 0x100-MemManager_TMP_PGS_LOW1MB_RESERVED_SIZE);
    /* copy 1MB and up pages */
    MemMan_Copy(physPages+(lastPg-MemManager_TMP_PGS), tmpPhysPgs, MemManager_TMP_PGS-0x100);
    freePhysPgs = nPhysPgFree - (MemManager_TMP_PGS-freePhysPgs);
    physPgVEnd = lastPg;
    /* create MemManager caches */
    areaCache = MemManager_CreateObjCache(sizeof(MemManager_MemArea), 0, 0, NULL, NULL);
    pageCache = MemManager_CreateObjCache(sizeof(MemManager_ShPage), 0, 0, NULL, NULL);
    if (IS_PTR_ERROR(areaCache) || IS_PTR_ERROR(pageCache)) {
        DEBUG_MSG_PRE(PTR_TO_ERROR(pageCache));
        return 1;
    }
    return 0;
}

static __inline__ int ProtectShMemoryFixupAll(MemManager_TaskInfo *taskInfo, MemManager_MemArea *area, int prop) {
    area->prop = prop;
    MergeMemArea(taskInfo, area->start, area->end);
    return E_OK;
}

static __inline__ int ProtectShMemoryFixupStart(MemManager_TaskInfo *taskInfo, MemManager_MemArea *area, unsigned int end, int prop) {
    MemManager_MemArea *new;

    new = MemManager_AllocObj(areaCache);
    if (IS_PTR_ERROR(new))
        return PTR_TO_ERROR(new);
    new->offset = area->offset;
    area->offset += end - area->start;
    new->start = area->start;
    area->start = end;
    new->end = end;
    new->prop = prop;
    new->shmem = area->shmem;
    new->shmem->count++;
    InsertMemArea(taskInfo, new);
    MergeMemArea(taskInfo, new->start, new->end);
    return E_OK;
}

static __inline__ int ProtectShMemoryFixupEnd(MemManager_TaskInfo *taskInfo, MemManager_MemArea *area, unsigned int start, int prop) {
    MemManager_MemArea *new;

    new = MemManager_AllocObj(areaCache);
    if (IS_PTR_ERROR(new))
        return PTR_TO_ERROR(new);
    new->offset = area->offset + start - area->start;
    new->end = area->end;
    area->end = start;
    new->start = start;
    new->prop = prop;
    new->shmem = area->shmem;
    new->shmem->count++;
    InsertMemArea(taskInfo, new);
    MergeMemArea(taskInfo, new->start, new->end);
    return E_OK;
}

static __inline__ int ProtectShMemoryFixupMiddle(MemManager_TaskInfo *taskInfo, MemManager_MemArea *area, unsigned int start, unsigned int end, int prop) {
    MemManager_MemArea *left, *right;
    int ret;

    left = MemManager_AllocObj(areaCache);
    if (IS_PTR_ERROR(left))
        return PTR_TO_ERROR(left);
    right = MemManager_AllocObj(areaCache);
    if (IS_PTR_ERROR(right)) {
        ret = MemManager_FreeObj(areaCache, left);
        ASSERT_ERROR(ret);
        return PTR_TO_ERROR(right);
    }
    *left = *area;
    left->end = start;
    *right = *area;
    right->start = end;
    right->offset = left->offset + end - left->start;
    area->start = start;
    area->offset = left->offset + start - left->start;
    area->end = end;
    area->prop = prop;
    area->shmem->count += 2;
    /* don't need to merge */
    InsertMemArea(taskInfo, left);
    InsertMemArea(taskInfo, right);
    return E_OK;
}

void ChangeProtection(unsigned int start, unsigned int end, int prop, TaskManager_Task *task) {
    int newP, ret;

    /* setup page protection */
    if (prop & MemManager_PROT_READ) {
        newP = 5;
        if (prop & MemManager_PROT_WRITE)
            newP |= 0x2;
    } else
        newP = 0;
    if (!task) {
        unsigned int *pdes, *ptes;

        pdes = &MemManager_PG_DIR[start>>22];
        ptes = &MemManager_PG_TABS[start>>12];
        while (*pdes) {
            do {
                if (*ptes)
                    *ptes = (*ptes & 0xfffffff8) | newP;
                start += 4096;
                ptes++;
            } while (((unsigned int)ptes & 0xfff) && start < end);
            if (start >= end)
                break;
            pdes++;
        }
        /* we need to invalidate TLB */
        __asm__ __volatile__ (
            "movl %%cr3, %%ecx\n"
            "movl %%ecx, %%cr3"
            : : : "%ecx" );
    } else {
        int pgd;
        unsigned int *pdes, *savedPDES, *savedPTES;

        pgd = TaskManager_GetTaskMMap(task);
        /* map PGD */
        pdes = MemManager_i386_MapPhysPages(NULL, (unsigned int)pgd, 1, MemManager_i386_RW_PAGE+MemManager_i386_ALLOC_PAGES);
        if (IS_PTR_ERROR(pdes)) {
            DEBUG_MSG_PRE(PTR_TO_ERROR(pdes));
            return;
        }
        savedPDES = pdes;
        /* alloc logical page to map PDEs */
        savedPTES = MemManager_i386_AllocPages(1, 0);
        if (IS_PTR_ERROR(savedPTES)) {
            DEBUG_MSG_PRE(PTR_TO_ERROR(savedPTES));
            return;
        }
        while (*pdes) {
            unsigned int *ptes;

            /* map PDE */
            ptes = MemManager_i386_MapPhysPages(savedPTES, *pdes, 1, MemManager_i386_RW_PAGE);
            ASSERT_ERROR_PTR(ptes);
            do {
                if (*ptes)
                    *ptes = (*ptes & 0xfffffff8) | newP;
                start += 4096;
                ptes++;
            } while (((unsigned int)ptes & 0xfff) && start < end);
            if (start >= end)
                break;
            pdes++;
        }
        /* free logical page to map PDEs */
        ret = MemManager_i386_FreePages(savedPTES, 1, MemManager_i386_DONT_FREE_PHYS);
        ASSERT_ERROR(ret);
        /* unmap PGD */
        ret = MemManager_i386_FreePages(savedPDES, 1, MemManager_i386_DONT_FREE_PHYS);
        ASSERT_ERROR(ret);
    }
}

int ProtectShMemoryFixup(MemManager_TaskInfo *taskInfo, MemManager_MemArea *area, unsigned int start, unsigned int end, int prop, TaskManager_Task *task) {
    int ret;

    /* nothing to do */
    if (prop == area->prop)
        return E_OK;
    if (start == area->start) {
        if (end == area->end)
            ret = ProtectShMemoryFixupAll(taskInfo, area, prop);
        else
            ret = ProtectShMemoryFixupStart(taskInfo, area, end, prop);
    } else if (end == area->end)
        ret = ProtectShMemoryFixupEnd(taskInfo, area, start, prop);
    else
        ret = ProtectShMemoryFixupMiddle(taskInfo, area, start, end, prop);
    if (ret)
        return ret;
    ChangeProtection(start, end, prop, task);
    return E_OK;
}

int MemManager_ProtectShMemory(MemManager_ShMem *shmem, const void *start, unsigned int size, int prop, TaskManager_Task *task) {
    MemManager_MemArea *area;
    MemManager_TaskInfo *taskInfo;
    unsigned int nend;
    int error, newProp;

    ASSERT_RET(shmem, E_BUG);
    ASSERT_RET(size, E_BUG);
    ASSERT_RET(!(size & 0xfff), E_BUG);
    ASSERT_RET(!((unsigned int)start & 0xfff), E_BUG);

    if (size > shmem->size || (unsigned int)start > (MemManager_START_SYS_MEM-size))
        return E_FAULT;
    /* check special SH_PDE conditions */
    if (shmem->prop & MemManager_SH_PDE && ((unsigned int)start & 0x3fffff || size & 0x3ffff))
        return E_MISMATCHED_VALUES;
    taskInfo = TaskManager_GetTaskInfo(task, ID);
    if (IS_PTR_ERROR(taskInfo)) {
        DEBUG_MSG_PRE(PTR_TO_ERROR(taskInfo));
        return E_NOT_USER_TASK;
    }
    newProp = prop & MemManager_MAP_FLAGS_MASK;
    /* i386 doesn't has an exec flag, but to exec we need to read */
    /* on a i386 if one can write it can read too */
    if (newProp & (MemManager_PROT_EXEC + MemManager_PROT_WRITE))
        newProp |= MemManager_PROT_READ;
    TaskManager_P(&taskInfo->areasSem, 0, 0);
    /* find first area that may be in the range */
    area = FindFirstMemArea(taskInfo->areas, (unsigned int)start + size);
    /* check if it's in the range */
    error = E_FAULT;
    if (!area || area->end <= (unsigned int)start)
        goto returnV;

    /* process each area */
    for (nend = (unsigned int)start+size; ; ) {
        unsigned int tmp;
        MemManager_MemArea *next;

        /* can we change this area? */
        if (!area->shmem) {
            error = E_FAULT;
            break;
        }
        if (area->prop & MemManager_ONLY_SYS && !(prop & MemManager_CHANGE_SYS)) {
            error = E_PERMISSION_DENIED;
            goto returnV;
        }

        /* end? */
        if ((unsigned int)start >= area->start) {
            error = ProtectShMemoryFixup(taskInfo, area, (unsigned int)start, nend, newProp, task);
            break;
        }
        tmp = area->start;
        next = area->next;
        error = ProtectShMemoryFixup(taskInfo, area, tmp, nend, newProp, task);
        if (error)
            break;
        nend = tmp;
        area = next;
        if (!area ||  area->end != nend) {
            error = E_FAULT;
            break;
        }
    }

returnV:
    TaskManager_V(&taskInfo->areasSem, 0);
    return error;
}

int MemManager_i386_SetPagesProp(void *start, unsigned int n, int prop) {
    unsigned int pageProp, *ptes;

    ASSERT_RET((unsigned int)start >= MemManager_START_SYS_MEM, E_BUG);
    ASSERT_RET(n <= ~(unsigned int)start, E_BUG);

    pageProp = ((prop << 1) & 0x1f);
    ptes = &MemManager_PG_TABS[(unsigned int)start>>12];
    TaskManager_P(&lgPgSem, 0, 0);
    do {
        *ptes = (*ptes & 0xffffffe1) | pageProp;
        ptes++;
    } while (--n);
    TaskManager_V(&lgPgSem, 0);
    if (!(prop & MemManager_i386_DONT_INV_TLB)) {
        __asm__ __volatile__ (
            "movl %%cr3, %%ecx\n"
            "movl %%ecx, %%cr3"
            : : : "%ecx");
    }
    return E_OK;
}

static __inline__ int UserPagesProtFixupAll(MemManager_TaskInfo *taskInfo, MemManager_MemArea *area, int prop) {
    area->prop = prop;
    MergeMemArea(taskInfo, area->start, area->end);
    return E_OK;
}

static __inline__ int UserPagesProtFixupStart(MemManager_TaskInfo *taskInfo, MemManager_MemArea *area, unsigned int end, int prop) {
    MemManager_MemArea *new;

    new = MemManager_AllocObj(areaCache);
    if (IS_PTR_ERROR(new))
        return PTR_TO_ERROR(new);
    new->start = area->start;
    area->start = end;
    new->end = end;
    new->prop = prop;
    new->shmem = NULL;
    InsertMemArea(taskInfo, new);
    MergeMemArea(taskInfo, new->start, new->end);
    return E_OK;
}

static __inline__ int UserPagesProtFixupEnd(MemManager_TaskInfo *taskInfo, MemManager_MemArea *area, unsigned int start, int prop) {
    MemManager_MemArea *new;

    new = MemManager_AllocObj(areaCache);
    if (IS_PTR_ERROR(new))
        return PTR_TO_ERROR(new);
    new->end = area->end;
    area->end = start;
    new->start = start;
    new->prop = prop;
    new->shmem = NULL;
    InsertMemArea(taskInfo, new);
    MergeMemArea(taskInfo, new->start, new->end);
    return E_OK;
}

static __inline__ int UserPagesProtFixupMiddle(MemManager_TaskInfo *taskInfo, MemManager_MemArea *area, unsigned int start, unsigned int end, int prop) {
    MemManager_MemArea *left, *right;
    int ret;

    left = MemManager_AllocObj(areaCache);
    if (IS_PTR_ERROR(left))
        return PTR_TO_ERROR(left);
    right = MemManager_AllocObj(areaCache);
    if (IS_PTR_ERROR(right)) {
        ret = MemManager_FreeObj(areaCache, left);
        ASSERT_ERROR(ret);
        return PTR_TO_ERROR(right);
    }
    *left = *area;
    left->end = start;
    *right = *area;
    right->start = end;
    area->start = start;
    area->end = end;
    area->prop = prop;
    /* don't need to merge */
    InsertMemArea(taskInfo, left);
    InsertMemArea(taskInfo, right);
    return E_OK;
}

int UserPagesProtFixup(MemManager_TaskInfo *taskInfo, MemManager_MemArea *area, unsigned int start, unsigned int end, int prop, TaskManager_Task *task) {
    int ret;

    /* nothing to do */
    if (prop == area->prop)
        return E_OK;
    if (start == area->start) {
        if (end == area->end)
            ret = UserPagesProtFixupAll(taskInfo, area, prop);
        else
            ret = UserPagesProtFixupStart(taskInfo, area, end, prop);
    } else if (end == area->end)
        ret = UserPagesProtFixupEnd(taskInfo, area, start, prop);
    else
        ret = UserPagesProtFixupMiddle(taskInfo, area, start, end, prop);
    if (ret)
        return ret;
    ChangeProtection(start, end, prop, task);
    return E_OK;
}

int MemManager_SetUserPagesProt(const void *start, unsigned int size, int prop, TaskManager_Task *task) {
    MemManager_MemArea *area;
    MemManager_TaskInfo *taskInfo;
    unsigned int nend;
    int error, newProp;

    ASSERT_RET(size, E_BUG);
    ASSERT_RET(!(size & 0xfff), E_BUG);
    ASSERT_RET(!((unsigned int)start & 0xfff), E_BUG);

    if (size > MemManager_START_SYS_MEM || (unsigned int)start > (MemManager_START_SYS_MEM-size))
        return E_FAULT;
    taskInfo = TaskManager_GetTaskInfo(task, ID);
    if (IS_PTR_ERROR(taskInfo)) {
        DEBUG_MSG_PRE(PTR_TO_ERROR(taskInfo));
        return E_NOT_USER_TASK;
    }
    newProp = prop & MemManager_MAP_FLAGS_MASK;
    /* i386 doesn't has an exec flag, but to exec we need to read */
    /* on a i386 if one can write it can read too */
    if (newProp & (MemManager_PROT_EXEC + MemManager_PROT_WRITE))
        newProp |= MemManager_PROT_READ;
    TaskManager_P(&taskInfo->areasSem, 0, 0);
    /* find first area that may be in the range */
    area = FindFirstMemArea(taskInfo->areas, (unsigned int)start + size);
    /* check if it's in the range */
    error = E_FAULT;
    if (!area || area->end <= (unsigned int)start)
        goto returnV;

    /* process each area */
    for (nend = (unsigned int)start+size; ; ) {
        unsigned int tmp;
        MemManager_MemArea *next;

        /* can we change this area? */
        if (area->shmem) {
            error = E_FAULT;
            break;
        }
        if (area->prop & MemManager_ONLY_SYS && !(prop & MemManager_CHANGE_SYS)) {
            error = E_PERMISSION_DENIED;
            goto returnV;
        }

        /* end? */
        if ((unsigned int)start >= area->start) {
            error = UserPagesProtFixup(taskInfo, area, (unsigned int)start, nend, newProp, task);
            break;
        }
        tmp = area->start;
        next = area->next;
        error = UserPagesProtFixup(taskInfo, area, tmp, nend, newProp, task);
        if (error)
            break;
        nend = tmp;
        area = next;
        if (!area ||  area->end != nend) {
            error = E_FAULT;
            break;
        }
    }

returnV:
    TaskManager_V(&taskInfo->areasSem, 0);
    return error;
}

void MemManager_TaskCreated(TaskManager_Task *creator, TaskManager_Task *created) {
    MemManager_TaskInfo *taskInfo;
    void *voidRet;

    taskInfo = MemManager_Alloc(sizeof(MemManager_TaskInfo), 0);
    if (IS_PTR_ERROR(taskInfo))
        return;
    taskInfo->areas = NULL;
    taskInfo->areasSem.count = 1;
    taskInfo->areasSem.data = NULL;
    voidRet = TaskManager_SetTaskInfo(created, ID, taskInfo, 0);
    ASSERT_ERROR_PTR(voidRet);
}

void MemManager_TaskDestroyed(TaskManager_Task *task) {
    MemManager_TaskInfo *taskInfo;
    MemManager_MemArea *area;
    int ret;

    taskInfo = TaskManager_DelTaskInfo(task, ID);
    if (IS_PTR_ERROR(taskInfo))
        return;
    /* free all areas - shared and not shared */
    area = taskInfo->areas;
    while (area) {
        MemManager_MemArea *tmp;

        if (!area->shmem) {
            /* free memory pages */
            FreePTES(area->start, area->end, 1, task);
            FreePDEs(area->start, area->end, 1, task);
            tmp = area->next;
        } else {
            /* shared PDE do nothing */
            if (!(area->shmem->prop & MemManager_SH_PDE)) {
                /* free memory pages */
                FreePTES(area->start, area->end, 0, task);
                FreePDEs(area->start, area->end, 1, task);
            }
            tmp = area->next;
            area->shmem->count--;
        }
        ret = MemManager_FreeObj(areaCache, area);
        ASSERT_ERROR(ret);
        area = tmp;
    }
    ret = MemManager_Free(taskInfo);
    ASSERT_ERROR(ret);
}

int MemManager_i386_Unblock(void *ptr, unsigned int size) {
    return E_OK;
}

void *MemManager_UAllocUserPages(void *start, unsigned int size, unsigned int align, int prop) {
    /* check args */
    if (!size || size & 0xfff || align & 0xfff || (unsigned int)start & 0xfff)
        return ERROR_TO_PTR(E_BUG);
    prop &= MemManager_MAP_USER_FLAGS_MASK;
    return MemManager_AllocUserPages(start, size, align, prop, NULL);
}

int MemManager_UFreeUserPages(void *start, unsigned int size) {
    /* check args */
    if (!size || size & 0xfff || (unsigned int)start & 0xfff)
        return E_BUG;
    return MemManager_FreeUserPages(start, size, 0, NULL);
}

int MemManager_USetUserPagesProt(const void *start, unsigned int size, int prop) {
    /* check args */
    if (!size || size & 0xfff || (unsigned int)start & 0xfff)
        return E_BUG;
    prop &= MemManager_MAP_USER_FLAGS_MASK;
    return MemManager_SetUserPagesProt(start, size, prop, NULL);
}
