/*
*****************************************
* This file implements the BlockDev interface for the partition manager
*
* Version: 0.1.0
* Author: Luiz Henrique Shigunov
* E-mail: luizshigunov@users.sourceforge.net
* WEB: http://www.sourceforge.net/projects/modulos
* Last change: 14/03/2004
*
* See the interface specification to know what each function does.
*
* Copyright (C) 2004, Luiz Henrique Shigunov
*
* This module is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2.1 of the License, or (at your option) any later version.
*
* This module is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this module; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*****************************************
*/

#include <stddef.h>
#include <blockdev.h>
#include <debug.h>
#include <diskpartition.h>
#include <errormanager.h>
#include <memmanager.h>
#include <stringman.h>
#include <sysdef.h>
#include <sysmodmanager.h>

typedef struct _Implementation {
    struct _Implementation *next;
    int (*readF)(BlockDev_IO*);
    int (*writeF)(BlockDev_IO*);
    SysModManager_Imp *imp;
} Implementation;

typedef struct _Partition {
    struct _Partition *next;
    unsigned int id;
    Implementation *imp;
    BlockDev_UnitDesc *unitDesc;
    unsigned int startBlock;
    unsigned int nBlocks;
} Partition;

typedef struct _DiskPartition {
    struct _DiskPartition *next;
    int (*readF)(SysModManager_Imp*,unsigned int,DiskPartition_Description*,unsigned int,unsigned int);
    SysModManager_Imp *imp;
} DiskPartition;

SysModManager_Module *ID;
Implementation *imps;
Partition *parts;
unsigned int nParts;
DiskPartition *diskParts;

int DoGetUnitDesc(Partition *part, BlockDev_UnitDesc *unitDesc, unsigned int size, int prop) {
    unsigned int unitSize, nameSize;
    char buf[32];

    ASSERT_RET(part, E_BUG);

    unitSize = sizeof(BlockDev_UnitDesc);
    nameSize = 0;
    if (!(prop & BlockDev_NO_NAME)) {
        nameSize = StringMan_Length(part->unitDesc->unitName);
        unitSize += nameSize + StringMan_PrintF(buf, ": Partition %u", part->id) + 1;
    }
    if (size < unitSize)
        return E_BUFFER_TOO_SMALL;

    unitDesc->unit = part->id;
    unitDesc->type = part->unitDesc->type;
    unitDesc->flags = part->unitDesc->flags;
    unitDesc->blockSize = part->unitDesc->blockSize;
    unitDesc->nBlocks = part->nBlocks;
    if (!(prop & BlockDev_NO_NAME)) {
        unitDesc->unitName = ((char*)&unitDesc->nBlocks)+sizeof(unitDesc->nBlocks);
        StringMan_Copy(unitDesc->unitName, part->unitDesc->unitName);
        StringMan_Copy(unitDesc->unitName+nameSize, buf);
    } else
        unitDesc->unitName = NULL;
    return E_OK;
}

int BlockDev_GetUnitDesc(unsigned int unit, BlockDev_UnitDesc *unitDesc, unsigned int size, int prop) {
    Partition *part;

    ASSERT_RET(unitDesc, E_BUG);

    part = parts;
    while (part) {
        if (part->id == unit)
            return DoGetUnitDesc(part, unitDesc, size, prop);
        part = part->next;
    }
    return E_NO_SUCH_UNIT;
}

int BlockDev_GetUnitDescIndex(unsigned int index, BlockDev_UnitDesc *unitDesc, unsigned int size, int prop) {
    Partition *part;

    ASSERT_RET(unitDesc, E_BUG);

    part = parts;
    while (index-- && part)
        part = part->next;
    if (part)
        return DoGetUnitDesc(part, unitDesc, size, prop);
    else
        return E_NO_SUCH_UNIT;
}

int BlockDev_Read(BlockDev_IO *request) {
    Partition *part;
    int error;

    ASSERT_RET(request, E_BUG);

    part = parts;
    while (part) {
        if (part->id == request->unit) {
            unsigned int origBlock, origN;

            origBlock = request->block;
            origN = request->n;
            /* check whether the request is inside partition */
            if (request->block >= part->nBlocks)
                return E_END_REACHED;
            if ((request->block+request->n) > part->nBlocks)
                request->n = part->nBlocks-request->block;
            request->block += part->startBlock;
            error = part->imp->readF(request);
            request->n = origN;
            request->block = origBlock;
            return error;
        }
        part = part->next;
    }
    return E_NO_SUCH_UNIT;
}

int BlockDev_Write(BlockDev_IO *request) {
    Partition *part;
    int error;

    ASSERT_RET(request, E_BUG);

    part = parts;
    while (part) {
        if (part->id == request->unit) {
            unsigned int origBlock, origN;

            origBlock = request->block;
            origN = request->n;
            /* check whether the request is inside partition */
            if (request->block >= part->nBlocks)
                return E_END_REACHED;
            if ((request->block+request->n) > part->nBlocks)
                request->n = part->nBlocks-request->block;
            request->block += part->startBlock;
            error = part->imp->writeF(request);
            request->n = origN;
            request->block = origBlock;
            return error;
        }
        part = part->next;
    }
    return E_NO_SUCH_UNIT;
}

void LoadDiskPartition(void) {
    SysModManager_Interface *diskPartID;
    SysModManager_Imp *impID;
    SysModManager_ImpInfo impInfo;
    DiskPartition *diskPart;
    int (*readF)(SysModManager_Imp*,unsigned int,DiskPartition_Description*,unsigned int,unsigned int);
    unsigned int impIndex;
    char buf[128];
    int ret;
    
    diskPartID = SysModManager_GetInterID("DiskPartition");
    if (IS_PTR_ERROR(diskPartID)) {
        DEBUG_MSG(("partmanager: no DiskPartition interface: %d\n", PTR_TO_ERROR(diskPartID)));
        return;
    }
    impIndex = 0;
    while (!SysModManager_GetImpInfoIndex(diskPartID, impIndex++, &impInfo, buf, sizeof(buf))) {
        impID = SysModManager_GetImpID("DiskPartition", impInfo.name);
        if (!IS_PTR_ERROR(impID)) {
            readF = SysModManager_GetFunction(impID, DiskPartition__READ, 0, NULL);
            if (IS_PTR_ERROR(readF)) {
                DEBUG_MSG(("partmanager: could not get DiskPartition read function: %s:%d\n", impInfo.name, PTR_TO_ERROR(readF)));
                ret = SysModManager_PutImpID(impID);
                ASSERT_ERROR(ret);
                continue;
            }
            diskPart = MemManager_Alloc(sizeof(DiskPartition), 0);
            if (IS_PTR_ERROR(diskPart)) {
                DEBUG_MSG(("partmanager: could not alloc mem for DiskPartition: %d\n", PTR_TO_ERROR(diskPart)));
                ret = SysModManager_PutImpID(impID);
                ASSERT_ERROR(ret);
                continue;
            }
            diskPart->imp = impID;
            diskPart->readF = readF;
            diskPart->next = diskParts;
            diskParts = diskPart;
        }
    }
}

/* Check whether the disk has a partition table.
   Returns: 0 - no partition
            1 - at least one partition
 */
int ScanDisk(Implementation *imp, BlockDev_UnitDesc *unitDesc) {
    Partition *part;
    DiskPartition *diskPart;
    DiskPartition_Description *desc;
    unsigned int descSize, i;
    int error, ret;

    /* go through all disk partitions */
    error = 0;
    descSize = 1024;
    desc = MemManager_Alloc(descSize, 0);
    if (IS_PTR_ERROR(desc)) {
        DEBUG_MSG(("partmanager: could not alloc desc struct: %d\n", PTR_TO_ERROR(desc)));
        return error;
    }
    diskPart = diskParts;
    while (diskPart) {
        ret = diskPart->readF(imp->imp, unitDesc->unit, desc, descSize, sizeof(DiskPartition_Partition));
        if (ret == E_BUFFER_TOO_SMALL) {
            ret = MemManager_Free(desc);
            ASSERT_ERROR(ret);
            descSize += 1024;
            desc = MemManager_Alloc(descSize, 0);
            if (IS_PTR_ERROR(desc)) {
                DEBUG_MSG(("partmanager: could not alloc desc struct: %d\n", PTR_TO_ERROR(desc)));
                return error;
            }
            continue;
        }
        if (!ret) {
            DEBUG_MSG(("partmanager: unit has %u partitions\n", desc->nParts));
            for (i=0; i < desc->nParts; i++ ) {
                if (desc->parts[i].flags & DiskPartition_F_EXTENDED)
                    continue;
                part = MemManager_Alloc(sizeof(Partition), 0);
                if (IS_PTR_ERROR(part)) {
                    DEBUG_MSG(("partmanager: could not alloc part struct: %d\n", PTR_TO_ERROR(part)));
                    break;
                }
                DEBUG_MSG(("partmanager: adding partition - start: %u  blocks: %u\n", desc->parts[i].startBlock, desc->parts[i].nBlocks));
                part->unitDesc = unitDesc;
                part->imp = imp;
                part->startBlock = desc->parts[i].startBlock;
                part->nBlocks = desc->parts[i].nBlocks;
                part->id = nParts;
                nParts++;
                part->next = parts;
                parts = part;
                /* at least one partition added */
                error = 1;
            }
            break;
        }
        diskPart = diskPart->next;
    }
    ret = MemManager_Free(desc);
    ASSERT_ERROR(ret);
    return error;
}

/* Check whether we have disks in this implementation.
   Returns: 0 - no partition added
            1 - at least one partition added
 */
int ScanImp(SysModManager_Imp *impID) {
    int (*getUnitDescIF)(unsigned int,BlockDev_UnitDesc*,unsigned int,int);
    int (*readF)(BlockDev_IO*);
    int (*writeF)(BlockDev_IO*);
    BlockDev_UnitDesc *unitDesc;
    Implementation *imp;
    unsigned int ind;
    int hasPart, ret;

    getUnitDescIF = SysModManager_GetFunction(impID, BlockDev__GETUNITDESCINDEX, 0, NULL);
    readF = SysModManager_GetFunction(impID, BlockDev__READ, 0, NULL);
    writeF = SysModManager_GetFunction(impID, BlockDev__WRITE, 0, NULL);
    if (IS_PTR_ERROR(getUnitDescIF) || IS_PTR_ERROR(readF) || IS_PTR_ERROR(writeF)) {
        DEBUG_MSG(("partmanager: could not get blockdev functions: %d,%d,%d\n", PTR_TO_ERROR(getUnitDescIF),
                  PTR_TO_ERROR(readF), PTR_TO_ERROR(writeF)));
        return 0;
    }
    /* alloc here so if there is an error we do not need to clean up partitions */
    imp = MemManager_Alloc(sizeof(Implementation), 0);
    if (IS_PTR_ERROR(imp)) {
        DEBUG_MSG(("partmanager: could not alloc imp struct: %d\n", PTR_TO_ERROR(imp)));
        return 0;
    }
    hasPart = 0;
    imp->readF = readF;
    imp->writeF = writeF;
    imp->imp = impID;
    unitDesc = MemManager_Alloc(sizeof(BlockDev_UnitDesc)+32, 0);
    if (IS_PTR_ERROR(unitDesc)) {
        DEBUG_MSG(("partmanager: could not alloc unit desc struct: %d\n", PTR_TO_ERROR(unitDesc)));
        goto freeImp;
    }
    /* go through all units */
    ind = 0;
    while (!getUnitDescIF(ind, unitDesc, sizeof(BlockDev_UnitDesc)+32, 0)) {
        if (unitDesc->type == BlockDev_TYPE_DISK && ScanDisk(imp, unitDesc)) {
            /* alloc a new unitDesc buf */
            unitDesc = MemManager_Alloc(sizeof(BlockDev_UnitDesc)+32, 0);
            if (IS_PTR_ERROR(unitDesc)) {
                DEBUG_MSG(("partmanager: could not alloc unit desc struct: %d\n", PTR_TO_ERROR(unitDesc)));
                break;
            }
            hasPart = 1;
        }
        ind++;
    }
    /* last unit desc was not used */
    if (!IS_PTR_ERROR(unitDesc)) {
        ret = MemManager_Free(unitDesc);
        ASSERT_ERROR(ret);
    }
    if (hasPart) {
        /* we have at least one partition in this implementation. Save it. */
        imp->next = imps;
        imps = imp;
    } else {
freeImp:
        ret = MemManager_Free(imp);
        ASSERT_ERROR(ret);
    }
    return hasPart;
}

void ScanDisks(void) {
    SysModManager_Interface *blockDevID;
    SysModManager_Imp *impID;
    SysModManager_ImpInfo impInfo;
    unsigned int impIndex;
    char buf[128];

    blockDevID = SysModManager_GetInterID("BlockDev");
    if (IS_PTR_ERROR(blockDevID)) {
        DEBUG_MSG(("partmanager: no blockdev interface: %d\n", PTR_TO_ERROR(blockDevID)));
        return;
    }
    /* go through all implementations looking for disks */
    impIndex = 0;
    while (!SysModManager_GetImpInfoIndex(blockDevID, impIndex, &impInfo, buf, sizeof(buf))) {
        impID = SysModManager_GetImpID("BlockDev", impInfo.name);
        /* do not check ourself! */
        if (StringMan_Compare(impInfo.name, "PartManager")) {
            if (!IS_PTR_ERROR(impID) && !ScanImp(impID))
                SysModManager_PutImpID(impID);
        }
        impIndex++;
    }
}

void UnloadDiskPartition(void) {
    DiskPartition *diskPart, *tmpDP;
    int ret;
    
    diskPart = diskParts;
    diskParts = NULL;
    while(diskPart) {
        ret = SysModManager_PutImpID(diskPart->imp);
        ASSERT_ERROR(ret);
        tmpDP = diskPart->next;
        ret = MemManager_Free(diskPart);
        ASSERT_ERROR(ret);
        diskPart = tmpDP;
    }
}

int Phase1Start(SysModManager_Module *modID) {
    int ret;

    ID = modID;
    /* start all used modules */
    if ((ret = SysModManager_StartModule((unsigned int)MemManager_Alloc)) ||
        (ret = SysModManager_StartModule((unsigned int)StringMan_Copy))) {
        DEBUG_MSG_PRE(ret);
        return 1;
    }
    /* get all disk partitions */
    LoadDiskPartition();
    /* get partition table from all disks */
    ScanDisks();
    /* not needed anymore */
    UnloadDiskPartition();
    return 0;
}

int Shutdown(void) {
    return 0;
}
