/*
*****************************************
* This file implements the SysModManager interface
*
* Version: 0.3.3
* Author: Luiz Henrique Shigunov
* E-mail: luizshigunov@users.sourceforge.net
* WEB: http://www.sourceforge.net/projects/modulos
* Last change: 24/06/2005
*
* See the interface specification to know what each function does.
*
* Copyright (C) 2000-2005, Luiz Henrique Shigunov
*
* This module is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2.1 of the License, or (at your option) any later version.
*
* This module is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this module; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*****************************************
*/

#include <stdarg.h>
#include <stddef.h>
#include <avltree_in.h>
#include <cfgmanager.h>
#include <debug.h>
#include <errormanager.h>
#include <intmanager_i386.h>
#include <memman.h>
#include <memmanager.h>
#include <memmanager_i386.h>
#include <rsysmodfile.h>
#include <stringman.h>
#include <sysmodmanager.h>
#include <taskmanager.h>
#include <textvideo.h>

#define ALIGN_DATA_UP(data, align) ((((unsigned int)(data)) + (align)-1) & ~((align)-1))
#define ALIGN_DATA_DOWN(data, align) (((unsigned int)(data)) & ~((align)-1))

/* modules properties */
#define SysModManager_MOD_PHASE0_STARTED 0x10000
#define SysModManager_MOD_PHASE1_STARTED 0x20000

#define SysModManager_REGISTER_PROP_MASK 0x02

struct _SysModManager_Module {
    SysModManager_Module *next;
    int prop;
    /*
    0x000001 - permanent module
    0x000002 - automatic start
    0x000004 - load at boot
    0x010000 - module phase 0 started
    0x020000 - module phase 1 started
    */
    unsigned int memStart;  /* = 0xffffffff module not loaded */
    unsigned int memSize;
    unsigned int refCounter;
    struct _SysModManager_ModData *data;
    TaskManager_RWSync dataSync;
    char *path;
    int (*phase0Start)(SysModManager_Module*);
    int (*phase1Start)(SysModManager_Module*);
    void (*shutdown)(void);
    char md5[16];
};

typedef struct _SysModManager_ModData {
    struct _SysModManager_ModData *next;
    SysModManager_Module *module;
    void *info;
} SysModManager_ModData;

struct _SysModManager_Interface {
    AVLTree_In_Node node;
    char *name;
    unsigned int nFunctions;    /* # functions */
    struct _SysModManager_Implementation *imps; /* implementation list */
};

typedef struct _SysModManager_Implementation {
    struct _SysModManager_Implementation *next;
    struct _SysModManager_Module *module;
    SysModManager_Interface *interface;
    char *name;
    unsigned short interN;
    unsigned short impN;
    struct _SysModManager_Function *functions;
} SysModManager_Implementation;

typedef struct _SysModManager_Function {
    char *code;
    unsigned char prop;
    /*
    0x01 - system function
    0x02 - not implemented
    */
    unsigned char words;
    short reserved;
} SysModManager_Function;

typedef struct {
    /* first module of the list */
    SysModManager_Module *modules;
    /* avl tree root */
    SysModManager_Interface *interfaces;
    char *pgDir;
    /* start of used memory */
    char *start;
    char *cfgFile;
    unsigned int cfgFileSize;
} SysModManager_StarterData;

typedef struct {
    unsigned int index;
    SysModManager_Interface *interface;
} IndexInfo;

int SysModManager_CmpInterface(const void*,const AVLTree_In_Node*);
int SysModManager_CmpAddInterface(const AVLTree_In_Node*,const AVLTree_In_Node*);

SysModManager_Module *ID;
AVLTree_In_Tree interfaces = {NULL, SysModManager_CmpInterface, SysModManager_CmpAddInterface, };
SysModManager_Module *modules;
TaskManager_Semaphore modulesSync = {1, NULL};
TaskManager_Semaphore loadMutex = {1, NULL};
char msgBuffer[1024];
int phase;

int SysModManager_LoadModule(SysModManager_Module*);
int SysModManager_StartModuleUtil(SysModManager_Module*);

void SysModManager_Broadcast(const char *interface, unsigned int function, void *param1, void *param2) {
    SysModManager_Interface *found;
    SysModManager_Implementation *imp;

    ASSERT_NO_RET(interface);

    TaskManager_P(&modulesSync, 0, 0);
    found = (SysModManager_Interface*)AVLTree_In_Find(&interfaces, interface);
    if (IS_PTR_ERROR(found) || function >= found->nFunctions)
        goto unlockSync;
    imp = found->imps;
    do {
        void (*funcCall)(void*,void*);

        if (imp->module->prop & SysModManager_MOD_PHASE1_STARTED) {
            /* don't free this module */
            imp->module->refCounter++;
            TaskManager_V(&modulesSync, 0);
            /* can only call system functions */
            if (!(imp->functions[function].prop & SysModManager_NOT_IMPLEMENTED) &&
                 (imp->functions[function].prop & SysModManager_SYS_FUNCTION)) {
                funcCall = (void(*)(void*,void*))imp->functions[function].code;
                funcCall(param1, param2);
            }
            TaskManager_P(&modulesSync, 0, 0);
            imp->module->refCounter--;
        }
        imp = imp->next;
    } while (imp);
unlockSync:
    TaskManager_V(&modulesSync, 0);
    return;
}

int SysModManager_CmpAddInterface(const AVLTree_In_Node *addNode, const AVLTree_In_Node *node) {
    return StringMan_Compare(((SysModManager_Interface*)addNode)->name, ((SysModManager_Interface*)node)->name);
}

int SysModManager_CmpInterface(const void *info, const AVLTree_In_Node *node) {
    return StringMan_Compare((char*)info, ((SysModManager_Interface*)node)->name);
}

void SysModManager_DebugMsg(const char *format, ...) {
    va_list args;
    int written;

    va_start(args, format);
    written = StringMan_NPrintFArgs(msgBuffer, sizeof(msgBuffer), format, args);
    if (written > sizeof(msgBuffer))
        written = sizeof(msgBuffer);
    TextVideo_Write(msgBuffer, written);
    va_end(args);
    return;
}

void *SysModManager_DelModInfo(SysModManager_Module *module, SysModManager_Module *which) {
    SysModManager_ModData *data, **prev;
    void *info;

    ASSERT_RET(module, NULL);
    ASSERT_RET(which, NULL);

    TaskManager_LockWrite(&module->dataSync, 0);
    data = module->data;
    prev = &module->data;
    info = NULL;
    while (data) {
        if (data->module == which) {
            int ret;
            info = data->info;
            *prev = data->next;
            ret = MemManager_Free(data);
            ASSERT_ERROR(ret);
            break;
        }
        prev = &data->next;
        data = data->next;
    }
    TaskManager_UnlockRW(&module->dataSync, 0);
    return info;
}

void *SysModManager_GetFunction(SysModManager_Implementation *imp, unsigned int function, int prop, int *words) {
    SysModManager_Function *functionF;
    int error;

    ASSERT_RET(imp, ERROR_TO_PTR(E_BUG));

    if (function >= imp->interface->nFunctions)
        return ERROR_TO_PTR(E_BAD_INDEX);
    functionF = &imp->functions[function];
    /* want sys function but is user or vice versa */
    if ((functionF->prop & SysModManager_SYS_FUNCTION) == prop)
        return ERROR_TO_PTR(E_BAD_FUNCTION_TYPE);
    /* not implemented */
    if (functionF->prop & SysModManager_NOT_IMPLEMENTED)
        return ERROR_TO_PTR(E_NOT_IMPLEMENTED);
    error = SysModManager_StartModuleUtil(imp->module);
    if (error)
        return ERROR_TO_PTR(error);
    if (words && !functionF->prop)
        *words = functionF->words;
    return functionF->code;
}

void CheckImp(AVLTree_In_Node *node, void *param1, void *param2) {
    SysModManager_Implementation *currImp;

    /* already found? */
    if (*(SysModManager_Implementation**)param2)
        return;
    currImp = ((SysModManager_Interface*)node)->imps;
    while (currImp) {
        if (currImp == (SysModManager_Implementation*)param1) {
            *(SysModManager_Implementation**)param2 = currImp;
            break;
        }
        currImp = currImp->next;
    }
}

int SysModManager_GetFunctionInfoIndex(SysModManager_Imp *imp, unsigned int index, SysModManager_FunctionInfo *info, void *data, unsigned int size) {
    SysModManager_Imp *impFound;
    int error;

    ASSERT_RET(imp, E_BUG);
    ASSERT_RET(info, E_BUG);
    ASSERT_RET(data, E_BUG);

    /* check whether imp is valid */
    TaskManager_P(&modulesSync, 0, 0);
    impFound = NULL;
    AVLTree_In_Walk(&interfaces, CheckImp, imp, (void*)&impFound);
    if (imp != impFound) {
        error = E_BUG;
        goto errorV;
    }
    if (index >= imp->interface->nFunctions) {
        error = E_BAD_INDEX;
        goto errorV;
    }
    info->code = (unsigned int)imp->functions[index].code;
    info->prop = (unsigned int)imp->functions[index].prop;
    info->words = (unsigned int)imp->functions[index].words;
    error = E_OK;
errorV:
    TaskManager_V(&modulesSync, 0);
    return error;
}

SysModManager_Implementation *SysModManager_GetImpID(const char *interface, const char *imp) {
    SysModManager_Implementation *impID;
    SysModManager_Interface *interF;

    ASSERT_RET(interface, ERROR_TO_PTR(E_BUG));

    TaskManager_P(&modulesSync, 0, 0);
    impID = ERROR_TO_PTR(E_NO_SUCH_INTERFACE);
    interF = (SysModManager_Interface*)AVLTree_In_Find(&interfaces, interface);
    if (IS_PTR_ERROR(interF))
        goto unlockSync;
    impID = interF->imps;
    if (imp) {
        do {
            if (!StringMan_Compare(impID->name, imp))
                break;
            impID = impID->next;
        } while (impID);
    }
    if (impID) {
        /* don't let someone remove this module */
        impID->module->refCounter++;
    } else
        impID = ERROR_TO_PTR(E_NO_SUCH_IMPLEMENTATION);
unlockSync:
    TaskManager_V(&modulesSync, 0);
    return impID;
}

int SysModManager_GetImpInfo(SysModManager_Implementation *imp, SysModManager_ImpInfo *info, void *data, unsigned int size) {
    unsigned int nameS;

    ASSERT_RET(imp, E_BAD_VALUE);
    ASSERT_RET(info, E_BAD_VALUE);
    ASSERT_RET(data, E_BAD_VALUE);

    info->imp = imp;
    info->module = imp->module;
    info->interface = imp->interface;
    info->name = data;
    nameS = StringMan_Length(imp->name)+1;
    if (size < nameS)
        return E_TOO_BIG;
    MemMan_Copy(data, imp->name, nameS);
    return E_OK;
}

void GetInterfaceID(AVLTree_In_Node *node, void *param1, void *param2) {
    IndexInfo *indexInfo;

    indexInfo = (IndexInfo *)param1;
    if (!indexInfo->index) {
        if (indexInfo->interface == (SysModManager_Interface *)node) {
            indexInfo->index = 1;
            return;
        }
    }
}

int SysModManager_GetImpInfoIndex(SysModManager_Interface *interface, unsigned int index, SysModManager_ImpInfo *info, void *data, unsigned int size) {
    IndexInfo indexInfo;
    SysModManager_Implementation *imp;
    unsigned int nameS;
    int error;

    ASSERT_RET(interface, E_BAD_VALUE);
    ASSERT_RET(info, E_BAD_VALUE);
    ASSERT_RET(data, E_BAD_VALUE);

    indexInfo.index = 0;
    indexInfo.interface = interface;
    TaskManager_P(&modulesSync, 0, 0);
    AVLTree_In_Walk(&interfaces, GetInterfaceID, &indexInfo, NULL);
    error = E_NOT_FOUND;
    if (!indexInfo.index)
        goto errorV;
    imp = interface->imps;
    while (index && imp) {
        index--;
        imp = imp->next;
    }
    error = E_BAD_INDEX;
    if (!imp)
        goto errorV;
    info->imp = imp;
    info->module = imp->module;
    info->interface = imp->interface;
    info->name = data;
    nameS = StringMan_Length(imp->name)+1;
    error = E_TOO_BIG;
    if (size < nameS)
        goto errorV;
    MemMan_Copy(data, imp->name, nameS);
    error = E_OK;

errorV:
    TaskManager_V(&modulesSync, 0);
    return error;
}

SysModManager_Interface *SysModManager_GetInterID(const char *interface) {
    SysModManager_Interface *inter;

    ASSERT_RET(interface, ERROR_TO_PTR(E_BUG));

    TaskManager_P(&modulesSync, 0, 0);
    inter = (SysModManager_Interface*)AVLTree_In_Find(&interfaces, interface);
    if (IS_PTR_ERROR(inter))
        inter = ERROR_TO_PTR(E_NO_SUCH_INTERFACE);
    TaskManager_V(&modulesSync, 0);
    return inter;
}

int SysModManager_GetInterInfo(SysModManager_Interface *interface, SysModManager_InterInfo *info, void *data, unsigned int size) {
    IndexInfo indexInfo;
    unsigned int nameS;
    int error;

    ASSERT_RET(interface, E_BAD_VALUE);
    ASSERT_RET(info, E_BAD_VALUE);
    ASSERT_RET(data, E_BAD_VALUE);

    indexInfo.index = 0;
    indexInfo.interface = interface;
    TaskManager_P(&modulesSync, 0, 0);
    AVLTree_In_Walk(&interfaces, GetInterfaceID, &indexInfo, NULL);
    error = E_NOT_FOUND;
    if (!indexInfo.index)
        goto errorV;
    info->interface = interface;
    info->nFunctions = interface->nFunctions;
    info->name = data;
    nameS = StringMan_Length(interface->name)+1;
    error = E_TOO_BIG;
    if (size < nameS)
        goto errorV;
    MemMan_Copy(data, interface->name, nameS);
    error = E_OK;

errorV:
    TaskManager_V(&modulesSync, 0);
    return error;
}

void GetInterfaceIndex(AVLTree_In_Node *node, void *param1, void *param2) {
    IndexInfo *indexInfo;

    indexInfo = (IndexInfo *)param1;
    if (!indexInfo->interface) {
        if (!indexInfo->index) {
            indexInfo->interface = (SysModManager_Interface *)node;
            return;
        }
        indexInfo->index--;
    }
}

int SysModManager_GetInterInfoIndex(unsigned int index, SysModManager_InterInfo *info, void *data, unsigned int size) {
    IndexInfo indexInfo;
    unsigned int nameS;
    int error;

    ASSERT_RET(info, E_BAD_VALUE);
    ASSERT_RET(data, E_BAD_VALUE);

    indexInfo.index = index;
    indexInfo.interface = NULL;
    TaskManager_P(&modulesSync, 0, 0);
    AVLTree_In_Walk(&interfaces, GetInterfaceIndex, &indexInfo, NULL);
    error = E_BAD_INDEX;
    if (!indexInfo.interface)
        goto errorV;
    info->interface = indexInfo.interface;
    info->nFunctions = indexInfo.interface->nFunctions;
    info->name = data;
    nameS = StringMan_Length(indexInfo.interface->name)+1;
    error = E_TOO_BIG;
    if (size < nameS)
        goto errorV;
    MemMan_Copy(data, indexInfo.interface->name, nameS);
    error = E_OK;

errorV:
    TaskManager_V(&modulesSync, 0);
    return error;
}

void *SysModManager_GetModInfo(SysModManager_Module *module, SysModManager_Module *which) {
    SysModManager_ModData *data;
    void *info;

    ASSERT_RET(module, NULL);
    ASSERT_RET(which, NULL);

    TaskManager_LockRead(&module->dataSync, 0);
    data = module->data;
    info = NULL;
    while (data) {
        if(data->module == which) {
            info = data->info;
            break;
        }
        data = data->next;
    }
    TaskManager_UnlockRW(&module->dataSync, 0);
    return info;
}

int SysModManager_GetModuleInfo(SysModManager_Module *module, SysModManager_ModuleInfo *info, void *data, unsigned int size) {
    SysModManager_Module *currModule;
    unsigned int pathS;
    int error;

    ASSERT_RET(module, E_BAD_VALUE);
    ASSERT_RET(info, E_BAD_VALUE);
    ASSERT_RET(data, E_BAD_VALUE);
    
    /* valid module? */
    TaskManager_P(&modulesSync, 0, 0);
    currModule = modules;
    while (currModule && currModule != module) {
        currModule = currModule->next;
    }
    error = E_NOT_FOUND;
    if (!currModule)
        goto errorV;
    info->module = module;
    info->prop = module->prop;
    info->memStart = module->memStart == 0xffffffff ? 0 : module->memStart;
    info->memSize = module->memSize;
    info->refCounter = module->refCounter;
    info->path = data;
    info->phase0Function = (unsigned int)module->phase0Start == 0xffffffff ? 0 : (unsigned int)module->phase0Start;
    info->phase1Function = (unsigned int)module->phase1Start == 0xffffffff ? 0 : (unsigned int)module->phase1Start;
    info->shutdownFunction = (unsigned int)module->shutdown == 0xffffffff ? 0 : (unsigned int)module->shutdown;
    MemMan_Copy(info->md5, module->md5, sizeof(module->md5));
    pathS = StringMan_Length(module->path)+1;
    error = E_TOO_BIG;
    if (size < pathS)
        goto errorV;
    MemMan_Copy(data, module->path, pathS);
    error = E_OK;

errorV:
    TaskManager_V(&modulesSync, 0);
    return error;
}

int SysModManager_GetModuleInfoIndex(unsigned int index, SysModManager_ModuleInfo *info, void *data, unsigned int size) {
    SysModManager_Module *module;
    unsigned int pathS;
    int error;

    ASSERT_RET(info, E_BAD_VALUE);
    ASSERT_RET(data, E_BAD_VALUE);
    
    TaskManager_P(&modulesSync, 0, 0);
    module = modules;
    while (index && module) {
        index--;
        module = module->next;
    }
    error = E_BAD_INDEX;
    if (!module)
        goto errorV;
    info->module = module;
    info->prop = module->prop;
    info->memStart = module->memStart == 0xffffffff ? 0 : module->memStart;
    info->memSize = module->memSize;
    info->refCounter = module->refCounter;
    info->path = data;
    info->phase0Function = (unsigned int)module->phase0Start == 0xffffffff ? 0 : (unsigned int)module->phase0Start;
    info->phase1Function = (unsigned int)module->phase1Start == 0xffffffff ? 0 : (unsigned int)module->phase1Start;
    info->shutdownFunction = (unsigned int)module->shutdown == 0xffffffff ? 0 : (unsigned int)module->shutdown;
    MemMan_Copy(info->md5, module->md5, sizeof(module->md5));
    pathS = StringMan_Length(module->path)+1;
    error = E_TOO_BIG;
    if (size < pathS)
        goto errorV;
    MemMan_Copy(data, module->path, pathS);
    error = E_OK;

errorV:
    TaskManager_V(&modulesSync, 0);
    return error;
}

int SysModManager_PutImpID(SysModManager_Implementation *imp) {
    ASSERT_RET(imp, E_BAD_VALUE);

    TaskManager_AtomicDec(&imp->module->refCounter);
    return E_OK;
}

/* caller must lock loadMutex and guarantee that module wont be freed */
int SysModManager_LoadModule(SysModManager_Module *module) {
    SysModManager_Interface *currInt;
    RSysModFile_File *file;
    RSysModFile_UsedFuncRelocInfo usedFuncRInfo;
    RSysModFile_InterInfo intInfo;
    unsigned int nUsedFunc, nInter;
    int error, ret;

    ASSERT_RET(module, E_BAD_VALUE);

    if (module->memStart == 0xffffffff) {
        RSysModFile_ModuleInfo modInfo;
        unsigned int dataStart;
        error = RSysModFile_Open(module->path, &file);
        if (error)
            goto errorEnd;
        module->memStart = (unsigned int)MemManager_i386_AllocPages(module->memSize >> 12, MemManager_i386_RW_PAGE+MemManager_i386_ALLOC_PHYS);
        if (IS_PTR_ERROR(module->memStart)) {
            error = PTR_TO_ERROR(module->memStart);
            goto errorAllocMem;
        }
        if ((unsigned int)module->phase0Start != 0xffffffff)
            module->phase0Start = (int (*)(SysModManager_Module*))((unsigned int)module->phase0Start + module->memStart);
        if ((unsigned int)module->phase1Start != 0xffffffff)
            module->phase1Start = (int (*)(SysModManager_Module*))((unsigned int)module->phase1Start + module->memStart);
        if ((unsigned int)module->shutdown != 0xffffffff)
            module->shutdown = (void (*)(void))((unsigned int)module->shutdown + module->memStart);
        error = RSysModFile_GetModuleInfo(file, &modInfo);
        if (error)
            goto errorGetMod;
        dataStart = module->memStart + modInfo.codeSize;
        dataStart += (4 - (dataStart & 0x3)) & 0x3;
        error = RSysModFile_LoadCode(file, 0, modInfo.codeSize, (void*)module->memStart, (void*)module->memStart, (void*)dataStart);
        if (error)
            goto errorGetMod;
        if (modInfo.dataSize) {
            error = RSysModFile_LoadData(file, 0, modInfo.dataSize, (void*)dataStart, (void*)module->memStart, (void*)dataStart);
            if (error)
                goto errorGetMod;
        }
        nUsedFunc = 0;
        while (1) {
            SysModManager_Implementation *currImp;
            SysModManager_Function *currFunc;
            error = RSysModFile_GetUsedFuncReloc(file, nUsedFunc, &usedFuncRInfo);
            if (error) {
                if (error != E_BAD_INDEX)
                    goto errorUsedFunc;
                break;
            }
            error = E_NOT_FOUND;
            currImp = SysModManager_GetImpID(usedFuncRInfo.interface, usedFuncRInfo.imp);
            if (IS_PTR_ERROR(currImp))
                goto errorUsedFunc;
            error = E_BAD_INDEX;
            if (usedFuncRInfo.function >= currImp->interface->nFunctions)
                goto errorPutImp;
            currFunc = &currImp->functions[usedFuncRInfo.function];
            error = E_BAD_FUNCTION_TYPE;
            if (!(currFunc->prop & SysModManager_SYS_FUNCTION))
                goto errorPutImp;
            error = E_NOT_IMPLEMENTED;
            if (currFunc->prop & SysModManager_NOT_IMPLEMENTED)
                goto errorPutImp;
            TaskManager_AtomicInc(&currImp->module->refCounter);
            nUsedFunc++;
            error = SysModManager_LoadModule(currImp->module);
            if (error) {
errorPutImp:
                ret = SysModManager_PutImpID(currImp);
                ASSERT_ERROR(ret);
                goto errorUsedFunc;
            }
            /* relative or absolute relocation? */
            if (usedFuncRInfo.prop)
                *(int*)(module->memStart+usedFuncRInfo.offset) = (int)currImp->functions[usedFuncRInfo.function].code;
            else
                *(int*)(module->memStart+usedFuncRInfo.offset) = (unsigned int)currImp->functions[usedFuncRInfo.function].code - (module->memStart+usedFuncRInfo.offset+4);
            ret = SysModManager_PutImpID(currImp);
            ASSERT_ERROR(ret);
        }
        /* fixup all implementations functions */
        nInter = 0;
        do {
            error = RSysModFile_GetInterInfo(file, nInter, &intInfo);
            if (error) {
                if (error != E_BAD_INDEX)
                    goto errorUsedFunc;
                break;
            }
            TaskManager_P(&modulesSync, 0, 0);
            currInt = (SysModManager_Interface*)AVLTree_In_Find(&interfaces, intInfo.name);
            if (!IS_PTR_ERROR(currInt)) {
                SysModManager_Implementation *currImp;

                /* go through all implementations */
                currImp = currInt->imps;
                while (currImp) {
                    if (currImp->module == module) {
                        unsigned int i;
                        for (i=0; i < currInt->nFunctions; i++)
                            currImp->functions[i].code += module->memStart;
                    }
                    currImp = currImp->next;
                }
            }
            TaskManager_V(&modulesSync, 0);
            nInter++;
        } while (1);
        ret = RSysModFile_Close(file);
        ASSERT_ERROR(ret);
    }
    return E_OK;

errorUsedFunc:
    while (nUsedFunc--) {
        SysModManager_Implementation *currImp;
        if (RSysModFile_GetUsedFuncReloc(file, nUsedFunc, &usedFuncRInfo))
            break;
        currImp = SysModManager_GetImpID(usedFuncRInfo.interface, usedFuncRInfo.imp);
        if (IS_PTR_ERROR(currImp))
            break;
        TaskManager_AtomicDec(&currImp->module->refCounter);
        ret = SysModManager_PutImpID(currImp);
        ASSERT_ERROR(ret);
    }
errorGetMod:
    ret = MemManager_i386_FreePages((void*)module->memStart, module->memSize >> 12, 0);
    ASSERT_ERROR(ret);
errorAllocMem:
    ret = RSysModFile_Close(file);
    ASSERT_ERROR(ret);
errorEnd:
    return error;
}

int Phase0Start(SysModManager_StarterData *sData) {
    SysModManager_Module *currM;
    CfgManager_Group *modulesGroup;
    unsigned int index, *pdes, *ptes, physPg, i;
    int ret;

    /* start modules and interfaces register */
    modules = sData->modules;
    interfaces.root = (AVLTree_In_Node*)sData->interfaces;

    /* get our module ID */
    currM = modules;
    while (currM && ((unsigned int)Phase0Start < currM->memStart || (unsigned int)Phase0Start >= (currM->memStart+currM->memSize)))
        currM = currM->next;
    ID = currM;
    /* this module is already started */
    currM->prop |= SysModManager_MOD_PHASE0_STARTED;

    /* setup paging */
    pdes = (unsigned int*)sData->pgDir;
    /* last PDE always has the PGD */
    physPg = (unsigned int)pdes - 0xe0000000;
    pdes[1023] = physPg | 0x3;
    ptes = pdes + 1024;
    /* free LOGICAL pages of PGD and first page table */
    ptes[physPg>>12] = 0;
    ptes[(physPg>>12)+1] = 0;
    /* free LOGICAL page of PGD */
    pdes[0] = 0;
    /* free starter pages */
    for (i = 0x100; i < ((unsigned int)sData->start-0xe0000000)>>12; i++) {
        ptes[i] = 0;
    }
    __asm__ __volatile__ (
        "movl %%cr3, %%ecx\n"
        "movl %%ecx, %%cr3"
        : : : "%ecx");


    /* start phase 0 modules */
    /* start TextVideo, so DebugMsg will work */
    /* Start IntManager to catch exception ASAP */
    if (SysModManager_StartModule((unsigned int)TextVideo_Write) ||
        SysModManager_StartModule((unsigned int)IntManager_i386_CatchException) ||
        SysModManager_StartModule((unsigned int)MemManager_Alloc) ||
        SysModManager_StartModule((unsigned int)CfgManager_SetSystemFile) ||
        SysModManager_StartModule((unsigned int)TaskManager_P) ||
        SysModManager_StartModule((unsigned int)StringMan_Compare)) {
        SysModManager_DebugMsg("SysModManager: Cannot start phase 0 modules!");
        while (1)
            __asm__ __volatile__ ("hlt");
    }

    /* set MemManager_i386 BIOS mem info */
/*    MemManager_i386_SetBIOSMemInfo(sData->memInfo);*/

    /* set CfgManager system config file */
    if (CfgManager_SetSystemFile(sData->cfgFile, sData->cfgFileSize)) {
        SysModManager_DebugMsg("SysModManager: Error parsing system configuration file!");
        while (1)
            __asm__ __volatile__ ("hlt");
    }

    /* now dinamiclly alloc memory for modules and interfaces register */
/*    SysModManager_StartModulesReg();
    SysModManager_StartInterfacesReg();*/

    /* and free staticlly alloc pages */
/*    SysModManager_FreeStaticMemory();*/

    /* that's it! Let's start to register permanent modules */
    phase = 1;
    /* start phase 1 modules */
    if (SysModManager_StartModule((unsigned int)IntManager_i386_CatchException) ||
        SysModManager_StartModule((unsigned int)TextVideo_Write) ||
        SysModManager_StartModule((unsigned int)MemManager_Alloc) ||
        SysModManager_StartModule((unsigned int)CfgManager_SetSystemFile) ||
        SysModManager_StartModule((unsigned int)TaskManager_P) ||
        SysModManager_StartModule((unsigned int)RSysModFile_Open) ||
        SysModManager_StartModule((unsigned int)StringMan_Compare) ||
        SysModManager_StartModule((unsigned int)MemMan_Copy) ||
        SysModManager_StartModule((unsigned int)AVLTree_In_Find)) {
        SysModManager_DebugMsg("SysModManager: Cannot start phase 1 modules!");
        while (1)
            __asm__ __volatile__ ("hlt");
    }

    DEBUG_MSG(("Registering permanent modules...\n"));

    /* open modules base key */
    modulesGroup = CfgManager_OpenGroup(NULL, "/system/SysModManager/modules");
    if (!modulesGroup) {
        SysModManager_DebugMsg("SysModManager: Cannot open base register key!");
        while (1)
            __asm__ __volatile__ ("hlt");
    }

    index = 0;
    while (1) {
        char buf[128];
        unsigned int bufSize;
        int prop;
        CfgManager_Group *modGroup;

        modGroup = CfgManager_OpenGroupIndex(modulesGroup, index);
        if (!modGroup) {
            break;
        }
        /* get module path */
        bufSize = 128;
        if ((ret = CfgManager_GetValue(modGroup, "path", CfgManager_STRING, buf, &bufSize))) {
            SysModManager_DebugMsg("Cannot read path: %x\n", ret);
            goto nextModule;
        }
        /* get module prop */
        bufSize = sizeof(prop);
        if ((ret = CfgManager_GetValue(modGroup, "prop", CfgManager_HEXNUMBER, &prop, &bufSize))) {
            SysModManager_DebugMsg("Cannot read prop: %x\n", ret);
            goto nextModule;
        }
        /* it's a permanent module */
        prop |= SysModManager_PERMANENT;
        ret = SysModManager_Register(buf, prop);
        if (ret && ret != E_REGISTERED) {
            SysModManager_DebugMsg("Error. Can't register module '");
            SysModManager_DebugMsg(buf);
            SysModManager_DebugMsg("' - %x.\n", ret);
        }
nextModule:
        ret = CfgManager_CloseGroup(modGroup);
        ASSERT_ERROR(ret);
        index++;
    }
    ret = CfgManager_CloseGroup(modulesGroup);
    ASSERT_ERROR(ret);

    DEBUG_MSG(("Done.\n"));

    /* start automatic start modules */
    DEBUG_MSG(("Starting automatic start modules...\n"));
    TaskManager_P(&modulesSync, 0, 0);
    currM = modules;
    do {
        if ((currM->prop & SysModManager_AUTOMATIC) && !(currM->prop & SysModManager_MOD_PHASE1_STARTED)) {
            currM->refCounter++;
            TaskManager_V(&modulesSync, 0);
            DEBUG_MSG(("Starting module: "));
            DEBUG_MSG((currM->path));
            DEBUG_MSG(("\n"));
            if(SysModManager_StartModuleUtil(currM)) {
                SysModManager_DebugMsg("Error. Can't start automatic start module: ");
                SysModManager_DebugMsg(currM->path);
                SysModManager_DebugMsg("\n");
            }
            TaskManager_P(&modulesSync, 0, 0);
        }
        currM = currM->next;
    } while (currM);
    TaskManager_V(&modulesSync, 0);
    DEBUG_MSG(("Done.\n"));

    /* that's it! */
    DEBUG_MSG(("System started!\n"));

    /* destroy this thread */
    if (TaskManager_DestroyThread(NULL)) {
        SysModManager_DebugMsg("SysModManager: Cannot destroy start thread!");
        while (1)
            __asm__ __volatile__ ("hlt");
    }
    return 0;
}

int SysModManager_Register(const char *path, int prop) {
    SysModManager_Module *currM;
    SysModManager_Module *module;
    SysModManager_Interface *currInt;
    SysModManager_Implementation *currImp;
    RSysModFile_File *file;
    RSysModFile_InterInfo intInfo;
    RSysModFile_ImpInfo impInfo;
    RSysModFile_ModuleInfo modInfo;
    int error, ret;
    unsigned int nInter, nImp;

    /* FIXME: Major rewrite needed! */
    error = E_REGISTERED;
    TaskManager_P(&modulesSync, 0, 0);
    currM = modules;
    while (currM && StringMan_Compare(currM->path, path))
        currM = currM->next;
    TaskManager_V(&modulesSync, 0);
    if (currM) {
        goto errorEnd;
    }
    /* lets start to register it. */
    error = RSysModFile_Open(path, &file);
    if (error) {
        goto errorEnd;
    }
    error = RSysModFile_GetModuleInfo(file, &modInfo);
    if (error) {
        goto errorCloseFile;
    }
    module = MemManager_Alloc(sizeof(SysModManager_Module), MemManager_BLOCK_MEM);
    if (IS_PTR_ERROR(module)) {
        error = PTR_TO_ERROR(module);
        goto errorCloseFile;
    }
    MemMan_Set(module, 0, sizeof(SysModManager_Module));
    module->path = MemManager_Alloc(StringMan_Length(path)+1, MemManager_BLOCK_MEM);
    if (IS_PTR_ERROR(module->path)) {
        error = PTR_TO_ERROR(module->path);
        goto errorFreeModule;
    }
    StringMan_Copy(module->path, path);
    MemMan_Copy(module->md5, modInfo.md5, 16);
    module->memStart = 0xffffffff;
    module->memSize = modInfo.codeSize;
    /* align in 4 bytes */
    module->memSize = ALIGN_DATA_UP(module->memSize, 4);
    module->memSize += modInfo.dataSize;
    /* align in 4096 bytes */
    module->memSize = ALIGN_DATA_UP(module->memSize, 0x1000);
    module->prop = prop & SysModManager_REGISTER_PROP_MASK;
    /* if module doesn't has a start function mark it as started */
    if ((unsigned int)modInfo.phase0Start == 0xffffffff) {
        module->prop |= SysModManager_MOD_PHASE0_STARTED;
    }
    if ((unsigned int)modInfo.phase1Start == 0xffffffff) {
        module->prop |= SysModManager_MOD_PHASE1_STARTED;
    }
    module->phase0Start = modInfo.phase0Start;
    module->phase1Start = modInfo.phase1Start;
    module->shutdown = modInfo.shutdown;
    TaskManager_P(&modulesSync, 0, 0);
    module->next = modules;
    modules = module;
    TaskManager_V(&modulesSync, 0);
    nInter = 0;
    do {
        error = RSysModFile_GetInterInfo(file, nInter, &intInfo);
        if (error) {
            if (error != E_BAD_INDEX)
                goto errorFreeModule;
            break;
        }
        TaskManager_P(&modulesSync, 0, 0);
        currInt = (SysModManager_Interface*)AVLTree_In_Find(&interfaces, intInfo.name);
        TaskManager_V(&modulesSync, 0);
        /* if not found, add it */
        if (IS_PTR_ERROR(currInt)) {
            SysModManager_Interface *interface;
            interface = MemManager_Alloc(sizeof(SysModManager_Interface), MemManager_BLOCK_MEM);
            if (IS_PTR_ERROR(interface)) {
                error = PTR_TO_ERROR(interface);
                goto errorFreeModule;
            }
            interface->name = MemManager_Alloc(StringMan_Length(intInfo.name)+1, MemManager_BLOCK_MEM);
            if (IS_PTR_ERROR(interface->name)) {
                error = PTR_TO_ERROR(interface->name);
                ret = MemManager_Free(interface);
                ASSERT_ERROR(ret);
                goto errorFreeModule;
            }
            StringMan_Copy(interface->name, intInfo.name);
            interface->nFunctions = intInfo.nFunctions;
            interface->imps = NULL;
            TaskManager_P(&modulesSync, 0, 0);
            currInt = (SysModManager_Interface*)AVLTree_In_FindAdd(&interfaces, &interface->node);
            TaskManager_V(&modulesSync, 0);
            ASSERT_ERROR_PTR(currInt);
        }
        nImp = 0;
        do {
            SysModManager_Implementation *imp;
            error = RSysModFile_GetImpInfo(file, nInter, nImp, &impInfo);
            if (error) {
                if (error != E_BAD_INDEX)
                    goto errorFreeModule;
                break;
            }
            /* check if we already have this implementation */
            TaskManager_P(&modulesSync, 0, 0);
            error = E_IN_USE;
            currImp = currInt->imps;
            while (currImp) {
                if (!StringMan_Compare(currImp->name, impInfo.name)) {
                    TaskManager_V(&modulesSync, 0);
                    goto errorFreeModule;
                }
                currImp = currImp->next;
            }
            TaskManager_V(&modulesSync, 0);
            imp = MemManager_Alloc(sizeof(SysModManager_Implementation), MemManager_BLOCK_MEM);
            if (IS_PTR_ERROR(imp)) {
                error = PTR_TO_ERROR(imp);
                goto errorFreeModule;
            }
            imp->functions = MemManager_Alloc(sizeof(SysModManager_Function)*currInt->nFunctions, MemManager_BLOCK_MEM);
            if (IS_PTR_ERROR(imp->functions)) {
                error = PTR_TO_ERROR(imp->functions);
                goto errorFreeImp;
            }
            imp->name = MemManager_Alloc(StringMan_Length(impInfo.name)+1, MemManager_BLOCK_MEM);
            if (IS_PTR_ERROR(imp->name)) {
                error = PTR_TO_ERROR(imp->name);
                goto errorFreeFunctions;
            }
            StringMan_Copy(imp->name, impInfo.name);
            error = RSysModFile_GetFuncsInfo(file, nInter, nImp, (RSysModFile_FunctionInfo*)imp->functions);
            if (error) {
                ret = MemManager_Free(imp->name);
                ASSERT_ERROR(ret);
errorFreeFunctions:
                ret = MemManager_Free(imp->functions);
                ASSERT_ERROR(ret);
errorFreeImp:
                ret = MemManager_Free(imp);
                ASSERT_ERROR(ret);
                goto errorFreeModule;
            }
            imp->module = module;
            imp->interface = currInt;
            imp->interN = nInter;
            imp->impN = nImp;
            TaskManager_P(&modulesSync, 0, 0);
            imp->next = currInt->imps;
            currInt->imps = imp;
            TaskManager_V(&modulesSync, 0);
            nImp++;
        } while (1);
        nInter++;
    } while (1);
    error = E_OK;
    goto errorEnd;

errorFreeModule:
    ret = MemManager_Free(module);
    ASSERT_ERROR(ret);
errorCloseFile:
    ret = RSysModFile_Close(file);
    ASSERT_ERROR(ret);
errorEnd:
    return error;
}

int SysModManager_SetModInfo(SysModManager_Module *module, SysModManager_Module *which, void *info, int prop, void **prev) {
    int error;
    SysModManager_ModData *data;

    ASSERT_RET(module, E_BAD_VALUE);
    ASSERT_RET(which, E_BAD_VALUE);
    ASSERT_RET(info, E_BAD_VALUE);

    TaskManager_LockWrite(&module->dataSync, 0);
    data = module->data;
    error = E_NOT_FOUND;
    while (data) {
        if(data->module == which) {
            error = E_IN_USE;
            if (!prop) {
                *prev = data->info;
                data->info = info;
                error = E_OK;
            }
            break;
        }
        data = data->next;
    }
    if (!data) {
        data = MemManager_Alloc(sizeof(SysModManager_ModData), MemManager_BLOCK_MEM);
        if (!IS_PTR_ERROR(data)) {
            data->module = which;
            data->info = info;
            data->next = module->data;
            module->data = data;
            error = E_OK;
        } else
            error = PTR_TO_ERROR(data);
    }
    TaskManager_UnlockRW(&module->dataSync, 0);
    return error;
}

int SysModManager_StartModule(unsigned int function) {
    SysModManager_Module *currM;
    int error;

    TaskManager_P(&modulesSync, 0, 0);
    currM = modules;
    while (currM && (function < currM->memStart || function >= (currM->memStart+currM->memSize)))
        currM = currM->next;
    error = E_BAD_VALUE;
    if (!currM) {
        goto unlockSync;
    }
    error = E_OK;
    if (!(currM->prop & SysModManager_MOD_PHASE0_STARTED)) {
        currM->prop |= SysModManager_MOD_PHASE0_STARTED;
        TaskManager_V(&modulesSync, 0);
        if (currM->phase0Start(currM)) {
            return E_STARTING_MODULE;
        } else {
            TaskManager_P(&modulesSync, 0, 0);
        }
    }
    if (!(currM->prop & SysModManager_MOD_PHASE1_STARTED) && phase == 1) {
        currM->prop |= SysModManager_MOD_PHASE1_STARTED;
        TaskManager_V(&modulesSync, 0);
        if (currM->phase1Start(currM)) {
            return E_STARTING_MODULE;
        } else {
            return E_OK;
        }
    }
unlockSync:
    TaskManager_V(&modulesSync, 0);
    return error;
}

/* caller must guarantee that module wont be freed */
int SysModManager_StartModuleUtil(SysModManager_Module *module) {
    int error;

    /* only one thread can load a module */
    TaskManager_P(&loadMutex, 0, 0);
    error = SysModManager_LoadModule(module);
    TaskManager_V(&loadMutex, 0);
    if (error) {
        return error;
    }
    TaskManager_P(&modulesSync, 0, 0);
    if (!(module->prop & SysModManager_MOD_PHASE0_STARTED)) {
        module->prop |= SysModManager_MOD_PHASE0_STARTED;
        TaskManager_V(&modulesSync, 0);
        if (module->phase0Start(module)) {
            return E_STARTING_MODULE;
        } else {
            TaskManager_P(&modulesSync, 0, 0);
        }
    }
    if (!(module->prop & SysModManager_MOD_PHASE1_STARTED) && phase == 1) {
        module->prop |= SysModManager_MOD_PHASE1_STARTED;
        TaskManager_V(&modulesSync, 0);
        if (module->phase1Start(module)) {
            return E_STARTING_MODULE;
        } else {
            return E_OK;
        }
    }
    TaskManager_V(&modulesSync, 0);
    return E_OK;
}

int SysModManager_Unregister(const char *path, int prop) {
    return E_UNSUPPORTED;
}
