/*
*****************************************
* This file implements the TaskManager interface
*
* Version: 0.3.7
* Author: Luiz Henrique Shigunov
* E-mail: luizshigunov@users.sourceforge.net
* WEB: http://www.sourceforge.net/projects/modulos
* Last change: 01/07/2005
*
* See the interface specification to know what each function does.
*
* Copyright (C) 2000-2005, Luiz Henrique Shigunov
*
* This module is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2.1 of the License, or (at your option) any later version.
*
* This module is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this module; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*****************************************
*/

#define __NO_TASKMANAGER_TYPES 1

/* must be declared first to avoid incomplete type errors */
struct _TaskManager_Semaphore {
    int count;
    struct _TaskManager_Thread *wait;
};

struct _TaskManager_RWSync {
    struct _TaskManager_Thread *waitWrite;
    struct _TaskManager_Thread *waitRead;
    struct _TaskManager_Thread *writer;
    int nActive;
};

struct _TaskManager_WaitQueue {
    struct _TaskManager_Thread *wait;
};

#include <stddef.h>
#include <debug.h>
#include <errormanager.h>
#include <intmanager_i386.h>
#include <memmanager.h>
#include <memmanager_i386.h>
#include <stringman.h>
#include <sysdef.h>
#include <sysmodmanager.h>
#include <taskmanager.h>
#include <timercontroller_i386.h>
#include <userspacemanager.h>

#define TaskManager_GET_CURRENT_THREAD_MACRO(x) __asm__ __volatile__ ("andl %%esp,%0; ":"=r" (x) : "0" (~4095UL))

/* round robin task quantum */
#define TaskManager_RR_QUANTUM 40
/* thread quantum */
#define TaskManager_THREAD_QUANTUM 4
/* thread property - suspended */
#define TaskManager_THREAD_SUSP 0x01
/* thread property - waiting */
#define TaskManager_THREAD_WAIT 0x02
/* thread property - been destroyed */
#define TaskManager_THREAD_DESTROYED 0x08
/* task property - system task */
#define TaskManager_TASK_SYS 0x01
/* task property - system task */
#define TaskManager_TASK_DESTROYED 0x02

#define PIDHASH_SIZE 256
#define TaskManager_PidHashFunc(id) ((((id) >> 8) ^ (id)) & (PIDHASH_SIZE-1))

struct _TaskManager_Task {
    struct _TaskManager_Task *nextTask;     /* used in the ready/suspended lists */
    struct _TaskManager_Task *prevTask;    /* used in the ready/suspended lists */
    struct _TaskManager_Thread *runThreads; /* ready thread list */
    struct _TaskManager_Thread *susThreads; /* suspended thread list */
    unsigned int PGD;       /* physical address */
    unsigned char pri;      /* current priority */
    unsigned char quantum;  /* current quantum */
    unsigned short res;
    int prop;
    /*
    0x01 - system task;
    0x02 - task been destroyed
    */
    unsigned int undestroyable;
    struct _TaskManager_Task *nextPIDHash;
    struct _TaskManager_Task **pprevPIDHash;
    int pid;
    struct _TaskManager_Task *parent;
    struct _TaskManager_Semaphore dataSem;  /* semaphore to use the data list */
    struct _TaskManager_ModData *data;      /* linked list with the data kept */
};

struct _TaskManager_Thread {
    struct _TaskManager_Thread *nextThread; /* used in the ready/suspended lists */
    struct _TaskManager_Thread *prevThread; /* used in the ready/suspended lists */
    struct _TaskManager_Thread *next;       /* used when the thread is in some list */
    struct _TaskManager_Thread *prev;       /* used when the thread is in some list */
    TaskManager_Task *task;
    unsigned int *ESP0;   /* system stack pointer */
    unsigned int *ESP;    /* current stack pointer */
    int prop;
    /*
    0x01 - suspended thread;
    0x02 - waiting thread;
    0x04 - interruptible
    0x08 - thread been destroyed
    */
    unsigned int undestroyable;
    struct _TaskManager_Semaphore dataSem;  /* semaphore to use the list */
    struct _TaskManager_ModData *data;      /* linked list with the data kept */
    unsigned int waitTime;  /* time to wait */
    int tid;
    long errorCode;
    int errorFunc;
    char errorName[32];
};

typedef struct _TaskManager_ModData {
    struct _TaskManager_ModData *next;
    SysModManager_Module *module;
    void *info;
} TaskManager_ModData;

typedef struct {
    unsigned int backLink;
    unsigned int ESP0;
    unsigned int SS0;
    unsigned int ESP1;
    unsigned int SS1;
    unsigned int ESP2;
    unsigned int SS2;
    unsigned int CR3;
    unsigned int EIP;
    unsigned int EFLAGS;
    unsigned int EAX;
    unsigned int ECX;
    unsigned int EDX;
    unsigned int EBX;
    unsigned int ESP;
    unsigned int EBP;
    unsigned int ESI;
    unsigned int EDI;
    unsigned int ES;
    unsigned int CS;
    unsigned int SS;
    unsigned int DS;
    unsigned int FS;
    unsigned int GS;
    unsigned int LDT;
    unsigned int IOMap;
} TaskManager_TSS;

SysModManager_Module *ID;
unsigned int msCounter; /* ms counter */
TaskManager_Task *readyTasks;  /* ready tasks circular linked list */
TaskManager_Task *idleTasks;   /* idle tasks circular linked list */
TaskManager_Task *susTasks;    /* suspended tasks linked list */
TaskManager_Task *currTask;    /* current task executing */
TaskManager_Task *destTask;    /* task that destroy tasks/threads */
TaskManager_Task *startTask;   /* task for the first thread */
TaskManager_Task idleTask;     /* system idle task */

TaskManager_Task *tasksToDest;     /* linked list of tasks to be destroyed */
TaskManager_Thread *threadsToDest; /* linked list of threads to be destroyed */
TaskManager_Semaphore ttToDestSem;

TaskManager_Thread *wThreads;  /* linked list of threads waiting */

unsigned int threadTime = TaskManager_THREAD_QUANTUM;  /* execution time of a thread */
TaskManager_TSS	currTSS;  /* system TSS */

MemManager_Cache *taskCache;  /* task structure cache */
MemManager_Cache *dataCache;  /* ModData structure cache */

int taskPIDHint;
int threadTIDHint;
TaskManager_Task *pidHash[PIDHASH_SIZE];

void TaskManager_DoThreadDestroy(TaskManager_Thread*);
int TaskManager_ServeInterrupt(IntManager_i386_Regs*);
void TaskManager_StartThread(void);

int TaskManager_CancelWait(TaskManager_Thread *thread, int prop) {
    int flags;

    ASSERT_RET(thread, E_BAD_VALUE);
    ASSERT(thread->undestroyable);

    flags = IntManager_i386_NoInt();
    if (thread->prop & TaskManager_THREAD_WAIT) {
        thread->prop &= ~TaskManager_THREAD_WAIT;
        if (thread->next && thread->prev) {
            /* remove from the middle of the list */
            thread->prev->next = thread->next;
            thread->next->prev = thread->prev;
        } else if (thread->prev)
            /* we're in the end of the list */
            thread->prev->next = NULL;
        else {
            /* we're in the start of the list */
            if (thread->next)
                thread->next->prev = NULL;
            wThreads = thread->next;
        }
        if (!(prop & 0x01)) {
            int ret;
            ret = TaskManager_Resume(thread, prop>>1);
            ASSERT_ERROR(ret);
        }
    }
    IntManager_i386_RestoreInt(flags);
    return E_OK;
}

static __inline__ void TaskManager_InsertPIDHash(TaskManager_Task *task) {
    TaskManager_Task **table;

    table = &pidHash[TaskManager_PidHashFunc(task->pid)];
    task->nextPIDHash = *table;
    if (*table)
        (*table)->pprevPIDHash = &task->nextPIDHash;
    *table = task;
    task->pprevPIDHash = table;
}

static __inline void TaskManager_RemovePIDHash(TaskManager_Task *task) {
    if (task->nextPIDHash)
        task->nextPIDHash->pprevPIDHash = task->pprevPIDHash;
    *task->pprevPIDHash = task->nextPIDHash;
}

static __inline__ int TaskManager_AllocPID(unsigned int page) {
    taskPIDHint = (taskPIDHint + 1) & 0xff;
    return (page >> 4) + taskPIDHint;
}

static __inline__ int TaskManager_AllocTID(unsigned int page) {
    threadTIDHint = (threadTIDHint + 1) & 0xff;
    return (page >> 4) + threadTIDHint;
}

int TaskManager_CreateTask(void *startCode, void *arg, int prop, TaskManager_Task **task, TaskManager_Thread **thread) {
    TaskManager_Task *newTask;
    TaskManager_Thread *newThread;
    int flags, ret;

    ASSERT_RET((unsigned int)startCode >= 0xe0000000, E_BAD_VALUE);
    ASSERT_RET(task, E_BAD_VALUE);
    ASSERT_RET(thread, E_BAD_VALUE);

    newTask = MemManager_AllocObj(taskCache);
    if (IS_PTR_ERROR(newTask))
        goto errorEnd;
    newThread = MemManager_i386_AllocPages(1, MemManager_i386_RW_PAGE+MemManager_i386_ALLOC_PHYS);
    if (IS_PTR_ERROR(newThread))
        goto errorThread;
    if (!(prop & TaskManager_CREATE_SYSTEM)) {
        newTask->PGD = MemManager_i386_CreatePGD();
        if (IS_PTR_ERROR(newTask->PGD))
            goto errorMmap;
    }
    newTask->pri = prop >> 24;
    newTask->quantum = (newTask->pri<<1)+30;
    newTask->prop = (prop & TaskManager_CREATE_SYSTEM)>>1;
    newTask->undestroyable = 1;
    newTask->dataSem.count = 1;
    newTask->parent = currTask;
    newTask->pid = TaskManager_AllocPID((unsigned int)newThread);
    TaskManager_InsertPIDHash(newTask);
    newThread->tid = TaskManager_AllocTID((unsigned int)newThread);
    newThread->task = newTask;
    newThread->nextThread = NULL;
    newThread->prevThread = NULL;
    newThread->prop = prop & 0x01;
    newThread->undestroyable = 1;
    newThread->dataSem.count = 1;
    newThread->dataSem.wait = NULL;
    newThread->data = NULL;
    newThread->errorCode = 0;
    newThread->errorFunc = 0;
    newThread->errorName[0] = 0;
    newThread->ESP0 = (unsigned int *)((unsigned int)(newThread) + 4096);
    newThread->ESP = newThread->ESP0 - 1;
    *(newThread->ESP) = 0x0;              /* parameter for DestroyThread */
    (newThread->ESP)--;
    *(newThread->ESP) = (int)arg;         /* parameter for startcode */
    (newThread->ESP)--;
    *(newThread->ESP) = (int)TaskManager_DestroyThread;  /* to destroy the thread if it returns */
    (newThread->ESP)--;
    *(newThread->ESP) = 0x00000202;       /* EFLAGS */
    (newThread->ESP)--;
    *(newThread->ESP) = DPL0CodeSegment;  /* CS */
    (newThread->ESP)--;
    *(newThread->ESP) = (int)startCode;   /* EIP */
    (newThread->ESP)--;
    *(newThread->ESP) = (int)TaskManager_StartThread; /* function that does iret */
    flags = IntManager_i386_NoInt();
    if (!(prop & TaskManager_CREATE_SUS)) {
        TaskManager_Task **list;
        if (newTask->pri)
            list = &readyTasks;
        else
            list = &idleTasks;
        if (!*list) {
            *list = newTask;
            newTask->nextTask = newTask;
        }
        (*list)->nextTask->prevTask = newTask;
        newTask->nextTask = (*list)->nextTask;
        (*list)->nextTask = newTask;
        newTask->prevTask = *list;
        newTask->runThreads = newThread;
    } else {
        newTask->susThreads = newThread;
        newTask->nextTask = susTasks;
        newTask->prevTask = NULL;
        if (susTasks)
            susTasks->prevTask = newTask;
        susTasks = newTask;
    }
    IntManager_i386_RestoreInt(flags);
    *task = newTask;
    *thread = newThread;
    return E_OK;

errorMmap:
    ret = MemManager_i386_FreePages(newThread, 1, 0);
    ASSERT_ERROR(ret);
errorThread:
    ret = MemManager_FreeObj(taskCache, newTask);
    ASSERT_ERROR(ret);
errorEnd:
    return E_NO_MEMORY;
}

int TaskManager_CreateThread(TaskManager_Task *task, void *startCode, void *arg, int prop, TaskManager_Thread **thread) {
    TaskManager_Thread *newThread, **list;
    int flags;

    ASSERT_RET((unsigned int)startCode >= 0xe0000000, E_BAD_VALUE);
    ASSERT_RET(thread, E_BAD_VALUE);

    if (!task)
        task = currTask;
    else
        ASSERT(task->undestroyable);
    newThread = MemManager_i386_AllocPages(1, MemManager_i386_RW_PAGE+MemManager_i386_ALLOC_PHYS);
    if (IS_PTR_ERROR(newThread))
        return PTR_TO_ERROR(newThread);;
    newThread->tid = TaskManager_AllocTID((int)newThread);
    newThread->prop = prop & TaskManager_CREATE_SUS;
    newThread->undestroyable = 1;
    newThread->prevThread = NULL;
    newThread->dataSem.count = 1;
    newThread->dataSem.wait = NULL;
    newThread->data = NULL;
    newThread->task = task;
    newThread->errorCode = 0;
    newThread->errorFunc = 0;
    newThread->errorName[0] = 0;
    newThread->ESP0 = (unsigned int *)((unsigned int)(newThread) + 4096);
    newThread->ESP = newThread->ESP0 - 1;
    *(newThread->ESP) = 0x0;              /* parameter for DestroyThread */
    (newThread->ESP)--;
    *(newThread->ESP) = (int)arg;         /* parameter for startcode */
    (newThread->ESP)--;
    *(newThread->ESP) = (int)TaskManager_DestroyThread;  /* to destroy the thread if it returns */
    (newThread->ESP)--;
    *(newThread->ESP) = 0x00000202;       /* EFLAGS */
    (newThread->ESP)--;
    *(newThread->ESP) = DPL0CodeSegment;  /* CS */
    (newThread->ESP)--;
    *(newThread->ESP) = (int)startCode;   /* EIP */
    (newThread->ESP)--;
    *(newThread->ESP) = (int)TaskManager_StartThread; /* function that does iret */
    TaskManager_AtomicInc(&task->undestroyable);
    list = &task->runThreads;
    if (prop & TaskManager_CREATE_SUS)
        list = &task->susThreads;
    flags = IntManager_i386_NoInt();
    newThread->nextThread = *list;
    if (*list)
        (*list)->prevThread = newThread;
    *list = newThread;
    IntManager_i386_RestoreInt(flags);
    *thread = newThread;
    return E_OK;
}

void *TaskManager_DelTaskInfo(TaskManager_Task *task, SysModManager_Module *module) {
    int ret;
    void *error;
    TaskManager_ModData *data, **prev;

    ASSERT_RET(module, ERROR_TO_PTR(E_BAD_VALUE));

    if (!task)
        task = currTask;
    else
        ASSERT(task->undestroyable);
    error = ERROR_TO_PTR(E_NOT_FOUND);
    TaskManager_P(&task->dataSem, 0, 0);
    data = task->data;
    prev = &task->data;
    while (data) {
        if (data->module == module) {
            *prev = data->next;
            error = data->info;
            ret = MemManager_FreeObj(dataCache, data);
            ASSERT_ERROR(ret);
            break;
        }
        prev = &data->next;
        data = data->next;
    }
    TaskManager_V(&task->dataSem, 0);
    return error;
}

void *TaskManager_DelThreadInfo(TaskManager_Thread *thread, SysModManager_Module *module) {
    int ret;
    void *error;
    TaskManager_ModData *data, **prev;

    ASSERT_RET(module, ERROR_TO_PTR(E_BAD_VALUE));

    if (!thread)
        TaskManager_GET_CURRENT_THREAD_MACRO(thread);
    else
        ASSERT(thread->undestroyable);
    error = ERROR_TO_PTR(E_NOT_FOUND);
    TaskManager_P(&thread->dataSem, 0, 0);
    data = thread->data;
    prev = &thread->data;
    while (data) {
        if (data->module == module) {
            *prev = data->next;
            error = data->info;
            ret = MemManager_FreeObj(dataCache, data);
            ASSERT_ERROR(ret);
            break;
        }
        prev = &data->next;
        data = data->next;
    }
    TaskManager_V(&thread->dataSem, 0);
    return error;
}

void TaskManager_Destroyable(TaskManager_Task *task) {
    int ret;

    if (!task)
        task = currTask;
    ASSERT(task->undestroyable);
    /* DestroyTask must be called with an undestroyable task */
    if (task->prop & TaskManager_TASK_DESTROYED) {
        ret = TaskManager_DestroyTask(task);
        ASSERT_ERROR(ret);
    }
    TaskManager_AtomicDec(&task->undestroyable);
}

void TaskManager_DestroyableThread(TaskManager_Thread *thread) {
    TaskManager_Task *task;
    int ret;

    if (!thread)
        TaskManager_GET_CURRENT_THREAD_MACRO(thread);
    task = thread->task;
    ASSERT(thread->undestroyable);
    ASSERT(task->undestroyable);
    if (task->prop & TaskManager_TASK_DESTROYED) {
        ret = TaskManager_DestroyTask(task);
        ASSERT_ERROR(ret);
    }
    if (thread->prop & TaskManager_THREAD_DESTROYED) {
        ret = TaskManager_DestroyThread(thread);
        ASSERT_ERROR(ret);
    }
    TaskManager_AtomicDec(&thread->undestroyable);
    TaskManager_AtomicDec(&task->undestroyable);
}

int TaskManager_DestroyTask(TaskManager_Task *task) {
    int flags, ret;
    unsigned int i;

    flags = IntManager_i386_NoInt();
    if (!task)
        task = currTask;
    /* task must be undestroyable */
    ASSERT(task->undestroyable);
    task->undestroyable--;
    if (task->undestroyable) {
        TaskManager_Thread *thread;

        if (!(task->prop & TaskManager_TASK_DESTROYED)) {
            /* set task as been destroyed */
            task->prop |= TaskManager_TASK_DESTROYED;
            /* interrupt all threads that are interruptible */
            thread = task->susThreads;
            while (thread) {
                if (thread->prop & TaskManager_THREAD_INT) {
                    /* thread must be undestroyable */
                    thread->undestroyable++;
                    ret = TaskManager_Resume(thread, TaskManager_RESUME_KEEP_CPU);
                    thread->undestroyable--;
                    ASSERT_ERROR(ret);
                }
                thread = thread->nextThread;
            }
        }
        /* stop caller thread until we can destroy the task */
        TaskManager_GET_CURRENT_THREAD_MACRO(thread);
        if (task == thread->task) {
            /* thread must be undestroyable */
            thread->undestroyable++;
            ret = TaskManager_Suspend(thread);
            thread->undestroyable--;
            ASSERT_MSG_PRE(ret);
        }
    } else {
        if (task->nextTask && task->prevTask) {
            /* remove from the middle of the suspended list, from the ready list or from idle list */
            if (task->nextTask != task) {
                task->prevTask->nextTask = task->nextTask;
                task->nextTask->prevTask = task->prevTask;
            } else
                task->nextTask = NULL;
            if (task->pri) {
                if (readyTasks == task)
                    readyTasks = task->nextTask;
            } else {
                if (idleTasks == task)
                    idleTasks = task->nextTask;
            }
        } else if (task->prevTask)
           /* we're in the end of the suspended list */
           task->prevTask->nextTask = NULL;
        else {
            /* we're in the start of the suspended list */
            if (task->nextTask)
                task->nextTask->prevTask = NULL;
            susTasks = task->nextTask;
        }
        TaskManager_RemovePIDHash(task);
        /* reparent our children */
        for (i = 0; i < PIDHASH_SIZE; i++) {
            TaskManager_Task *child;

            child = pidHash[i];
            for ( ; child; child = child->nextPIDHash) {
                if (child->parent == task)
                    child->parent = &idleTask;
            }
        }
        /* if it's the current task, this will cause the Schedule function to change of task */
        task->quantum = 1;
        task->nextTask = tasksToDest;
        tasksToDest = task;
        TaskManager_V(&ttToDestSem, TaskManager_RESUME_KEEP_CPU);
        if (currTask == task)
            TaskManager_Schedule();
    }
    IntManager_i386_RestoreInt(flags);
    return E_OK;
}

int TaskManager_DestroyThread(TaskManager_Thread *thread) {
    TaskManager_Thread *curr;
    TaskManager_Task *task;
    int flags, ret;

    flags = IntManager_i386_NoInt();
    if (!thread)
        TaskManager_GET_CURRENT_THREAD_MACRO(thread);
    /* thread must be undestroyable */
    task = thread->task;
    ASSERT(thread->undestroyable);
    ASSERT(task->undestroyable);
    thread->undestroyable--;
    task->undestroyable--;
    if (thread->undestroyable) {
        if (!(thread->prop & TaskManager_THREAD_DESTROYED)) {
            /* set thread as been destroyed */
            thread->prop |= TaskManager_THREAD_DESTROYED;
            /* interrupt thread if it's interruptible */
            if (thread->prop & TaskManager_THREAD_INT) {
                ret = TaskManager_Resume(thread, TaskManager_RESUME_KEEP_CPU);
                ASSERT_ERROR(ret);
            }
        }
        TaskManager_GET_CURRENT_THREAD_MACRO(curr);
        if (curr == thread) {
            /* stop caller thread until we can destroy the thread */
            ret = TaskManager_Suspend(thread);
            ASSERT_MSG_PRE(ret);
        }
    } else {
        if (!thread->nextThread && !thread->prevThread && (!task->runThreads || !task->susThreads)) {
            int error;
            
            /* task must be undestroyable for DestroyTask */
            task->undestroyable++;
            error = TaskManager_DestroyTask(task);
            IntManager_i386_RestoreInt(flags);
            return error;
        }
        /* remove from run list */
        if (thread->nextThread && thread->prevThread) {
            /* remove from the middle of the suspended or ready list */
            thread->prevThread->nextThread = thread->nextThread;
            thread->nextThread->prevThread = thread->prevThread;
        } else if (thread->prevThread)
            /* we're in the end of the list */
            thread->prevThread->nextThread = NULL;
        else {
            /* we're in the start of the list */
            if (thread->nextThread)
                thread->nextThread->prevThread = NULL;
            if (thread->prop & TaskManager_THREAD_SUSP)
                task->susThreads = thread->nextThread;
            else
                task->runThreads = thread->nextThread;
        }
        if (!task->runThreads) {
            /* if it's the current task, this will cause the Schedule function to change of task */
            task->quantum = 1;
            /* remove from run list */
            if (task->nextTask != task) {
                task->prevTask->nextTask = task->nextTask;
                task->nextTask->prevTask = task->prevTask;
            } else
                task->nextTask = NULL;
            if (task->pri) {
                if (readyTasks == task)
                    readyTasks = task->nextTask;
            } else {
                if (idleTasks == task)
                    idleTasks = task->nextTask;
            }
            task->nextTask = susTasks;
            task->prevTask = NULL;
            if (susTasks)
                susTasks->prevTask = task;
            susTasks = task;
        }
        thread->nextThread = threadsToDest;
        threadsToDest = thread;
        TaskManager_V(&ttToDestSem, TaskManager_RESUME_KEEP_CPU);
        TaskManager_GET_CURRENT_THREAD_MACRO(curr);
        if (curr == thread) {
            /* this will cause a change in the current thread */
            threadTime = 1;
            TaskManager_Schedule();
        }
    }
    IntManager_i386_RestoreInt(flags);
    return E_OK;
}

void TaskManager_DoNothing(void) {
    /* if there's more task, stop */
    while (1) {
        if (currTask->nextTask != currTask)
            currTask->quantum = 1;
        __asm__ __volatile__ ("hlt");
    }
}

void TaskManager_DoTTDestroy(void) {
    int flags, ret;
    TaskManager_Task *task;
    TaskManager_Thread *thread;

    while (1) {
        flags = IntManager_i386_NoInt();
        task = tasksToDest;
        tasksToDest = NULL;
        thread = threadsToDest;
        threadsToDest = NULL;
        IntManager_i386_RestoreInt(flags);
        while (task) {
            TaskManager_Task *tmpT;

            /* event handler expect an undestroyable task */
            task->undestroyable++;
            TaskManager_DoThreadDestroy(task->runThreads);
            TaskManager_DoThreadDestroy(task->susThreads);
            SysModManager_Broadcast("TaskManagerE", 0, task, NULL);
            while(task->data) {
                TaskManager_ModData *tmp = task->data;
                task->data = task->data->next;
                ret = MemManager_FreeObj(dataCache, tmp);
                ASSERT_ERROR(ret);
            }
            if (!(task->prop & TaskManager_TASK_SYS)) {
                ret = MemManager_i386_FreePhysPages(task->PGD, 1);
                ASSERT_ERROR(ret);
            }
            task->undestroyable = 0;
            task->runThreads = NULL;
            task->susThreads = NULL;
            task->dataSem.wait = NULL;
            tmpT = task;
            task = task->nextTask;
            ret = MemManager_FreeObj(taskCache, tmpT);
            ASSERT_ERROR(ret);
        }
        TaskManager_DoThreadDestroy(thread);
        TaskManager_P(&ttToDestSem, 0, 0);
    }
}

void TaskManager_DoThreadDestroy(TaskManager_Thread *thread) {
    int ret;

    while(thread) {
        TaskManager_Thread *tmpT;
        TaskManager_ModData *currD;
        /* event handler expect an undestroyable thread */
        thread->undestroyable++;
        SysModManager_Broadcast("TaskManagerE", 1, thread->task, thread);
        currD = thread->data;
        while(currD) {
            TaskManager_ModData *tmp = currD;
            currD = currD->next;
            ret = MemManager_FreeObj(dataCache, tmp);
            ASSERT_ERROR(ret);
        }
        tmpT = thread;
        thread = thread->nextThread;
        ret = MemManager_i386_FreePages(tmpT, 1, 0);
        ASSERT_ERROR(ret);
    }
    return;
}

long TaskManager_GetLastError(const char **name, int *function) {
    TaskManager_Thread *curr;

    TaskManager_GET_CURRENT_THREAD_MACRO(curr);
    if (name != NULL)
        *name = curr->errorName;
    if (function != NULL)
        *function = curr->errorFunc;
    return curr->errorCode;
}

unsigned int TaskManager_GetMsCounter(void) {
    return msCounter;
}

TaskManager_Task *TaskManager_GetParentTaskHandle(TaskManager_Task *task) {
    TaskManager_Task *parent;
    
    if (task) {
        ASSERT(task->undestroyable);
        parent = task->parent;
    } else
        parent = currTask->parent;
    TaskManager_AtomicInc(&parent->undestroyable);
    return parent;
}

static __inline__ TaskManager_Task *GetTaskHandleIDHash(int taskID) {
    TaskManager_Task *task, **table;

    table = &pidHash[TaskManager_PidHashFunc(taskID)];
    for (task = *table; task && task->pid != taskID; task = task->nextPIDHash)
        ;
    return task;
}

TaskManager_Task *TaskManager_GetTaskHandleID(int taskID) {
    TaskManager_Task *task;
    int flags;

    if (taskID) {
        flags = IntManager_i386_NoInt();
        task = GetTaskHandleIDHash(taskID);
        if (task)
            task->undestroyable++;
        IntManager_i386_RestoreInt(flags);
    } else {
        task = currTask;
        TaskManager_AtomicInc(&task->undestroyable);
    }
    return task;
}

TaskManager_Task *TaskManager_GetTaskHandle(TaskManager_Thread *thread) {
    TaskManager_Task *task;
    
    if (thread) {
        ASSERT(thread->undestroyable);
        task = thread->task;
    } else
        task = currTask;
    TaskManager_AtomicInc(&task->undestroyable);
    return task;
}

int TaskManager_GetTaskID(TaskManager_Task *task) {
    if (task) {
        ASSERT(task->undestroyable);
        return task->pid;
    }
    return currTask->pid;
}

void *TaskManager_GetTaskInfo(TaskManager_Task *task, SysModManager_Module *module) {
    void *error;
    TaskManager_ModData *data;

    ASSERT_RET(module, ERROR_TO_PTR(E_BAD_VALUE));

    if (!task)
        task = currTask;
    else
        ASSERT(task->undestroyable);
    error = ERROR_TO_PTR(E_NOT_FOUND);
    TaskManager_P(&task->dataSem, 0, 0);
    data = task->data;
    while (data) {
        if (data->module == module) {
            error = data->info;
            break;
        }
        data = data->next;
    }
    TaskManager_V(&task->dataSem, 0);
    return error;
}

int TaskManager_GetTaskMMap(TaskManager_Task *task) {
    if (!task)
        task = currTask;
    else
        ASSERT(task->undestroyable);
    return (int)task->PGD;
}

TaskManager_Thread *TaskManager_GetThreadHandleID(TaskManager_Task *task, int threadID) {
    TaskManager_Thread *curr;
    int flags;

    if (!threadID) {
        TaskManager_GET_CURRENT_THREAD_MACRO(curr);
        TaskManager_AtomicInc(&curr->undestroyable);
        TaskManager_AtomicInc(&curr->task->undestroyable);
        return curr;
    }
    if (!task)
        task = currTask;
    else
        ASSERT(task->undestroyable);
    flags = IntManager_i386_NoInt();
    curr = task->runThreads;
    while (curr) {
        if (curr->tid == threadID) {
            curr->undestroyable++;
            curr->task->undestroyable++;
            goto out;
        }
        curr = curr->nextThread;
    }
    curr = task->susThreads;
    while (curr) {
        if (curr->tid == threadID) {
            curr->undestroyable++;
            curr->task->undestroyable++;
            break;
        }
        curr = curr->nextThread;
    }
out:
    IntManager_i386_RestoreInt(flags);
    return curr;
}

TaskManager_Thread *TaskManager_GetThreadHandle(void) {
    TaskManager_Thread *thread;

    TaskManager_GET_CURRENT_THREAD_MACRO(thread);
    TaskManager_AtomicInc(&thread->undestroyable);
    TaskManager_AtomicInc(&thread->task->undestroyable);
    return thread;
}

int TaskManager_GetThreadID(TaskManager_Thread *thread) {
    if (thread) {
        ASSERT(thread->undestroyable);
        return thread->tid;
    }
    TaskManager_GET_CURRENT_THREAD_MACRO(thread);
    return thread->tid;
}

void *TaskManager_GetThreadInfo(TaskManager_Thread *thread, SysModManager_Module *module) {
    void *error;
    TaskManager_ModData *data;

    ASSERT_RET(module, ERROR_TO_PTR(E_BAD_VALUE));

    if (!thread)
        TaskManager_GET_CURRENT_THREAD_MACRO(thread);
    else
        ASSERT(thread->undestroyable);
    error = ERROR_TO_PTR(E_NOT_FOUND);
    TaskManager_P(&thread->dataSem, 0, 0);
    data = thread->data;
    while (data) {
        if (data->module == module) {
            error = data->info;
            break;
        }
        data = data->next;
    }
    TaskManager_V(&thread->dataSem, 0);
    return error;
}

int TaskManager_GetThreadStatus(TaskManager_Thread *thread) {
    if (!thread)
        return 1;
    else
        ASSERT(thread->undestroyable);
    if (thread->prop & TaskManager_THREAD_SUSP)
        return 0;
    return 1;
}

int TaskManager_LockRead(TaskManager_RWSync *rw, int prop) {
    int flags, error;
    TaskManager_Thread *currT;

    ASSERT_RET(rw, E_BUG);

    TaskManager_GET_CURRENT_THREAD_MACRO(currT);
    flags = IntManager_i386_NoInt();
    currT->prop |= (prop & TaskManager_THREAD_INT);
    /* we only interrupt threads once, so see if task is already been destroyed */
    if (currT->prop & TaskManager_THREAD_INT && (currT->prop & TaskManager_THREAD_DESTROYED || currT->task->prop & TaskManager_TASK_DESTROYED))
        error = E_INTERRUPTED;
    else {
        rw->nActive++;
        error = E_OK;
        if (rw->writer) {
            if (prop & TaskManager_LOCK_DONT_BLOCK) {
                rw->nActive--;
                error = E_WOULD_BLOCK;
            } else {
                currT->next = rw->waitRead;
                rw->waitRead = currT;
                currT->undestroyable++;
                TaskManager_Suspend(NULL);
                currT->undestroyable--;
                if (currT->prop & TaskManager_THREAD_INT && (currT->prop & TaskManager_THREAD_DESTROYED || currT->task->prop & TaskManager_TASK_DESTROYED)) {
                    TaskManager_Thread *thread, **pnext;

                    /* see if thread was interrupted */
                    pnext = &rw->waitRead;
                    thread = rw->waitRead;
                    while (thread) {
                        if (thread == currT) {
                            *pnext = thread->next;
                            rw->nActive--;
                            error = E_INTERRUPTED;
                            break;
                        }
                        pnext = &thread->next;
                        thread = thread->next;
                    }
                }
            }
        }
    }
    currT->prop &= ~TaskManager_THREAD_INT;
    IntManager_i386_RestoreInt(flags);
    return error;
}

int TaskManager_LockWrite(TaskManager_RWSync *rw, int prop) {
    int flags, error;
    TaskManager_Thread *currT;

    ASSERT_RET(rw, E_BUG);

    TaskManager_GET_CURRENT_THREAD_MACRO(currT);
    flags = IntManager_i386_NoInt();
    currT->prop |= (prop & TaskManager_THREAD_INT);
    /* we only interrupt threads once, so see if task is already been destroyed */
    if (currT->prop & TaskManager_THREAD_INT && (currT->prop & TaskManager_THREAD_DESTROYED || currT->task->prop & TaskManager_TASK_DESTROYED))
        error = E_INTERRUPTED;
    else {
        error = E_OK;
        if (rw->nActive || rw->writer) {
            if (prop & TaskManager_LOCK_DONT_BLOCK) {
                error = E_WOULD_BLOCK;
                goto returnRest;
            }
            currT->next = rw->waitWrite;
            rw->waitWrite = currT;
            currT->undestroyable++;
            TaskManager_Suspend(NULL);
            currT->undestroyable--;
            if (currT->prop & TaskManager_THREAD_INT && (currT->prop & TaskManager_THREAD_DESTROYED || currT->task->prop & TaskManager_TASK_DESTROYED)) {
                TaskManager_Thread *thread, **pnext;

                /* see if thread was interrupted */
                pnext = &rw->waitWrite;
                thread = rw->waitWrite;
                while (thread) {
                    if (thread == currT) {
                        *pnext = thread->next;
                        error = E_INTERRUPTED;
                        goto returnRest;
                    }
                    pnext = &thread->next;
                    thread = thread->next;
                }
            }
        }
        rw->writer = currT;
    }
returnRest:
    currT->prop &= ~TaskManager_THREAD_INT;
    IntManager_i386_RestoreInt(flags);
    return error;
}

int TaskManager_P(TaskManager_Semaphore *semaphore, int prop, unsigned int timeout) {
    int flags, error;
    TaskManager_Thread *currT;

    ASSERT_RET(semaphore, E_BUG);

    TaskManager_GET_CURRENT_THREAD_MACRO(currT);
    flags = IntManager_i386_NoInt();
    currT->prop |= (prop & TaskManager_THREAD_INT);
    /* we only interrupt threads once, so see if task is already been destroyed */
    if (currT->prop & TaskManager_THREAD_INT && (currT->prop & TaskManager_THREAD_DESTROYED || currT->task->prop & TaskManager_TASK_DESTROYED))
        error = E_INTERRUPTED;
    else {
        if (prop & TaskManager_LOCK_DONT_BLOCK && semaphore->count <= 0) {
            error = E_WOULD_BLOCK;
            goto returnRest;
        }
        semaphore->count--;
        error = E_OK;
        if (semaphore->count < 0) {
            currT->next = semaphore->wait;
            semaphore->wait = currT;
            currT->undestroyable++;
            if (!timeout) {
                /* wait till get it */
                TaskManager_Suspend(NULL);
            } else {
                /* wait till get it or time out */
                TaskManager_Wait(NULL, timeout);
                if (currT->prop & TaskManager_THREAD_WAIT)
                    TaskManager_CancelWait(currT, TaskManager_WAIT_SUS|TaskManager_WAIT_KEEP_CPU);
                else {
                    /* timed out */
                    error = E_TIMED_OUT;
                }
            }
            currT->undestroyable--;
            /* see whether thread was interrupted */
            if (currT->prop & TaskManager_THREAD_INT && (currT->prop & TaskManager_THREAD_DESTROYED || currT->task->prop & TaskManager_TASK_DESTROYED))
                error = E_INTERRUPTED;
            /* remove from queue if thread did not get sem */
            if (error) {
                TaskManager_Thread *thread, **pnext;

                pnext = &semaphore->wait;
                thread = semaphore->wait;
                while (thread) {
                    if (thread == currT) {
                        *pnext = thread->next;
                        semaphore->count++;
                        break;
                    }
                    pnext = &thread->next;
                    thread = thread->next;
                }
            }
        }
    }
returnRest:
    currT->prop &= ~TaskManager_THREAD_INT;
    IntManager_i386_RestoreInt(flags);
    return error;
}

int Phase0Start(SysModManager_Module *modID) {
    TaskManager_Thread *newThread;
    int ret;

    ID = modID;
    idleTasks = &idleTask;
    idleTask.nextTask = &idleTask;
    idleTask.prevTask = &idleTask;
    idleTask.parent = &idleTask;
    TaskManager_InsertPIDHash(&idleTask);
    if ((ret = SysModManager_StartModule((unsigned int)MemManager_AllocObj)) ||
        (ret = SysModManager_StartModule((unsigned int)IntManager_i386_NoInt))) {
        DEBUG_MSG_PRE(ret);
        return 1;
    }
    /* allocate the thread structure and system stack */
    newThread = MemManager_i386_AllocPages(1, MemManager_i386_RW_PAGE+MemManager_i386_ALLOC_PHYS);
    if (IS_PTR_ERROR(newThread)) {
        DEBUG_MSG_PRE(PTR_TO_ERROR(newThread));
        return 1;
    }
    idleTask.prop = 0x01;
    idleTask.undestroyable = 1;
    idleTask.runThreads = newThread;
    idleTask.quantum = 30;
    newThread->tid = TaskManager_AllocTID((int)newThread);
    newThread->task = &idleTask;
    newThread->nextThread = NULL;
    newThread->prevThread = NULL;
    newThread->prop = 0;
    newThread->undestroyable = 1;
    newThread->ESP0 = (unsigned int *)((unsigned int)(newThread) + 4096);
    newThread->ESP = newThread->ESP0 - 1;
    *(newThread->ESP) = 0x00000202;       /* EFLAGS */
    (newThread->ESP)--;
    *(newThread->ESP) = DPL0CodeSegment;  /* CS */
    (newThread->ESP)--;
    *(newThread->ESP) = (int)TaskManager_DoNothing;   /* EIP */
    (newThread->ESP)--;
    *(newThread->ESP) = (int)TaskManager_StartThread; /* function that does iret */
    taskCache = MemManager_CreateObjCache(sizeof(TaskManager_Task), 0, MemManager_CLEAR_OBJS, NULL, NULL);
    dataCache = MemManager_CreateObjCache(sizeof(TaskManager_ModData), 0, 0, NULL, NULL);
    if (IS_PTR_ERROR(taskCache) || IS_PTR_ERROR(dataCache)) {
        DEBUG_MSG_PRE(0);
        return 1;
    }
    startTask = MemManager_AllocObj(taskCache);
    if (IS_PTR_ERROR(startTask)) {
        DEBUG_MSG_PRE(PTR_TO_ERROR(startTask));
        return 1;
    }
    /* create current task and thread */
    currTask = startTask;
    TaskManager_GET_CURRENT_THREAD_MACRO(newThread);
    startTask->prop = TaskManager_TASK_SYS;
    startTask->runThreads = newThread;
    startTask->pri = 100;
    startTask->undestroyable = 1;
    startTask->quantum = (100<<1)+30;
    startTask->dataSem.count = 1;
    startTask->parent = &idleTask;
    startTask->pid = TaskManager_AllocPID((int)newThread);
    TaskManager_InsertPIDHash(startTask);
    newThread->tid = TaskManager_AllocTID((int)newThread);
    newThread->task = startTask;
    newThread->nextThread = NULL;
    newThread->prevThread = NULL;
    newThread->prop = 0;
    newThread->undestroyable = 1;
    newThread->dataSem.count = 1;
    newThread->dataSem.wait = NULL;
    newThread->data = NULL;
    newThread->ESP0 = (unsigned int *)((unsigned int)(newThread) + 4096);
    newThread->ESP = newThread->ESP0;
    /* insert in the ready list */
    if (!readyTasks) {
        readyTasks = startTask;
        startTask->nextTask = startTask;
    }
    readyTasks->nextTask->prevTask = startTask;
    startTask->nextTask = readyTasks->nextTask;
    readyTasks->nextTask = startTask;
    startTask->prevTask = readyTasks;
    return 0;
}

int Phase1Start(SysModManager_Module *modID) {
    unsigned short segment;
    MemManager_i386_SegDescriptor descriptor;
    TaskManager_Thread *tmpThread;
    int ret;

    if ((ret = SysModManager_StartModule((unsigned int)MemManager_AllocObj)) ||
        (ret = SysModManager_StartModule((unsigned int)IntManager_i386_NoInt))) {
        DEBUG_MSG_PRE(ret);
        return 1;
    }
    /* create task to destroy tasks/threads */
    if ((ret = TaskManager_CreateTask(TaskManager_DoTTDestroy, NULL, 0x01000002, &destTask, &tmpThread))) {
        DEBUG_MSG_PRE(ret);
        return 1;
    }
    if ((ret = IntManager_i386_Catch(modID, 0, TaskManager_ServeInterrupt, 0x01+0x04))) {
        DEBUG_MSG_PRE(ret);
        goto errorDestroyTask;
    }
    /* setup IRQ 0 to fire each ms */
    if ((ret = TimerController_i386_Alloc(modID, 0))) {
        DEBUG_MSG_PRE(ret);
        goto errorFreeInt;
    }
    TimerController_i386_Setup(ID, 0, 0x4A9, 0x04);
    currTSS.SS0 = DPL0DataSegment;
    descriptor.limit15 = 0x67;
    descriptor.base15 = (short)&currTSS;
    descriptor.base23 = ((unsigned int)(&currTSS) >> 16);
    descriptor.type = 0x9;
    descriptor.s = 0;
    descriptor.dpl = 0;
    descriptor.p = 1;
    descriptor.limit19 = 0;
    descriptor.avl = 0;
    descriptor.reserved = 0;
    descriptor.db = 0;
    descriptor.g = 0;
    descriptor.base31 = ((unsigned int)(&currTSS) >> 24);
    ret = MemManager_i386_CreateDescriptor(descriptor, &segment);
    ASSERT_ERROR(ret);
    __asm__ __volatile__ ("ltr %0" : : "q" (segment));
    return 0;

errorFreeInt:
    ret = IntManager_i386_Free(modID, 0, 0x01);
    ASSERT_ERROR(ret);
errorDestroyTask:
    ret = TaskManager_DestroyTask(destTask);
    ASSERT_ERROR(ret);
    return 1;
}

int TaskManager_Resume(TaskManager_Thread *thread, int prop) {
    TaskManager_Task *task;
    int flags;

    ASSERT_RET(thread, E_BAD_VALUE);
    ASSERT(thread->undestroyable);

    flags = IntManager_i386_NoInt();
    if (thread->prop & TaskManager_THREAD_SUSP) {
        thread->prop &= ~TaskManager_THREAD_SUSP;
        task = thread->task;
        /* remove form sus list */
        if (thread->nextThread && thread->prevThread) {
            /* remove from the middle of the list */
            thread->prevThread->nextThread = thread->nextThread;
            thread->nextThread->prevThread = thread->prevThread;
        } else if (thread->prevThread)
            /* we're in the end of the list */
            thread->prevThread->nextThread = NULL;
        else {
            /* we're in the start of the list */
            if (thread->nextThread)
                thread->nextThread->prevThread = NULL;
            task->susThreads = thread->nextThread;
        }
        /* insert in run list */
        thread->nextThread = task->runThreads;
        thread->prevThread = NULL;
        if (task->runThreads)
            task->runThreads->prevThread = thread;
        task->runThreads = thread;
        if (!task->runThreads->nextThread) {
            TaskManager_Task **list;
            /* remove task from sus list */
            if (task->nextTask && task->prevTask) {
                /* remove from the middle of the list */
                task->prevTask->nextTask = task->nextTask;
                task->nextTask->prevTask = task->prevTask;
            } else if (task->prevTask)
                /* we're in the end of the list */
                task->prevTask->nextTask = NULL;
            else {
                /* we're in the start of the list */
                if (task->nextTask)
                    task->nextTask->prevTask = NULL;
                susTasks = task->nextTask;
            }
            /* insert in run list */
            if (task->pri)
                list = &readyTasks;
            else
                list = &idleTasks;
            if (!*list) {
                *list = task;
                task->nextTask = task;
            }
            (*list)->nextTask->prevTask = task;
            task->nextTask = (*list)->nextTask;
            (*list)->nextTask = task;
            task->prevTask = *list;
        }
    }
    IntManager_i386_RestoreInt(flags);
    return E_OK;
}

#ifdef _INCLUDE_ASM
/* must be changed in assembly */
void TaskManager_Schedule(void) {
    TaskManager_Thread *currT;

    /* don't save any register! */
    currTask->quantum--;
    if (!currTask->quantum) {
        TaskManager_Task *toRun;
        /* more priority, greater quantum */
        currTask->quantum = (currTask->pri<<1)+30;
        if (readyTasks) {
            readyTasks = readyTasks->nextTask;
            toRun = readyTasks;
        } else {
            idleTasks = idleTasks->nextTask;
            toRun = idleTasks;
        }
        if (toRun != currTask) {
            currTask = toRun;
            if (!(currTask->prop & 0x01))
                __asm__ __volatile__ ("movl %0, %%cr3" : : "q" (currTask->PGD));
            threadTime = TaskManager_THREAD_QUANTUM;
            TaskManager_GET_CURRENT_THREAD_MACRO(currT);
            /* currT->ESP = ESP */
            __asm__ __volatile__ ("movl %%esp, (%0)" : : "q" (&currT->ESP));
            currT = currTask->runThreads;
            currTSS.ESP0 = (int)currT->ESP0;
            /* ESP = currT->ESP */
            __asm__ __volatile__ ("movl %0, %%esp" : : "q" (currT->ESP));
        }
        return;
    }
    threadTime--;
    if (!threadTime) {
        threadTime = TaskManager_THREAD_QUANTUM;
        TaskManager_GET_CURRENT_THREAD_MACRO(currT);
        /* currT->ESP = ESP */
        __asm__ __volatile__ ("movl %%esp, (%0)" : : "q" (&currT->ESP));
        currT = currT->nextThread;
        if (!currT)
            currT = currTask->runThreads;
        currTSS.ESP0 = (int)currT->ESP0;
        /* ESP = currT->ESP */
        __asm__ __volatile__ ("movl %0, %%esp" : : "q" (currT->ESP));
    }
    return;
}
#endif

int TaskManager_ServeInterrupt(IntManager_i386_Regs *regs) {
    TaskManager_Thread *thread;
    int ret;

    msCounter++;
    thread = wThreads;
    while (thread && thread->waitTime <= msCounter) {
        thread->prop &= ~TaskManager_THREAD_WAIT;
        /* thread was set undestroyable by Wait */
        ret = TaskManager_Resume(thread, TaskManager_RESUME_KEEP_CPU);
        ASSERT_ERROR(ret);
        thread = thread->next;
    }
    wThreads = thread;
    if (thread)
        thread->prev = NULL;
    return E_INVOKE_SCHEDULER;
}

long TaskManager_SetLastError(long error, const char *name, int function) {
    TaskManager_Thread *curr;

    ASSERT_RET(name != NULL, E_BAD_VALUE);

    TaskManager_GET_CURRENT_THREAD_MACRO(curr);
    // check whether name is too big
    if (StringMan_NCopy(curr->errorName, name, sizeof(curr->errorName)) != 0)
        return E_TOO_BIG;
    curr->errorCode = error;
    curr->errorFunc = function;
    return E_OK;
}

void *TaskManager_SetTaskInfo(TaskManager_Task *task, SysModManager_Module *module, void *info, int prop) {
    void *error;
    TaskManager_ModData *data;

    ASSERT_RET(module, ERROR_TO_PTR(E_BAD_VALUE));

    if (!task)
        task = currTask;
    else
        ASSERT(task->undestroyable);
    TaskManager_P(&task->dataSem, 0, 0);
    data = task->data;
    while (data) {
        if (data->module == module) {
            error = ERROR_TO_PTR(E_IN_USE);
            if (!prop) {
                error = data->info;
                data->info = info;
            }
            goto out;
        }
        data = data->next;
    }
    data = MemManager_AllocObj(dataCache);
    if (IS_PTR_ERROR(data)) {
        error = data;
        goto out;
    }
    data->module = module;
    data->info = info;
    data->next = task->data;
    task->data = data;
    error = NULL;
out:
    TaskManager_V(&task->dataSem, 0);
    return error;
}

void *TaskManager_SetThreadInfo(TaskManager_Thread *thread, SysModManager_Module *module, void *info, int prop) {
    void *error;
    TaskManager_ModData *data;

    ASSERT_RET(module, ERROR_TO_PTR(E_BAD_VALUE));

    if (!thread)
        TaskManager_GET_CURRENT_THREAD_MACRO(thread);
    else
        ASSERT(thread->undestroyable);
    TaskManager_P(&thread->dataSem, 0, 0);
    data = thread->data;
    while (data) {
        if (data->module == module) {
            error = ERROR_TO_PTR(E_IN_USE);
            if (!prop) {
                error = data->info;
                data->info = info;
            }
            goto out;
        }
        data = data->next;
    }
    data = MemManager_AllocObj(dataCache);
    if (IS_PTR_ERROR(data)) {
        error = data;
        goto out;
    }
    data->module = module;
    data->info = info;
    data->next = thread->data;
    thread->data = data;
    error = NULL;
out:
    TaskManager_V(&thread->dataSem, 0);
    return error;
}

void TaskManager_SpinLock(int *lock) {
    *lock = IntManager_i386_NoInt();
}

void TaskManager_SpinUnlock(int *lock) {
    IntManager_i386_RestoreInt(*lock);
}

#ifdef _INCLUDE_ASM
/* Must be modified in assembly */
void TaskManager_StartThread(void) {
    __asm__ __volatile__ ("iret");
}
#endif

int TaskManager_Suspend(TaskManager_Thread *thread) {
    TaskManager_Task *task;
    int flags;

    if (!thread)
        TaskManager_GET_CURRENT_THREAD_MACRO(thread);
    else
        ASSERT(thread->undestroyable);
    flags = IntManager_i386_NoInt();
    if (!(thread->prop & TaskManager_THREAD_SUSP)) {
        thread->prop |= TaskManager_THREAD_SUSP;
        task = thread->task;
        /* remove from run list */
        if (thread->nextThread && thread->prevThread) {
            /* remove from the middle of the list */
            thread->prevThread->nextThread = thread->nextThread;
            thread->nextThread->prevThread = thread->prevThread;
        } else if (thread->prevThread)
            /* we're in the end of the list */
            thread->prevThread->nextThread = NULL;
        else {
            /* we're in the start of the list */
            if (thread->nextThread)
                thread->nextThread->prevThread = NULL;
            task->runThreads = thread->nextThread;
        }
        /* insert in sus list */
        thread->nextThread = task->susThreads;
        thread->prevThread = NULL;
        if (task->susThreads)
            task->susThreads->prevThread = thread;
        task->susThreads = thread;
        if (!task->runThreads) {
            /* remove from run list */
            if (task->nextTask != task) {
                task->prevTask->nextTask = task->nextTask;
                task->nextTask->prevTask = task->prevTask;
            } else
                task->nextTask = NULL;
            if (task->pri) {
                if (readyTasks == task)
                    readyTasks = task->nextTask;
            } else {
                if (idleTasks == task)
                    idleTasks = task->nextTask;
            }
            /* insert in sus list */
            task->nextTask = susTasks;
            task->prevTask = NULL;
            if (susTasks)
                susTasks->prevTask = task;
            susTasks = task;
            if (currTask == task) {
                task->quantum = 1;
                TaskManager_Yield();
            }
        } else {
            TaskManager_Thread *currT;
            TaskManager_GET_CURRENT_THREAD_MACRO(currT);
            if (currT == thread) {
                threadTime = 1;
                TaskManager_Yield();
            }
        }
    }
    IntManager_i386_RestoreInt(flags);
    return E_OK;
}

#ifdef _INCLUDE_ASM
/* Must be modified in assembly */
void TaskManager_ToUserLevel(TaskManager_Regs *regs) {
    __asm__ __volatile__ (
        "pushl TaskManager_i386_SS(%%eax)\n\t"
        "pushl TaskManager_i386_ESP(%%eax)\n\t"
        "pushl TaskManager_i386_EFLAGS(%%eax)\n\t"
        "pushl TaskManager_i386_CS(%%eax)\n\t"
        "pushl TaskManager_i386_EIP(%%eax)\n\t"
        "movl  TaskManager_i386_EBX(%%eax), %%ebx\n\t"
        "movl  TaskManager_i386_ECX(%%eax), %%ecx\n\t"
        "movl  TaskManager_i386_EDX(%%eax), %%edx\n\t"
        "movl  TaskManager_i386_ESI(%%eax), %%esi\n\t"
        "movl  TaskManager_i386_EDI(%%eax), %%edi\n\t"
        "movl  TaskManager_i386_EBP(%%eax), %%ebp\n\t"
        "movl  TaskManager_i386_DS(%%eax) , %%ds\n\t"
        "movl  TaskManager_i386_ES(%%eax) , %%es\n\t"
        "movl  TaskManager_i386_EAX(%%eax), %%eax\n\t"
        "iret": : "a" (regs));
}
#endif

void TaskManager_UndestroyableThread(void) {
    TaskManager_Thread *currT;
    TaskManager_Task *task;
    int ret;

    TaskManager_GET_CURRENT_THREAD_MACRO(currT);
    task = currT->task;
    TaskManager_AtomicInc(&currT->undestroyable);
    TaskManager_AtomicInc(&task->undestroyable);
    if (task->prop & TaskManager_TASK_DESTROYED) {
        ret = TaskManager_DestroyTask(task);
        ASSERT_ERROR(ret);
    }
    if (currT->prop & TaskManager_THREAD_DESTROYED) {
        ret = TaskManager_DestroyThread(currT);
        ASSERT_ERROR(ret);
    }
}

int TaskManager_UnlockRW(TaskManager_RWSync *rw, int prop) {
    int flags, ret;

    ASSERT_RET(rw, E_BAD_VALUE);

    flags = IntManager_i386_NoInt();
    if (rw->writer) {
        rw->writer = NULL;
        if (rw->waitRead) {
            while (rw->waitRead) {
                TaskManager_Thread *tmp;
                tmp = rw->waitRead;
                rw->waitRead = tmp->next;
                /* thread was set undestroyable by LockRead or LockWrite */
                ret = TaskManager_Resume(tmp, prop);
                ASSERT_ERROR(ret);
            }
            goto errorEnd;
        }
        if (rw->waitWrite)
            goto changeWriter;
        goto errorEnd;
    }
    rw->nActive -= 1;
    if (!rw->nActive && rw->waitWrite) {
changeWriter:
        rw->writer = rw->waitWrite;
        rw->waitWrite = rw->waitWrite->next;
        /* thread was set undestroyable by LockRead or LockWrite */
        ret = TaskManager_Resume(rw->writer, prop);
        ASSERT_ERROR(ret);
    }
errorEnd:
    IntManager_i386_RestoreInt(flags);
    return E_OK;
}

int TaskManager_V(TaskManager_Semaphore *semaphore, int prop) {
    int flags, ret;

    ASSERT_RET(semaphore, E_BAD_VALUE);

    flags = IntManager_i386_NoInt();
    semaphore->count += 1;
    if (semaphore->count <= 0) {
        TaskManager_Thread *tmp;
        tmp = semaphore->wait;
        semaphore->wait = semaphore->wait->next;
        /* thread was set undestroyable by P */
        ret = TaskManager_Resume(tmp, prop);
        ASSERT_ERROR(ret);
    }
    IntManager_i386_RestoreInt(flags);
    return E_OK;
}

int TaskManager_Wait(TaskManager_Thread *thread, unsigned int time) {
    int flags, ret;
    TaskManager_Thread *currT, *prevT, **tmp;

    if (!thread)
        TaskManager_GET_CURRENT_THREAD_MACRO(thread);
    else
        ASSERT(thread->undestroyable);
    flags = IntManager_i386_NoInt();
    thread->prop |= TaskManager_THREAD_WAIT;
    thread->waitTime = msCounter+time;
    currT = wThreads;
    prevT = NULL;
    while (currT && currT->waitTime < thread->waitTime) {
        prevT = currT;
        currT = currT->next;
    }
    thread->prev = prevT;
    tmp = &prevT->next;
    if (!prevT)
        tmp = &wThreads;
    thread->next = *tmp;
    if (*tmp)
        (*tmp)->prev = thread;
    *tmp = thread;
    thread->undestroyable++;
    ret = TaskManager_Suspend(thread);
    thread->undestroyable--;
    ASSERT_ERROR(ret);
    IntManager_i386_RestoreInt(flags);
    return E_OK;
}

int TaskManager_WaitQueueAdd(TaskManager_WaitQueue *waitQ, TaskManager_Semaphore *sem, int prop) {
    TaskManager_Thread *currT;
    int flags, error, ret;

    ASSERT_RET(waitQ, E_BAD_VALUE);
    ASSERT_RET(sem, E_BAD_VALUE);

    TaskManager_GET_CURRENT_THREAD_MACRO(currT);
    flags = IntManager_i386_NoInt();
    currT->prop |= (prop & TaskManager_THREAD_INT);
    /* we only interrupt threads once, so see if task is already been destroyed */
    if (currT->prop & TaskManager_THREAD_INT && (currT->prop & TaskManager_THREAD_DESTROYED || currT->task->prop & TaskManager_TASK_DESTROYED))
        error = E_INTERRUPTED;
    else {
        currT->next = waitQ->wait;
        waitQ->wait = currT;
        currT->undestroyable++;
        TaskManager_V(sem, TaskManager_RESUME_KEEP_CPU);
        ret = TaskManager_Suspend(currT);
        currT->undestroyable--;
        ASSERT_ERROR(ret);
        error = E_OK;
        if (currT->prop & TaskManager_THREAD_INT && (currT->prop & TaskManager_THREAD_DESTROYED || currT->task->prop & TaskManager_TASK_DESTROYED)) {
            TaskManager_Thread *thread, **pnext;

            /* see if thread was interrupted */
            pnext = &waitQ->wait;
            thread = waitQ->wait;
            while (thread) {
                if (thread == currT) {
                    *pnext = thread->next;
                    error = E_INTERRUPTED;
                    break;
                }
                pnext = &thread->next;
                thread = thread->next;
            }
        }
    }
    currT->prop &= ~TaskManager_THREAD_INT;
    IntManager_i386_RestoreInt(flags);
    return error;
}

int TaskManager_WaitQueueWakeup(TaskManager_WaitQueue *waitQ) {
    int flags, ret;
    TaskManager_Thread *threads;

    ASSERT_RET(waitQ, E_BAD_VALUE);

    flags = IntManager_i386_NoInt();
    threads = waitQ->wait;
    waitQ->wait = NULL;
    IntManager_i386_RestoreInt(flags);
    while (threads) {
        TaskManager_Thread *tmp;
        tmp = threads;
        threads = threads->next;
        /* thread was set undestroyable by WaitQueueAdd */
        ret = TaskManager_Resume(tmp, 0);
        ASSERT_ERROR(ret);
    }
    return E_OK;
}

#ifdef _INCLUDE_ASM
/* Must be modified in assembly */
void TaskManager_Yield(void) {
    int flags;

    /* save all registers */
    __asm__ __volatile__ (
    "pushal"::);
    flags = IntManager_i386_NoInt();
    currTask->quantum = 1;
    TaskManager_Schedule();
    IntManager_i386_RestoreInt(flags);
    /* restore all registers */
    __asm__ __volatile__ (
    "popal"::);
}
#endif

typedef struct {
    TaskManager_Semaphore sem;
    TaskManager_WaitQueue waitQ;
    void *code, *stack;
} WaitCreate;

void StartUserThread(WaitCreate *wait) {
    TaskManager_Regs regs;

    regs.eax = 0;
    regs.ebx = 0;
    regs.ecx = 0;
    regs.edx = 0;
    regs.esi = 0;
    regs.edi = 0;
    regs.ebp = 0;
    regs.esp = (unsigned int)wait->stack;
    regs.eip = (unsigned int)wait->code;
    regs.eflags = 0x00000202;
    regs.ds = DPL3DataSegment;
    regs.es = DPL3DataSegment;
    regs.ss = DPL3DataSegment;
    regs.cs = DPL3CodeSegment;
    /* signal we already have the args */
    TaskManager_WaitQueueAdd(&wait->waitQ, &wait->sem, 0);
    TaskManager_ToUserLevel(&regs);
}

int TaskManager_UCreateThread(void *stack, void *startCode, int prop) {
    int ret, tid;
    TaskManager_Thread *sysThread;
    WaitCreate wait;

    wait.sem.count = 0;
    wait.sem.wait = NULL;
    wait.waitQ.wait = NULL;
    wait.code = startCode;
    wait.stack = stack;
    ret = TaskManager_CreateThread(NULL, StartUserThread, &wait, 0, &sysThread);
    if (ret)
        return -ret;
    /* wait StartUserThread get args */
    TaskManager_P(&wait.sem, 0, 0);
    if (!(prop & TaskManager_CREATE_SUS)) {
        int error;
        error = TaskManager_Resume(sysThread, 0);
        ASSERT_ERROR(error);
    }
    tid = sysThread->tid;
    /* set sysThread destroyable */
    TaskManager_DestroyableThread(sysThread);
    return tid;
}

int TaskManager_UDestroyThread(int tid, void *stack, unsigned int size) {
    TaskManager_Thread *thread;
    int ret;

    if (!tid)
        TaskManager_GET_CURRENT_THREAD_MACRO(thread);
    else {
        thread = TaskManager_GetThreadHandleID(NULL, tid);
        if (!thread)
            return E_BAD_VALUE;
    }
    /* free stack space */
    ret = MemManager_UFreeUserPages(stack, size);
    if (ret)
        return ret;
    ret = TaskManager_DestroyThread(thread);
    ASSERT_ERROR(ret);
    return ret;
}

long TaskManager_UGetLastError(char *name, int *function) {
    const char *sysName;
    int sysFunc;
    long sysError;

    sysError = TaskManager_GetLastError(&sysName, &sysFunc);
    if (name != NULL)
        UserSpaceManager_CopyToUser(name, sysName, 32);
    if (function != NULL)
        UserSpaceManager_CopyToUser(function, &sysFunc, sizeof(sysFunc));
    return sysError;
}

int TaskManager_UGetParentTaskID(void) {
    return currTask->parent->pid;
}

int TaskManager_UGetTaskID(void) {
    return currTask->pid;
}

int TaskManager_UGetThreadID(void) {
    TaskManager_Thread *currT;

    TaskManager_GET_CURRENT_THREAD_MACRO(currT);
    return currT->tid;
}

/* check if thread is a valid suspended thread pointer */
static __inline__ int CheckThreadPointerSus(TaskManager_Thread *thread) {
    TaskManager_Thread *curr;
    int flags;

    flags = IntManager_i386_NoInt();
    curr = currTask->susThreads;
    while (curr && curr != thread)
        curr = curr->nextThread;
    IntManager_i386_RestoreInt(flags);
    if (!curr)
        return E_BAD_VALUE;
    return E_OK;
}

int TaskManager_UP(TaskManager_Semaphore *semaphore) {
    int flags, ret;
    TaskManager_Semaphore sysSem;
    TaskManager_Thread *currT;

    TaskManager_GET_CURRENT_THREAD_MACRO(currT);
    flags = IntManager_i386_NoInt();
    ret = UserSpaceManager_CopyFromUser((char*)&sysSem, (char*)semaphore, sizeof(TaskManager_Semaphore));
    if (ret)
        goto returnRest;
    sysSem.count--;
    if (sysSem.count < 0) {
        /* check if wait is a valid pointer */
        if (sysSem.wait) {
            ret = CheckThreadPointerSus(sysSem.wait);
            if (ret)
                goto returnRest;
        }
        currT->next = sysSem.wait;
        sysSem.wait = currT;
        ret = UserSpaceManager_CopyToUser((char*)semaphore, (char*)&sysSem, sizeof(TaskManager_Semaphore));
        if (ret)
            goto returnRest;
        TaskManager_Suspend(NULL);
    } else
        ret = UserSpaceManager_CopyToUser((char*)semaphore, (char*)&sysSem, sizeof(TaskManager_Semaphore));
returnRest:
    IntManager_i386_RestoreInt(flags);
    return ret;
}

long TaskManager_USetLastError(long error, const char *name, int function) {
    char sysName[32];
    unsigned int size;
    long ret;

    ASSERT_RET(name != NULL, E_BAD_VALUE);

    sysName[sizeof(sysName)-1] = 0;
    ret = UserSpaceManager_StrNCopy(sysName, name, sizeof(sysName)-1, &size);
    if (ret)
        return ret;
    return TaskManager_SetLastError(error, sysName, function);
}

int TaskManager_UTryP(TaskManager_Semaphore *semaphore) {
    int flags, ret;
    unsigned int sysCount;

    flags = IntManager_i386_NoInt();
    ret = UserSpaceManager_CopyFromUser((char*)&sysCount, (char*)&semaphore->count, 4);
    if (ret)
        goto returnRest;
    ret = E_IN_USE;
    if (sysCount > 0) {
        sysCount--;
        ret = UserSpaceManager_CopyToUser((char*)&semaphore->count, (char*)&sysCount, 4);
    }
returnRest:
    IntManager_i386_RestoreInt(flags);
    return ret;
}

int TaskManager_UV(TaskManager_Semaphore *semaphore) {
    int flags, ret;
    TaskManager_Semaphore sysSem;

    flags = IntManager_i386_NoInt();
    ret = UserSpaceManager_CopyFromUser((char*)&sysSem, (char*)semaphore, sizeof(TaskManager_Semaphore));
    if (ret)
        goto returnRest;
    sysSem.count++;
    if (sysSem.count <= 0) {
        TaskManager_Thread *tmp;
        int error;
        /* check if wait is a valid pointer */
        ret = CheckThreadPointerSus(sysSem.wait);
        if (ret)
            goto returnRest;
        tmp = sysSem.wait;
        sysSem.wait = sysSem.wait->next;
        /* thread was set undestroyable by UP */
        error = TaskManager_Resume(tmp, 1);
        ASSERT_ERROR(error);
    }
    ret = UserSpaceManager_CopyToUser((char*)semaphore, (char*)&sysSem, sizeof(TaskManager_Semaphore));
returnRest:
    IntManager_i386_RestoreInt(flags);
    return ret;
}
