/*
*****************************************
* This file implements the UserModManager interface
*
* Version: 0.4.2
* Author: Luiz Henrique Shigunov
* E-mail: luizshigunov@users.sourceforge.net
* WEB: http://www.sourceforge.net/projects/modulos
* Last change: 25/06/2005
*
* See the interface specification to know what each function does.
*
* Copyright (C) 2000-2005, Luiz Henrique Shigunov
*
* This module is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2.1 of the License, or (at your option) any later version.
*
* This module is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this module; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*****************************************
*/

#include <stddef.h>
#include <avltree_in.h>
#include <cfgmanager.h>
#include <debug.h>
#include <errormanager.h>
#include <execloader.h>
#include <intmanager_i386.h>
#include <memman.h>
#include <memmanager.h>
#include <memmanager_i386.h>
#include <rexecmodfile.h>
#include <rlibmodfile.h>
#include <stringman.h>
#include <sysdef.h>
#include <sysmodmanager.h>
#include <taskmanager.h>
#include <taskmanager_i386.h>
#include <usermodmanager.h>
#include <userspacemanager.h>

#define ALIGN_DATA_UP(data, align) ((((unsigned int)(data)) + (align)-1) & ~((align)-1))
#define ALIGN_DATA_DOWN(data, align) (((unsigned int)(data)) & ~((align)-1))
/* start of brokers functions */
#define START_BROKERS 0x1000
#define SIZE_BROKERS 0x3fe000
/* start of module area */
#define START_EXEC_MOD 0x400000
#define START_EXEC_MOD_PDE 0x1
/* start of lib data area */
#define START_LIB_MOD_DATA 0x80000000
/* size of lib data area */
#define SIZE_LIB_MOD_DATA 0x20000000
/* start of lib code and read only data area */
#define START_LIB_MOD_CODE 0xa0000000
#define START_LIB_MOD_CODE_PDE 0x280
/* size of lib code and read only data area */
#define SIZE_LIB_MOD_CODE 0x40000000
#define SIZE_LIB_MOD_CODE_PDE 0x100
/* size of broker function aligned in 4 bytes */
#define SIZE_BROKER 0x10
/* min size of thread stack */
#define MIN_STACK_SIZE (32*1024)
#define MIN_STACK_SIZE_2 15
/* default size of thread stack */
#define DEFAULT_STACK_SIZE (128*1024)

typedef struct _ExecMod {
    struct _ExecMod *next;
    MemManager_ShMem *codeROShMem;
    unsigned int codeROSize;
    unsigned int utilCounter;   /* utilization counter */
    unsigned int startBrk;  /* start of heap */
    unsigned int stackSize; /* size of thread stack in bytes */
    char md5[16];
} ExecMod;

#define BAD_EXIT_TYPE 0xff

typedef struct {
    void *exception;
    char **argv;
    char **envp;
    ExecMod *exec;
    struct _LibModule *libs;    /* libs in use */
    int exitCode;
    int exitType;
} TaskInfo;

typedef struct {
    AVLTree_In_Node node;
    SysModManager_Imp *impID;
    unsigned int function;
    char *broker;
} SysFunction;

typedef struct _Interface {
    AVLTree_In_Node node;
    char *name;
    unsigned int nFunctions;        /* # functions */
    struct _Implementation *imps;   /* implementation list */
} Interface;

typedef struct _Implementation {
    struct _Implementation *next;
    struct _LibModule *module;
    char *name;
    unsigned short interN;
    unsigned short impN;
    struct _Function *functions;
} Implementation;

#define NOT_IMP_FUNCTION 0x01

typedef struct _Function {
    char *code;
    int prop;
    /* 0x01 - not implemented */
} Function;

typedef struct _LibModule {
    struct _LibModule *next;
    unsigned int utilCounter;
    unsigned int codeStart; /* = 0xffffffff module doesn't has a position */
    unsigned int dataStart;
    unsigned int codeROSize;  /* in bytes */
    unsigned int dataSize;    /* in bytes */
    unsigned int startF;
    unsigned int shutdownF;
    int isLoaded;
    char *path;
    char md5[16];
} LibModule;

typedef struct {
    char *dummy;
} Library;

typedef struct _RMap {
    struct _RMap *next;
    unsigned int start, size;
} RMap;

void Broker(void);
int CmpInterface(const void*,const AVLTree_In_Node*);
int CmpAddInterface(const AVLTree_In_Node*,const AVLTree_In_Node*);
int CmpSysFuncs(const void*,const AVLTree_In_Node*);
int CmpAddSysFuncs(const AVLTree_In_Node*,const AVLTree_In_Node*);
int FillBrokersPage(void*,void*,unsigned int);
int GetLibFunctionPos(const char*,const char*,unsigned int,unsigned int*);
int GetSysFunctionPos(const char*,const char*,unsigned int,unsigned int*);
int LoadLibrary(TaskInfo*,LibModule*);
int SetLibraryPos(LibModule*,Interface*,Implementation*);
int StartFunctionUtil(Interface*,Implementation*,unsigned int);
int SysCall(IntManager_i386_Regs*);

SysModManager_Module *ID;
TaskManager_Semaphore modMutex = {1, NULL};
ExecMod *modules;
/* must be used to access libraries or interfaces */
TaskManager_Semaphore libMutex = {1, NULL};
AVLTree_In_Tree interfaces = {NULL, CmpInterface, CmpAddInterface, 0};
LibModule *libraries;
AVLTree_In_Tree sysFuncs = {NULL, CmpSysFuncs, (int (*)(const AVLTree_In_Node*,const AVLTree_In_Node*))CmpSysFuncs, 0};
MemManager_Cache *sysFuncCache;

/* number of logical pages for sysCallTable */
#define NUMBER_SYS_CALL_PAGES 0x200
unsigned int nextSysCall;
char *sysCallTable;

MemManager_ShMemFunctions brokersShFunctions = {FillBrokersPage};
MemManager_ShMem *brokersShMem;
MemManager_ShMem *libCodeShMem;

MemManager_ShMemFunctions nullShFunctions;

/* keeps free library's data memory areas */
RMap *dataArea;
/* keeps free library's code memory areas */
RMap *codeArea;

int ArrayBound(IntManager_i386_ExRegs *regs) {
    TaskInfo *taskInfo;
    int ret;

    /* task running an user module? */
    taskInfo = TaskManager_GetTaskInfo(NULL, ID);
    if (IS_PTR_ERROR(taskInfo))
        return E_UNHANDLED_INT;

    if (taskInfo->exception) {
        int tmp;
        regs->ESP -= 4;
        /* put code address where exception occurred */
        if (UserSpaceManager_CopyToUser((char*)regs->ESP, (char*)&regs->EIP, sizeof(int)))
           goto errorUserTask;
        tmp = 0x02;
        regs->ESP -= 4;
        /* put exception code */
        if (UserSpaceManager_CopyToUser((char*)regs->ESP, (char*)&tmp, sizeof(int)))
           goto errorUserTask;
        /* ok, now return to user mode */
        regs->EIP = (unsigned int)taskInfo->exception;
        return E_HANDLED_INT;
    }
errorUserTask:
    taskInfo->exitCode = UserModManager_CHILD_DUMPED_BOUND;
    taskInfo->exitType = UserModManager_CHILD_DUMPED;

    DEBUG_MSG(("Caught array bound exceeded exception.\n At = 0x%x\n", regs->EIP));
    /* Exit expect an undestroyable thread */
    TaskManager_UndestroyableThread();
    ret = UserSpaceManager_Exit(NULL, UserSpaceManager_TASK_DUMPED, UserSpaceManager_TASK_DUMPED_BOUND);
    ASSERT_ERROR(ret);
    return E_HANDLED_INT;
}

/* libmutex must be locked before calling this function */
int CheckParameters(const char *interface, const char *imp, unsigned int function, Interface **intF, Implementation **impF, Function **funcF) {
    int error;
    Implementation *tmpImp;
    Interface *interF;

    if (!interface)
        return E_BAD_VALUE;
    interF = (Interface*)AVLTree_In_Find(&interfaces, interface);
    if (IS_PTR_ERROR(interF))
        return PTR_TO_ERROR(interF);
    tmpImp = interF->imps;
    /* find a specific implementation */
    if (imp) {
        do {
            if (!StringMan_Compare(tmpImp->name, imp))
                break;
            tmpImp = tmpImp->next;
        } while (tmpImp);
        if (!tmpImp)
            return E_NOT_FOUND;
    }
    error = StartFunctionUtil(interF, tmpImp, function);
    if (error)
        return error;
    /* check if function is implemented */
    if (tmpImp->functions[function].prop & NOT_IMP_FUNCTION)
        return E_NOT_IMPLEMENTED;
    *intF = interF;
    *impF = tmpImp;
    *funcF = &tmpImp->functions[function];
    return E_OK;
}

int CmpAddInterface(const AVLTree_In_Node *addNode, const AVLTree_In_Node *node) {
	return StringMan_Compare(((Interface*)addNode)->name, ((Interface*)node)->name);
}

int CmpInterface(const void *info, const AVLTree_In_Node *node) {
	return StringMan_Compare((char*)info, ((Interface*)node)->name);
}

int CmpSysFuncs(const void *info, const AVLTree_In_Node *node) {
    int ret;
    ret = (int)((SysFunction*)info)->impID - (int)((SysFunction*)node)->impID;
    if (!ret)
        ret = (int)(((SysFunction*)info)->function - ((SysFunction*)node)->function);
    return ret;
}

int DivideByZero(IntManager_i386_ExRegs *regs) {
    TaskInfo *taskInfo;
    int ret;

    /* task running an user module? */
    taskInfo = TaskManager_GetTaskInfo(NULL, ID);
    if (IS_PTR_ERROR(taskInfo))
        return E_UNHANDLED_INT;

    if (taskInfo->exception) {
        int tmp;
        regs->ESP -= 4;
        /* put code address where exception occurred */
        if (UserSpaceManager_CopyToUser((char*)regs->ESP, (char*)&regs->EIP, sizeof(int)))
           goto errorUserTask;
        tmp = 0x00;
        regs->ESP -= 4;
        /* put exception code */
        if (UserSpaceManager_CopyToUser((char*)regs->ESP, (char*)&tmp, sizeof(int)))
           goto errorUserTask;
        /* ok, now return to user mode */
        regs->EIP = (unsigned int)taskInfo->exception;
        return E_HANDLED_INT;
    }
errorUserTask:
    taskInfo->exitCode = UserModManager_CHILD_DUMPED_DIV0;
    taskInfo->exitType = UserModManager_CHILD_DUMPED;

    DEBUG_MSG(("Caught divide by zero exception.\n At = 0x%x\n", regs->EIP));
    /* Exit expect an undestroyable thread */
    TaskManager_UndestroyableThread();
    ret = UserSpaceManager_Exit(NULL, UserSpaceManager_TASK_DUMPED, UserSpaceManager_TASK_DUMPED_DIV0);
    ASSERT_ERROR(ret);
    return E_HANDLED_INT;
}

int Exit(TaskManager_Task *task, int exitCode) {
    TaskInfo *info;

    info = TaskManager_GetTaskInfo(task, ID);
    if (IS_PTR_ERROR(info))
        return E_BAD_VALUE;

    info->exitCode = exitCode;
    info->exitType = UserModManager_CHILD_EXITED;
    return UserSpaceManager_Exit(task, UserSpaceManager_TASK_EXITED, exitCode);
}

int FillBrokersPage(void *arg, void *page, unsigned int offset) {
    char *broker;
    unsigned int sysCallNumber;

    for (broker = (char*)page, sysCallNumber = offset >> 3; broker < (char*)page + 0x1000; broker += SIZE_BROKER, sysCallNumber++) {
        /* copy broker function */
        MemMan_Copy(broker, Broker, SIZE_BROKER);
        /* store sys call number */
        *(unsigned int*)(broker+1) = sysCallNumber;
    }
    return 0;
}

/* libmutex must be locked before calling this function */
int GetLibFunctionPos(const char *interP, const char *impP, unsigned int funcP, unsigned int *pos) {
    int error;
    Interface *interface;
    Implementation *imp;
    Function *function;

    error = CheckParameters(interP, impP, funcP, &interface, &imp, &function);
    if (error)
        return error;
    error = SetLibraryPos(imp->module, interface, imp);
    if (error)
        return error;
    *pos = (unsigned int)function->code;
    return error;
}

int GetSysFunctionPos(const char *inter, const char *imp, unsigned int func, unsigned int *pos) {
    int error, ret;
    SysModManager_Imp *funcImp;
    SysFunction *sysFunc, sysFunc2Find;

    /* try system function */
    funcImp = SysModManager_GetImpID(inter, imp);
    if (IS_PTR_ERROR(funcImp))
        return E_MODULE;
    sysFunc2Find.impID = funcImp;
    sysFunc2Find.function = func;
    sysFunc = (SysFunction*)AVLTree_In_Find(&sysFuncs, &sysFunc2Find);
    if (IS_PTR_ERROR(sysFunc)) {
        void *function;
        int nParams;

        function = SysModManager_GetFunction(funcImp, func, SysModManager_USER_FUNCTION, &nParams);
        if (!IS_PTR_ERROR(function)) {
            /* do we need to setup more pages for sysCallTable? */
            /* each page (4096 bytes) has 512 (0x200) syscalls */
            /* FIXME: free resources if this fails */
            if (!(nextSysCall & 0x1ff)) {
                unsigned int physPg;
                void *voidRet;

                physPg = MemManager_i386_AllocPhysPages(1, 0);
                if (IS_PTR_ERROR(physPg)) {
                    error = PTR_TO_ERROR(physPg);
                    return error;
                }
                voidRet = MemManager_i386_MapPhysPages(sysCallTable+(nextSysCall<<3), physPg, 1, MemManager_i386_RW_PAGE);
                if (IS_PTR_ERROR(voidRet)) {
                    ret = MemManager_i386_FreePhysPages(physPg, 1);
                    ASSERT_ERROR(ret);
                    return PTR_TO_ERROR(voidRet);
                }
            }
            /* do we need more broker pages? */
            if (!(nextSysCall & 0x1ff)) {
                error = MemManager_ProtectShMemory(brokersShMem, (void*)(START_BROKERS+(nextSysCall<<3)), 0x1000, MemManager_PROT_EXEC+MemManager_PROT_READ+MemManager_ONLY_SYS+MemManager_CHANGE_SYS, NULL);
                ASSERT_ERROR(error);
                if (error)
                    return error;
            }
            sysFunc = MemManager_AllocObj(sysFuncCache);
            if (IS_PTR_ERROR(sysFunc))
                return PTR_TO_ERROR(sysFunc);
            *(void**)(sysCallTable+(nextSysCall<<3)) = function;
            *(int*)(sysCallTable+(nextSysCall<<3)+4) = nParams;
            sysFunc->impID = funcImp;
            sysFunc->function = func;
            sysFunc->broker = (char*)(START_BROKERS+(nextSysCall*SIZE_BROKER));
            ret = AVLTree_In_Add(&sysFuncs, (AVLTree_In_Node*)sysFunc);
            ASSERT_ERROR(ret);
            nextSysCall++;
        } else {
            ret = SysModManager_PutImpID(funcImp);
            ASSERT_ERROR(ret);
            return E_NOT_FOUND;
        }
    } else {
        ret = SysModManager_PutImpID(funcImp);
        ASSERT_ERROR(ret);
    }
    *pos = (unsigned int)sysFunc->broker;
    return E_OK;
}

int InvalidInstruction(IntManager_i386_ExRegs *regs) {
    TaskInfo *taskInfo;
    int ret;

    /* task running an user module? */
    taskInfo = TaskManager_GetTaskInfo(NULL, ID);
    if (IS_PTR_ERROR(taskInfo))
        return E_UNHANDLED_INT;

    if (taskInfo->exception) {
        int tmp;
        regs->ESP -= 4;
        /* put code address where exception occurred */
        if (UserSpaceManager_CopyToUser((char*)regs->ESP, (char*)&regs->EIP, sizeof(int)))
           goto errorUserTask;
        tmp = 0x03;
        regs->ESP -= 4;
        /* put exception code */
        if (UserSpaceManager_CopyToUser((char*)regs->ESP, (char*)&tmp, sizeof(int)))
           goto errorUserTask;
        /* ok, now return to user mode */
        regs->EIP = (unsigned int)taskInfo->exception;
        return E_HANDLED_INT;
    }
errorUserTask:
    taskInfo->exitCode = UserModManager_CHILD_DUMPED_ILINST;
    taskInfo->exitType = UserModManager_CHILD_DUMPED;

    DEBUG_MSG(("Caught invalid instruction exception.\n At = 0x%x\n", regs->EIP));
    /* Exit expect an undestroyable thread */
    TaskManager_UndestroyableThread();
    ret = UserSpaceManager_Exit(NULL, UserSpaceManager_TASK_DUMPED, UserSpaceManager_TASK_DUMPED_ILINST);
    ASSERT_ERROR(ret);
    return E_HANDLED_INT;
}

/* libmutex must be locked before calling this function */
int LoadLibrary(TaskInfo *taskInfo, LibModule *module) {
    int error, ret;
    LibModule *currLib;

    /* check if task is already using lib */
    currLib = taskInfo->libs;
    while (currLib && MemMan_Compare(currLib->md5, module->md5, sizeof(module->md5)))
        currLib = currLib->next;
    if (!currLib) {
        unsigned int i;
        RLibModFile_File *file;
        RLibModFile_UsedFuncInfo usedFunc;
        char *roDataStart;
        RLibModFile_ModuleInfo modInfo;

        error = RLibModFile_Open(module->path, &file);
        if (error)
            goto errorEnd;
        error = RLibModFile_GetModuleInfo(file, &modInfo);
        if (error)
            goto errorCloseFile;
        roDataStart = (char*)module->codeStart+ALIGN_DATA_UP(modInfo.codeSize, 4);
        /* do we have to load the library code? */
        if (!module->isLoaded) {
            RLibModFile_UsedFuncRelocInfo usedFuncReloc;

            error = MemManager_ProtectShMemory(libCodeShMem, (void*)module->codeStart, module->codeROSize, MemManager_PROT_WRITE+MemManager_PROT_READ+MemManager_ONLY_SYS+MemManager_CHANGE_SYS, NULL);
            if (error)
                goto errorCloseFile;
            error = RLibModFile_LoadCode(file, 0, modInfo.codeSize, (char*)module->codeStart, (char*)module->codeStart, roDataStart, (char*)module->dataStart);
            if (error)
                goto errorCloseFile;
            if (modInfo.roDataSize) {
                error = RLibModFile_LoadROData(file, 0, modInfo.roDataSize, roDataStart, (char*)module->codeStart, roDataStart, (char*)module->dataStart);
                if (error)
                    goto errorCloseFile;
            }
            /* do used function relocation */
            i = 0;
            while ((error = RLibModFile_GetUsedFuncReloc(file, i++, &usedFuncReloc)) == 0) {
                unsigned int value;
                if (usedFuncReloc.function.prop) {
                    error = GetSysFunctionPos(usedFuncReloc.function.interface, usedFuncReloc.function.imp, usedFuncReloc.function.function, &value);
                } else {
                    TaskManager_P(&libMutex, 0, 0);
                    error = GetLibFunctionPos(usedFuncReloc.function.interface, usedFuncReloc.function.imp, usedFuncReloc.function.function, &value);
                    TaskManager_V(&libMutex, 0);
                }
                if (error)
                    goto errorCloseFile;
                /* relative or absolute relocation? */
                if (usedFuncReloc.prop)
                    *(unsigned int*)(usedFuncReloc.offset+module->codeStart) = value;
                else
                    *(unsigned int*)(usedFuncReloc.offset+module->codeStart) = value-(usedFuncReloc.offset+module->codeStart+4);
            }
            if (error != E_BAD_INDEX)
                goto errorCloseFile;
            module->isLoaded = 1;
        }
        currLib = MemManager_Alloc(sizeof(LibModule), 0);
        if (IS_PTR_ERROR(currLib)) {
            error = PTR_TO_ERROR(currLib);
            goto errorCloseFile;
        }
        MemMan_Copy(currLib, module, sizeof(LibModule));
        currLib->next = taskInfo->libs;
        taskInfo->libs = currLib;
        /* change code and RO data pages properties */
        error = MemManager_ProtectShMemory(libCodeShMem, (void*)module->codeStart, module->codeROSize, MemManager_PROT_EXEC+MemManager_PROT_READ+MemManager_ONLY_SYS+MemManager_CHANGE_SYS, NULL);
        if (error)
            goto errorCloseFile;
        /* do we have to load library's data? */
        if (module->dataSize) {
            /* make pages read/write */
            error = MemManager_SetUserPagesProt((void*)module->dataStart, module->dataSize, MemManager_PROT_WRITE+MemManager_PROT_READ+MemManager_CHANGE_SYS, NULL);
            if (error)
                goto errorCloseFile;
            error = RLibModFile_LoadData(file, 0, modInfo.dataSize, (char*)module->dataStart, (char*)module->codeStart, roDataStart, (char*)module->dataStart);
            if (error)
                goto errorCloseFile;
        }
        /* load used libraries */
        i = 0;
        while ((error = RLibModFile_GetUsedFunc(file, i++, &usedFunc)) == 0) {
            if (!usedFunc.prop) {
                Interface *interface;
                Implementation *imp;
                Function *function;

                /* search for a user module */
                if (CheckParameters(usedFunc.interface, usedFunc.imp, usedFunc.function, &interface, &imp, &function) ||
                    LoadLibrary(taskInfo, imp->module)) {
                        error = E_NOT_FOUND;
                        break;
                    }
            }
        }
        if (error == E_BAD_INDEX)
            error = E_OK;
errorCloseFile:
        ret = RLibModFile_Close(file);
        ASSERT_ERROR(ret);
        if (error)
            goto errorEnd;
    }
    error = E_OK;
errorEnd:
    return error;
}

int Overflow(IntManager_i386_ExRegs *regs) {
    TaskInfo *taskInfo;
    int ret;

    /* task running an user module? */
    taskInfo = TaskManager_GetTaskInfo(NULL, ID);
    if (IS_PTR_ERROR(taskInfo))
        return E_UNHANDLED_INT;

    if (taskInfo->exception) {
        int tmp;
        regs->ESP -= 4;
        /* put code address where exception occurred */
        if (UserSpaceManager_CopyToUser((char*)regs->ESP, (char*)&regs->EIP, sizeof(int)))
           goto errorUserTask;
        tmp = 0x01;
        regs->ESP -= 4;
        /* put exception code */
        if (UserSpaceManager_CopyToUser((char*)regs->ESP, (char*)&tmp, sizeof(int)))
           goto errorUserTask;
        /* ok, now return to user mode */
        regs->EIP = (unsigned int)taskInfo->exception;
        return E_HANDLED_INT;
    }
errorUserTask:
    taskInfo->exitCode = UserModManager_CHILD_DUMPED_INTO;
    taskInfo->exitType = UserModManager_CHILD_DUMPED;

    DEBUG_MSG(("Caught overflow exception.\n At = 0x%x\n", regs->EIP));
    /* Exit expect an undestroyable thread */
    TaskManager_UndestroyableThread();
    ret = UserSpaceManager_Exit(NULL, UserSpaceManager_TASK_DUMPED, UserSpaceManager_TASK_DUMPED_INTO);
    ASSERT_ERROR(ret);
    return E_HANDLED_INT;
}

int Register(const char *pathname, int prop) {
    LibModule *currL;
    LibModule *module;
    Interface *currInt;
    Implementation *currImp;
    RLibModFile_File *file;
    RLibModFile_InterInfo intInfo;
    RLibModFile_ImpInfo impInfo;
    RLibModFile_ModuleInfo modInfo;
    int error, ret;
    unsigned int nInter, nImp;

    error = E_REGISTERED;
    /* only one thread can register a module */
    TaskManager_P(&libMutex, 0, 0);
    currL = libraries;
    while (currL && StringMan_Compare(currL->path, pathname))
        currL = currL->next;
    if (currL)
        goto errorEnd;
    error = RLibModFile_Open(pathname, &file);
    if (error)
        goto errorEnd;
    error = RLibModFile_GetModuleInfo(file, &modInfo);
    if (error)
        goto errorGetMod;
    module = MemManager_Alloc(sizeof(LibModule), 0);
    if (IS_PTR_ERROR(module)) {
        error = PTR_TO_ERROR(module);
        goto errorGetMod;
    }
    module->path = MemManager_Alloc(StringMan_Length(pathname)+1, 0);
    if (IS_PTR_ERROR(module->path)) {
        error = PTR_TO_ERROR(module->path);
        goto errorAllocPath;
    }
    StringMan_Copy(module->path, pathname);
    MemMan_Copy(module->md5, modInfo.md5, 16);
    module->utilCounter = 0;
    module->isLoaded = 0;
    module->codeStart = 0xffffffff;
    module->dataStart = 0xffffffff;
    module->startF = (unsigned int)modInfo.start;
    module->shutdownF = (unsigned int)modInfo.shutdown;
    /* align in 4 bytes */
    module->codeROSize = ALIGN_DATA_UP(modInfo.codeSize, 4);
    module->codeROSize += modInfo.roDataSize;
    /* align in 4096 bytes */
    module->codeROSize = ALIGN_DATA_UP(module->codeROSize, 0x1000);
    /* align in 4096 bytes */
    module->dataSize = ALIGN_DATA_UP(modInfo.dataSize, 0x1000);
    module->next = libraries;
    libraries = module;
    nInter = 0;
    do {
        error = RLibModFile_GetInterInfo(file, nInter, &intInfo);
        if (error) {
            if (error != E_BAD_INDEX)
                goto errorReg;
            break;
        }
        /* if not found, add it */
        currInt = (Interface *)AVLTree_In_Find(&interfaces, intInfo.name);
        if (IS_PTR_ERROR(currInt)) {
            Interface *interface;
            interface = MemManager_Alloc(sizeof(Interface), 0);
            if (IS_PTR_ERROR(interface)) {
                error = PTR_TO_ERROR(interface);
                goto errorReg;
            }
            interface->name = MemManager_Alloc(StringMan_Length(intInfo.name)+1, 0);
            if (IS_PTR_ERROR(interface->name)) {
                error = PTR_TO_ERROR(interface->name);
                ret = MemManager_Free(interface);
                ASSERT_ERROR(ret);
                goto errorReg;
            }
            StringMan_Copy(interface->name, intInfo.name);
            interface->nFunctions = intInfo.nFunctions;
            interface->imps = NULL;
            currInt = (Interface*)AVLTree_In_FindAdd(&interfaces, &interface->node);
            ASSERT_ERROR_PTR(currInt);
        }
        nImp = 0;
        do {
            Implementation *imp;
            error = RLibModFile_GetImpInfo(file, nInter, nImp, &impInfo);
            if (error) {
                if (error != E_BAD_INDEX)
                    goto errorReg;
                break;
            }
            /* check if we already have this implementation */
            error = E_IN_USE;
            currImp = currInt->imps;
            while (currImp) {
                if (!StringMan_Compare(currImp->name, impInfo.name))
                    goto errorReg;
                currImp = currImp->next;
            }
            imp = MemManager_Alloc(sizeof(Implementation), 0);
            if (IS_PTR_ERROR(imp)) {
                error = PTR_TO_ERROR(imp);
                goto errorReg;
            }
            imp->name = MemManager_Alloc(StringMan_Length(impInfo.name)+1, 0);
            if (IS_PTR_ERROR(imp->name)) {
                error = PTR_TO_ERROR(imp->name);
                ret = MemManager_Free(imp);
                ASSERT_ERROR(ret);
                goto errorReg;
            }
            StringMan_Copy(imp->name, impInfo.name);
            imp->module = module;
            imp->interN = nInter;
            imp->impN = nImp;
            imp->functions = NULL;
            imp->next = currInt->imps;
            currInt->imps = imp;
        } while (++nImp);
    } while (++nInter);
    error = E_OK;
    goto errorGetMod;

errorReg:
    libraries = module->next;
errorAllocPath:
    ret = MemManager_Free(module);
    ASSERT_ERROR(ret);
errorGetMod:
    ret = RLibModFile_Close(file);
    ASSERT_ERROR(ret);
errorEnd:
    TaskManager_V(&libMutex, 0);
    return error;
}

int SegmentationFaultUser(unsigned int where, IntManager_i386_ExRegs *regs) {
    TaskInfo *taskInfo;
    int ret;

    /* task running an user module? */
    taskInfo = TaskManager_GetTaskInfo(NULL, ID);
    if (IS_PTR_ERROR(taskInfo))
        return E_UNHANDLED_INT;

    /* send segmenation exception */
    if (taskInfo->exception) {
        int tmp;
        regs->ESP -= 4;
        /* put address that caused the exception */
        if (UserSpaceManager_CopyToUser((char*)regs->ESP, (char*)&where, sizeof(int)))
           goto errorUserTask;
        tmp = regs->ErrorCode & 0x02;
        regs->ESP -= 4;
        /* put type of access (read (0) or write (2)) */
        if (UserSpaceManager_CopyToUser((char*)regs->ESP, (char*)&tmp, sizeof(int)))
           goto errorUserTask;
        regs->ESP -= 4;
        /* put code address where exception occurred */
        if (UserSpaceManager_CopyToUser((char*)regs->ESP, (char*)&regs->EIP, sizeof(int)))
           goto errorUserTask;
        tmp = 0x05;
        regs->ESP -= 4;
        /* put exception code */
        if (UserSpaceManager_CopyToUser((char*)regs->ESP, (char*)&tmp, sizeof(int)))
           goto errorUserTask;
        /* ok, now return to user mode */
        regs->EIP = (unsigned int)taskInfo->exception;
        return E_HANDLED_INT;
    }
errorUserTask:
    taskInfo->exitCode = UserModManager_CHILD_DUMPED_MEMREF;
    taskInfo->exitType = UserModManager_CHILD_DUMPED;

    DEBUG_MSG(("Caught segmentation fault exception.\nAt = 0x%x where = 0x%x prop = 0x%x\n", regs->EIP, where, regs->ErrorCode));
    /* Exit expect an undestroyable thread */
    TaskManager_UndestroyableThread();
    ret = UserSpaceManager_Exit(NULL, UserSpaceManager_TASK_DUMPED, UserSpaceManager_TASK_DUMPED_MEMREF);
    ASSERT_ERROR(ret);
    return E_HANDLED_INT;
}

/* libmutex must be locked before calling this function */
int SetLibraryPos(LibModule *module, Interface *interface, Implementation *imp) {
    if (module->codeStart == 0xffffffff) {
        RMap *curr, *currC;
        unsigned int codePgs, dataPgs;

        /* if there's no data, this won't change nothing */
        curr = dataArea;
        dataPgs = module->dataSize >> 12;
        while (curr && curr->size < dataPgs)
            curr = curr->next;
        if (!curr)
            return E_NO_MEMORY;
        currC = codeArea;
        codePgs = module->codeROSize >> 12;
        while (currC && currC->size < codePgs)
            currC = currC->next;
        if (!currC)
            return E_NO_MEMORY;
        module->codeStart = currC->start;
        currC->start += module->codeROSize;
        currC->size -= codePgs;
        module->dataStart = curr->start;
        curr->start += module->dataSize;
        curr->size -= dataPgs;
        if (module->startF != 0xffffffff)
            module->startF += module->codeStart;
        if (module->shutdownF != 0xffffffff)
            module->shutdownF += module->codeStart;
    }
    if ((unsigned int)imp->functions[0].code < START_LIB_MOD_CODE) {
        unsigned int i;
        for (i=0; i < interface->nFunctions; i++)
            imp->functions[i].code += module->codeStart;
    }
    return E_OK;
}

int Shutdown(void) {
    return 0;
}

int Phase1Start(SysModManager_Module *modID) {
    CfgManager_Group *baseGroup;
    unsigned int index;
    int ret;

    ID = modID;

    /* start used modules */
    if ((ret = SysModManager_StartModule((unsigned int)AVLTree_In_Add)) ||
        (ret = SysModManager_StartModule((unsigned int)CfgManager_OpenGroup)) ||
        (ret = SysModManager_StartModule((unsigned int)MemManager_Alloc)) ||
        (ret = SysModManager_StartModule((unsigned int)TaskManager_P)) ||
        (ret = SysModManager_StartModule((unsigned int)InterManager_SendEvent)) ||
        (ret = SysModManager_StartModule((unsigned int)IntManager_i386_Catch)) ||
        (ret = SysModManager_StartModule((unsigned int)MemMan_Copy)) ||
        (ret = SysModManager_StartModule((unsigned int)StringMan_Compare)) ||
        (ret = SysModManager_StartModule((unsigned int)RExecModFile_Close)) ||
        (ret = SysModManager_StartModule((unsigned int)RLibModFile_Close))) {
        DEBUG_MSG_PRE(ret);
        goto errorReturn;
    }

    dataArea = MemManager_Alloc(sizeof(RMap), 0);
    if (IS_PTR_ERROR(dataArea)) {
        DEBUG_MSG_PRE(PTR_TO_ERROR(dataArea));
        goto errorReturn;
    }
    codeArea = MemManager_Alloc(sizeof(RMap), 0);
    if (IS_PTR_ERROR(codeArea)) {
        DEBUG_MSG_PRE(PTR_TO_ERROR(codeArea));
        goto errorFreeDataArea;
    }
    dataArea->next = NULL;
    dataArea->start = START_LIB_MOD_DATA;
    dataArea->size = SIZE_LIB_MOD_DATA >> 12;
    codeArea->next = NULL;
    codeArea->start = START_LIB_MOD_CODE;
    codeArea->size = SIZE_LIB_MOD_CODE >> 12;
    sysFuncCache = MemManager_CreateObjCache(sizeof(SysFunction), 0, 0, NULL, NULL);
    if (IS_PTR_ERROR(sysFuncCache)) {
        DEBUG_MSG_PRE(PTR_TO_ERROR(sysFuncCache));
        goto errorFreeCodeArea;
    }
    /* create brokers shared mem */
    brokersShMem = MemManager_CreateShMemory(SIZE_BROKERS, MemManager_SH_CLEAR, &brokersShFunctions, NULL);
    if (IS_PTR_ERROR(brokersShMem)) {
        DEBUG_MSG_PRE(PTR_TO_ERROR(brokersShMem));
        goto errorDestroySysFuncCache;
    }
    /* create syscall table */
    sysCallTable = MemManager_i386_AllocPages(NUMBER_SYS_CALL_PAGES, MemManager_i386_RW_PAGE);
    if (IS_PTR_ERROR(sysCallTable)) {
        DEBUG_MSG_PRE(PTR_TO_ERROR(sysCallTable));
        goto errorDestroyBrokersShMem;
    }
    /* create library code shared mem */
    libCodeShMem = MemManager_CreateShMemory(SIZE_LIB_MOD_CODE, 0, &nullShFunctions, NULL);
    if (IS_PTR_ERROR(libCodeShMem)) {
        DEBUG_MSG_PRE(PTR_TO_ERROR(libCodeShMem));
        goto errorFreeSysCallTable;
    }
    /* catch int that call system functions */
    if ((ret = IntManager_i386_Catch(modID, 0x88, SysCall, 0))) {
        DEBUG_MSG_PRE(ret);
        goto errorDestroyLibCodeShMem;
    }

    /* register exception handlers */
    if ((ret = MemManager_i386_Catch(NULL, NULL, SegmentationFaultUser, MemManager_i386_EX_USER_MODE))) {
        DEBUG_MSG_PRE(ret);
        goto errorFreeSysCallInt;
    }
    if ((ret = IntManager_i386_CatchException(0x00, DivideByZero, (void*)0, (void*)0xdfffffff))) {
        DEBUG_MSG_PRE(ret);
        goto errorFreeSegEx;
    }
    if ((ret = IntManager_i386_CatchException(0x04, Overflow, (void*)0, (void*)0xdfffffff))) {
        DEBUG_MSG_PRE(ret);
        goto errorFreeDivEx;
    }
    if ((ret = IntManager_i386_CatchException(0x05, ArrayBound, (void*)0, (void*)0xdfffffff))) {
        DEBUG_MSG_PRE(ret);
        goto errorFreeOverflowEx;
    }
    if ((ret = IntManager_i386_CatchException(0x06, InvalidInstruction, (void*)0, (void*)0xdfffffff))) {
        DEBUG_MSG_PRE(ret);
        goto errorFreeArrayEx;
    }

    DEBUG_MSG(("Getting UserModManager settings...\n"));

    /* open modules base key */
    baseGroup = CfgManager_OpenGroup(NULL, "/system/UserModManager/modules");
    if (!baseGroup) {
        DEBUG_MSG_PRE(0);
        goto errorFreeInvalidEx;
    }

    index = 0;
    while (1) {
        CfgManager_Group *modGroup;
        char buf[128];
        unsigned int bufSize;

        modGroup = CfgManager_OpenGroupIndex(baseGroup, index);
        if (!modGroup) {
            break;
        }
        /* get module path */
        bufSize = 128;
        if ((ret = CfgManager_GetValue(modGroup, "path", CfgManager_STRING, buf, &bufSize))) {
            SysModManager_DebugMsg("Cannot read path: 0x%x\n", ret);
            goto nextModule;
        }
        ret = Register(buf, 0);
        if (ret && ret != E_REGISTERED) {
            SysModManager_DebugMsg("Error. Can't register module '%s' - 0x%x.\n", buf, ret);
        }
nextModule:
        ret = CfgManager_CloseGroup(modGroup);
        ASSERT_ERROR(ret);
        index++;
    }
    ret = CfgManager_CloseGroup(baseGroup);
    ASSERT_ERROR(ret);

    DEBUG_MSG(("Done.\n"));

    return E_OK;

errorFreeInvalidEx:
    ret = IntManager_i386_FreeException(0x06, InvalidInstruction, (void*)0);
    ASSERT_ERROR(ret);
errorFreeArrayEx:
    ret = IntManager_i386_FreeException(0x05, ArrayBound, (void*)0);
    ASSERT_ERROR(ret);
errorFreeOverflowEx:
    ret = IntManager_i386_FreeException(0x04, Overflow, (void*)0);
    ASSERT_ERROR(ret);
errorFreeDivEx:
    ret = IntManager_i386_FreeException(0x00, DivideByZero, (void*)0);
    ASSERT_ERROR(ret);
errorFreeSegEx:
    ret = MemManager_i386_FreeException(SegmentationFaultUser, (void*)0, 0);
    ASSERT_ERROR(ret);
errorFreeSysCallInt:
    ret = IntManager_i386_Free(modID, 0x88, 0);
    ASSERT_ERROR(ret);
errorDestroyLibCodeShMem:
    ret = MemManager_DestroyShMemory(libCodeShMem);
    ASSERT_ERROR(ret);
errorFreeSysCallTable:
    ret = MemManager_i386_FreePages(sysCallTable, NUMBER_SYS_CALL_PAGES, 0);
    ASSERT_ERROR(ret);
errorDestroyBrokersShMem:
    ret = MemManager_DestroyShMemory(brokersShMem);
    ASSERT_ERROR(ret);
errorDestroySysFuncCache:
    ret = MemManager_DestroyObjCache(sysFuncCache);
    ASSERT_ERROR(ret);
errorFreeCodeArea:
    ret = MemManager_Free(codeArea);
    ASSERT_ERROR(ret);
errorFreeDataArea:
    ret = MemManager_Free(dataArea);
    ASSERT_ERROR(ret);
errorReturn:
    return 1;
}

/* libmutex must be locked before calling this function */
int StartFunctionUtil(Interface *interface, Implementation *imp, unsigned int function) {
    int error, ret;
    RLibModFile_File *file;

    if (!imp->functions) {
        RLibModFile_ModuleInfo modInfo;
        error = RLibModFile_Open(imp->module->path, &file);
        if (error)
            goto errorReturn;
        error = RLibModFile_GetModuleInfo(file, &modInfo);
        if (error)
            goto errorCloseFile;
        error = E_LOADING_MODULE;
        if (MemMan_Compare(imp->module->md5, modInfo.md5, 16))
            goto errorCloseFile;
        imp->functions = MemManager_Alloc(sizeof(Function)*interface->nFunctions, MemManager_BLOCK_MEM);
        if (IS_PTR_ERROR(imp->functions)) {
            error = PTR_TO_ERROR(imp->functions);
            goto errorCloseFile;
        }
        error = RLibModFile_GetFuncsInfo(file, imp->interN, imp->impN, (RLibModFile_FunctionInfo*)imp->functions);
        if (error)
            goto errorFreeFunctions;
        ret = RLibModFile_Close(file);
        ASSERT_ERROR(ret);
    }
    error = E_OK;
    if (function >= interface->nFunctions)
        error = E_BAD_INDEX;
    goto errorReturn;

errorFreeFunctions:
    ret = MemManager_Free(imp->functions);
    ASSERT_ERROR(ret);
errorCloseFile:
    ret = RLibModFile_Close(file);
    ASSERT_ERROR(ret);
errorReturn:
    return error;
}

void UserSpaceManagerE_TaskDestroyed(TaskManager_Task *task, UserSpaceManager_ExitInfo *exitInfo) {
    TaskManager_Task *taskP;
    ExecMod *info;
    TaskInfo *taskInfo, *taskInfoP, tmpTaskInfo;
    int flags, ret;
    LibModule *currLib;

    /* is an user module? */
    taskInfo = TaskManager_DelTaskInfo(task, ID);
    if (IS_PTR_ERROR(taskInfo)) {
        /* do this so we can check parent */
        tmpTaskInfo.exitType = exitInfo->exitType;
        tmpTaskInfo.exitCode = exitInfo->exitCode;
        taskInfo = &tmpTaskInfo;
    }
#ifdef DEBUG_ASSERT
    if (taskInfo->exitType == BAD_EXIT_TYPE) {
        /* something wrong. This shouldn't happen! */
        ASSERT_MSG_PRE(0);
        taskInfo->exitType = UserModManager_CHILD_DUMPED;
        taskInfo->exitCode = UserModManager_CHILD_DUMPED_MEMREF;
    }
#endif
    /* parent is an user module too? */
    taskP = TaskManager_GetParentTaskHandle(task);
    taskInfoP = TaskManager_GetTaskInfo(taskP, ID);
    if (!IS_PTR_ERROR(taskInfoP)) {
        UserModManager_Event event;

        /* send Child event */
        event.uEvent.type = UserModManager_EVENT_CHILD;
        event.uEvent.child.code = taskInfo->exitType;
        event.uEvent.child.exitCode = taskInfo->exitCode;
        event.uEvent.child.pid = TaskManager_GetTaskID(task);
        ret = InterManager_SendEvent(taskP, &event.iEvent);
        ASSERT_ERROR(ret);
    }
    /* because of GetParentTaskHandle() above */
    TaskManager_Destroyable(taskP);
    /* only parent is user module? */
    if (taskInfo == &tmpTaskInfo)
        return;
    currLib = taskInfo->libs;
    while (currLib) {
        LibModule *tmp;

        tmp = currLib->next;
        ret = MemManager_Free(currLib);
        ASSERT_ERROR(ret);
        currLib = tmp;
    }
    info = taskInfo->exec;
    ret = MemManager_Free(taskInfo);
    ASSERT_ERROR(ret);
    TaskManager_SpinLock(&flags);
    info->utilCounter--;
    TaskManager_SpinUnlock(&flags);
}

int Unregister(const char *pathname, int prop) {
    return E_UNSUPPORTED;
}

char *UGetUserStr(const char *uStr) {
    char *page;
    int error, ret;
    unsigned int len;

    if (!uStr)
        return ERROR_TO_PTR(E_BAD_VALUE);
    page = MemManager_i386_AllocPages(1, MemManager_i386_RW_PAGE+MemManager_i386_ALLOC_PHYS);
    if (IS_PTR_ERROR(page))
        return page;
    error = UserSpaceManager_StrNCopy(page, uStr, 4096, &len);
    if (!error) {
        if (len != 4096)
            return page;
        error = E_TOO_BIG;
    }
    ret = MemManager_i386_FreePages(page, 1, 0);
    ASSERT_ERROR(ret);
    return ERROR_TO_PTR(error);
}

static __inline__ void UPutUserStr(char *uStr) {
    int ret;

    ret = MemManager_i386_FreePages(uStr, 1, 0);
    ASSERT_ERROR(ret);
}

int UCatchException(void *function) {
    TaskInfo *taskInfo;

    if ((unsigned int)function >= 0xe0000000)
        return E_BAD_VALUE;
    taskInfo = TaskManager_GetTaskInfo(NULL, ID);
    ASSERT_ERROR_PTR_RET(taskInfo, E_BAD_VALUE);
    taskInfo->exception = function;
    return E_OK;
}

void UExit(int exitCode) {
    Exit(NULL, exitCode);
}

void *UGetFunction(const char *interface, const char *imp, unsigned int function, int type) {
    char *uInterface, *uImp;
    unsigned int value;
    int ret;

    uInterface = UGetUserStr(interface);
    if (IS_PTR_ERROR(uInterface))
        return uInterface;
    uImp = NULL;
    if (imp) {
        uImp = UGetUserStr(imp);
        if (IS_PTR_ERROR(uImp)) {
            ret = PTR_TO_ERROR(uImp);
            goto errorFreeUInterface;
        }
    }
    if (type) {
        ret = GetSysFunctionPos(uInterface, uImp, function, &value);
    } else {
        TaskManager_P(&libMutex, 0, 0);
        ret = GetLibFunctionPos(uInterface, uImp, function, &value);
        /* load the library */
        if (!ret) {
            Interface *interface;
            Implementation *imp;
            Function *func;
            TaskInfo *taskInfo;

            taskInfo = TaskManager_GetTaskInfo(NULL, ID);
#ifdef DEBUG_ASSERT
            if (IS_PTR_ERROR(taskInfo)) {
                ret = PTR_TO_ERROR(taskInfo);
                ASSERT_MSG_PRE(ret);
                goto errorVLib;
            }
#endif
            /* search for a user module */
            ret = CheckParameters(uInterface, uImp, function, &interface, &imp, &func);
            if (ret) {
                goto errorVLib;
            }
            ret = LoadLibrary(taskInfo, imp->module);
        }
errorVLib:
        TaskManager_V(&libMutex, 0);
    }

    if (imp) {
        UPutUserStr(uImp);
    }
errorFreeUInterface:
    UPutUserStr(uInterface);
    if (ret)
        return ERROR_TO_PTR(ret);
    return (void*)value;
}

void *UGetLibFunction(Library *lib, const char *interface, const char *imp, unsigned int function) {
    return ERROR_TO_PTR(E_UNSUPPORTED);
}

void *UGetShutdownFunctionI(unsigned int index) {
    LibModule *curr;
    TaskInfo *taskInfo;

    taskInfo = TaskManager_GetTaskInfo(NULL, ID);
    ASSERT_ERROR_PTR_RET(taskInfo, NULL);
    curr = taskInfo->libs;
    while (curr && index--)
        curr = curr->next;
    if (!curr)
        return NULL;
    return (void*)curr->shutdownF;
}

unsigned int UGetStackSize(void) {
    TaskInfo *taskInfo;

    taskInfo = TaskManager_GetTaskInfo(NULL, ID);
    ASSERT_ERROR_PTR_RET(taskInfo, 0);
    return taskInfo->exec->stackSize;
}

void *UGetStartFunction(unsigned int function) {
    LibModule *curr;
    TaskInfo *taskInfo;

    taskInfo = TaskManager_GetTaskInfo(NULL, ID);
    ASSERT_ERROR_PTR_RET(taskInfo, NULL);
    curr = taskInfo->libs;
    while (curr && (function < curr->codeStart || function >= (curr->codeStart+curr->codeROSize)))
        curr = curr->next;
    if (!curr || curr->startF == 0xffffffff)
        return NULL;
    return (void*)curr->startF;
}

void *UGetStartFunctionI(unsigned int index) {
    LibModule *curr;
    TaskInfo *taskInfo;

    taskInfo = TaskManager_GetTaskInfo(NULL, ID);
    ASSERT_ERROR_PTR_RET(taskInfo, NULL);
    curr = taskInfo->libs;
    while (curr && index--)
        curr = curr->next;
    if (!curr)
        return NULL;
    return (void*)curr->startF;
}

int ULoad(const char *pathname, Library **lib) {
    return E_UNSUPPORTED;
}

int UUnload(Library *lib) {
    return E_UNSUPPORTED;
}

int ExecLoader_CheckExec(UserSpaceManager_ExecInfo *exec) {
    if (RExecModFile_CheckFormat(exec->fileBuf, sizeof(exec->fileBuf)))
        return E_UNSUPPORTED;
    return E_OK;
}

int ExecLoader_LoadExec(UserSpaceManager_ExecInfo *exec, TaskManager_Regs *regs) {
    int error, ret, tmpInt;
    char *newStack, *stackTop, *argStart;
    unsigned int codeROSize, i, argv, envp, argc, envc;
    RExecModFile_UsedFuncInfo usedFunc;
    RExecModFile_ModuleInfo modInfo;
    RExecModFile_File *modFile;
    ExecMod *currM;
    TaskInfo *taskInfo;
    void *tmp;

    error = RExecModFile_OpenFD(exec->fd, RExecModFile_DO_NOT_CLOSE, &modFile);
    if (error)
        goto errorReturn;
    error = RExecModFile_GetModuleInfo(modFile, &modInfo);
    if (error)
        goto errorCloseFile;
    /* setup guard page for NULL pointers */
    tmp = MemManager_AllocUserPages((void*)0, 0x1000, 0, MemManager_PROT_NONE+MemManager_ONLY_SYS+MemManager_FIXED_ADDR, NULL);
    if (IS_PTR_ERROR(tmp)) {
        error = PTR_TO_ERROR(tmp);
        goto errorCloseFile;
    }
    /* attach shared mem with brokers */
    tmp = MemManager_AttachShMemory(brokersShMem, (void*)START_BROKERS, 0, SIZE_BROKERS, MemManager_PROT_NONE+MemManager_ONLY_SYS+MemManager_FIXED_ADDR, NULL);
    if (IS_PTR_ERROR(tmp)) {
        error = PTR_TO_ERROR(tmp);
        goto errorCloseFile;
    }
    /* make used brokers available */
    error = MemManager_ProtectShMemory(brokersShMem, (void*)START_BROKERS, ALIGN_DATA_UP((nextSysCall+1)<<3, 0x1000), MemManager_PROT_EXEC+MemManager_PROT_READ+MemManager_ONLY_SYS+MemManager_CHANGE_SYS, NULL);
    if (error)
        goto errorCloseFile;
    /* setup guard page after brokers */
    tmp = MemManager_AllocUserPages((void*)(START_EXEC_MOD-0x1000), 0x1000, 0, MemManager_PROT_NONE+MemManager_ONLY_SYS+MemManager_FIXED_ADDR, NULL);
    if (IS_PTR_ERROR(tmp)) {
        error = PTR_TO_ERROR(tmp);
        goto errorCloseFile;
    }
    TaskManager_P(&modMutex, 0, 0);
    /* check if module is already loaded */
    currM = modules;
    while (currM && MemMan_Compare(modInfo.md5, currM->md5, sizeof(modInfo.md5)))
        currM = currM->next;
    if (!currM) {
        RExecModFile_UsedFuncRelocInfo usedFuncReloc;
        MemManager_ShMem *codeROShMem;

        /* align RO data in 4 bytes */
        codeROSize = ALIGN_DATA_UP(modInfo.codeSize, 4);
        codeROSize += modInfo.roDataSize;
        codeROSize = ALIGN_DATA_UP(codeROSize, 0x400000);
        /* create shared mem for code and RO data */
        codeROShMem = MemManager_CreateShMemory(codeROSize, MemManager_SH_PDE, &nullShFunctions, NULL);
        if (IS_PTR_ERROR(codeROShMem)) {
            error = PTR_TO_ERROR(codeROShMem);
            goto errorVModMutex;
        }
        /* attach code and RO data shared mem */
        tmp = MemManager_AttachShMemory(codeROShMem, (void*)START_EXEC_MOD, 0, codeROSize, MemManager_PROT_WRITE+MemManager_PROT_READ+MemManager_ONLY_SYS+MemManager_FIXED_ADDR, NULL);
        if (IS_PTR_ERROR(tmp)) {
            error = PTR_TO_ERROR(tmp);
            goto errorDestroyCodeShMem;
        }
        error = RExecModFile_LoadCode(modFile, 0, modInfo.codeSize, (char*)START_EXEC_MOD);
        if (error)
            goto errorDetachShMem;
        if (modInfo.roDataSize) {
            error = RExecModFile_LoadROData(modFile, 0, modInfo.roDataSize, (char*)START_EXEC_MOD+ALIGN_DATA_UP(modInfo.codeSize, 4));
            if (error)
                goto errorDetachShMem;
        }
        /* do used function relocation */
        i = 0;
        while ((error = RExecModFile_GetUsedFuncReloc(modFile, i++, &usedFuncReloc)) == 0) {
            unsigned int value;
            if (usedFuncReloc.function.prop) {
                error = GetSysFunctionPos(usedFuncReloc.function.interface, usedFuncReloc.function.imp, usedFuncReloc.function.function, &value);
            } else {
                TaskManager_P(&libMutex, 0, 0);
                error = GetLibFunctionPos(usedFuncReloc.function.interface, usedFuncReloc.function.imp, usedFuncReloc.function.function, &value);
                TaskManager_V(&libMutex, 0);
            }
            if (error)
                goto errorDetachShMem;
            /* relative or absolute relocation? */
            if (usedFuncReloc.prop)
                *(unsigned int*)(usedFuncReloc.offset+START_EXEC_MOD) = value;
            else
                *(unsigned int*)(usedFuncReloc.offset+START_EXEC_MOD) = value-(usedFuncReloc.offset+START_EXEC_MOD+4);
        }
        if (error != E_BAD_INDEX)
           goto errorDetachShMem;
        /* change code and RO data pages properties */
        error = MemManager_ProtectShMemory(codeROShMem, (void*)START_EXEC_MOD, codeROSize, MemManager_PROT_EXEC+MemManager_PROT_READ+MemManager_ONLY_SYS+MemManager_CHANGE_SYS, NULL);
        if (error)
            goto errorDetachShMem;
        currM = MemManager_Alloc(sizeof(ExecMod), 0);
        if (IS_PTR_ERROR(currM)) {
            error = PTR_TO_ERROR(currM);
errorDetachShMem:
            ret = MemManager_DetachShMemory(codeROShMem, (void*)START_EXEC_MOD, codeROSize, MemManager_CHANGE_SYS, NULL);
            ASSERT_ERROR(ret);
errorDestroyCodeShMem:
            ret = MemManager_DestroyShMemory(codeROShMem);
            ASSERT_ERROR(ret);
            goto errorVModMutex;
        }
        currM->codeROShMem = codeROShMem;
        currM->codeROSize = codeROSize;
        currM->startBrk = START_EXEC_MOD + codeROSize + ALIGN_DATA_UP(modInfo.dataSize, 0x1000);
        if (!modInfo.stackSize)
            currM->stackSize = DEFAULT_STACK_SIZE;
        else if (modInfo.stackSize < MIN_STACK_SIZE_2)
            currM->stackSize = 1 << MIN_STACK_SIZE_2;
        else
            currM->stackSize = 1 << modInfo.stackSize;
        currM->utilCounter = 1;
        MemMan_Copy(currM->md5, modInfo.md5, 16);
        /* insert in register */
        currM->next = modules;
        modules = currM;
    } else {
        /* attach code and RO data shared mem */
        tmp = MemManager_AttachShMemory(currM->codeROShMem, (void*)START_EXEC_MOD, 0, currM->codeROSize, MemManager_PROT_EXEC+MemManager_PROT_READ+MemManager_ONLY_SYS+MemManager_FIXED_ADDR, NULL);
        if (IS_PTR_ERROR(tmp)) {
            error = PTR_TO_ERROR(tmp);
            goto errorVModMutex;
        }
        currM->utilCounter++;
    }
    TaskManager_V(&modMutex, 0);
    if (modInfo.dataSize) {
        unsigned int dataSize;
        char *dataStart;

        dataSize = ALIGN_DATA_UP(modInfo.dataSize, 0x1000);
        /* alloc user pages for data */
        dataStart = MemManager_AllocUserPages((void*)(START_EXEC_MOD+currM->codeROSize), dataSize, 0, MemManager_PROT_WRITE+MemManager_PROT_READ+MemManager_FIXED_ADDR, NULL);
        if (IS_PTR_ERROR(dataStart)) {
            error = PTR_TO_ERROR(dataStart);
            goto errorPModMutex;
        }
        error = RExecModFile_LoadData(modFile, 0, modInfo.dataSize, dataStart);
        if (error)
            goto errorPModMutex;
    }

    /* attach library code and ro data */
    tmp = MemManager_AttachShMemory(libCodeShMem, (void*)START_LIB_MOD_CODE, 0, SIZE_LIB_MOD_CODE, MemManager_PROT_NONE+MemManager_ONLY_SYS+MemManager_FIXED_ADDR, NULL);
    if (IS_PTR_ERROR(tmp)) {
        error = PTR_TO_ERROR(tmp);
        goto errorPModMutex;
    }

    /* reserve library dada */
    tmp = MemManager_AllocUserPages((void*)START_LIB_MOD_DATA, SIZE_LIB_MOD_DATA, 0, MemManager_PROT_NONE+MemManager_ONLY_SYS+MemManager_FIXED_ADDR, NULL);
    if (IS_PTR_ERROR(tmp)) {
        error = PTR_TO_ERROR(tmp);
        goto errorPModMutex;
    }

    taskInfo = MemManager_Alloc(sizeof(TaskInfo), 0);
    if (IS_PTR_ERROR(taskInfo)) {
        error = PTR_TO_ERROR(taskInfo);
        goto errorPModMutex;
    }

    /* start task stuff */
    taskInfo->exception = NULL;
    taskInfo->exec = currM;
    taskInfo->libs = NULL;
    /* something wrong if we exit with this code */
    taskInfo->exitType = BAD_EXIT_TYPE;
    if (IS_PTR_ERROR(TaskManager_SetTaskInfo(NULL, ID, taskInfo, 0x01)))
        goto errorFreeTaskInfo;

    /* load all used libraries */
    TaskManager_P(&libMutex, 0, 0);
    i = 0;
    while ((error = RExecModFile_GetUsedFunc(modFile, i++, &usedFunc)) == 0) {
        if (!usedFunc.prop) {
            Interface *interface;
            Implementation *imp;
            Function *function;

            /* search for a user module */
            if (CheckParameters(usedFunc.interface, usedFunc.imp, usedFunc.function, &interface, &imp, &function) ||
                LoadLibrary(taskInfo, imp->module)) {
                    error = E_NOT_FOUND;
                    break;
                }
        }
    }
    TaskManager_V(&libMutex, 0);
    if (error != E_BAD_INDEX)
        goto errorDelTaskInfo;

    /* alloc stack space */
    newStack = MemManager_AllocUserPages((void*)0, currM->stackSize, currM->stackSize, MemManager_PROT_WRITE+MemManager_PROT_READ, NULL);
    if (IS_PTR_ERROR(newStack)) {
        error = PTR_TO_ERROR(newStack);
        goto errorDelTaskInfo;
    }
    /* map stack physical pages */
    error = UserSpaceManager_MapStackPages(exec, newStack, currM->stackSize);
    if (error)
        goto errorDelTaskInfo;
    argStart = (char*)exec->stackTop;
    /* store arguments for main function */
    stackTop = (char*)ALIGN_DATA_DOWN(exec->stackTop, 4);
    stackTop -= (exec->envc+1) * sizeof(int);
    envp = (unsigned int)stackTop;
    stackTop -= (exec->argc+1) * sizeof(int);
    argv = (unsigned int)stackTop;
    /* envp, argv, argc and NULL for return address of main function */
    stackTop -= 4 * sizeof(int);
    exec->stackTop = (unsigned int)stackTop;
    exec->startBrk = currM->startBrk;
    error = E_TOO_BIG;
    if (stackTop < newStack)
        goto errorDelTaskInfo;
    tmpInt = 0;
    error = UserSpaceManager_CopyToUser(stackTop, (char*)&tmpInt, sizeof(tmpInt));
    if (error)
        goto errorDelTaskInfo;
    stackTop += sizeof(int);
    tmpInt = exec->argc;
    error = UserSpaceManager_CopyToUser(stackTop, (char*)&tmpInt, sizeof(tmpInt));
    if (error)
        goto errorDelTaskInfo;
    stackTop += sizeof(int);
    tmpInt = (int)argv;
    error = UserSpaceManager_CopyToUser(stackTop, (char*)&tmpInt, sizeof(tmpInt));
    if (error)
        goto errorDelTaskInfo;
    stackTop += sizeof(int);
    tmpInt = (int)envp;
    error = UserSpaceManager_CopyToUser(stackTop, (char*)&tmpInt, sizeof(tmpInt));
    if (error)
        goto errorDelTaskInfo;
    stackTop += sizeof(int);
    argc = exec->argc;
    while (argc--) {
        unsigned int len;
        
        tmpInt = (int)argStart;
        error = UserSpaceManager_CopyToUser(stackTop, (char*)&tmpInt, sizeof(tmpInt));
        if (error)
            goto errorDelTaskInfo;
        error = UserSpaceManager_StrLen(argStart, &len);
        if (error)
            goto errorDelTaskInfo;
        stackTop += sizeof(int);
        argStart += len + 1;
    }
    tmpInt = 0;
    error = UserSpaceManager_CopyToUser(stackTop, (char*)&tmpInt, sizeof(tmpInt));
    if (error)
        goto errorDelTaskInfo;
    envc = exec->envc;
    while (envc--) {
        unsigned int len;

        tmpInt = (int)argStart;
        error = UserSpaceManager_CopyToUser(stackTop, (char*)&tmpInt, sizeof(tmpInt));
        if (error)
            goto errorDelTaskInfo;
        error = UserSpaceManager_StrLen(argStart, &len);
        if (error)
            goto errorDelTaskInfo;
        stackTop += sizeof(int);
        argStart += len + 1;
    }
    tmpInt = 0;
    error = UserSpaceManager_CopyToUser(stackTop, (char*)&tmpInt, sizeof(tmpInt));
    if (error)
        goto errorDelTaskInfo;
    /* start all registers */
    regs->eax = 0;
    regs->ebx = 0;
    regs->ecx = 0;
    regs->edx = 0;
    regs->esi = 0;
    regs->edi = 0;
    regs->ebp = 0;
    regs->esp = exec->stackTop;
    regs->eip = START_EXEC_MOD;
    regs->eflags = 0x00000202;
    regs->ds = DPL3DataSegment;
    regs->es = DPL3DataSegment;
    regs->ss = DPL3DataSegment;
    regs->cs = DPL3CodeSegment;
    error = E_OK;
    goto errorCloseFile;

errorDelTaskInfo:
    taskInfo = TaskManager_DelTaskInfo(NULL, ID);
    ASSERT_ERROR_PTR(taskInfo);
errorFreeTaskInfo:
    ret = MemManager_Free(taskInfo);
    ASSERT_ERROR(ret);
errorPModMutex:
    TaskManager_P(&modMutex, 0, 0);
    currM->utilCounter--;
errorVModMutex:
    TaskManager_V(&modMutex, 0);
errorCloseFile:
    ret = RExecModFile_Close(modFile);
    ASSERT_ERROR(ret);
errorReturn:
    return error;
}
