/*
*****************************************
* This file implements the UserSpaceManager interface
*
* Version: 0.1.2
* Author: Luiz Henrique Shigunov
* E-mail: luizshigunov@users.sourceforge.net
* WEB: http://www.sourceforge.net/projects/modulos
* Last change: 18/06/2005
*
* See the interface specification to know what each function does.
*
* Copyright (C) 2003-2005, Luiz Henrique Shigunov
*
* This module is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2.1 of the License, or (at your option) any later version.
*
* This module is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this module; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*****************************************
*/

#include <stddef.h>
#include <cfgmanager.h>
#include <debug.h>
#include <errormanager.h>
#include <execloader.h>
#include <fsmanager.h>
#include <intermanager.h>
#include <memman.h>
#include <memmanager.h>
#include <memmanager_i386.h>
#include <stringman.h>
#include <sysdef.h>
#include <sysmodmanager.h>
#include <taskmanager.h>
#include <taskmanager_i386.h>
#include <userspacemanager.h>

/* number of pages for argv/envp */
#define NUM_ARGS_PG 16

#define ALIGN_DATA_UP(data, align) ((((unsigned int)(data)) + (align)-1) & ~((align)-1))

typedef struct {
    UserSpaceManager_ExitInfo pub;
    unsigned int brk;       /* current heap limit */
    unsigned int startBrk;  /* start of heap */
    TaskManager_Semaphore brkSem;
} TaskInfo;

typedef struct _ExecLoader {
    struct _ExecLoader *next;
    int (*checkExec)(UserSpaceManager_ExecInfo*);
    int (*loadExec)(UserSpaceManager_ExecInfo*,TaskManager_Regs*);
    unsigned int refCounter;
    SysModManager_Imp *imp;
} ExecLoader;

typedef struct {
    UserSpaceManager_ExecInfo pub;
    unsigned int pages[NUM_ARGS_PG];
    unsigned int cachePage;
    char *cache;
    TaskManager_Semaphore sem;
    TaskManager_WaitQueue waitQ;
    ExecLoader *loader;
    int ret;
} ExecInfo;

int CommonCopyUser(char*,const char*,unsigned int);
int CopyFromUser(char*,const char*,unsigned int);
int CopyToUser(char*,const char*,unsigned int);
void LoadExec(ExecInfo*);
int MapStackPages(ExecInfo*,char*,unsigned int);
int PutString(const char*,int,ExecInfo*);
int StrLen(const char*,unsigned int*);
int StrNCopy(char*,const char*,unsigned int,unsigned int*);

extern unsigned int CommonCopyUser_newEIP;
extern unsigned int StrNCopy_newEIP;
extern unsigned int StrLen_newEIP;

SysModManager_Module *ID;

ExecLoader *execLoaders;
TaskManager_RWSync execLoadersRW;

unsigned int Brk(unsigned int brk) {
    TaskInfo *taskInfo;
    unsigned int ret;

    taskInfo = TaskManager_GetTaskInfo(NULL, ID);
    if (IS_PTR_ERROR(taskInfo)) {
        DEBUG_MSG_PRE(0);
        return 0;
    }

    TaskManager_P(&taskInfo->brkSem, 0, 0);
    if (brk < taskInfo->startBrk)
        goto returnV;
    brk = ALIGN_DATA_UP(brk, 0x1000);

    if (brk == taskInfo->brk)
        goto returnV;

    /* shrink brk */
    if (brk < taskInfo->brk) {
        if (!MemManager_FreeUserPages((void*)brk, taskInfo->brk - brk, 0, NULL))
            goto setBrk;
        goto returnV;
    }

    /* alloc new brk */
    if (IS_PTR_ERROR(MemManager_AllocUserPages((void*)taskInfo->brk, brk - taskInfo->brk, 0, MemManager_PROT_WRITE+MemManager_PROT_READ+MemManager_FIXED_ADDR+MemManager_FAIL_MAPPED, NULL)))
        goto returnV;

setBrk:
    taskInfo->brk = brk;
returnV:
    ret = taskInfo->brk;
    TaskManager_V(&taskInfo->brkSem, 0);
    return ret;
}

int CopyStrings(unsigned int count, char *const *strs, int prop, ExecInfo *exec) {
    int error;

    error = E_OK;
    while (count--) {
        char *str;

        if (prop & UserSpaceManager_RUN_USER) {
            /* user strings */
            error = CopyFromUser((char*)&str, (char*)(strs+count), 4);
            if (error)
                break;
        } else
            str = *(strs+count);
        error = PutString(str, prop, exec);
        if (error)
            break;
    }
    return error;
}

/* count the number of args in the arg array. count is not initialized to 0 */
int CountArgs(char *const *arg, int prop, unsigned int *count) {
    int error;
    
    error = E_OK;
    if (prop & UserSpaceManager_RUN_USER) {
        char *tmp2;
        /* user strings */
        while (!(error = CopyFromUser((char*)&tmp2, (char*)arg, sizeof(char*))) && tmp2) {
            (*count)++;
            arg++;
        }
    } else {
        while (*arg) {
            (*count)++;
            arg++;
        }
    }
    return error;
}

int Exit(TaskManager_Task *task, unsigned int exitType, unsigned int exitCode) {
    TaskInfo *taskInfo;

    taskInfo = TaskManager_GetTaskInfo(task, ID);
    if (IS_PTR_ERROR(taskInfo))
        return E_BAD_TYPE;

    taskInfo->pub.exitType = exitType;
    taskInfo->pub.exitCode = exitCode;
    return TaskManager_DestroyTask(task);
}

int FindExecLoader(ExecInfo *exec, int prop, TaskManager_Task **task, TaskManager_Thread **thread) {
    ExecLoader *curr;
    int error, ret;

    error = E_UNSUPPORTED;
    /* find loader */
    TaskManager_LockRead(&execLoadersRW, 0);
    curr = execLoaders;
    while (curr) {
        error = curr->checkExec(&exec->pub);
        /* don't let this loader be unregistered */
        if (error == E_OK)
            curr->refCounter++;
        if (error != E_UNSUPPORTED)
            break;
        curr = curr->next;
    }
    TaskManager_UnlockRW(&execLoadersRW, 0);
    if (error != E_OK)
        goto errorEnd;
    /* ok, we have a loader. Now create task and load it! */
    exec->loader = curr;
    /* task with priority 30 */
    error = TaskManager_CreateTask(LoadExec, exec, 0x1e000000, task, thread);
    if (error)
        goto errorDecExecLoader;
    /* wait LoadExec result */
    TaskManager_P(&exec->sem, 0, 0);
    /* success? */
    if (!exec->ret && !(prop & UserSpaceManager_RUN_SUSPENDED)) {
        /* run! */
        ret = TaskManager_WaitQueueWakeup(&exec->waitQ);
        ASSERT_ERROR(ret);
    }
    error = exec->ret;

errorDecExecLoader:
    TaskManager_LockRead(&execLoadersRW, 0);
    curr->refCounter--;
    TaskManager_UnlockRW(&execLoadersRW, 0);
errorEnd:
    return error;
}

void LoadExec(ExecInfo *exec) {
    TaskManager_Regs regs;
    TaskManager_Task *parent, *task;
    TaskInfo *taskInfo;
    int error, ret;
    void *voidRet;

    /* say it's a user task */
    taskInfo = MemManager_Alloc(sizeof(TaskInfo), 0);
    if (IS_PTR_ERROR(taskInfo)) {
        error = PTR_TO_ERROR(taskInfo);
        goto errorV;
    }
    taskInfo->pub.exitType = UserSpaceManager_TASK_KILLED;
    taskInfo->pub.exitCode = 0;
    taskInfo->brk = 0;
    taskInfo->startBrk = 0xffffffff;
    taskInfo->brkSem.count = 1;
    taskInfo->brkSem.data = NULL;
    voidRet = TaskManager_SetTaskInfo(NULL, ID, taskInfo, TaskManager_SET_INFO_KEEP_DATA);
    if (IS_PTR_ERROR(voidRet)) {
        error = PTR_TO_ERROR(voidRet);
        ret = MemManager_Free(taskInfo);
        ASSERT_ERROR(ret);
        goto errorV;
    }
    /* send UserSpaceManagerE TaskCreated event */
    parent = TaskManager_GetParentTaskHandle(NULL);
    task = TaskManager_GetTaskHandle(NULL);
    SysModManager_Broadcast("UserSpaceManagerE", 0x00, parent, task);
    TaskManager_Destroyable(parent);
    TaskManager_Destroyable(NULL);
    /* load it! */
    error = exec->loader->loadExec(&exec->pub, &regs);
    if (error != E_OK)
        goto errorV;
    /* update brk info */
    taskInfo->brk = exec->pub.startBrk;
    taskInfo->startBrk = exec->pub.startBrk;
    /* wait, till we may run */
    TaskManager_WaitQueueAdd(&exec->waitQ, &exec->sem, 0);
    TaskManager_ToUserLevel(&regs);

errorV:
    exec->ret = error;
    TaskManager_V(&exec->sem, 0);
    ret = TaskManager_DestroyTask(NULL);
    ASSERT_ERROR(ret);
}

int MapStackPages(ExecInfo *exec, char *stack, unsigned int size) {
    char *logPage;
    unsigned int i;
    int error;

    /* fixup stack pointer */
    logPage = stack + size - (NUM_ARGS_PG * 4096);
    exec->pub.stackTop += (unsigned int)logPage;
    exec->pub.exec += (unsigned int)logPage;
    logPage += (NUM_ARGS_PG * 4096) - 4096;
    error = E_OK;
    for (i = NUM_ARGS_PG - 1; i < NUM_ARGS_PG && exec->pages[i]; i--, logPage -= 4096) {
        error = MemManager_MapPageAtUserPage(logPage, exec->pages[i], NULL);
        if (error)
            break;
        /* we don't need to free this page in case of error */
        exec->pages[i] = 0;
    }
    return error;
}

int PutInt(unsigned int pos, unsigned int integer, ExecInfo *exec) {
    unsigned int i, page;
    void *voidRet;

    i = pos >> 12;
    page = exec->pages[i];
    if (!page) {
        page = MemManager_i386_AllocPhysPages(1, 0);
        if (IS_PTR_ERROR(page))
            return PTR_TO_ERROR(page);
        exec->pages[i] = page;
    }
    if (exec->cachePage != page) {
        exec->cachePage = page;
        voidRet = MemManager_i386_MapPhysPages(exec->cache, page, 1, MemManager_i386_RW_PAGE);
        ASSERT_ERROR_PTR(voidRet);
    }
    *(unsigned int*)(exec->cache + (pos & 0xfff)) = integer;
    return E_OK;
}

int PutString(const char *str, int prop, ExecInfo *exec) {
    unsigned int len, pos;
    int error;

    if (prop & UserSpaceManager_RUN_USER) {
        /* user strings */
        error = StrLen(str, &len);
        if (error)
            return error;
    } else
        len = StringMan_Length(str);
    len++;
    if (exec->pub.stackTop < len)
        return E_TOO_BIG;
    exec->pub.stackTop -= len;
    pos = exec->pub.stackTop;
    while (len > 0) {
        unsigned int offset, i, new, copy, page;

        offset = pos & 0xfff;
        i = pos >> 12;
        page = exec->pages[i];
        new = 0;
        if (!page) {
            page = MemManager_i386_AllocPhysPages(1, 0);
            if (IS_PTR_ERROR(page))
                return PTR_TO_ERROR(page);
            exec->pages[i] = page;
            new = 1;
        }
        if (exec->cachePage != page) {
            void *voidRet;
            
            exec->cachePage = page;
            voidRet = MemManager_i386_MapPhysPages(exec->cache, page, 1, MemManager_i386_RW_PAGE);
            ASSERT_ERROR_PTR(voidRet);
        }

        if (new && offset)
            MemMan_Set(exec->cache, 0, offset);
        copy = 4096 - offset;
        if (copy > len)
            copy = len;
        if (prop & UserSpaceManager_RUN_USER) {
            /* user strings */
            error = CopyFromUser(exec->cache + offset, str, copy);
            if (error)
                return error;
        } else
            MemMan_Copy(exec->cache + offset, str, copy);

        pos += copy;
        str += copy;
        len -= copy;
    }
    return E_OK;
}

int RegisterLoader(const char *impName, int prop) {
    int error, ret;
    ExecLoader *newExec;
    int (*checkExecF)(UserSpaceManager_ExecInfo*);
    int (*loadExecF)(UserSpaceManager_ExecInfo*,TaskManager_Regs*);
    SysModManager_Imp *imp;

#ifdef DEBUG_ASSERT
    if (!impName) {
        ASSERT_MSG_PRE(0);
        return E_BAD_VALUE;
    }
#endif
    /* get imp ID */
    imp = SysModManager_GetImpID("ExecLoader", impName);
    if (IS_PTR_ERROR(imp))
        return E_NOT_FOUND;
    TaskManager_LockWrite(&execLoadersRW, 0);
    error = E_MODULE;
    checkExecF = SysModManager_GetFunction(imp, ExecLoader__CHECKEXEC, 0, NULL);
    loadExecF = SysModManager_GetFunction(imp, ExecLoader__LOADEXEC, 0, NULL);
    if (IS_PTR_ERROR(checkExecF) | IS_PTR_ERROR(loadExecF))
        goto errorReg;
    newExec = MemManager_Alloc(sizeof(ExecLoader), 0);
    if (IS_PTR_ERROR(newExec)) {
        error = PTR_TO_ERROR(newExec);
errorReg:
        TaskManager_UnlockRW(&execLoadersRW, 0);
        ret = SysModManager_PutImpID(imp);
        ASSERT_ERROR(ret);
        return error;
    }
    newExec->imp = imp;
    newExec->refCounter = 0;
    newExec->checkExec = checkExecF;
    newExec->loadExec = loadExecF;
    newExec->next = execLoaders;
    execLoaders = newExec;
    TaskManager_UnlockRW(&execLoadersRW, 0);
    return E_OK;
}

int Run(const char *pathname, char *const *argv, char *const *envp, int prop, TaskManager_Task **task, TaskManager_Thread **thread) {
    ExecInfo exec;
    FSManager_FDIO fdIO;
    FSManager_EntryStatus statBuf;
    unsigned int i;
    int error, ret;

#ifdef DEBUG_ASSERT
    if (!pathname || !argv || !envp || !task || !thread) {
        ASSERT_MSG_PRE(0);
        return E_BAD_VALUE;
    }
#endif
    MemMan_Set((char*)&exec, 0, sizeof(exec));
    error = FSManager_Open(pathname, FSManager_READ, &exec.pub.fd);
    if (error)
        goto errorReturn;
    error = FSManager_StatusFD(exec.pub.fd, &statBuf, 0);
    if (error)
        goto errorClose;
    if (!(statBuf.mode & FSManager_TYPE_FILE) || !(statBuf.mode & (FSManager_USER_EXEC+FSManager_GROUP_EXEC+FSManager_OTHER_EXEC))) {
        error = E_PERMISSION_DENIED;
        goto errorClose;
    }
    /* count argv and envc */
    error = CountArgs(argv, prop, &exec.pub.argc);
    if (error)
        goto errorClose;
    error = CountArgs(envp, prop, &exec.pub.envc);
    if (error)
        goto errorClose;
    /* read some bytes from file to identify it */
    fdIO.fd = exec.pub.fd;
    fdIO.flags = 0;
    fdIO.buf = exec.pub.fileBuf;
    fdIO.size = sizeof(exec.pub.fileBuf);
    error = FSManager_Read(&fdIO);
    if (error)
        goto errorClose;
    exec.pub.stackTop = NUM_ARGS_PG * 4096;
    exec.pub.stackSize = NUM_ARGS_PG * 4096;
    /* current mapped stack page */
    exec.cachePage = 0xffffffff;
    exec.cache = MemManager_i386_AllocPages(1, MemManager_i386_RW_PAGE);
    if (IS_PTR_ERROR(exec.cache)) {
        error = PTR_TO_ERROR(exec.cache);
        goto errorClose;
    }
    error = PutString(pathname, 0, &exec);
    if (error)
        goto errorFreePages;
    exec.pub.exec = exec.pub.stackTop;
    /* copy envp */
    error = CopyStrings(exec.pub.envc, envp, prop, &exec);
    if (error)
        goto errorFreePages;
    /* copy argv */
    error = CopyStrings(exec.pub.argc, argv, prop, &exec);
    if (error)
        goto errorFreePages;
    error = FindExecLoader(&exec, prop, task, thread);

    /* if there's no error this will do nothing */
errorFreePages:
    for (i = 0; i < NUM_ARGS_PG; i++)
        if (exec.pages[i]) {
            ret = MemManager_i386_FreePhysPages(exec.pages[i], 1);
            ASSERT_ERROR(ret);
        }
    ret = MemManager_i386_FreePages(exec.cache, 1, MemManager_i386_DONT_FREE_PHYS);
    ASSERT_ERROR(ret);
errorClose:
    ret = FSManager_Close(exec.pub.fd);
    ASSERT_ERROR(ret);
errorReturn:
    return error;
}

int Shutdown(void) {
    return 0;
}

int Phase1Start(SysModManager_Module *modID) {
    TaskManager_Task *task;
    TaskManager_Thread *thread;
    CfgManager_Group *baseGroup, *devGroup;
    char buf[128];
    unsigned int bufSize, index;
    char *argv[2];
    char *envp[1];
    int ret;

    ID = modID;

    /* start used modules */
    if ((ret = SysModManager_StartModule((unsigned int)CfgManager_OpenGroup)) ||
        (ret = SysModManager_StartModule((unsigned int)MemManager_Alloc)) ||
        (ret = SysModManager_StartModule((unsigned int)TaskManager_P)) ||
        (ret = SysModManager_StartModule((unsigned int)InterManager_RegisterInput)) ||
        (ret = SysModManager_StartModule((unsigned int)MemMan_Set)) ||
        (ret = SysModManager_StartModule((unsigned int)StringMan_Length))) {
        DEBUG_MSG_PRE(ret);
        goto errorReturn;
    }

    /* get page fault exceptions */
    if ((ret = MemManager_i386_Catch(CommonCopyUser, (void*)CommonCopyUser_newEIP, (void*)CommonCopyUser_newEIP, MemManager_i386_EX_RETURN))) {
        DEBUG_MSG_PRE(ret);
        goto errorReturn;
    }
    if ((ret = MemManager_i386_Catch(StrNCopy, (void*)StrNCopy_newEIP, (void*)StrNCopy_newEIP, MemManager_i386_EX_RETURN))) {
        DEBUG_MSG_PRE(ret);
        goto errorFreeCommonCopyEx;
    }
    if ((ret = MemManager_i386_Catch(StrLen, (void*)StrLen_newEIP, (void*)StrLen_newEIP, MemManager_i386_EX_RETURN))) {
        DEBUG_MSG_PRE(ret);
        goto errorFreeStrNCopyEx;
    }

    DEBUG_MSG(("Getting UserSpaceManager settings...\n"));

    /* register permanent exec loaders */
    baseGroup = CfgManager_OpenGroup(NULL, "/system/UserSpaceManager/loader");
    if (!baseGroup) {
        DEBUG_MSG_PRE(0);
        goto errorFreeStrLenEx;
    }

    index = 0;
    while (1) {
        CfgManager_Group *loaderGroup;

        loaderGroup = CfgManager_OpenGroupIndex(baseGroup, index);
        if (!loaderGroup) {
            break;
        }
        /* get name of the implementation */
        bufSize = 128;
        if ((ret = CfgManager_GetValue(loaderGroup, "imp", CfgManager_STRING, buf, &bufSize))) {
            SysModManager_DebugMsg("Cannot read loader imp: %x\n", ret);
            goto nextLoader;
        }
        if ((ret = RegisterLoader(buf, 0))) {
            SysModManager_DebugMsg("Cannot register binary loader '%s': %x\n", buf, ret);
        }
nextLoader:
        ret = CfgManager_CloseGroup(loaderGroup);
        ASSERT_ERROR(ret);
        index++;
    }
    ret = CfgManager_CloseGroup(baseGroup);
    ASSERT_ERROR(ret);

    /* let's start the first program */
    baseGroup = CfgManager_OpenGroup(NULL, "/system/UserSpaceManager/init");
    if (!baseGroup) {
        DEBUG_MSG_PRE(0);
        goto errorFreeStrLenEx;
    }
    /* get program path */
    bufSize = 128;
    if ((ret = CfgManager_GetValue(baseGroup, "path", CfgManager_STRING, buf, &bufSize))) {
        DEBUG_MSG_PRE(ret);
        goto errorCloseBaseGroup;
    }
    /* find exec name */
    argv[0] = StringMan_FindLastChar(buf, '/');
    if (!argv[0])
        argv[0] = buf;
    else
        argv[0]++;
    argv[1] = NULL;
    envp[0] = NULL;
    if ((ret = Run(buf, argv, envp, UserSpaceManager_RUN_SUSPENDED, &task, &thread))) {
        DEBUG_MSG_PRE(ret);
        goto errorCloseBaseGroup;
    }
    /* register in and output devices */
    devGroup = CfgManager_OpenGroup(baseGroup, "input");
    if (!devGroup) {
        DEBUG_MSG_PRE(0);
        goto errorDestroyTask;
    }

    index = 0;
    while (1) {
        CfgManager_Group *impGroup;

        impGroup = CfgManager_OpenGroupIndex(devGroup, index);
        if (!impGroup) {
            break;
        }
        /* get name of the implementation */
        bufSize = 128;
        if ((ret = CfgManager_GetValue(impGroup, "imp", CfgManager_STRING, buf, &bufSize))) {
            SysModManager_DebugMsg("Cannot read input imp: %x\n", ret);
            goto nextImpInput;
        }
        if ((ret = InterManager_RegisterInput(task, buf))) {
            SysModManager_DebugMsg("Cannot register input dev '%s': %x\n", buf, ret);
        }
nextImpInput:
        ret = CfgManager_CloseGroup(impGroup);
        ASSERT_ERROR(ret);
        index++;
    }
    ret = CfgManager_CloseGroup(devGroup);
    ASSERT_ERROR(ret);

    devGroup = CfgManager_OpenGroup(baseGroup, "output");
    if (!devGroup) {
        DEBUG_MSG_PRE(0);
        goto errorDestroyTask;
    }

    index = 0;
    while (1) {
        CfgManager_Group *impGroup;

        impGroup = CfgManager_OpenGroupIndex(devGroup, index);
        if (!impGroup) {
            break;
        }
        /* get name of the implementation */
        bufSize = 128;
        if ((ret = CfgManager_GetValue(impGroup, "imp", CfgManager_STRING, buf, &bufSize))) {
            SysModManager_DebugMsg("Cannot read output imp: %x\n", ret);
            goto nextImpOutput;
        }
        if ((ret = InterManager_RegisterOutput(task, buf))) {
            SysModManager_DebugMsg("Cannot register output dev '%s': %x\n", buf, ret);
        }
nextImpOutput:
        ret = CfgManager_CloseGroup(impGroup);
        ASSERT_ERROR(ret);
        index++;
    }
    ret = CfgManager_CloseGroup(devGroup);
    ASSERT_ERROR(ret);
    ret = CfgManager_CloseGroup(baseGroup);
    ASSERT_ERROR(ret);

    DEBUG_MSG(("Done.\n"));

    ret = TaskManager_Resume(thread, 0);
    ASSERT_ERROR(ret);
    /* Run has set it undestroyable */
    TaskManager_DestroyableThread(thread);
    return E_OK;

errorDestroyTask:
    ret = Exit(task, UserSpaceManager_TASK_KILLED, 0);
    ASSERT_ERROR(ret);
errorCloseBaseGroup:
    ret = CfgManager_CloseGroup(baseGroup);
    ASSERT_ERROR(ret);
errorFreeStrLenEx:
    ret = MemManager_i386_FreeException(StrLen, (void*)StrLen_newEIP, 0);
    ASSERT_ERROR(ret);
errorFreeStrNCopyEx:
    ret = MemManager_i386_FreeException(StrNCopy, (void*)StrNCopy_newEIP, 0);
    ASSERT_ERROR(ret);
errorFreeCommonCopyEx:
    ret = MemManager_i386_FreeException(CommonCopyUser, (void*)CommonCopyUser_newEIP, 0);
    ASSERT_ERROR(ret);
errorReturn:
    return 1;
}

int UnregisterLoader(const char *impName, int prop) {
    int error, ret;
    ExecLoader *curr, **pNext;
    SysModManager_Imp *imp;

#ifdef DEBUG_ASSERT
    if (!impName) {
        ASSERT_MSG_PRE(0);
        return E_BAD_VALUE;
    }
#endif
    /* get imp ID */
    imp = SysModManager_GetImpID("ExecLoader", impName);
    if (IS_PTR_ERROR(imp))
        return E_NOT_FOUND;
    TaskManager_LockWrite(&execLoadersRW, 0);
    error = E_NOT_FOUND;
    curr = execLoaders;
    pNext = &execLoaders;
    while (curr) {
        if (curr->imp == imp) {
            error = E_BUSY;
            if (!curr->refCounter) {
                ret = SysModManager_PutImpID(curr->imp);
                ASSERT_ERROR(ret);
                *pNext = curr->next;
                error = E_OK;
            }
            break;
        }
        pNext = &curr->next;
        curr = curr->next;
    }
    TaskManager_UnlockRW(&execLoadersRW, 0);
    ret = SysModManager_PutImpID(imp);
    ASSERT_ERROR(ret);
    return error;
}

void TaskManagerE_TaskDestroyed(TaskManager_Task *task) {
    TaskInfo *taskInfo;
    int ret;

    /* is an user module? */
    taskInfo = TaskManager_DelTaskInfo(task, ID);
    if (IS_PTR_ERROR(taskInfo))
        return;
    /* send UserSpaceManagerE TaskDestroyed event */
    SysModManager_Broadcast("UserSpaceManagerE", 1, task, &taskInfo->pub);
    ret = MemManager_Free(taskInfo);
    ASSERT_ERROR(ret);
}

int UGetUserStr(const char *uStr, char **sysStr) {
    char *page;
    int error, ret;
    unsigned int len;

    error = E_BAD_VALUE;
    if (!uStr)
        goto returnError;
    page = MemManager_i386_AllocPages(1, MemManager_i386_RW_PAGE+MemManager_i386_ALLOC_PHYS);
    if (IS_PTR_ERROR(page)) {
        error = PTR_TO_ERROR(page);
        goto returnError;
    }
    error = StrNCopy(page, uStr, 4096, &len);
    if (!error) {
        if (len != 4096) {
            *sysStr = page;
            goto returnError;
        }
        error = E_TOO_BIG;
    }
    ret = MemManager_i386_FreePages(page, 1, 0);
    ASSERT_ERROR(ret);
returnError:
    return error;
}

static __inline__ void UPutUserStr(char *uStr) {
    int ret;

    ret = MemManager_i386_FreePages(uStr, 1, 0);
    ASSERT_ERROR(ret);
}

unsigned int UBrk(unsigned int brk) {
    return Brk(brk);
}

int URun(const char *pathname, char *const argv[], char *const envp[]) {
    int error;
    char *uPath;
    TaskManager_Task *task;
    TaskManager_Thread *thread;

    if (!argv || !envp)
        return E_BAD_VALUE;
    error = UGetUserStr(pathname, &uPath);
    if (error)
        return error;
    /* we must check this because system modules must not use empty paths */
    error = -E_BAD_VALUE;
    if (!*uPath)
        goto returnEnd;
    error = -Run(uPath, argv, envp, UserSpaceManager_RUN_USER, &task, &thread);
    if (!error) {
        error = TaskManager_GetTaskID(task);
        /* Run has set it undestroyable */
        TaskManager_DestroyableThread(thread);
    }
returnEnd:
    UPutUserStr(uPath);
    return error;
}
