/*
*****************************************
* This file implements the BlockBuffer interface
*
* Version: 0.2.6
* Author: Luiz Henrique Shigunov
* E-mail: luizshigunov@users.sourceforge.net
* WEB: http://www.sourceforge.net/projects/modulos
* Last change: 25/01/2004
*
* See the interface specification to know what each function does.
*
* Copyright (C) 2000-2004, Luiz Henrique Shigunov
*
* This module is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2.1 of the License, or (at your option) any later version.
*
* This module is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this module; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*****************************************
*/

#define __NO_BLOCKBUFFER_TYPES

#include <stddef.h>
#include <blockbuffer.h>
#include <blockdev.h>
#include <debug.h>
#include <errormanager.h>
#include <memmanager.h>
#include <memmanager_i386.h>
#include <stringman.h>
#include <sysmodmanager.h>
#include <taskmanager.h>

#define BLOCK_FLAG_BUSY 0x01
#define BLOCK_FLAG_NOT_VALID 0x02

/* very simple now, will get better when we get a better memory manager */
struct _BlockBuffer_Block {
    char *data;
    void *avail1;
    void *avail2;
    unsigned int flags;
    unsigned int blockN;
    struct _BlockBuffer_Block *hashNext;
    struct _BlockBuffer_Block *hashPrior;
    TaskManager_RWSync lock;
    TaskManager_WaitQueue waitB;
};

#define NUM_FUNCTIONS 3
#define READ_FUNC 0
#define WRITE_FUNC 1
#define GETUD_FUNC 2

typedef struct _Device {
    void *functions[NUM_FUNCTIONS];
    struct _Device *next;
    struct _Device *prior;
    SysModManager_Imp *imp;
    struct _BlockBuffer_Unit *units;
    unsigned int refCount;
} Device;

struct _BlockBuffer_Unit {
    Device *dev;
    TaskManager_Semaphore htSem;
    BlockBuffer_Block **hashTable;
    unsigned int blocksInUse;
    unsigned int unit;
    unsigned int LBS;
    unsigned int HBS;
    struct _BlockBuffer_Unit *next;
};

Device *devs;
TaskManager_Semaphore devsMutex = {1, NULL};
unsigned int functionNumber[NUM_FUNCTIONS] = {BlockDev__READ, BlockDev__WRITE, BlockDev__GETUNITDESC};

/* linked list of free blocks */
BlockBuffer_Block *freeBlocks[4];
TaskManager_Semaphore freeBlocksMutex = {1, NULL};

static __inline__ BlockBuffer_Block *AllocBlock(BlockBuffer_Unit *unit, unsigned int blockN) {
    unsigned int i, bSize;
    BlockBuffer_Block *block;

    /* check if there is free blocks */
    i = 0;
    bSize = unit->LBS >> 10;
    while (bSize) {
        i++;
        bSize >>= 1;
    }
    block = NULL;
    TaskManager_P(&freeBlocksMutex, 0, 0);
    if (!freeBlocks[i]) {
        void *page;
        unsigned int n;
        BlockBuffer_Block *newBlock, *tmp;
        /* we have to fill the free blocks list */
        page  = MemManager_i386_AllocPages(1, MemManager_i386_RW_PAGE+MemManager_i386_ALLOC_PHYS);
        if (IS_PTR_ERROR(page))
            goto returnV;
        n = 4096 >> (9+i);
        newBlock = NULL;
        do {
            tmp = MemManager_Alloc(sizeof(BlockBuffer_Block), 0);
            if (IS_PTR_ERROR(tmp)) {
                int ret;
                /* no memory, free everything */
                while (newBlock) {
                    tmp = newBlock->hashNext;
                    ret = MemManager_Free(newBlock);
                    ASSERT_ERROR(ret);
                    newBlock = tmp;
                }
                ret = MemManager_i386_FreePages(page, 1, 0);
                ASSERT_ERROR(ret);
                goto returnV;
            }
            n--;
            tmp->data = (char*)page+(unit->LBS*n);
            tmp->lock.data[0] = NULL;
            tmp->lock.data[1] = NULL;
            tmp->lock.data[2] = NULL;
            tmp->lock.data[3] = NULL;
            tmp->waitB.data = NULL;
            tmp->hashNext = newBlock;
            newBlock = tmp;
        } while (n);
        freeBlocks[i] = newBlock;
    }
    block = freeBlocks[i];
    freeBlocks[i] = block->hashNext;
    /* start block */
    block->blockN = blockN;
returnV:
    TaskManager_V(&freeBlocksMutex, 0);
    return block;
}

static __inline__ BlockBuffer_Block *GetCachedBlock(BlockBuffer_Unit *unit, unsigned int blockN) {
    BlockBuffer_Block **head, *tmp, *ret;

    head = unit->hashTable + (blockN & 0x3ff);
    tmp = *head;
    ret = NULL;
    while (tmp) {
        BlockBuffer_Block *block;
        block = tmp;
        tmp = tmp->hashNext;
        if (block->blockN != blockN)
            continue;
        TaskManager_AtomicInc(&unit->blocksInUse);
        ret = block;
        break;
    }
    return ret;
}

static __inline__ int GetDiskBlock(BlockBuffer_Unit *unit, unsigned int blockN, BlockBuffer_Block *block) {
    int (*read)(BlockDev_IO*);
    BlockDev_IO rq;
    unsigned int n;

    /* get number of physical blocks per logical block */
    n = unit->LBS / unit->HBS;
    rq.unit = unit->unit;
    rq.block = blockN*n;
    rq.n = n;
    rq.buffer = block->data;
    read = unit->dev->functions[READ_FUNC];
    if (read(&rq))
        return E_IO_ERROR;
    return E_OK;
}

static __inline__ void PutBlock(BlockBuffer_Unit *unit, BlockBuffer_Block *block) {
    TaskManager_AtomicDec(&unit->blocksInUse);
}

int FreeUnit(BlockBuffer_Unit *unit) {
    int error, ret;

#ifdef DEBUG_ASSERT
    if (!unit) {
        ASSERT_MSG_PRE(0);
        return E_BAD_VALUE;
    }
#endif

    TaskManager_P(&devsMutex, 0, 0);
    error = E_IN_USE;
    if (!unit->blocksInUse) {
        Device *dev;
        dev = unit->dev;
        if (dev->units == unit)
            dev->units = unit->next;
        else {
            BlockBuffer_Unit *prior;
            /* remove unit from linked list */
            prior = dev->units;
            while (prior->next != unit)
                prior = prior->next;
            prior->next = unit->next;
        }
        ret = MemManager_i386_FreePages(unit->hashTable, 1, 0);
        ASSERT_ERROR(ret);
        ret = MemManager_Free(unit);
        ASSERT_ERROR(ret);
        dev->refCount--;
        if (!dev->refCount) {
            /* free device */
            ret = SysModManager_PutImpID(dev->imp);
            ASSERT_ERROR(ret);
            /* remove from device list */
            if (dev->next && dev->prior) {
                /* remove from the middle of the list */
                dev->prior->next = dev->next;
                dev->next->prior = dev->prior;
            } else if (dev->prior)
                /* we're in the end of the list */
                dev->prior->next = NULL;
            else {
                /* we're in the start of the list */
                if (dev->next)
                    dev->next->prior = NULL;
                devs = dev->next;
            }
            ret = MemManager_Free(dev);
            ASSERT_ERROR(ret);
        }
        error = E_OK;
    }
    TaskManager_V(&devsMutex, 0);
    return error;
}

int GetBlock(BlockBuffer_Unit *unit, unsigned int blockN, int prop, BlockBuffer_Block **block) {
    int error;
    BlockBuffer_Block *tmpB;

#ifdef DEBUG_ASSERT
    if (!unit || !block) {
        ASSERT_MSG_PRE(0);
        return E_BAD_VALUE;
    }
#endif

    /* first check hash table */
    TaskManager_P(&unit->htSem, 0, 0);
    tmpB = GetCachedBlock(unit, blockN);
    /* if block is not valid, try to read it again */
    if (tmpB && !(tmpB->flags & BLOCK_FLAG_NOT_VALID)) {
        /* block is busy? If so, wait. */
        while (tmpB->flags & BLOCK_FLAG_BUSY) {
            TaskManager_WaitQueueAdd(&tmpB->waitB, &unit->htSem, 0);
            TaskManager_P(&unit->htSem, 0, 0);
        }
        /* ok, now we know the block is not busy. is it valid? */
        if (tmpB->flags & BLOCK_FLAG_NOT_VALID) {
            PutBlock(unit, tmpB);
            TaskManager_V(&unit->htSem, 0);
            return E_IO_ERROR;
        }
getIt:
        TaskManager_V(&unit->htSem, 0);
        /* there is no problem here, we DO NOT free blocks yet! */
        if (prop & 0x02)
            error = TaskManager_LockWrite(&tmpB->lock, prop & 0x01);
        else
            error = TaskManager_LockRead(&tmpB->lock, prop & 0x01);
        if (!error)
            *block = tmpB;
        else
            PutBlock(unit, tmpB);
        return error;
    }
    /* we have to read it from disk */
    if (!tmpB) {
        BlockBuffer_Block **head;
        /* alloc a block and put it in the hash table. */
        tmpB = AllocBlock(unit, blockN);
        if (!tmpB) {
            TaskManager_V(&unit->htSem, 0);
            return E_NO_MEMORY;
        }
        head = unit->hashTable + (blockN & 0x3ff);
        tmpB->hashPrior = NULL;
        tmpB->hashNext = *head;
        if (*head)
            (*head)->hashPrior = tmpB;
        *head = tmpB;
        TaskManager_AtomicInc(&unit->blocksInUse);
    }
    /* make block busy, so no one can get it */
    tmpB->flags = BLOCK_FLAG_BUSY;
    /* now read it */
    TaskManager_V(&unit->htSem, 0);
    error = GetDiskBlock(unit, blockN, tmpB);
    TaskManager_P(&unit->htSem, 0, 0);
    /* wake up everyone waiting in the wait queue */
    TaskManager_WaitQueueWakeup(&tmpB->waitB);
    tmpB->flags = 0;
    if (!error)
        goto getIt;
    /* block is bad */
    tmpB->flags = BLOCK_FLAG_NOT_VALID;
    PutBlock(unit, tmpB);
    TaskManager_V(&unit->htSem, 0);
    return error;
}

int GetBlockSize(BlockBuffer_Unit *unit, unsigned int *size) {
#ifdef DEBUG_ASSERT
    if (!unit || !size) {
        ASSERT_MSG_PRE(0);
        return E_BAD_VALUE;
    }
#endif

    *size = unit->LBS;
    return E_OK;
}

int GetHardBlockSize(BlockBuffer_Unit *unit, unsigned int *size) {
#ifdef DEBUG_ASSERT
    if (!unit || !size) {
        ASSERT_MSG_PRE(0);
        return E_BAD_VALUE;
    }
#endif

    *size = unit->HBS;
    return E_OK;
}

int RegisterUnit(const char *devImpName, unsigned int unitN, BlockBuffer_Unit **unit) {
    int error, ret;
    Device *dev;
    unsigned int hbs;
    BlockDev_UnitDesc unitDesc;
    BlockBuffer_Unit *newUnit;
    SysModManager_Imp *devImp;
    int (*getUD)(unsigned int,BlockDev_UnitDesc*,unsigned int,int);

#ifdef DEBUG_ASSERT
    if (!devImpName || !unit) {
        ASSERT_MSG_PRE(0);
        return E_BAD_VALUE;
    }
#endif

    devImp = SysModManager_GetImpID("BlockDev", devImpName);
    if (IS_PTR_ERROR(devImp))
        return E_MODULE;
    TaskManager_P(&devsMutex, 0, 0);
    /* find device */
    dev = devs;
    while (dev && dev->imp != devImp)
        dev = dev->next;
    if (!dev) {
        unsigned int i;
        /* register a new device */
        dev = MemManager_Alloc(sizeof(Device), 0);
        if (IS_PTR_ERROR(dev)) {
            error = PTR_TO_ERROR(dev);
            goto errorPutImp;
        }
        /* get module functions */
        error = E_MODULE;
        i = 0;
        do {
            dev->functions[i] = SysModManager_GetFunction(devImp, functionNumber[i], 0, NULL);
            if (IS_PTR_ERROR(dev->functions[i]))
                goto returnFreeFunctions;
        } while (++i < NUM_FUNCTIONS);
        /* register new unit */
        getUD = dev->functions[GETUD_FUNC];
        error = getUD(unitN, &unitDesc, sizeof(unitDesc), BlockDev_NO_NAME);
        if (error)
            goto returnFreeFunctions;
        hbs = unitDesc.blockSize ? unitDesc.blockSize : 512;
        newUnit = MemManager_Alloc(sizeof(BlockBuffer_Unit), 0);
        if (IS_PTR_ERROR(newUnit)) {
            error = PTR_TO_ERROR(newUnit);
            goto returnFreeFunctions;
        }
        /* for now, each hash table is a page */
        newUnit->hashTable = MemManager_i386_AllocPages(1, MemManager_i386_RW_PAGE+MemManager_i386_ALLOC_PHYS+MemManager_i386_CLEAR_PAGES);
        if (IS_PTR_ERROR(newUnit->hashTable)) {
            error = PTR_TO_ERROR(newUnit->hashTable);
            ret = MemManager_Free(newUnit);
            ASSERT_ERROR(ret);
returnFreeFunctions:
            ret = MemManager_Free(dev);
            ASSERT_ERROR(ret);
errorPutImp:
            ret = SysModManager_PutImpID(devImp);
            ASSERT_ERROR(ret);
            goto returnV;
        }
        newUnit->htSem.count = 1;
        newUnit->htSem.data = NULL;
        newUnit->blocksInUse = 0;
        newUnit->HBS = hbs;
        newUnit->LBS = hbs;
        newUnit->dev = dev;
        newUnit->unit = unitN;
        newUnit->next = dev->units;
        dev->units = newUnit;
        dev->imp = devImp;
        dev->refCount = 1;
        dev->prior = NULL;
        dev->next = devs;
        if (devs)
            devs->prior = dev;
        devs = dev;
    } else {
        ret = SysModManager_PutImpID(devImp);
        ASSERT_ERROR(ret);
        /* check if unit is in use */
        newUnit = dev->units;
        error = E_IN_USE;
        do {
            if (newUnit->unit == unitN)
                goto returnV;
            newUnit = newUnit->next;
        } while (newUnit);
        /* register new unit */
        getUD = dev->functions[GETUD_FUNC];
        error = getUD(unitN, &unitDesc, sizeof(unitDesc), BlockDev_NO_NAME);
        if (error)
            goto returnV;
        hbs = unitDesc.blockSize ? unitDesc.blockSize : 512;
        newUnit = MemManager_Alloc(sizeof(BlockBuffer_Unit), 0);
        if (IS_PTR_ERROR(newUnit)) {
            error = PTR_TO_ERROR(newUnit);
            goto returnV;
        }
        /* for now, each hash table is a page */
        newUnit->hashTable = MemManager_i386_AllocPages(1, MemManager_i386_RW_PAGE+MemManager_i386_ALLOC_PHYS+MemManager_i386_CLEAR_PAGES);
        if (IS_PTR_ERROR(newUnit->hashTable)) {
            error = PTR_TO_ERROR(newUnit->hashTable);
            ret = MemManager_Free(newUnit);
            ASSERT_ERROR(ret);
            goto returnV;
        }
        newUnit->htSem.count = 1;
        newUnit->htSem.data = NULL;
        newUnit->blocksInUse = 0;
        newUnit->HBS = hbs;
        newUnit->LBS = hbs;
        newUnit->dev = dev;
        newUnit->unit = unitN;
        newUnit->next = dev->units;
        dev->units = newUnit;
        dev->refCount++;
    }
    *unit = newUnit;
    error = E_OK;
returnV:
    TaskManager_V(&devsMutex, 0);
    return error;
}

int Release(BlockBuffer_Unit *unit, BlockBuffer_Block *block) {
#ifdef DEBUG_ASSERT
    if (!unit || !block) {
        ASSERT_MSG_PRE(0);
        return E_BAD_VALUE;
    }
#endif

    TaskManager_AtomicDec(&unit->blocksInUse);
    TaskManager_UnlockRW(&block->lock, 0);
    return E_OK;
}

int SetBlockSize(BlockBuffer_Unit *unit, unsigned int size) {
    int error;
    unsigned int i, j, bSize;

#ifdef DEBUG_ASSERT
    if (!unit || (size != 512 && size != 1024 && size != 2048 && size != 4096)) {
        ASSERT_MSG_PRE(0);
        return E_BAD_VALUE;
    }
#endif

    TaskManager_P(&unit->htSem, 0, 0);
    error = E_IN_USE;
    if (unit->blocksInUse)
        goto returnV;
    j = 0;
    bSize = unit->LBS >> 10;
    while (bSize) {
        j++;
        bSize >>= 1;
    }
    TaskManager_P(&freeBlocksMutex, 0, 0);
    /* free all blocks */
    for (i=0; i<1024; i++)
        if (unit->hashTable[i]) {
            BlockBuffer_Block *block, *tmp;
            block = unit->hashTable[i];
            unit->hashTable[i] = NULL;
            do {
                tmp = block->hashNext;
                block->hashNext = freeBlocks[j];
                freeBlocks[j] = block;
                block = tmp;
            } while (block);
        }
    TaskManager_V(&freeBlocksMutex, 0);
    unit->LBS = size;
returnV:
    TaskManager_V(&unit->htSem, 0);
    return error;
}

int Phase1Start(SysModManager_Module *modID) {
    int ret;

    /* start all used modules */
    if ((ret = SysModManager_StartModule((unsigned int)MemManager_Alloc)) ||
        (ret = SysModManager_StartModule((unsigned int)StringMan_Compare)) ||
        (ret = SysModManager_StartModule((unsigned int)TaskManager_LockRead))) {
        DEBUG_MSG_PRE(ret);
        return 1;
    }
    return 0;
}
