/*
*****************************************
* This file implements the ExecLoader interface for ELF files
*
* Version: 0.1.2
* Author: Luiz Henrique Shigunov
* E-mail: luizshigunov@users.sourceforge.net
* WEB: http://www.sourceforge.net/projects/modulos
* Last change: 18/06/2005
*
* See the interface specification to know what each function does.
*
* Copyright (C) 2003-2005, Luiz Henrique Shigunov
*
* This module is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2.1 of the License, or (at your option) any later version.
*
* This module is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this module; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*****************************************
*/

/* This code was modified from the Linux 2.4 kernel - binfmt_elf.c */

#include <stddef.h>
#include <debug.h>
#include <errormanager.h>
#include <execdomain.h>
#include <execloader.h>
#include <fsmanager.h>
#include <memman.h>
#include <memmanager.h>
#include <sysdef.h>
#include <sysmodmanager.h>
#include <userspacemanager.h>

#define ALIGN_DATA_UP(data, align) ((((unsigned int)(data)) + (align)-1) & ~((align)-1))

#define PATH_MAX 4096
#define PAGE_SIZE 4096
#define ELF_EXEC_PAGESIZE 4096
#define unlikely(x) __builtin_expect((x),0)

#define CLOCKS_PER_SEC 1

#define PROT_READ MemManager_PROT_READ
#define PROT_WRITE MemManager_PROT_WRITE
#define PROT_EXEC MemManager_PROT_EXEC
#define MAP_PRIVATE 0
#define MAP_DENYWRITE 0
#define MAP_EXECUTABLE 0
#define MAP_FIXED MemManager_FIXED_ADDR

#define DLINFO_ITEMS 13

#define ELF_ET_DYN_BASE  (0xc0000000 / 3 * 2)

#if ELF_EXEC_PAGESIZE > PAGE_SIZE
# define ELF_MIN_ALIGN	ELF_EXEC_PAGESIZE
#else
# define ELF_MIN_ALIGN	PAGE_SIZE
#endif

#define ELF_PAGESTART(_v) ((_v) & ~(unsigned long)(ELF_MIN_ALIGN-1))
#define ELF_PAGEOFFSET(_v) ((_v) & (ELF_MIN_ALIGN-1))
#define ELF_PAGEALIGN(_v) (((_v) + ELF_MIN_ALIGN - 1) & ~(ELF_MIN_ALIGN - 1))

/* These constants define the various ELF target machines */
#define EM_386   3
#define EM_486   6   /* Perhaps disused */

#define elf_check_arch(x) (((x)->e_machine == EM_386) || ((x)->e_machine == EM_486))

/* These constants define the different elf file types */
#define ET_EXEC   2
#define ET_DYN    3

#define ELFMAG          "\177ELF"
#define SELFMAG         4

#define EI_NIDENT       16

/* These constants are for the segment types stored in the image headers */
#define PT_NULL    0
#define PT_LOAD    1
#define PT_DYNAMIC 2
#define PT_INTERP  3
#define PT_NOTE    4
#define PT_SHLIB   5
#define PT_PHDR    6
#define PT_LOPROC  0x70000000
#define PT_HIPROC  0x7fffffff

/* These constants define the permissions on sections in the program
   header, p_flags. */
#define PF_R            0x4
#define PF_W            0x2
#define PF_X            0x1

/* Symbolic values for the entries in the auxiliary table
   put on the initial stack */
#define AT_NULL   0     /* end of vector */
#define AT_IGNORE 1     /* entry should be ignored */
#define AT_EXECFD 2     /* file descriptor of program */
#define AT_PHDR   3     /* program headers for program */
#define AT_PHENT  4     /* size of program header entry */
#define AT_PHNUM  5     /* number of program headers */
#define AT_PAGESZ 6     /* system page size */
#define AT_BASE   7     /* base address of interpreter */
#define AT_FLAGS  8     /* flags */
#define AT_ENTRY  9     /* entry point of program */
#define AT_NOTELF 10    /* program is not ELF */
#define AT_UID    11    /* real uid */
#define AT_EUID   12    /* effective uid */
#define AT_GID    13    /* real gid */
#define AT_EGID   14    /* effective gid */
#define AT_PLATFORM 15  /* string identifying CPU for optimizations */
#define AT_HWCAP  16    /* arch dependent hints at CPU capabilities */
#define AT_CLKTCK 17    /* frequency at which times() increments */

/* 32-bit ELF base types. */
typedef unsigned int     Elf32_Addr;
typedef unsigned short   Elf32_Half;
typedef unsigned int     Elf32_Off;
typedef int              Elf32_Sword;
typedef unsigned int     Elf32_Word;

typedef struct elf32_hdr{
  unsigned char e_ident[EI_NIDENT];
  Elf32_Half    e_type;
  Elf32_Half    e_machine;
  Elf32_Word    e_version;
  Elf32_Addr    e_entry;  /* Entry point */
  Elf32_Off     e_phoff;
  Elf32_Off     e_shoff;
  Elf32_Word    e_flags;
  Elf32_Half    e_ehsize;
  Elf32_Half    e_phentsize;
  Elf32_Half    e_phnum;
  Elf32_Half    e_shentsize;
  Elf32_Half    e_shnum;
  Elf32_Half    e_shstrndx;
} Elf32_Ehdr;

typedef struct elf32_phdr{
  Elf32_Word    p_type;
  Elf32_Off     p_offset;
  Elf32_Addr    p_vaddr;
  Elf32_Addr    p_paddr;
  Elf32_Word    p_filesz;
  Elf32_Word    p_memsz;
  Elf32_Word    p_flags;
  Elf32_Word    p_align;
} Elf32_Phdr;

#define elfhdr          elf32_hdr
#define elf_phdr        elf32_phdr
#define elf_addr_t unsigned long
#define elf_caddr_t char *

int ExecLoader_CheckExec(UserSpaceManager_ExecInfo *exec) {
    struct elfhdr *elf_ex;
    
    /* Get the exec-header */
    elf_ex = (struct elfhdr *)exec->fileBuf;

    /* First of all, some simple consistency checks */
    if (MemMan_Compare(elf_ex->e_ident, ELFMAG, SELFMAG))
        return E_UNSUPPORTED;
    if (elf_ex->e_type != ET_EXEC && elf_ex->e_type != ET_DYN)
        return E_UNSUPPORTED;
    if (!elf_check_arch(elf_ex))
        return E_UNSUPPORTED;
    return E_OK;
}

static void set_brk(unsigned long start, unsigned long end) {
    void *tmp;

    start = ELF_PAGEALIGN(start);
    end = ELF_PAGEALIGN(end);
    if (end <= start)
        return;
    tmp = MemManager_AllocUserPages((void*)start, end - start, 0, MemManager_PROT_WRITE+MemManager_PROT_READ+MemManager_FIXED_ADDR, NULL);
    if (IS_PTR_ERROR(tmp))
        ASSERT_MSG_PRE(PTR_TO_ERROR(tmp));
}

/* quick and dirty! */
static void clear_user(void *to, unsigned long n) {
    char zero[64];
    unsigned int size;
    
    MemMan_Set(zero, 0, sizeof(zero));
    while (n) {
        size = n;
        if (size > sizeof(zero))
            size = sizeof(zero);
        if (UserSpaceManager_CopyToUser(to, zero, size))
            return;
        n -= size;
        to = (char*)to + size;
    }
}

/* We need to explicitly zero any fractional pages
   after the data section (i.e. bss).  This would
   contain the junk from the file that should not
   be in memory */
static void padzero(unsigned long elf_bss) {
    unsigned long nbyte;

    nbyte = ELF_PAGEOFFSET(elf_bss);
    if (nbyte) {
        nbyte = ELF_MIN_ALIGN - nbyte;
        clear_user((void *) elf_bss, nbyte);
    }
}

static elf_addr_t *
create_elf_tables(UserSpaceManager_ExecInfo *execInfo, char *p, int argc, int envc,
        struct elfhdr * exec,
        unsigned long load_addr,
        unsigned long load_bias,
        unsigned long interp_load_addr, int ibcs)
{
    elf_caddr_t *argv;
    elf_caddr_t *envp;
    elf_addr_t *sp, *csp;
    char *k_platform, *u_platform;
    long hwcap = 0;
    size_t len;
    int tmpInt, ret;

    k_platform = NULL;
    u_platform = p;

    /*
    * Force 16 byte _final_ alignment here for generality.
    */
    sp = (elf_addr_t *)(~15UL & (unsigned long)(u_platform));
    csp = sp;
    csp -= (1+DLINFO_ITEMS)*2 + (k_platform ? 2 : 0);
    csp -= envc+1;
    csp -= argc+1;
    csp -= (!ibcs ? 3 : 1);	/* argc itself */
    if ((unsigned long)csp & 15UL)
        sp -= ((unsigned long)csp & 15UL) / sizeof(*sp);

    /*
    * Put the ELF interpreter info on the stack
    */
#define NEW_AUX_ENT(nr, id, val) \
    tmpInt = id; \
    UserSpaceManager_CopyToUser(sp+(nr*2), (char*)&tmpInt, sizeof(tmpInt)); \
    tmpInt = val; \
    UserSpaceManager_CopyToUser(sp+(nr*2+1), (char*)&tmpInt, sizeof(tmpInt)); \

    sp -= 2;
    NEW_AUX_ENT(0, AT_NULL, 0);
    if (k_platform) {
        sp -= 2;
        NEW_AUX_ENT(0, AT_PLATFORM, (elf_addr_t)(unsigned long) u_platform);
    }
    sp -= DLINFO_ITEMS*2;
    NEW_AUX_ENT( 0, AT_HWCAP, hwcap);
    NEW_AUX_ENT( 1, AT_PAGESZ, ELF_EXEC_PAGESIZE);
    NEW_AUX_ENT( 2, AT_CLKTCK, CLOCKS_PER_SEC);
    NEW_AUX_ENT( 3, AT_PHDR, load_addr + exec->e_phoff);
    NEW_AUX_ENT( 4, AT_PHENT, sizeof (struct elf_phdr));
    NEW_AUX_ENT( 5, AT_PHNUM, exec->e_phnum);
    NEW_AUX_ENT( 6, AT_BASE, interp_load_addr);
    NEW_AUX_ENT( 7, AT_FLAGS, 0);
    NEW_AUX_ENT( 8, AT_ENTRY, load_bias + exec->e_entry);
    NEW_AUX_ENT( 9, AT_UID, (elf_addr_t) 0);
    NEW_AUX_ENT(10, AT_EUID, (elf_addr_t) 0);
    NEW_AUX_ENT(11, AT_GID, (elf_addr_t) 0);
    NEW_AUX_ENT(12, AT_EGID, (elf_addr_t) 0);
#undef NEW_AUX_ENT

    sp -= envc+1;
    envp = (elf_caddr_t *) sp;
    sp -= argc+1;
    argv = (elf_caddr_t *) sp;
    if (!ibcs) {
        tmpInt = (int)envp;
        UserSpaceManager_CopyToUser(--sp, (char*)&tmpInt, sizeof(tmpInt));
        tmpInt = (int)argv;
        UserSpaceManager_CopyToUser(--sp, (char*)&tmpInt, sizeof(tmpInt));
    }

    tmpInt = argc;
    UserSpaceManager_CopyToUser(--sp, (char*)&tmpInt, sizeof(tmpInt));
    while (argc-->0) {
        tmpInt = (int)p;
        UserSpaceManager_CopyToUser(argv++, (char*)&tmpInt, sizeof(tmpInt));
        ret = UserSpaceManager_StrLen(p, &len);
        len++;
        if (ret || len > execInfo->stackSize)
            return NULL;
        p += len;
    }
    tmpInt = 0;
    UserSpaceManager_CopyToUser(argv, (char*)&tmpInt, sizeof(tmpInt));
    while (envc-->0) {
        tmpInt = (int)p;
        UserSpaceManager_CopyToUser(envp++, (char*)&tmpInt, sizeof(tmpInt));
        ret = UserSpaceManager_StrLen(p, &len);
        len++;
        if (ret || len > execInfo->stackSize)
            return NULL;
        p += len;
    }
    tmpInt = 0;
    UserSpaceManager_CopyToUser(envp, (char*)&tmpInt, sizeof(tmpInt));
    return sp;
}

static inline unsigned long
elf_map (FSManager_Handle *filep, unsigned long addr, struct elf_phdr *eppnt, int prot, int type) {
    unsigned long map_addr;
    FSManager_FDIO fdio;

    map_addr = (unsigned long)MemManager_AllocUserPages((void*)ELF_PAGESTART(addr), ALIGN_DATA_UP(eppnt->p_filesz + ELF_PAGEOFFSET(eppnt->p_vaddr), 4096),
               0, MemManager_PROT_WRITE+MemManager_PROT_READ+MemManager_FIXED_ADDR, NULL);
    if (IS_PTR_ERROR(map_addr))
        return 0xffffffff;
    if (FSManager_Seek(filep, eppnt->p_offset - ELF_PAGEOFFSET(eppnt->p_vaddr), FSManager_SEEK_START, NULL))
        return 0xffffffff;

    fdio.fd = filep;
    fdio.flags = FSManager_USER_BUF_IO;
    fdio.buf = (void*)map_addr;
    fdio.size = eppnt->p_filesz + ELF_PAGEOFFSET(eppnt->p_vaddr);
    if (FSManager_Read(&fdio))
        return 0xffffffff;
    if (MemManager_SetUserPagesProt((void*)map_addr, ALIGN_DATA_UP(eppnt->p_filesz + ELF_PAGEOFFSET(eppnt->p_vaddr), 4096), prot|type, NULL))
        return 0xffffffff;
    return map_addr;
}

static unsigned long load_elf_interp(struct elfhdr * interp_elf_ex, FSManager_Handle *interpreter, unsigned long *interp_load_addr) {
    struct elf_phdr *elf_phdata;
    struct elf_phdr *eppnt;
    unsigned long load_addr = 0;
    int load_addr_set = 0;
    unsigned long last_bss = 0, elf_bss = 0;
    unsigned long error = ~0UL;
    int retval, i, size;
    FSManager_FDIO fdio;

    /* First of all, some simple consistency checks */
    if (interp_elf_ex->e_type != ET_EXEC &&
        interp_elf_ex->e_type != ET_DYN)
        goto out;
    if (!elf_check_arch(interp_elf_ex))
        goto out;

    /*
    * If the size of this structure has changed, then punt, since
    * we will be doing the wrong thing.
    */
    if (interp_elf_ex->e_phentsize != sizeof(struct elf_phdr))
        goto out;
    if (interp_elf_ex->e_phnum > 65536U / sizeof(struct elf_phdr))
        goto out;

    /* Now read in all of the header information */

    size = sizeof(struct elf_phdr) * interp_elf_ex->e_phnum;
    if (size > ELF_MIN_ALIGN)
        goto out;
    elf_phdata = (struct elf_phdr *) MemManager_Alloc(size, 0);
    if (IS_PTR_ERROR(elf_phdata))
        goto out;

    retval = FSManager_Seek(interpreter, interp_elf_ex->e_phoff, FSManager_SEEK_START, NULL);
    if (retval)
        goto out_close;
    
    fdio.fd = interpreter;
    fdio.flags = 0;
    fdio.buf = elf_phdata;
    fdio.size = size;
    retval = FSManager_Read(&fdio);
    if (retval)
        goto out_close;

    eppnt = elf_phdata;
    for (i=0; i<interp_elf_ex->e_phnum; i++, eppnt++) {
        if (eppnt->p_type == PT_LOAD) {
            int elf_type = MAP_PRIVATE | MAP_DENYWRITE;
            int elf_prot = 0;
            unsigned long vaddr = 0;
            unsigned long k, map_addr;

            if (eppnt->p_flags & PF_R) elf_prot =  PROT_READ;
            if (eppnt->p_flags & PF_W) elf_prot |= PROT_WRITE;
            if (eppnt->p_flags & PF_X) elf_prot |= PROT_EXEC;
            vaddr = eppnt->p_vaddr;
            if (interp_elf_ex->e_type == ET_EXEC || load_addr_set)
                elf_type |= MAP_FIXED;

            map_addr = elf_map(interpreter, load_addr + vaddr, eppnt, elf_prot, elf_type);
            if (map_addr == 0xffffffff)
                goto out_close;

            if (!load_addr_set && interp_elf_ex->e_type == ET_DYN) {
                load_addr = map_addr - ELF_PAGESTART(vaddr);
                load_addr_set = 1;
            }

            /*
            * Find the end of the file mapping for this phdr, and keep
            * track of the largest address we see for this.
            */
            k = load_addr + eppnt->p_vaddr + eppnt->p_filesz;
            if (k > elf_bss)
                elf_bss = k;

            /*
            * Do the same thing for the memory mapping - between
            * elf_bss and last_bss is the bss section.
            */
            k = load_addr + eppnt->p_memsz + eppnt->p_vaddr;
            if (k > last_bss)
                last_bss = k;
        }
    }

    /* Now use mmap to map the library into memory. */

    /*
    * Now fill out the bss section.  First pad the last page up
    * to the page boundary, and then perform a mmap to make sure
    * that there are zero-mapped pages up to and including the
    * last bss page.
    */
    padzero(elf_bss);
    elf_bss = ELF_PAGESTART(elf_bss + ELF_MIN_ALIGN - 1);	/* What we have mapped so far */

    /* Map the last of the bss segment */
    if (last_bss > elf_bss) {
        void *tmp;

        tmp = MemManager_AllocUserPages((void*)elf_bss, last_bss - elf_bss, 0, MemManager_PROT_WRITE+MemManager_PROT_READ+MemManager_FIXED_ADDR, NULL);
        if (IS_PTR_ERROR(tmp))
            ASSERT_MSG_PRE(PTR_TO_ERROR(tmp));
    }

    *interp_load_addr = load_addr;
    error = ((unsigned long) interp_elf_ex->e_entry) + load_addr;

out_close:
    MemManager_Free(elf_phdata);
out:
    return error;
}

#define INTERPRETER_NONE 0
#define INTERPRETER_AOUT 1
#define INTERPRETER_ELF 2

int ExecLoader_LoadExec(UserSpaceManager_ExecInfo *exec, TaskManager_Regs *regs) {
    FSManager_Handle *interpreter = NULL; /* to shut gcc up */
    unsigned long load_addr = 0, load_bias = 0;
    int load_addr_set = 0;
    char * elf_interpreter = NULL;
    unsigned int interpreter_type = INTERPRETER_NONE;
    unsigned long error;
    struct elf_phdr * elf_ppnt, *elf_phdata;
    unsigned long elf_bss, k, elf_brk;
    int retval, i, ret;
    unsigned int size;
    unsigned long elf_entry, interp_load_addr = 0;
    unsigned long start_code, end_code, start_data, end_data;
    struct elfhdr elf_ex;
    struct elfhdr interp_elf_ex;
    FSManager_FDIO fdio;
    char *newStack;

    /* Get the exec-header */
    elf_ex = *((struct elfhdr *) exec->fileBuf);

    /* Now read in all of the header information */
    retval = E_NO_MEMORY;
    if (elf_ex.e_phentsize != sizeof(struct elf_phdr))
        goto out;
    if (elf_ex.e_phnum > 65536U / sizeof(struct elf_phdr))
        goto out;
    size = elf_ex.e_phnum * sizeof(struct elf_phdr);
    elf_phdata = (struct elf_phdr *) MemManager_Alloc(size, 0);
    if (IS_PTR_ERROR(elf_phdata))
        goto out;

    retval = FSManager_Seek(exec->fd, elf_ex.e_phoff, FSManager_SEEK_START, NULL);
    if (retval)
        goto out_free_ph;

    fdio.fd = exec->fd;
    fdio.flags = 0;
    fdio.buf = elf_phdata;
    fdio.size = size;
    retval = FSManager_Read(&fdio);
    if (retval)
        goto out_free_ph;

    elf_ppnt = elf_phdata;
    elf_bss = 0;
    elf_brk = 0;

    start_code = ~0UL;
    end_code = 0;
    start_data = 0;
    end_data = 0;

    for (i = 0; i < elf_ex.e_phnum; i++) {
        if (elf_ppnt->p_type == PT_INTERP) {
            /* This is the program interpreter used for
            * shared libraries - for now assume that this
            * is an a.out format binary
            */

            retval = E_NO_MEMORY;
            if (elf_ppnt->p_filesz > PATH_MAX)
                goto out_free_file;
            elf_interpreter = (char *) MemManager_Alloc(elf_ppnt->p_filesz, 0);
            if (IS_PTR_ERROR(elf_interpreter))
                goto out_free_file;

            retval = FSManager_Seek(exec->fd, elf_ppnt->p_offset, FSManager_SEEK_START, NULL);
            if (retval)
                goto out_free_interp;
    
            fdio.buf = elf_interpreter;
            fdio.size = elf_ppnt->p_filesz;
            retval = FSManager_Read(&fdio);
            if (retval)
                goto out_free_interp;

            retval = FSManager_Open(elf_interpreter, FSManager_READ, &interpreter);
            if (retval)
                goto out_free_interp;
            retval = FSManager_Seek(interpreter, 0, FSManager_SEEK_START, NULL);
            if (retval)
                goto out_free_dentry;
            fdio.fd = interpreter;
            fdio.buf = exec->fileBuf;
            fdio.size = sizeof(exec->fileBuf);
            retval = FSManager_Read(&fdio);
            if (retval)
                goto out_free_dentry;

            /* Get the exec headers */
            interp_elf_ex = *((struct elfhdr *) exec->fileBuf);
            break;
        }
        elf_ppnt++;
    }
    
    /* Some simple consistency checks for the interpreter */
    if (elf_interpreter) {
        interpreter_type = INTERPRETER_ELF;

        if (MemMan_Compare(interp_elf_ex.e_ident, ELFMAG, SELFMAG))
            interpreter_type &= ~INTERPRETER_ELF;

        retval = E_UNSUPPORTED;
        if (!interpreter_type)
            goto out_free_dentry;
    }
    
    elf_entry = (unsigned long) elf_ex.e_entry;

    /* alloc stack space */
    newStack = MemManager_AllocUserPages((void*)0, exec->stackSize, 4096, MemManager_PROT_WRITE+MemManager_PROT_READ, NULL);
    if (IS_PTR_ERROR(newStack))
        goto out_free_dentry;
    retval = UserSpaceManager_MapStackPages(exec, newStack, exec->stackSize);
    if (retval)
        goto out_free_dentry;

    /* Now we do a little grungy work by mmaping the ELF image into
    the correct location in memory.  At this point, we assume that
    the image should be loaded at fixed address, not at a variable
    address. */

    for(i = 0, elf_ppnt = elf_phdata; i < elf_ex.e_phnum; i++, elf_ppnt++) {
        int elf_prot = 0, elf_flags;
        unsigned long vaddr;

        if (elf_ppnt->p_type != PT_LOAD)
            continue;

        if (unlikely (elf_brk > elf_bss)) {
            unsigned long nbyte;

            /* There was a PT_LOAD segment with p_memsz > p_filesz
            before this one. Map anonymous pages, if needed,
            and clear the area.  */
            set_brk (elf_bss + load_bias, elf_brk + load_bias);
            nbyte = ELF_PAGEOFFSET(elf_bss);
            if (nbyte) {
                nbyte = ELF_MIN_ALIGN - nbyte;
                if (nbyte > elf_brk - elf_bss)
                    nbyte = elf_brk - elf_bss;
                clear_user((void *) elf_bss + load_bias, nbyte);
            }
        }

        if (elf_ppnt->p_flags & PF_R) elf_prot |= PROT_READ;
        if (elf_ppnt->p_flags & PF_W) elf_prot |= PROT_WRITE;
        if (elf_ppnt->p_flags & PF_X) elf_prot |= PROT_EXEC;

        elf_flags = MAP_PRIVATE|MAP_DENYWRITE|MAP_EXECUTABLE;

        vaddr = elf_ppnt->p_vaddr;
        if (elf_ex.e_type == ET_EXEC || load_addr_set) {
            elf_flags |= MAP_FIXED;
        } else if (elf_ex.e_type == ET_DYN) {
            /* Try and get dynamic programs out of the way of the default mmap
            base, as well as whatever program they might try to exec.  This
                is because the brk will follow the loader, and is not movable.  */
            load_bias = ELF_PAGESTART(ELF_ET_DYN_BASE - vaddr);
        }

        error = elf_map(exec->fd, load_bias + vaddr, elf_ppnt, elf_prot, elf_flags);
        if (error == 0xffffffff)
            continue;

        if (!load_addr_set) {
            load_addr_set = 1;
            load_addr = (elf_ppnt->p_vaddr - elf_ppnt->p_offset);
            if (elf_ex.e_type == ET_DYN) {
                load_bias += error -
                            ELF_PAGESTART(load_bias + vaddr);
                load_addr += load_bias;
            }
        }
        k = elf_ppnt->p_vaddr;
        if (k < start_code) start_code = k;
        if (start_data < k) start_data = k;

        k = elf_ppnt->p_vaddr + elf_ppnt->p_filesz;

        if (k > elf_bss)
            elf_bss = k;
        if ((elf_ppnt->p_flags & PF_X) && end_code <  k)
            end_code = k;
        if (end_data < k)
            end_data = k;
        k = elf_ppnt->p_vaddr + elf_ppnt->p_memsz;
        if (k > elf_brk)
            elf_brk = k;
    }

    elf_entry += load_bias;
    elf_bss += load_bias;
    elf_brk += load_bias;
    start_code += load_bias;
    end_code += load_bias;
    start_data += load_bias;
    end_data += load_bias;
    
    if (elf_interpreter) {
        elf_entry = load_elf_interp(&interp_elf_ex, interpreter, &interp_load_addr);

        ret = FSManager_Close(interpreter);
        ASSERT_ERROR(ret);
        ret = MemManager_Free(elf_interpreter);
        ASSERT_ERROR(ret);

        if (elf_entry == 0xffffffff) {
            DEBUG_MSG(("Unable to load interpreter\n"));
            MemManager_Free(elf_phdata);
            return E_IO_ERROR;
        }
    }

    ret = MemManager_Free(elf_phdata);
    ASSERT_ERROR(ret);

    exec->stackTop = (unsigned long) create_elf_tables(exec, (char *)exec->stackTop, exec->argc, exec->envc,
            &elf_ex, load_addr, load_bias, interp_load_addr, 1);

    exec->startBrk = ALIGN_DATA_UP(elf_brk, 0x1000);

    /* Calling set_brk effectively mmaps the pages that we need
    * for the bss and break sections
    */
    set_brk(elf_bss, elf_brk);

    padzero(elf_bss);

#ifdef ELF_PLAT_INIT
    /*
    * The ABI may specify that certain registers be set up in special
    * ways (on i386 %edx is the address of a DT_FINI function, for
    * example.  This macro performs whatever initialization to
    * the regs structure is required.
    */
    ELF_PLAT_INIT(regs);
#endif

    /* start all registers */
    regs->eax = 0;
    regs->ebx = 0;
    regs->ecx = 0;
    regs->edx = 0;
    regs->esi = 0;
    regs->edi = 0;
    regs->ebp = 0;
    regs->esp = exec->stackTop;
    regs->eip = elf_entry;
    regs->eflags = 0x00000202;
    regs->ds = DPL3DataSegment;
    regs->es = DPL3DataSegment;
    regs->ss = DPL3DataSegment;
    regs->cs = DPL3CodeSegment;

    /* set this task as a Linux program */
    retval = ExecDomain_TaskCreated(NULL);
    if (retval)
        DEBUG_MSG_PRE(retval);
out:
    return retval;

    /* error cleanup */
out_free_dentry:
    ret = FSManager_Close(interpreter);
    ASSERT_ERROR(ret);
out_free_interp:
    if (elf_interpreter) {
        ret = MemManager_Free(elf_interpreter);
        ASSERT_ERROR(ret);
    }
out_free_file:
out_free_ph:
    ret = MemManager_Free(elf_phdata);
    ASSERT_ERROR(ret);
    goto out;
}

int Phase1Start(SysModManager_Module *modID) {
    int ret;

    /* start all used modules */
    if ((ret = SysModManager_StartModule((unsigned int)FSManager_Open)) ||
        (ret = SysModManager_StartModule((unsigned int)ExecDomain_TaskCreated)) ||
        (ret = SysModManager_StartModule((unsigned int)MemManager_Alloc)) ||
        (ret = SysModManager_StartModule((unsigned int)MemMan_Compare))) {
        DEBUG_MSG_PRE(ret);
        return 1;
    }
    return 0;
}
