/*
*****************************************
* This file implements the ExecDomain interface for Linux binaries
*
* Version: 0.1.1
* Author: Luiz Henrique Shigunov
* E-mail: luizshigunov@users.sourceforge.net
* WEB: http://www.sourceforge.net/projects/modulos
* Last change: 18/06/2005
*
* See the interface specification to know what each function does.
*
* Copyright (C) 2003-2005, Luiz Henrique Shigunov
*
* This module is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2.1 of the License, or (at your option) any later version.
*
* This module is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this module; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*****************************************
*/

#include <stddef.h>
#include "common.h"
#include <debug.h>
#include <errormanager.h>
#include <intermanager.h>
#include <intmanager_i386.h>
#include <memman.h>
#include <memmanager.h>
#include <memmanager_i386.h>
#include <sysdef.h>
#include <taskmanager.h>
#include <textvideo.h>
#include <userspacemanager.h>

int SysCall(IntManager_i386_Regs*);

SysModManager_Module *ID;

int ArrayBound(IntManager_i386_ExRegs *regs) {
    TaskInfo *taskInfo;
    int ret;

    /* task running a linux program? */
    taskInfo = TaskManager_GetTaskInfo(NULL, ID);
    if (IS_PTR_ERROR(taskInfo))
        return E_UNHANDLED_INT;

    DEBUG_MSG(("Caught array bound exceeded exception.\n At = 0x%x\n", regs->EIP));
    ret = UserSpaceManager_Exit(NULL, UserSpaceManager_TASK_DUMPED, UserSpaceManager_TASK_DUMPED_BOUND);
    ASSERT_ERROR(ret);
    return E_HANDLED_INT;
}

int DivideByZero(IntManager_i386_ExRegs *regs) {
    TaskInfo *taskInfo;
    int ret;

    /* task running a linux program? */
    taskInfo = TaskManager_GetTaskInfo(NULL, ID);
    if (IS_PTR_ERROR(taskInfo))
        return E_UNHANDLED_INT;

    DEBUG_MSG(("Caught divide by zero exception.\n At = 0x%x\n", regs->EIP));
    ret = UserSpaceManager_Exit(NULL, UserSpaceManager_TASK_DUMPED, UserSpaceManager_TASK_DUMPED_DIV0);
    ASSERT_ERROR(ret);
    return E_HANDLED_INT;
}

int InvalidInstruction(IntManager_i386_ExRegs *regs) {
    TaskInfo *taskInfo;
    int ret;

    /* task running a linux program? */
    taskInfo = TaskManager_GetTaskInfo(NULL, ID);
    if (IS_PTR_ERROR(taskInfo))
        return E_UNHANDLED_INT;

    DEBUG_MSG(("Caught invalid instruction exception.\n At = 0x%x\n", regs->EIP));
    ret = UserSpaceManager_Exit(NULL, UserSpaceManager_TASK_DUMPED, UserSpaceManager_TASK_DUMPED_ILINST);
    ASSERT_ERROR(ret);
    return E_HANDLED_INT;
}

int Overflow(IntManager_i386_ExRegs *regs) {
    TaskInfo *taskInfo;
    int ret;

    /* task running a linux program? */
    taskInfo = TaskManager_GetTaskInfo(NULL, ID);
    if (IS_PTR_ERROR(taskInfo))
        return E_UNHANDLED_INT;

    DEBUG_MSG(("Caught overflow exception.\n At = 0x%x\n", regs->EIP));
    ret = UserSpaceManager_Exit(NULL, UserSpaceManager_TASK_DUMPED, UserSpaceManager_TASK_DUMPED_INTO);
    ASSERT_ERROR(ret);
    return E_HANDLED_INT;
}

int SegmentationFaultUser(unsigned int where, IntManager_i386_ExRegs *regs) {
    TaskInfo *taskInfo;
    int ret;

    /* task running a linux program? */
    taskInfo = TaskManager_GetTaskInfo(NULL, ID);
    if (IS_PTR_ERROR(taskInfo))
        return E_UNHANDLED_INT;

    DEBUG_MSG(("Caught segmentation fault exception.\nAt = 0x%x where = 0x%x prop = 0x%x\n", regs->EIP, where, regs->ErrorCode));
    /* MemManager_Catch says its destroyable */
    TaskManager_UndestroyableThread();
    ret = UserSpaceManager_Exit(NULL, UserSpaceManager_TASK_DUMPED, UserSpaceManager_TASK_DUMPED_MEMREF);
    ASSERT_ERROR(ret);
    return E_HANDLED_INT;
}

int sys_NotImplemented(int stack) {
    SysModManager_DebugMsg("sys_xxx function not implemented: %d\n", *(&stack+9));
    return -38;
}

int TaskCreated(TaskManager_Task *task) {
    TaskInfo *taskInfo;
    InterManager_OutputInfo *outInfo;
    char outBuf[256];
    int error, ret;
    void *voidRet;

    taskInfo = MemManager_Alloc(sizeof(TaskInfo), 0);
    if (IS_PTR_ERROR(taskInfo))
        return PTR_TO_ERROR(taskInfo);
    taskInfo->files = MemManager_Alloc(NUM_FILES * sizeof(FileInfo), 0);
    if (IS_PTR_ERROR(taskInfo->files)) {
        error = PTR_TO_ERROR(taskInfo->files);
        goto errorFreeTaskInfo;
    }
    MemMan_Set(taskInfo->files, 0, NUM_FILES * sizeof(FileInfo));
    voidRet = TaskManager_SetTaskInfo(task, ID, taskInfo, TaskManager_SET_INFO_KEEP_DATA);
    if (IS_PTR_ERROR(voidRet)) {
        error = PTR_TO_ERROR(voidRet);
        goto errorFreeFiles;
    }
    /* setup stdout and stderr */
    outInfo = (InterManager_OutputInfo*)outBuf;
    error = InterManager_GetOutputInfo(outInfo, sizeof(outBuf));
    if (error)
        goto errorDelTaskInfo;
    taskInfo->imp = SysModManager_GetImpID(outInfo->names, outInfo->names+outInfo->interSize);
    if (IS_PTR_ERROR(taskInfo->imp))
        goto errorDelTaskInfo;
    taskInfo->TVWrite = SysModManager_GetFunction(taskInfo->imp, TextVideo__WRITE, 0, NULL);
    if (IS_PTR_ERROR(taskInfo->TVWrite))
        goto errorPutImp;

    /* set stdin, stdout and stderr */
    taskInfo->files[0].type = FILE_TYPE_INTERMANAGER_IN;
    taskInfo->files[1].type = FILE_TYPE_INTERMANAGER_OUT;
    taskInfo->files[2].type = FILE_TYPE_INTERMANAGER_OUT;
    return error;

errorPutImp:
    ret = SysModManager_PutImpID(taskInfo->imp);
    ASSERT_ERROR(ret);
errorDelTaskInfo:
    taskInfo = TaskManager_DelTaskInfo(task, ID);
    ASSERT_ERROR_PTR(taskInfo);
errorFreeFiles:
    ret = MemManager_Free(taskInfo->files);
    ASSERT_ERROR(ret);
errorFreeTaskInfo:
    ret = MemManager_Free(taskInfo);
    ASSERT_ERROR(ret);
    return error;
}

void UserSpaceManagerE_TaskDestroyed(TaskManager_Task *task, UserSpaceManager_ExitInfo *exitInfo) {
    TaskInfo *taskInfo;
    int ret;

    /* task running a linux program? */
    taskInfo = TaskManager_DelTaskInfo(task, ID);
    if (IS_PTR_ERROR(taskInfo))
        return;
    ret = MemManager_Free(taskInfo->files);
    ASSERT_ERROR(ret);
    ret = SysModManager_PutImpID(taskInfo->imp);
    ASSERT_ERROR(ret);
    ret = MemManager_Free(taskInfo);
    ASSERT_ERROR(ret);
    return;
}

int Phase1Start(SysModManager_Module *modID) {
    int ret;

    ID = modID;

    /* start all used modules */
    if ((ret = SysModManager_StartModule((unsigned int)IntManager_i386_Catch)) ||
        (ret = SysModManager_StartModule((unsigned int)MemManager_i386_Catch)) ||
        (ret = SysModManager_StartModule((unsigned int)TaskManager_GetTaskInfo))) {
        DEBUG_MSG_PRE(ret);
        return 1;
    }
    
    /* catch int that call system functions */
    if ((ret = IntManager_i386_Catch(modID, 0x80, SysCall, 0))) {
        DEBUG_MSG_PRE(ret);
        goto errorReturn;
    }

    /* register exception handlers */
    if ((ret = MemManager_i386_Catch(NULL, NULL, SegmentationFaultUser, MemManager_i386_EX_USER_MODE))) {
        DEBUG_MSG_PRE(ret);
        goto errorFreeSysCallInt;
    }
    if ((ret = IntManager_i386_CatchException(0x00, DivideByZero, (void*)0, (void*)0xdfffffff))) {
        DEBUG_MSG_PRE(ret);
        goto errorFreeSegEx;
    }
    if ((ret = IntManager_i386_CatchException(0x04, Overflow, (void*)0, (void*)0xdfffffff))) {
        DEBUG_MSG_PRE(ret);
        goto errorFreeDivEx;
    }
    if ((ret = IntManager_i386_CatchException(0x05, ArrayBound, (void*)0, (void*)0xdfffffff))) {
        DEBUG_MSG_PRE(ret);
        goto errorFreeOverflowEx;
    }
    if ((ret = IntManager_i386_CatchException(0x06, InvalidInstruction, (void*)0, (void*)0xdfffffff))) {
        DEBUG_MSG_PRE(ret);
        goto errorFreeArrayEx;
    }
    return 0;
    
    ret = IntManager_i386_FreeException(0x06, InvalidInstruction, (void*)0);
    ASSERT_ERROR(ret);
errorFreeArrayEx:
    ret = IntManager_i386_FreeException(0x05, ArrayBound, (void*)0);
    ASSERT_ERROR(ret);
errorFreeOverflowEx:
    ret = IntManager_i386_FreeException(0x04, Overflow, (void*)0);
    ASSERT_ERROR(ret);
errorFreeDivEx:
    ret = IntManager_i386_FreeException(0x00, DivideByZero, (void*)0);
    ASSERT_ERROR(ret);
errorFreeSegEx:
    ret = MemManager_i386_FreeException(SegmentationFaultUser, (void*)0, 0);
    ASSERT_ERROR(ret);
errorFreeSysCallInt:
    ret = IntManager_i386_Free(modID, 0x80, 0);
    ASSERT_ERROR(ret);
errorReturn:
    return 1;
}
