/*
*****************************************
* This file implements the sys_xxx functions from Linux kernel
*
* Version: 0.1.1
* Author: Luiz Henrique Shigunov
* E-mail: luizshigunov@users.sourceforge.net
* WEB: http://www.sourceforge.net/projects/modulos
* Last change: 18/06/2005
*
* See the interface specification to know what each function does.
*
* Copyright (C) 2003-2005, Luiz Henrique Shigunov
*
* This module is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2.1 of the License, or (at your option) any later version.
*
* This module is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this module; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*****************************************
*/

/* some code used from the Linux kernel 2.4 */

#include <stddef.h>
#include "common.h"
#include "errno.h"

#include <debug.h>
#include <errormanager.h>
#include <memmanager.h>
#include <sysmodmanager.h>
#include <userspacemanager.h>

struct new_utsname {
    char sysname[65];
    char nodename[65];
    char release[65];
    char version[65];
    char machine[65];
    char domainname[65];
};

struct iovec {
    void *iov_base;
    int iov_len;
};

#define UIO_FASTIOV     8
#define UIO_MAXIOV      1024

/* for now. Where should this be? */
struct new_utsname system_utsname = {
"Linux",
"(none)",
"2.4.18-mod",
"#1 Thu May 22 18:41:00 EDT 2003",
"unknown",
"(none)"
};

extern SysModManager_Module *ID;

unsigned long sys_brk(unsigned long brk) {
    return (unsigned long)UserSpaceManager_Brk((void*)brk);
}

void sys_exit(int error) {
    UserSpaceManager_Exit(NULL, UserSpaceManager_TASK_EXITED, error & 0xff);
}

long sys_newuname(struct new_utsname *name) {
    if (UserSpaceManager_CopyToUser(name, &system_utsname, sizeof(*name)))
        return -EFAULT;
    return 0;
}

long sys_ni_syscall(void) {
    return -ENOSYS;
}

int TranslateError(int error) {
    switch (error) {
    case E_FAULT:
        return EFAULT;
    default:
        return EINVAL;
    }
}

int DoRead(TaskInfo *taskInfo, FileInfo *file, void *buf, int count) {
    return -EINVAL;
}

int DoWrite(TaskInfo *taskInfo, FileInfo *file, void *buf, int count) {
    int error;

    switch (file->type) {
    case FILE_TYPE_INTERMANAGER_OUT:
        error = taskInfo->TVWrite(buf, count);
        if (!error)
            return count;
        return -TranslateError(error);
    default:
        return -EINVAL;
    }
}

#define DO_READ 1
#define DO_WRITE 2

int DoReadvWritev(int type, TaskInfo *taskInfo, FileInfo *file, const struct iovec * vector, unsigned long count) {
    int tot_len;
    struct iovec iovstack[UIO_FASTIOV];
    struct iovec *iov=iovstack;
    int ret, i;

    /*
    * First get the "struct iovec" from user memory and
    * verify all the pointers
    */
    ret = 0;
    if (!count)
        goto out_nofree;
    ret = -EINVAL;
    if (count > UIO_MAXIOV)
        goto out_nofree;
    if (count > UIO_FASTIOV) {
        ret = -ENOMEM;
        iov = MemManager_Alloc(count*sizeof(struct iovec), 0);
        if (IS_PTR_ERROR(iov))
            goto out_nofree;
    }
    ret = -EFAULT;
    if (UserSpaceManager_CopyFromUser(iov, vector, count*sizeof(*vector)))
        goto out;

    /*
    * Single unix specification:
    * We should -EINVAL if an element length is not >= 0 and fitting an ssize_t
    * The total length is fitting an ssize_t
    *
    * Be careful here because iov_len is a size_t not an ssize_t
    */

    tot_len = 0;
    ret = -EINVAL;
    for (i = 0 ; i < count ; i++) {
        int tmp = tot_len;
        int len = (int) iov[i].iov_len;
        if (len < 0)    /* size_t not fitting an ssize_t .. */
            goto out;
        tot_len += len;
        if (tot_len < tmp) /* maths overflow on the ssize_t */
            goto out;
    }

    ret = 0;
    vector = iov;
    while (count > 0) {
        void * base;
        unsigned int len;
        int nr;

        base = vector->iov_base;
        len = vector->iov_len;
        vector++;
        count--;

        if (type == DO_WRITE)
            nr = DoWrite(taskInfo, file, base, len);
        else
            nr = DoRead(taskInfo, file, base, len);

        if (nr < 0) {
            if (!ret) ret = nr;
                break;
        }
        ret += nr;
        if (nr != len)
            break;
    }

out:
    if (iov != iovstack) {
        int tmpRet;

        tmpRet = MemManager_Free(iov);
        ASSERT_ERROR(tmpRet);
    }

out_nofree:
    return ret;
}

int sys_writev(unsigned long fd, const struct iovec * vector, unsigned long count) {
    TaskInfo *taskInfo;

    if (fd >= NUM_FILES)
        return -EBADF;

    taskInfo = TaskManager_GetTaskInfo(NULL, ID);
    if (IS_PTR_ERROR(taskInfo)) {
        DEBUG_MSG_PRE(0);
        return -EINVAL;
    }

    if (!taskInfo->files[fd].type)
        return -EBADF;

    return DoReadvWritev(DO_WRITE, taskInfo, &taskInfo->files[fd], vector, count);
}
