/*
*****************************************
* This file implements the RExecModFile interface
*
* Version: 0.2.9
* Author: Luiz Henrique Shigunov
* E-mail: luizshigunov@users.sourceforge.net
* WEB: http://www.sourceforge.net/projects/modulos
* Last change: 01/07/2005
*
* See the interface specification to know what each function does.
*
* Copyright (C) 2000-2005, Luiz Henrique Shigunov
*
* This module is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2.1 of the License, or (at your option) any later version.
*
* This module is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this module; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*****************************************
*/

#include <stddef.h>
#include <debug.h>
#include <errormanager.h>
#include <fsmanager.h>
#include <memman.h>
#include <memmanager.h>
#include <rexecmodfile.h>
#include <sysmodmanager.h>

/* EM04 */
#define HEADER_SIGN 0x34304d45

struct _Header {
    char md5[16];
    char sign[4];
    unsigned int stackSize;
    unsigned int code;
    unsigned int codeSize;
    unsigned int roData;
    unsigned int roDataSize;
    unsigned int data;
    unsigned int dataSize;
    unsigned int uDataSize;
    unsigned int usedFuncs;
    unsigned int usedFuncsSize;
    unsigned int usedFuncReloc;
    unsigned int usedFuncRelocSize;
    unsigned int strings;
    unsigned short int stringsSize;
    unsigned short int comment;
} __attribute__ ((packed));

typedef struct _Header Header;

struct _UsedFunc {
    unsigned short interface;
    unsigned short imp;
    unsigned int funcProp;
} __attribute__ ((packed));

typedef struct _UsedFunc UsedFunc;

struct _UsedFuncReloc {
    unsigned int offset;
    unsigned int propInd;
} __attribute__ ((packed));

typedef struct _UsedFuncReloc UsedFuncReloc;

struct _RExecModFile_File {
    Header header;
    char *strings;
    FSManager_Handle *fd;
    UsedFunc *usedFuncs;
    UsedFuncReloc *usedFuncRelocs;
    int prop;
};

int Close(RExecModFile_File *file) {
    int ret;

#ifdef DEBUG_ASSERT
    if (!file) {
        ASSERT_MSG_PRE(0);
        return E_BAD_VALUE;
    }
#endif
    if (!(file->prop & RExecModFile_DO_NOT_CLOSE)) {
        ret = FSManager_Close(file->fd);
        ASSERT_ERROR(ret);
    }
    ret = MemManager_Free(file->strings);
    ASSERT_ERROR(ret);
    if (file->usedFuncs) {
        ret = MemManager_Free(file->usedFuncs);
        ASSERT_ERROR(ret);
    }
    if (file->usedFuncRelocs) {
        ret = MemManager_Free(file->usedFuncRelocs);
        ASSERT_ERROR(ret);
    }
    ret = MemManager_Free(file);
    ASSERT_ERROR(ret);
    return E_OK;
}

int CheckFormat(void *buf, unsigned int size) {
    Header *header;
    
#ifdef DEBUG_ASSERT
    if (!buf) {
        ASSERT_MSG_PRE(0);
        return E_BAD_VALUE;
    }
#endif
    header = (Header*)buf;
    if (size < 20 || *(unsigned int*)header->sign != HEADER_SIGN)
        return E_BAD_TYPE;
    return E_OK;
}

int GetModuleInfo(RExecModFile_File *file, RExecModFile_ModuleInfo *info) {
#ifdef DEBUG_ASSERT
    if (!file) {
        ASSERT_MSG_PRE(0);
        return E_BAD_VALUE;
    }
#endif
    MemMan_Copy(info->md5, file->header.md5, sizeof(info->md5));
    info->comment = &file->strings[file->header.comment];
    info->codeSize = file->header.codeSize;
    info->dataSize = file->header.dataSize + file->header.uDataSize;
    info->roDataSize = file->header.roDataSize;
    info->stackSize = file->header.stackSize;
    return E_OK;
}

int GetUsedFunc(RExecModFile_File *file, unsigned int n, RExecModFile_UsedFuncInfo *info) {
#ifdef DEBUG_ASSERT
    if (!file) {
        ASSERT_MSG_PRE(0);
        return E_BAD_VALUE;
    }
#endif
    if (n >= (file->header.usedFuncsSize/sizeof(UsedFunc)))
        return E_BAD_INDEX;
    if (!file->usedFuncs) {
        FSManager_FDIO io;
        file->usedFuncs = MemManager_Alloc(file->header.usedFuncsSize, 0);
        if (IS_PTR_ERROR(file->usedFuncs))
            return PTR_TO_ERROR(file->usedFuncs);
        io.fd = file->fd;
        io.flags = 0;
        io.buf = (char*)file->usedFuncs;
        io.size = file->header.usedFuncsSize;
        if (FSManager_Seek(file->fd, file->header.usedFuncs, FSManager_SEEK_START, NULL) || FSManager_Read(&io))
            return E_IO_ERROR;
    }
    info->interface = &file->strings[file->usedFuncs[n].interface];
    info->imp = NULL;
    if (file->usedFuncs[n].imp)
        info->imp = &file->strings[file->usedFuncs[n].imp];
    info->function = file->usedFuncs[n].funcProp & 0xffffff;
    info->prop = file->usedFuncs[n].funcProp >> 24;
    return E_OK;
}

int GetUsedFuncReloc(RExecModFile_File *file, unsigned int n, RExecModFile_UsedFuncRelocInfo *info) {
    unsigned int ind;

#ifdef DEBUG_ASSERT
    if (!file) {
        ASSERT_MSG_PRE(0);
        return E_BAD_VALUE;
    }
#endif
    if (n >= (file->header.usedFuncRelocSize/sizeof(UsedFuncReloc)))
        return E_BAD_INDEX;
    if (!file->usedFuncRelocs) {
        FSManager_FDIO io;
        file->usedFuncRelocs = MemManager_Alloc(file->header.usedFuncRelocSize, 0);
        if (IS_PTR_ERROR(file->usedFuncRelocs))
            return PTR_TO_ERROR(file->usedFuncRelocs);
        io.fd = file->fd;
        io.flags = 0;
        io.buf = (char*)file->usedFuncRelocs;
        io.size = file->header.usedFuncRelocSize;
        if (FSManager_Seek(file->fd, file->header.usedFuncReloc, FSManager_SEEK_START, NULL) || FSManager_Read(&io))
            return E_IO_ERROR;
        if (!file->usedFuncs) {
            file->usedFuncs = MemManager_Alloc(file->header.usedFuncsSize, 0);
            if (IS_PTR_ERROR(file->usedFuncs))
                return PTR_TO_ERROR(file->usedFuncs);
            io.buf = (char*)file->usedFuncs;
            io.size = file->header.usedFuncsSize;
            if (FSManager_Seek(file->fd, file->header.usedFuncs, FSManager_SEEK_START, NULL) || FSManager_Read(&io))
                return E_IO_ERROR;
        }
    }
    info->offset = file->usedFuncRelocs[n].offset;
    info->prop = (char)file->usedFuncRelocs[n].propInd;
    ind = file->usedFuncRelocs[n].propInd >> 8;
    info->function.interface = &file->strings[file->usedFuncs[ind].interface];
    info->function.imp = NULL;
    if (file->usedFuncs[ind].imp)
        info->function.imp = &file->strings[file->usedFuncs[ind].imp];
    info->function.function = file->usedFuncs[ind].funcProp & 0xffffff;
    info->function.prop = file->usedFuncs[ind].funcProp >> 24;
    return E_OK;
}

int LoadCode(RExecModFile_File *file, unsigned int start, unsigned int size, void *where) {
    int error;
    FSManager_FDIO io;

    error = E_BAD_VALUE;
#ifdef DEBUG_ASSERT
    if (!file) {
        ASSERT_MSG_PRE(0);
        goto end;
    }
#endif
    if (start || start >= file->header.codeSize || (start+size) > file->header.codeSize)
        goto end;
    io.fd = file->fd;
    io.flags = 0;
    io.buf = (char*)where;
    io.size = size;
    error = E_IO_ERROR;
    if (FSManager_Seek(file->fd, file->header.code+start, FSManager_SEEK_START, NULL) || FSManager_Read(&io))
        goto end;
    error = E_OK;
end:
    return error;
}

int LoadData(RExecModFile_File *file, unsigned int start, unsigned int size, void *where) {
    int error;
    unsigned int sizeRead;

    error = E_BAD_VALUE;
#ifdef DEBUG_ASSERT
    if (!file) {
        ASSERT_MSG_PRE(0);
        goto end;
    }
#endif
    if (start || start >= (file->header.dataSize+file->header.uDataSize) || (start+size) > (file->header.dataSize+file->header.uDataSize))
        goto end;
    sizeRead = 0;
    if (file->header.dataSize && start < file->header.dataSize) {
        FSManager_FDIO io;
        sizeRead = size > file->header.dataSize-start ? file->header.dataSize-start : size;
        io.fd = file->fd;
        io.flags = 0;
        io.buf = (char*)where;
        io.size = sizeRead;
        error = E_IO_ERROR;
        if (FSManager_Seek(file->fd, file->header.data+start, FSManager_SEEK_START, NULL) || FSManager_Read(&io))
            goto end;
    }
    if (file->header.uDataSize && (start+size) > file->header.dataSize)
        MemMan_Set(where+sizeRead, 0, size-sizeRead);
    error = E_OK;
end:
    return error;
}

int LoadROData(RExecModFile_File *file, unsigned int start, unsigned int size, void *where) {
    int error;
    FSManager_FDIO io;

    error = E_BAD_VALUE;
#ifdef DEBUG_ASSERT
    if (!file) {
        ASSERT_MSG_PRE(0);
        goto end;
    }
#endif
    if (start || start >= file->header.roDataSize || (start+size) > file->header.roDataSize)
        goto end;
    io.fd = file->fd;
    io.flags = 0;
    io.buf = (char*)where;
    io.size = size;
    error = E_IO_ERROR;
    if (FSManager_Seek(file->fd, file->header.roData+start, FSManager_SEEK_START, NULL) || FSManager_Read(&io))
        goto end;
    error = E_OK;
end:
    return error;
}

int Open(const char *pathname, RExecModFile_File **file) {
    RExecModFile_File *tmpR;
    FSManager_Handle *tmpF;
    FSManager_FDIO io;
    int error, ret;

    error = FSManager_Open(pathname, FSManager_READ, &tmpF);
    if (error)
        goto errorReturn;
    tmpR = MemManager_Alloc(sizeof(RExecModFile_File), 0);
    if (IS_PTR_ERROR(tmpR)) {
        error = PTR_TO_ERROR(tmpR);
        goto errorCloseFile;
    }
    io.fd = tmpF;
    io.flags = 0;
    io.buf = &tmpR->header;
    io.size = sizeof(Header);
    error = E_IO_ERROR;
    if (FSManager_Read(&io))
        goto errorFreeFile;
    error = E_BAD_TYPE;
    if (*(unsigned int*)tmpR->header.sign != HEADER_SIGN)
        goto errorFreeFile;
    tmpR->strings = MemManager_Alloc(tmpR->header.stringsSize, 0);
    if (IS_PTR_ERROR(tmpR->strings)) {
        error = PTR_TO_ERROR(tmpR->strings);
        goto errorFreeFile;
    }
    error = E_IO_ERROR;
    if (FSManager_Seek(tmpF, tmpR->header.strings, FSManager_SEEK_START, NULL))
        goto errorFreeStrings;
    io.buf = tmpR->strings;
    io.size = tmpR->header.stringsSize;
    if (FSManager_Read(&io))
        goto errorFreeStrings;
    tmpR->fd = tmpF;
    tmpR->usedFuncs = NULL;
    tmpR->usedFuncRelocs = NULL;
    tmpR->prop = 0;
    *file = tmpR;
    return E_OK;

errorFreeStrings:
    ret = MemManager_Free(tmpR->strings);
    ASSERT_ERROR(ret);
errorFreeFile:
    ret = MemManager_Free(tmpR);
    ASSERT_ERROR(ret);
errorCloseFile:
    ret = FSManager_Close(tmpF);
    ASSERT_ERROR(ret);
errorReturn:
    return error;
}

int OpenFD(FSManager_Handle *fd, int prop, RExecModFile_File **file) {
    RExecModFile_File *tmpR;
    FSManager_FDIO io;
    int error, ret;

    tmpR = MemManager_Alloc(sizeof(RExecModFile_File), 0);
    if (IS_PTR_ERROR(tmpR)) {
        error = PTR_TO_ERROR(tmpR);
        goto errorReturn;
    }
    error = E_IO_ERROR;
    if (FSManager_Seek(fd, 0, FSManager_SEEK_START, NULL))
        goto errorFreeFile;
    io.fd = fd;
    io.flags = 0;
    io.buf = &tmpR->header;
    io.size = sizeof(Header);
    if (FSManager_Read(&io))
        goto errorFreeFile;
    error = E_BAD_TYPE;
    if (*(unsigned int*)tmpR->header.sign != HEADER_SIGN)
        goto errorFreeFile;
    tmpR->strings = MemManager_Alloc(tmpR->header.stringsSize, 0);
    if (IS_PTR_ERROR(tmpR->strings)) {
        error = PTR_TO_ERROR(tmpR->strings);
        goto errorFreeFile;
    }
    error = E_IO_ERROR;
    if (FSManager_Seek(fd, tmpR->header.strings, FSManager_SEEK_START, NULL))
        goto errorFreeStrings;
    io.buf = tmpR->strings;
    io.size = tmpR->header.stringsSize;
    if (FSManager_Read(&io))
        goto errorFreeStrings;
    tmpR->fd = fd;
    tmpR->usedFuncs = NULL;
    tmpR->usedFuncRelocs = NULL;
    tmpR->prop = prop;
    *file = tmpR;
    return E_OK;

errorFreeStrings:
    ret = MemManager_Free(tmpR->strings);
    ASSERT_ERROR(ret);
errorFreeFile:
    ret = MemManager_Free(tmpR);
    ASSERT_ERROR(ret);
errorReturn:
    return error;
}

int Phase1Start(SysModManager_Module *modID) {
    int ret;

    /* start all used modules */
    if ((ret = SysModManager_StartModule((unsigned int)FSManager_Open)) ||
        (ret = SysModManager_StartModule((unsigned int)MemManager_Alloc)) ||
        (ret = SysModManager_StartModule((unsigned int)MemMan_Set))) {
        DEBUG_MSG_PRE(ret);
        return 1;
    }
    return 0;
}
