/*
*****************************************
* This file implements the RLibModFile interface
*
* Version: 0.2.8
* Author: Luiz Henrique Shigunov
* E-mail: luizshigunov@users.sourceforge.net
* WEB: http://www.sourceforge.net/projects/modulos
* Last change: 18/06/2005
*
* See the interface specification to know what each function does.
*
* Copyright (C) 2000-2005, Luiz Henrique Shigunov
*
* This module is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2.1 of the License, or (at your option) any later version.
*
* This module is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this module; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*****************************************
*/

#include <stddef.h>
#include <debug.h>
#include <errormanager.h>
#include <fsmanager.h>
#include <memman.h>
#include <memmanager.h>
#include <rlibmodfile.h>
#include <sysmodmanager.h>

/* LM04 */
#define HEADER_SIGN 0x34304d4c

struct _Header {
    char md5[16];
    char sign[4];
    unsigned int code;
    unsigned int codeSize;
    unsigned int roData;
    unsigned int roDataSize;
    unsigned int data;
    unsigned int dataSize;
    unsigned int uDataSize;
    unsigned int usedFuncs;
    unsigned int usedFuncsSize;
    unsigned int usedFuncReloc;
    unsigned int usedFuncRelocSize;
    unsigned int interfaces;
    unsigned int interfacesSize;
    unsigned int roDataReloc;
    unsigned int roDataRelocSize;
    unsigned int dataReloc;
    unsigned int dataRelocSize;
    unsigned int codeReloc;
    unsigned int codeRelocSize;
    unsigned int strings;
    unsigned short int stringsSize;
    unsigned short int version;
    unsigned short int prop;
    unsigned short int comment;
    unsigned int startF;
    unsigned int shutdownF;
} __attribute__ ((packed));

typedef struct _Header Header;

struct _Function {
    unsigned int code;
    unsigned short prop;
} __attribute__ ((packed));

typedef struct _Function Function;

struct _Implementation {
    Function *functions;
    unsigned short name;
} __attribute__ ((packed));

typedef struct _Implementation Implementation;

struct _Interface {
    unsigned short name;
    unsigned short nFunctions;
    unsigned short nImp;
    Implementation imps[1];
} __attribute__ ((packed));

typedef struct _Interface Interface;

struct _UsedFunc {
    unsigned short interface;
    unsigned short imp;
    unsigned int funcProp;
} __attribute__ ((packed));

typedef struct _UsedFunc UsedFunc;

struct _UsedFuncReloc {
    unsigned int offset;
    unsigned int propInd;
} __attribute__ ((packed));

typedef struct _UsedFuncReloc UsedFuncReloc;

struct _RLibModFile_File {
    Header header;
    char *strings;
    FSManager_Handle *fd;
    Interface *interfaces;
    UsedFunc *usedFuncs;
    UsedFuncReloc *usedFuncRelocs;
};

int Close(RLibModFile_File *file) {
    int ret;

    ASSERT_RET(file, E_BAD_VALUE);
 
    ret = FSManager_Close(file->fd);
    ASSERT_ERROR(ret);
    ret = MemManager_Free(file->strings);
    ASSERT_ERROR(ret);
    ret = MemManager_Free(file->interfaces);
    ASSERT_ERROR(ret);
    if (file->usedFuncs) {
        ret = MemManager_Free(file->usedFuncs);
        ASSERT_ERROR(ret);
    }
    if (file->usedFuncRelocs) {
        ret = MemManager_Free(file->usedFuncRelocs);
        ASSERT_ERROR(ret);
    }
    ret = MemManager_Free(file);
    ASSERT_ERROR(ret);
    return E_OK;
}

int GetFuncsInfo(RLibModFile_File *file, unsigned int interface, unsigned int imp, RLibModFile_FunctionInfo *info) {
    Interface *currInt;
    int error, ret;

    ASSERT_RET(file, E_BAD_VALUE);
 
    error = E_BAD_INDEX;
    currInt = file->interfaces;
    for(; interface; interface--) {
        currInt = (Interface*)((char*)currInt+sizeof(Interface)-sizeof(Implementation)+sizeof(Implementation)*currInt->nImp);
        if ((unsigned int)currInt >= ((unsigned int)file->interfaces+file->header.interfacesSize))
            goto end;
    }
    if (imp < currInt->nImp) {
        Function *funcs;
        FSManager_FDIO io;
        unsigned int i;

        funcs = MemManager_Alloc(sizeof(Function)*currInt->nFunctions, 0);
        if (IS_PTR_ERROR(funcs)) {
            error = PTR_TO_ERROR(funcs);
            goto end;
        }
        error = E_IO_ERROR;
        if (FSManager_Seek(file->fd, (unsigned int)currInt->imps[imp].functions, FSManager_SEEK_START, NULL))
            goto errorFree;
        io.fd = file->fd;
        io.flags = 0;
        io.buf = (char*)funcs;
        io.size = currInt->nFunctions*sizeof(Function);
        if (FSManager_Read(&io))
            goto errorFree;
        error = E_OK;
        for (i=0; i < currInt->nFunctions; i++) {
            info[i].codeStart = funcs[i].code;
            info[i].prop = funcs[i].prop;
        }
errorFree:
        ret = MemManager_Free(funcs);
        ASSERT_ERROR(ret);
    }
end:
    return error;
}

int GetImpInfo(RLibModFile_File *file, unsigned int interface, unsigned int imp, RLibModFile_ImpInfo *info) {
    Interface *currInt;
    int error;

    ASSERT_RET(file, E_BAD_VALUE);
 
    error = E_BAD_INDEX;
    currInt = file->interfaces;
    for(; interface; interface--) {
        currInt = (Interface*)((char*)currInt+sizeof(Interface)-sizeof(Implementation)+sizeof(Implementation)*currInt->nImp);
        if ((unsigned int)currInt >= ((unsigned int)file->interfaces+file->header.interfacesSize))
            goto end;
    }
    if (imp < currInt->nImp) {
        info->name = &file->strings[currInt->imps[imp].name];
        error = E_OK;
    }
end:
    return error;
}

int GetInterInfo(RLibModFile_File *file, unsigned int interface, RLibModFile_InterInfo *info) {
    Interface *currInt;

    ASSERT_RET(file, E_BAD_VALUE);
 
    currInt = file->interfaces;
    for(; interface; interface--) {
        currInt = (Interface*)((char*)currInt+sizeof(Interface)-sizeof(Implementation)+sizeof(Implementation)*currInt->nImp);
        if ((unsigned int)currInt >= ((unsigned int)file->interfaces+file->header.interfacesSize))
            return E_BAD_INDEX;
    }
    info->name = &file->strings[currInt->name];
    info->nFunctions = currInt->nFunctions;
    return E_OK;
}

int GetModuleInfo(RLibModFile_File *file, RLibModFile_ModuleInfo *info) {

    ASSERT_RET(file, E_BAD_VALUE);
 
    MemMan_Copy(info->md5, file->header.md5, 16);
    info->comment = &file->strings[file->header.comment];
    info->start = (void*)file->header.startF;
    info->shutdown = (void*)file->header.shutdownF;
    info->codeSize = file->header.codeSize;
    info->roDataSize = file->header.roDataSize;
    info->dataSize = file->header.dataSize + file->header.uDataSize;
    info->version = file->header.version;
    info->prop = file->header.prop;
    return E_OK;
}

int GetUsedFunc(RLibModFile_File *file, unsigned int n, RLibModFile_UsedFuncInfo *info) {

    ASSERT_RET(file, E_BAD_VALUE);
 
    if (n >= (file->header.usedFuncsSize/sizeof(UsedFunc)))
        return E_BAD_INDEX;
    if (!file->usedFuncs) {
        FSManager_FDIO io;
        file->usedFuncs = MemManager_Alloc(file->header.usedFuncsSize, 0);
        if (IS_PTR_ERROR(file->usedFuncs))
            return PTR_TO_ERROR(file->usedFuncs);
        io.fd = file->fd;
        io.flags = 0;
        io.buf = (char*)file->usedFuncs;
        io.size = file->header.usedFuncsSize;
        if (FSManager_Seek(file->fd, file->header.usedFuncs, FSManager_SEEK_START, NULL) || FSManager_Read(&io))
            return E_IO_ERROR;
    }
    info->interface = &file->strings[file->usedFuncs[n].interface];
    info->imp = NULL;
    if (file->usedFuncs[n].imp)
        info->imp = &file->strings[file->usedFuncs[n].imp];
    info->function = file->usedFuncs[n].funcProp & 0xffffff;
    info->prop = file->usedFuncs[n].funcProp >> 24;
    return E_OK;
}

int GetUsedFuncReloc(RLibModFile_File *file, unsigned int n, RLibModFile_UsedFuncRelocInfo *info) {
    unsigned int ind;

    ASSERT_RET(file, E_BAD_VALUE);
 
    if (n >= (file->header.usedFuncRelocSize/sizeof(UsedFuncReloc)))
        return E_BAD_INDEX;
    if (!file->usedFuncRelocs) {
        FSManager_FDIO io;
        file->usedFuncRelocs = MemManager_Alloc(file->header.usedFuncRelocSize, 0);
        if (IS_PTR_ERROR(file->usedFuncRelocs))
            return PTR_TO_ERROR(file->usedFuncRelocs);
        io.fd = file->fd;
        io.flags = 0;
        io.buf = (char*)file->usedFuncRelocs;
        io.size = file->header.usedFuncRelocSize;
        if (FSManager_Seek(file->fd, file->header.usedFuncReloc, FSManager_SEEK_START, NULL) || FSManager_Read(&io))
            return E_IO_ERROR;
        if (!file->usedFuncs) {
            file->usedFuncs = MemManager_Alloc(file->header.usedFuncsSize, 0);
            if (IS_PTR_ERROR(file->usedFuncs))
                return PTR_TO_ERROR(file->usedFuncs);
            io.buf = (char*)file->usedFuncs;
            io.size = file->header.usedFuncsSize;
            if (FSManager_Seek(file->fd, file->header.usedFuncs, FSManager_SEEK_START, NULL) || FSManager_Read(&io))
                return E_IO_ERROR;
        }
    }
    info->offset = file->usedFuncRelocs[n].offset;
    info->prop = (char)file->usedFuncRelocs[n].propInd;
    ind = file->usedFuncRelocs[n].propInd >> 8;
    info->function.interface = &file->strings[file->usedFuncs[ind].interface];
    info->function.imp = NULL;
    if (file->usedFuncs[ind].imp)
        info->function.imp = &file->strings[file->usedFuncs[ind].imp];
    info->function.function = file->usedFuncs[ind].funcProp & 0xffffff;
    info->function.prop = file->usedFuncs[ind].funcProp >> 24;
    return E_OK;
}

int LoadCode(RLibModFile_File *file, unsigned int start, unsigned int size, void *where, void *codeStart, void *roDataStart, void *dataStart) {
    int error, ret;
    FSManager_FDIO io;

    ASSERT_RET(file, E_BAD_VALUE);
 
    error = E_BAD_VALUE;
    if (start || start >= file->header.codeSize || (start+size) > file->header.codeSize)
        goto end;
    io.fd = file->fd;
    io.flags = 0;
    io.buf = (char*)where;
    io.size = size;
    error = E_IO_ERROR;
    if (FSManager_Seek(file->fd, file->header.code+start, FSManager_SEEK_START, NULL) || FSManager_Read(&io))
        goto end;
    /* do relocations */
    error = E_OK;
    if (file->header.codeRelocSize) {
        char *reloc, *currR;
        unsigned int roDataRSize, dataRSize, codeRSize;

        reloc = MemManager_Alloc(file->header.codeRelocSize, 0);
        if (IS_PTR_ERROR(reloc)) {
            error = PTR_TO_ERROR(reloc);
            goto end;
        }
        io.buf = reloc;
        io.size = file->header.codeRelocSize;
        error = E_IO_ERROR;
        if (FSManager_Seek(file->fd, file->header.codeReloc, FSManager_SEEK_START, NULL) || FSManager_Read(&io))
            goto errorFree;
        currR = reloc;
        roDataRSize = (*(unsigned int*)currR)>>2;
        dataRSize = (*(unsigned int*)(currR+4))>>2;
        codeRSize = (*(unsigned int*)(currR+8))>>2;
        currR += 12;
        for(; roDataRSize; roDataRSize--, currR += 4)
            *(unsigned int*)((char*)where+*(unsigned int*)currR) = (unsigned int)(*(unsigned int*)((char*)where+*(unsigned int*)currR)+(char*)roDataStart);
        for(; dataRSize; dataRSize--, currR += 4)
            *(unsigned int*)((char*)where+*(unsigned int*)currR) = (unsigned int)(*(unsigned int*)((char*)where+*(unsigned int*)currR)+(char*)dataStart);
        for(; codeRSize; codeRSize--, currR += 4)
            *(unsigned int*)((char*)where+*(unsigned int*)currR) = (unsigned int)(*(unsigned int*)((char*)where+*(unsigned int*)currR)+(char*)codeStart);
        error = E_OK;
errorFree:
        ret = MemManager_Free(reloc);
        ASSERT_ERROR(ret);
    }
end:
    return error;
}

int LoadData(RLibModFile_File *file, unsigned int start, unsigned int size, void *where, void *codeStart, void *roDataStart, void *dataStart) {
    int error, ret;
    unsigned int sizeRead;
    FSManager_FDIO io;

    ASSERT_RET(file, E_BAD_VALUE);
 
    error = E_BAD_VALUE;
    if (start || start >= (file->header.dataSize+file->header.uDataSize) || (start+size) > (file->header.dataSize+file->header.uDataSize))
        goto end;
    sizeRead = 0;
    io.fd = file->fd;
    io.flags = 0;
    if (file->header.dataSize && start < file->header.dataSize) {
        sizeRead = size > file->header.dataSize-start ? file->header.dataSize-start : size;
        io.buf = (char*)where;
        io.size = sizeRead;
        error = E_IO_ERROR;
        if (FSManager_Seek(file->fd, file->header.data+start, FSManager_SEEK_START, NULL) || FSManager_Read(&io))
            goto end;
    }
    if (file->header.uDataSize && (start+size) > file->header.dataSize)
        MemMan_Set(where+sizeRead, 0, size-sizeRead);
    /* do relocations */
    error = E_OK;
    if (file->header.dataRelocSize) {
        char *reloc, *currR;
        unsigned int roDataRSize, dataRSize, codeRSize;

        reloc = MemManager_Alloc(file->header.dataRelocSize, 0);
        if (IS_PTR_ERROR(reloc)) {
            error = PTR_TO_ERROR(reloc);
            goto end;
        }
        io.buf = reloc;
        io.size = file->header.dataRelocSize;
        error = E_IO_ERROR;
        if (FSManager_Seek(file->fd, file->header.dataReloc, FSManager_SEEK_START, NULL) || FSManager_Read(&io))
            goto errorFree;
        currR = reloc;
        roDataRSize = (*(unsigned int*)currR)>>2;
        dataRSize = (*(unsigned int*)(currR+4))>>2;
        codeRSize = (*(unsigned int*)(currR+8))>>2;
        currR += 12;
        for(; roDataRSize; roDataRSize--, currR += 4)
            *(unsigned int*)((char*)where+*(unsigned int*)currR) = (unsigned int)(*(unsigned int*)((char*)where+*(unsigned int*)currR)+(char*)roDataStart);
        for(; dataRSize; dataRSize--, currR += 4)
            *(unsigned int*)((char*)where+*(unsigned int*)currR) = (unsigned int)(*(unsigned int*)((char*)where+*(unsigned int*)currR)+(char*)dataStart);
        for(; codeRSize; codeRSize--, currR += 4)
            *(unsigned int*)((char*)where+*(unsigned int*)currR) = (unsigned int)(*(unsigned int*)((char*)where+*(unsigned int*)currR)+(char*)codeStart);
        error = E_OK;
errorFree:
        ret = MemManager_Free(reloc);
        ASSERT_ERROR(ret);
    }
end:
    return error;
}

int LoadROData(RLibModFile_File *file, unsigned int start, unsigned int size, void *where, void *codeStart, void *roDataStart, void *dataStart) {
    int error, ret;
    FSManager_FDIO io;

    ASSERT_RET(file, E_BAD_VALUE);
 
    error = E_BAD_VALUE;
    if (start || start >= file->header.roDataSize || (start+size) > file->header.roDataSize)
        goto end;
    io.fd = file->fd;
    io.flags = 0;
    io.buf = (char*)where;
    io.size = size;
    error = E_IO_ERROR;
    if (FSManager_Seek(file->fd, file->header.roData+start, FSManager_SEEK_START, NULL) || FSManager_Read(&io))
        goto end;
    /* do relocations */
    error = E_OK;
    if (file->header.roDataRelocSize) {
        char *reloc, *currR;
        unsigned int roDataRSize, dataRSize, codeRSize;

        reloc = MemManager_Alloc(file->header.roDataRelocSize, 0);
        if (IS_PTR_ERROR(reloc)) {
            error = PTR_TO_ERROR(reloc);
            goto end;
        }
        io.buf = reloc;
        io.size = file->header.roDataRelocSize;
        error = E_IO_ERROR;
        if (FSManager_Seek(file->fd, file->header.roDataReloc, FSManager_SEEK_START, NULL) || FSManager_Read(&io))
            goto errorFree;
        currR = reloc;
        roDataRSize = (*(unsigned int*)currR)>>2;
        dataRSize = (*(unsigned int*)(currR+4))>>2;
        codeRSize = (*(unsigned int*)(currR+8))>>2;
        currR += 12;
        for(; roDataRSize; roDataRSize--, currR += 4)
            *(unsigned int*)((char*)where+*(unsigned int*)currR) = (unsigned int)(*(unsigned int*)((char*)where+*(unsigned int*)currR)+(char*)roDataStart);
        for(; dataRSize; dataRSize--, currR += 4)
            *(unsigned int*)((char*)where+*(unsigned int*)currR) = (unsigned int)(*(unsigned int*)((char*)where+*(unsigned int*)currR)+(char*)dataStart);
        for(; codeRSize; codeRSize--, currR += 4)
            *(unsigned int*)((char*)where+*(unsigned int*)currR) = (unsigned int)(*(unsigned int*)((char*)where+*(unsigned int*)currR)+(char*)codeStart);
        error = E_OK;
errorFree:
        ret = MemManager_Free(reloc);
        ASSERT_ERROR(ret);
    }
end:
    return error;
}

int Open(const char *pathname, RLibModFile_File **file) {
    RLibModFile_File *tmpR;
    FSManager_Handle *tmpF;
    FSManager_FDIO io;
    int error, ret;

    error = FSManager_Open(pathname, FSManager_READ, &tmpF);
    if (error)
        goto errorEnd;

    tmpR = MemManager_Alloc(sizeof(RLibModFile_File), 0);
    if (IS_PTR_ERROR(tmpR)) {
        error = PTR_TO_ERROR(tmpR);
        goto errorAllocH;
    }
    io.fd = tmpF;
    io.flags = 0;
    io.buf = &tmpR->header;
    io.size = sizeof(Header);
    error = E_IO_ERROR;
    if (FSManager_Read(&io))
        goto errorRead;
    error = E_BAD_TYPE;
    if (*(unsigned int*)tmpR->header.sign != HEADER_SIGN)
        goto errorRead;
    tmpR->strings = MemManager_Alloc(tmpR->header.stringsSize, 0);
    if (IS_PTR_ERROR(tmpR->strings)) {
        error = PTR_TO_ERROR(tmpR->strings);
        goto errorRead;
    }
    tmpR->interfaces = MemManager_Alloc(tmpR->header.interfacesSize, 0);
    if (IS_PTR_ERROR(tmpR->interfaces)) {
        error = PTR_TO_ERROR(tmpR->interfaces);
        goto errorAllocInt;
    }
    error = E_IO_ERROR;
    if (FSManager_Seek(tmpF, tmpR->header.strings, FSManager_SEEK_START, NULL))
        goto errorSeek;
    io.buf = tmpR->strings;
    io.size = tmpR->header.stringsSize;
    if (FSManager_Read(&io))
        goto errorSeek;
    if (FSManager_Seek(tmpF, tmpR->header.interfaces, FSManager_SEEK_START, NULL))
        goto errorSeek;
    io.buf = tmpR->interfaces;
    io.size = tmpR->header.interfacesSize;
    if (FSManager_Read(&io))
        goto errorSeek;
    tmpR->fd = tmpF;
    tmpR->usedFuncs = NULL;
    tmpR->usedFuncRelocs = NULL;
    *file = tmpR;
    return E_OK;

errorSeek:
    ret = MemManager_Free(tmpR->interfaces);
    ASSERT_ERROR(ret);
errorAllocInt:
    ret = MemManager_Free(tmpR->strings);
    ASSERT_ERROR(ret);
errorRead:
    ret = MemManager_Free(tmpR);
    ASSERT_ERROR(ret);
errorAllocH:
    ret = FSManager_Close(tmpF);
    ASSERT_ERROR(ret);
errorEnd:
    return error;
}

int Phase1Start(SysModManager_Module *modID) {
    int ret;

    /* start all used modules */
    if ((ret = SysModManager_StartModule((unsigned int)FSManager_Open)) ||
        (ret = SysModManager_StartModule((unsigned int)MemManager_Alloc)) ||
        (ret = SysModManager_StartModule((unsigned int)MemMan_Set))) {
        DEBUG_MSG_PRE(ret);
        return 1;
    }
    return 0;
}
