/*
*****************************************
* This file implements the DiskPartition interface for msdos partitions
*
* Version: 0.1.0
* Author: Luiz Henrique Shigunov
* E-mail: luizshigunov@users.sourceforge.net
* WEB: http://www.sourceforge.net/projects/modulos
* Last change: 14/03/2004
*
* See the interface specification to know what each function does.
*
* Copyright (C) 2004, Luiz Henrique Shigunov
*
* This module is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2.1 of the License, or (at your option) any later version.
*
* This module is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this module; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*****************************************
*/

#include <stddef.h>
#include <blockdev.h>
#include <debug.h>
#include <diskpartition.h>
#include <errormanager.h>
#include <sysdef.h>
#include <sysmodmanager.h>

typedef struct  __attribute__((__packed__)) {
    unsigned char boot;        /* 0x80 - bootable */
    unsigned char head;        /* starting head */
    unsigned char sector;      /* starting sector */
    unsigned char cyl;         /* starting cylinder */
    unsigned char id;          /* partition id */
    unsigned char endHead;     /* end head */
    unsigned char endSector;   /* end sector */
    unsigned char endCyl;      /* end cylinder */
    unsigned int startSector;  /* starting sector counting from 0 */
    unsigned int nSectors;     /* number of sectors in partition */
} MsdosPartition;

SysModManager_Module *ID;

__inline__ static int ReadExtended(unsigned int extendedStart, unsigned int extendedSize, BlockDev_IO *rq, int (*readF)(BlockDev_IO*),
                                   DiskPartition_Partition **descPart, DiskPartition_Description *desc, unsigned int descSize) {
    unsigned char buf[512];
    MsdosPartition *part;
    int i, ret, loopCount;
    unsigned int currStart, currSize;

    /* number of links followed without finding a data partition */
    loopCount = 0;
    currStart = extendedStart;
    currSize = extendedSize;
    while (1) {
        if (++loopCount > 100)
            return E_OK;
        /* read extended partition info */
        rq->block = currStart;
        rq->n = 1;
        rq->buffer = buf;
        ret = readF(rq);
        if (ret) {
            DEBUG_MSG(("partmsdos: read extended error: %d\n", ret));
            return ret;
        }
        /* valid MBR? */
        if (buf[510] != 0x55 || buf[511] != 0xaa) {
            DEBUG_MSG(("partmsdos: invalid extended sign: %02x%02x\n", (unsigned int)buf[510], (unsigned int)buf[511]));
            return E_OK;
        }
        /*
         * Usually, the first entry is the real data partition,
         * the 2nd entry is the next extended partition, or empty,
         * and the 3rd and 4th entries are unused.
         * However, DRDOS sometimes has the extended partition as
         * the first entry (when the data partition is empty),
         * and OS/2 seems to use all four entries.
         */

        /*
         * First process the data partition(s)
         */
        part = (MsdosPartition*)(buf+0x1be);
        for (i = 0; i < 4; i++, part++) {
            unsigned int start;
            
            /* check buffer size */
            if ((((char*)*descPart)-(char*)desc+sizeof(DiskPartition_Partition)) > descSize)
                return E_BUFFER_TOO_SMALL;
            /* used? Or is an extended partition? */
            if (!part->nSectors || part->id == 0x05 || part->id == 0x0f || part->id == 0x85)
                continue;
                
            /* Check the 3rd and 4th entries - these sometimes contain random garbage */
            start = currStart + part->startSector;
            if (i > 1) {
                if ((part->startSector + part->nSectors) > currSize)
                    continue;
                if (start < extendedStart)
                    continue;
                if ((start + part->nSectors) > (extendedStart + extendedSize))
                    continue;
            }

            /* add it */
            (*descPart)->startBlock = start;
            (*descPart)->nBlocks = part->nSectors;
            (*descPart)->id = part->id;
            (*descPart)->flags = part->boot ? DiskPartition_F_BOOT : 0;
            (*descPart)->flags |= DiskPartition_F_LOGICAL;
            desc->nParts++;
            (*descPart)++;
            loopCount = 0;
        }
        /*
         * Next, process the (first) extended partition, if present.
         * It should be a link to the next logical partition.
         */
        part -= 4;
        for (i = 0; i < 4; i++, part++)
            if (part->nSectors && (part->id == 0x05 || part->id == 0x0f || part->id == 0x85))
                break;
        if (i == 4)
            break;

        currStart = extendedStart + part->startSector;
        currSize = part->nSectors;
    }
    return E_OK;
}

int Read(SysModManager_Imp *imp, unsigned int unit, DiskPartition_Description *desc, unsigned int descSize, unsigned int partSize) {
    int (*readF)(BlockDev_IO*);
    MsdosPartition *part;
    DiskPartition_Partition *descPart;
    BlockDev_IO rq;
    unsigned char buf[512];
    int i, ret;

    ASSERT_RET(imp, E_BUG);
    ASSERT_RET(desc, E_BUG);
    
    if (descSize < sizeof(DiskPartition_Description))
        return E_BUFFER_TOO_SMALL;
    
    readF = SysModManager_GetFunction(imp, BlockDev__READ, 0, NULL);
    if (IS_PTR_ERROR(readF)) {
        DEBUG_MSG(("partmsdos: could not get blockdev functions: %d\n", PTR_TO_ERROR(readF)));
        return E_IO_ERROR;
    }
    /* read MBR */
    rq.unit = unit;
    rq.block = 0;
    rq.n = 1;
    rq.buffer = buf;
    ret = readF(&rq);
    if (ret) {
        DEBUG_MSG(("partmsdos: read error: %d\n", ret));
        return ret;
    }
    /* valid MBR? */
    if (buf[510] != 0x55 || buf[511] != 0xaa) {
        DEBUG_MSG(("partmsdos: invalid sign: %02x%02x\n", (unsigned int)buf[510], (unsigned int)buf[511]));
        return E_BAD_TYPE;
    }
    /* go through all partitions */
    descPart = desc->parts;
    desc->nParts = 0;
    part = (MsdosPartition*)(buf+0x1be);
    for (i = 0; i < 4; i++, part++) {
        /* check buffer size */
        if (((char*)descPart-(char*)desc+sizeof(DiskPartition_Partition)) > descSize)
            return E_BUFFER_TOO_SMALL;
        /* used? */
        if (!part->nSectors)
            continue;
        /* add it */
        descPart->startBlock = part->startSector;
        descPart->nBlocks = part->nSectors;
        descPart->id = part->id;
        descPart->flags = part->boot ? DiskPartition_F_BOOT : 0;
        desc->nParts++;
        descPart++;
        /* extended partition? (dos, win98 and linux extended partitions)*/
        if (part->id == 0x05 || part->id == 0x0f || part->id == 0x85) {
            (descPart-1)->flags |= DiskPartition_F_EXTENDED;
            DEBUG_MSG(("partmsdos: adding extended - start: %u  blocks: %u\n", part->startSector, part->nSectors));
            ret = ReadExtended(part->startSector, part->nSectors, &rq, readF, &descPart, desc, descSize);
            if (ret)
                return ret;
        }
    }
    return E_OK;
}

int Phase1Start(SysModManager_Module *modID) {
    ID = modID;
    return 0;
}

int Shutdown(void) {
    return 0;
}
