/*
*****************************************
* This file implements the Thread interface
*
* Version: 0.1.4
* Author: Luiz Henrique Shigunov
* E-mail: luizshigunov@users.sourceforge.net
* WEB: http://www.sourceforge.net/projects/modulos
* Last change: 06/06/2003
*
* See the interface specification to know what each function does.
*
* Copyright (C) 2000-2003, Luiz Henrique Shigunov
*
* This module is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2.1 of the License, or (at your option) any later version.
*
* This module is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this module; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*****************************************
*/

#include <stddef.h>
#include <errormanager.h>
#include <lib/mutex.h>
#include <taskmanager.h>
#include <memmanager.h>
#include <usermodmanager.h>

#define GET_THREAD_ID(x) __asm__ __volatile__ ("andl %%esp,%0;" : "=r" (x) : "0" (~(stackSize-1)))

/* number of thread specific data keys */
#define NUM_TSD 256

typedef struct _Thread {
    int thread;
    int (*function)(void*);
    void *arg;
    struct _Thread *next;
    struct _Thread *prev;
    void *tsds[NUM_TSD];
} Thread;

/* is a power of 2 */
static unsigned int stackSize;
static char keys[NUM_TSD];
static User_Mutex_Mutex keysMutex;
/* circular list of threads */
static Thread *threads;
static User_Mutex_Mutex threadsMutex;
static int started;

static int StartThread(void*);
int Fast_Lock(User_Mutex_Mutex*);
int Fast_Unlock(User_Mutex_Mutex*);
int Rec_Init(User_Mutex_Mutex*,int);

int Thread_AllocTSD(unsigned int *key, void *destructor) {
    unsigned int i;
    int ret;

    ret = E_NO_MEMORY;
    Fast_Lock(&keysMutex);
    for (i = 0; i < NUM_TSD; i++) {
        if (!keys[i]) {
            /* set key as in use. destructor is not used yet */
            keys[i] = 1;
            *key = i;
            ret = E_OK;
            break;
        }
    }
    Fast_Unlock(&keysMutex);
    return ret;
}

int Thread_Create(int (*function)(void*), void *arg, int prop, Thread **id) {
    char *stack;
    Thread *newThread;

    /* alloc stack space */
    stack = MemManager_UAllocUserPages((void*)0, stackSize, stackSize, MemManager_PROT_WRITE+MemManager_PROT_READ);
    if (IS_PTR_ERROR(stack))
        return PTR_TO_ERROR(stack);
    /* create a guard page */
    MemManager_USetUserPagesProt(stack+4096, 4096, MemManager_PROT_NONE);
    newThread = (Thread*)stack;
    stack += stackSize;
    /* start new thread structure */
    newThread->function = function;
    newThread->arg = arg;
    /* set args */
    stack -= 4;
    *(Thread**)stack = newThread;
    stack -= 4;
    /* create the new thread */
    newThread->thread = TaskManager_UCreateThread(stack, StartThread, prop);
    if (newThread->thread <= 0) {
        MemManager_UFreeUserPages((char*)newThread, stackSize);
        return E_NO_MEMORY;
    }
    /* insert in the threads list */
    Fast_Lock(&threadsMutex);
    if (!threads) {
        threads = newThread;
        newThread->next = newThread;
    }
    threads->next->prev = newThread;
    newThread->next = threads->next;
    threads->next = newThread;
    newThread->prev = threads;
    Fast_Unlock(&threadsMutex);
    *id = newThread;
    return E_OK;
}

int Thread_Destroy(Thread *id) {
    if (!id)
        GET_THREAD_ID(id);
    /* remove from threads list */
    Fast_Lock(&threadsMutex);
    if (id->next != id) {
        id->prev->next = id->next;
        id->next->prev = id->prev;
    } else
        id->next = NULL;
    if (threads == id)
        threads = id->next;
    Fast_Unlock(&threadsMutex);
    TaskManager_UDestroyThread(id->thread, (char*)id, stackSize);
    /* failed */
    /* reinsert in the threads list */
    Fast_Lock(&threadsMutex);
    if (!threads) {
        threads = id;
        id->next = id;
    }
    threads->next->prev = id;
    id->next = threads->next;
    threads->next = id;
    id->prev = threads;
    Fast_Unlock(&threadsMutex);
    return E_BAD_VALUE;
}

int Thread_FreeTSD(unsigned int key) {
    int ret;
    Thread *curr, *self;

    Fast_Lock(&keysMutex);
    ret = E_BAD_INDEX;
    if (key >= NUM_TSD || !keys[key])
        goto returnUnlock;
    keys[key] = 0;
    /* Set the value of the key to NULL in all threads, so that if the key is
       reallocated later, its associated values will be NULL in all threads. */
    GET_THREAD_ID(self);
    curr = self;
    do {
        curr->tsds[key] = NULL;
        curr = curr->next;
    } while (curr != self);
    ret = E_OK;
returnUnlock:
    Fast_Unlock(&keysMutex);
    return ret;
}

Thread *Thread_GetID(void) {
    Thread *id;

    GET_THREAD_ID(id);
    return id;
}

void *Thread_GetTSD(unsigned int key) {
    Thread *id;

    if (key >= NUM_TSD || !keys[key])
        return NULL;
    GET_THREAD_ID(id);
    return id->tsds[key];
}

int Thread_SetTSD(unsigned int key, void *value) {
    Thread *id;

    if (key >= NUM_TSD || !keys[key])
        return E_BAD_INDEX;
    GET_THREAD_ID(id);
    id->tsds[key] = value;
    return E_OK;
}

static int StartThread(void *arg) {
    Thread *thread;

    thread = (Thread*)arg;
    thread->function(thread->arg);
    return Thread_Destroy(NULL);
}

int Start(void) {
    Thread *id;

    if (started)
        return E_OK;
    started = 1;
    stackSize = UserModManager_UGetStackSize();
    Rec_Init(&keysMutex, 0);
    Rec_Init(&threadsMutex, 0);
    /* start first thread struct */
    GET_THREAD_ID(id);
    /* create a guard page */
    MemManager_USetUserPagesProt((char*)id+4096, 4096, MemManager_PROT_NONE);
    id->thread = TaskManager_UGetThreadID();
    return E_OK;
}
