/*
*****************************************
* This file implements the TextVideo and the OutputDev interfaces for a VGA adapter
*
* Version: 0.1.3
* Author: Luiz Henrique Shigunov
* E-mail: luizshigunov@users.sourceforge.net
* WEB: http://www.sourceforge.net/projects/modulos
* Last change: 18/06/2005
*
* See the interface specification to know what each function does.
*
* Copyright (C) 2000-2005, Luiz Henrique Shigunov
*
* This module is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2.1 of the License, or (at your option) any later version.
*
* This module is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this module; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*****************************************
*/

#include <stddef.h>
#include <debug.h>
#include <errormanager.h>
#include <ioportmanager.h>
#include <memman.h>
#include <memmanager_i386.h>
#include <outputdev.h>
#include <sysmodmanager.h>
#include <taskmanager.h>
#include <textvideo.h>
#include <userspacemanager.h>

#define PORT_REG 0x3d4
#define PORT_VAL 0x3d5

#define FONT_HEIGHT 16

typedef int VideoData;

SysModManager_Module *ID;
char *buffer;
char currentAttrib;
int cursorAttrib;
int cursorPos;
int cursorType;
TaskManager_Semaphore mutex = {1, NULL};

int Write(const char*,int);

static __inline__ void WriteWordVGAReg(unsigned char reg, unsigned int value) {
    unsigned int comp1, comp2;

    comp1 = reg + (value & 0xff00);
    comp2 = reg + 1 + ((value << 8) & 0xff00);
    __asm__ __volatile__ ("outw %%ax, %%dx" :: "a" (comp1), "d" (PORT_REG));
    __asm__ __volatile__ ("outw %%ax, %%dx" :: "a" (comp2), "d" (PORT_REG));
}

void Clear(unsigned int row, unsigned int col, unsigned int height, unsigned int width) {
    unsigned int start;

    if (row >= 25 || col >= 80 || (row+height) > 25 || (col+width) > 80) {
        DEBUG_MSG_PRE(0);
        return;
    }
    TaskManager_P(&mutex, 0, 0);
    start = (row * 80) + col;
    if (width == 80) {
        unsigned int count, tmp;
        count = height * 40;
        tmp = 0x200020;
        tmp |= ((unsigned int)currentAttrib << 24) | ((unsigned int)currentAttrib << 8);
        for (; count; count--, start++)
            ((unsigned int*)buffer)[start] = tmp;
    } else {
        for (; height; height--) {
            unsigned int count;
            for (count = width; count; count--) {
                buffer[start] = 0x20;
                buffer[start+1] = currentAttrib;
                start += 2;
            }
            /* go to next line */
            start += 160;
            start -= 2 * width;
        }
    }
    TaskManager_V(&mutex, 0);
}

unsigned int GetCurrentMode(void) {
    /* we only support one mode */
    return 0;
}

int GetModeDesc(unsigned int mode, TextVideo_ModeDesc *desc) {

    ASSERT_RET(desc, E_BUG);

    if (mode)
        return E_BAD_INDEX;
    /* we only support mode 25x80 */
    desc->rows = 25;
    desc->cols = 80;
    desc->flags = TextVideo_MODE_COLOR+TextVideo_MODE_REVERSE+TextVideo_MODE_BLINK+TextVideo_MODE_INTENSITY;
    return E_OK;
}

int GetStatus(TextVideo_Status *status) {

    ASSERT_RET(status, E_BUG);

    TaskManager_P(&mutex, 0, 0);
    status->row = cursorPos / 160;
    status->col = (cursorPos % 160) >> 1;
    status->attrib = cursorAttrib;
    status->cursorType = 0;
    TaskManager_V(&mutex, 0);
    return E_OK;
}

int Scroll(unsigned int top, unsigned int bottom, unsigned int lines, int dir) {
    unsigned int count, tmp, *cl;
    char *dst, *src;

    if (top+lines > bottom)
        lines = bottom - top;
    if (bottom >= 25 || top >= bottom || !lines) {
        DEBUG_MSG_PRE(0);
        return E_BAD_VALUE;
    }
    if (!dir) {
        dst = buffer + (top * 160);
        src = dst + (lines * 160);
    } else {
        src = buffer + (top * 160);
        dst = src + (lines * 160);
    }
    TaskManager_P(&mutex, 0, 0);
    MemMan_Move(dst, src, (bottom-top-lines+1)*160);
    count = lines * 40;
    tmp = 0x200020;
    tmp |= ((unsigned int)currentAttrib << 24) | ((unsigned int)currentAttrib << 8);
    cl = (unsigned int*)(buffer + ((bottom-lines+1) * 160));
    for (; count; count--, cl++)
        *cl = tmp;
    TaskManager_V(&mutex, 0);
    return E_OK;
}

void SetAttrib(int attrib) {
    char fcolor, bcolor;

    /* change red/blue brown/cyan colors */
    fcolor = attrib >> 24;
    switch (fcolor) {
        case 0x01:
            fcolor <<= 2;
            break;
        case 0x04:
            fcolor >>= 2;
            break;
        case 0x03:
            fcolor <<= 1;
            break;
        case 0x06:
            fcolor >>= 1;
            break;
    }
    bcolor = attrib >> 16;
    switch (bcolor) {
        case 0x01:
            bcolor <<= 2;
            break;
        case 0x04:
            bcolor >>= 2;
            break;
        case 0x03:
            bcolor <<= 1;
            break;
        case 0x06:
            bcolor >>= 1;
            break;
    }
    TaskManager_P(&mutex, 0, 0);
    /* blink */
    currentAttrib = (attrib & 0x08) << 4;
    /* intensity */
    currentAttrib |= (attrib << 2) & 0x08;
    /* reverse */
    if (attrib & 0x10)
        currentAttrib |= (bcolor & 0x07) | ((fcolor & 0x07) << 4);
    else
        currentAttrib |= (fcolor & 0x07) | ((bcolor & 0x07) << 4);
    cursorAttrib = attrib;
    TaskManager_V(&mutex, 0);
}

void SetCursorPos(unsigned int row, unsigned int col) {
    if (row >= 25 || col >= 80) {
        DEBUG_MSG_PRE(0);
        return;
    }
    TaskManager_P(&mutex, 0, 0);
    cursorPos = (row * 160) + (col << 1);
    WriteWordVGAReg(14, cursorPos/2);
    TaskManager_V(&mutex, 0);
}

void SetCursorType(int type) {
    int from, to, curStart, curEnd;

    if (cursorType == type)
        return;

    switch (type) {
        case TextVideo_CURSOR_INVISIBLE:
            from = 31;
            to = 30;
            break;
        case TextVideo_CURSOR_LOWER_THIRD:
            from = (FONT_HEIGHT*2)/3;
            to = FONT_HEIGHT - 2;
            break;
        case TextVideo_CURSOR_LOWER_HALF:
            from = FONT_HEIGHT / 2;
            to = FONT_HEIGHT - 2;
            break;
        case TextVideo_CURSOR_TWO_THIRDS:
            from = FONT_HEIGHT / 3;
            to = FONT_HEIGHT - 2;
            break;
        case TextVideo_CURSOR_BLOCK:
            from = 1;
            to = FONT_HEIGHT;
            break;
        case TextVideo_CURSOR_UNDERLINE:
        default:
            from = FONT_HEIGHT - 3;
            to = FONT_HEIGHT - 2;
            break;
    }
    TaskManager_P(&mutex, 0, 0);
    cursorType = type;
    /* cursor start */
    __asm__ __volatile__ ("outb %%al, %%dx" :: "a" (0x0a), "d" (PORT_REG));
    __asm__ __volatile__ ("inb %%dx, %%al" : "=a" (curStart): "d" (PORT_VAL));
    /* cursor end */
    __asm__ __volatile__ ("outb %%al, %%dx" :: "a" (0x0b), "d" (PORT_REG));
    __asm__ __volatile__ ("inb %%dx, %%al" : "=a" (curEnd): "d" (PORT_VAL));

    curStart = (curStart & 0xc0) | from;
    curEnd = (curEnd & 0xe0) | to;

    /* cursor start */
    __asm__ __volatile__ ("outb %%al, %%dx" :: "a" (0x0a), "d" (PORT_REG));
    __asm__ __volatile__ ("outb %%al, %%dx" :: "a" (curStart), "d" (PORT_VAL));
    /* cursor end */
    __asm__ __volatile__ ("outb %%al, %%dx" :: "a" (0x0b), "d" (PORT_REG));
    __asm__ __volatile__ ("outb %%al, %%dx" :: "a" (curEnd), "d" (PORT_VAL));
    TaskManager_V(&mutex, 0);
    return;
}

int SetMode(unsigned int mode) {
    int i;

    /* we only support mode 0 - 25x80 */
    if (mode) {
        DEBUG_MSG_PRE(0);
        return E_BAD_VALUE;
    }
    TaskManager_P(&mutex, 0, 0);
    /* clear screen and set cursor position to 0,0 */
    for(i=0; i<1000; i++)
        ((int*)buffer)[i] = 0x07200720;
    cursorPos = 0;
    WriteWordVGAReg(14, 0);
    cursorAttrib = 0x07000011;
    currentAttrib = 0x07;
    TaskManager_V(&mutex, 0);
    SetCursorType(TextVideo_CURSOR_DEFAULT);
    return E_OK;
}

int Shutdown(void) {
    int ret;

    ret = MemManager_i386_FreePages(buffer, 1, MemManager_i386_DONT_FREE_PHYS);
    ASSERT_ERROR(ret);
    ret = IOPortManager_Free(ID, PORT_REG);
    ASSERT_ERROR(ret);
    return 0;
}

int Phase0Start(SysModManager_Module *modID) {
    int ret;

    ID = modID;
    if ((ret = SysModManager_StartModule((unsigned int)MemMan_Move)) ||
        (ret = SysModManager_StartModule((unsigned int)TaskManager_P)) ||
        (ret = SysModManager_StartModule((unsigned int)IOPortManager_Alloc)) ||
        (ret = SysModManager_StartModule((unsigned int)MemManager_i386_MapPhysPages))) {
        DEBUG_MSG_PRE(ret);
        goto returnEnd;
    }
    buffer = MemManager_i386_MapPhysPages(NULL, 0xb8000, 1, MemManager_i386_RW_PAGE+MemManager_i386_WT_PAGE+MemManager_i386_ALLOC_PAGES);
    if (IS_PTR_ERROR(buffer)) {
        DEBUG_MSG_PRE(PTR_TO_ERROR(buffer));
        goto returnEnd;
    }
    if ((ret = IOPortManager_Alloc(modID, PORT_REG, PORT_VAL, 0))) {
        DEBUG_MSG_PRE(ret);
        goto returnUnmap;
    }
    /* set default settings */
    SetMode(0);
    return E_OK;

returnUnmap:
    ret = MemManager_i386_FreePages(buffer, 1, MemManager_i386_DONT_FREE_PHYS);
    ASSERT_ERROR(ret);
returnEnd:
    return 1;
}

int Phase1Start(SysModManager_Module *modID) {
    int ret;

    if ((ret = SysModManager_StartModule((unsigned int)MemMan_Move)) ||
        (ret = SysModManager_StartModule((unsigned int)TaskManager_P)) ||
        (ret = SysModManager_StartModule((unsigned int)IOPortManager_Alloc)) ||
        (ret = SysModManager_StartModule((unsigned int)MemManager_i386_MapPhysPages))) {
        DEBUG_MSG_PRE(ret);
        return 1;
    }
    return E_OK;
}

int Write(const char *text, int size) {
    int error;
    int count, bytesLeft;
    const char *textPtr, *currUserPtr;
    static char sysBuf[1024];

    TaskManager_P(&mutex, 0, 0);
    bytesLeft = size;
    textPtr = text;
    currUserPtr = text;
    while (bytesLeft) {
        if (text < (char*)0xe0000000) {
            /* user space */
            if (bytesLeft > sizeof(sysBuf))
                count = sizeof(sysBuf);
            else
                count = bytesLeft;
            error = UserSpaceManager_CopyFromUser(sysBuf, currUserPtr, count);
            if (error)
                goto errorV;
            bytesLeft -= count;
            currUserPtr += count;
            textPtr = sysBuf;
        } else {
            count = bytesLeft;
            bytesLeft = 0;
        }
        while (count--) {
            if (*textPtr == '\xd' || *textPtr == '\xa') {
                cursorPos = (cursorPos/160)*160;
                if (*textPtr == '\xa')
                    cursorPos += 160;
            } else {
                buffer[cursorPos] = *textPtr;
                buffer[cursorPos+1] = currentAttrib;
                cursorPos += 2;
            }
            textPtr++;
            /* end of screen? */
            if (cursorPos == 4000) {
                unsigned int i, tmp;
                tmp = 0x200020;
                tmp |= ((unsigned int)currentAttrib << 24) | ((unsigned int)currentAttrib << 8);
                MemMan_Move(buffer, buffer+160, 4000-160);
                for(i=(4000-160)/4; i<1000; i++)
                    ((unsigned int*)buffer)[i] = tmp;
                cursorPos = 4000-160;
            }
        }
    }

    error = E_OK;

errorV:
    WriteWordVGAReg(14, cursorPos/2);
    TaskManager_V(&mutex, 0);
    return error;
}

int OutputDev_GetImpInfo(OutputDev_ImpInfo *info) {

    ASSERT_RET(info, E_BUG);

    info->interface = "TextVideo";
    info->implementation = "VGA";
    return E_OK;
}

int OutputDev_Register(TaskManager_Task *task) {
    void *voidRet;

    voidRet = TaskManager_SetTaskInfo(task, ID, (void*)0x1234, 0);
    if (IS_PTR_ERROR(voidRet))
        return PTR_TO_ERROR(voidRet);
    return E_OK;
}

int UClear(unsigned int row, unsigned int col, unsigned int height, unsigned int width) {
    VideoData *info;

    info = TaskManager_GetTaskInfo(NULL, ID);
    if (IS_PTR_ERROR(info))
        return E_NOT_REGISTERED;
    Clear(row, col, height, width);
    return E_OK;
}

unsigned int UGetCurrentMode(void) {
    VideoData *info;

    info = TaskManager_GetTaskInfo(NULL, ID);
    if (IS_PTR_ERROR(info))
        return E_NOT_REGISTERED;
    return GetCurrentMode();
}

int UGetModeDesc(unsigned int mode, TextVideo_ModeDesc *desc) {
    VideoData *info;
    TextVideo_ModeDesc localDesc;
    int error;

    info = TaskManager_GetTaskInfo(NULL, ID);
    if (IS_PTR_ERROR(info))
        return E_NOT_REGISTERED;
    error = GetModeDesc(mode, &localDesc);
    if (!error) {
        error = UserSpaceManager_CopyToUser(desc, &localDesc, sizeof(localDesc));
    }
    return error;
}

int UGetStatus(TextVideo_Status *status) {
    VideoData *info;
    TextVideo_Status localStatus;
    int error;

    info = TaskManager_GetTaskInfo(NULL, ID);
    if (IS_PTR_ERROR(info))
        return E_NOT_REGISTERED;
    error = GetStatus(&localStatus);
    if (!error) {
        error = UserSpaceManager_CopyToUser(status, &localStatus, sizeof(localStatus));
    }
    return error;
}

int UScroll(unsigned int top, unsigned int bottom, unsigned int lines, int dir) {
    VideoData *info;

    info = TaskManager_GetTaskInfo(NULL, ID);
    if (IS_PTR_ERROR(info))
        return E_NOT_REGISTERED;
    return Scroll(top, bottom, lines, dir);
}

int USetAttrib(int attrib) {
    VideoData *info;

    info = TaskManager_GetTaskInfo(NULL, ID);
    if (IS_PTR_ERROR(info))
        return E_NOT_REGISTERED;
    SetAttrib(attrib);
    return E_OK;
}

int USetCursorPos(unsigned int row, unsigned int col) {
    VideoData *info;

    info = TaskManager_GetTaskInfo(NULL, ID);
    if (IS_PTR_ERROR(info))
        return E_NOT_REGISTERED;
    SetCursorPos(row, col);
    return E_OK;
}

int USetCursorType(int type) {
    VideoData *info;

    info = TaskManager_GetTaskInfo(NULL, ID);
    if (IS_PTR_ERROR(info))
        return E_NOT_REGISTERED;
    SetCursorType(type);
    return E_OK;
}

int USetMode(unsigned int mode) {
    VideoData *info;

    info = TaskManager_GetTaskInfo(NULL, ID);
    if (IS_PTR_ERROR(info))
        return E_NOT_REGISTERED;
    return SetMode(mode);
}

int UWrite(const char *text, int size) {
    VideoData *info;

    /* must point to user space */
    if (text >= (char*)0xe0000000)
        return E_FAULT;
    info = TaskManager_GetTaskInfo(NULL, ID);
    if (IS_PTR_ERROR(info))
        return E_NOT_REGISTERED;
    return Write(text, size);
}
